#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "glesp.h"

#define PR      fprintf
#define ER      stderr

#define MOD(a,p)   ((a) - (p) * (int)((a)/(p)))

extern void a_lm_(int *, int *, int *, int *, int *, double *, double *,
	 double *, int *,
	 int *, double *, double *, double *, int *, double *,
	 double *, double *, double *, int *, int *, int *);
extern void alm_dipole_(int *, int *, int *, int *, int *, double *,
	 double *, double *, double *);
extern void alm_lgndr_m_(int *, int *, int *, int *, int *, int *, int *,
	 int *, double *,
	 int *, double *, double *, double *, double *, double *, double *);
extern void alm_lgndr_l_(int *, int *, int *, int *, int *, int *, int *,
	 double *,
	 int *, double *, double *, double *, double *, double *, double *);
extern void start_clm_(int *, int *, double *, double *);
extern void lgndr_0_(int *, double *, double *);
extern void lgndr_l_(int *, int *, double *, double *, double *);
extern void lgndr_m_(int *, int *, double *, double *, double *, double *);
extern void alm_phi_  (int *, int *, int *, double *, int *, int *,
	     double *, double *, double *, double *, double *);
extern void alm_fast3_(int *, int *, int *, int *, double *, int *,
	     double *, double *, double *, double *, double *);

/*
 *
 *
 *  input parameters: x(Nx) - coordinates on
 *  the sphere, f(Nx,Nph(ix)) - signal.
 *  Nph - number of cells vs. x
 *  0< phi < phi2 for the direct transformation
 *  output: - f(l,m+1) - real a_lm, f(l,Lmx+m+2) - imag a_lm,
 *  a((l,-m)=(-1)^m a(l,m)*
 *-------------------------------------------------------------------- *
 *
 *
 *      DIMENSION x(Nmx),pw(Nmx),f(Nmx*Nmp),Nph(Nmx)
 *      DIMENSION apm((Jmx+1)*Nmx),bpm((Jmx+1)*Nmx)
 *      DIMENSION wsave(2*Nmp+15),cf(Nmp)
 */
extern int max_double_();
void a_lm_(Lmn,Lmx,Mmn,Mmx,Nx,Vnyq,eps,phi2, Nmx,Nmp, x,pw,f,Nph,
	       wsave,cf,apm,bpm, K_fast,Ksym,Kml)
int *Lmn, *Lmx, *Mmn, *Mmx, *Nx;
double *Vnyq, *eps, *phi2;
int *Nmx, *Nmp;
double *x, *pw, *f;
int *Nph;
double *wsave, *cf, *apm, *bpm;
int *K_fast, *Ksym, *Kml;
{
	register int m, mm, i, j, ii;
//      double suma, sumb;
	int i_dip_flag = 0;
	_lmin_gl = max_double_();
	_enorm_gl = pow(10., (double)_lmin_gl);   /* 10^Lmin */

	for(mm=m=0; m<= *Lmx; m++, mm+=*Nmx) {
	   for(i=0; i< *Nmx; i++) apm[mm+i]=bpm[mm+i]=0.;
	}

	if(*Kml == 100 && *Lmx <= 0) { /* also calculation of dipole */
	    *Kml = -1;
	    *Lmx = - *Lmx;
	    i_dip_flag = 1;
	}

	if(*Kml == 0 && *Lmx < 0) { /* only dipole calculation */
	     *Lmx = - *Lmx;
	     i_dip_flag = 1;
	}
       /*
	*   loop over phi - SUM_phi[f*exp(-m*phi)
	*   for rectangular map x1< x< x2, 0< phi < phi2
	*/
	if(*K_fast < 0)
	    alm_phi_(Nmx,Nmp, Lmx,Vnyq,Nx,Nph,f,phi2,wsave,apm,bpm);

	/* fast-fourier */

	if(*K_fast > 0)
	    alm_fast3_(Nmx,Nmp,Nx, Lmx,Vnyq,Nph,f,cf,apm,bpm,wsave);

	for(ii=i=0; i< *Nmx; i++, ii+=*Nmp) {
	     for(j=0; j< *Nmp; j++)  f[ii+j]=0.0;
	}

/*
	suma=sumb=0.0;
	for (i=0; i< *Nx; i++) {
	     for (m=0; m<= *Lmx; m++) {
		    int mm = m * *Nmx + i;
		    suma += apm[mm]*apm[mm];
		    sumb += bpm[mm]*bpm[mm];
	     }
	}
PR(ER, "s_amp=%15.8e,  s_bpm=%15.8e\n", sumb, suma);
*/

	/*
	 *   loop over l - SUM_l[a_{lm}P_l^m(x_p)]
	 */
	if(i_dip_flag == 1) {
	    alm_dipole_(Nmx, Nmp, Lmx,Ksym,Nx,x,apm,bpm,f);
	    if (*Kml < 0)  *Lmn = 2;
	}

	if(*Kml > 0) {
/*
PR(ER, "alm-lgndr-L\n");
*/
	    alm_lgndr_l_(Nmx,Nmp,Lmn,Lmx,Mmn,Mmx,Ksym,eps,Nx,x,apm,bpm,
		cf,wsave,f);
	}
	if(*Kml < 0) {
/*
PR(ER, "alm-lgndr-M\n");
*/
	    alm_lgndr_m_(Nmx,Nmp,Nph, Lmn,Lmx,Mmn,Mmx,Ksym,eps,Nx,x,apm,bpm,
		cf,wsave,f);
	}

	return;
}


/*
 *   loop over l - SUM_l[a_{lm}P_l^m(x_p)]
 *   input: x(Nx),Nph(Jmx),apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
 *   ouput: f(Nmx,Nmp)
 *
 *     DIMENSION f(Nmx*Nmp),x(Nx),Nph(Nmx)
 *     DIMENSION apm((Jmx+1)*Nmx),bpm((Jmx+1)*Nmx)
 *     DIMENSION c(2*Lmx+2),glm(Lmx+1)
 *
 */
void alm_lgndr_m_(Nmx,Nmp,Nph, Lmn,Lmx, Mmn, Mmx, flag_sym,
			eps,Nx,x,apm,bpm,glm,c,f)
int *Nmx, *Nmp, *Nph;
int *Lmn, *Lmx, *Mmn, *Mmx;
int *flag_sym;
double *eps;
int *Nx;
double *x, *apm, *bpm;
double *glm, *c, *f;
{
	register int l,m, mb, ix;
/*
	double sq2   = M_SQRT2,         \* 1.41421356237309504880; *\
	       sq_pi = 2./ M_2_SQRTPI,  \* = 1.772453851           *\
*/
	double sq = M_2_SQRTPI / 2. / M_SQRT2; /* 1/sqrt(2pi)=0.3989422804*/
	double cnorm;
	int Kx;
	double cmm;
	int mmin= (*Mmn < 0 ? 0 : *Mmn),      /*               */
	    mmax= (*Mmx < 0 ? *Lmx : *Mmx);   /* use in future */
	register int mm = *Nmx * mmin;
	int Lmx1 = *Lmx + 1, nx1 = *Nx - 1;
	double vnyq = 0.5;
	register int *mx = (int *)0;

	cnorm = -sq;  /* sq = 1./dsqrt(2.*pi) */
	Kx = *Nx;
	if(*flag_sym > 0)  Kx = (*Nx+1)/2;

	if ((mx = (int *)calloc((unsigned)Kx, sizeof *mx)) == (int *)0) {
	    PR(ER, "alm_lgndr_m_: no memory...\n");
	    exit(1);
	}
	for(m=0; m<Kx; m++) {
	    mx[m] = vnyq * Nph[m];
	    if (mx[m] > *Lmx) mx[m] = *Lmx;
	}


	for(m=mmin; m <= mmax; m++, mm+=*Nmx) { /* for all m */
	   int mv = m;
	   int L0 = m;
	   register int llp;
	   mb = Lmx1 + m;

	   start_clm_(Lmx,&mv,&cmm,c);

	   for (ix=0; ix<Kx; ix++) { /* for all x(ix) */
	      double xc = x[ix];
	      register double api = apm[mm+ix], bpi = bpm[mm+ix];
	      register double apj=0.0, bpj=0.0;
	      register int jx = nx1-ix;

	      if (m > mx[ix])  continue;

	      if(Kx < *Nx)  {
		    apj = apm[mm+jx],
		    bpj = bpm[mm+jx];
	      }

	      lgndr_m_(Lmx, &mv, &cmm, c, &xc, glm);

	      if(L0 < *Lmn)  L0 = *Lmn;

	      llp = L0 * *Nmp;
	      for(l=L0; l<= *Lmx; l++,llp+=*Nmp) {  /* for all m<L<Lmx */
		  register double gl = glm[l+1];
		  if(ABS(gl) > *eps) {
			f[llp+m]   += gl*api;
			f[llp+mb]  += gl*bpi;

			if(Kx < *Nx)  {  /* for symmetric x */
			   if(ix < jx) {
				int ax = 1 - 2 * MOD((l+m),2);
				if (ax < 0) {
				    f[llp+m]  -= gl*apj;
				    f[llp+mb] -= gl*bpj;
				} else {
				    f[llp+m]  += gl*apj;
				    f[llp+mb] += gl*bpj;
				}
			   }
			}         /* for symmetric x    */
		  }             /* for glm< eps       */
	      }               /* for all m<L< Lmx   */
	   }                /* for all x(ix)      */

	   /* normalization and record of alm */
	   llp = *Lmn * *Nmp;
	   for (l= *Lmn; l<= *Lmx; l++,llp+=*Nmp) {
	      f[llp+m]  *=  cnorm;
	      f[llp+mb] *= -cnorm;
	   }
	}     /*  over all m */

	if (mx != (int *)0)  free((void *)mx);
	return;
}

/*
 *    loop over l - SUM_l[a_{lm}P_l^m(x_p)]
 *    input: x(Nx),apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
 *    ouput: f(Nmx,Nmp)
 *
 *    DIMENSION f(Nmx*Nmp),x(Nx)
 *    DIMENSION apm((Jmx+1)*Nmx),bpm((Jmx+1)*Nmx)
 *    DIMENSION abc(4*Lmx),glm(2*Lmx+1)
 *
 */
void  alm_lgndr_l_(Nmx,Nmp, Lmn,Lmx, Mmn, Mmx,Ksym, eps,Nx,x,apm,bpm,glm,
		   abc,f)
int *Nmx, *Nmp;
int *Lmn, *Lmx, *Mmn, *Mmx;
int *Ksym;
double *eps;
int *Nx;
double *x, *apm, *bpm;
double *glm, *abc, *f;
{
	register int ix, l, m, mb = *Lmx+1;
	double sq2   = M_SQRT2,         /* 1.41421356237309504880; */
	       sq_pi = 2./ M_2_SQRTPI;  /* = 1.772453851           */
	double cnorm;
	int mmin= (*Mmn <  0 ? 0 : *Mmn),      /*               */
	    mmax= (*Mmx <  0 ? *Lmx : *Mmx);   /* use in future */
	int Kx = *Nx;

	if(*Ksym > 0)  Kx = (*Nx+1)/2;
	for (ix=0; ix<Kx; ix++) { /*  for all x(ix) */
	     double xc = x[ix];

	     lgndr_0_(Lmx,&xc,glm);         /* P_L^0(ix),L<Lmx */

	     for(l= *Lmn; l<= *Lmx; l++) {   /* for all L       */
		 int LL = l * *Nmp;
		 int LL0 = *Lmx+l+1;
		 int lv = l;
		 int mmn=mmin, mmx = mmax;

		 /* for m=0 */
		 if (mmn == 0) {
		    if (fabs(glm[LL0]) > *eps) {
			f[LL] += glm[LL0]* apm[0+ix];
			if(Kx < *Nx) {
			   int jx = *Nx-1-ix;
			   if(jx > ix) {
				 double ax = 1-2*MOD(l,2);
				 f[LL] += glm[LL0]*ax*apm[0+jx];
			   }
			}
		    }      /*  for glm< eps */
		 }
		 /* for m=1, ..., L */
		 lgndr_l_(Lmx,&lv, &xc, abc, glm); /* P_L^m(ix),1<m<L */

		 if (mmn <= 1)  mmn = 1;
		 if (mmx >  l)  mmx = l;

		 for(m=mmn; m<=mmx; m++) {
		   if(fabs(glm[m]) > *eps)  {
		       int mm = m * *Nmx;
		       f[LL+m]    += glm[m]*apm[mm+ix];
		       f[LL+mb+m] += glm[m]*bpm[mm+ix];
		       if ( Kx < *Nx) {
			 int jx = *Nx - 1-ix;
			 if(jx > ix) {
				double ax = 1 - 2*MOD((l+m),2);
				f[LL+m]    += glm[m]*ax*apm[mm+jx];
				f[LL+mb+m] += glm[m]*ax*bpm[mm+jx];
			 }
		       }
		   }              /* for glm< eps  */
		 }              /* over all m    */
	     }                /* over all L    */
	}                   /* end for x(ix) */

	/* normalization and record of alm */
	cnorm  = -1. / sq_pi;
	cnorm *= sq2;
	for (l= *Lmn; l<= *Lmx; l++)  { /* for all L */
	    int LL = l * *Nmp;
	    for (m=0; m<=l; m++) {    /* for all m */
		f[LL+m]    *=  cnorm;
		f[LL+mb+m] *= -cnorm;
	    }                         /* over all m */
	    /*  f[LL] /= sq2; */ /*  m=0 */
	}                              /* over all l */

	return;
}

/*
 *
 *   loop over l - SUM_l[a_{lm}P_l^m(x_p)]
 *   input: x(Nx),Nph(Jmx),apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
 *   ouput: f(Nmx,Nmp)
 *
 *     DIMENSION f(Nmx*Nmp),x(Nx)                !,Nph(Nmx)
 *     DIMENSION c(2*Lmx+2),glm(Lmx+1)
 *     DIMENSION apm((Jmx+1)*Nmx),bpm((Jmx+1)*Nmx)
 *
 */
void alm_dipole_(Nmx, Nmp, Lmx, Ksym,  Nx,x,apm,bpm,f)
int *Nmx, *Nmp, *Lmx;
int *Ksym;
int *Nx;
double *x, *apm, *bpm, *f;
{
	register int ix, mb = *Lmx + 2, L1= *Nmp;
	int Kx = *Nx, nx1=*Nx-1;
	double cnorm,
	       pi  = M_PI,
	       sq2 = M_SQRT2;

	if(*Ksym > 0) Kx=(*Nx+1)/2;

	/* for m=0,1,-1 */
	for(ix=0; ix<Kx; ix++) { /* for all x(ix) */
	    double xc = -sqrt(1. - x[ix]*x[ix]);
	    f[L1+0]  +=  x[ix]*apm[ix];        /* a_10    */
	    f[L1+1]  +=  xc*apm[*Nmx+ix];      /* Re a_11 */
	    f[L1+mb] +=  xc*bpm[*Nmx+ix];      /* Im a_11 */
	    if(Kx < *Nx) { /* for symmetric x */
		int jx = nx1-ix;
		if(ix < jx) {
		    f[L1+0]  += -x[ix]*apm[jx];       /* a_10    */
		    f[L1+1]  += +xc*apm[*Nmx+jx];     /* Re a_11 */
		    f[L1+mb] += +xc*bpm[*Nmx+jx];     /* Im a_11 */
		}
	    }                 /* for symmetric x  */
	}                 /* for all x(ix)       */

	/* normalization and record of alm */
	cnorm = -sqrt(1.5/pi);
	f[L1+0]  *=   cnorm / sq2;         /* a_10  */
	f[L1+1]  *=   cnorm / 2.;          /* Re a_11 */
	f[L1+mb] *=  -cnorm / 2.;          /* Im a_11 */

	return;
}


void alm_dipole_T(Nmx, nbuf, Lmx, Ksym,  Nx,x,apm,bpm,f)
int Nmx, nbuf, Lmx;
int Ksym;
int Nx;
double *x, *apm, *bpm, *f;
{
	register int ix, mb = Lmx + 2;
	int Kx = Nx, nx1=Nx-1;
	double cnorm,
	       pi  = M_PI,
	       sq2 = M_SQRT2;

	if(Ksym > 0) Kx=(Nx+1)/2;

	/* for m=0,1,-1 */
	for(ix=0; ix<Kx; ix++) { /* for all x(ix) */
	    double xc = -sqrt(1. - x[ix]*x[ix]);
	    f[1]  +=  x[ix]*apm[ix];        /* a_10    */
	    f[nbuf+1]    +=  xc*apm[Nmx+ix];       /* Re a_11 */
	    f[nbuf*mb+1] +=  xc*bpm[Nmx+ix];       /* Im a_11 */
	    if(Kx < Nx) { /* for symmetric x */
		int jx = nx1-ix;
		if(ix < jx) {
		    f[1]  += -x[ix]*apm[jx];      /* a_10    */
		    f[nbuf+1]    += +xc*apm[Nmx+jx];     /* Re a_11 */
		    f[nbuf*mb+1] += +xc*bpm[Nmx+jx];     /* Im a_11 */
		}
	    }                 /* for symmetric x  */
	}                 /* for all x(ix)       */

	/* normalization and record of alm */
	cnorm = -sqrt(1.5/pi);
	f[1]  *=   cnorm / sq2;         /* a_10  */
	f[nbuf+1]    *=   cnorm / 2.;     /* Re a_11 */
	f[nbuf*mb+1] *=  -cnorm / 2.;     /* Im a_11 */

	return;
}


