#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"

/*
 * Program calculates some characteristics C(l), phase(l,m), amplitude(l,m)
 * on the alm(l,m) and others
 *
 *  created  by Oleg Verkhodanov 2002-2004
 *  last changes 17.01.2005, 21/07/2006, 10/05/07, 26/03/2009, 15/07/2011,
 *               28/09/2011, 29/11/2011, 26/01/2012, 12/2015
 */
#define PR      fprintf
#define ER      stderr
#define ST      stdout
#define ABS(x)  ((x) < 0 ? -(x) : (x))

char alm_file[161], xlm_file[161], outp_file[161];
int print_amp=0, print_phase=0;
int print_pol_amp=0, print_pol_phase=0;
int  print_alm_flag = 0;
int  print_alm_gnuplot_flag = 0;
int outp_fflag=0;                   /* output alm in FITS */

int  lmax = 0;
int  lmin_print = -1;
int  lmax_print = -1;
int  dl_flag = 1;      /* flag to print Cl or Dl    */
int  cut_lmin = -1;    /* cut value for a_lm        */
int  cut_lmax = 0;     /* cut value for a_lm        */
int kolmorov_smirnov_test = 0;  /* flag to execute Kolmogorov-Smirnov test */
int sin_flag = 0;      /* print sum of sin(phase)   */
int zero_flag = 0;
int pol3alm_flag= 0;  /* polarization alm-file       */
int flag_tl = 0;      /* Tl statistics               */
double d_step = -1.0; /* rotation step for Tl        */
int flag_pol_sp = 0;  /* to calc TE or TB spectra    */
int even_harmonics_flag = 0;
int average_phase = 0;

int sd_l=0, sd_m = 0;     /* value of shifting step for l & m   */
int shift_multip_flag = 0;/* flag to use multipole shifting     */

extern int sun_data_key;

extern void setopt(int, char **), usage(void), usage2(void);
extern double ph_calc(double, double);

int    naxis2;
char *myname="alm2dl";

int main(argc, argv)
int argc;
char **argv;
{
    register int l, m=0;
    double *alm = (double *)0, *Cl = (double *)0,
	   *phase = (double *)0, *Clp = (double *)0;
    double *almp1 = (double *)0, *almp2 = (double *)0;
    int nalm=0, nbuf=0;
    FILE *fp_out = (FILE *)0;
    int file_type;

    xlm_file[0] = alm_file[0] = outp_file[0] = '\0';

    if (glesp_comments < 0) {
	if (getenv("GLESP_COMMENTS") != NULL) {
	    char str[21];
	    strcpy(str, getenv("GLESP_COMMENTS"));
	    glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
	    if (glesp_comments) {
		if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
		    glesp_comments = 0;
	    }
	} else
	    glesp_comments = 1;
    }

    /* loop over external parameters */
    setopt(argc, argv);

    /* test of input GLESP file type:
    *      flag = -1       it is not GLESP file
    *      flag = 0        F-file
    *      flag = -2       FITS-file, but not GLESP map
    *      flag = -3       ASCII-file
    *      flag = 1        FITS-file
    *      flag = 2        FITS-file, alm-file
    *      flag = 3        FITS-file, 3 maps (T,Q,U)
    *      flag = 4        FITS-file, 3 alm-s (A,E,B)
    *      flag = 5        FITS-file, Hpix map
    *      flag = 6        FITS-file, Hpix 3-map
    */
    file_type = test_glesp_map(alm_file);
    switch(file_type) {
    default:
    case 3 :
	    PR(ER, "%s: '%s' is not GLESP alm-file. It contains a map\n",
		myname, alm_file);
	    exit(1);
    case -1:
    case  0:  /* F-file    */
    case -2:  /* FITS-file */
    case  1:  /* FITS-file */
	    PR(ER, "%s: '%s' is not a GLESP file.\n",  myname, alm_file);
	    exit(1);
    case  4:  /* FITS-file 3 alm-s (A,E,B)  */
	    if (!pol3alm_flag)  pol3alm_flag = 1;
	    break;
    case  -3: /* ASCII - may be alm ? */
    case   2: /* FITS  alm-file       */
	    ;  /* Ok */
	    break;
    }
    /* read input file */
    if (pol3alm_flag) {
       /* read 3-alm file */
       read_hp_palm(&alm, &almp1, &almp2, &lmax, &nbuf, alm_file);
    }  else {
       /* read ASCII or 1-alm file */
       if (lmax ==0 && lmax_print > 0) {
	    lmax = -lmax_print;
	    nbuf = -2 * (lmax+1);
	    nalm = -(lmax + 1);
       }
       read_alm_table(alm_file, &alm, &nalm, &nbuf, &lmax);
       if (lmax < 0 ) lmax = -lmax;
    }

    /* read second file for correlation analysis of polarization */
    /* if the 1st file is not 3-alm file */
    if (flag_pol_sp && !pol3alm_flag) {
	/* read additional e_lm or b_lm file */
	int nalm1=0, nbuf1=0, lmax1=0;

	/* read 2nd file to built correlation spectrum */
	if (*xlm_file != '\0') {
	    read_alm_table(xlm_file, &almp1, &nalm1, &nbuf1, &lmax1);
	} else {
	    PR(ER, "%s: no 2nd input file with coefficients...\n",
		myname);
	    exit(1);
	}
	/* set minimal value of lmax & lmax1 */
	if ( lmax1 > lmax ) {
	     alm2alm(&almp1, &lmax1, &nbuf1, lmax, nbuf);
	} else if ( lmax1 < lmax ) {
	     alm2alm(&alm, &lmax, &nbuf, lmax1, nbuf1);
	} else if (nbuf1 != nbuf ) {
	     alm2alm(&alm, &lmax, &nbuf, lmax1, nbuf1);
	}
    }

    /* separate even and odd harmonics */
    if (even_harmonics_flag) {
	for (l=0; l<=lmax; l++) {
	    int ll = l * nbuf;
	    for (m=0; m<=l; m++) {
		if (even_harmonics_flag < 0) { /* odd */
		  if ( (l+m)%2 == 0 ) { /* if even, then nulling them */
		      alm[ll + m] = alm[ll + m + lmax+1] = 0.0;
		      if (almp1 != (double *)0)
			  almp1[ll + m] = almp1[ll + m + lmax+1] = 0.0;
		      if (almp2 != (double *)0)
			  almp2[ll + m] = almp2[ll + m + lmax+1] = 0.0;
		  }
		} else {
		  if ( (l+m)%2 == 1 ) { /* if odd, then nulling them */
		      alm[ll + m] = alm[ll + m + lmax+1] = 0.0;
		      if (almp1 != (double *)0)
			  almp1[ll + m] = almp1[ll + m + lmax+1] = 0.0;
		      if (almp2 != (double *)0)
			  almp2[ll + m] = almp2[ll + m + lmax+1] = 0.0;
		  }
		}
	    }
	}
    }

    /* cut byl L-limits */
    if (cut_lmin > 0) {
	    for(l=0; l<cut_lmin; l++) {
		int ll = l * nbuf;
		alm[ll] = 0.0;
		for(m=1; m<=l; m++) {
			alm[ll + m] = alm[ll + m + lmax+1] = 0.0;
		}
	    }
    }

    if (lmax_print < 0 || lmax_print > lmax)
	    lmax_print = lmax;
    if (lmin_print < 0 || lmin_print > lmax)
	    lmin_print = 0;

    if (outp_file[0] == '\0' || strcmp(outp_file, "stdout") == 0)
	fp_out = stdout;
    else
	if ((fp_out = fopen(outp_file, "w")) == (FILE *)0) {
	    PR(ER, "%s: can not create file '%s'\n", myname, outp_file);
	    exit(1);
	}

    /* shift mutlipoles by l and m */
    if (shift_multip_flag) {
	extern void shift_multipoles(double **, int *, int *, int, int);
	shift_multipoles(&alm, &lmax, &nbuf, sd_l, sd_m);
    }

    if (print_alm_flag) {
	double *f = alm;

	/* same operations for E,B-modes */
	switch (pol3alm_flag) {
	case 'E':
	case 'e':
		f = almp1;
		break;
	case 'B':
	case 'b':
		f = almp2;
		break;
	default :
		break;
	}
	/* take memory for phase array */
	if ((phase = (double *)calloc((unsigned)(2*(lmax+2)*(lmax+2)),
	     sizeof *phase)) == (double *)0) {
nomem:
	    PR(ER, "%s: no memory for arrays of %d elements...\n",
		    myname, 2*(lmax+2)*(lmax+2));
	    exit(1);
	}
	/* calculate phase */
	phase_calc(f, nbuf, phase, lmax);

	if (cut_lmax > 0) lmax_print = cut_lmax;
	print_alm_table (fp_out, f, lmax, nbuf, phase,
	    lmin_print, lmax_print);

    } else if (print_alm_gnuplot_flag) {
	if (cut_lmax > 0) lmax_print = cut_lmax;
	if (pol3alm_flag) {
	    switch(pol3alm_flag) {
	    case 'e':
	    case 'E':
		 print_alm_plot(fp_out,almp1,lmax,nbuf,
				 lmin_print,lmax_print);
		 break;
	    case 'b':
	    case 'B':
		 print_alm_plot(fp_out,almp2,lmax,nbuf,
				 lmin_print,lmax_print);
		 break;
	    case 't':
	    case 'T':
		 print_alm_plot (fp_out, alm, lmax, nbuf,
				 lmin_print, lmax_print);
		 break;
	    default :
		 print_alm_plot (fp_out, alm, lmax, nbuf,
				 lmin_print, lmax_print);
		 PR(fp_out, "#---------------------------------------\n");
		 print_alm_plot(fp_out,almp1,lmax,nbuf,
				 lmin_print,lmax_print);
		 PR(fp_out, "#---------------------------------------\n");
		 print_alm_plot(fp_out,almp2,lmax,nbuf,
				 lmin_print,lmax_print);
		 break;
	   }
	} else {
	   if (even_harmonics_flag)
	     print_alm_plot (fp_out,alm,lmax,nbuf,lmin_print,-lmax_print);
	   else
	     print_alm_plot (fp_out,alm,lmax,nbuf,lmin_print,lmax_print);
	}
    } else if (flag_pol_sp) {
	/* calculate TE-, TB-spectra */
	extern void alm2pcl (double *, double *, int,
	      double *, double *, double *, int);
	extern void print_2eb_table(FILE *, double *, double *, int,int);
	double *cle = (double *)0, *clb = (double *)0;

	if ((cle =(double *)calloc((unsigned)(lmax+1), sizeof *cle)) ==
	     (double *)0)
		 goto nomem;
	if ((clb =(double *)calloc((unsigned)(lmax+1), sizeof *clb)) ==
	     (double *)0)
		 goto nomem;

	if  (pol3alm_flag == 'B') {
	    /* calculate BB-spectrum */
	    alm2cl (clb, lmax, almp2, nbuf);
	} else if  (pol3alm_flag == 'E') {
	    /* calculate EE-spectrum */
	    alm2cl (cle, lmax, almp1, nbuf);
	} else {
	    /* calculate TE-, TB-spectra */
	    alm2pcl (cle, clb, lmax, alm, almp1, almp2, nbuf);
	    if (almp2 == NULL && flag_pol_sp == 'b') {
		free ((void *)clb);
		clb = cle;
	    }
	}

	switch (flag_pol_sp) {
	case 'e':
	  print_cl_table (fp_out,cle,(double*)0, lmin_print,lmax_print);
	  break;
	case 'b':
	  print_cl_table (fp_out,clb,(double*)0, lmin_print,lmax_print);
	  break;
	case  1 :
	  print_2eb_table (fp_out, cle, clb, lmin_print,lmax_print);
	  break;
	}
	if (clb == cle)  clb = (double *)0;

    } else if (flag_tl) {
	double pi2 = M_PI / 2.;
	double *vpsi, *vtheta;
	if ((vpsi = (double *)calloc((unsigned)(lmax+1),sizeof *vpsi))
	     == (double *)0)
		goto nomem;
	if ((vtheta=(double*)calloc((unsigned)(lmax+1),sizeof *vtheta))
	     == (double *)0)
		goto nomem;
	if ((Cl =(double *)calloc((unsigned)(lmax+1), sizeof *Cl)) ==
	     (double *)0)
		 goto nomem;

	/* detect maximum of Tl in rotation */
	alm2tl(alm, lmax, nbuf, Cl, vpsi, vtheta, d_step, lmax_print);

	for (l=lmin_print; l<=lmax_print; l++) {
	    double t = vtheta[l], a = vpsi[l];
	    double sth = sin(t), cth = cos(t),
		   sps = sin(a), cps = cos(a);
	    double x = sth * cps, y = sth * sps, z = cth;
	    PR(fp_out,
		"%4d   %12.5e  (l=%5.1f,b=%5.1f)  (%7.4f,%7.4f,%7.4f)\n",
		l, Cl[l], a*180/M_PI, (pi2-t)*180/M_PI,
		x,y,z);
	}

	free((void *)vpsi);
	free((void *)vtheta);
    } else if (print_amp || print_phase) {
	/* print amplitude or phase */
	double pi = M_PI;
	double *f = alm;

	/* same operations for E,B-modes */
	switch (pol3alm_flag) {
	case 'E':
	case 'e':
		f = almp1;
		break;
	case 'B':
	case 'b':
		f = almp2;
		break;
	default :
		break;
	}

	for(l=lmin_print; l<=lmax_print; l++)   {
	    int ll = l * nbuf;
	    for(m=0; m<=l; m++) {
		double re = f[ll + m],
		       im = f[ll + m + lmax+1];
		double v = print_amp ? sqrt(re*re+im*im) :
		       (m == 0 ? (re < 0. ? pi : 0.) : ph_calc(re,im));
		/* if (print_amp && fabs(im)<1.e-30) v=re; */
		if (print_phase == 2)
			v = sin(v);
		PR(fp_out, "%4d  %4d  %16.9e\n", l,m, v);
	    }
	}
    } else if (print_pol_amp || print_pol_phase) {
	double pi = M_PI;
	for(l=lmin_print; l<=lmax_print; l++)   {
	    int ll = l * nbuf;
	    for(m=0; m<=l; m++) {
		double re = alm[ll + m], im = alm[ll + m + lmax+1];
		double v = print_amp ? sqrt(re*re+im*im) :
		       (m == 0 ? (re < 0. ? pi : 0.) : ph_calc(re,im));
		/* if (print_amp && fabs(im)<1.e-30) v=re; */
		if (print_pol_phase == 2)
			v = sin(v);
		PR(fp_out, "%4d  %4d  %16.9e\n", l,m, v);
	    }
	}
    } else if (average_phase) {
	/* calculate average phase vs l */
	extern double ph_calc(double, double);
	double phase_ch=0.;
	for (l=lmin_print; l<=lmax_print; l++) {
	    register int ll = l * nbuf;
	    double si_av = 0., co_av = 0., sum_w = 0.0;

	    if (l == 0) continue;

	    for (m=1; m<=l; m++) {
		double re=alm[ll+m], im = alm[ll+m+lmax+1];
		double phase = ph_calc(re, im) + phase_ch;
		double w = 1.0;

		if (average_phase == 2)  w = 1. / sqrt(re*re+im*im);
		sum_w += w;

		if (m==0) {
		    continue;
//                      if (re < 0.0)  phase = M_PI; else phase = 0.0;
//                      if(ABS(phase_ch-M_PI) < 0.001) {
//                              phase += M_PI;
//                      }
//                      si_av += sin (phase);
//                      co_av += cos (phase);
		} else {
		    phase = ph_calc(re, im) + phase_ch;
		    si_av += w * sin (phase);
		    co_av += w * cos (phase);
		}
	    }
	    si_av /= sum_w;
	    co_av /= sum_w;
	    printf("%5d %12.5e\n", l, atan2(si_av,co_av));
	}
    } else if (sin_flag) {
	for(l=2; l<=lmax_print; l++) {
	    double s = 0.;
	    int ll = l * nbuf;
	    for(m=0; m<=l; m++) {
		double re = alm[ll + m], im = alm[ll + m + lmax+1];
		double ss = sin(ph_calc(re,im));
		s += ss*ss;
	    }
	    PR(fp_out, "%4d  %12.5e\n", l, s/(l+1));
	}
    } else if (cut_lmax > 0 || (pol3alm_flag && !flag_pol_sp && dl_flag)){
	if (cut_lmax <= 0) cut_lmax = lmax;
	if (0 < lmax_print && lmax_print < cut_lmax)
	    cut_lmax = lmax_print;
	switch (pol3alm_flag) {
	case 'e':
	     out_glesp_alm(almp1, lmax, nbuf, cut_lmax, fp_out);
	case 'b':
	     out_glesp_alm(almp2, lmax, nbuf, cut_lmax, fp_out);
	case 't':
	default :
	     out_glesp_alm(alm, lmax, nbuf, cut_lmax, fp_out);
	case  0 :
	    if (outp_fflag)
		ff_write_alm(alm, nbuf, lmax, cut_lmax, fp_out);
	    else
		out_glesp_alm(alm, lmax, nbuf, cut_lmax, fp_out);
	    break;
	}
    } else {
	if ((Cl =(double *)calloc((unsigned)(lmax+1), sizeof *Cl)) ==
	     (double *)0)
		 goto nomem;

	if (cut_lmax > 0) lmax_print = cut_lmax;
	if (!dl_flag) {
	    if (glesp_comments > 0)
	       PR(ER, "C(l) power spectrum is calculted for file '%s'\n",
		   alm_file);
	    alm2clz (Cl, lmax, alm, nbuf, zero_flag);
	} else {
	    /* calculate Dl */
	    alm2dlz (Cl, lmax, alm, nbuf, zero_flag);

	    if (glesp_comments > 0) {
	      if (dl_flag==2)
		PR(ER,
		  "D(l)=l(l+1)C(l)/(2pi) is calculted for file '%s'\n",
		  alm_file);
	      else
		PR(ER,
		  "D(l)=l(l+1)C(l)/(2piT^2) is calculted for file '%s'\n",
		  alm_file);
	    }
	    if (dl_flag==2) {
	       double T2 = 2.726*2.726;
	       for(l=lmin_print; l<=lmax_print; l++)  Cl[l] *= T2;
	    }
	}
	/* print Dl     */
	print_cl_table (fp_out, Cl, Clp, lmin_print, lmax_print);
    }

    /* free memory */
    if (Cl != (double *)0) free((void *)Cl);
    if (alm != (double *)0) free((void *)alm);
    if (phase != (double *)0) free((void *)phase);

    /* close file */
    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
	register int i;
	myname = argv[0];
	if (argc == 1)
		usage();
	for(i=1; i<argc; i++) {
	    switch(argv[i][0]) {
	    case '-':
		switch(argv[i][1]) {
		case '0':
			zero_flag = 1;
			break;
		case 'a':
			if (strncmp(argv[i]+1, "amP", 3) == 0 ||
			    strncmp(argv[i]+1, "ampP", 4) == 0) {
				print_pol_amp = 1;
			} else if (strncmp(argv[i]+1, "am", 2) == 0) {
				print_amp = 1;
			} else if (strncmp(argv[i]+1, "apw", 3) == 0) {
			       /* calc weighted average phase vs l */
			       average_phase = 2;
			} else if (strncmp(argv[i]+1, "ap", 2) == 0) {
				/* calc average phase vs l */
				average_phase = 1;
			} else if (strncmp(argv[i]+1, "ao", 2) == 0) {
				strcpy(outp_file, argv[++i]);
			} else
				strcpy(alm_file, argv[++i]);
			break;
		case 'b':
			if (argv[i][strlen(argv[i])-1] == 'l') {
			    pol3alm_flag = 'B';
			    flag_pol_sp = 'b';
			} else {
			    pol3alm_flag = 'b';
			    flag_pol_sp = 0;
			}
			break;
		case 'B':
			flag_pol_sp = 'b';
			pol3alm_flag = 'B';
			break;
		case 'c':
			if (strncmp(argv[i]+1, "clma", 4) == 0) {
				cut_lmax = atoi(argv[++i]);
			} else if (strncmp(argv[i]+1, "clmi", 4) == 0) {
				cut_lmin = atoi(argv[++i]);
			} else {
				/* print Cl */
				dl_flag = 0;
			}
			break;
		case 'C':
			dl_flag = 0;
			break;
		case 'd':
			if (strcmp(argv[i]+1, "dl") == 0) {
			    dl_flag = 1;
			} else if (strcmp(argv[i]+1, "dlt") == 0) {
			    dl_flag = 2;
			} else {
			    /* rotation step for Tl */
			    d_step = atof(argv[++i]);
			}
			break;
		case 'D':
			if (strcmp(argv[i]+1, "Dl") == 0) {
			    dl_flag = 1;
			} else if (strcmp(argv[i]+1, "Dlt") == 0) {
			    dl_flag = 2;
			} else {
			    PR(ER,"%s: illegal flag '%s'\n", myname,argv[i]);
			    usage();
			}
			break;
		case 'e':
			if (strncmp(argv[i]+1, "ev", 2) == 0) {
			   /* even harmonics */
			   even_harmonics_flag = 1;
			} else if (argv[i][strlen(argv[i])-1] == 'l') {
			    pol3alm_flag = 'E';
			    flag_pol_sp = 'e';
			} else {
			    pol3alm_flag = 'e';
			    flag_pol_sp = 0;
			}
			break;
		case 'E':
			pol3alm_flag = 'E';
			flag_pol_sp = 'e';
			break;
		case 'g': /* print alm in GNU Plot format */
			print_alm_gnuplot_flag = 1;
			break;
		case 'h':
			if (strcmp(argv[i]+1,"h2") == 0)
			    usage2();
			else
			    usage();
			break;
		case 'l':
			if (strncmp(argv[i]+1, "lmn",3) == 0 ||
			    strncmp(argv[i]+1, "lmi",3) == 0 )
				lmin_print = atoi(argv[++i]);
			else
				lmax_print = atoi(argv[++i]);
			break;
		case 'L':
			lmax = atoi(argv[++i]);
			break;
		case 'n':
			if (strncmp(argv[i]+1,"noc",3) == 0) {
			    glesp_comments = 0;
			} else {
			    PR(ER,"%s: illegal flag '%s'\n", myname,argv[i]);
			    usage();
			}
			break;
		case 'k':
			kolmorov_smirnov_test = 1;
			break;
		case 'o':
			if (strncmp(argv[i]+1, "od", 2) == 0) {
			   /* odd harmonics */
			   even_harmonics_flag = -1;
			} else {
			   if (strcmp(argv[i]+1, "of") == 0) {
				   outp_fflag = 1;
			   }
			   strcpy(outp_file, argv[++i]);
			}
			break;
		case 'p':
			if (strncmp(argv[i]+1, "ph", 2) == 0) {
				print_phase = 1;
				if (strncmp(argv[i]+1, "phs", 3) == 0)
				    print_phase = 2;
			} else if (strcmp(argv[i]+1, "pl") == 0) {
				print_alm_gnuplot_flag = 1;
				print_alm_flag = 2;
			} else
				print_alm_flag = 1;
			break;
		case 's':
			switch(argv[i][2]) {
			case 'u':   /* sum */
			    if( strncmp(argv[i], "-sun", 4) == 0) {
				 sun_data_key = 1;
			    }
			    break;
			case 'i':  /* si */
			    sin_flag = 1; break;
			case 'l': /* shift by step sl */
			    sd_l = atoi(argv[++i]);
			    shift_multip_flag = 1;
			    break;
			case 'm':
			    sd_m = atoi(argv[++i]);
			    shift_multip_flag = 1;
			    break;
			default:
			    PR(ER,"%s: illegal flag '%s'\n", myname,argv[i]);
			    usage();
			}
			break;
		case 't':
			if( strcmp(argv[i], "-t") == 0) {
			    pol3alm_flag = 't';
			    flag_pol_sp = 0;
			} else if( strncmp(argv[i], "-tl", 3) == 0) {
			    flag_tl = 1;
			} else if( strncmp(argv[i], "-tbl", 4) == 0) {
			    if ( flag_pol_sp == 'e' )
				    flag_pol_sp = 1;
			    else
				    flag_pol_sp = 'b';
			} else if( strncmp(argv[i], "-tel", 4) == 0) {
			    if ( flag_pol_sp == 'b' )
				    flag_pol_sp = 1;
			    else
				    flag_pol_sp = 'e';
			} else {
			    PR(ER,"%s: illegal flag '%s'\n", myname,argv[i]);
			    usage();
			}
			break;
		case 'v':
			PR(ER, "%s: version 0.6 (29/12/2015), GLESP 2.0\n",
			    myname);
			exit(0);
		default :
			usage();
			break;
		}
		break;
	    case '+':
		switch(argv[i][1]) {
		case 'h':
			usage();
		default :
			usage();
			break;
		}
		break;
	    default :
		if (*alm_file == '\0')
		    strcpy(alm_file, argv[i]);
		else
		    strcpy(xlm_file, argv[i]);
		break;
	    }
	}

	return;
}

void usage()
{
PR(ER,"usage: %s -a alm_file [{e,b}-lm file] [-o out_file] [-g(GNUPlot output)]\n",
myname);
PR(ER,"[-l lmax_print] [-lmin lmin_print] [-cl/dlt] [-nocom]\n");
PR(ER,"[-clmin cut_lmin] [-clmax cut_lmax] [-v(version)]\n");
PR(ER,"[-{t,e,b}(separate from 3alm-file)] [-{tbl,tel,bbl,eel}(cross-spectra)]\n");
PR(ER,"[-ph[s](print l,m,phase)] [-amp(print l,m,amplitude)] [-{odd,even}]\n");
PR(ER,"[-ampP (print l,m,Pol-amplitude=sqrt(E^2+B^2))]\n");
PR(ER,"[-tl(max.rot.)] [-sin(print Sum_sin] [-ap{w} (calc.aver.phases)]\n");
PR(ER,"[-sl shift_dl] [-sm shift_dm]\n");
exit(2);
}

void usage2()
{
PR(ST, "usage: %s xlm-file [FLAGS]\n", myname);
PR(ST, "---------------------------------------------------------------\n");
PR(ST, "-a alm_file - processes alm_file; same as without flag '-a' \n");
PR(ST, "-amp - calculates and prints data: l, m, amplitude using alm-file\n");
PR(ST, "-ampP - prints: l,m,Pol-amplitude=sqrt(E^2+B^2)\n");
PR(ST, "-ap  - calculates averaged phases\n");
PR(ST, "-apw - calculates weighted averaged phases\n");
PR(ST, "-b - extracts anisotropy B-data from 3alm-file\n");
PR(ST, "-bbl 3-alm_file - prepares BB-cross-spectrum for 3-alm_file\n");
PR(ST, "-cl - calculates C(l)-spectrum using xlm-file data\n");
PR(ST, "-dl - calculates D(l)-spectrum=l(l+1)C(l)/(2piT) using xlm-file data\n");
PR(ST, "-dlt - calculates T*D(l)-spectrum=l(l+1)C(l)/2pi using xlm-file data\n");
PR(ST, "-e - extracts anisotropy E-data from 3alm-file\n");
PR(ST, "-eel 3-alm_file - prepares EE-cross-spectrum for 3-alm_file\n");
PR(ST, "-even - extracts even harmonics (l+m) from alm-file\n");
PR(ST, "-g - prints values of x_{lm}-s in ASCII format\n");
PR(ST, "-lmax LMAX - sets the upper limit of the l-multipole at LMAX for output\n");
PR(ST, "-lmin LMIN - sets the lower limit of the l-multipole at LMIN for output\n");
PR(ST, "-nocom -no output comments\n");
PR(ST, "-o output_file - outputs result to file output_file\n");
PR(ST, "-odd - extracts odd harmonics (l+m) from alm-file\n");
PR(ST, "-ph - calculates and prints data: l, m, phase using alm-file\n");
PR(ST, "-phs - calculates and prints data: l, m, sin(phase) using alm-file\n");
PR(ST, "-sin - prints sum of sines of multipoole mode phases\n");
PR(ST, "-sl dl_to_shift - set value for shifting by l-multipole\n");
PR(ST, "-sm dm_to_shift - set value for shifting by m-multipole\n");
PR(ST, "-t - extracts anisotropy T-data from 3alm-file\n");
PR(ST, "-tbl alm_file xlm_file - prepares TB-cross-spectrum for alm_file & xlm-file\n");
PR(ST, "-tbl 3-alm_file - prepares TB-cross-spectrum for 3-alm_file\n");
PR(ST, "-tel alm_file xlm_file - prepares TE-cross-spectrum for alm_file & xlm-file\n");
PR(ST, "-tel 3-alm_file - prepares TE-cross-spectrum for 3-alm_file\n");
PR(ST, "-tl - rotates and finds maximum Sum_m (m^2*amp^2)\n");
PR(ST, "-v - prints version number of '%s'\n", myname);
exit(0);
}
