#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PR      fprintf
#define ER      stderr

extern double correlation(double *, double *, int);

void  alm_correl(alm1, alm2, nbuf, lmax, l_calc, m_calc, fp_d)
double *alm1, *alm2;
int nbuf, lmax;
int l_calc, m_calc;
FILE *fp_d;
{
	register int l, m;
	int part_flag = 3;

	double *r1 = (double *)0, *r2 = (double *)0;

	if ((r1 = (double *)calloc(lmax+1, sizeof *r1)) == (double *)0) {
nomem:
		PR(ER, "no memory for correlation arrays...\n");
		exit(2);
	}
	if ((r2 = (double *)calloc(lmax+1, sizeof *r2)) == (double *)0)
		goto nomem;

	for (l=2; l<=lmax; l++) {
		register int ll = l * nbuf;
		double correl_factor;
		for (m=0; m<=l; m++) {
		    double alm_re1, alm_im1;
		    double alm_re2, alm_im2;

		    alm_re1 = alm1[ll+m];
		    alm_im1 = alm1[ll+m+lmax+1];
		    alm_re2 = alm2[ll+m];
		    alm_im2 = alm2[ll+m+lmax+1];

		    switch (part_flag) {
		    case 1:
			r1[m] = alm_re1; r2[m] = alm_re2;
			break;
		    case 2:
			r1[m] = alm_im1; r2[m] = alm_im2;
			break;
		    case 3:
			r1[m] = atan(alm_im1/alm_re1);
			r2[m] = atan(alm_im2/alm_re2);
		    default:
			r1[m] = sqrt(alm_re1*alm_re1+alm_im1*alm_im1);
			r2[m] = sqrt(alm_re2*alm_re2+alm_im2*alm_im2);
			break;
		    }

		}
		correl_factor = correlation(r1, r2, l+1);

		PR(fp_d, "%4d %6.4f\n", l, correl_factor);
	}

	if (r1 != (double *)0)  free((void *)r1);
	if (r2 != (double *)0)  free((void *)r2);

	return;
}

double correlation(r1, r2, n)
double *r1, *r2;
int n;
{
     register int i;
     double c=0.0, a1=0.0, a2=0.0, s1=0.0, s2=0.0;

     for (i=0; i<n; i++) {
	   a1 += r1[i];
	   a2 += r2[i];
     }
     a1 /= (double)n;
     a2 /= (double)n;

     for (i=0; i<n; i++) {
	   double v1 = r1[i] - a1;
	   double v2 = r2[i] - a2;
	   s1 += v1 * v1;
	   s2 += v2 * v2;
	   c += v1 * v2;
     }
     s1 = sqrt ( s1 / (double)n );
     s2 = sqrt ( s2 / (double)n );
     c = c / (double)n;

     c = c / s1 / s2;

     return c;
}


void k_l1(alm1, alm2, lmax, nbuf, fp_d, flag)
double *alm1, *alm2;
int lmax, nbuf;
FILE *fp_d;
int flag;
{
    register int l, m;
    int k, j=0, lmik=2, lmak=lmax;
    extern double ph_calc();

    if (flag == 'm')  {
	  k = -1;
	  lmik = lmik-k;
    } else if (flag == 'p')  {
	  k = 1;
	  lmak = lmak-k;
    } else k=0;

    for(l=lmik; l<=lmak; l++) {
	register int ll = l * nbuf, lk=(l+k)*nbuf;
	double s1 = 0.0, s2a = 0.0, s2b = 0.0, v;
	double am1 = alm1[ll], am2 = alm2[lk];

	s1 += am1 * am2;
	s2a += am1 * am1;
	s2b += am2 * am2;
	for(m=1; m<=l; m++) {
	    double re1 = alm1[ll+m],   im1 = alm1[ll+m+lmax+1];
	    double re2 = alm2[lk+m+j], im2 = alm2[lk+m+j+lmax+1];
	    double ph1 = ph_calc(re1,im1), ph2 = ph_calc(re2,im2);
	    double amp1=sqrt(re1*re1+im1*im1), amp2=sqrt(re2*re2 + im2*im2);

	    s1 += 2 * amp1*amp2*cos(ph1-ph2);
	    s2a += 2 * amp1*amp1;
	    s2b += 2 * amp2*amp2;
	}
	v = s1 / sqrt(s2a*s2b);
	PR(fp_d, "%4d %7.4f\n", l,  v);
    }

    return;
}


void k_l(alm1, alm2, lmax, nbuf, fp_d)
double *alm1, *alm2;
int lmax, nbuf;
FILE *fp_d;
{
    register int l, m;

    for(l=2; l<=lmax; l++) {
	register int ll = l * nbuf;
	double s1 = 0.0, s2a = 0.0, s2b = 0.0, v;

	s1 += alm1[ll] * alm2[ll];
	// s2 += sqrt(alm1[ll] * alm2[ll]);
	s2a += alm1[ll] * alm1[ll];
	s2b += alm2[ll] * alm2[ll];
	for(m=1; m<=l; m++) {
	    double re1 = alm1[ll+m], im1 = alm1[ll+m+lmax+1];
	    double re2 = alm2[ll+m], im2 = alm2[ll+m+lmax+1];

	    s1 += 2 * (re1 * re2 + im1 * im2);
     //     s2 += 2 * ((re1*re1 + im1*im1) * (re2*re2 + im2*im2));
	    s2a += 2 * (re1*re1 + im1*im1);
	    s2b += 2 * (re2*re2 + im2*im2);
	}
	v = s1 / sqrt(s2a*s2b);
	PR(fp_d, "%4d %7.4f\n", l,  v);
    }

    return;
}
