#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"

#define  PR     fprintf
#define  ER     stderr

extern void spline (double *, double *, int, double, double, double *);
extern void splint (double *, double *, double *, int, double, double *);
extern void gauleg (double, double, double *, double *, int);
extern int test_flag; /* = 0; */

/*
 * appropriate pixel to Gauss-Legendre quadrature grid
 *
 */
void approp_hpix(f, nx, nphi, nph, x, x1, x2)
double *f;
double *x;
int *nph;
int nx, nphi;
double x1, x2;
{
    register int i, j;
    double *xl = (double *)0, *f_n = (double *)0;
    int *nph2 = (int *)0;
    int nphi_eq_flag = 1;
    int np, npmax;

    i = nphi > nx ? nphi : nx;
    if ((xl = (double *)calloc((unsigned)i, sizeof *xl)) == (double *)0)
	   goto nomem;
    if ((nph2 = (int *)calloc((unsigned)i, sizeof *nph2)) == (int *)0)
	   goto nomem;
    i = nphi * nx;
    if ((f_n = (double *)calloc((unsigned)i, sizeof *f_n)) == (double *)0){
nomem:
	   PR(ER, "approp_hpix: no memory for temporal arrays...\n");
	   exit(1);
    }

    /* check equal or different numbers of phi */
    npmax=np=nph[0];
    for(i=1; i<nx; i++) {
	   if (np  !=  nph[i])  nphi_eq_flag = 0;
	   if (npmax < nph[i])  npmax = nph[i];
    }

{
double mx = -1.e30, mn=1.e30;
i = nx * nphi;
for(j=0; j<i; j++) {
if (f[j] > mx) mx=f[j];
if (f[j] < mn) mn=f[j];
}
PR(ER,"0,mx=%g, mn=%g\n", mx, mn);
}
    /* copy map by phi: restore equal step by phi */
/*
    if ( ! nphi_eq_flag ) {
       for(i=0; i<nx; i++) {
	   int ii = i * nphi;

	   if (nph[i] != npmax) {
		double dp_rat = (double)nph[i] / npmax;
		for(j=0; j<npmax; j++) {
		     int j_k = j * dp_rat + 0.5;
		     f_n[ii+j] = f[ii+j_k];
		}
	   }
	   nph2[i] = npmax;
       }
    }
*/
for(i=0; i<nx; i++) {
    int ii = i * nphi;
    for(j=0; j<nph[i]; j++) {
	  f_n[ii+j] = f[ii+j];
    }
    nph2[i] = nph[i];
    xl[i] = x[i];
}


{
double mx = -1.e30, mn=1.e30;
i = nx * nphi;
for(j=0; j<i; j++) {
if (f_n[j] > mx) mx=f[j];
if (f_n[j] < mn) mn=f[j];
}
PR(ER,"1,mx=%g, mn=%g\n", mx, mn);
}

    /* select and calculate grid for Legendre polynomials */
    gauleg(x1, x2, xl, (double *)0, nx);

    /* set pixel grid with equal areas */
    set_nphi_by_x2(xl, nx, nphi, nph);

    /*-----------------------------------------------*/
    /* copy pixels to corresponding place            */
    /*  (f_n,nx,nphi,nph2,x) -> (f,nx,nphi,nph,xl);  */
    /*-----------------------------------------------*/
    cop_grid(f_n, nx, nphi, nph2, x,  f, nx, nphi, nph, xl);
    for(j=0; j<nx; j++)  x[j] = xl[j];

    /* free memory */
    if(xl != (double *)0) free((void *)xl);
    if(f_n!= (double *)0) free((void *)f_n);
    if(nph2  != (int *)0) free((void *)nph2);

    return;
}

/* repixelize (f1,nx1,np1,nph1,x1) -> (f2,nx2,np2,nph2,x2) */
void cop_grid(f1, nx1, np1, nph1, x1, f2, nx2, np2, nph2, x2)
double *f1, *f2;
double *x1, *x2;
int nx1, nx2, np1, np2;
int *nph1, *nph2;
{
    register int i, j;
    int *count = (int *)0;

    i = nx2 * np2;
    if ((count = (int *)calloc((unsigned)i, sizeof *count)) == (int *)0) {
	   PR(ER, "cop_pix: no memory for temporal arrays...\n");
	   exit(1);
    }
    /*------------------------------------*/
    /* copy pixels to corresponding place */
    /*------------------------------------*/
    for(j=0; j<i; j++)  f2[j] = 0.0;
    for(i=0; i<nx2; i++) {
	int ii = i * np2;
	for(j=0; j<nph2[i]; j++) {
	    int p2 = ii+j;
	    /* old: f1, new: f2 */
	    /* convert: x1,nph1,nx1,jj -> x2,nph2,nx2,ip    */
	    /* i.e. obtain 'pix1' by 'pix2' and appropriate */
	    /*  f2[p2] = f1[p1]                             */
	    int p1 = pix2pix(x1, nx1, np1, nph1, p2, x2, nx2, np2, nph2);

	    if (p1 >= 0) {
		  if (!test_flag) {
			 f2[p2] += f1[p1];
			 count[p2]++;
		  } else {
			 if (fabs(f1[p1]) > 1.e-15)
				f2[p2] = 1.0;
		  }
	    }
	}
    }
    if (!test_flag) {
	i = nx2 * np2;
	for(j=0; j<i; j++)
	    if (count[j]>1)  f2[j] /= count[j];
    }

    /* free memory */
    if(count != (int *)0) free((void *)count);
    return;
}
/* repixelize (f1,nx1,np1,nph1,x1) -> (f2,nx2,np2,nph2,x2) */
void repix_grid(f1, nx1, np1, nph1, x1, f2, nx2, np2, nph2, x2)
double *f1, *f2;
double *x1, *x2;
int nx1, nx2, np1, np2;
int *nph1, *nph2;
{
    register int i, j;
    int *count = (int *)0;

    i = nx2 * np2;
    if ((count = (int *)calloc((unsigned)i, sizeof *count)) == (int *)0) {
	   PR(ER, "cop_pix: no memory for temporal arrays...\n");
	   exit(1);
    }
    /*------------------------------------*/
    /* copy pixels to corresponding place */
    /*------------------------------------*/
    for(j=0; j<i; j++)  f2[j] = 0.0;
    for(i=0; i<nx1; i++) {
	int ii = i * np1;
	for(j=0; j<nph1[i]; j++) {
	    int p1 = ii+j;
	    /* old: f1, new: f2 */
	    /* convert: x1,nph1,nx1,p1 -> x2,nph2,nx2,p2    */
	    /* i.e. obtain 'pix2' by 'pix1' and appropriate */
	    /*  f2[p2] = f1[p1]                             */
	    int p2 = pix2pix(x2, nx2, np2, nph2, p1, x1, nx1, np1, nph1);

	    if (p2 >= 0) {
		  if (!test_flag) {
			 f2[p2] += f1[p1];
			 count[p2]++;
		  } else {
			 if (fabs(f1[p1]) > 1.e-15)
				f2[p2] = 1.0;
		  }
	    }
	}
    }
    if (!test_flag) {
	i = nx2 * np2;
	for(j=0; j<i; j++)
	    if (count[j]>1)  f2[j] /= (double)count[j];
    }

    /* free memory */
    if(count != (int *)0) free((void *)count);
    return;
}

/* repixelize (f1,nx1,np1,nph1,x1) -> (f2,nx2,np2,nph2,x2) */
/* phi-s can have offset                                   */
void cop_hgrid(f1, nx1, np1, nph1, x1, phi_offset, f2, nx2, np2, nph2, x2)
double *f1, *f2;
double *x1, *x2;
double *phi_offset;
int nx1, nx2, np1, np2;
int *nph1, *nph2;
{
    register int i, j;
    int *count = (int *)0;

    if (phi_offset == (double *)0) {
	/* repixelize (f1,nx1,np1,nph1,x1) -> (f2,nx2,np2,nph2,x2) */
	repix_grid(f1, nx1, np1, nph1, x1, f2, nx2, np2, nph2, x2);
	return;
    }

    i = nx2 * np2;
    if ((count = (int *)calloc((unsigned)i, sizeof *count)) == (int *)0) {
	   PR(ER, "cop_pix: no memory for temporal arrays...\n");
	   exit(1);
    }
    /*------------------------------------*/
    /* copy pixels to corresponding place */
    /*------------------------------------*/
    for(j=0; j<i; j++)  f2[j] = 0.0;
    for(i=0; i<nx2; i++) {
	int ii = i * np2;
	for(j=0; j<nph2[i]; j++) {
	    int p2 = ii+j;
	    /* old: f1, new: f2 */
	    /* convert: x1,nph1,nx1,jj -> x2,nph2,nx2,ip    */
	    /* i.e. obtain 'pix1' by 'pix2' and appropriate */
	    /*  f2[p2] = f1[p1]                             */
	    int p1 = hpix2pix(x1, nx1, np1, nph1, phi_offset,
			  p2, x2, nx2, np2, nph2);

	    if (p1 >= 0) {
		  if (!test_flag) {
			 f2[p2] += f1[p1];
			 count[p2]++;
		  } else {
			 if (fabs(f1[p1]) > 1.e-15)
				f2[p2] = 1.0;
		  }
	    }
	}
    }
    if (!test_flag) {
	i = nx2 * np2;
	for(j=0; j<i; j++)
	    if (count[j]>1)  f2[j] /= count[j];
    }

    /* free memory */
    if(count != (int *)0) free((void *)count);
    return;
}
