#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr

void beam_tikh(alm, lmax, nbuf, beam_l, alp)
double *alm;
int lmax, nbuf;
double *beam_l;
double alp;
{
    register int l, m;
    double f;

    for(l=0; l<=lmax; l++) {
	int ll = l * nbuf;
	double ex = beam_l[l] + alp;
	if (fabs(ex) < 1.e-30)  f = 0.0;  else f = 1./ex;
	for(m=0; m<=l; m++) {
		alm [ll+m]        *= f;
		alm [ll+m+lmax+1] *= f;
	}
    }

    return;
}

void beam_smooth(alm, lmax, nbuf, beam_l)
double *alm;
int lmax, nbuf;
double *beam_l;
{
    register int l, m;
    double f;

    for(l=0; l<=lmax; l++) {
	int ll = l * nbuf;
	f = beam_l[l];
	for(m=0; m<=l; m++) {
		alm [ll+m]        *= f;
		alm [ll+m+lmax+1] *= f;
	}
    }
    return;
}

void beam_gauss_smooth(alm, lmax, nbuf, fwhm)
double *alm;
int lmax, nbuf;
double fwhm;
{
    register int l, m;
    double f;

    /* transform beam size to window size */
    fwhm *= M_PI/180. / 60. /  2.354820045;  /* 2.35=sqrt(8*ln 2):sig->fw */
    fwhm *= fwhm * 0.5;

    for(l=0; l<=lmax; l++) {
	int ll = l * nbuf;
	f = exp(-l*(double)(l+1)*fwhm);
	for(m=0; m<=l; m++) {
		alm [ll+m]        *= f;
		alm [ll+m+lmax+1] *= f;
	}
    }
    return;
}

void beam_tikh_gauss(alm, lmax, nbuf, fwhm, alp)
double *alm;
int lmax, nbuf;
double alp;
double fwhm;
{
    register int l, m;
    double f;

    /* transform beam size to window size */
    fwhm *= M_PI/180. / 60. /  2.354820045;
    fwhm *= fwhm * 0.5;

    for(l=0; l<=lmax; l++) {
	int ll = l * nbuf;
	double ex;
	f = exp(-l*(double)(l+1)*fwhm);
	ex = f + alp;
	if (fabs(ex) < 1.e-30)  f = 0.0;  else f = 1./ex;
	for(m=0; m<=l; m++) {
		alm [ll+m]        *= f;
		alm [ll+m+lmax+1] *= f;
	}
    }

    return;
}


void beam_cfg_smooth(alm, lmax, nbuf, bb)
double *alm;
int lmax, nbuf;
double bb;
{
    register int l, m;
    double f;

    /* transform beam size to window size */

    for(l=0; l<=lmax; l++) {
	int ll = l * nbuf;
	f = exp(-l*(double)(l+1)/bb);
	for(m=0; m<=l; m++) {
		alm [ll+m]        *= f;
		alm [ll+m+lmax+1] *= f;
	}
    }
    return;
}
