#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr

/*
 *
 *  Oleg Verkhodanov (NBI), vo@sao.ru
 *          last modification, 5/10/2004
 */
extern void   setopt(int, char **), usage(void);


extern int sun_data_key;
extern int swap_write_key;

int nx_new = -1, np_new = -1;

char tmpcat[161];
char map_file[161], outfile[161];
int  bitpix_map = GLESPPIX; /* default: Float Point Pixels for map */
HEAD hd, *ph;

char *myname="mapcut";

int main(argc, argv)
int argc;
char **argv;
{
    register int i, j;
    double *f = (double *)0;
    double *x = (double *)0;
    int *nph = (int *)0, *nph_n = (int *)0;
    int nx, np;
    double *w_n = (double *)0, *x_n = (double *)0, *f_n = (double *)0;
    FILE *fp_out = (FILE *)0;
    char x_type[40], grid_type[40];
    double pi2 = M_PI * 2.;

    /* select grid output keyword values */
    strcpy(x_type, "COS(THETA)");
    strcpy(grid_type, "PHI(X)+LEGENDRE");
    strcpy(tmpcat, "/tmp");

    /* loop over external parameters */
    setopt(argc, argv);

    /* read a GLESP map */
    read_glesp_map(&f, &x, &nx, &np, &nph, map_file);

    if (nx_new < 0 && np_new < 0)  {nx_new = nx; np_new = np;}
    if (nx_new < 0 && np_new > 0)  nx_new = np_new/2;
    if (nx_new > 0 && np_new < 0)  np_new = nx_new*2;

    /* take memory for a new map */
    if((x_n = (double*)calloc((unsigned)nx_new, sizeof *x_n))==(double *)0){
nomem:
	   PR(ER,"%s: no memory for working arrays\n", myname);
	   exit(1);
    }
    if((nph_n=(int *)calloc((unsigned)nx_new, sizeof *nph_n))==(int *)0)
	   goto nomem;
    if((f_n = (double *)calloc((unsigned)(nx_new*np_new), sizeof *f_n))
	== (double *)0)
	   goto nomem;
    if((w_n=(double *)calloc((unsigned)nx_new, sizeof *w_n))==(double*)0)
	   goto nomem;


    /* change resolution */
    if (spline_interpolation) {
	/* spline interpolation */
    } else {
	/* copy pixels */
	if (nx_new > nx || (nx == nx_new && np_new > np) {
	    /* to higher resolution: just coping */
	} else {
	    /* to lower resolution: use average */

	}
    }

    /********************/
    /* open output file */
    /********************/
    if (strcmp(outfile, "stdout") == 0)  strcpy(outfile, "-");
    if (*outfile == '\0' || strcmp(outfile, "-") == 0)
	fp_out = stdout;
    else {
	if((fp_out = fopen(outfile, "w")) == (FILE *)0) {
	    PR(ER, "%s: can not create file '%s'\n",
		myname, outfile);
	    exit(1);
	}
    }

    /*  binary */
    if (map_out_flag)  {
	if (ffile_flag) {
	    /* map output in F-format */
	    ff_write_ext2_map(f_n,x_n,nx_new,np_new,nph_n,
		     fp_out, x_type,grid_type);
	} else {
	    /* map output in FITS-format */
	    outp_glesp_map(f_n, x_n, nx_new, np_new,nph_n,
		     fp_out, x_type,grid_type);
	}
    }

    fclose(fp_out);

    /* free memory */
    if (r != (double *)0)  free((void *)r);
    if (f != (double *)0)  free((void *)f);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    char s1[50], s2[50], s3[50], s4[50];
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'd':
			if (strcmp("double", argv[i]+1) == 0) {
				bitpix_map = DOUBLEPIX;
			} else {
			     PR(ER, "%s: illegal flag '%s'\n",
				  myname, argv[i]);
			     usage();
			}
			break;
		case 'f':
			if (strcmp("float", argv[i]+1) == 0) {
				bitpix_map = FLOATPIX;
			} else {
			     PR(ER, "%s: illegal flag '%s'\n",
				  myname, argv[i]);
			     usage();
			}
			break;
		case 'h':
			usage();
			break;
		case 'o':
			if(strncmp("of", argv[i]+1, 2) == 0) {
				ffile_flag = 1;
			}
			strcpy(outfile, argv[++i]);
			break;
		case 's':
			switch(argv[i][2]) {
			case 'u':
			     sun_data_key = 1;
			     break;
			case 'w':
			     swap_write_key = 1;
			     break;
			default :
			     usage();
			     break;
			}
			break;
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		strcpy(map_file, argv[i]);
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s file  -o output_file [-nx new_nx] [-np new_np]\n");
PR(ER, "[-{double,float}]\n");
exit(2);
}
