/*
 * cl2map - map generation and spectral calculation for CMB,
 *          basic procedure for TAC4CMB package
 *
 * uses Andrej G. Doroshkevich's FORTAN subroutines signal.f and alm.f
 *     form alm->map and map->alm conversions
 *
 *
 * Oleg V. Verkhodanov, (TAC; SAO) Version 0.0 (stopped on 3-Dec-2002)
 *      Version 0.1 (started on 3-Dec-2002)
 *         Variable Nph[i] are set
 *      Version 0.2 (started on 17-Feb-2003) - pixelization selected.
 *         Parallel OMP-procedure started to be included (with V.Stolyarov)
 *      Version 1.0 (started on 3-Sep-2004)
 *      Changes on 21-Sep-2004 at NBI, 19-Oct-2004, SAO
 *                 04/2009 (NBI),  06/2009 (SAO)
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"

#define ABS(x)  ((x) < 0 ? -(x) : (x))
#define PR      fprintf
#define ER      stderr
#define ST      stdout

extern void signal_(int *, int *, int *, int *, int *, double *, double *,
		  int *, int *, double *, double *, double *, int *,
		  double *, double *, double *, double *,
		  int *, int *, int *);
extern void a_lm_(int *, int *, int *, int *, int *,
		  double *, double *, double *,
		  int *, int *, double *, double *, double *, int *,
		  double *, double *, double *, double *,
		  int *, int *, int *);
extern void coord(int, int, double, double, double, double *, double *,
		  int, int *, int);
extern void weights(int, double, double, double, double *, double *,
		  int, int *);

extern void read_pars_file(char *);
extern void setopt(int, char **), usage(void), usage2(void);
extern void set_sig_fatal(void);

int file_pars_flag = 0;     /* read parameters from file */
int debug_flag = 0;         /* debug flag                */
int debug_sum_flag = 0;     /* debug flag                */
double fwhm = -1.;          /* if > 0, the smoothing beam in arcmin */
int Cl_input=0;             /* flag showing reading Cl  from file   */
int  outp_fflag = 0;   /* default output binary data in FITS format */
char alm_out_file[161];

int seed_flag=0,            /* flag to use user seed                */
    random_seed=0;          /* input random seed                    */

/*********************/
/* common parameters */
/* from pars-file    */
/*********************/
char file_C_l[161];   /* file with l, Cl tables                 */
char inp_file[161];   /* intput file for F(x,phi) or  a_{lm}    */
char outp_file[161];  /* output file for a_{lm} or F(x,phi)     */
int  calc_flag=-1;    /* <0, transforms a_lm to map, else back  */
int  calc_grid= 1;    /* <0 - dx=2/(Nx-1), >0 - special grid    */
int  calc_number=1;   /* number of calculation                  */
int  num_x=-1;        /* number of points along cos(theta)      */
int  num_phi=-1;      /* number of points along phi             */
int  l_min=-1;        /* Lmin                                   */
int  l_max=-1;        /* Lmax                                   */
int  m_min=-1;        /* Mmin                                   */
int  m_max=-1;        /* Mmax                                   */
int nomap=0;          /* stop calc-s after alm-simulation       */

int  map_input;       /* flag to read map from file             */
char map_file[161];   /* file with map                          */

char file_alm[161];   /* file with alm tables                   */
int  alm_input=0;     /* flag to read  a_lm from file or str    */
char alm_str[51];     /* string with a_{lm},format as ASCII-file*/
int  fft_flag=1;      /* default flag for FFT                   */
int  legendre_calc_scheme = 0;  /* if flag=-1, then direct calculations, */
				/* else from largest M to lower          */
int  sym_grid = 1;    /* symmetric grid over x                  */
int  fill_0 = 0;      /* fill arrays by 0 till common length for FFT       */
double  Vnyq =0.5;    /* Nyquist-factor for max m in respect with Nph(ix), */
		      /*  It is importnat for varying Nph(ix)              */
//double eps = 1.e-300; /* limit for neglecting (Plm < eps equal 0)        */
//int lx_add = -10;     /* addition to Nphi numbers for 'coord'-function   */
double eps = 1.e-20;  /* limit for neglecting (Plm < eps equal 0)          */
int lx_add =  0;      /* addition to Nphi numbers for 'coord'-function     */
		      /* 0 - pin 0, -10 - old scheme                       */
double calibration_factor = 1.0;
int  dipole_flag = 1;
int cosbell_flag = 0;       /* multiply by cos-bell             */
int output_format_flag = 2; /* default:0=1 - FITS, 2 - F-format */
int cl_output_flag = 1;
int dl_output_flag = 0; /* for Dl output */
char dl_out_file[161];  /*               */

/**********************/
/*  common parameters */
/*  from setopt func  */
/**********************/
char so_i_file[161];   /* intput file for F(x,phi) or  a_{lm}    */
char so_o_file[161];   /* output file for a_{lm} or F(x,phi)     */
int  so_c_flag = -10;  /* <0, transforms a_lm to map, else back  */
int  so_c_grid = -10;  /* <0 - dx=2/(Nx-1), >0 - special grid    */
int  so_num_x  = -10;  /* number of points along cos(theta)      */
int  so_num_phi= -10;  /* number of points along phi             */
int  so_l_min  =   0;  /* Lmin                                   */
int  so_l_max  = -10;  /* Lmax                                   */
int  so_m_min  = -10;  /* Mmin                                   */
int  so_m_max  = -10;  /* Mmax                                   */
int  so_l_c    = -10;  /* index of input harmonique for f(x,phi) */
int  so_m_c = -100000; /* index of input harmonique for f(x,phi) */
int monopole=1;        /* calculate monopole */

int  NMX = 0;
int  der2_flag = 0;
int  sin2_flag = 0;
char file_par[160];

char *myname="cl2map";

int main (argc, argv)
int argc;
char **argv;
{
    double *x=(double *)0, /* array of arguments x=cos theta, [-1;1] */
	   *pw=(double *)0;

    /*---------------------------------------------------------------*/
    /* in FORTRAN f(NMX,2*NMX): f(l,m), l=1,lmax, m=0,l, + real,imag */
    /* or map: f(x,phi), x=cos theta, x=-1,1, phi=0,2pi,             */
    /*---------------------------------------------------------------*/
    double *f=(double *)0;       /* main array */
    double *wsave=(double *)0;   /* buffer     */
    int    *Nph = (int *)0;      /* array of phi dimensions */

    int Nmx,
	Nmphi,          /* Nmp - max number of cells */
	Jmx;            /* Jmx - max value of Lmx    */
    int Nphi=0;
    FILE *fp_out = (FILE *)0;
    char str[160];
    int map_flag,   /* flag: >0,transforms a_lm to map,else back */
	Kx;         /* flag: <0 - dx=2/(Nx-1), >0 - special grid */
    int Lmn=2, Lmx=2, Mmn=0, Mmx=2, Nx;
    double sm, sum;
    register int i,j;
    char file_nnm[40];
    double *cl    = (double *)0, /* array for C_l   */
	   *cpl   = (double *)0; /* array for Cp_l  */
    double *xbuf  = (double *)0;
    double *cf   = (double *)0,
	   *apm  = (double *)0,
	   *bpm  = (double *)0;
    double x1 = -1.0, x2=1.0, phi2 = M_PI * 2;
    double alm00=0.0;

    map_file[0]  = inp_file[0]  = outp_file[0] = '\0';
    so_i_file[0] = so_o_file[0] = '\0';
    file_C_l[0]  = '\0';
    file_par[0] = '\0';
    strcpy(tmpcat, ".");
    theta_grid=0;       /*if=1,flag to restore grid for theta   */

    /* signal processing */
    set_sig_fatal();

    alm_out_file[0] = '\0';

    /* loop over external parameters */
    setopt(argc, argv);

    if (glesp_comments < 0) {
	if (getenv("GLESP_COMMENTS") != NULL) {
	    char str[21];
	    strcpy(str,getenv("GLESP_COMMENTS"));
	    glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
	    if (glesp_comments) {
		 if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
		     glesp_comments = 0;
	    }
	} else
	    glesp_comments = 1;
    }

    Nmx = NMX;
    Nmphi=2*NMX;    /* Nmp - max number of cells */
    Jmx=(Nmx+1)/2;  /* Jmx - max value of Lmx    */

    if (fft_flag == 0) fft_flag = -1;
    if (sym_grid == 0) sym_grid = -1;

    /* check default value of array sizes */
    if (Nmx < so_num_x) {
	    num_x = so_num_x;
	    Nmx = num_x + 10;
	    Jmx=(Nmx+1)/2;      /* Jmx - max value of Lmx    */
    }
    if ((Nmphi  = 2*Nmx) < so_num_phi)  {
	    num_phi = so_num_phi;
	    if (num_phi%2)  /* odd value  */
		    Nmx = (num_phi+1) / 2;
	    else            /* even value */
		    Nmx = num_phi / 2;
	    Nmphi = 2 * Nmx;
	    Jmx=(Nmx+1)/2;      /* Jmx - max value of Lmx    */
    }

    /*------------------------------*/
    /* analysis of input parameters */
    /*------------------------------*/
    if (*file_par != '\0') {
       /* read parameters from file */
       read_pars_file(file_par);
    }
    if (legendre_calc_scheme == 0) legendre_calc_scheme = -1;
    {
	if ( *so_i_file != '\0')     strcpy(inp_file,  so_i_file);
	if ( *so_o_file != '\0')     strcpy(outp_file, so_o_file);
	if ( so_c_flag  >  -10 )     calc_flag   = so_c_flag ;
	if ( so_c_grid  >  -10 )     calc_grid   = so_c_grid ;
	if ( so_num_x   >  -10 )     num_x       = so_num_x  ;
	if ( so_num_phi >  -10 )     num_phi     = so_num_phi;
	if ( so_l_min   >  -10 )     l_min       = so_l_min  ;
	if ( so_l_max   >  -10 )     l_max       = so_l_max  ;
	if ( so_m_min   >  -10 )     m_min       = so_m_min  ;
	if ( so_m_max   >  -10 )     m_max       = so_m_max  ;

	/********************/
	/* check parameters */
	/********************/
	if ( ! Cl_input && ! alm_input && ! map_input ) {
	    if (*inp_file  == '\0')  {
		PR(ER,"%s:  input file name, please => ", myname);
		fgets(inp_file, 160, stdin);
		if ( *inp_file == '\0' || *inp_file == '\n') {
		     PR(ER,"%s: sorry, no input file name...\n", myname);
		     exit(1);
		}
	    }
	}
	if (num_x < 0 && ! map_input)  {
	    *str = '\0';
	    PR(ER,"%s:  number of points by x (cos theta) => ", myname);
	    fgets(str, 160, stdin);
	    if ( *str == '\0' || *str == '\n') {
		 PR(ER,"%s: sorry, no input \\theta-step...\n", myname);
		 exit(1);
	    }
	    num_x = atoi(str);
	}
	if (num_phi < 0 && ! map_input)  {
	    *str = '\0';
	    PR(ER,"%s:  number of points by \\phi          => ", myname);
	    fgets(str, 160, stdin);
	    if ( *str == '\0' || *str == '\n') {
		 PR(ER,"%s: sorry, no input \\phi-step...\n", myname);
		 exit(1);
	    }
	    num_phi = atoi(str);
	}

	if ( l_min  < 0 )       l_min = 0 ;
	if ( m_min  < 0 )       m_min = 0  ;
	if ( m_max  < 0 )       m_max = l_max ;


	/*-------------------*/
	/* set old variables */
	/*-------------------*/
	/* old: K_alm_xphi */
	map_flag = calc_flag;/* <0,transforms a_lm to map,else back */
	Kx  = calc_grid;     /* <0 - dx=2/(Nx-1), >0 - special grid */
	Nx   = num_x;
	Nphi = num_phi;
	Lmn = l_min;
	Lmx = l_max;
	Mmn = m_min;
	Mmx = m_max;

    }
    if (Lmn > 0) monopole=0; /* no calculation of monopole */

    sprintf(file_nnm, "%s.dat", myname);

    if (!nomap) {
	if (*outp_file  == '\0')  {
	    strcpy (outp_file, file_nnm);
	    if (glesp_comments) {
	      if(map_flag < 0)
		PR(ER, "The output map will be written to the file '%s'.\n",
		    outp_file);
	      else
		PR(ER,"The output a_{lm} will be written to the file '%s'.\n",
		    outp_file);
	    }
	}

	if (*outp_file  == '\0')  {
	    PR(ER,"%s: output file name, please => ", myname);
	    fgets(outp_file, 160, stdin);
	    if ( *outp_file == '\0' || *outp_file == '\n') {
		PR(ER,"%s: sorry, no output file name...\n", myname);
		exit(1);
	    }
	}

	if ((fp_out = fopen(outp_file, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n", myname, outp_file);
		exit(1);
	}
    }
    /*-------------------------*/
    /*  start of calculations  */
    /*-------------------------*/
    if(map_flag < 0)   {
	/*--------------------------------*/
	/* map preparation and simulation */
	/*--------------------------------*/

	/***************************************/
	/* simulation of a_lm by C_l and phase */
	/***************************************/
	if ( Cl_input ) {
	    /* create C_l and Cp_l arrays and  */
	    /* reads data from file CMBfast    */
	    if (ABS(Cl_input) == 2)
		   read_dl_table (file_C_l, &cl, &cpl, &Lmx);
	    else
		   read_cl_table (file_C_l, &cl, &cpl, &Lmx);

	    /* set limits of array dimensions */
	    /*--------------------------------*/
	    if (l_max > Lmx) {
		cl = (double *)realloc(cl, (unsigned)(l_max+1)*sizeof *cl);
		if (cl == (double *)0)  goto nomem_w;
		for(i=Lmx+1; i<=l_max; i++)  cl[i] = 0.0;
		Lmx = l_max;
	    }
	    Jmx = Lmx + 1;
	    if (Nx   < 0)     Nx    = 2*Lmx+1;
	    if (Nphi < 0)     Nphi  = 2*Nx;
	    if (Nx   > Nmx)   Nmx   = Nx;
	    if (Nphi > Nmphi) Nmphi = Nphi;

	    /* take memory for arrays */
	    /*------------------------*/
	    if ((f=(double*)calloc((unsigned)(Nmx*Nmphi),sizeof *f))==
		(double*)0) {
nomem_w:
		PR(ER,"%s: no memory for working arrays\n", myname);
		exit(1);
	    }

	    /* simulation a_lm by C_l without phase */
	    cl2alm (cl, Lmx, f, Nmphi, fwhm);

	} else if (alm_input) {
	    if (alm_input == 1) {
		/* read alm from ASCII or binary file */
		/* full size (Nmx, Nmphi) for Lmx)    */
		/*------------------------------------*/
		int flag;
		int test_glesp_map(char *),
		    read_alm_3table (char *, double **, int *, int *, int *);

		flag = test_glesp_map(file_alm);

		if (flag == 4) {
		    Nmx = -Nmx;
		    Nmphi = -Nmphi;
		    read_alm_3table (file_alm, &f, &Nmx, &Nmphi, &Lmx);
		    if(Nmx < 0)   Nmx = -Nmx;
		    if(Nmphi < 0) Nmphi = -Nmphi;
		} else /* if (flag == -3 || flag == 2) */
		    /* ASCII or FITS */
		    read_alm_table (file_alm, &f, &Nmx, &Nmphi, &Lmx);
	    } else {
		/* read alm from ASCII string         */
		/*------------------------------------*/
		Lmx = get_lmax_str(alm_str);

		/* set limits & take memory */
		/*--------------------------*/
		j = l_max < Lmx ? (Lmx+2)*(Lmx+2)*8 : (l_max+2)*(l_max+2)*8;
		j = Nx*Nphi < j ? j : Nx*Nphi;
		if ((f=(double*)calloc((unsigned)j, sizeof *f))==
		    (double*)0)
			goto nomem_w;
		j = l_max < Lmx ? Lmx : l_max;
		Nmx = Nx > 2*j+1 ? Nx : 2*j+1;
		Nmphi = 2 * Nx;
		/* read alm from ASCII string */
		/*----------------------------*/
		read_alm_str (alm_str, f, Nmphi, Lmx);
	    }
	    if (dipole_flag || monopole) {
		    /* do it not to calculate twice */
		    if (Lmn == 0)  Lmn = 2;
		    if (Lmn == 1)  Lmn = 2;
		    if (Lmx == 1)  l_max = 2;
	    }

	    /* take more memory if we need */
	    if (l_max == -1 && Lmx > 0)  l_max = Lmx;
	    j = l_max < Lmx ? (Lmx+2)*(Lmx+2)*8 : (l_max+2)*(l_max+2)*8;
	    if (Nmx*Nmphi < j || Nx>Nmx || Nphi>Nmphi) {
		/* increase basic array */
		i = Nx < Nmx ? Nmx : Nx;
		j = i*Nphi  < j ? j : i*Nphi;
		j = i*Nmphi < j ? j : i*Nmphi;
		f = (double *)realloc(f, (unsigned)j * sizeof *f);
		if (f == (double *)0)  goto nomem_w;
		/* after 'realloc' fill with zeros */
		for(i=Nmx*Nmphi; i<j; i++)  f[i] = 0.0;

		if (Nx*Nphi > Nmx*Nmphi) {
		    j = -10;  /* flag: we changed arrays with lmax */
		} else {
		    if (Nmx < Nx)     Nmx = Nx;
		    if (Nmphi < Nphi) Nmphi = Nphi;
		}

	    }

	     /* reconfiguration of a_lm - array */
	     if ( (l_max > 0 && Lmx != l_max && Lmx > l_max) || j== -10 ) {
		  /* and cut Nmphi down to Nphi and Nmx to Nx */
		  alm2alm(&f, &Lmx, &Nmphi, l_max, (j == -10 ? -Nphi : Nphi));
		  Nmx = Nx;
	     }
	}
	Jmx = Nmx/2; /* maximum lmax */

	if (Lmx < 1) {
	    PR(ER, "%s: Lmax is not set. See flag '-lmax'.\n", myname);
	    exit(1);
	}

	/* create file and save a_lm data */
	if ( *alm_out_file != '\0' ) {
	    FILE *fp;

	    if (glesp_comments)
		PR(ER, "The file '%s' of a_{lm} is created (lmax=%d)...",
			 alm_out_file, Lmx);
	    fp = fopen(alm_out_file, "w");
	    if (fp == (FILE *)0) {
		PR(ER,"\n%s: can not create this file...\n", myname);
	    } else {
		if (outp_fflag)
		    ff_write_alm(f, Nmphi, Lmx, Lmx, fp);
		else
		    out_glesp_alm(f, Lmx, Nmphi, Lmx, fp);
		fclose (fp);
		if (glesp_comments)
		    PR(ER, "Done.\n");
	    }

	    if (nomap) {
		PR(ER, "\nNo map calculations.\n");
		exit(0);
	    }
	}

	/* take memory for description arrays */
	if ((Nph=(int *)calloc((unsigned)Nmx, sizeof *Nph))==(int *)0)
		goto nomem_w;
	if ((x=(double *)calloc((unsigned)Nmx, sizeof *x))==(double *)0)
		goto nomem_w;
	if ((pw=(double *)calloc((unsigned)Nmx, sizeof *pw)) ==(double*)0)
		goto nomem_w;


       /*------------------------*/
       /* select coordinate grid */
       /*------------------------*/
       if (lx_add < -1000) { /* construct GLESP grid from polar cap */
	    Kx = 0; j = -(lx_add + 1000);  /* number of pixels near pole */
       } else if (lx_add == 'P') {
	    Kx = 2;
	    j = lx_add;
       } else j = lx_add;
       coord(Kx, j, x1, x2, phi2, x, pw, Nx, Nph, Nphi);

       /* memory for temporal arrays */
       if ((cf = (double *)calloc((unsigned)Nmphi, sizeof *cf)) ==
	  (double *)0) {
no_mem_temp:
	     PR(ER, "%s: no memory for temporal arrays...\n", myname);
	     exit(1);
	}
	if ((wsave=(double*)calloc((unsigned)(2*Nmphi+15),sizeof *wsave))==
	    (double*)0)
		goto no_mem_temp;
	if ((apm = (double *)calloc((unsigned)((Jmx+1)*Nmx), sizeof *apm))
	   == (double *)0)
	     goto no_mem_temp;
	if ((bpm = (double *)calloc((unsigned)((Jmx+1)*Nmx), sizeof *bpm))
	   == (double *)0)
	     goto no_mem_temp;
      /*
       * Simulate map:
       *
       ****************************************
       *  here f(i,j) is the map of signal,   *
       *  i=1:Nx - cos(theta), j=1:2*Nx - phi *
       ****************************************
       * subroutine takes a_lm[][] from f[][]
       * and writes map[][] to f[][]
       */

       if ( dipole_flag ) {
	    Lmx = -Lmx;
	    if ( dipole_flag < 0)
		 legendre_calc_scheme= 0;
	    else
		 legendre_calc_scheme= 100;
       }
       if (monopole)
	    alm00 = f[0] / sqrt(4. * M_PI);

       /*
	*   map preparation, alm_rd -- input of alm,
	*   signal -- transforms alm--to--map,
	*   alm_tst -- transforms back map to alm
	*/
       signal_(&Lmn,&Lmx,&Mmn,&Mmx,&Nx,&eps,&phi2,
	    &Nmx,&Nmphi, x,pw,f,Nph,
	    wsave,cf,apm,bpm,
	    &fft_flag, &sym_grid, &legendre_calc_scheme);

       /************************/
       /* derivative transform */
       /************************/
       if (der2_flag == 1) { /* 1st derivative */
	    register int ii;
	    for (ii=i=0; i<Nx; i++, ii+=Nmphi) {
		 double sth = sqrt(1. - x[i]*x[i]);
		 for(j=0; j<Nph[i]; j++)  {
		      f[ii+j] = (f[ii+j] + alm00) / sth;
		 }
	    }
       } else if (der2_flag == 2) { /* 2nd derivative */
	    register int ii;
	    for (ii=i=0; i<Nx; i++, ii+=Nmphi) {
		 double sth2 = 1. - x[i]*x[i];
		 for(j=0; j<Nph[i]; j++)  {
		      f[ii+j] = (f[ii+j] + alm00) / sth2;
		 }
	    }
       }
       /* add monopole */
       if (monopole) {
	   for (i=0; i<Nx; i++) {
		register int ll = Nmphi * i;
		for(j=0; j<Nph[i]; j++)  f[ll+j] += alm00;
	   }
       }

       /* show minimum and maximum values */
       {
	    double v_min= 1.e+30, v_max= -1.e+30;
	    register int ii, n;
	    for (ii=i=0; i<Nx; i++,ii+=Nmphi) {
		 n = Nph[i];
		 for(j=0; j<n; j++)  {
		      double v = f[ii+j];
		      if (v > v_max) v_max = v;
		      if (v < v_min) v_min = v;
		 }
	    }
	    if (glesp_comments)
		PR(ER,"v_min=%15.8e, v_max=%15.8e\n", v_min, v_max);
       }
       /****************/
       /*  output  map */
       /****************/
       {
	    char x_type[40], grid_type[40];

	    /* select grid output keyword values */
	    strcpy(x_type, "COS(THETA)");
	    strcpy(grid_type, "PHI(X)+LEGENDRE");

	    /* map output */
	    if (outp_fflag) /* F format output */
		    ff_write_ext2_map(f, x, Nx, Nmphi, Nph, fp_out,
			   x_type, grid_type);
	    else    /* FITS format output */
		    outp_glesp_map(f, x, Nx, Nmphi, Nph, fp_out,
			   x_type, grid_type);
       }
       /*******************/
       /* debugging check */
       /*******************/
       if (debug_flag) {
	    sum = 0.0;
	    for (i=0; i<Nx; i++) {
		 register int ll = Nmphi * i;
		 for(j=0; j<Nmphi; j++) {
		    if (pw[i] > 0) {
			    sm = f[ll+j];
			    sum += sm * sm / pw[i];
			    f[ll+j] *= pw[i];
		    }
		 }
	    }
	    fflush(stdout);
	    if (glesp_comments)
		PR(ER, "sum_signal_tst=%19.12e\n", sum);

	    if ( dipole_flag ) {
		 Lmx = -Lmx;
		 if ( dipole_flag < 0)
		      legendre_calc_scheme= 0;
		 else
		      legendre_calc_scheme= 100;
	    }

	    /* calculate a_{lm} and write to f(l, m) */
	    a_lm_(&Lmn,&Lmx,&Mmn,&Mmx,&Nx,&Vnyq,&eps,&phi2,
		&Nmx,&Nmphi, x,pw,f,Nph,
		wsave,cf,apm,bpm,
		&fft_flag, &sym_grid, &legendre_calc_scheme);

	    sum = 0.0;
	    for (i=0; i<Nx; i++) {
		 register int ll = Nmphi * i;
		 for(j=0; j<Nmphi; j++) {
		    sm = f[ll+j];
		    sum += sm * sm;
		 }
	    }
	    fflush(stdout);
	    if (glesp_comments)
		PR(ER, "sum_alm_tst=%15.8e\n", sum);
	    goto label_100;
       }
    }  else {
       /***********************/
       /* read map: f(x, phi) */
       /***********************/
       register int ii;
       i = test_glesp_map(map_file);

       if (i != 0 && i != 1 && i != 3) {
	   PR(ER, "%s: file '%s' is not a GLESP map file.\n",
		myname, map_file);
	   exit(1);
       }


       if (i == 0)  { /* F-format map */
	    /* detect file type, read map and check size of arrays     */
	    /* if arrays are short, then free old and catch new memory */
	    /* It returns type of grid in input file                   */
	    i = read_map2(&f, &Nx, &Nphi, &Nph, &x, map_file);

	    /* free memory for working arrays */
	    if (pw    != (double *)0)  free((void *)pw);
	    if (wsave != (double *)0)  free((void *)wsave);

	    /* select Nmx according Nx and Nphi */
	    /* (we need Nmx for FORTRAN calls)  */
	    Nmx   = Nx;
	    Nmphi = Nphi;

	    /* catch memory for working arrays */
	    if ((pw = (double *)calloc((unsigned)Nx, sizeof *pw)) ==
		(double*)0)
		    goto nomem_w;

	    /*------------------------------------------*/
	    /* select new coordinate grid if we need it */
	    /*------------------------------------------*/
	    if(!((i==GRID_LEGENDRE ||i== GRID_PHI_X_LEGENDRE) &&Kx==1) &&
		!(i==GRID_UNIFORM_COS && Kx == -1)  ) {

		 /* take memory for temporal X-argument array */
		 if((xbuf=(double*)calloc((unsigned)Nx, sizeof *xbuf)) ==
		     (double*)0)
			 goto nomem_w;

		 coord(Kx, lx_add, x1, x2, phi2,
			 xbuf, pw, Nx, Nph, Nphi);
	    } else {
		 /* calculate weights 'pw' */
		 weights(Kx, x1, x2, phi2, x, pw, Nx, Nph);
	    }

	    /*******************************************/
	    /* interpolate map according to grid       */
	    /* and transpose it: f(phi,x) -> f(x, phi) */
	    /* also involve Nmx-dimension parameter    */
	    /* to use 'f'-array in FORTRAN calls       */
	    /*******************************************/
	    recalc_map2(&f, Nx, Nphi, Nmx, Nmphi, Nph, x, xbuf, i, Kx);
	    if ( xbuf != (double *)0) {
	       /* change arguments to present values and free memory */
	       PR(ER,
		  "%s: I have changed a grid according your needs...\n",
		       myname);
	       for(i=0; i<Nx; i++)  x[i] = xbuf[i];
	       free((void *)xbuf);
	    }
       } else {
	    /* read FITS format map */
	    extern void in_glesp_1map(double **, double **, int *, int *,
			      int **, char *);
	    if (i==3)
		in_glesp_1map(&f, &x, &Nx, &Nphi, &Nph, map_file);
	    else
		in_glesp_map(&f, &x, &Nx, &Nphi, &Nph, map_file);

	    /* catch memory for working arrays */
	    if ((pw = (double *)calloc((unsigned)Nx, sizeof *pw)) ==
		(double*)0)
		    goto nomem_w;

	    /* calculate weights 'pw' */
	    weights(Kx, x1, x2, phi2, x, pw, Nx, Nph);
	    Nmx   = Nx;
	    Nmphi = Nphi;
       }

       /* control parameters */
       if (Nmx < 2*l_max+1) {
	   PR(ER,"\n%s:\nyour map resolution parameters are (%d,%d).\n",
		 myname, Nx, Nphi);
	   PR(ER,"Thus, the maximum L should be Lmax <= (%d-1)/2=%d.\n",
		   Nx, (Nx-1)/2);
	   PR(ER,"But you use Lmax=%d... Please, set the proper Lmax.\n",
		   l_max);
	   exit(2);
       }

       if (Nmphi < 4*l_max+1) {
	   PR(ER,"\n%s:\nyour map resolution parameters are (%d,%d).\n",
		 myname, Nx, Nphi);
	   PR(ER,"Thus, the maximum L should be Lmax <= (%d-1)/2=%d, ",
		   Nx, (Nx-1)/2);
	   PR(ER,"and L<=(%d-1)/4=%d by phi-direction.\n",
		   Nphi,(Nphi-1)/4);
	   PR(ER,"You use Lmax=%d... ", l_max);
	   PR(ER,"Please, use Lmax corresponding to rules.\n");
	   exit(2);
       }

       for (ii=i=0; i<Nx; i++, ii+=Nmphi) {
	    register int n = Nph[i];
	    for(j=0; j<n; j++) {
	       if (pw[i] > 0) {
		       f[ii+j] *= pw[i];
	       }
	    }
      }
      if ( cosbell_flag ) {
	    int i_c, j_c;
	    double dt1, dt2;
	    double wss;

	    /* detect position of area */
	    area_pos(f, Nx, Nmphi, Nph, x,  &i_c, &j_c);

	    /* detect size of area */
	    area_size(f, Nx, Nmphi, Nph, x, i_c, j_c, &dt1, &dt2);

	    /* calc cosbell and mutiply by array */
	    cosbell(f, Nx, Nmphi, Nph, x, i_c, j_c, dt1, dt2, &wss, 1);
       }

       /* take memory for temporal arrays */
       if ((wsave=(double*)calloc((unsigned)(2*Nmphi+15),sizeof *wsave))
	   == (double*)0)
	       goto nomem_w;

       if ((cf = (double *)calloc((unsigned)Nmphi, sizeof *cf)) ==
	  (double *)0)
	       goto nomem_w;
       Jmx=(Nmx+1)/2;  /* Jmx - max value of Lmx    */
       if ((apm = (double *)calloc((unsigned)((Jmx+1)*Nmx), sizeof *apm))
	  == (double *)0)
	       goto nomem_w;
       if ((bpm = (double *)calloc((unsigned)((Jmx+1)*Nmx), sizeof *bpm))
	  == (double *)0)
	       goto nomem_w;

       if (debug_sum_flag) {
	    sum=0.0;
	    for (i=0; i<Nx; i++) {
		 register int ll = Nmphi * i;
		 for(j=0; j<Nphi; j++) {
		    if (pw[i] > 0) {
			 sm = f[ll+j];
			 sum += sm * sm / pw[i];
		    }
		 }
	    }
	    if (glesp_comments)
		PR(ER,"Sum=%15.8e, Lmx=%d, Nx=%d, Nphi=%d\n",sum,Lmx,Nx,Nphi);
       }

       if(sin2_flag) {
	    for (i=0; i<Nx; i++) {
		 register int ii = Nmphi * i;
		 for(j=0; j<Nph[i]; j++) {
		    f[ii+j] *= (1-x[i]*x[i]);
		 }
	    }
       }
       /* calculate a_{lm} and write to f(l, m) */
       if ( dipole_flag ) {
	    Lmx = -Lmx;
	    if ( dipole_flag < 0)
		 legendre_calc_scheme= 0;
	    else
		 legendre_calc_scheme= 100;
       }
       /* calc monopole */
       if (monopole)  {
	   for (i=0; i<Nx; i++) {
	       register int ii = Nmphi * i;
	       for(j=0; j<Nph[i]; j++)  alm00 += f[ii+j];
	   }
	   alm00 /= sqrt(4. * M_PI);
       }


       a_lm_(&Lmn,&Lmx,&Mmn,&Mmx,&Nx,&Vnyq,&eps,&phi2,
	    &Nmx,&Nmphi, x,pw,f,Nph,
	    wsave,cf,apm,bpm,
	    &fft_flag, &sym_grid, &legendre_calc_scheme);

       if (monopole)  f[0] = alm00;


label_100:
       /* array for C_l   */
       if (cl != (double *)0) free((void *)cl);
       if ((cl = (double *)calloc((unsigned)Lmx+2, sizeof *cl)) ==
	   (double *)0)
		goto nomem_w;


       /************************/
       /*  output loop for alm */
       /************************/
       {
	    double *clp = (double *)0;

	    /* create file and save a_lm data */
	    if ( *alm_out_file != '\0' ) {
		FILE *fp;

		if (debug_flag)
			 strcat(alm_out_file, "_2");
		if (glesp_comments)
		    PR(ER, "%s: file '%s' of a_{lm} is created (lmax=%d)...",
			 myname, alm_out_file, Lmx);
		fp = fopen(alm_out_file, "w");
		if (fp == (FILE *)0) {
		    PR(ER,"\n%s: can not create this file...\n", myname);
		} else {
		    if (outp_fflag)
			ff_write_alm(f, Nmphi, Lmx, Lmx, fp);
		    else
			out_glesp_alm(f, Lmx, Nmphi, Lmx, fp);
		    fclose (fp);
		    if (glesp_comments)
			PR(ER, "Done.\n");
		}
	    }

	    /* calculate Cl/Dl */
	    if (cl_output_flag) {
		alm2cl (cl, Lmx, f, Nmphi);
		print_cl_table (fp_out, cl, clp, 0, Lmx);
	    }
	    if (dl_output_flag) {
		FILE *fp = fopen(dl_out_file, "w");
		if (fp == (FILE *)0) {
		    PR(ER,"\n%s: can not create file '%s' for Dl...\n",
			myname, dl_out_file);
		} else {
		    alm2dl (cl, Lmx, f, Nmphi);
		    /* print Dl     */
		    print_cl_table (fp, cl, clp, 0, Lmx);
		    fclose (fp);
		}
	    }

       }

       if (debug_sum_flag) {
	    sum=0.00;
	    for(i=0; i<=Lmx; i++) {
		 register int ll = Nmphi * i, m;
		 for (m=0; m<=i; m++) {
		    double f1=f[ll+m], f2=f[ll+Lmx+m+1];
		    double sm = f1*f1 + f2*f2;
		    sum += sm;
		 }
	    }
	    if (glesp_comments)
		PR(ER, "sum_fin=%15.8e\n", sum);
       }
     }


    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);

    if (cl    != (double *)0)  free((void *)cl);
    if (cpl   != (double *)0)  free((void *)cpl);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
	register int i;
	myname = argv[0];

	for(i=1; i<argc; i++) {
	    switch(argv[i][0]) {
	    case '-':
		switch(argv[i][1]) {
		case '1':
			der2_flag = 1;
			break;
		case '2':
			der2_flag = 2;
			break;
		case 'a':
			if (strncmp(argv[i]+1, "alm", 3) == 0) {
			    strcpy(alm_str, argv[++i]);

			    alm_input = 2; /* alm from string */
			    /* <0,transforms a_lm to map */
			    calc_flag = -1;
			} else if (strncmp(argv[i]+1, "ao", 2) == 0) {
			    if (strncmp(argv[i]+2, "of", 2) == 0) {
				outp_fflag = 1; /* output in F-format */
			    }
			    strcpy(alm_out_file, argv[++i]);
			} else if (strncmp(argv[i]+1, "ai", 2) == 0) {
			    strcpy(file_alm, argv[++i]);
			    alm_input = 1; /* alm from file */
			    /* <0,transforms a_lm to map */
			    calc_flag = -1;
			} else {
			    /* < 0,transforms a_lm to map, else back  */
			    so_c_flag = -1;
			}
			break;
		case 'b':
			if (strncmp(argv[i]+1, "bitp", 4) == 0) {
				/* set bitpix for output map */
				bitpix_map = atoi(argv[++i]);
			}
			break;
		case 'c':
		case 'C':
			if (strcmp(argv[i]+1, "cb") == 0) {
			      cosbell_flag = 1;
			} else {
			      /* read Cl-file */
			      strcpy(file_C_l, argv[++i]);
			      Cl_input = 1;
			      /* <0,transforms a_lm to map */
			      calc_flag = -1;
			}
			break;
		case 'd':
			if (strncmp(argv[i]+1, "di", 2) == 0) {
			      int l = strlen(argv[i]);
			      if (argv[i][l-1] == '+')
				    dipole_flag = 1;
			      else if (argv[i][l-1] == '-')
				    dipole_flag = 0;
			      else
				    dipole_flag = -1;
			} else if (strcmp(argv[i]+1, "dl") == 0) {
				strcpy(file_C_l, argv[++i]);
				Cl_input = 2;
				/* <0,transforms a_lm to map */
				calc_flag = -1;

			} else if (strcmp(argv[i]+1, "double") == 0) {
				bitpix_map = DOUBLEPIX;

			} else {
				/* debugging flags */
				debug_flag = 1;
				debug_sum_flag = 1;
			}
			break;
		case 'D':
			strcpy(file_C_l, argv[++i]);
			Cl_input = 2;
			/* <0,transforms a_lm to map */
			calc_flag = -1;
			break;
		case 'f':
			switch(argv[i][2]) {
			case 'a':
				strcpy(file_alm, argv[++i]);
				alm_input = 1; /* alm from file */
				/* <0,transforms a_lm to map */
				calc_flag = -1;
				break;
			case 'c':
				strcpy(file_par, argv[++i]);
				break;
			case 'f':
				/* real Fourier transform, switch off FFT */
				fft_flag = 0;
				break;
			case 'l':
				if (strncmp(argv[i]+1, "flo",4) == 0)
				     bitpix_map = FLOATPIX;
				break;
			case 'w':
				fwhm = atof(argv[++i]);
				if (fwhm < 1.e-8) {
				  PR(ER, "%s: you can not smooth data with ",
					     myname);
				  PR(ER, "such beam size: '-fw %g'\n", fwhm);
				  PR(ER, "Please, set positive value.\n");
				  exit(1);
				}
				if (argv[i-1][strlen(argv[i-1])-1] == '2') {
				    /* last symbol after 'fw' is '2' */
				    fwhm *= M_SQRT2;
				}
				break;
			default :
				PR(ER, "%s: Illegal flag '%s'.\n",
					myname, argv[i]);
				usage();
				break;
			}
			break;
		case 'g':
			if (strncmp(argv[i]+1, "gr", 2) == 0) {
			  if (argv[i][3] != '\0') {
			    int c = argv[i][3];
			    switch (c) {
			    case 'a': case 'A': lx_add = -10; break;
			    case 'n': case 'N': lx_add =   0; break;
			    case 's': case 'S': lx_add =   4; break;
			    case 'i': goto lab1;
			    default :
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			    }
			  } else {
lab1:
			    i++;
			    if (argv[i][0] == 'a' || argv[i][0] == 'A') {
				/* quasi-equal area pixel */
				lx_add = -10;  /* old scheme */
			    } else if (argv[i][0]=='n' || argv[i][0]=='N') {
				/* equal pixel number in rings */
				lx_add = 0; /* pin 0 */
			    } else if (argv[i][0]=='s' || argv[i][0]=='S') {
				/* equal pixel number in rings */
				lx_add = 4; /* pin 4 */
			    }
			  }
			} else {
			    /* <0 - dx=2/(Nx-1), >0 - special grid    */
			    PR(ER, "%s: flag '%s' doesn't work now.\n",
				    myname, argv[i]);
			    usage();
			    so_c_grid = -1;
			}
			break;
		case 'h':
			if (strncmp(argv[i]+1,"h2",2) == 0)
			    usage2();
			else
			    usage();
		case 'i':
			switch(argv[i][2]) {
			default :
				/* input file for F(x,phi) or a_{lm} */
				strcpy(so_i_file, argv[++i]);
				break;
			}
			break;
		case 'l': /* index of input harmonique for f(x,phi) */
			switch(argv[i][2]) {
			case 'm':
				if (strncmp(argv[i]+2,"mi",2) == 0)
				    so_l_min = atoi(argv[++i]); /* Lmin */
				else if (strncmp(argv[i]+2,"ma",2) == 0)
				    so_l_max = atoi(argv[++i]); /* Lmax */
				else
				    so_l_min = atoi(argv[++i]); /* Lmin */
				break;
			default : /* ind of input harmonique for f(x,phi) */
				so_l_c = atoi(argv[++i]);
				so_l_min = so_l_max = so_l_c;
				break;
			}
			break;
		case 'L':
			if ('0' <= argv[i][2] && argv[i][2] <= '9')
			    so_l_max = so_l_min = atoi(argv[i]+2);
			else if (strncmp(argv[i]+2,"mi",2) == 0)
			    so_l_min = atoi(argv[++i]); /* Lmin */
			else if (strncmp(argv[i]+2,"ma",2) == 0)
			    so_l_max = atoi(argv[++i]); /* Lmax */
			else {
			    /* Lmin = Lmax */
			    so_l_max = so_l_min = atoi(argv[++i]);
			}
			if (so_l_min > 1)
			     dipole_flag = 0;
			break;
		case 'm': /* index of input harmonique for f(x,phi) */
			switch(argv[i][2]) {
			case 'a':
				if (strncmp(argv[i]+1,"map",3) == 0) {
					strcpy(map_file, argv[++i]);
					map_input = 1;
					calc_flag = 1;
				} else {
				    PR(ER,"%s: illegal flag '%s'\n",
					myname, argv[i]);
				    usage();
				}
				break;
			case 'm':
				if (strncmp(argv[i]+2,"mi",2) == 0)
				    so_m_min = atoi(argv[++i]); /* Mmin */
				else if (strncmp(argv[i]+2,"ma",2) == 0)
				    so_m_max = atoi(argv[++i]); /* Mmax */
				else
				    so_m_min = atoi(argv[++i]); /* Mmin */
				break;
			default : /* ind of input harmonique for f(x,phi) */
				so_m_c = atoi(argv[++i]);
				so_m_min = so_m_max = so_m_c;
				break;
			}
			break;
		case 'M':
			if (strncmp(argv[i]+2,"mi",2) == 0)
			    so_m_min = atoi(argv[++i]); /* Mmin */
			else if (strncmp(argv[i]+2,"ma",2) == 0)
			    so_m_max = atoi(argv[++i]); /* Mmax */
			else {
			    /* Mmin = Mmax */
			    so_m_max = so_m_min = atoi(argv[++i]);
			}
			break;
		case 'n':
			switch(argv[i][2]) {
			case 'x': /* number of points along cos(theta) */
				so_num_x = atoi(argv[++i]);
				break;
			case 't': /* number of points along cos(theta) */
				so_num_x = atoi(argv[++i]);
				break;
			case 'o':
				if (strncmp(argv[i]+1, "nom", 3)==0){
				    nomap = 1; /* no anisotropy calc */
				}
				break;
			case 'p': /* number of points along phi */
				so_num_phi = atoi(argv[++i]);
				break;
			default :
				PR(ER, "%s: Illegal flag '%s'.\n",
					myname, argv[i]);
				usage();
				break;
			}
			break;
		case 'N':
			NMX = atoi(argv[++i]);
			break;
		case 'o':
			{
			    /* output file for a_{lm} or F(x,phi) */
			    if (strncmp(argv[i]+1, "of", 2) == 0) {
				outp_fflag = 1; /* output in F-format */
				strcpy(so_o_file, argv[++i]);
			    } else if (strcmp("-oc", argv[i]) == 0 ) {
				cl_output_flag = 2;
				strcpy(so_o_file, argv[++i]);
			    } else if (strcmp("-od", argv[i]) == 0 ) {
				dl_output_flag = 1;
				strcpy(dl_out_file, argv[++i]);
			    } else {
				strcpy(so_o_file, argv[++i]);
			    }
			}
			break;
		case 'p':
			if (strncmp(argv[i]+1, "pin", 3) == 0) {
				/* phi increment number :       */
				/* addition to Nphi numbers for */
				/* 'coord'-function             */
				int v = atoi(argv[++i]);
				if (argv[i][0] == 'o')   v = -10;
				/* fill arrays by 0 for Fourier transform */
				else if (argv[i][0]=='p' || argv[i][0]=='P')
				     lx_add = 'P';
				else if (v == -1)    fill_0 = 1;
				else if (v < -1)
				     lx_add = -1000 + v;
				else lx_add = v;
			}
			break;
		case 's':
			if (strncmp(argv[i]+1, "si", 2) == 0)
			      sin2_flag = 1;
			else if (i < argc-1) {
			/* if flag=1, then direct calculations, */
			/* else from largest M to lower         */
			    if (argv[i+1][0] != '-') {
				i++;
				if(argv[i][0] == 'L' || argv[i][0] == '1') {
					legendre_calc_scheme = 1;
				} else
					legendre_calc_scheme = -1;
			    } else {
				legendre_calc_scheme = !legendre_calc_scheme;
			    }
			} else /* flag '-s' is the last parameter */
				legendre_calc_scheme = !legendre_calc_scheme;
			break;
		case 'r': /* random seed */
			random_seed = atoi(argv[++i]);
			seed_flag = 1;
			break;
		case 't':
			if (strcmp(argv[i]+1, "tmp") == 0) {
				strcpy(tmpcat, argv[++i]);
			} else {
				/* flag to keep cos(thetha)/restore grid */
				/* for theta                             */
				theta_grid=0;
			}
			break;
		case 'v': /* version */
			PR(ER, "%s: GLESP 2.0, version 1.0c (30.11.2011)\n",
			    myname);
			exit(0);
		case 'w': /* beam size */
			fwhm = atof(argv[++i]);
			break;
		default :
			PR(ER, "%s: Illegal flag '%s'.\n", myname, argv[i]);
			usage();
			break;
		}
		break;
	    case '+':
		switch(argv[i][1]) {
		case 'a': /* > 0,transforms a_lm to map, else back  */
			so_c_flag = 1;
			break;
		case 'f':
			/* switch on FFT (defautlt) */
			fft_flag = 1;
			break;
		case 'g': /* <0 - dx=2/(Nx-1), >0 - special grid    */
			so_c_grid = 1;
			break;
		case 't':
			/* flag to keep cos(thetha)/restore grid for theta */
			theta_grid=1;
			break;
		case 'h':
			usage();
		default :
			PR(ER, "%s: Illegal flag '%s'.\n", myname, argv[i]);
			usage();
			break;
		}
		break;
	    default :
		PR(ER, "%s: illegal flag '%s'.\n", myname, argv[i]);
		usage();
		break;
	    }
	}
	if (alm_input == 2 && so_l_max < 0)
	      so_l_max = get_lmax_str(alm_str);

	if ( so_l_max > 0) {
	    if (so_num_x < 0)  {
		if (so_l_max <= 50)  so_num_x = 201;
		else if (so_l_max <= 100)  so_num_x = 401;
		else so_num_x =  so_l_max * 3;
	    }
	    if (so_num_phi < 0)  so_num_phi = 2 * so_num_x;
	}

	if ( so_l_min == 0) {
		monopole = 1;
	}
	if (so_l_min == 1) {
		dipole_flag = 1;
	}
	if (so_l_min > 1)  {dipole_flag = 0; monopole = 0;}

	if (dipole_flag) {
	    if (so_l_min < 0)  so_l_min = 1;
	    if (so_l_max < 0)  so_l_max = 1;
	}
	if ( so_l_max < so_l_min)  so_l_max = so_l_min;

	/* if no map output then a_lm-s data file should be */
	if (nomap) {
	    PR(ER, "\n");
	    if (*alm_out_file == '\0')
		PR(ER, "%s: name for output a_lm-file is absent...\n", myname);
	    if (*file_C_l == '\0')
		PR(ER, "%s: name for input C(l)-file is absent...\n", myname);
	    if (*alm_out_file == '\0' || *file_C_l == '\0') {
		PR(ER, "\n");
PR(ER, "For '-nomap' flag, the command string format should look like this:\n");
PR(ER, "\tcl2map -r N -nx NX -np NP -lmax LM -nomap -ao ALM.FTS -cl CL.DAT\n");
		PR(ER, "\n");
		exit(1);
	    }
	}

	return;
}

void usage()
{
PR(ER,"usage: %s [-fc cll_file] [-l l-harm] [-m m-harm] [-/+g (grid:eq./spec.)]\n",
myname);
PR(ER,"[-i inp_file] [-o outp_file] [-/+a (map->alm/alm->map)]\n");
PR(ER,"[-lmin lmin] [-lmax lmax] [-mmin mmin] [-mmax mmax] [-Cl file_with_Cl]\n");
PR(ER,"[-fw{2} fwhm(')] [-pin phi_incr]\n");
PR(ER,"[-/+t(cos(thetha) grid)] [-ext(no extension output)] [-nx nx] [-np nphi]\n");
PR(ER,"[-ip(phase input)] [-r random_seed] [+/-fft] [-s {LM01}(polyn.calc.scheme)]\n");
PR(ER,"[-falm file_with_alm(format: l,m,re,im] [-alm l,m,re,im] [-dipole]\n");
PR(ER,"[-map map_file] [-tmp tmpcat] [-ao output_alm_file] [-bitp map_pix]\n");
PR(ER,"[-cb(cos-bell mutipl.)] [-{double,float}] [-nomap] [-v(version)]\n");
exit(2);
}


void usage2()
{
PR(ST, "usage: %s [FLAGS]\n", myname);
PR(ST, "-----------------------------------------------------------------\n");
PR(ST,"-fc cll_file\n");
PR(ST,"-/+g (grid:eq./spec.)\n");
PR(ST,"-i inp_file \n");
PR(ST,"-/+a (map->alm/alm->map)\n");
PR(ST,"-/+t(cos(thetha) grid)\n");
PR(ST,"-bitp map_pix\n");
PR(ST,"-ext(no extension output)\n");
PR(ST,"-cb(cos-bell mutipl.)\n");
PR(ST,"-ip(phase input)\n");
//PR(ST,"+/-fft\n");
PR(ST,"-alm l,m,re,im - string of input a_lm-s to synthesyze a map\n");
PR(ST,"-ao output_alm_file - file to save a_lm-s when calculation of Cl\n");
PR(ST,"-Cl Cl-file - use Cl-file to simulate random map\n");
PR(ST,"-Dl Dl-file - use Dl-file to simulate random map (see '-r','-ao')\n");
PR(ST,"-dipole - swithes off calculation of dipole and monopole\n");
PR(ST,"-double - output file in 64-bit float format (default) (see '-float')\n");
PR(ST,"-falm alm_file(format: l,m,re,im) - a_lm-s input file to prepare a map\n");
PR(ST,"-float - output file in 32-bit float format (see '-double')\n");
PR(ST,"-fw{2} fwhm(') - smooth a map with gaussian beam of 'fwhm' arcmin\n");
PR(ST,"-l l-harm - selects only l-multipole to prepare map or to extract\n");
PR(ST,"-lmin lmin - sets minimum l-multipole number to calculate\n");
PR(ST,"-lmax lmax - sets maximum l-multipole number to calculate\n");
PR(ST, "-LN - same to '-lmin N -lmax N'\n");
PR(ST,"-m m-harm - selects only m-multipole to prepare map or to extract\n");
PR(ST,"-map map_file - file with map to be used for harmonical expansion\n");
PR(ST,"-mmin mmin - sets minimum m-multipole number to calculate\n");
PR(ST,"-mmax mmax - sets maximum m-multipole number to calculate\n");
PR(ST, "-MN - same to '-mmin N -mmax N'\n");
PR(ST,"-nx nx - resolution parameter, number of rings (nx=2lmax+1)\n");
PR(ST,"-np nphi - resolution parameter, number of pixels on equator (np=2nx)\n");
PR(ST,"-nomap - no map calcultion, just generation of random a_lms (see '-Cl','-r')\n");
PR(ST,"-o outp_file - output file: map or cl-spectrum\n");
PR(ST,"-pin phi_incr - increment for changing pixel number per ring \n");
PR(ST,"-r random_seed - random seed for random map simulation\n");
PR(ST,"-s {LM01} - sets a scheme for Legendre's polynomial calculation\n");
PR(ST,"-tmp tmpcat - change temporal directory (default: /tmp)\n");
PR(ST,"-v - prints version of '%s'\n", myname);
exit(0);
}


