#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>

/*
 * some local functions for GLESP package
 *      Oleg Verkhodanov, 2003-2004
 */
#include "glesp.h"

#define PR      fprintf
#define ER      stderr

extern char tmpcat[161];
extern char *myname;

void short_form(form)
char *form;
{
	register char *cs;

	/* remove ' ', '\t', '\n' in name: before and after */
	cs = form;
	while ( *cs && (*cs == ' ' || *cs == '\t' || *cs == '\n')) {
		int i = 0;
		while ( (cs[i] = cs[i+1]) )  i++;
		/* if ( *cs ) cs++; */
	}
	cs = form + strlen(form) - 1;
	while ( *cs == ' ' || *cs == '\t' || *cs == '\n' )  cs--;
	cs++; *cs = '\0';

	/* remove double symbols ' ', '\t' in name */
	cs = form;
	while ( *cs ) {
	    if ((*cs==' ' || *cs=='\t') && (cs[1]==' ' || cs[1]=='\t')) {
		int i = 0;
		while ( (cs[i] = cs[i+1]) )  i++;
	    } else
		cs++;
	}
	return;
}

char *skip_wd(n, str, csq)
int n;             /* number (from zero) of a field to be read */
char *str,         /* input string                */
     *csq;         /* string with requested field */
{
	register char *cs, *rs, *fs;
	register int i;
	cs = str;
	for(i=0; i<n; i++) {
		if ( *cs == '\0' || *cs == '\n') {
			*csq = '\0';
			return cs;
		}
		/* miss emty fields from begining */
		while (*cs == ' ' || *cs == '\t'||*cs == ';'||*cs == ',')
		      cs++;
		/* read till separator */
		while (*cs && *cs != ' ' && *cs != '\t' && *cs != '\n'
		   && *cs != ';' && *cs != ',')
		      cs++;
	}
	if ( *cs == '\0' || *cs == '\n') {
		*csq = '\0';
		return cs;
	}
	/* miss empty fields and separators after current field */
	while (*cs == ' ' || *cs == '\t' || *cs == ';'||*cs == ',')   cs++;
	if ( *cs == '\0' || *cs == '\n') {
		*csq = '\0';
		return cs;
	}
	rs = cs;
	fs = csq;
	while (*rs && *rs != ' ' && *rs != '\t' && *rs != '\n'
	      && *rs != ';' && *rs != ',') {
		*fs = *rs;
		fs++; rs++;
	}
	*fs = '\0';
	return cs;
}

/* return number of parameters in a string */
int  n_skip_wd(str)
char *str;         /* input string                */
{
	register char *cs;
	register int i=0;
	cs = str;
	while(*cs) {
		/* miss emty fields from begining */
		while (*cs ==' ' || *cs =='\t'||*cs ==';'||*cs ==',')
		      cs++;

		if (*cs && *cs != '\n')  i++; /* count parameter */

		/* read parameter till separator */
		while (*cs && *cs != ' ' && *cs != '\t' && *cs != '\n'
		   && *cs != ';' && *cs != ',')
		      cs++;
		if (*cs) cs++;
	}
	return i;
}

/* transpose_map:           */
/* fi(nb, na) -> fi(na, nb) */
void transpose_map(f, na, nb)
double *f;      /* map f(na,nb): 1-dim array of 2-dim map */
int na, nb;
{
	register int i, j;
	int nn = na*nb;
	double *v;
	char tmpfile[161];

	if ((v = (double *)calloc((unsigned)nn, sizeof *v)) == (double *)0) {
		FILE *fp_tmp;
		PR(ER,"%s: no memory for temporal transposed array...\n",
			myname);
		PR(ER, "%s: trying to use disk space for temporal array ",
			myname);
		PR(ER, "of same size...");

		/* open temporal file and save array to disk */
		sprintf(tmpfile, "%s/ftr_XXXXXX", tmpcat);
		if((fp_tmp=fopen(mktemp(tmpfile), "w+")) == (FILE *)0) {
			PR(ER,"%s: can not create temporal file '%s'\n",
			     myname, tmpfile);
			exit(1);
		}
		if ((v = (double *)calloc((unsigned)na, sizeof *v))
			== (double *)0) {
			PR(ER,"\n%s: Sorry. No memory at all...\n",
				myname);
			fclose(fp_tmp);
			unlink(tmpfile);
			exit(1);
		} else {
		     /* transpose matrix:      */
		     /* f(nb, na) -> v(na, nb) */
		     /* and write to disk      */
		     for(i=0; i<nb; i++) {
			for(j=0; j<na; j++) v[j] = f[j*nb+i];

			if (fwrite(v,  sizeof *v,  na, fp_tmp) != na) {
			     PR(ER,"%s: error during writing...\n", myname);
			     fclose(fp_tmp);
			     unlink(tmpfile);
			     exit(1);
			}
		     }
		     free((void *)v);
		     v = (double *)0;

		     fseek(fp_tmp, 0l, 0);
		     for(i=0; i<nb; i++) {
			if (fread(f+i*na,  sizeof *f, na, fp_tmp) != na) {
			     PR(ER,"%s: error during reading...\n", myname);
			     exit(1);
			}
		     }
		     fclose(fp_tmp);
		     unlink(tmpfile);
		}

		PR(ER, "Done.\n");
	} else {
		for(i=0; i<na; i++) {
			register int k = i * nb;
			for(j=0; j<nb; j++) {
				 v[j*na+i] = f[k+j];
			}
		}
		for(i=0; i<nn; i++)  f[i] = v[i];
		free((void *)v);
	}
	return;
}

#ifndef __DARWIN__      /* added by Tristan Beau <beau@in2p3.fr> */
#include <values.h>
#endif
int max_double_()
{
#ifndef MAXDOUBLE
#define MAXDOUBLE       MAXFLOAT
#endif
	return (int)(log10(MAXDOUBLE))-8;
}

void get_val_str(str, v1, v2)
char  *str;
double *v1, *v2;
{
	char str1[20], str2[20];
	register char *cs, *ps;
	float sgn = 1.0;

	ps = str;
	cs = str1;
	while ( ((*cs = *ps) != ',' && *cs != ';' ) && *cs ) {
		cs++; ps++;
	}
	*cs = '\0';
	if ( *ps ) {
		 ps++;
		 cs = str2;
		 while ( ((*cs = *ps) != ',' && *cs != ';' ) && *cs ) {
			 cs++; ps++;
		 }
		 *cs = '\0';
	}
	/* different conditions for 1st string */

	if (str1[0] == '-') {
		sgn = -1.0;
		cs = str1;
		ps = str1 + 1;
		while ((*cs = *ps)) { cs++; ps++; }
	} else if (str1[0] == '+') {
		cs = str1;
		ps = str1 + 1;
		while ((*cs = *ps)) { cs++; ps++; }
	}
	if (str1[0] == '\0')
		*v1 = -FMAX;
	else *v1 = atof(str1);
	*v1 *= sgn;

	sgn = 1.0;

	/* different conditions for 2nd string  */

	if (str2[0] == '-') {
		sgn = -1.0;
		cs = str2;
		ps = str2 + 1;
		while ((*cs = *ps)) { cs++; ps++; }
	} else if (str2[0] == '+') {
		cs = str2;
		ps = str2 + 1;
		while ((*cs = *ps)) { cs++; ps++; }
	}

	if (str2[0] == '\0')
		*v2 = FMAX;
	else *v2 = atof(str2);

	*v2 *= sgn;

	return;
}
void get_2_str(str, str1, str2)
char  *str;
char *str1, *str2;
{
	register char *cs, *ps;

	*str1 = *str2 = '\0';
	ps = str;
	cs = str1;
	while ( ((*cs = *ps) != ',' && *cs != ';' ) && *cs ) {
		cs++; ps++;
	}
	*cs = '\0';
	if ( *ps ) {
		 ps++;
		 cs = str2;
		 while ( ((*cs = *ps) != ',' && *cs != ';' ) && *cs ) {
			 cs++; ps++;
		 }
		 *cs = '\0';
	}
	return;
}
void get_4_str(str, str1, str2, str3, str4)
char  *str;
char *str1, *str2, *str3, *str4;
{
	*str1 = *str2 = *str3 = *str4 = '\0';
	skip_wd(0, str, str1);
	skip_wd(1, str, str2);
	skip_wd(2, str, str3);
	skip_wd(3, str, str4);
	return;
}

/* subroutine of search for ':'-symbol in a string */
int semicol(s)
char *s;
{
   register char *cs,c;
   cs = s;
   while((c = *cs++))   if(c == ':')   return 1;
   return 0;
}

/* convert one of strings: hh:mm:ss.s, ss.s, Nm (N minutes),           */
/* mN (N min), hN,Nh (N hours), gN,Ng (defs), N' (N arcmin) to radians */
double deg2rad(str)
char *str;
{
	double rad=0.0, pi=M_PI, cf = M_PI/180.;
	char *st = str;
	int sg = 1;

again:
	switch(*st) {
	case '+':
		st++;
		goto again;
	case '-':
		st++;
		cf = -cf;
		sg = -1;
		goto again;
	case '\'':
	case 'm':
		rad = atof(st+1) / 60. * cf;
		break;
	case 'g':
	case 'd':
	case 'h':
		rad = atof(st+1) * cf;
		break;
	case '\"':
	case 's':
		rad = atof(st+1)/3600.*cf;
		break;
	case 'p':
		if(strcmp(st, "pi") == 0)
		       rad = pi * sg;
		else if(strcmp(st, "pi2") == 0)
		       rad = pi/2. * sg;
		else if(strcmp(st, "pi3") == 0)
		       rad = pi/3. * sg;
		else if(strcmp(st, "pi4") == 0)
		       rad = pi/4. * sg;
		else if(isdigit(st[1]) || st[1] == '.') {
			rad = -1.e+6 - atof(st+1);
		}
		break;
	default :
		if(semicol(st)) {
		     long d=0, m=0;
		     float s = 0.;
		     sscanf(st,"%ld%*c%ld%*c%f",&d,&m,&s);
		     rad = (d + m/60. + s/3600.)*cf;
		 } else if( (st[2]== 'h'&& (st[3] != ' '&& st[3] != '\0'))
		   || (st[1]== 'h'&& (st[2] != ' '&& st[2] != '\0'))
		   || (st[1]== 'd'&& (st[2] != ' '&& st[2] != '\0'))
		   || (st[2]== 'd'&& (st[3] != ' '&& st[3] != '\0'))
		   || (st[3]== 'd'&& (st[4] != ' '&& st[4] != '\0'))){
		     /* format:  HHhMMmSS.SS */
		     long h=0, m=0;
		     float ti = 0.;
		     sscanf(st,"%ld%*c%ld%*c%f",&h,&m,&ti);
		     rad = ((float)h + (float)m / 60. + ti / 3600.)*cf;
		 } else if((st[2]== ' '&&(st[3] !=' '&& st[3] !='\0'))
		   || (st[1]== ' '&& (st[2] != ' '&& st[2] != '\0'))
		   || (st[3]== ' '&& (st[4] != ' '&& st[4] != '\0'))){
		     /* format:  "HH MM SS.SS" */
		     long h=0, m=0;
		     float ti = 0.;
		     sscanf(st,"%ld%ld%f",&h,&m,&ti);
		     rad = ((float)h + (float)m / 60. + ti / 3600.)*cf;
		} else {
		     switch ( st[strlen(st)-1] )  {
		     case '\'':
		     case 'm':
			     rad = atof(st) / 60. * cf;
			     break;
		     case 'd':
		     case 'g':
		     case 'h':
			     rad = atof(st) * cf;
			     break;
		     case '\"':
		     case 's':
			     rad = atof(st)/3600.*cf;
			     break;
		     case 'i':
			     if (st[strlen(st)-2] == 'p')
				   rad = atof(st) * pi * sg;
			     else
				   rad = atof(st) * sg;
			     break;
		     case 'p':
			     rad = -1.e+6 - atof(st);
			     break;
		     default:
			     rad = atof(st) * sg;
			     break;
		     }
		}
	}
	return rad;
}

/* convert one of strings: hh:mm:ss.s, ss.s, Nm (N minutes),       */
/* mN (N min), hN,Nh (N hours), gN,Ng (defs), N' (N arcmin) to deg */
double str2deg(st)
char *st;
{
	double ang = 0.0;
	int sig=1;
	char *cs=st;
again:
	switch(*cs) {
	case '\0':
	case '\033':
		break;
	case ' ':
	case '+':
		cs++;
		goto again;
	case '-':
		sig = -1;
		cs++;
		goto again;
	case '\'':
	case 'm':
		ang = atof(cs+1) / 60.;
		break;
	case '\"':
	case 's':
		ang = atof(cs+1) / 3600. ;
		break;
	case 'h':
	case 'd':
	case 'g':
		ang = atof(cs+1);
		break;
	default :
		if(semicol(cs)) {  /* if symbol ':' */
		     long h=0, m=0;
		     float tang = 0.;

		     sscanf(cs,"%ld%*c%ld%*c%f",&h,&m,&tang);
		     ang = h + m / 60. + tang / 3600.;
		} else {
		      if( (cs[2]== 'h'&& (cs[3] != ' '&& cs[3] != '\0'))
			|| (cs[1]== 'h'&& (cs[2] != ' '&& cs[2] != '\0'))
			|| (cs[1]== 'd'&& (cs[2] != ' '&& cs[2] != '\0'))
			|| (cs[2]== 'd'&& (cs[3] != ' '&& cs[3] != '\0'))
			|| (cs[3]== 'd'&& (cs[4] != ' '&& cs[4] != '\0'))){
			    /* format:  HHhMMmSS.SS */
			    long h=0, m=0;
			    float ti = 0.;
			    sscanf(cs,"%ld%*c%ld%*c%f",&h,&m,&ti);
			    ang = (float)h + (float)m / 60. + ti / 3600.;
		      }else if((cs[2]== ' '&&(cs[3] !=' '&& cs[3] !='\0'))
			|| (cs[1]== ' '&& (cs[2] != ' '&& cs[2] != '\0'))
			|| (cs[3]== ' '&& (cs[4] != ' '&& cs[4] != '\0'))){
			    /* format:  "HH MM SS.SS" */
			    long h=0, m=0;
			    float ti = 0.;
			    sscanf(cs,"%ld%ld%f",&h,&m,&ti);
			    ang = (float)h + (float)m / 60. + ti / 3600.;
		      } else { /* if one of symbols at the end */
			      /* format: SS.SS{hdms'"} */
			      switch ( cs[strlen(cs)-1] )  {
			      case 'd':
			      case 'g':
			      case 'h':
				      ang = atof(cs);
				      break;
			      case '\'':
			      case 'm':
				      ang = atof(cs) / 60.;
				      break;
			      case 'n':
				      if(strncmp(cs+strlen(cs)-3,"min",3)==0)
					  ang = atof(cs) / 60.;
				      break;
			      case 'c':
				      if(strncmp(cs+strlen(cs)-3,"sec",3)==0)
					  ang = atof(cs) / 3600.;
				      break;
			      case '\"':
			      case 's':
				    {
				      char *qq;
				      int qq_key = 0;
				      qq = cs;
				      while(*qq) {
					    if (*qq == 'm' ||
						*qq == 'h' || *qq == 'd' ||
						*qq == '\'')
						    qq_key++;
					    qq++;
				      }
				      if (qq_key == 2) {
					    /* format:  "HHhMMmSS.SSs" */
					    long h=0, m=0;
					    float ti = 0.;
					    sscanf(cs,"%2ld%*c%2ld%*c%f",
						    &h,&m,&ti);
					    ang = (float)h + (float)m / 60.
						      + ti / 3600.;
				      } else {
					    ang = atof(cs) / 3600.;
				      }
				      break;
				    }
			      default:
				      ang = atof(cs) / 3600.;
				      break;
			      }
		      }
		}
	}
	return (sig*ang);
}

void  read_data_eq2gal(cr, n, filename)
struct _coor_ **cr;
int *n;
char *filename;
{
	register int i=0;
	char str[161], s1[31], s2[31];
	FILE *fp;
	double p = M_PI/180.;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}

	/* strings count */
	while (fgets(str, 160, fp) != NULL) {
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;
		i++;
	}
	fseek(fp, 0l, 0);

	if (i == 0) {
		PR(ER, "%s: problem with content of file '%s'\n",
			myname, filename);
		exit(1);
	}

	*n = i;
	/* find memory for arrays */
	if ((*cr = (struct _coor_ *)calloc((unsigned)i, sizeof **cr)) ==
	    (struct _coor_ *)0) {
		PR(ER,"%s: can not find memory for array...\n", myname);
		exit(1);
	}

	/* strings reading */
	i=0;
	while (fgets(str, 160, fp) != NULL) {
		double alp, del, lat, lon;
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;

		sscanf(str, "%s%s", s1, s2);
		alp = str2deg(s1);    /* hours   */
		del = str2deg(s2);    /* degrees */
		equat2gal(alp, del, &lat, &lon);
		(*cr)[i].lat = lat * p;
		(*cr)[i].lon = lon * p;
		i++;
	}

	fclose(fp);
}
void  read_data_gal(cr, n, filename)
struct _coor_ **cr;
int *n;
char *filename;
{
	register int i=0;
	char str[161], s1[31], s2[31];
	FILE *fp;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}

	/* strings count */
	while (fgets(str, 160, fp) != NULL) {
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;
		i++;
	}
	fseek(fp, 0l, 0);

	if (i == 0) {
		PR(ER, "%s: problem with content of file '%s'\n",
			myname, filename);
		exit(1);
	}

	*n = i;
	/* find memory for arrays */
	if ((*cr = (struct _coor_ *)calloc((unsigned)i, sizeof **cr)) ==
	    (struct _coor_ *)0) {
		PR(ER,"%s: can not find memory for array...\n", myname);
		exit(1);
	}

	/* strings reading */
	i=0;
	while (fgets(str, 160, fp) != NULL) {
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;

		sscanf(str, "%s%s", s1, s2);
		(*cr)[i].lat = deg2rad(s1);
		(*cr)[i].lon = deg2rad(s2);
		i++;
	}

	fclose(fp);
}
void  read_data_pol(cr, n, filename)
struct _coor_ **cr;
int *n;
char *filename;
{
	register int i=0;
	char str[161], s1[31], s2[31];
	FILE *fp;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}

	/* strings count */
	while (fgets(str, 160, fp) != NULL) {
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;
		i++;
	}
	fseek(fp, 0l, 0);

	if (i == 0) {
		PR(ER, "%s: problem with content of file '%s'\n",
			myname, filename);
		exit(1);
	}

	*n = i;
	/* find memory for arrays */
	if ((*cr = (struct _coor_ *)calloc((unsigned)i, sizeof **cr)) ==
	    (struct _coor_ *)0) {
		PR(ER,"%s: can not find memory for array...\n", myname);
		exit(1);
	}

	/* strings reading */
	i=0;
	while (fgets(str, 160, fp) != NULL) {
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;

		sscanf(str, "%s%s", s1, s2);
		(*cr)[i].lat = deg2rad(s1);
		(*cr)[i].lon = deg2rad(s2);
		i++;
	}

	fclose(fp);
}
void pol2gal(theta, phi, lat, lon)
double *lat, *lon;
double theta, phi;
{
	*lat = M_PI/2 - theta;
	*lon = phi;
	return;
}

void gal2pol(lat, lon, theta, phi)
double lat, lon;
double *theta, *phi;
{
	*theta = M_PI/2 - lat;
	*phi = lon;
	return;
}

/*
 *     Convert equatorial coordinates to Galactic
 *     and vice versa:
 *           if key=0, then (alpha,delta) --> (b2,l2)
 *           if key=1, then (b2,l2) --> (alpha,delta)
 *     alpha in hours;   delta,b2d,l2d  in  degrees
 *     version 2.0, 1994-Sep; author: Trushkin S.A.
 */
#ifdef __GNUC__
int gal_coor(double *alp, double *del, double *b, double *l, int key)
#else
int gal_coor(alp, del, b, l, key)
double *alp, *del;  /* alpha in hours, delta in degs        */
double *b, *l;      /* gal. latitude and longitude in degs  */
int    key;         /* key = 0: (alpha, delta) --> (b2, l2) */
#endif
{
	double pi=M_PI,
//           a0=282.25, d0=62.6,  l0=33.;   /*   1950.0 */
//           a0=283.0, d0=62.83, l0=33.;    /* ? 2000.0 */
//arXiv:1010.3773:  ???
//  a_new =      12h51m36s.7151981
//  d_new =     +27d06'11".193172
//  theta_new= 123d.0075021536
//          a0=282.90298, d0=62.89689079, l0=33.0075021536;
// ICRS:
//      12h51m26s.27469
//     +27d07'41".7087
//      122d.93192526
	  a0=282.8594779, d0=62.87174758, l0=32.93192526;

	double l2, b2, rd, sinl, cosl, l33, tl=0., dl;
	double decr, alpha1, sind, cosd, sina1, cosa1, n0=0., sd0, cd0;
	double cosb, sinb, decd;

	rd = pi / 180.;
	sd0 = sin (d0 * rd);
	cd0 = cos (d0 * rd);
	switch (key) {
	case  0 :    /* (alpha, delta)  -->  (b2, l2) */
		decr   = *del * rd;
		alpha1 = (*alp * 15. - a0) * rd;
		sind   = sin(decr);
		cosd   = cos(decr);
		sina1  = sin(alpha1);
		cosa1  = cos(alpha1);
		b2 = asin((-cosd) * sina1 * sd0 + sind * cd0);
		*b = b2 / rd;    /* determination of b2 */

		cosl = cosd * cosa1;
		sinl = cosd * sina1 * cd0 + sind * sd0;
		if(cosl != n0)    tl = sinl / cosl;
		dl = n0;
		if(sinl <= n0 && cosl > n0)   dl = 2. * pi;
		if(cosl < n0)   dl = pi;
		l33 = pi / 2. + dl;
		if(cosl != n0)   l33 = atan(tl) + dl;
		*l = l33 / rd + l0;     /* determination of l2 in deg */
		if(*l > 360.)    *l -= 360.;
		break;

	case  1 :         /*  (b2, l2)  -->  (alpha, delta)  */
		l2 = (*l - l0) * rd;
		b2 = *b * rd;
		cosb = cos(b2);
		sinb = sin(b2);
		sind = cosb * sin(l2) * sd0 + sinb * cd0;
		decr = asin(sind);
		decd = decr/rd;
		*del = decd;          /* determination of delta(deg) */

		cosa1 = cosb * cos(l2);
		sina1 = (sind * cd0 - sinb) / sd0;
		tl = sina1 / cosa1;
		dl = n0;
		if(sina1 <= n0 && cosa1 > n0)  dl = 2. * pi;
		if(cosa1 < n0) dl=pi;
		alpha1 = pi/2 + dl;
		if(cosa1 != n0) alpha1 = atan(tl) + dl;
		alpha1 = alpha1 / rd + a0;
		if(alpha1 > 360.)   alpha1 -= 360.;
		*alp = alpha1 / 15.;  /* determination of alfa(h) */
		break;
	default:
		return key;
	}
	return 0;
}

void  equat2gal(alp, del, lat, lon)
double alp, del, *lat, *lon;
{
    gal_coor(&alp, &del, lat, lon, 0);
    return;
}

void  gal2equat(lat, lon, alp, del)
double *alp, *del, lat, lon;
{
    gal_coor(alp, del, &lat, &lon, 1);
    return;
}

/*
 *  Read data from string 'str' and convert them to (theta, phi)
 */
void str2sph(str, theta, phi, t, gal_flag)
char   *str;
double *theta, *phi, *t;
{
	double epoch = 2000.0;
	int n=0;
	double a,b, lat, lon;
	int rh, rm, dd, dm, sg = 1;
	double rs, ds;
	char st[81];

	/* number of parameters in a string */
	n = n_skip_wd(str);

	switch(n) {
	case 3:
	case 4:
		a = str2deg((skip_wd(0, str, st),st));
		b = str2deg((skip_wd(1, str, st),st));
		if (n==4)  {
		    skip_wd(2, str, st);
		    if (*st == 'B' || *st == 'b')      epoch = 1950.0;
		    else if (*st == 'J' || *st == 'j') epoch = 2000.0;
		    else epoch = atof(st);
		    *t = atof((skip_wd(3, str, st),st));
		} else
		    *t = atof((skip_wd(2, str, st),st));
		break;
	case 7:
	case 8:
		sscanf(str, "%d%d%lf%d%d%lf", &rh, &rm, &rs, &dd, &dm, &ds);
		skip_wd(3, str, st);
		if (*st == '-')  { sg = -1; dd = -dd; }
		a = rh + rm / 60. + rs / 3600.;
		b = sg * (dd + dm / 60. + ds / 3600.);
		if (n==8)  {
		    skip_wd(6, str, st);
		    if (*st == 'B' || *st == 'b')      epoch = 1950.0;
		    else if (*st == 'J' || *st == 'j') epoch = 2000.0;
		    else epoch = atof(st);
		    *t = atof((skip_wd(7, str, st),st));
		} else
		    *t = atof((skip_wd(6, str, st),st));
		break;
	default:
		PR(ER, "%s: problems with format of point source file...\n",
		    myname);
		exit(1);
	}
	if ( gal_flag )  {
		lat = a;
		lon = b;
	} else {
		/* equatorial coordinates */
		if (fabs(epoch - 2000.0) > 0.01) {
		    /* convert to 2000.0 */
		    /*-------------------*/
		    char s1[30], s2[30];
		    char tmp_file[81], command[121];
		    char *tmp_pat="cl2mfXXXXXX";
		    FILE *fp;

		    /* create temporal file name */
		    sprintf(tmp_file, "%s/%s", tmpcat, tmp_pat);
		    mktemp(tmp_file);

		    /**********************************************/
		    /* convert observational coordinates to epoch */
		    /**********************************************/
		    /* command to be executed */
		    sprintf(command,
			  "epoch -ei %g -eo J -a %9.6f -d %10.6f > %s",
			  epoch, a, b, tmp_file);
		    if (system(command) != 0) {
			PR(ER,
	     "%s: problems with execution of the following command:\n%s\n",
			    myname, command);
			exit(1);
		    }
		    /* read coordinates to array */
		    if((fp  = fopen(tmp_file, "r")) == (FILE *)0) {
			PR(ER, "%s: can not open file '%s'\n",
			    myname, tmp_file);
			exit(1);
		    }
		    fgets(st, 80, fp);
		    sscanf(st, "%s%s", s1, s2);
		    a = str2deg(s1);
		    b = str2deg(s2);
		    fclose(fp);
		    unlink(tmp_file);
		}
		equat2gal(a, b, &lat, &lon);
	}

	lat *= M_PI/180.;
	lon *= M_PI/180.;

	gal2pol(lat, lon, theta, phi);

	return;
}

int map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)
int nx1, np1, *nph1;
int nx2, np2, *nph2;
{
    int flag=0;

    if (nx2 != nx1 || np2 != np1 || nph1[nx1/2-1] != nph2[nx1/2-1] ||
	nph1[1] != nph2[1] || nph1[2] != nph2[2])
	   flag = 1;

    return flag;
}

/*
     cos be * cos le = cos Dec * cos RA
     cos be * sin le = cos Dec * sin RA * cos eps + sin Dec * sin eps
     sin be = - cos Dec * sin RA * sin eps + sin Dec * cos eps
     eps = 23:26:21.448
*/
void equat2eclip(ra,dec, betta, lambda)
double ra, dec;          /* hours; degs */
double *betta, *lambda;  /* degs;  degs */
{
    double cos_eps = 0.9174820621, sin_eps = 0.3977771559;
    double alp = ra / 12. * M_PI, del = dec / 180. * M_PI;
    double cos_ra  = cos(alp), sin_ra  = sin(alp),
	   cos_dec = cos(del), sin_dec = sin(del);
    double cos_bet, cos_lam=0., sin_lam=0.;
    *betta  = asin (-cos_dec * sin_ra * sin_eps + sin_dec * cos_eps);
    cos_bet = cos(*betta);
    if (ABS(cos_bet) > 1.e-10) {
	cos_lam = cos_dec * cos_ra / cos_bet;
	sin_lam = (cos_dec*sin_ra*cos_eps + sin_dec*sin_eps) / cos_bet;
	*lambda = atan2 (sin_lam, cos_lam);
	if (*lambda < 0.)  *lambda += 2 * M_PI;
    } else {
	*lambda = 0.;
    }
    *betta  *= 180/M_PI;
    *lambda *= 180/M_PI;

    return;
}


/*
     sin Dec = cos eps * sin be + sin eps * cos be * sin le
     cos Dec cos RA = cos be * cos le
     cos Dec sin RA = -sin be * sin eps + cos be * cos eps * sin le
     eps = 23:26:21.448

     return RA in hours, Dec in degrees
*/
void eclip2equat(betta, lambda, ra,dec)
double betta, lambda;
double *ra, *dec;
{
    double cos_eps = 0.9174820621, sin_eps = 0.3977771559;
    double bet = betta * M_PI/180., lam = lambda * M_PI/180.;
    double cos_bet = cos(bet), sin_bet = sin(bet),
	   cos_lam = cos(lam), sin_lam = sin(lam);
    double cos_dec, cos_ra, sin_ra;

    *dec = asin ( cos_eps * sin_bet + sin_eps * cos_bet * sin_lam );
    cos_dec = cos (*dec);
    if (ABS(cos_dec) > 1.e-10) {
	cos_ra = cos_bet * cos_lam / cos_dec;
	sin_ra = (-sin_bet * sin_eps + cos_bet * cos_eps * sin_lam)/cos_dec;
	*ra = atan2 (sin_ra, cos_ra);
	if (*ra < 0.)  *ra += 2 * M_PI;
    } else {
	*ra = 0.;
    }
    *ra  *= 12. /M_PI;
    *dec *= 180./M_PI;

    return;
}


