#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr

/*
 *
 *  Oleg Verkhodanov (TAC), vo@sao.ru, 2002
 *        to read HEALPix maps and convert them to GLESP maps
 *        last modification, 2/12/2004, 13/03/2006
 */

extern int F_sun_data_key;

extern void setopt(int, char **), usage(void);

int lx_add = 4;       /* addition to Nphi numbers for 'coord'-function */
extern int test_flag; /* = 0; */

char filename[161];
char outp_file[161];
char mask_file[161];
char zone_file[81];    /* file with list of zones to be processed */
int  set_Legendre_grid = 1;
int  spline_repixelization = 0;
int  nx_g = 0, np_g = 0, lmax_g = -1; /* dimensions of output map */
int  equiv_flag = 0;
int gp_out = 0;          /* flag to print out coordinates and T */
int second_field = 0;    /* make output of sectond filed */
int reflection = 0;
int ffile_flag = 0;      /* output to FITS file            */
int flag_bf_hp=0;        /* flag for Basic FITS in Hpix-format */

double conv_factor = 1.0;   /* conversion factor of temperature */
int inp_ff_flag = 0;        /* input F-file                     */
double radius = -1.;        /* radius of removed circle in degs */
double delta_dec = 0.0;
int circle = 0;
int pol_map = 0;            /* read HEALPix polarization maps   */

char cr_file[161];
double cut_low = -1.e-30, cut_up = 1.e+30;
double lat0 = -1000., lon0 = -1000., lat1 = -1000., lon1 = -1000.;
double latR0 = -1000., lonR0 = -1000., latR1 = -1000., lonR1 = -1000.;
int cut_values = 0;
int remove_flag=0, remain_flag=0;
int convert_mask_2_glesp_flag = 0;
double dummy_value = 0.0;
int eq_input = 0;
int galactic_coors = 0;     /* input table in polar coordinates */
int equat_coors = 0;
struct _cross_sect_ {       /* input cross section by RA        */
	double ra, dec;
	double st, st2;
	int n, n2;
} cs;

char *myname="cmap";


int main(argc, argv)
int argc;
char **argv;
{
    register int i=0,j=0;
    FILE *fp_out = (FILE *)0;
    double *f = (double *)0,
	   *q = (double *)0,
	   *u = (double *)0,
	   *x = (double *)0,
	   *r = (double *)0,
	   *offs = (double *)0;
    int *nph = (int *)0;
    double *f_g = (double *)0,
	   *q_g = (double *)0,
	   *u_g = (double *)0,
	   *x_g = (double *)0;
    int *nph_g = (int *)0;
    int *counts = (int *)0;
    int nx, nphi, nside=0;
    double theta0, theta1, phi0, phi1;
    char grid_type[30], x_type[30];
    int  grid_flag = 0;
    double pi = M_PI;
    int map_out_flag = 1;
    double pi2 = M_PI * 2.;

    filename[0] = outp_file[0] = mask_file[0] = cr_file[0] = '\0';
    cs.n = cs.n2 = -1;
    strcpy(tmpcat, ".");

    /* loop over external parameters */
    setopt(argc, argv);

    /* read Hpix-file */
/*
    if ( second_field && !gp_out) second_field = 0;
*/
    if (convert_mask_2_glesp_flag) {
	;
    } else {
	if ( second_field ) {
		read_hp_file_nobs(&counts, &nph, &nx, &nphi, &x, filename);
		i = nx*nphi;
		if((f=(double*)calloc((unsigned)i,sizeof *f))==(double*)0)
		    goto nomem_e;
		for(i=0; i<nx; i++) {
		    int ii = i * nphi;
		    for(j=0; j<nph[i]; j++) f[ii+j] = counts[ii+j];
		}
		if (counts != (int *)0)   free((void *)counts);
	} else if (pol_map) {
		read_hp_pol(&f, &q, &u, &nph, &nx, &nphi, &x, filename);
	} else {
	    extern void read_bhp_file(double **, int **, int *, int *,
		 double **, char *);
	    if(flag_bf_hp)  /* flag for Basic FITS in Hpix-format */
		read_bhp_file(&f, &nph, &nx, &nphi, &x, filename);
	    else /* just Hpix-format */
		read_hp_file(&f, &nph, &nx, &nphi, &x, filename);
	}
	nside = (nx+1) / 4;
	i=nx;
	if((offs=(double*)calloc((unsigned)i,sizeof *offs))==(double*)0)
	    goto nomem_e;

	/* calculate arrays: x & nph (again), offs (new) */
	hp_arrays(x, offs, nph, nside);

	/* read mask file: Binary or ASCII */
	/*   and multiply it by data       */
	if (*mask_file != '\0') {
	    int *nph_m = (int *)0;
	    int nx_m, nphi_m;
	    double *x_m = (double *)0;

	    read_mask(&counts, &nph_m, &nx_m, &nphi_m, &x_m, filename, nside);

	    if(nx_m != nx || nphi_m != nphi ) {
	       PR(ER,
		  "%s: different resolution in datafile and mask-file...\n",
		  myname);
	       exit(1);
	    }
	    if (nph_m != (int *)0)  free((void *)nph_m);
	    if (x_m != (double *)0) free((void *)x_m);

	    /* set mask */
	    for(i=0; i<nx; i++) {
		int ii = i * nphi;
		for(j=0; j<nph[i]; j++) f[ii+j] *= counts[ii+j];
	    }

	    if (counts != (int *)0)   free((void *)counts);
	}
    }

if (!second_field) {
register int i,j;
double vmin= 1.e30, vmax= -1.e30,
       qmin= 1.e30, qmax= -1.e30,
       umin= 1.e30, umax= -1.e30;
for (i=0;i<nx;i++) {
int ii=i*nphi;
for (j=0;j<nph[i];j++) {
    if ( f[ii+j] < vmin) vmin = f[ii+j];
    if ( f[ii+j] > vmax) vmax = f[ii+j];
    if (pol_map) {
	if ( q[ii+j] < qmin) qmin = q[ii+j];
	if ( q[ii+j] > qmax) qmax = q[ii+j];
	if ( u[ii+j] < umin) umin = u[ii+j];
	if ( u[ii+j] > umax) umax = u[ii+j];
    }
}}
PR(ER, "input values:\n");
PR(ER, "vmin=%15.8e, vmax=%15.8e\n", vmin, vmax);
if (pol_map) {
PR(ER, "qmin=%15.8e, qmax=%15.8e\n", qmin, qmax);
PR(ER, "umin=%15.8e, umax=%15.8e\n", umin, umax);
}
}

    if (cut_values && f != (double *)0) {
	  for(i=0; i<nx; i++) {
	      int ii = i * nphi;
	      for(j=0; j<nph[i]; j++) {
		  if(f[ii+j] > cut_up )  f[ii+j] = cut_up ;
		  if(f[ii+j] < cut_low)  f[ii+j] = cut_low ;
	      }
	  }
    }

    /* remove or keep zones */
    if (f != (double *)0 && (remove_flag || remain_flag) ) {
       register int is=0;
       int ns = 1;
       struct _zone_ {
	    double latR0, lonR0;
	    double latR1, lonR1;
	    double radius;
	    int type;               /* zone type */
       } *zz;
       extern void hmap_zone(struct _zone_ *, int, int,
	       double *, double *, double *, int, int, int *, double);
       int function_type = remove_flag ? 1 : 0;
       int zone_type = circle ? 0 : 1;

       if (zone_file[0] != '\0') {
	    FILE *fz;
	    char str[81];
	    /* read data from file with zones */
	    if((fz = fopen(zone_file, "r")) == (FILE *)0) {
		PR(ER, "%s: I can not open file '%s'...\n",
		    myname, zone_file);
		exit(1);
	    }
	    /* get number of zones */
	    is =0;
	    while(fgets(str, 80, fz) != NULL) {
		if (*str == '#' || *str == '\n')
		    continue;
		is++;
	    }
	    /* allocate memory */
	    i = ns = is;
	    zz = (struct _zone_ *)calloc(i, sizeof *zz);
	    if (zz == (struct _zone_ *)0) goto nomem;

	    /* read zones  */
	    is=0;
	    fseek(fz, 0l, 0);
	    while(fgets(str, 80, fz) != NULL) {
		char s1[30], s2[30], s3[30], s4[30];
		char *cs = str;
		int n_par;  /* number of parameters per string */
		if (*str == '#' || *str == '\n')
			continue;
		while(*cs) if (*cs == '#') {*cs = '\0'; break;} else cs++;

		n_par = n_skip_wd(str);

		switch (n_par) {
		case 1:
		    PR(ER,"%s: there is only 1 parameter in the string:\n",
			    myname);
		    PR(ER,"%s", str);
		    PR(ER, "\n%s: I've skipped it...\n", myname);
		    break;
		case 2:
		    zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		    zz[is].radius = radius;
		    zz[is].type = 0;
		    break;
		case 3:
		    zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		    if (radius > 0.)
			zz[is].radius = radius;
		    else
			zz[is].radius = deg2rad((skip_wd(2,str,s3),s3));
		    zz[is].type = 0;
		    break;
		case 4:
		    zz[is].latR0 = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0 = deg2rad((skip_wd(1,str,s2),s2));
		    zz[is].latR1 = deg2rad((skip_wd(2,str,s3),s3));
		    zz[is].lonR1 = deg2rad((skip_wd(3,str,s4),s4));
		    if (zone_type == 0) {
			    zz[is].radius = zz[is].latR1;
			    zz[is].type = 0;
		    } else  zz[is].type = 1;
		    break;
		}
		is++;
	    }

	    fclose(fz);
       } else {
	    /* allocate memory for 1 zone */
	    i = ns = 1;
	    zz = (struct _zone_ *)calloc(i, sizeof *zz);
	    if (zz == (struct _zone_ *)0) goto nomem;
	    zz[0].latR0 = latR0;  zz[0].lonR0 = lonR0;
	    zz[0].latR1 = latR1;  zz[0].lonR1 = lonR1;
	    zz[0].radius = radius;
	    if (latR1 < -900.) {
		    zz[0].type = 0;
	    } if (lonR1 < -900.) {
		    if (radius > 0.)
			zz[is].radius = radius;
		    else
			zz[is].radius = latR1;
		    zz[0].type = 0;  /* circle */
	    } else {
		    zz[0].type = 1;  /* rectangular */
	    }
       }
//       /* convert position of zones by phi */
//       /* phi(HEALPix) = 2pi - phi(GLESP)  */
//       /* check why ?   (VO, 10/12/04)     */
//       /* the same problem is also below   */
//       for(is=0; is<ns; is++) {
//            if(zz[is].lonR0 > -1000.)  zz[is].lonR0 = pi2 - zz[is].lonR0;
//            if(zz[is].lonR1 > -1000.)  zz[is].lonR1 = pi2 - zz[is].lonR1;
//       }


       /* process zones */
       hmap_zone(zz,ns, function_type, f,x, offs, nx,nphi,nph, dummy_value);

       /* free memory from zone-array */
       free((void *)zz);
    }

    /********************/
    /* open output file */
    /********************/
    if (*outp_file == '\0')  fp_out = stdout;
    else {
	if (strcmp(outp_file,"-")==0 || strcmp(outp_file,"stdout")==0)
	    fp_out = stdout;
	else if ((fp_out = fopen(outp_file, "w")) == (FILE *)0) {
	    PR(ER, "%s: can not create file '%s'\n", myname, outp_file);
	    exit(1);
	}
    }

    /************************************************************/
    /* extract zones from initial file and write them to output */
    /************************************************************/
    if (f != (double *)0 ) { /* i.e. we do not work with 'counts'-array */
	if( *cr_file != '\0') {
	    int n_cr;
	    double p = M_PI/180;
	    /******************/
	    /* extract vector */
	    /******************/
	    struct _coor_  *cr = (struct _coor_ *)0;

	    if (inp_ff_flag) {
		/* read (lat,lon) from F-file */
		read_ff(&cr, &n_cr, cr_file, 1, delta_dec);

		i = n_cr;
		if ((r =(double*)calloc((unsigned)i,sizeof *r))==(double *)0)
		    goto nomem;

		/* read vector */
		for(i=0; i<n_cr; i++) {
		     int k;
		     double alp=cr[i].lat,
			    del=cr[i].lon,
			    lat=cr[i].lat, lon=cr[i].lon;
		     if (eq_input)  {
			    alp /= 15.*p;
			    del /= p;
			    equat2gal(alp, del, &lat, &lon);
			    lat *= p;
			    lon *= p;
		     }
		     gal2pol(lat, lon, &theta1, &phi1);

//                     /* convert position of zones by phi */
//                     /* phi(HEALPix) = 2pi - phi(GLESP)  */
//                     /* check why ?   (VO, 10/12/04)     */
//                     /* the same problem is also below   */
//                     phi1 = pi2 - phi1;

		     k = hang2n(theta1, phi1, x, nx, nphi, nph, offs);
		     r[i] = f[k] * conv_factor;
		}

		/* write to F-file */
		write_ff(r, n_cr, fp_out);
	    } else {
		if (equat_coors) {
			/* read (alp,del)in radians from file      */
			/* and converse them to (lat,lon) radians  */
			read_data_eq2gal(&cr, &n_cr, cr_file);
			galactic_coors = 1;
		} else if (galactic_coors) {
			/* read (lat,lon) in radians from file */
			/* and converse them to radians        */
			read_data_gal(&cr, &n_cr, cr_file);
		} else {
			/* read (theta,phi) in radians from file */
			read_data_pol(&cr, &n_cr, cr_file);
		}

		i = n_cr;
		if((r=(double *)calloc((unsigned)i,sizeof *r))==(double *)0)
		    goto nomem_e;

		/* read vector */
		for(i=0; i<n_cr; i++) {
		     int k;
		     if (galactic_coors) {
			    gal2pol(cr[i].lat, cr[i].lon, &theta1, &phi1);
		     } else {
			    theta1 = cr[i].lat;
			    phi1 = cr[i].lon;
		     }
//                     /* convert position of zones by phi */
//                     /* phi(HEALPix) = 2pi - phi(GLESP)  */
//                     /* check why ?   (VO, 10/12/04)     */
//                     /* the same problem is also below   */
//                     phi1 = pi2 - phi1;

		     /* get pixel number */
		     k = hang2n(theta1, phi1, x, nx, nphi, nph, offs);
		     r[i] = f[k] * conv_factor;
//                     phi1 = pi2 - phi1;

		     if (galactic_coors) {
			double lat = cr[i].lat/M_PI*180,
			       lon = cr[i].lon/M_PI*180;
			int dd1, dd2, mm1, mm2, c1='+';
			double ss1, ss2;

			if (equat_coors) {
			    double alp, del;
			    /* convert galactic to equatorial */
			    gal2equat(lat, lon, &alp, &del);
			    if (del < 0.0) { c1 = '-'; del = -del; }
			    dd1 = (int)del;
			    mm1 = (int)((del - dd1)*60.);
			    ss1 = (del - dd1 - mm1/60.) * 3600.;
			    dd2 = (int)alp;
			    mm2 = (int)((alp - dd2)*60.);
			    ss2 = (alp - dd2 - mm2/60.) * 3600.;

			    PR(fp_out,
			     "%02d:%02d:%05.2f  %c%02d:%02d:%04.1f %12.5e\n",
			     dd2,mm2,ss2, c1,dd1,mm1,ss1, r[i]);
			} else {
			    if (lat < 0.0) { c1 = '-'; lat = -lat; }
			    dd1 = (int)lat;
			    mm1 = (int)((lat - dd1)*60.);
			    ss1 = (lat - dd1 - mm1/60.) * 3600.;
			    if (lon < 0.)       lon += 2*M_PI;
			    if (lon > 2.*M_PI)  lon -= 2*M_PI;
			    dd2 = (int)lon;
			    mm2 = (int)((lon - dd2)*60.);
			    ss2 = (lon - dd2 - mm2/60.) * 3600.;

			    PR(fp_out,
			     "%c%02d:%02d:%04.1f  %03d:%02d:%04.1f %12.5e\n",
			     c1,dd1,mm1,ss1, dd2,mm2,ss2, r[i]);
			}
		     }
		     else
			   /* print vector */
			   PR(fp_out, "%12.5e  %12.5e  %12.5e\n",
				  theta1, phi1, r[i]);
		}

	    }
	    if (cr != (struct _coor_ *)0) free((void *)cr);
	    map_out_flag = 0;
	} if (cs.n > 0) {
	    /* extract vector using parameters of cross-section */
	    int n_cr = cs.n;
	    double p = M_PI/180;
	    double lat = eq_input ? cs.ra  : cs.ra * 15. * p,
		   lon = eq_input ? cs.dec : cs.dec * p,
		   st  = eq_input ? cs.st  : cs.st * 15. * p;
	    double st2 = eq_input ? cs.st2 : cs.st2 * p;

	    if (cs.n2 > 0)  i = cs.n * cs.n2; else i = n_cr;
	    if ((r=(double*)calloc((unsigned)i, sizeof *r))==(double *)0)
		goto nomem;

	    if (cs.n2 > 0) {
		/* read vector */
		for(i=0; i<cs.n2; i++) {
		    int ii = n_cr * i;
		    double del = lon + st2 * j;
		    for(j=0; j<n_cr; j++) {
			int k;
			double lat1, lon1;
			if (eq_input)  {
			       double alp=lat+st*j;
			       equat2gal(alp, del, &lat1, &lon1);
			       lat1 *= p;
			       lon1 *= p;
			} else {
			       lat1=lat+st*j; lon1=del;
			}
			gal2pol(lat1, lon1, &theta1, &phi1);

//                        /* convert position of zones by phi */
//                        /* phi(HEALPix) = 2pi - phi(GLESP)  */
//                        /* check why ?   (VO, 10/12/04)     */
//                        phi1 = pi2 - phi1;

			k = hang2n(theta1, phi1, x, nx, nphi, nph, offs);
			r[ii+j] = f[k] * conv_factor;
		    }
		}

		if (ffile_flag) {
		   /* write ff-map */
		   write_simple_map(fp_out, r, n_cr, cs.n2, lon, lat, st, st2,
			"GLON", "GLAT", 1., "T");
		} else {
		   /* write ff-map */
		   write_2d_fits(fp_out, r, n_cr, cs.n2, lon, lat, st, st2,
			"GLON", "GLAT", 1., "T");
		}
	    } else {
		/* read vector */
		for(i=0; i<n_cr; i++) {
		     int k;
		     double lat1, lon1;
		     if (eq_input)  {
			    double alp=lat+st*i, del =lon;
			    equat2gal(alp, del, &lat1, &lon1);
			    lat1 *= p;
			    lon1 *= p;
		     } else {
			    lat1=lat+st*i; lon1=lon;
		     }
		     gal2pol(lat1, lon1, &theta1, &phi1);

//                     /* convert position of zones by phi */
//                     /* phi(HEALPix) = 2pi - phi(GLESP)  */
//                     /* check why ?   (VO, 10/12/04)     */
//                     phi1 = pi2 - phi1;

		     k = hang2n(theta1, phi1, x, nx, nphi, nph, offs);
		     r[i] = f[k] * conv_factor;
		}

		/* prepare head of F-file */
		create_ff(FLOATPIX,(long)n_cr,"K", 1.0,0.0,
		       cs.ra, cs.st * 3600., 1.0, "TIME");

		/* write to F-file */
		write_ff(r, n_cr, fp_out);

		if (r != (double *)0) { free((void *)r);  r = (double *)0; }
		map_out_flag = 0;
	    }
	} else if (lat0 > -1000. ) {
	    if (lat1 > -1000.) {
		 int nx0, nx1, np0, np1;
		 int nn, nm;
		 double dt, dp;

		 /* extract rectangular */
		 gal2pol(lat0, lon0, &theta0, &phi0);
		 gal2pol(lat1, lon1, &theta1, &phi1);
		 nx0 = ind_x(nx, x, cos(theta0));
		 nx1 = ind_x(nx, x, cos(theta1));
		 nn = ABS(nx1-nx0+1);
		 dt = (theta1 - theta0) / (double)nn;

		 np0 = lon0 / pi2 * (double)nph[nx0];
		 np1 = lon1 / pi2 * (double)nph[nx0];
		 i   = lon0 / pi2 * (double)nph[nx1];
		 j   = lon1 / pi2 * (double)nph[nx1];
		 if ( ABS(np1-np0) > ABS(j-i) )
		       nm = ABS(np1-np0+1);
		 else
		       nm = ABS(j-i+1);
		 dp = (phi1 - phi0) / (double)nm;

		 i = nn*nm;
		 r = (double *)calloc((unsigned)i, sizeof *r);
		 if ( r == (double *)0) {
nomem_e:
		      PR(ER, "%s: no memory for arrays of '%d' elements...\n",
			  myname, i);
		      exit(1);
		 }
		 for(i=0; i<nn; i++) {
		      int ii = i * nm, k;
		      double tn = theta0 + i * dt, pn;
		      for(j=0; j<nm; j++) {
			  pn = phi0 + j * dp;

//                          /* convert position of zones by phi */
//                          /* phi(HEALPix) = 2pi - phi(GLESP)  */
//                          /* check why ?   (VO, 10/12/04)     */
//                          pn = pi2 - pn;

			  k = hang2n(tn, pn, x, nx, nphi, nph, offs);
			  if (k>=0)
			      r[ii+j] = f[k] * conv_factor;
		      }
		 }
		 if (ffile_flag) {
		    /* write ff-map */
		    write_simple_map(fp_out, r, nm, nn, lon0, lat0, dp, dt,
			 "GLON", "GLAT", 1., "T");
		 } else {
		    /* write fits-map */
		    write_2d_fits(fp_out, r, nm, nn, lon0, lat0, dp, dt,
			 "GLON", "GLAT", 1., "T");
		 }
		 map_out_flag = 0;
	    } else {
		 /*********************/
		 /* print pixel value */
		 /*********************/

		 gal2pol(lat0, lon0, &theta1, &phi1);

// /* convert position of zones by phi */
// /* check why ?   (VO, 10/12/04)     */
//for(i=0; i<nx; i++) {
//int np  = nph[i] - 1;
//int np2 = nph[i] / 2;
//int ii = i * nphi;
//for(j=0; j<np2; j++) {
//int jn = np-j; double v = f[ii+j];
//f[ii+j]  = f[ii+jn]; f[ii+jn] = v;
//}}


		 i = hang2n(theta1, phi1, x, nx, nphi, nph, offs);
/*
PR(ER, "i=%d, ntot=%d, nx=%d, nphi=%d\n", i, nx*nphi, nx, nphi);
*/
		 PR(fp_out,"theta = %g, phi = %g, t = %g\n",
		    theta1, phi1, f[i] * conv_factor);
/*
for(i=0; i<nx;i++)
for(j=0; j<nphi;j++)
PR(ER, "%d %d %d  %g\n",i,j,i*nphi+j, f[i*nphi+j]);
*/
	    }
	    map_out_flag = 0;
	}
    }
    if (r != (double *)0)     free((void *)r);
    if (offs != (double *)0)  free((void *)offs);

    /********************************/
    /* map recalculation for output */
    /********************************/
    if (map_out_flag) {
	if (!gp_out) {
	    /* recalc to Legendre grid */
	    /* (repixelization)        */
	    if ( set_Legendre_grid ) {
		double x1= -1.0, x2 = 1.0;
		int nside = nphi / 4, nx2;
		double pixel_size = 4*pi / (12.*nside*nside);

		if (lmax_g < 0) {
		    if (nx_g <= 0) {
			 double dt = sqrt(pixel_size);
			 nx_g = pi / dt;
			 if(!(nx_g%2))  nx_g++;
			 lmax_g = (nx_g - 1) * 2;
		    }
		}
		if (nx_g <= 0) nx_g = 2 * lmax_g + 1;
		if (np_g <= 0) np_g = 2 * nx_g;

//                int lmax = lmax_g > 0 ? lmax_g : 1.5 * nside;


PR(ER, "Legendre grid is selected.\n");
		if (equiv_flag || nx_g || np_g) {
		    double dt = sqrt(pixel_size);

		    /* select equal size pixels for repixelized map */
		    if (nx_g == 0 && np_g  > 0)  nx_g = np_g / 2;
		    else if (nx_g >  0 && np_g == 0)  np_g = nx_g * 2;
		    else if (nx_g == 0 && np_g == 0 && equiv_flag) {
			nx_g = pi / dt;
		    } else if (nx_g < 0 || np_g < 0) {
			    PR(ER, "%s: bad values of input parameters...\n",
				myname);
			    exit(1);
		    }
		    if((nph_g=(int*)calloc(nx_g, sizeof *nph_g))==(int*)0) {
nomem:
			    PR(ER,"%s: no memory...\n", myname);
			    exit(1);
		    }
		    if((x_g =(double*)calloc(nx_g, sizeof *x_g))==(double *)0)
			    goto nomem;

		    gauleg(x1, x2, x_g, (double *)0, nx_g);

		    nx2 = (nx_g+1)/2;
		    dt = fabs(x_g[nx2+1] - x_g[nx2-1]) / 2.;

		    if (nx_g == 0 && np_g == 0 && equiv_flag)
			np_g = 2*pi / dt + 0.5;
		    if (pol_map)
			nx2 = set_nphi_by_x2p(x_g, nx_g, np_g, nph_g);
		    else {
			nx2 = set_nphi_by_x2(x_g, nx_g, np_g, nph_g);
		    }

PR(ER, "nside=%d, nx=%d, np=%d, ntot_hp=%d, ntot=%d, n_pol=%d\n",
nside, nx_g, np_g, nside*nside*12, nx2, nph_g[0]);
		    nx2 = nx_g * np_g;
		    if((f_g =(double*)calloc(nx2, sizeof *f_g))==(double *)0)
			    goto nomem;
		    /* appropriate pixels -> 1st grid to 2nd grid */
		    cop_grid(f, nx, nphi, nph, x,
				f_g, nx_g, np_g, nph_g, x_g);

		    if (pol_map) {
		       if((q_g=(double*)calloc(nx2,sizeof *q_g))==(double*)0)
			    goto nomem;
		       if((u_g=(double*)calloc(nx2,sizeof *u_g))==(double*)0)
			    goto nomem;

		       /* appropriate pixels -> 1st grid to 2nd grid */
		       cop_grid(q, nx, nphi, nph, x,
			    q_g, nx_g, np_g, nph_g, x_g);
		       cop_grid(u, nx, nphi, nph, x,
			    u_g, nx_g, np_g, nph_g, x_g);
		    }

		    /* free memory from old arrays */
		    free((void *)f);
		    if (pol_map) {
			free((void *)q);
			free((void *)u);
		    }
		    free((void *)x);
		    free((void *)nph);

		    /* use old variables for new arrays */
		    f = f_g;
		    if (pol_map) {
			q = q_g;
			u = u_g;
		    }
		    x = x_g;
		    nph  = nph_g;
		    nx   = nx_g;
		    nphi = np_g;
PR(ER, "nside=%d, nx=%d, np=%d, ntot_hp=%d, nx*np=%d\n",
nside, nx_g, np_g, nside*nside*12, nx_g*np_g);
		} else {
		    if (spline_repixelization) {
			    calc2gl(f, nx, nphi, nph, x, x1, x2);
			    if (pol_map) {
				calc2gl(q, nx, nphi, nph, x, x1, x2);
				calc2gl(u, nx, nphi, nph, x, x1, x2);
			    }
		    } else {
			    /* reappropriation of pixel values */
			    approp_hpix(f, nx, nphi, nph, x, x1, x2);
			    if (pol_map) {
				approp_hpix(q, nx, nphi, nph, x, x1, x2);
				approp_hpix(u, nx, nphi, nph, x, x1, x2);
			    }
		    }
		}
		grid_flag = GRID_PHI_X_LEGENDRE;
	    }

	    if (!second_field) {
		register int i,j;
		double vmin= 1.e30, vmax= -1.e30;
		double umin= 1.e30, umax= -1.e30;
		double qmin= 1.e30, qmax= -1.e30;
		for (i=0;i<nx;i++) {
		   int ii=i*nphi;
		   for (j=0;j<nph[i];j++) {
		      if ( f[ii+j] < vmin) vmin = f[ii+j];
		      if ( f[ii+j] > vmax) vmax = f[ii+j];
		      if (pol_map) {
			    if ( q[ii+j] < qmin) qmin = q[ii+j];
			    if ( q[ii+j] > qmax) qmax = q[ii+j];
			    if ( u[ii+j] < umin) umin = u[ii+j];
			    if ( u[ii+j] > umax) umax = u[ii+j];
		      }
		   }
		}
		PR(ER, "output values:\n");
		PR(ER, "tmin=%15.8e, tmax=%15.8e\n", vmin, vmax);
		if (pol_map) {
		    PR(ER, "qmin=%15.8e, qmax=%15.8e\n", qmin, qmax);
		    PR(ER, "umin=%15.8e, umax=%15.8e\n", umin, umax);
		}
	    }
	}

	/* symmetric reflection */
	if( reflection ) {
	    /* meridian reflection */
	    if (reflection % 2 == 1) {  /* flag == 1 || == 3 */
		for(i=0; i<nx; i++) {
		    int np  = nph[i] - 1;
		    int np2 = nph[i] / 2;
		    int ii = i * nphi;
		    for(j=0; j<np2; j++) {
			int jn = np-j;
			double v = f[ii+j];
			f[ii+j]  = f[ii+jn];
			f[ii+jn] = v;
			if (pol_map) {
			    v = q[ii+j];  q[ii+j] = q[ii+jn];  q[ii+jn] = v;
			    v = u[ii+j];  u[ii+j] = u[ii+jn];  u[ii+jn] = v;
			}
		    }
		}
	    }
	    /* equator reflection */
	    if (reflection > 1)      {  /* flag == 2 || == 3 */
		int nx2 = nx / 2;
		for(i=0; i<nx2; i++) {
		    register int ii = i * nphi;
		    register int in = (nx-1-i) * nphi;
		    for(j=0; j<nph[i]; j++) {
			double v = f[in+j];
			f[in+j] = f[ii+j];
			f[ii+j] = v;
			if (pol_map) {
			    v = q[in+j];  q[in+j] = q[ii+j];  q[ii+j] = v;
			    v = u[in+j];  u[in+j] = u[ii+j];  u[ii+j] = v;
			}
		    }
		}
	    }
	}

	/**************/
	/* map output */
	/**************/
	if (gp_out) {
	    for(i=0; i<nx; i++) {
		int ii = i * nphi;
		int cf = (i+1 <nside || i+1 >nx-nside ? 1 : (i+2-nside)%2);
		double theta = acos(x[i]);
		double dphi = 2. * M_PI / nph[i];
		for(j=0; j<nph[i]; j++) {
		    double phi = j * dphi;
		    double p = phi + cf*dphi/2.;

		    if ( second_field )
			PR(fp_out, "%8.6f %8.6f %d\n",
				theta, p, (int)(f[ii+j]+.5));
		    else
			PR(fp_out, "%13.8f %13.8f  %15.8e\n",
				theta, p, f[ii+j]);
		}
	    }
	} else {
	    if ( set_Legendre_grid ) {
		/* select grid output keyword values */
		strcpy(x_type, "COS(THETA)");
		switch(grid_flag) {
		case GRID_PHI_X_LEGENDRE :
			strcpy(grid_type, "PHI(X)+LEGENDRE");
			break;
		default:
		case GRID_PHI_X:
			strcpy(grid_type, "PHI(X)+X");
			break;
		}

		/* write GLESP-file */
		if (ffile_flag) {
		    ff_write_ext2_map(f,x,nx,nphi,nph,
			    fp_out, x_type,grid_type);
		} else {
		    /* map output in FITS-format */
		    outp_glesp_map(f,x,nx,nphi,
			    nph, fp_out, x_type, grid_type);
		    if (pol_map == 1) {
			FILE *fp_p;
			char o_f[81];
			/* Q-map */
			strcpy(o_f, outp_file);
			strcat(o_f, "_q");
			if ((fp_p = fopen(o_f, "w")) == (FILE *)0) {
			  PR(ER,"%s: can not create file '%s'\n",myname,o_f);
			  exit(1);
			}
			outp_glesp_map(q,x,nx,nphi,
			    nph, fp_p, x_type, grid_type);
			fclose(fp_p);

			/* U-map */
			strcpy(o_f, outp_file);
			strcat(o_f, "_u");
			if ((fp_p = fopen(o_f, "w")) == (FILE *)0) {
			  PR(ER,"%s: can not create file '%s'\n",myname,o_f);
			  exit(1);
			}
			outp_glesp_map(u,x,nx,nphi,
			    nph, fp_p, x_type, grid_type);
			fclose(fp_p);
		    } else if (pol_map == 2) {
			FILE *fp_p;
			char o_f[81];
			/* Q-map */
			strcpy(o_f, outp_file);
			if ((fp_p = fopen(o_f, "w")) == (FILE *)0) {
			  PR(ER,"%s: can not create file '%s'\n",myname,o_f);
			  exit(1);
			}
			outp_glesp_3map(f,q,u, x, nx, nphi, nph, fp_p,
			       x_type,grid_type);
			fclose(fp_p);

		    }
		}
	    } else {
		for(i=0; i<nx; i++) {
		    int np  = nph[i] - 1;
		    int np2 = nph[i] / 2;
		    int ii = i * nphi;
		    for(j=0; j<np2; j++) {
			int jn = np-j;
			double v = f[ii+j];
			f[ii+j]  = f[ii+jn];
			f[ii+jn] = v;
		    }
		}

		/* Basic FITS output as 2-d map */
		if (ffile_flag) {
			ff_write_map2d(f, x, nx, nphi, nph, fp_out);
		} else {
			/* write ff-map */
			glesp_2_map2d(f, x, nx, nphi, nph, fp_out);
		}
	    }
	}
    }

    /* close output file */
    if (fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    char s1[50], s2[50], s3[30], s4[30];
    int flag_cr = 0;

    myname = argv[0];

    if (argc == 1)
	    usage();

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
	    switch(argv[i][1]) {
	    case 'b':
		set_Legendre_grid = 0;
		break;
	    case 'c':
		switch(argv[i][2]) {
		case 'i':
		     circle = 1;
		     break;
		case 'r': /* get value by coordinates */
		     get_2_str(argv[++i], s1, s2);
		     if(*s1 !='\0') lat0=deg2rad(s1); else lat0= -1001.0;
		     if(*s2 !='\0') lon0=deg2rad(s2); else lon0= -1001.0;
		     if (argv[i-1][3] == 'p') {
			    lat0 = M_PI/2 - lat0;
		     }
		     flag_cr = 1;
		     break;
		case 's':  /* section */
		     cs.ra  = str2deg((skip_wd(0, argv[++i], s1),s1));
		     cs.dec = str2deg((skip_wd(1, argv[i], s1),s1));
		     cs.st  = str2deg((skip_wd(2, argv[i], s1),s1));
		     cs.n   = atoi((skip_wd(3, argv[i], s1),s1));
		     *s1 = '\0';
		     cs.st2 = str2deg((skip_wd(4, argv[i], s1),s1));
		     *s1 = '\0';
		     cs.n2  = atoi((skip_wd(5, argv[i], s1),s1));
		     break;
		default :
		     second_field = 1;
		     break;
		}
		break;
	    case 'C':
		    get_val_str(argv[++i], &cut_low, &cut_up);
		    cut_values = 1;
		    break;
	    case 'd':
		    if (strcmp(argv[i]+1, "d") == 0) {
			    dummy_value = atof(argv[++i]);
		    } else if (strcmp(argv[i]+1, "double") == 0) {
			    bitpix_map = DOUBLEPIX;
		    } else {
			    PR(ER,"%s: illegal flag '%s'\n",
				myname, argv[i]);
			    usage();
		    }
		    break;
	    case 'e':
		    if (strncmp("eq", argv[i]+1, 2) == 0) {
			    eq_input = 1;
		    } else {
			    equiv_flag = 1;
		    }
		    break;
	    case 'f':
		    if (strncmp("fc", argv[i]+1, 2) == 0) {
			    conv_factor = atof(argv[++i]);
		    } else if (strcmp("float", argv[i]+1) == 0) {
			bitpix_map = FLOATPIX;
		    } else {
			PR(ER,"%s: illegal flag '%s'\n",myname,argv[i]);
			usage();
		    }
		    break;
	    case 'g':
		    gp_out = 1;
		    break;
	    case 'h':
		    usage();
		    break;
	    case 'i':
		    if (strcmp(argv[i]+1, "if") == 0) {
			/* input f-file for cut */
			inp_ff_flag = 1;
			strcpy(cr_file, argv[++i]);
		    } else {
			if (strcmp(argv[i]+1, "ib") == 0)
			    /* basic FITS for Hpix format */
			    flag_bf_hp=1;
			strcpy(filename, argv[++i]);
		    }
		    break;
	    case 'k':
		    if (strncmp("ke", argv[i]+1, 2) == 0) {
			 get_4_str(argv[++i], s1, s2, s3, s4);
			 if(*s1 !='\0') latR0=deg2rad(s1);else latR0= -1000.;
			 if(*s2 !='\0') lonR0=deg2rad(s2);else lonR0= -1000.;
			 if(*s3 !='\0') latR1=deg2rad(s3);else latR1= -1000.;
			 if(*s4 !='\0') lonR1=deg2rad(s4);else lonR1= -1000.;
			 remain_flag = 1;
		    } else {
			 PR(ER, "%s: illegal flag '%s'\n",
			      myname, argv[i]);
			 usage();
		    }
		    break;
	    case 'l':
		    if(strncmp("lma",argv[i]+1,3) == 0) {
			    lmax_g = atoi(argv[++i]);
		    } else
			    set_Legendre_grid = 1;
		    break;
	    case 'm':
		    strcpy(mask_file, argv[++i]);
		    break;
	    case 'n':
		    switch(argv[i][2]) {
		    case 'x':
			    nx_g = atoi(argv[++i]);
			    break;
		    case 'p':
			    np_g = atoi(argv[++i]);
			    break;
		    default:
			    usage();
			    break;
		    }
		    break;
	    case 'o':
		    if(strncmp("of", argv[i]+1, 2) == 0) {
			    ffile_flag = 1;
		    }
		    strcpy(outp_file, argv[++i]);
		    break;
	    case 'p':
		    if (strcmp(argv[i]+1, "pol3") == 0 ||
			strcmp(argv[i]+1, "p3") == 0) {
			pol_map = 2;
		    } else if (strncmp(argv[i]+1, "po", 2) == 0) {
			pol_map = 1;
		    } else
			gp_out = 1;
		    break;
	    case 'r':
		    switch(argv[i][2]) {
		    case 'a':
			 radius = deg2rad(argv[++i]);
			 circle = 1;
			 break;
		    case 'm':
			 get_4_str(argv[++i], s1, s2, s3, s4);
			 if(*s1 !='\0') latR0=deg2rad(s1);else latR0= -1001.;
			 if(*s2 !='\0') lonR0=deg2rad(s2);else lonR0= -1001.;
			 if(*s3 !='\0') latR1=deg2rad(s3);else latR1= -1001.;
			 if(*s4 !='\0') lonR1=deg2rad(s4);else lonR1= -1001.;
			 remove_flag = 1;
			 break;
		    case '\0':
		    case '1':
			 if (strcmp(argv[i]+1, "r12") == 0)
				reflection = 3;
			 else {
			      if (reflection == 2)
				     reflection = 3;
			      else
				     reflection = 1;
			 }
			 break;
		    case '2':
			 if (strcmp(argv[i]+1, "r21") == 0)
				reflection = 3;
			 else {
			      if (reflection == 1)
				     reflection = 3;
			      else
				     reflection = 2;
			 }
			 break;
		    default :
			 PR(ER,"%s: illegal flag '%s'\n",myname,argv[i]);
			 usage();
			 break;
		    }
		    break;
	    case 's':
		    switch(argv[i][2]) {
		    case 'q': /* get value by coordinates */
			 get_4_str(argv[++i], s1, s2, s3, s4);
			 if(*s1 !='\0') lat0=deg2rad(s1); else lat0= -1001.;
			 if(*s2 !='\0') lon0=deg2rad(s2); else lon0= -1001.;
			 if(*s3 !='\0') lat1=deg2rad(s3); else lat1= -1001.;
			 if(*s4 !='\0') lon1=deg2rad(s4); else lon1= -1001.;
			 break;
		    case 'p':
			  spline_repixelization = 1;
			 break;
		    case 'f': /* name of file with coordinates list */
			 if (strncmp(argv[i]+1, "sfg", 3) == 0)
				galactic_coors = 1;
			 else if (strncmp(argv[i]+1, "sfe", 3) == 0)
				equat_coors = 1;
			 strcpy(cr_file, argv[++i]);
			 break;
		    case '\0':
			 F_sun_data_key = !F_sun_data_key;
			 break;
		    default :
			 PR(ER,"%s: illegal flag '%s'\n",myname,argv[i]);
			 usage();
			 break;
		    }
		    break;
	    case 't':
		    if (strcmp(argv[i]+1, "test") == 0) {
			    test_flag = 1;
		    } else if (strcmp(argv[i]+1, "tmp") == 0) {
			    strcpy(tmpcat, argv[++i]);
		    } else
			    usage();
		    break;
	    case 'z':
		    if (strncmp(argv[i]+1, "zc", 2) == 0) {
			get_2_str(argv[++i], s1, s2);
			if(*s1 !='\0') latR0=deg2rad(s1); else latR0= -1001.;
			if(*s2 !='\0') lonR0=deg2rad(s2); else lonR0= -1001.;
			circle = 1;
		    } else {
			/* file with list of zones to be processed */
			strcpy(zone_file, argv[++i]);
		    }
		    break;
	    default:
		    break;
	    }
	    break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default :
			usage();
			break;
		}
		break;
	default:
		strcpy(filename, argv[i]);
		break;
	}
    }
    if (flag_cr && circle) {
	latR0 = lat0;
	lonR0 = lon0;
    }


    return;
}

void usage()
{
PR(ER, "%s usage: -i[b] hpix_file [[-o] out_tac_file] [-l(set Legendre grid)] [-test]\n",
myname);
PR(ER, "[-sp(spline_repixelization)] [-gp(print out theta,phi,T))] [-ff] [-bf]\n");
PR(ER, "[-cn(counts field output)] [-m mask_file(FITS or ASCII)]\n");
PR(ER, "[-sq lat1,lon1 lat2,lon2] [-sf file_to_select_by_lat,lon]\n");
PR(ER, "[-{rm,keep} lat1,lon1 lat2,lon2(remove/keep sky zone)] [-d dummy]\n");
PR(ER, "[-C cut_low,cut_up] [-cr lat,lon] [-r{12}(reflection)] [-{float,double}]\n");
PR(ER, "[-cs ra,dec,dra,n1(cross-section by RA),ddec,n2] [-eq] [-fc factor]\n");
PR(ER, "[-lmax lmax] [-nx nx] [-np nphi] [-if ffile]\n");
PR(ER, "[-pol[3]]\n");
exit(2);
}
