#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"

#define PR  fprintf
#define ER  stderr

//struct _zone_ {
//     double latR0, lonR0;
//     double latR1, lonR1;
//     double radius;
//     int type;
//};

void  hmap_zone(zz, nz, function_type, f, x, offs, nx, np, nph, dummy_value)
struct _zone_ *zz;  /* zone coordinates          */
int nz;             /* number of zones           */
int function_type;  /* remain=0 or remove=1      */
double *f, *x, *offs;
int nx, np, *nph;
double dummy_value;
{
    register int iz, i, j;
    char *m = NULL;
    double lat0, lon0, lat1, lon1, radius;
    double theta0, theta1, phi0, phi1;
    double pi2 = 2. * M_PI;
    int nx0, nx1, np0=0, np1=0;
    int zone_type; /* circle=0 or rectangular=1 */

    /***********************************************/
    /* if save several zones,  we should set masks */
    /***********************************************/
    /* Basic formula for distance between points is
     *     d = arccos(cos(lon2-lon1)cos(lat1)cos(lat2)+sin(lat1)sin(lat2))
     * for polar coordinates"
     *     d = acos(cos(dphi)sin(th1)sin(th2)+cos(th1)*cos(th2))
     */
    if ( function_type == 0 ) { /* remain_flag */
	 m = (char *)calloc(nx*np, sizeof *m);
	 if (m == NULL) {
	     PR(ER, "%s: no memory for masking removed zones.", myname);
	     exit(1);
	 }
	 for(iz=0; iz<nz; iz++) {
	     double cth0, sth0, v;
	     zone_type = zz[iz].type; /* circle=0 or rectangular=1 */
	     if (zone_type == 0) {  /* circle */
	       int nx_c, np_c;   /* center of circle */
	       lat0  = zz[iz].latR0;
	       lon0  = zz[iz].lonR0;
	       radius = zz[iz].radius;

	       /* extract circle */
	       gal2pol(lat0, lon0, &theta0, &phi0);
	       cth0 = cos(theta0);
	       sth0 = sqrt(1-cth0*cth0);
	       nx_c = ind_x(nx, x, cth0);
	       np_c = phi0 / pi2 * (double)nph[nx_c];

	       nx0 = ind_x(nx, x, cos(theta0-radius));
	       nx1 = ind_x(nx, x, cos(theta0+radius));
	       if ( nx0 > nx1 ) { j=nx0; nx0=nx1; nx1=j; }

	       if (theta0-radius < 0. || theta0+radius > M_PI ) {
		 /* process pixels near pole */
		 for(i=0; i<nx; i++) {
		   int ii = i * np;
		   double cth = x[i], sth = sqrt(1.-cth*cth);
		   double dphi = pi2 / (double)nph[i];
		   double dp;   /* dt = theta0 - acos(cth), */

		   if (!((theta0+radius > M_PI &&
		     ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))) ||
		     (theta0-radius < 0.   &&
		     ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))))) {
			for(j=0; j<nph[i]; j++) {
			    dp = phi0 - j*dphi - offs[i];
			    v = acos(cos(dp)*sth0*sth + cth0*cth);
			    if (v <= radius)  {
				    m[ii+j] = 1;
				    continue;
			    }
			}
		   }
		 }
	       } else {
		  for(i=0; i<nx; i++) {
		    int ii = i * np;
		    double cth = x[i],  sth = sqrt(1-cth*cth);
		    double dphi = pi2 / (double)nph[i];
		    double dp; /* dt = theta0 - acos(cth), */
		    /* make checking interval a little larger */
		    double rl = radius/sth0/sth;

		    if (i >= nx0 && i <= nx1) {
			np0 = (phi0-rl-offs[i]) / dphi;
			np1 = (phi0+rl+offs[i]) / dphi;
			if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

			if (np0 < 0) {
			     int npR = nph[i] + np0;
			     for(j=0; j<nph[i]; j++) {
			       if (npR <= j || j < np1) {
				 dp = (2*M_PI+phi0) - j*dphi - offs[i];
				 v = acos(cos(dp)*sth0*sth + cth0*cth);
				 if (v <= radius) {
					 m[ii+j] = 1;
					 continue;
				 }
			       }
			     }
			}
			if (np1 > nph[i]) {
			    int npR = np1 - nph[i];
			    for(j=0; j<nph[i]; j++) {
			      if (npR < j || j < np0) {
				dp = -phi0 + j*dphi + offs[i] + 2*M_PI;
				v = acos(cos(dp)*sth0*sth + cth0*cth);
				if (v <= radius) {
					m[ii+j] = 1;
					continue;
				}
			      }
			    }
			}
			for(j=0; j<nph[i]; j++) {
			    if (j >= np0 && j <= np1) {
				  dp = phi0 - j*dphi - offs[i];
				  v = acos(cos(dp)*sth0*sth + cth0*cth);
				  if (v <= radius) {
					  m[ii+j] = 1;
					  continue;
				  }
			    }
			}
		    }
		  }
	       }
	     } else {
	       lat0  = zz[iz].latR0;
	       lon0  = zz[iz].lonR0;
	       lat1  = zz[iz].latR1;
	       lon1  = zz[iz].lonR1;

	       /* set mask for rectangular */
	       gal2pol(lat0, lon0, &theta0, &phi0);
	       gal2pol(lat1, lon1, &theta1, &phi1);
	       nx0 = ind_x(nx, x, cos(theta0));
	       nx1 = ind_x(nx, x, cos(theta1));

	       if (nx0 > nx1)  {i=nx0; nx0=nx1; nx1=i;}

	       for(i=0; i<nx; i++) {
		    int ii = i * np;
		    if (i >= nx0 && i < nx1) {
			np0 = phi0 / pi2 * (double)nph[i];
			np1 = phi1 / pi2 * (double)nph[i];
			if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }
			/* process pixels near center */
			if (np0 < 0) {
			      int npR = nph[i] + np0;
			      for(j=0; j<nph[i]; j++) {
				    if (j <= np1 || j >= npR) {
					    m[ii+j] = 1;
					    continue;
				    }
			      }
			}
			if (np1 >= nph[i]) {
			    int npR = np1 - nph[i];
			    for(j=0; j<nph[i]; j++) {
				    if (j >= np0 || j <= npR) {
					    m[ii+j] = 1;
					    continue;
				    }
			    }
			}
			for(j=0; j<nph[i]; j++) {
			    if (j >= np0 && j <= np1) {
				m[ii+j] = 1;
				continue;
			    }
			}
		    }
	       }
	     }
	 }
    }

    /******************/
    /* process zones  */
    /******************/
    /* Basic formula for distance between points is
     *     d = arccos(cos(lon2-lon1)cos(lat1)cos(lat2)+sin(lat1)sin(lat2))
     * for polar coordinates"
     *     d = acos(cos(dphi)sin(th1)sin(th2)+cos(th1)*cos(th2))
     */

    for(iz=0; iz<nz; iz++) {
     zone_type = zz[iz].type; /* circle=0 or rectangular=1 */
     if (zone_type == 0) { /* circle */
       int nx_c, np_c;   /* center of circle */
       double cth0, sth0, v;

       lat0  = zz[iz].latR0;
       lon0  = zz[iz].lonR0;
       radius = zz[iz].radius;

       /* extract circle */
       gal2pol(lat0, lon0, &theta0, &phi0);
       cth0 = cos(theta0);
       sth0 = sqrt(1.-cth0*cth0);
       nx_c = ind_x(nx, x, cth0);
       np_c = phi0 / pi2 * (double)nph[nx_c];

       nx0 = ind_x(nx, x, cos(theta0-radius));
       nx1 = ind_x(nx, x, cos(theta0+radius));
       if ( nx0 > nx1 ) { j=nx0; nx0=nx1; nx1=j; }

       if (theta0-radius < 0. || theta0+radius > M_PI ) {
	 /* process pixels near pole */
	 for(i=0; i<nx; i++) {
	    int ii = i * np;
	    double cth = x[i], sth = sqrt(1.-cth*cth);
	    double dphi = pi2 / (double)nph[i];
	    double dp;  /* dt = theta0 - acos(cth), */

	    if ( function_type == 1) {  /* remove_flag */
		if (theta0+radius > M_PI) {
		    if ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))
			  continue;
		} else {
		    if ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))
			  continue;
		}

		for(j=0; j<nph[i]; j++) {
		    dp = phi0 - j*dphi - offs[i];
		    v = acos(cos(dp)*sth0*sth + cth0*cth);
		    if (v > radius)
			    continue;
		    f[ii+j] = dummy_value;
		}
	    }

	    if ( function_type == 0) {  /* remain_flag */
		if (!((theta0+radius > M_PI &&
		    ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))) ||
		    (theta0-radius < 0.   &&
		    ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))))) {
			for(j=0; j<nph[i]; j++) {
			    dp = phi0 - j*dphi - offs[i];
			    v = acos(cos(dp)*sth0*sth + cth0*cth);
			    if (v <= radius)
				    continue;
			    if (m[ii+j] != 1)  f[ii+j] = dummy_value;
			}
		} else {
		     for(j=0; j<nph[i]; j++)
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		}
	    }
	 }
       } else {
	 for(i=0; i<nx; i++) {
	      int ii = i * np;
	      double cth = x[i], sth = sqrt(1-cth*cth);
	      double dphi = pi2 / (double)nph[i];
	      double dp;  /* dt = theta0 - acos(cth),  */
	      /* make checking interval a little larger */
	      double rl = radius/sth0/sth;

	      if ( function_type == 1 ) {  /* remove_flag */
		     if (i < nx0 || i > nx1)
			     continue;
		     np0 = (phi0-rl-offs[i]) / dphi;
		     np1 = (phi0+rl+offs[i]) / dphi;
		     if ( np1 < np0 ) { j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			 int npR = nph[i] + np0;
			 for(j=0; j<nph[i]; j++) {
			     if (np1<j  && j < npR)
				     continue;
			     dp = (2*M_PI+phi0) - j*dphi - offs[i];
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (v > radius)
				     continue;
			     f[ii+j] = dummy_value;
			 }
		     }
		     if (np1 >= nph[i]) {
			 int npR = np1 - nph[i];
			 for(j=0; j<nph[i]; j++) {
			     if (npR < j && j < np0)
				     continue;
			     dp = -phi0 + j*dphi + offs[i] + 2*M_PI;
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (v > radius) {
				     continue;
			     }
			     f[ii+j] =dummy_value;
			 }
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if (j < np0 || j > np1)
				 continue;

			 dp = phi0 - j*dphi - offs[i];
			 v = acos(cos(dp)*sth0*sth + cth0*cth);
			 if (v > radius)
				 continue;
			 f[ii+j] = dummy_value;
		     }
	      }
	      if ( function_type == 0 ) { /* remain_flag */
		 if (i >= nx0 && i <= nx1) {
		     np0 = (phi0-rl) / dphi;
		     np1 = (phi0+rl) / dphi;
		     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			 int npR = nph[i] + np0;
			 for(j=0; j<nph[i]; j++) {
			   if (npR < j || j < np1) {
			     dp = 2*M_PI-phi0 + j*dphi + offs[i];
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (v <= radius)  {
				     continue;
			     }
			   }
			   if (m[ii+j] !=1) f[ii+j] =dummy_value;
			 }
		     }
		     if (np1 >= nph[i]) {
			 int npR = np1 - nph[i];
			 for(j=0; j<nph[i]; j++) {
			   if (npR <= j || j < np0) {
			     dp = phi0 + j*dphi + offs[i] - 2*M_PI;
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (v <= radius)  {
				     continue;
			     }
			   }
			   if (m[ii+j] !=1) f[ii+j] =dummy_value;
			 }
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if (j >= np0 && j <= np1) {
			       dp = phi0 - j*dphi - offs[i];
			       v = acos(cos(dp)*sth0*sth + cth0*cth);
			       if (v <= radius)  {
				       continue;
			       }
			 }
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		     }
		 } else {
		     for(j=0; j<nph[i]; j++)
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		 }
	      }
	 }
       }
     } else {
	 lat0  = zz[iz].latR0;
	 lon0  = zz[iz].lonR0;
	 lat1  = zz[iz].latR1;
	 lon1  = zz[iz].lonR1;

	 /* extract rectangular */
	 gal2pol(lat0, lon0, &theta0, &phi0);
	 gal2pol(lat1, lon1, &theta1, &phi1);
	 nx0 = ind_x(nx, x, cos(theta0));
	 nx1 = ind_x(nx, x, cos(theta1));

	 if (nx0 > nx1)  {i=nx0; nx0=nx1; nx1=i;}

	 for(i=0; i<nx; i++) {
	      int ii = i * np;
	      if ( function_type == 1 ) {  /* remove_flag */
		     if (i < nx0 || i >= nx1)
			     continue;
		     np0 = phi0 / pi2 * (double)nph[i];
		     np1 = phi1 / pi2 * (double)nph[i];

		     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			 int npR = nph[i] + np0;
			 for(j=0; j<nph[i]; j++) {
				 if (j > np1  && j < npR)
					 continue;
				 f[ii+j] = dummy_value;
			 }
		     }
		     if (np1 >= nph[i]) {
			 int npR = np1 - nph[i];
			 for(j=0; j<nph[i]; j++) {
				 if (j < np0 && j > npR)
					 continue;
				 f[ii+j] = dummy_value;
			 }
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if (j < np0 || j > np1)
				 continue;
			 f[ii+j] = dummy_value;
		     }
	      }
	      if ( function_type == 0 ) { /* remain_flag */
		 if (i >= nx0 && i < nx1) {
		     np0 = phi0 / pi2 * (double)nph[i];
		     np1 = phi1 / pi2 * (double)nph[i];
		     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			   int npR = nph[i] + np0;
			   for(j=0; j<nph[i]; j++) {
				 if (j <= np1 || j >= npR)
				    continue;
				 if (m[ii+j] !=1) f[ii+j] =dummy_value;
			   }
		     }
		     if (np1 >= nph[i]) {
			 int npR = np1 - nph[i];
			 for(j=0; j<nph[i]; j++) {
				 if (j >= np0 || j <= npR)
				    continue;
				 if (m[ii+j] !=1) f[ii+j] =dummy_value;
			 }
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if (j >= np0 && j <= np1)
				 continue;
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		     }
		 } else {
		     for(j=0; j<nph[i]; j++)
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		 }
	      }
	 }
       }
    }

    if (m != NULL) free((void *)m);
    return;
}

