#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/*
 * convert 'nested/ pixel number to 'ring' pixel number
 */
extern void n2r(int, int, int *);
extern void init_pix2xy(int *, int *);

int convert_nested2ring(pix_number, nside)
int pix_number;  /* 'nested'-pixel number */
int nside;
{
      int pix;
      n2r(nside, pix_number, &pix);
      return pix;
}

/*
 * modified 'HEALPix' procedures to convert nested scheme to 'ring'
 * See 'HEALPix' description.
 */
void n2r(nside, n_nest, n_ring)
int  nside;
int  n_nest;
int *n_ring;
{
    int ntot, np_face, face_number, ncap, n_b;
    int p_f, p_low, p_trunc, p_med, p_hi;
    int xp, yp, crt, cr, nr, cpt, cp, sh=0, nside4;
    int nside_max=8192;
    static int init_pix2xy_flag = 0;
    static int p2x[1024], p2y[1024];
    int crl[12], cpl[12]; /* coordinate of the lowest corner of each face */

    /* set values for working arrays */
    crl[0] = crl[1] = crl[2]  = crl[3]  = 2;
    crl[4] = crl[5] = crl[6]  = crl[7]  = 3;
    crl[8] = crl[9] = crl[10] = crl[11] = 4;
    cpl[0] = cpl[8] =1;   cpl[4]=0;
    cpl[1] = cpl[9] =3;   cpl[5]=2;
    cpl[2] = cpl[10]=5;   cpl[6]=4;
    cpl[3] = cpl[11]=7;   cpl[7]=6;

    if( nside<1 || nside>nside_max ) {
	   fprintf(stderr, "convert: nside = %d out of range (0<nside<%d)\n",
		nside, nside_max);
	   exit(1);
    }
    ntot = 12 * nside*nside;
    if( n_nest<0 || n_nest> ntot-1 ) {
	   fprintf(stderr, "'nest' number (%d) out of range (0<nside<%d)\n",
		n_nest, ntot);
	   exit(1);
    }

    if (!init_pix2xy_flag) {init_pix2xy(p2x, p2y); init_pix2xy_flag = 1;}

    /* number of points in the North Polar cap  */
    ncap   = 2*nside*(nside-1);
    nside4 = 4*nside;

    /*
     *    finds the face, and the number in the face
     */
    np_face = nside*nside;

    face_number = n_nest/np_face;
    /* pixel number in the face {0,np_face-1} */
    p_f = (int)fmod(n_nest, np_face);

    /*
     *     finds the x,y on the face (starting from the lowest corner)
     *     from the pixel number
     */
    p_low = (int)fmod(p_f,1024);     /* last 10 bits         */
    p_trunc =   p_f/1024;            /* truncate ast 10 bits */
    p_med = (int)fmod(p_trunc,1024); /* next 10 bits         */
    p_hi  =     p_trunc/1024;        /* high weight 10 bits  */

    /* count << "xp = " << xp << " yp = " << yp << endl; */
    xp = 1024*p2x[p_hi] + 32*p2x[p_med] + p2x[p_low];
    yp = 1024*p2y[p_hi] + 32*p2y[p_med] + p2y[p_low];

    /* transforms this in (horizontal, vertical) coordinates */
    crt = xp + yp;  /*   'vertical' in {0,2*(nside-1)}       */
    cpt = xp - yp;  /*   'horizontal' in {-nside+1,nside-1}  */

    /*
     *  computes the z coordinate on the sphere
     *  cr =  crl[face_number+1]*nside - crt - 1;
     *                ring number in {1,4*nside-1}
     */
    cr =  crl[face_number]*nside - crt - 1;

    if( cr < nside ) { /* north pole region */
	nr = cr;
	n_b = 2 * nr * (nr - 1);
    } else if( cr > 3*nside ) { /* south pole region */
	nr = nside4 - cr;
	n_b = ntot - 2 * (nr + 1) * nr;
    } else {
	nr = nside;   /*  equatorial region */
	n_b = ncap + nside4 * (cr - nside);
	sh = (int)fmod(cr - nside, 2);
    }

    /*
     *  computes the phi-coordinate on the sphere, in [0,2pi]
     *  cp = (cpl[face_number+1]*nr + cpt + 1 + sh)/2;
     *             'phi' number in the ring in {1,4*nr}
     */
    cp = (cpl[face_number]*nr + cpt + 1 + sh)/2;

    if( cp>nside4 ) cp -= nside4;
    if( cp<1 )      cp += nside4;

    *n_ring = n_b + cp - 1;  /*  in [0,ntot-1] */

    return;
}

/*
 * the array giving x and y in the face from pixel number
 * for the nested (quad-cube like) ordering of pixels
 *
 * the bits corresponding to x and y are interleaved in the pixel number
 * one breaks up the pixel number by even and odd bits
 */
void init_pix2xy(x, y)
int *x;
int *y;
{
  register int i, j;
  int xp, yp, bp, bv;

  for( i=0; i<1024; i++ ) {
    j = i;
    xp = yp = 0;
    bp = 1 ; /* bit position (in x and y)  */
    while ( j ) { /* loop over all the bits    */
      /* bit values */
      bv = (int)fmod(j,2);  j /= 2;   xp += bv * bp;
      bv = (int)fmod(j,2);  j /= 2;   yp += bv * bp;
      bp *= 2; /* next bit */
    }
    x[i] = xp;   /*  in 0,31  */
    y[i] = yp;   /*  in 0,31  */
  }
  
  return;
}
