#include <stdio.h>
#include <math.h>

#define PR      fprintf
#define ER      stderr
#define ABS(x)  ((x) < 0 ? -(x) : (x))

extern void gauleg(double, double, double *, double *, int);
extern void coord(int *, int *, double *, double *, double *,
	   double *, double *, int *, int *, int *);

/* select and calculate grid for Legendre polynomials */
void coord_(coord_flag, fft_flag, lx_add, x1, x2, phi2, x, pw, Nx, Nph, Nphi)
int *fft_flag;
int *lx_add;  /* addition to Nphi numbers */
double *x1, *x2,
      *phi2;
double *x,  *pw;
int *coord_flag; /* flag of coordinate grid selection */
int    *Nx, *Nphi;
int *Nph;
{
	register int i;
	double xx, pp = 0.;
	double dx, dphi;
	int NNphi = 0;

	if(*coord_flag < 0) {
		/* uniform grid */
		dx = (*x2 - *x1) / (*Nx -1.0);
		xx = *x1;
		for(i=0; i< *Nx; i++) {
			pw[i] = dx / 1.5; /* for Sympson integral */
			x[i] = xx;
			xx += dx;
		}
		pw[0]    = pw[0]    / 2.;
		pw[*Nx-1] = pw[*Nx-1] / 2.;

		pp = pw[0]+pw[*Nx-1];
		for(i=1; i< *Nx-1; i+=2) {
			pw[i] += pw[i];
			pp += pw[i]+pw[i+1];
		}
	} else {  /* special grid for Legendre polynomials */
		gauleg(*x1, *x2, x, pw, *Nx);
		pp=0.;
		for(i=0; i< *Nx; i++) {
		    pp += pw[i];
		}
	}
	if( x[*Nx-1] > 1.0)   x[*Nx-1] = 1.0;

	if(fft_flag ) {
	       /*
		*   arbitrary grid for sphere and fast-fourier
		*/
		int Imx = (*Nx+1)/2;
		int lx = lx_add;

		/* old value: 4, starting length, near Pole */

		/* Nph[0] = Nphi; */
		switch(lx) {
		case 0:
			Nph[0] = *Nphi;
			break;
		default:
			Nph[0] = 10;
			break;
		}
		Nph[*Nx-1] = Nph[0];
		NNphi = Nph[0] + Nph[*Nx-1];
		dphi = *phi2 / (Nph[0]-0.0);
		pw[0]    = dphi * pw[0];
		pw[*Nx-1] = dphi * pw[*Nx-1];

		for(i=1; i<Imx; i++) {
			int jx;
			/* Nph[i] = Nph[0]; */
			if (lx == 0)
				Nph[i] = *Nphi;
			else
				Nph[i] = Nph[i-1] + lx;

			if(Nph[i] > *Nphi) Nph[i] = *Nphi;
			dphi  = *phi2 / (Nph[i]-0.0);
			pw[i] *= dphi;
			NNphi += Nph[i];

			jx = *Nx - i - 1;
			if(jx > i) {
				pw[jx]  *= dphi;
				Nph[jx] = Nph[i];
				NNphi += Nph[jx];
			}
		}
	} else {
	       /*
		*   Number of cells over phi for a given x
		*   homogeneous grid for rectangular area
		*/
		Nph[0] = *Nphi;
		dphi = *phi2 / (Nph[0]-0.0);
		for (i=0; i< *Nx; i++) {
			Nph[i] = Nph[0];
			pw[i]  = pw[i] * dphi;
		}
	}

	pp=0.;
	for(i=0; i< *Nx; i++) {
	    pp += pw[i] * Nph[i];
	}

	return;
}

/* special grid for Legendre polynomials */
void gauleg(x1,x2,x,w,n)
double x1, x2;
double *x, *w;
register int n;
{
	double eps = 3.e-15,
	       pi = M_PI;
	double p1, p2, p3, pp, xm, xl, z, z1;
	register int i,j,m;

	m = (n+1) / 2;
	xm = 0.5 * (x2 + x1);
	xl = 0.5 * (x2 - x1);

	for (i=0; i<m; i++) {
		/* FORTRAN: z=cos(pi*(i-.25d0)/(n+.5d0)) */
		z = cos(pi * ((i+1)-.25) / (n+.50));
lab_1:
		p1 = 1.0;
		p2 = 0.0;
		for(j=0; j<n; j++) {
		    p3 = p2;
		    p2 = p1;
		    /* FORTRAN: p1=((2.d0*j-1.d0)*z*p2-(j-1.d0)*p3)/j */
		    p1 = ((2*j+1)*z*p2-j*p3)/(j+1);
		}
		pp = n*(z*p1-p2)/(z*z-1.0);
		z1 = z;
		z = z1-p1/pp;

		if(ABS(z-z1) > eps)  goto lab_1;

		x[i] = xm - xl * z;
		/* FORTAN: x(n+1-i)=xm+xl*z */
		x[n-1-i] = xm + xl * z;
		if (w != (double *)0) {
			w[i] = 2.0*xl / ((1. - z*z) * pp*pp);
			w[n-1-i] = w[i];
		}
	}

	return;
}

/* select and calculate weights at grid for Legendre polynomials */
void weights(coord_flag, fft_flag, x1, x2, phi2, x, pw, Nx, Nph)
int fft_flag;
double x1, x2,
      phi2;
double *x,  *pw;
int coord_flag; /* flag of coordinate grid selection */
int    Nx;
int *Nph;
{
	register int i;
	double xx;
	double dx, dphi;

	if(coord_flag < 0) {
		/* uniform grid */
		dx = (x2-x1) / (Nx -1.0);
		xx = x1;
		for(i=0; i<Nx; i++) {
			pw[i] = dx / 1.5; /* for Sympson integral */
			x[i] = xx;
			xx += dx;
		}
		pw[0]    = pw[0]    / 2.;
		pw[Nx-1] = pw[Nx-1] / 2.;

		for(i=1; i<Nx-1; i+=2) {
			pw[i] += pw[i];
		}
	} else {  /* special grid for Legendre polynomials */
		gauleg(x1, x2, x, pw, Nx);
	}
	if(fft_flag ) {
		int Imx = (Nx+1)/2;
		dphi = phi2 / (Nph[0]-0.0);
		pw[0]    = dphi * pw[0];
		pw[Nx-1] = dphi * pw[Nx-1];

		for(i=1; i<Imx; i++) {
			int jx;
			dphi  = phi2 / (Nph[i]-0.0);
			pw[i] *= dphi;

			jx = Nx - i - 1;
			if(jx > i) {
				pw[jx]  *= dphi;
			}
		}
	} else {
		dphi = phi2 / (Nph[0]-0.0);
		for (i=0; i<Nx; i++) {
			pw[i]  = pw[i] * dphi;
		}
	}

	return;
}
