#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr

/* detect position of area */
void area_pos(f, nx, nphi, nph, x,  i_x, j_phi)
double *f;
int nx, nphi;
int *nph;
double *x;
int *i_x, *j_phi;
{
	register int i, j;
	double pi = M_PI, pi2 = 2 * M_PI;
	double *y = (double *)0;
	double f_s=0.0,
	       f_i =0.0, f_j=0.0,
	       f_a =0.0, f_b=0.0;
	int i_w=0, j_w=0;
	double a, b, a_c, b_c;
	double theta, phi, t_w, p_w;

/*
	if ((y = (double *)calloc(nphi, sizeof *y)) == (double *)0) {
		PR(ER, "%s: no memory for temporal buffer.\n", myname);
		exit(1);
	}
*/
	for(i=0; i<nx; i++) {
	   int ii = i * nphi;
	   double sth = sqrt(1. - x[i]*x[i]);
	   double dphi = pi2 / nph[i];
/*
	   for(j=0; j<nphi; j++)
		     y[j] = ff[ii+(int)((double)nphi/nph[i]+0.5)];
*/
	   for(j=0; j<nph[i]; j++) {
		double ff = f[ii+j];
		if ( fabs(ff) > 1.e-20 )  ff = 1.; else ff = 0.0;
		f_s += ff;
		a = cos(j*dphi) * sth;
		b = sin(j*dphi) * sth;
/*
		f_i += ff * i;
		f_j += ff * j;
*/
		f_i += ff * acos(x[i]);
		f_j += ff * j * dphi;
		f_a += ff * a;
		f_b += ff * b;
	   }
	}
	t_w = f_i / f_s + 0.5;
	p_w = f_j / f_s + 0.5;
	a_c = f_a / f_s;
	b_c = f_b / f_s;

	theta = asin(sqrt(a_c*a_c + b_c*b_c));
	phi = atan(b_c/a_c);

	if (t_w > M_PI/2)       theta =  pi - theta;

	if (b_c < 0. && a_c < 0.) {
		/* IV */
		phi = pi2 - phi;
	} else if (b_c > 0. && a_c < 0.) {
		/* II */
		phi = pi - phi;
	}

PR(ER, "thet=%g, phi=%g, t_w=%g, p_w=%g, x0=%g, x1=%g, ct=%g\n",
theta,phi, t_w, p_w, x[0], x[1], cos(theta));
	if (theta < 0.0)  theta += pi;
	if (phi   < 0.0)  phi   += pi2;

	angle2pixel(x, nx, nph, i_x, j_phi, theta, phi);

PR(ER, "i_w=%d,j_w=%d, a_c=%g,b_c=%g, \nt=%g,p=%g, i_x=%d,j_phi=%d\n",
i_w,j_w,a_c,b_c,
theta,phi, *i_x, *j_phi);
	if (y != (double *)0)  free((void *)y);

	return;
}

/* detect size of area */
/*
 *    B  = \sqrt {\frac{1}{I} (X_{t} - X)^{2} I_{t} }, \qquad t=1,r
*/
void area_size(f, nx, nphi, nph, x, i_c, j_c, dt1, dt2)
double *f;
int nx, nphi;
int *nph;
double *x;
int i_c, j_c;
double *dt1, *dt2;
{
    register int i, j;
    double pi2 = 2. * M_PI;
    double t_c = acos(x[i_c]),
	   p_c = j_c * pi2 / nph[i_c];
    double a_c = cos(p_c)*sin(t_c),
	   b_c = sin(p_c)*sin(t_c);
    double a, b, l_a, l_b, s_a=0., s_b=0., f_s=0.0, tt, dd;

    for(i=0; i<nx; i++) {
	int ii = i * nphi;
	double sth = sqrt(1. - x[i]*x[i]);
	double dphi = pi2 / nph[i];

	for(j=0; j<nph[i]; j++) {
	     double ff = f[ii+j];
	     double da, db;
	     if ( fabs(ff) > 1.e-20 )  ff = 1.; else ff = 0.0;
	     f_s += ff;
	     a = cos(j*dphi) * sth;
	     b = sin(j*dphi) * sth;

	     da = a - a_c;
	     db = b - b_c;
	     s_a += (da * da) * ff;
	     s_b += (db * db) * ff;
	}
    }
/*
PR(ER, "s_a=%g, s_b=%g, f_s=%g\n", s_a, s_b, f_s);
PR(ER, "t_c=%g, p_c=%g, a_c=%g, b_c=%g\n", t_c, p_c, a_c, b_c);
*/
    l_a  = sqrt (s_a / f_s);
    l_b  = sqrt (s_b / f_s);

    a = a_c - l_a;
    b = b_c;
    tt = asin(sqrt(a*a + b*b));
    if (t_c > M_PI/2)  tt = M_PI - tt;
    dd = acos(cos(tt)*x[i_c]+sin(tt)*sin(t_c));

/*
PR(ER, "tt_1=%g, dd_1=%g, dt1=%g\n", tt, dd, *dt1);
    *dt1 = fabs(asin(sqrt(a*a + b*b)) - t_c);
*/
    *dt1 = dd;

    a = a_c;
    b = b_c - l_b;
    tt = asin(sqrt(a*a + b*b));
    if (t_c > M_PI/2)  tt = M_PI - tt;
    dd = acos(cos(tt)*x[i_c]+sin(tt)*sin(t_c));
    *dt2 = dd;


    return;
}



void cosbell(f, nx, nphi, nph, x, ic, jc, dt1, dt2, wss,dir)
double *f;    /* (na,nb) */
double dt1, dt2;
double *x;
double *wss;
int dir;
int nx, nphi;
int *nph;
int ic, jc;
{
    register int i,j;
    double pi=M_PI, factor;
    double rt = dt1 < dt2 ? dt1 : dt2;
    int count = 0;

    rt *= 0.7;
    *wss=0.;

    for (i=0; i<nx; i++) {
	int ii = i * nphi;

	double t_c = acos(x[ic]);
	double t_i = acos(x[i]);

	if (fabs(t_i - t_c) > rt)
	    factor = 0.;
	else{
	    double ct = cos(pi*(t_i-t_c)/2./rt);
	    factor = ct * ct;
	}
	for (j=0; j<nph[i]; j++) {
/*
	    factor = cos((pi/na)*(i-ic)) * cos((pi/nb)*(j-jc));
*/
	    *wss += factor*factor;

	    if (fabs(factor) < 1.e-20 && dir == -1) {
		f[ii+j] = 0.;
	    } else {
		f[ii+j] *= (dir == -1 ? 1./factor : factor);
		count++;
	    }
	}
    }
/*
    *wss /= (double)(nx*ny);
*/
    if (count == 0) count++;
    *wss /= (double)count;

    return;
}
