#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#define PR  fprintf
#define ER  stderr
#define ST  stdout
#define NFF 12000
#define MAXFLOAT    1.e+30

/*
 * Program calculates some arithmetics over C(l)
 *
 *  created  by Oleg Verkhodanov 2009
 *  last changes 1/03/09, 26/03/09, 21/09/10, 2011, 2012, 05/04/2013
 */
extern void setopt(int, char **), usage(void), usage2(void);
extern void read_cl_dat(double **, int *, int *, char *, int);
extern void print_cl_dat(double *, int, int, FILE *, int);
extern void print_x_dat(double *, double *, int, FILE *);
extern void cl2dl(double *, int), dl2cl(double *, int);

char outfile[161];
char cl_file1[81],  cl_file2[81];
int lmin = -1, lmax = -1;
int lmin_print = -1, lmax_print = -1;
int one_column_input = 0;
int one_column_output = 0;
double c1=1.0, c2=1.0, cf = 1.0, cs=0.0;
int action = 0;        /* action: 0 - difference, 1 - sum      */
int cldl_flag = 0;     /* dl <--> cl */
int limits=0;          /* calculate limits by input sets of Cl */
int confid=0;          /* calculate probability                */
char **filename;
int nf = 0;            /* number of input Cl-sets              */
int stat_flag = 0;
int noneg =0;          /* no negative values in limits         */
double v_prob = 0.0;
int event_count=0;
char file_list[81];
char pfile[81];
char hg_file[161];     /* output file with  histogram     */
int n_hg = 0;          /* number of bins in histogram     */
double s_hg0 = 1.e+30, s_hgF= -1.e+30;
int hist_flag = 0;     /* flag for histogram  calculation */
int mix_flag = 0;      /* take 1st column from 1st file, 2nd from 2nd */
int n1_column = -1, n2_column = -1;
int ef1_flag = 0;      /* flag to set 1st column float    */
int bin_flag = 0;
int bin_window = 0;

int nf_m=0;
int ind_nf[NFF];
int sum_odd =0, sum_eve =0, sum_interp=0;
char *myname = "difcl";

int main(argc, argv)
int argc;
char **argv;
{
    register int l;
    FILE *fp_out = stdout, *fp_in;
    double *cl1 = (double *)0, *cl2 = (double *)0, *x = (double *)0;
    int lmin1 = -1, lmax1 = -1;
    int lmin2 = -1, lmax2 = -1;

    cl_file1[0] =  cl_file2[0] = outfile[0] = '\0';
    pfile[0] = file_list[0] = '\0';

    /* loop over input parameters */
    setopt(argc, argv);

    /*****************************************/
    /* calculate variation limits by Cl-sets */
    /*****************************************/
    if (limits || confid) {
	register int n_str = 0;
	double **data;
	char str[161], fn[161], *cs;
	int i, i_min = -1, i_max = -1, j;
	double *aver = (double *)0, *sigma = (double *)0,
	       max = -MAXFLOAT, min = MAXFLOAT;
	int *ll = (int *)0;
	FILE *fpl = (FILE *)0;

	if (*file_list != '\0') {
	    nf = NFF;
	    /* read filename from list in file */
	    if ((fpl = fopen(file_list, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'.\n", myname, file_list);
		exit(1);
	    }
	    while (((cs = fgets(fn, 160, fpl)) != NULL) && (*fn == '#' ||
		*fn == '\n' || *fn == '\r' || *fn == '\0')) ;
	    if (cs == NULL) {
		PR(ER, "%s: no accesible data in '%s'.\n", myname, file_list);
		exit(1);
	    }
	    if (fn[strlen(fn)-1] == '\n')
		fn[strlen(fn)-1] = '\0';
	}  else {
	    /* read filename from list in command string */
	    j=0;
	    strcpy (fn, filename[j]);
	}

	/*  get sizes of arrays from the 1st file (number of multipoles) */
	if ((fp_in = fopen(fn, "r")) == (FILE *)0) {
file_problem:
	    PR(ER, "%s: can not open file '%s'.\n", myname, fn);
	    exit(1);
	}

	while(fgets(str, 160, fp_in) != NULL) {
	    if (*str == '#' || *str == '\n' || *str == '\r')
		continue;
	    n_str++;
	}
	fclose(fp_in);

	/* create data arrays */
	if ((data = (double **)calloc((unsigned)nf, sizeof *data)) ==
	  (double **)0) {
nomem:
	    PR(ER, "%s: problems to find memory for array of %d elements.\n",
		myname, n_str);
	    exit(1);
	}
	for (i=0; i<nf; i++) {
	    if((data[i]=(double*)calloc(n_str, sizeof *data[0]))==(double*)0)
		goto nomem;
	}

	if((aver = (double *)calloc(n_str, sizeof *aver))==(double *)0)
		goto nomem;
	if((sigma = (double *)calloc(n_str, sizeof *sigma))==(double *)0)
		goto nomem;
	if((ll = (int *)calloc(n_str, sizeof *ll))==(int *)0)  goto nomem;
	if (one_column_output || ef1_flag)
	    if((x = (double *)calloc(n_str, sizeof *x))==(double *)0)
		    goto nomem;

	if (*file_list != '\0')  /* move pointer to begining of file list */
		fseek(fpl, 0l, 0);

	/* fill in arrays */
	for (j=0; j<nf; j++) {
	    int k;

	    if (*file_list != '\0') {
		while (((cs = fgets(fn, 160, fpl)) != NULL) && (*fn == '#' ||
		    *fn == '\n' || *fn == '\r' || *fn == '\0')) ;
		if (cs == NULL) {/* end of list */
		    break;
		}
		if (fn[strlen(fn)-1] == '\n')
		    fn[strlen(fn)-1] = '\0';
	    }  else {
		/* read filename from list in command string */
		strcpy (fn, filename[j]);
	    }
	    fp_in = fopen(fn, "r");

	    if (fp_in == (FILE *)0)
		goto file_problem;

	    i =  0;
	    k = -1;

	    while(fgets(str, 160, fp_in) != NULL) {
		double v, vv, vv2;
		if (*str == '#' || *str == '\n' || *str == '\r')
		    continue;

		if (i >= n_str) {
		    PR(ER, "varing number of strings in files.\n");
		    exit(1);
		}

		if (one_column_input) {
		    switch (one_column_input) {
		    case 1:
			sscanf(str, "%lf", &v);
			break;
		    case 2:
			sscanf(str, "%lf %lf", &vv, &v);
			break;
		    case 3:
			sscanf(str, "%lf %lf %lf", &vv, &vv2, &v);
			break;
		    }
		    k++;
		} else {
		    double kv;
		    sscanf(str, "%lf%lf", &kv, &v);
		    k = kv + .5;
		    if (ef1_flag)  x[i] = kv;
		}

		data[j][i] = v;
		ll[i] = k;
		if (one_column_output) {
		    if (one_column_output==1)
			x[i] = vv;
		    else
			x[i] = vv2;
		}
		i++;
	    }
	    fclose(fp_in);
	}

	if (j != nf) nf = j;

	if (*file_list != '\0')  /* close file list */
		fclose(fpl);

	PR(ER, "%d files to be processed.\n", nf);

	if (*outfile == '\0' || strcmp(outfile,"-") == 0
	   || strcmp(outfile,"stdout") == 0) {
	       fp_out = stdout;
	} else {
		/* open output file */
		if((fp_out = fopen(outfile, "w")) == (FILE *)0) {
		    PR(ER, "%s: can not create file '%s'\n",
			myname, outfile);
		    exit(1);
		}
	}

	/* calculate aver & sigma for each multipole l[i] */
	if (limits) {
	 for(i=0; i<n_str; i++) {
	    max = -MAXFLOAT, min = MAXFLOAT;
	    for(j=0; j<nf; j++) {
		aver[i] += data[j][i];
		if (min > data[j][i]) {min = data[j][i]; i_min = i;}
		if (max < data[j][i]) {max = data[j][i]; i_max = i;}
	    }
	    aver[i] /= nf;
	    for(j=0; j<nf; j++) {
		sigma[i] += (data[j][i] - aver[i]) * (data[j][i] - aver[i]);
	    }
	    sigma[i] = sqrt(sigma[i]/nf);

	    if (limits == 3) {
	       if (one_column_output)
		    PR(fp_out, "%12.5e  ", x[i]);
	       else {
		    if (ef1_flag) PR(fp_out,"%12.5e  ", x[i]);
		    else PR(fp_out, "%4d  ", ll[i]);
	       }
	       PR(fp_out, "%12.5e  %12.5e  %12.5e  %12.5e  ",
		   aver[i], sigma[i],
		   noneg && (aver[i]-sigma[i]<0.) ? 0. : aver[i]-sigma[i],
		   aver[i]+sigma[i]);
	       PR(fp_out, "%12.5e  %12.5e  %12.5e  %12.5e\n",
		  noneg && (aver[i]-2*sigma[i]<0.) ? 0. : aver[i]-2*sigma[i],
		  aver[i]+2*sigma[i],
		  noneg && (aver[i]-3*sigma[i]<0.) ? 0. : aver[i]-3*sigma[i],
		  aver[i]+3*sigma[i]);
	    } else {
	       if (one_column_output)
		    PR(fp_out, "%12.5e  ", x[i]);
	       else {
		    if (ef1_flag)
			PR(fp_out, "%12.5e  ", x[i]);
		    else
			PR(fp_out, "%4d  ", ll[i]);
	       }
	       PR(fp_out, "%12.5e  %12.5e  %12.5e  %12.5e\n",
		    aver[i], sigma[i],
		    noneg && (aver[i]-sigma[i]<0.) ? 0. : aver[i]-sigma[i],
		    aver[i]+sigma[i]);
	    }
	 }
	} else { /* if probability */
	    /* read file with limits */
	    if (*pfile != '\0') {
	       int k = -1;
	       FILE *fp;
	       if ((fp = fopen(pfile, "r")) == (FILE *)0) {
		   PR(ER, "%s: problem to open file '%s'\n", myname, pfile);
		   exit(1);
	       }
	       i = 0;
	       while (fgets(str, 160, fp) != NULL) {
		   double v,vv,vv2;
		   if (*str == '#' || *str == '\n' || *str == '\r')
		       continue;
		   switch (one_column_input) {
		   case 1:
		       sscanf(str, "%lf", &v);
		       k++;
		       break;
		   case 2:
		       sscanf(str, "%lf %lf", &vv, &v);
		       k++;
		       break;
		   case 3:
		       sscanf(str, "%lf %lf %lf", &vv, &vv2, &v);
		       k++;
		       break;
		   case 0:
		       {
			 double kv;
			 sscanf(str, "%lf%lf", &kv, &v);
			 k = kv + .5;
		       }
		       break;
		   }
		   sigma[i] = v;
		   i++;
	       }
	       fclose(fp);
	    } else {
	       /* set equal level for all multipoles */
	       for(i=0; i<n_str; i++) sigma[i] = v_prob;
	    }
	    for(i=0; i<n_str; i++) {
	       for(j=0; j<nf; j++) {
		   if (data[j][i] >  sigma[i])  aver[i] += 1.0;
	       }
	       if (!event_count)
		   aver[i] /= nf;

	       if (one_column_output)
		    PR(fp_out, "%12.5e  ", x[i]);
	       else
		    PR(fp_out, "%3d  ", ll[i]);
	       if (!event_count)
		   PR(fp_out, "%12.5e\n", aver[i]);
	       else
		   PR(fp_out, "%7d\n", (int)(aver[i]+.5));
	    }
	}
	if (fp_out != stdout)  fclose(fp_out);

	exit(0);
    }

    if (mix_flag) {
	extern void read_col_dat();
	lmin1=lmin2=0;
	if (n1_column < 1) n1_column = 1;
	if (n2_column < 1) n2_column = 1;

	read_col_dat(&cl1, &lmax1, cl_file1, n1_column);
	if (*cl_file2 == 0) {
	    PR(ER, "%s: flag '-mix' need 2 data files.\n", myname);
	    exit(1);
	}
	read_col_dat(&cl2, &lmax2, cl_file2, n2_column);
	if (lmin < 0) {
	    lmin = lmin1 > lmin2 ? lmin1 : lmin2;
	}
	if (lmax < 0) {
	    lmax = lmax1 < lmax2 ? lmax1 : lmax2;
	}

	if (lmin_print > 0)
	    if (lmin < lmin_print)  lmin = lmin_print;
	if (lmax_print > 0)
	    if (lmax > lmax_print)  lmax = lmax_print;

	if (*outfile == '\0' || strcmp(outfile,"-") == 0
	   || strcmp(outfile,"stdout") == 0) {
	       fp_out = stdout;
	} else {
		/* open output file */
		if((fp_out = fopen(outfile, "w")) == (FILE *)0) {
		    PR(ER, "%s: can not create file '%s'\n",
			myname, outfile);
		    exit(1);
		}
	}

	for (l=lmin; l<lmax; l++) {
	    PR(fp_out, "%12.5e  %12.5e\n", cl1[l], cl2[l]);
	}
	fclose(fp_out);
	exit(0);
    }

    /*************************/
    /* arithmetic operations */
    /*************************/

    /* read data */
    read_cl_dat(&cl1, &lmin1, &lmax1, cl_file1, one_column_input);

    if ( *cl_file2 != 0 && !stat_flag) {
	read_cl_dat(&cl2, &lmin2, &lmax2, cl_file2, one_column_input);
	if (lmin < 0) {
	    lmin = lmin1 > lmin2 ? lmin1 : lmin2;
	}
	if (lmax < 0) {
	    lmax = lmax1 < lmax2 ? lmax1 : lmax2;
	}

	if (lmin_print > 0)
	    if (lmin < lmin_print)  lmin = lmin_print;
	if (lmax_print > 0)
	    if (lmax > lmax_print)  lmax = lmax_print;


	/**************************************************/
	/* make difference, sum, devision, multiplication */
	/**************************************************/
	c1 *= cf;
	c2 *= cf;
	for(l=lmin; l<=lmax; l++) {
	    switch (action) {
	    case 1 :  /* sum */
		 cl1[l] = c1 * cl1[l] + c2 * cl2[l] + cs;
		 break;
	    case 2 :  /* multiplication */
		 cl1[l] = c1 * cl1[l] * c2 * cl2[l] + cs;
		 break;
	    case 3 :  /* division  */
		 cl1[l] = c1 * cl1[l] / c2 / cl2[l] + cs;
		 break;
	    case 'r' :  /* relative spectra,  */
		 cl1[l] = (c1 * cl1[l] - c2 * cl2[l]) / c2 / cl2[l] + cs;
		 break;
	    case 'a' :  /* relative spectra,  */
		 cl1[l] = fabs(c1 * cl1[l] - c2 * cl2[l]) / c2 / cl2[l] + cs;
		 break;
	    case 0 : /* difference */
	    default:
		 cl1[l] = c1 * cl1[l] - c2 * cl2[l] + cs;
		 break;
	    }
	    if (noneg && cl1[l] < 0.)  cl1[l] = 0.;
	}
    } else if ( *cl_file2 == '\0' && fabs(cs)>1.e-20) {
	lmin = lmin1;
	lmax = lmax1;
	if (lmin_print > 0)
	    if (lmin < lmin_print)  lmin = lmin_print;
	if (lmax_print > 0)
	    if (lmax > lmax_print)  lmax = lmax_print;

	for(l=lmin; l<=lmax; l++) cl1[l] += cs;
    } else if (stat_flag) {
	double aver=0.0,  sigma=0.0;
	int n;
	if ( *cl_file2 != 0) {
	    read_cl_dat(&cl2, &lmin2, &lmax2, cl_file2, one_column_input);
	    if (lmin < 0) {
		lmin = lmin1 > lmin2 ? lmin1 : lmin2;
	    }
	    if (lmax < 0) {
		lmax = lmax1 < lmax2 ? lmax1 : lmax2;
	    }

	    if (lmin_print > 0)
		if (lmin < lmin_print)  lmin = lmin_print;
	    if (lmax_print > 0)
		if (lmax > lmax_print)  lmax = lmax_print;
	}

	if (lmin < 0) lmin = lmin1;
	if (lmax < 0) lmax = lmax1;

	if (lmin_print > 0)
	    if (lmin < lmin_print)  lmin = lmin_print;
	if (lmax_print > 0)
	    if (lmax > lmax_print)  lmax = lmax_print;

	n = lmax - lmin + 1;
	for(l=lmin; l<=lmax; l++) aver += cl1[l];
	aver /= n;
	for(l=lmin; l<=lmax; l++) sigma += (cl1[l] - aver)*(cl1[l] - aver);
	sigma /= n;
	sigma = sqrt(sigma);
	if ( *cl_file2 != 0) {
	    double aver2=0.0, sigma2=0.0;
	    double corr_cf = 0.0;
	    for(l=lmin; l<=lmax; l++) aver2 += cl2[l];
	    aver2 /= n;
	    for(l=lmin; l<=lmax; l++) sigma2 +=(cl2[l]-aver2)*(cl2[l]-aver2);
	    sigma2 /= n;
	    sigma2 = sqrt(sigma2);
	    for(l=lmin; l<=lmax; l++) {
		double v1 = cl1[l] - aver;
		double v2 = cl2[l] - aver2;
		corr_cf += v1 * v2;
	    }
	    corr_cf /= (double)n;
	    corr_cf /= sigma * sigma2;
	    PR(stdout, "N_points = %d, aver1 = %g, sig1 = %g, ",n,aver,sigma);
	    PR(stdout, "aver2 = %g, sig2 = %g\n", aver2, sigma2);
	    PR(stdout, "corr.coefficient = %g\n", corr_cf);
	} else
	    PR(stdout, "N_points = %d, aver = %g, sigma = %g\n",n,aver,sigma);

	exit(0);
    } else if (hist_flag) {
	double min = 1e+30, max = -1e+30;
	int nn, i;
	FILE *fp_hg;
	int *hg = (int *)0;
	double step_hg;

	lmin = lmin1;
	lmax = lmax1;
	nn = lmax-lmin+1, i;
	for(i=lmin; i<= lmax; i++) {
	    if (cl1[i] < min) min = cl1[i];
	    if (cl1[i] > max) max = cl1[i];
	}

	/* histogram intervals */
	if (n_hg <= 0)  {
	    n_hg = 1. + log ((double)nn) / log(2.);
	}

	/* set limit for histogram */
	if (s_hg0 <  1.e+30)
	    min = s_hg0;
	if (s_hgF > -1.e+30)
	    max = s_hgF;

	/* open output file */
	if (*hg_file != '\0') {
	    if((fp_hg = fopen(hg_file, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n", myname, hg_file);
		fp_hg = (FILE *)0;
	    }
	} else if (*outfile != '\0') {
	    if((fp_hg = fopen(outfile, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n", myname, outfile);
		fp_hg = (FILE *)0;
	    }
	} else {
	    fp_hg = stdout;
	}

	if((hg = (int *)calloc(n_hg, sizeof *hg)) == (int *)0) {
	    PR(ER, "%s: no memory for histogram array of %d elements\n",
		  myname, n_hg);
	    fclose(fp_hg);
	    unlink(hg_file);
	    fp_hg = (FILE *)0;
	    exit(1);
	}
	step_hg = (max - min) / n_hg;

	for(i=0; i<nn; i++) {
	    int bin = (cl1[i] - min) / step_hg + 0.5;
	    if (bin > n_hg)    continue;
	    else if (bin > n_hg-1) continue;
	    else if (bin < 0)    continue;
	    hg[bin]++;
	}

	for(i=0; i<n_hg; i++) {
	   PR(fp_hg, "%12.5e  %12.5e\n",
		min+(i+0.5)*step_hg, hg[i]/(double)nn);
	}

	if(fp_hg != stdout) fclose(fp_hg);
	free((void *)hg);
	exit(0);

    } else if (sum_odd || sum_eve) {
	double *sc;
	int l0 = lmin1;
	if (lmin < 0) lmin = lmin1;
	if (lmax < 0) lmax = lmax1;
	if((sc = (double *)calloc(lmax+1, sizeof *sc)) == (double *)0) {
	    PR(ER, "%s: no memory for %d elements\n",   myname, lmax+1);
	    exit(1);
	}
	if ((sum_odd && l0%2==0) || (sum_eve && l0%2==1)) l0++;
	sc[l0]   = cl1[l0];
	sc[l0+1] = sc[l0];
	for(l=l0+2; l<=lmax; l+=2)  {sc[l] = sc[l-2]+cl1[l]; sc[l+1]=sc[l];}
	if (sum_interp) {
	    for(l=l0+1; l<=lmax-1; l+=2)   sc[l] = (sc[l-1]+sc[l+1])/2.;
	}
	for(l=lmin; l<=lmax; l++)  cl1[l] = sc[l];
    } else if (bin_flag) {  /* binning */
	if (bin_window < 1) bin_window=1;
	for(l=lmin1; l<=lmax1; l+=bin_window)  {
	    int k, k1=l, k2=k1+bin_window;
	    int la=0;
	    double cla = 0.0;
	    if (k2 <= lmax1) {
		for (k=k1; k<k2; k++) { la += k; cla += cl1[k]; }
		printf("%d %12.5e\n",
		      (int)(la/(double)bin_window), cla/bin_window);
	    }
	}
	exit(0);
    } else {
	if (lmin < 0) lmin = lmin1;
	if (lmax < 0) lmax = lmax1;

	if (lmin_print > 0)
	    if (lmin < lmin_print)  lmin = lmin_print;
	if (lmax_print > 0)
	    if (lmax > lmax_print)  lmax = lmax_print;

	for(l=lmin; l<=lmax; l++) cl1[l] *= cf;
    }

    if (cldl_flag) {
	/* convert  cl<-->dl */
	if (cldl_flag == 1) {
		cl2dl(cl1, lmax);
	} else if (cldl_flag == 2) {
		double T2 = 2.726*2.726;
		cl2dl(cl1, lmax);
		for (l=lmin; l<=lmax; l++) cl1[l] *= T2;
	} else if (cldl_flag == 3)  {
	       /* convert  dl --> cl */
		dl2cl(cl1, lmax);
	} else if (cldl_flag == 4) {
		double T2 = 2.726*2.726;
	       /* convert  dl*T^2 --> cl */
		for (l=lmin; l<=lmax; l++) cl1[l] /= T2;
		dl2cl(cl1, lmax);
	}
    }

    if (*outfile == '\0' || strcmp(outfile,"-") == 0
       || strcmp(outfile,"stdout") == 0) {
	   fp_out = stdout;
    } else {
	    /* open output file */
	    if((fp_out = fopen(outfile, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, outfile);
		exit(1);
	    }
    }
    if (one_column_output)
	print_x_dat(cl1, x, lmax+1, fp_out);
    else
	print_cl_dat(cl1, lmin, lmax, fp_out, one_column_input);
    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
	    switch(argv[i][1]) {
	    case '1':
		one_column_input = 1;
		break;
	    case '2':
		one_column_input = 2;
		break;
	    case 'b':
		bin_window = atoi(argv[++i]);
		bin_flag = 1;
		break;
	    case 'c':
		switch(argv[i][2]) {
		case '1':
			c1 = atof(argv[++i]);
			break;
		case '2':
			if (argv[i][3] == 'd') {
			    if (argv[i][4] == 't')
				cldl_flag = 2; /* cl -> dl*T^2 */
			    else
				cldl_flag = 1; /* cl -> dl */
			} else
			    c2 = atof(argv[++i]);
			break;
		case 'd':
			cf = 1./atof(argv[++i]);
			break;
		case 'n':
			event_count=1;
			break;
		default :
			cf = atof(argv[++i]);
			break;
		}
		break;
	    case 'd':
		if (strncmp(argv[i]+1, "dif", 3) == 0) {
		   action = 0; /* difference */
		} else if (strncmp(argv[i]+1, "div", 3) == 0) {
		   action = 3; /* division */
		} else if (strncmp(argv[i]+1, "de", 2) == 0) {
		   action = 3; /* division */
		} else if (strncmp(argv[i]+1, "d2c", 3) == 0) {
		   cldl_flag = 3; /* dl -> cl */
		} else if (strncmp(argv[i]+1, "dt2c", 4) == 0) {
		   cldl_flag = 4; /* dl*T^2 -> cl */
		} else if (strncmp(argv[i]+1, "dlt2c", 5) == 0) {
		   cldl_flag = 4; /* dl*T^2 -> cl */
		} else {
		   PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
		   usage();
		}
		break;
	    case 'e':
		ef1_flag = 1;
		break;
	    case 'f':
		strcpy(file_list, argv[++i]);
		break;
	    case 'h':
		if (strncmp(argv[i]+1, "hf", 2) == 0) {
			/* output file for ASCII histogram data */
			strcpy(hg_file, argv[++i]);
			hist_flag = 1;
		} else if (strncmp(argv[i]+1, "hn", 2) == 0) {
			/* number of bins for histogram */
			n_hg = atoi(argv[++i]);
			hist_flag = 1;
		} else if (strncmp(argv[i]+1, "h0", 2) == 0) {
			/* start value for histogram */
			s_hg0 = atof(argv[++i]);
			hist_flag = 1;
		} else if (strncmp(argv[i]+1, "he", 2) == 0) {
			/* start value for histogram */
			s_hgF = atof(argv[++i]);
			hist_flag = 1;
		} else if (strncmp(argv[i]+1, "h2", 2) == 0) {
			usage2();
		} else
			usage();
		break;
	    case 'l':
		/* determine limits of variations with sets of Cl */
		if (strncmp(argv[i]+1,"lma",3) == 0) {
		    lmax_print = atoi(argv[++i]);
		} else if (strncmp(argv[i]+1,"lmi",3) == 0) {
		    lmin_print = atoi(argv[++i]);
		} else {
		    /* set flag for printed columns */
		    if (argv[i][strlen(argv[i])-1] == '3')
			/* use 3sigma level too */
			limits = 3;
		    else
			limits = 1;
		}
		break;
	    case 'm':
		if (strncmp(argv[i]+1, "mu", 2) == 0) {
		   action = 2; /* multiplication */
		} else if (strncmp(argv[i]+1, "mi", 2) == 0) {
		   /* take 1st column from 1st file, 2nd from 2nd */
		   mix_flag = 1;
		   i++;
		   if (argv[i][0] != ',') {
		       if (strncmp(argv[i], "none", 4) != 0) {
			    n1_column = atoi(argv[i]);
		       }
		   }
		   {
		       char *ps = strchr(argv[i], ',');
		       if (ps != NULL && *ps && *(ps+1)) {
			    if (strncmp(ps+1, "none", 4) != 0) {
				 n2_column = atof(ps+1);
			    }
		       }
		   }
		} else {
		   PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
		   usage();
		}
		break;
	    case 'n':
		noneg = 1;
		break;
	    case 'o':
		strcpy(outfile, argv[++i]);
		break;
	    case 'p':
		if (strncmp(argv[i]+1, "pf", 2) == 0) {
		    strcpy(pfile, argv[++i]);
		} else {
		    v_prob = atof(argv[++i]);
		}
		confid = 1;
		break;
	    case 'r':
		if (strncmp(argv[i]+1, "r2", 2) == 0)
		    action = 'a';
		else
		    action = 'r';
		break;
	    case 's':
		if (strncmp(argv[i]+1, "st", 2) == 0) {
		       stat_flag = 1;
		} else if (strncmp(argv[i]+1, "sumo", 4) == 0) {
		       sum_odd = 1; /* sum of odd C_l */
		} else if (strncmp(argv[i]+1, "sume", 4) == 0) {
		       sum_eve = 1; /* sum of even C_l */
		} else if (strncmp(argv[i]+1, "sumi", 4) == 0) {
		       sum_interp = 1; /* interpolation of sum  */
		} else if (strncmp(argv[i]+1, "su", 2) == 0) {
		       action = 1; /* sum */
		} else {
		   PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
		   usage();
		}
		break;
	   case 'v':
		printf("%s: version 0.8 (02/01/2016), GLESP 2.0\n",
		    myname);
		exit(0);
	    default:
		break;
	    }
	    break;
	case '+':
	    switch(argv[i][1]) {
	    case '1':
		one_column_output = 1;
		break;
	    case '2':
		one_column_output = 2;
		break;
	    default:
		PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
		usage();
		break;
	    }
	    break;
	default :
		if (*cl_file1 == '\0')      strcpy(cl_file1, argv[i]);
		else if (*cl_file2 == '\0') strcpy(cl_file2, argv[i]);
		if(nf_m == NFF)  break;
		ind_nf[nf_m] = i;  nf_m++;
		nf++;
		break;
	}
    }

    if (nf && (limits || confid) ) {
	/* take memory for filenames */
	if ((filename = (char**)calloc(nf, sizeof *filename)) == (char**)0) {
no_f:
	    PR(ER, "%s: no memory for file names.\n", myname);
	    exit(1);
	}
	for (i=0; i<nf; i++) {
	    if ((filename[i] = (char *)calloc(81, sizeof (char))) == NULL)
		goto no_f;
	}

	/* read file names */
	nf = 0;
	for(i=1; i<argc; i++) {
	    switch(argv[i][0]) {
	    case '-':
		    switch(argv[i][1]) {
		    case '1':
		    case '2':
		    case 'c':
		    case 'd':
		    case 'h':
		    case 'l':
		    case 'm':
		    case 'n':
			break;
		    case 'o':
		    case 'p':
			i++;
			break;
		    case 'r':
		    case 's':
			break;
		    }
		    break;
	    case '+':
		    break;
	    default:
		    strcpy(filename[nf], argv[i]);
		    nf++;
		    break;
	    }
	}
    }

    return;
}

void usage()
{
PR(ER, "%s [-c1 c1] [-c2 c2] [-cf cf] [-cs cs] [-{sum,mul,div}] [-{c2d[t],d[t]2c}]\n",
    myname);
PR(ER, "[-o output_file] [-r((d1-d2)/d2)] [-r2(|d1-d2|/d2)] [-1(1col_input)]\n");
PR(ER, "[-lim[3] f1 f2 ... fn [-ef]] [-stat] [-2 (take 2nd column)] [+1] [-ne(noneg)]\n");
PR(ER, "[-lmin lmin] [-lmax lmax] [-p lev(calc.probability)] [-cn(counts)]\n");
PR(ER, "[-pf file_dat(calc.probability for values)] [-mix n,m] [-fl filelist]\n");
PR(ER, "[-hn n_hg] [-hf hg_file] [-h0 start_his.value] [-he stop_his.val.]\n");
PR(ER, "[-bw l_bin_window]\n");
exit(2);
}

void usage2()
{
PR(ST, "usage: %s file1 [file2] -o output_file [FLAGS]\n", myname);
PR(ST, "-----------------------------------------------------------------\n");
PR(ST, "-1 - set one column file input\n");
PR(ST, "-2 - take 2nd column from files\n");
PR(ST, "+1 - set one column output\n");
PR(ST, "-bw l_win - calculates binning of C(l) with l-window\n");
PR(ST, "-c1 k1  -c2 k2 - calc expression for files: k1*cl1.dat -/+ k2*cl2.fts\n");
PR(ST, "-c2d  - convert C(l) to D(l)\n");
PR(ST, "-c2dt - convert C(l) to D(l)*T^2\n");
PR(ST, "-cf value - calc expression: value * cl.dat (see '-c1', '-cs')\n");
PR(ST, "-cn - calculate number of events (see 'p','-pf')\n");
PR(ST, "-cs value - calc expression: value + cl.dat (see '-c1', '-cf')\n");
PR(ST, "-d2c  - converts D(l) to C(l)\n");
PR(ST, "-div - calculates divison: cl1.dat / cl2.dat\n");
PR(ST, "-dt2c - converts D(l)*T^2 to C(l)\n");
PR(ST, "-ef   - sets 1st column in float format for '-lim'\n");
PR(ST, "-fl filelist - reads file names from file 'filelist', not from com.string\n");
PR(ST, "-he v2 - calc histogram upto value v2 (see:'-h0','-hf','-hn')\n");
PR(ST, "-hf hist.dat - output filename for histogram data (see:'-h0','-he','-hn')\n");
PR(ST, "-hn num - set number of histogram bins 'num' (see:'-h0','-he','-hf')\n");
PR(ST, "-lim  f1 f2 ... fn - calculate +/-1sigma limits from set of files (see '-ef')\n");
PR(ST, "-lim3 f1 f2 ... fn - calculate +/-3sigma limits from set of files (see '-ef')\n");
PR(ST, "-lmin lmin - set lmin\n");
PR(ST, "-lmax lmax - set lmax\n");
PR(ST, "-mix n,m - take n-th column from 1st file, m-th column from 2nd file\n");
PR(ST, "-mul - calculate multiplication: cl1.dat * cl2.dat\n");
PR(ST, "-ne - change negative value to zero\n");
PR(ST, "-o output_file - set name of output file\n");
PR(ST, "-p lev - set level of probabilty to be calculated (see '-pf','-cn')\n");
PR(ST, "-pf file - calculate probability for values from 'file' (see '-p','-cn')\n");
PR(ST, "-r  - calculate value: (cl1.dat-cl2.dat) / cl2.dat\n");
PR(ST, "-r2 - calculate value: |cl1.dat-cl2.dat| / cl2.dat\n");
PR(ST, "-sum - calculate sum: cl1.dat + cl2.dat\n");
PR(ST, "-stat file1 - calculate average and sigma of a sample of file1 data\n");
PR(ST, "-stat f1 f2 - calculate average and sigma of a sample of f1 and f2 data\n");
exit(0);
}

void read_cl_dat(cl, lmin, lmax, file, one_column_input)
double **cl;
int *lmin, *lmax;
char *file;
int one_column_input;
{
    register int n=0;
    int lmi = 1000000, lma = -1;
    int ll=0; double v=0.0;
    FILE *fp;
    char str[161];

    if((fp = fopen(file, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'\n", myname, file);
	exit(1);
    }
    if (fgets(str, 160, fp) != NULL) {
	if (strncmp (str, "SIMPLE  =", 9) == 0) {
	    extern int read_cl_fits (char *, double **, int *);
	    fclose(fp);
	    read_cl_fits (file, cl, &lma);
	    *lmin = 0;
	    *lmax = lma;
	    return;
	} else {
	    fseek(fp, 0l, 0);
	}
    } else {
	    PR(ER, "%s: can not read file '%s'\n", myname, file);
	    exit(1);
    }

    while(fgets(str, 160, fp) != NULL) {
	if (*str == '#' || *str == '\n' || *str == '\r')
	    continue;
	if (! one_column_input) {
	    sscanf(str, "%d%lf", &ll, &v);
	    if (ll < 0) {
		PR(ER, "%s: wrong format for the 1st column in file '%s':\n",
			myname, file);
		PR(ER, "'l' should be positive.\n");
		exit (1);
	    }
	    if (ll > lma) lma = ll;
	    if (ll < lmi) lmi = ll;
	}
	n++;
    }
    if (one_column_input) { lmi=0; lma=n-1; }

    /* create data arrays */
    if ((*cl =(double*)calloc((unsigned)(lma+1),sizeof **cl))==(double*)0) {
	PR(ER, "%s: problems to find memory for array of %d elements.\n",
	    myname, lma);
	exit(1);
    }
    fseek(fp, 0l, 0);
    ll = -1;
    while(fgets(str, 160, fp) != NULL) {
	if (*str == '#' || *str == '\n' || *str == '\r')
	    continue;
	if (one_column_input) {
	    sscanf(str, "%lf", &v);
	    ll++;
	} else
	    sscanf(str, "%d%lf", &ll, &v);
	(*cl)[ll] = v;

    }
    fclose (fp);

    *lmin = lmi;
    *lmax = lma;

    return;
}

void print_cl_dat(cl, lmin, lmax, fp_out, one_column_flag)
double *cl;
int lmin, lmax;
FILE *fp_out;
int one_column_flag;
{
    register int l;

    for(l=lmin; l<=lmax; l++)
	if (one_column_flag)
	    PR(fp_out, "%14.7e\n", cl[l]);
	else
	    PR(fp_out, "%4d %14.7e\n", l, cl[l]);

    return;
}


void print_x_dat(y, x, n, fp_out)
double *y, *x;
int n;
FILE *fp_out;
{
    register int i;

    for(i=0; i<n; i++)
	    PR(fp_out, "%14.7e %14.7e\n", x[i], y[i]);

    return;
}


void cl2dl(cl, lmax)
double *cl;
int lmax;
{
    register int l;
    double pi = M_PI, T=2.726, T2=T*T;
    for(l=0; l<=lmax; l++)  cl[l] *= l*(l+1) / 2. / pi / T2;
    return;
}
void dl2cl(dl, lmax)
double *dl;
int lmax;
{
    register int l;
    double pi = M_PI, T=2.726, T2=T*T;
    for(l=1; l<=lmax; l++)  dl[l] /= l*(l+1) / 2. / pi / T2;
    return;
}


void read_col_dat(d, nn, file, n_column)
double **d;
int *nn;
char *file;
int n_column;
{
    register int n=0;
    FILE *fp;
    char str[161];
    char s1[40], s2[40], s3[40], s4[40], s5[40], s6[40], s7[40], s8[40];

    if((fp = fopen(file, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'\n", myname, file);
	exit(1);
    }

    if (fgets(str, 160, fp) != NULL) {
	if (strncmp (str, "SIMPLE  =", 9) == 0) {
	    extern int read_cl_fits (char *, double **, int *);
	    fclose(fp);
	    read_cl_fits (file, d, nn);
	    return;
	} else {
	    fseek(fp, 0l, 0);
	}
    } else {
	    PR(ER, "%s: can not read file '%s'\n", myname, file);
	    exit(1);
    }


    while(fgets(str, 160, fp) != NULL) {
	if (*str == '#' || *str == '\n' || *str == '\r')
	    continue;
	n++;
    }
    *nn = n;

    /* create data arrays */
    if ((*d =(double*)calloc((unsigned)n,sizeof **d))==(double*)0) {
	PR(ER, "%s: problems to find memory for array of %d elements.\n",
	    myname, *nn);
	exit(1);
    }
    fseek(fp, 0l, 0);
    n=0;
    while(fgets(str, 160, fp) != NULL) {
	if (*str == '#' || *str == '\n' || *str == '\r')
	    continue;

	s1[0]=s2[0]=s3[0]=s4[0]=s5[0]=s6[0]=s7[0]=s8[0]='\0';
	sscanf(str, "%s%s%s%s%s%s%s%s", s1,s2,s3,s4,s5,s6,s7,s8);

	switch (n_column) {
	default:
	case 1: (*d)[n] = atof(s1); break;
	case 2: (*d)[n] = atof(s2); break;
	case 3: (*d)[n] = atof(s3); break;
	case 4: (*d)[n] = atof(s4); break;
	case 5: (*d)[n] = atof(s5); break;
	case 6: (*d)[n] = atof(s6); break;
	case 7: (*d)[n] = atof(s7); break;
	case 8: (*d)[n] = atof(s8); break;
	}
	n++;
    }
    fclose (fp);

    return;
}
