#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

/*
 *  program of arithmetic (and not only) operations over maps
 *
 *  created by Oleg Verkhodanov, 2002-2004,
 *  2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016
 *
 */
#include "glesp.h"

#define PR      fprintf
#define ER      stderr
#define ST      stdout
#define NFF     50

extern void setopt(int, char **), usage(void), usage2(void);

double c1=1.0, c2=1.0, cf = 1.0, sc = 0.0;
int action = 0;        /* action: 0 - difference, 1 - sum     */
char map1_file[161], map2_file[161], outfile[161];
char q_out_file[161], u_out_file[161];
char outp_angle_file[161];
char hg_file[161];     /* output file with  histogram         */
int robust_average = 0;
int max_flag = 0;
int min_flag = 0;
int mask_flag = 0;
int mask_file_flag = 0;
int dif_mask = 1;
int level_flag = 0;
double eps = 0.1;
double level = 0.0;   /* level to calc number of pixels */
int ffile_flag = 0;   /* output to FITS file            */
double dummy = 0.0;
int abs_flag = 0;
int do_stacking = 0;  /* stacking procedure for zones    */
char file_stack[81];  /* list of stacking zones          */

int stat_flag = 0;  /* statistics estimation */
int corr_flag = 0;  /* calculation of correlation coefficient of 2 maps */
int convert_coors=0;/* transform coordinates                            */
int polariz_flag=0; /* calc I=(q^2+u^2)^.5; a=.5atan(q/u)               */
double cut_level = -1.e+30;
int map_type = 't'; /* type of map: T, Q or U-map, P-map, or 3-map */
int mirror_flag=0;  /* mirror reflection          */
char out_extr_file[161];
double dp_ex = -1.; /* radius to search for extrema, rad */
int ex_stat = 0;    /* flag to calc extrema              */
int dipole_fitting = 0;
int apply_mask = 0; /* apply mask                        */
int estimator_type = 0; /* mapping estimator type to calc.rotat.anisotropy */
int esl_lmax_stat = -1; /* lmax for temporal maps in calculayed rotation   */

int n_hg = 0;       /* number of bins in histogram */
double s_hg0 = 1.e+30, s_hgF= -1.e+30;
int norm_flag=0;
int nf_m=0;
int ind_nf[NFF];
int simple_average = 0;
int rotate_flag = 0;
double dphi = 0.0, dtheta = 0.0;
int grad_flag = 0;  /* flag to calculate gradients */
double corr_win = -1.;
double corr_rwin = -1.;
double level_low = 0; /* low level of pixels value to get pixels number */
double level_up  = 0; /* upper level of pixels value to get pixels number */
char file_stat_interv[161];
char fit_map_patt_file[161]; /* fitting map pattern file */

int zmask_flag = 0;        /* set mask                                 */
double zmask_value = M_PI * 1.e+30;
int zone_circle = 0;
char zone_file[81];       /* file with list of zones to be processed   */
double zone_radius = -1.; /* radius of removed circle in degs          */
double zone_latR0 = -1000., zone_lonR0 = -1000.,
       zone_latR1 = -1000., zone_lonR1 = -1000.;
int    z_remove_flag = 1,
       z_remain_flag = 0;
char hp_mask_file[161];     /* HPix mask file to overlay mask   */
char gl_mask_file[161];     /* GLESP mask file to overlay mask  */
int    f_mask_flag = 1;     /* remove hp_mask_zone:1, remain: 0 */
int  t_cos_th_flag = 0;     /* flag to calc T(th) vs cos(th)    */
int  inverse_flag = 0;      /* inversion of a map               */
int  mink_func = 0;         /* flag for Minkowski func. cal-tion*/

int print_flag=0;
char format_out[81];        /* ASCII output format              */
double repix_win = -1;
int repix_flag = 0;
int coor_type = 'g';        /* coordinate system type for circular zones */

int rot_q = 0;      /* number of angle in a queue to be rotated */
int rot_max = 10.;  /* maximum number of angles */
struct _rot_   {
    int cr;        /* coordinate type */
    double ang;    /* angle */
} rotate_queue [10];

extern void dhlbeg_(int *), dhlend_(void);
extern double drhodl_(double  *);

double lat_c,  /* galactic coordinate of ring center or rectang. corner */
       lon_c,  /* galactic coordinate of ring center or rectang. corner*/
       lat1_c, /* galactic coordinate of another rectang. corner */
       lon1_c, /* galactic coordinate of another rectang. corner */
       r_c,    /* radius of inner circle */
       w_c;    /* radius of ring width   */
int flag_conc_fig = 0;  /* 1-circle, 2 - rectangular */
int  flag_kZS = 0;
int sig_win_flag = 0;   /* calculate sigma in window: 1-new size, */
			/* 2-running  window                      */
double sig_win = 0.;

char *myname="difmap";


int main(argc, argv)
int argc;
char **argv;
{
    register int i=0, j;
    double *f1 = (double *)0, *f2 = (double *)0;
    double *x1 = (double *)0, *x2 = (double *)0;
    double *mf[NFF], *zmask = (double *)0;;
    int *nph1 = (int *)0, *nph2 = (int *)0;
    int nx1, nx2, np1, np2;
    FILE *fp_out = (FILE *)0;
    char x_type[40], grid_type[40];
    double *q1=(double *)0, *u1=(double *)0;
    double *q2=(double *)0, *u2=(double *)0;
    struct _zone_ *zz = (struct _zone_ *)0;
    int function_ztype;
    int zns = 1;
    extern void gal2ec_map(double *, double *, int, int, int *, double *);


    map1_file[0] = map2_file[0] = outfile[0] = hg_file[0] =
	 gl_mask_file[0] = hp_mask_file[0] = outp_angle_file[0] = '\0';
    out_extr_file[0] = file_stat_interv[0] = '\0';
    q_out_file[0] = u_out_file[0] = '\0';
    fit_map_patt_file[0] = '\0';

    strcpy(format_out, "%15.8e");

    /* select grid output keyword values */
    strcpy(x_type, "COS(THETA)");
    strcpy(grid_type, "PHI(X)+LEGENDRE");

    /* loop over external parameters */
    setopt(argc, argv);

    if (glesp_comments < 0) {
	if (getenv("GLESP_COMMENTS") != NULL) {
	    char str[21];
	    strcpy(str,getenv("GLESP_COMMENTS"));
	    glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
	    if (glesp_comments) {
		 if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
		     glesp_comments = 0;
	    }
	} else
	    glesp_comments = 1;
    }


    function_ztype = z_remove_flag ? 1 : 0; /* remove_flag */

    /* read regions from file for masking */
    /*-------------------------*/
    if ( zmask_flag ) {
	   register int is=0;
	   extern void map_zone(struct _zone_ *, int, int,
		   double *, double *, int, int, int *, double);
	   int zone_type = zone_circle ? 0 : 1;

	   /* if to read zone-file */
	   if (zone_file[0] != '\0') {
		FILE *fz;
		char str[81];
		/* read data from file with zones */
		if((fz = fopen(zone_file, "r")) == (FILE *)0) {
		    PR(ER, "%s: I can not open file '%s'...\n",
			myname, zone_file);
		    exit(1);
		}
		/* get number of zones */
		is =0;
		while(fgets(str, 80, fz) != NULL) {
		    if (*str == '#' || *str == '\n')
			continue;
		    is++;
		}
		/* allocate memory */
		i = zns = is;
		zz = (struct _zone_ *)calloc(i, sizeof *zz);
		if (zz == (struct _zone_ *)0) goto nomem;

		/* read zones  */
		is=0;
		fseek(fz, 0l, 0);
		while(fgets(str, 80, fz) != NULL) {
		  char s1[30], s2[30], s3[30], s4[30];
		  char *cs = str;
		  int n_par;  /* number of parameters per string */
		  if (*str == '#' || *str == '\n')
			  continue;
		  while(*cs) if (*cs == '#') {*cs = '\0'; break;} else cs++;

		  n_par = n_skip_wd(str);
		  switch (n_par) {
		  case 1:
		      PR(ER,"%s: there is only 1 parameter in the string:\n",
			      myname);
		      PR(ER,"%s", str);
		      PR(ER, "\n%s: I've skipped it...\n", myname);
		      break;
		  case 2:
		      zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		      zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		      if ( coor_type == 'e') { /* equatorial */
			    /* convert (RA_hr,Dec_deg)-->(GLAT_deg,GLON_deg)*/
			    double alp, del, lat, lon;
			    /* radians -> degrees and hours */
			    alp = zz[is].latR0 *  24/M_PI;  /* hours */
			    del = zz[is].lonR0 * 180/M_PI;  /* degrees */
			    equat2gal(alp, del, &lat, &lon);
			    zz[is].latR0 = lat * M_PI/180.; /* radians */
			    zz[is].lonR0 = lon * M_PI/180.;
		      }
		      if (zone_radius > 0.)
			  zz[is].radius = zone_radius;
		      else {
			  zz[is].radius = deg2rad((skip_wd(2,str,s3),s3));
		      }
		      zz[is].type = 0;
		      break;
		  case 3:
		      zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		      zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		      if ( coor_type == 'e') { /* equatorial */
			    /* convert (RA_hr,Dec_deg)-->(GLAT_deg,GLON_deg)*/
			    double alp, del, lat, lon;
			    /* radians -> degrees and hours */
			    alp = zz[is].latR0 *  24/M_PI;  /* hours */
			    del = zz[is].lonR0 * 180/M_PI;  /* degrees */
			    equat2gal(alp, del, &lat, &lon);
			    zz[is].latR0 = lat * M_PI/180.; /* radians */
			    zz[is].lonR0 = lon * M_PI/180.;
		      }
		      if (zone_radius > 0.)
			  zz[is].radius = zone_radius;
		      else {
			  zz[is].radius = deg2rad((skip_wd(2,str,s3),s3));
		      }
		      zz[is].type = 0;
		      break;
		  case 4:
		      zz[is].latR0 = deg2rad((skip_wd(0,str,s1),s1));
		      zz[is].lonR0 = deg2rad((skip_wd(1,str,s2),s2));
		      zz[is].latR1 = deg2rad((skip_wd(2,str,s3),s3));
		      zz[is].lonR1 = deg2rad((skip_wd(3,str,s4),s4));
		      if (zone_type == 0) {
			      zz[is].radius = zz[is].latR1;
			      zz[is].type = 0;
		      } else  zz[is].type = 1;
		      break;
		  }
		  is++;
		}

		fclose(fz);
	   } else {
		/* allocate memory for 1 zone */
		/* from command string */
		i = zns = 1;
		zz = (struct _zone_ *)calloc(i, sizeof *zz);
		if (zz == (struct _zone_ *)0) goto nomem;
		zz[0].latR0 = zone_latR0;  zz[0].lonR0 = zone_lonR0;
		zz[0].latR1 = zone_latR1;  zz[0].lonR1 = zone_lonR1;
		zz[0].radius = zone_radius;
		if (zone_latR1 < -900.) {
			zz[0].type = 0;
		} if (zone_lonR1 < -900.) {
			if (zone_radius > 0.)
			    zz[0].radius = zone_radius;
			else
			    zz[0].radius = zone_latR1;
			zz[0].type = 0;  /* circle */
		} else {
			zz[0].type = 1;  /* rectangular */
		}
	   }
    }

    if (robust_average) {
	double arr[NFF], sum=0.0;
	int ntot=0;
	if(nf_m > NFF) {
	    PR(ER,
	      "%s: too big number of averaged files: %d (upper limit: %d)\n",
		    myname, nf_m, NFF);
	    exit(2);
	}
	mf[0] = (double *)0;

	if (map_type == 't' || map_type == 'T')
	   read_glesp_tmap(&mf[0], &x1, &nx1, &np1, &nph1, argv[ind_nf[0]]);
	else if (map_type == 'u')
	   read_glesp_umap(&mf[0], &x1, &nx1, &np1, &nph1, argv[ind_nf[0]]);
	else
	   read_glesp_qmap(&mf[0], &x1, &nx1, &np1, &nph1, argv[ind_nf[0]]);

	for(i=1; i<nf_m; i++) {
	    mf[i] = (double *)0;

	    if (map_type == 't' || map_type == 'T')
	     read_glesp_tmap(&mf[i], &x2, &nx2, &np2, &nph2, argv[ind_nf[i]]);
	    else if (map_type == 'u')
	     read_glesp_umap(&mf[i], &x2, &nx2, &np2, &nph2, argv[ind_nf[i]]);
	    else
	     read_glesp_qmap(&mf[i], &x2, &nx2, &np2, &nph2, argv[ind_nf[i]]);

	    if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
		PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		       myname, argv[ind_nf[0]], argv[ind_nf[i]]);
		PR(ER, "Sorry...\n");
		exit(1);
	    }
	}
	dhlbeg_(&nf_m);  /* start Hodges-Lehmann estimator */

	f1 = mf[0];
	for(i=0; i<nx1; i++) {
	    register int k;
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		for(k=0; k<nf_m; k++) arr[k] = mf[k][ii+j];
		f1[ii+j] = drhodl_(arr);   /* Hodges-Lehmann estimate */
		sum += f1[ii+j];
		ntot++;
	    }
	}
	dhlend_();
	sum /= ntot;
    } else if (simple_average) {
	double a;
	if(nf_m > NFF) {
	    PR(ER,
	      "%s: too big number of averaged files: %d (upper limit: %d)\n",
		    myname, nf_m, NFF);
	    exit(2);
	}
	mf[0] = (double *)0;
	if (map_type == 't' || map_type == 'T')
	    read_glesp_tmap(&mf[0], &x1, &nx1, &np1, &nph1, argv[ind_nf[0]]);
	else if (map_type == 'u')
	    read_glesp_umap(&mf[0], &x1, &nx1, &np1, &nph1, argv[ind_nf[0]]);
	else
	    read_glesp_qmap(&mf[0], &x1, &nx1, &np1, &nph1, argv[ind_nf[0]]);

	for(i=1; i<nf_m; i++) {
	    mf[i] = (double *)0;

	    if (map_type == 't' || map_type == 'T')
	     read_glesp_tmap(&mf[i], &x2, &nx2, &np2, &nph2, argv[ind_nf[i]]);
	    else if (map_type == 'u')
	     read_glesp_umap(&mf[i], &x2, &nx2, &np2, &nph2, argv[ind_nf[i]]);
	    else
	     read_glesp_qmap(&mf[i], &x2, &nx2, &np2, &nph2, argv[ind_nf[i]]);

	    if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
		PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		       myname, argv[ind_nf[0]], argv[ind_nf[i]]);
		PR(ER, "Sorry...\n");
		exit(1);
	    }
	}
	f1 = mf[0];
	for(i=0; i<nx1; i++) {
	    register int k;
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		a = 0.;
		for(k=0; k<nf_m; k++) a += mf[k][ii+j];
		f1[ii+j] = a/nf_m;
	    }
	}

    } else if (dipole_fitting) {
	double tx,ty,tz, dtx,dty,dtz, t_ex,b_ex,l_ex, dt_ex,db_ex,dl_ex;
	extern void dipole_fit(double*,double*,int,int,int*,int*,
	    double*,double*,double*, double*,double*,double*,
	    double*,double*,double*, double*,double*,double* );
	int *mask = (int *)0;

	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	if((mask=(int*)calloc((unsigned)(nx1*np1),sizeof *mask))==(int *)0)
	    goto nomem;

	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
	      int n = ii+j;
	      if (ABS(f1[n] - GLESPMASK) < 500)  mask[n] = 1;
	    }
	}
	dipole_fit(f1,x1,nx1,np1,nph1,mask, &tx,&ty,&tz, &dtx,&dty,&dtz,
	    &t_ex,&b_ex,&l_ex, &dt_ex,&db_ex,&dl_ex);

	PR(ER, "t_ex = %12.5e, (b,l) = (%-7.3f, %7.3f)\n",
		t_ex, b_ex*180/M_PI, l_ex*180/M_PI);
	PR(ER, "      [%12.5e]        [(%-7.3f, %7.3f)]\n",
		ABS(dt_ex), ABS(db_ex)*180/M_PI, ABS(dl_ex)*180/M_PI);
	PR(ER, "(tx,ty,tz) = (%12.5e, %12.5e, %12.5e)\n",  tx,ty,tz);
	PR(ER, "             [%12.5e, %12.5e, %12.5e]\n",  dtx,dty,dtz);
    } else if (min_flag || max_flag) {
	if(nf_m > NFF) {
	    PR(ER,
	      "%s: too big number of averaged files: %d (upper limit: %d)\n",
		    myname, nf_m, NFF);
	    exit(2);
	}
	mf[0] = (double *)0;
	if (map_type == 't' || map_type == 'T')
	    read_glesp_tmap(&mf[0], &x1, &nx1, &np1, &nph1, argv[ind_nf[0]]);
	else if (map_type == 'u')
	    read_glesp_umap(&mf[0], &x1, &nx1, &np1, &nph1, argv[ind_nf[0]]);
	else
	    read_glesp_qmap(&mf[0], &x1, &nx1, &np1, &nph1, argv[ind_nf[0]]);

	for(i=1; i<nf_m; i++) {
	    mf[i] = (double *)0;
	    if (map_type == 't' || map_type == 'T')
	     read_glesp_tmap(&mf[i], &x2, &nx2, &np2, &nph2, argv[ind_nf[i]]);
	    else if (map_type == 'u')
	     read_glesp_umap(&mf[i], &x2, &nx2, &np2, &nph2, argv[ind_nf[i]]);
	    else
	     read_glesp_qmap(&mf[i], &x2, &nx2, &np2, &nph2, argv[ind_nf[i]]);

	    if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
		PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		       myname, argv[ind_nf[0]], argv[ind_nf[i]]);
		PR(ER, "Sorry...\n");
		exit(1);
	    }
	}
	f1 = mf[0];
	for(i=0; i<nx1; i++) {
	    register int k;
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		double a_min = 1.e+30, a_max = -1.e+30;
		int sg = 1;
		for(k=0; k<nf_m; k++) {
		     double v = mf[k][ii+j], fv=fabs(v);
		     if (min_flag) {
			   if (a_min > fv)  {
				 a_min = fv;
				 if (v < 0.) sg =  -1;
			   }
		     } else { /* max_flag */
			   if (a_max < fv)  {
				 a_max = fv;
				 if (v < 0.) sg =  -1;
			   }
		     }
		}
		f1[ii+j] = sg * (min_flag ? a_min : a_max) ;
	    }
	}

    } else if(mask_file_flag) {
	if (map_type == 't' || map_type == 'T') {
	    /* read data */
	    read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    /* read mask */
	    read_glesp_tmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	} else if (map_type == 'u') {
	    read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    read_glesp_umap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	} else {
	    read_glesp_qmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    read_glesp_qmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	}

//        if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
//            PR(ER,"%s: map '%s' and mask '%s' have different resolution. ",
//                   myname, map2_file, map1_file);
//            PR(ER, "Sorry...\n");
//            exit(1);
//        }
	for(i=0; i<nx1; i++) {
	    double theta, phi;
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		  int n;
		  pixel2angle(x1, nx1, nph1, i, j, &theta, &phi);
		  n = ang2n(theta, phi, x2, nx2, np2, nph2);
		  int m = (int)(f2[n]+0.5);
		  if (!m)  f1[ii+j] = dummy;
	    }
	}
    } else if(mask_flag && !mink_func) {
	if (map_type == 't' || map_type == 'T') {
	    read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    if (*map2_file != '\0')
		read_glesp_tmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	} else if (map_type == 'u') {
	    read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    read_glesp_umap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	} else {
	    read_glesp_qmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    read_glesp_qmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	}

	if (dif_mask != 5 && *map2_file != '\0' &&
	    map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
	    PR(ER,"%s: maps '%s' and '%s' have different resolution.\n",
		   myname, map2_file, map1_file);
	    PR(ER, "Sorry...\n");
	    exit(1);
	}

	if ( dif_mask != 5 && *map2_file != '\0') {
	    free((void *)x2);     x2  = (double *)0;
	    free((void *)nph2);   nph2 = (int *)0;
	}

	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		double  v=0.0, dd=0.0,
			v1 = f1[ii+j],
			v2 = 0.0;

		if (f2 != (double *)0) {
		    if (dif_mask != 5 && dif_mask != 6)
			/* same array dimensions */
			v2 = f2[ii+j];
		    else {  /* different dimensions of arrays */
			if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
			    double theta, phi;
			    int i2, j2;
			    pixel2angle(x1, nx1, nph1, i, j, &theta, &phi);
			    angle2pixel(x2, nx2, nph2, &i2, &j2, theta, phi);
			    v2 = f2[i2*np2+j2];
			} else
			    v2 = f2[ii+j];
		    }
		}
		switch(dif_mask) {
		case 0:
			if (v1*v2 > 0.0) v =  1.;
			else if (v1*v2 < 0.0) v = -1.;
			else v = 0.;
			break;
		case 1:
			dd = v1-v2;
			if (fabs(dd) > eps * fabs(v1)) {
				if (dd < 0.) v = -1.;
				else if (dd > 0.) v = 1.;
			}
			break;
		case 2:
			if (v1*v2 > 0.0 && v1 > 0.)  v =  1.;
			else if (v1*v2 > 0.0 && v1 < 0.)  v =  0.5;
			else if (v1*v2 < 0.0 )  v =  -1.;
			else v = 0.;
			break;
		case 3:
			dd = v1-v2;
			if (fabs(dd) > eps * fabs(v1))
			     v =  1.;
			else
			     v =  -1.;
			break;
		case 4:
			if (v1 > 0.0 && v2 > 0.0) v =  1.;
			else if (v1 < 0.0 && v2 < 0.0) v = -1.;
			else v = 0.;
			break;
		case 5: /* cut level by galaxy number per pixel */
			if (v2 < cut_level) v = 0.0;
			else v = v1;
			break;
		case 6: /* isolines */
			if (v2 > 1.e8)   v = -1.e30 * v2;
			else v = v1;
			break;
		case 7:
			if (ABS(v1 - GLESPMASK) < 500)   v = dummy;
			else
			    v = v1;
			break;
		}
//if (f1[ii+j] > 1e9) {
//PR(ER, "b,f1[n] = %15.8e, GM=%15.8e, d=%15.8e\n",
//f1[ii+j], GLESPMASK, f1[ii+j] - GLESPMASK);
//}
		f1[ii+j] = v;
	    }
	}
    } else if (corr_win > 0.) {
	/*-----------------------------------*/
	/* calculation of correlationmapping */
	/*-----------------------------------*/
	double *fcor = (double *)0;
	extern void  map_cor(double *, double *, double *, int, int, int *,
			 double, double **, double **, int *, int *, int **),
	   map_cor_mask(double*, double*, int*, double *, int, int, int *,
		  double, double **, double **, int *, int *, int **, double);
	int *imask = (int *)0;
	double *fmask = (double *)0, *gmask = (double *)0;
	double dummy = GLESPMASK;

	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	read_glesp_tmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);

	if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
	       PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		      myname, map2_file, map1_file);
	       PR(ER, "Sorry...\n");
	       exit(1);
	}
	free((void *)x2);
	free((void *)nph2);
	x2 = (double *)0;
	nph2 = (int *)0;

	/* preparation of memory for masks for correlation areas */
	/*-------------------------------------------------------*/
	if (zmask_flag) {
	    if ((zmask=(double*)calloc((unsigned)(nx1*np1), sizeof *zmask))==
	      (double *)0) {
no_mem_mask:
		PR(ER, "%s: no memory for mask array... No mask.\n", myname);
		zmask_flag = 0;
	    }
	    if ((imask=(int *)calloc((unsigned)(nx1*np1), sizeof *imask))==
	      (int *)0)
		goto no_mem_mask;
	}
	if (*hp_mask_file != '\0') {
	    if ((fmask=(double*)calloc((unsigned)(nx1*np1), sizeof *fmask))==
	      (double *)0) {
		PR(ER, "%s: no memory for mask array... No mask.\n", myname);
		zmask_flag = 0;
	    }
	}
	if (*gl_mask_file != '\0') {
	    if ((gmask=(double*)calloc((unsigned)(nx1*np1), sizeof *gmask))==
	      (double *)0) {
		PR(ER, "%s: no memory for mask array... No mask.\n", myname);
		zmask_flag = 0;
	    }
	}

	/* read region for masking */
	/*-------------------------*/
	if ( zmask_flag ) {
	   extern void map_zone(struct _zone_ *, int, int,
		   double *, double *, int, int, int *, double);

	   map_zone(zz,zns,function_ztype,zmask,x1,nx1,np1,nph1,zmask_value);

	   /* set mask from HPix-file */
	   if (*hp_mask_file != '\0') {
	      extern void mask_hp(double*,double*,int,int,int*,char*,double);
	      i = nx1;
	      mask_hp(fmask,x1,i,np1,nph1,hp_mask_file,zmask_value);
	   }
	   /* set mask from GLESP-file */
	   if (*gl_mask_file != '\0') {
	      extern void mask_gl(double*,double*,int,int,int*,char*,double);
	      i = nx1;
	      mask_gl(gmask,x1,i,np1,nph1,gl_mask_file,zmask_value);
	   }
	}

	/*-----------------------*/
	/* correlation procedure */
	/*-----------------------*/
	if (zmask_flag) {
	    /* correlation with mask */
	    /*-----------------------*/
	    for(i=0; i<nx1; i++) {
		int ii = i * np1;
		for(j=0; j<nph1[i]; j++) {
		    double zm = zmask[ii+j] - zmask_value;

		    if (*hp_mask_file != '\0') {
			  double fm = fmask[ii+j] - zmask_value;
			  fm = ABS(fm);

			  /* fm = 0, when zone is masked */
			  if ( ( (f_mask_flag && /* remove zone of mask */
				   /* mask has 'zmask_value' pixels */
				   fm <= 1.e-6)
			   || (!f_mask_flag &&
				   /* remain zone of mask */
				   fm >= 1.e-6)))
				imask [ii+j] = 1;
			  else
				imask [ii+j] = 0;
		    }
		    if (*gl_mask_file != '\0') {
			  double fm = gmask[ii+j] - zmask_value;
			  fm = ABS(fm);

			  /* fm = 0, when zone is masked */
			  if ( ( (f_mask_flag && /* remove zone of mask */
				   /* mask has 'zmask_value' pixels */
				   fm <= 1.e-6)
			   || (!f_mask_flag &&
				   /* remain zone of mask */
				   fm >= 1.e-6)))
				imask [ii+j] = 1;
			  else
				imask [ii+j] = 0;
		    }

		    if (ABS(zm) < 1.e-6)
		       imask [ii+j] = 1;
		}
	    }

	    map_cor_mask(f1,f2, imask, x1,nx1,np1,nph1, corr_win, &fcor,&x2,
	      &nx2, &np2, &nph2, dummy);

	    free((void *)imask);
	    free((void *)fmask);
	    free((void *)zmask);
	    free((void *)gmask);
	} else
	    /* correlation for total sphere */
	    /*------------------------------*/
	    map_cor(f1,f2,x1,nx1,np1,nph1,corr_win,&fcor,&x2,&nx2,&np2,&nph2);

	free((void *)f1);
	free((void *)f2);
	free((void *)x1);
	free((void *)nph1);

	/* set pointers for output */
	f1 = fcor;
	x1 = x2;
	nx1 = nx2;
	np1 = np2;
	nph1 = nph2;

    } else if (corr_rwin > 0.) {
	/* calculate mapping of correlation */

	double *fcor = (double *)0;
	extern void  map_rcor(double *, double *, double *, int, int, int *,
			 double, double **, double **, int *, int *, int **);

	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	read_glesp_tmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);

	if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
	       PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		      myname, map2_file, map1_file);
	       PR(ER, "Sorry...\n");
	       exit(1);
	}
	free((void *)x2);
	free((void *)nph2);
	x2 = (double *)0;
	nph2 = (int *)0;

	map_rcor(f1,f2, x1,nx1,np1,nph1,corr_rwin, &fcor,&x2,&nx2,&np2,&nph2);

	free((void *)f1);
	free((void *)f2);
	free((void *)x1);
	free((void *)nph1);

	/* set pointers for output */
	f1 = fcor;
	x1 = x2;
	nx1 = nx2;
	np1 = np2;
	nph1 = nph2;

    } else if (repix_flag) {
	extern void map_repix(double *, double *, int, int, int *, double,
	  double **, double **, int *, int *, int **);

	/* make repixelization decreasing pixel size */
	double *map_rp = (double *)0;

	/* read map */
	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	/* repixelize with decreasing pixel size */
	map_repix(f1,x1,nx1,np1,nph1,repix_win, &map_rp,&x2,&nx2,&np2,&nph2);

	free((void *)f1);
	free((void *)x1);
	free((void *)nph1);

	/* set pointers for output */
	f1 = map_rp;
	x1 = x2;
	nx1 = nx2;
	np1 = np2;
	nph1 = nph2;
    } else if (sig_win_flag) {
	/* calculate mapping of correlation */

	double *fsig = (double *)0;
	extern void  map_sig(double *, double *, int, int, int *,
			 double, double **, double **, int *, int *, int **);
	extern void  map_rsig(double *, double *, int, int, int *, double);

	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	if (sig_win_flag == 2) {
	    /* calculation sigma in running window */
	    map_rsig(f1,x1,nx1,np1,nph1, sig_win);
	} else {  /* if (sig_fin_flag == 1) */
	    /* calculate in larger pixel */

	    x2 = (double *)0;
	    nph2 = (int *)0;

	    map_sig(f1,x1,nx1,np1,nph1, sig_win, &fsig,&x2,&nx2,&np2,&nph2);

	    free((void *)f1);
	    free((void *)f2);
	    free((void *)x1);
	    free((void *)nph1);

	    /* set pointers for output */
	    f1 = fsig;
	    x1 = x2;
	    nx1 = nx2;
	    np1 = np2;
	    nph1 = nph2;
	}
    } else if (*fit_map_patt_file != '\0') {
	/* fitting map with pattern map */
	double sigma = 1.0, alpha;
	extern void patt_coef(double *, double *, double *, int,int,int*,
	    double, double *);

	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	read_glesp_tmap(&f2, &x2, &nx2, &np2, &nph2, fit_map_patt_file);
	if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
	       PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		      myname, map2_file, map1_file);
	       PR(ER, "Sorry...\n");
	       exit(1);
	}

	patt_coef(f1, f2, x1, nx1, np1, nph1, sigma, &alpha);
	printf ("alpha = %12.5e\n", alpha);

	exit(0);
    } else if (polariz_flag == 1 || polariz_flag == 2) {
	/* calculate I from Q & U */
	if (map_type == 't' || map_type == 'T') {
	    read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    read_glesp_tmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	} else if (map_type == 'u') {
	    read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    read_glesp_umap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	} else {
	    read_glesp_qmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    read_glesp_qmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	}
	if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
	       PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		      myname, map2_file, map1_file);
	       PR(ER, "Sorry...\n");
	       exit(1);
	}

	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		double q = f1[ii+j], u = f2[ii+j];
		double p_i = sqrt(q*q+u*u),
		       p_a = 0.5*atan2(u,q);
//                     p_a = 0.5*atan(u/q);

		if (polariz_flag == 2) /* radians --> degrees */
		    p_a *= 180./M_PI;

		f1[ii+j] = p_i;  /* intensity */
		f2[ii+j] = p_a;  /* angle     */
	    }
	}
	if (*outp_angle_file != '\0') {
	    /* open output file */
	    if((fp_out = fopen(outp_angle_file, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, outp_angle_file);
		exit(1);
	    }

	    /* angle output in FITS-format like map */
	    outp_glesp_map(f2, x1, nx1, np1, nph1, fp_out, x_type,grid_type);

	    fclose(fp_out);
	}
    } else if (polariz_flag == 3 || polariz_flag == 4) {
	/* calculate I from Q & U */
	if (map_type == 'p' || map_type == 'P') {
	    read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    read_glesp_umap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	}
	if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
	       PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		      myname, map2_file, map1_file);
	       PR(ER, "Sorry...\n");
	       exit(1);
	}

	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		double p_i = f1[ii+j],
		       p_a = f2[ii+j] * polariz_flag == 4 ? M_PI/180. : 1.;
		double q = p_i * cos(2.*p_a),
		       u = p_i * sin(2.*p_a);

		f1[ii+j] = q;  /* Q-parameter */
		f2[ii+j] = u;  /* U-parameter */
	    }
	}
	/* output map of U-parameter */
	if (*u_out_file != '\0') {
	    /* open output file */
	    if((fp_out = fopen(u_out_file, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, u_out_file);
		exit(1);
	    }

	    /* angle output in FITS-format like map */
	    outp_glesp_map(f2, x1, nx1, np1, nph1, fp_out, x_type,grid_type);

	    fclose(fp_out);
	}
	if (*q_out_file != '\0')
	    strcpy(outfile, q_out_file);

    } else if (grad_flag) {
      if (map_type == 't' || map_type == 'T')
	 read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
      else if (map_type == 'u')
	 read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
      else
	 read_glesp_qmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

      if (grad_flag == 1) {  /* gradient by phi */
	map_grad_phi(f1, x1, nx1, np1, nph1);
      } else if (grad_flag == 2) { /* gradient by theta */
	map_grad_theta(f1, x1, nx1, np1, nph1);
      } else if (grad_flag == 3) {
	double *q, *u, *p, *pa;
	/* take memory for arrays */
	if((q=(double*)calloc((unsigned)(nx1*np1),sizeof *q))==(double *)0) {
nomem:
	    PR(ER, "%s: no memory...\n", myname);
	    exit(1);
	}
	if((u=(double*)calloc((unsigned)(nx1*np1),sizeof *u))==(double *)0)
	    goto nomem;

	/* copy arrays */
	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++)
		q[ii+j] = u[ii+j] = f1[ii+j];
	}
	free((void *)f1);

	/* calculate gradients */
	map_grad_phi  (u, x1, nx1, np1, nph1);
	map_grad_theta(q, x1, nx1, np1, nph1);

	/* take more memory */
	if((p=(double*)calloc((unsigned)(nx1*np1),sizeof *p))==(double *)0)
	    goto nomem;
	if((pa=(double*)calloc((unsigned)(nx1*np1),sizeof *pa))==(double *)0)
	    goto nomem;

	/* calculate p=sqrt(u^2+q^2), and pa=atan(u/q) */
	map_grad_pa(q, u, p, pa, x1, nx1, np1, nph1);

	/* free change array pointers */
	free((void *)q);
	free((void *)u);
	f1 = p;
	f2 = pa;

	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++)
		pa[ii+j] = cos(pa[ii+j]);
	}


	if (*outp_angle_file != '\0') {
	    /* open output file */
	    if((fp_out = fopen(outp_angle_file, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, outp_angle_file);
		exit(1);
	    }

	    /* angle output in FITS-format like map */
	    outp_glesp_map(f2, x1, nx1, np1, nph1, fp_out, x_type,grid_type);

	    fclose(fp_out);
	}
      }
    } else if (stat_flag==3 || stat_flag==4 || stat_flag==5) {
	/* count pixels over given level: flag = 3 */
	/* and under given level: flag = 4 */
	int n_count =0;
	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		if (stat_flag == 3 && f1[ii+j] > level_low) n_count++;
		if (stat_flag == 4 && f1[ii+j] < level_up)  n_count++;
		if (stat_flag == 5 &&
		     level_low < f1[ii+j] && f1[ii+j] < level_up) n_count++;
	    }
	}
	if (stat_flag == 5)
	    printf("low_lim = %g, up_lim = %g, number of pixels = %d\n",
		level_low,level_up,n_count);
	else  if (stat_flag == 4)
	    printf("up_lim = %g, number of pixels = %d\n",level_up,n_count);
	else
	    printf("low_level = %g, number of pixels = %d\n",
		 level_low,n_count);
	exit(0);
    } else if (stat_flag==6) {
	struct _rr_ {
	    double lim_up, lim_low;
	    double aver;
	    double sig;
	    int n_count;
	} *rr;
	char str[161];
	int n_r = 0;
	FILE *fp;

	/* read number of intervals from file */
	if ((fp = fopen(file_stat_interv, "r")) == (FILE *)0) {
	    PR(ER, "%s: problem to open file '%s'\n",
		    myname, file_stat_interv);
	    exit(1);
	}
	while(fgets(str, 160, fp) != NULL) {
	    if (*str == '\0' || *str == '#' || *str == '\n'|| *str == '\r')
		continue;
	    n_r++;
	}
	if((rr = (struct _rr_ *)calloc(n_r, sizeof *rr)) == (struct _rr_ *)0){
	    PR(ER, "%s: no memory for intervals array...\n", myname);
	    exit(1);
	}
	fseek(fp, 0l, 0);
	i=0;
	while(fgets(str, 160, fp) != NULL) {
	    if (*str == '\0' || *str == '#' || *str == '\n'|| *str == '\r')
		continue;
	    sscanf(str, "%lf%lf", &rr[i].lim_low, &rr[i].lim_up);
	    rr[i].n_count = 0;
	    rr[i].aver = 0.0;
	    rr[i].sig = 0.0;
	    i++;
	}
	fclose(fp);


	/* read map to calc statistic */
	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	/* count pixels */
	for(i=0; i<nx1; i++) {
	    int ii = i * np1, c;
	    for(j=0; j<nph1[i]; j++) {
		double ff_ij = f1[ii+j];
		for (c=0; c<n_r; c++) {
		    if(rr[c].lim_low <= ff_ij && ff_ij < rr[c].lim_up){
			rr[c].n_count++;
			rr[c].aver += ff_ij;
		    }
		}
	    }
	}
	/* average */
	for (i=0; i<n_r; i++) {
	    rr[i].aver /= (double)rr[i].n_count;
	}
	/* sum of sig^2 */
	for(i=0; i<nx1; i++) {
	    int ii = i * np1, c;
	    for(j=0; j<nph1[i]; j++) {
		double ff_ij = f1[ii+j];
		for (c=0; c<n_r; c++) {
		    if(rr[c].lim_low <= ff_ij && ff_ij < rr[c].lim_up){
			rr[c].sig += (ff_ij-rr[c].aver)*(ff_ij -rr[c].aver);
		    }
		}
	    }
	}
	for (i=0; i<n_r; i++) {
	    rr[i].sig  = sqrt(rr[i].sig / (double)rr[i].n_count);
	}

	for (i=0; i<n_r; i++) {
	   printf("%12.5e %12.5e  %12.5e %12.5e  %10d\n",
	      rr[i].lim_low,rr[i].lim_up,rr[i].aver,rr[i].sig,rr[i].n_count);
	}

	free((void *)rr);
	exit(0);
    } else if (stat_flag==7) {
	struct _rr_ {
	    double lim_up, lim_low;
	    double aver;
	    double sig;
	    int n_count;
	} *rr, *dd, *ss;
	char str[161];
	int n_r = 0;
	FILE *fp;

	/* read number of intervals from file */
	if ((fp = fopen(file_stat_interv, "r")) == (FILE *)0) {
	    PR(ER, "%s: problem to open file '%s'\n",
		    myname, file_stat_interv);
	    exit(1);
	}
	while(fgets(str, 160, fp) != NULL) {
	    if (*str == '\0' || *str == '#' || *str == '\n'|| *str == '\r')
		continue;
	    n_r++;
	}
	if((rr = (struct _rr_ *)calloc(n_r, sizeof *rr)) == (struct _rr_ *)0){
nomem5:
	    PR(ER, "%s: no memory for intervals array...\n", myname);
	    exit(1);
	}
	if((dd = (struct _rr_ *)calloc(n_r, sizeof *dd)) == (struct _rr_ *)0)
	    goto nomem5;
	if((ss = (struct _rr_ *)calloc(n_r, sizeof *ss)) == (struct _rr_ *)0)
	    goto nomem5;

	/* read limits from file */
	fseek(fp, 0l, 0);
	i=0;
	while(fgets(str, 160, fp) != NULL) {
	    if (*str == '\0' || *str == '#' || *str == '\n'|| *str == '\r')
		continue;
	    sscanf(str, "%lf%lf", &rr[i].lim_low, &rr[i].lim_up);
	    rr[i].n_count = 0;
	    rr[i].aver = 0.0;
	    i++;
	}
	fclose(fp);

	/* read maps to calc statistic */
	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	read_glesp_tmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
	       PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		      myname, map2_file, map1_file);
	       PR(ER, "Sorry...\n");
	       exit(1);
	}

	/* count pixels */
	for(i=0; i<nx1; i++) {
	    int ii = i * np1, c;
	    for(j=0; j<nph1[i]; j++) {
		double ff1_ij = f1[ii+j];
		double ff2_ij = f2[ii+j];
		for (c=0; c<n_r; c++) {
		    if(rr[c].lim_low <= ff1_ij && ff1_ij < rr[c].lim_up){
			rr[c].n_count++;
			rr[c].aver += ff1_ij;
			dd[c].aver += ff2_ij;
			if (fabs(ff2_ij) < 1.e-40)
			    ss[c].aver += 0;
			else
			    ss[c].aver += ff1_ij / ff2_ij;
		    }
		}
	    }
	}
	/* average */
	for (i=0; i<n_r; i++) {
	    if ( rr[i].n_count == 0) {
		rr[i].aver = 0.0;
		dd[i].aver = 0.0;
		ss[i].aver = 0.0;
	    } else {
		rr[i].aver /= (double)rr[i].n_count;
		dd[i].aver /= (double)rr[i].n_count;
		ss[i].aver /= (double)rr[i].n_count;
	    }
	}
	/* sum of sig^2 */
	for(i=0; i<nx1; i++) {
	    int ii = i * np1, c;
	    for(j=0; j<nph1[i]; j++) {
		double ff1_ij = f1[ii+j];
		double ff2_ij = f2[ii+j];
		double sr = ff1_ij / ff2_ij;
		for (c=0; c<n_r; c++) {
		    if(rr[c].lim_low <= ff1_ij && ff1_ij < rr[c].lim_up){
			rr[c].sig += (ff1_ij-rr[c].aver)*(ff1_ij -rr[c].aver);
			dd[c].sig += (ff2_ij-dd[c].aver)*(ff2_ij -dd[c].aver);
			if (fabs(ff2_ij) < 1.e-40)
			    ss[c].sig += 0;
			else
			    ss[c].sig += (sr -ss[c].aver) * (sr -ss[c].aver);
		    }
		}
	    }
	}
	for (i=0; i<n_r; i++) {
	    if ( rr[i].n_count == 0) {
		rr[i].sig  = 0.;
		dd[i].sig  = 0.;
		ss[i].sig  = 0.;
	    } else {
		rr[i].sig  = sqrt(rr[i].sig / (double)rr[i].n_count);
		dd[i].sig  = sqrt(dd[i].sig / (double)rr[i].n_count);
		ss[i].sig  = sqrt(ss[i].sig / (double)rr[i].n_count);
	    }
	}

	for (i=0; i<n_r; i++) {
	   printf("%12.5e %12.5e  %12.5e %12.5e  %10d  %12.5e %12.5e  %12.5e %12.5e  %12.5e \n",
	      rr[i].lim_low,rr[i].lim_up,rr[i].aver,rr[i].sig,rr[i].n_count,
	      dd[i].aver,dd[i].sig, ss[i].aver, ss[i].sig,
	      1. / (1 + ss[i].aver));
	}

	free((void *)rr);
	free((void *)dd);
	free((void *)ss);
	exit(0);
    } else if (stat_flag) {
	/* calculate statistics */
	double min= 1.e+30, max= -1.e+30;
	double min_z= 1.e+30, max_z= -1.e+30;
	int i_min=0, j_min=0, i_max=0, j_max=0;
	int ntot=0, n_pix_lev = 0, ntot_real = 0;
	double sigma=0., aver=0., sigma_real=0., aver_real=0.;
	int *hg = (int *)0;
	double step_hg = 0.0;
	FILE *fp_hg = (FILE *)0;
	double *fmask=(double *)0, *gmask=(double *)0;
	double aver_in_circle= 0.0,  disp_in_circle = 0.0;

	if (map_type == 't' || map_type == 'T')
	    read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	else if (map_type == 'u')
	    read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	else
	    read_glesp_qmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	if (zmask_flag) {
	    if ((zmask=(double*)calloc((unsigned)(nx1*np1), sizeof *zmask))==
	      (double *)0) {
		PR(ER, "%s: no memory for mask array... No mask.\n", myname);
		zmask_flag = 0;
	    }
	}
	if (*hp_mask_file != '\0') {
	    if ((fmask=(double*)calloc((unsigned)(nx1*np1), sizeof *fmask))==
	      (double *)0) {
		PR(ER, "%s: no memory for mask array... No mask.\n", myname);
		zmask_flag = 0;
	    }
	}
	if (*gl_mask_file != '\0') {
	    if ((gmask=(double*)calloc((unsigned)(nx1*np1), sizeof *gmask))==
	      (double *)0) {
		PR(ER, "%s: no memory for mask array... No mask.\n", myname);
		zmask_flag = 0;
	    }
	}

	/* read region for masking */
	/*-------------------------*/
	if ( zmask_flag ) {
	   extern void map_zone(struct _zone_ *, int, int,
		   double *, double *, int, int, int *, double);
	   map_zone(zz,zns,function_ztype,zmask,x1,nx1,np1,nph1,zmask_value);

	   /* set mask from HPix-file */
	   if (*hp_mask_file != '\0') {
	      extern void mask_hp(double*,double*,int,int,int*,char*,double);
	      i = nx1;
	      mask_hp(fmask,x1,i,np1,nph1,hp_mask_file,zmask_value);
	   }
	   /* set mask from GLESP-file */
	   if (*gl_mask_file != '\0') {
	      extern void mask_gl(double*,double*,int,int,int*,char*,double);
	      i = nx1;
	      mask_gl(gmask,x1,i,np1,nph1,gl_mask_file,zmask_value);
	   }
	}

	/* calculate statistics in inner circle */
	if (flag_kZS) {
	    extern void get_circle_array (double, double, double, double *,
		double *, double *, int, int, int *, double **, int *);
	    double *ar = (double *)0;
	    int nar = 0;                /* number of pixels in array */

	    /* extract pixels from in radius r_c */
	    get_circle_array (lat_c, lon_c, r_c, zmask, f1, x1, nx1,np1,nph1,
		&ar, &nar);

	    for(i=0; i<nar; i++)  aver_in_circle += ar[i];
	    aver_in_circle /= nar;
	    for(i=0; i<nar; i++) {
		    double d = ar[i] - aver_in_circle;
		    disp_in_circle += d*d;
	    }
	    disp_in_circle = sqrt(disp_in_circle/nar);
	    if (ar != (double *)0) free((void *)ar);
	}


	/* set mask for            */
	/* concentric circles/ring */
	if (flag_conc_fig) {
	    extern void map_ring_zone();

	    i = zns = 1;
	    zz = (struct _zone_ *)calloc(i, sizeof *zz);
	    if (zz == (struct _zone_ *)0) goto nomem;
	    if ( flag_conc_fig == 1) {
		zz[0].latR0 = lat_c;  zz[0].lonR0 = lon_c;
		zz[0].radius = r_c;
		zz[0].dr = w_c;
		zz[0].type = 0; /* circle */
	    } else {
		zz[0].latR0 = lat_c;   zz[0].lonR0 = lon_c;
		zz[0].latR1 = lat1_c;  zz[0].lonR1 = lon1_c;
		zz[0].dr = w_c;
		zz[0].type = 1; /* rectangular */
	    }
	    map_ring_zone(zz,zns,function_ztype,zmask,x1,nx1,np1,nph1,
		   zmask_value);
	}

	/* calculate statistics */
	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		double v = f1[ii+j];

		if (v > max)   { max = v; i_max = i; j_max = j; }
		if (v < min)   { min = v; i_min = i; j_min = j; }
		if (level_flag) {
			if (fabs(level-v) < eps)  n_pix_lev++;
		}
		if (zmask_flag) {
		  double zm = zmask[ii+j] - zmask_value;

		  if (*hp_mask_file != '\0') {
			double fm = fmask[ii+j] - zmask_value;

			fm = ABS(fm);
			if (f_mask_flag) { /* remove zone of mask */
			    if ( fm <= 1.e-6 )
				continue;
			} else {           /* remain zone of mask */
			    if ( fm >= 1.e-6 )
				continue;
			}
		  }
		  if (*gl_mask_file != '\0') {
			double fm = gmask[ii+j] - zmask_value;

			fm = ABS(fm);
			if ( ( (f_mask_flag && /* remove zone of mask */
				 /* mask has 'zmask_value' pixels */
				 fm <= 1.e-6)
			 || (!f_mask_flag &&
				 /* remain zone of mask */
				 fm >= 1.e-6)))
			      continue;
		  }

		  if (ABS(zm) > 1.e-6) {
			if (v > max_z)   max_z = v;
			if (v < min_z)   min_z = v;
			ntot_real++;
			aver_real += v;
		  }
		}
		aver += v;
	    }
	    ntot += nph1[i];
	}
	aver /= (double)ntot;
	if (zmask_flag) aver_real /= (double)ntot_real;

	if (flag_kZS) {
	    PR(ER, "circle (r=%12.5e): a=%12.5e+/-%12.5e, ",
		    r_c, aver_in_circle, disp_in_circle);
	    PR(ER, "ring (R=%12.5e): %12.5e+/-%12.5e, d=%12.5e\n",
		    r_c+w_c, aver_in_circle, disp_in_circle,
		    aver_in_circle-aver_in_circle);
	}

	/* histohram intervals */
	if (n_hg <= 0)  {
	    if (zmask_flag) {
		n_hg = 1. + log ((double)ntot_real) / log(2.);
	    } else {
		n_hg = 1. + log ((double)ntot) / log(2.);
	    }
	}
	/* processing the histogram */
	if (*hg_file != '\0') {
	    /* set limit for histogram */
	    if (s_hg0 <  1.e+30)
		min_z = min = s_hg0;
	    if (s_hgF > -1.e+30)
		max_z = max = s_hgF;


	    if((fp_hg = fopen(hg_file, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n", myname, hg_file);
		fp_hg = (FILE *)0;
	    } else {
		if((hg = (int *)calloc(n_hg, sizeof *hg)) == (int *)0) {
		  PR(ER, "%s: no memory for histogram array of %d elements\n",
			myname, n_hg);
		  fclose(fp_hg);
		  unlink(hg_file);
		  fp_hg = (FILE *)0;
		} else {
		  if (zmask_flag)
			step_hg = (max_z - min_z) / n_hg;
		  else
			step_hg = (max - min) / n_hg;
		}
	    }
	}

	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		double v = f1[ii+j] - aver;

		if ( ABS(f1[ii+j] - GLESPMASK) < 1e-6 )
		    continue;

		if (zmask_flag) {
		    double zm = zmask[ii+j] - zmask_value;
		    if (ABS(zm) > 1.e-6) {
			  double v_r = f1[ii+j] - aver_real;
			  sigma_real += v_r*v_r;
		    }
		}


		sigma += v * v;
		if (hg != (int *)0) {
		    if (zmask_flag) {
			double zm = zmask[ii+j] - zmask_value;
			if (*hp_mask_file != '\0') {
			      double fm = fmask[ii+j] - zmask_value;

			      fm = ABS(fm);
			      if ((( f_mask_flag && fm <= 1.e-6 ) ||
				  (!f_mask_flag && fm >= 1.e-6 )))
				      continue;
			}
			if (*gl_mask_file != '\0') {
			      double fm = gmask[ii+j] - zmask_value;

			      fm = ABS(fm);
			      if ((( f_mask_flag && fm <= 1.e-6 ) ||
				  ( !f_mask_flag && fm >= 1.e-6 )))
				      continue;
			}
			if (ABS(zm) > 1.e-6) {
//                        if ((z_remove_flag && ABS(zm) > 1.e-6) ||
//                            (z_remain_flag && ABS(zm) >= 1.e-6 )) {
			      int bin = (f1[ii+j] - min_z) / step_hg ;
			      if (bin > n_hg)    continue;
			      else if (bin > n_hg-1) bin--;
			      else if (bin < 0)    continue;
			      hg[bin]++;

			}
		    } else {
			int bin = (f1[ii+j] - min) / step_hg ;
			if (bin > n_hg)    continue;
			else if (bin > n_hg-1) bin--;
			else if (bin < 0)    continue;
			hg[bin]++;
		    }
		}
	    }
	}
	sigma /= (double)ntot;
	if (zmask_flag) sigma_real /= (double)ntot_real;

	if (hg != (int *)0) {
	    if (zmask_flag) {
		double norm = 1.;
		if (norm_flag)
			norm /= (double)(norm_flag==1 ? ntot : ntot_real);
		for(i=0; i<n_hg; i++) {
		   PR(fp_hg, "%12.5e  %12.5e\n",
			min_z+(i+0.5)*step_hg, hg[i] * norm);
		}
	    } else {
		for(i=0; i<n_hg; i++) {
		   PR(fp_hg, "%12.5e  %12.5e\n",
		      min+(i+0.5)*step_hg,hg[i]/(double)(norm_flag ? ntot:1));
		}
	    }
	}

	PR(stdout, "ntot = %d, nx = %d, np = %d\n", ntot, nx1, np1);
	PR(stdout, "min = %15.8e at (%d,%d), max = %15.8e at (%d,%d)\n",
		min, i_min, j_min, max, i_max, j_max);
	PR(stdout,"(theta,phi)_min = (%12.5f,%12.5f)=(%9.6f,%10.6f)degs,\n",
		  acos(x1[i_min]),  2*M_PI/nph1[i_min]*j_min,
		  acos(x1[i_min])*180/M_PI,  360./nph1[i_min]*j_min);
	PR(stdout,"(theta,phi)_max = (%12.5f,%12.5f)=(%9.6f,%10.6f)degs\n",
		  acos(x1[i_max]),  2*M_PI/nph1[i_max]*j_max,
		  acos(x1[i_max])*180/M_PI,  360./nph1[i_max]*j_max);
	PR(stdout, "aver = %14.7e,  sigma^2 = %14.7e\n", aver, sigma);
	if (zmask_flag) {
	   PR(stdout, "ntot_real =%d, aver_r = %12.5e,  sigma_r^2 = %12.5e\n",
	      ntot_real, aver_real, sigma_real);
	}

	if (level_flag) {
	  PR(stdout,"level+-eps = (%12.5e)+-(%12.5e), n_pix=%d (%4.1f%%)\n",
		  level, eps, n_pix_lev,  (double)n_pix_lev/ntot*100.);
	}

	if (stat_flag == 2) {
	    /* some isolatitude ring statistics */
	    PR(stdout, "\nRing statistics:\n");
	    PR(stdout, "   N       Theta       Np     aver        sigma^2");
	    PR(stdout, "        min          max\n");
	    for(i=0; i<nx1; i++) {
		int ii = i * np1, n=nph1[i];
		sigma = aver = 0.0;
		min = 1.e+30; max = -min;
		for(j=0; j<n; j++)  aver += f1[ii+j];
		aver /= (double)n;
		for(j=0; j<n; j++)  {
		    double ff=f1[ii+j], v = ff - aver ;
		    sigma += v * v;
		    if (min > ff)  min = ff;
		    if (max < ff)  max = ff;
		}
		sigma /= (double)n;
		PR(stdout, "%5d  %12.5e %5d %12.5e %12.5e %12.5e %12.5e\n",
		      i, acos(x1[i]), n, aver, sigma, min, max);
	    }
	}

	/* calculate extrema */
	if (ex_stat) {
	    int *mex = (int *)0;
	    int n_max=0, n_min = 0;
	    extern int map_extrem(), map_extrem_mask();
	    FILE *fp_e = (FILE *)0;

	    mex = (int *)calloc((unsigned)(np1*nx1), sizeof *mex);
	    if (mex == (int *)0)
		goto nomem;

	    if (*out_extr_file != '\0') {
		fp_e = fopen(out_extr_file, "w");
		if (fp_e == (FILE *)0){
		  PR(ER,"%s: can not open file '%s'\n",myname,out_extr_file);
		  exit(1);
		}
	    }

	    if (zmask_flag) {
		/* find extrema outside mask */
		int *imask = (int *)0;
		imask = (int *)calloc((unsigned)(np1*nx1), sizeof *imask);
		if (imask == (int *)0)
		    goto nomem;
		for(i=0; i<nx1; i++) {
		    int ii = i * np1;
		    for(j=0; j<nph1[i]; j++) {
			double zm = zmask[ii+j] - zmask_value;
			/* outside mask */
			if (!((z_remove_flag && ABS(zm) > 1.e-6) ||
			    (z_remain_flag && ABS(zm) >= 1.e-6 ))) {
			      imask[ii+j] = 1;
			}
		    }
		}
		map_extrem_mask(f1, x1, nx1, np1, nph1, imask, mex, dp_ex);
		free ((void *)imask);
	    } else
		/* find extrema on a total map */
		map_extrem(f1, x1, nx1, np1, nph1, mex, dp_ex);

	    for(i=0; i<nx1; i++) {
		int ii = i * np1;
		double dp = 2*M_PI / nph1[i];
		for(j=0; j<nph1[i]; j++) {
		    if (mex[ii+j] ==  1) n_max++;
		    if (mex[ii+j] == -1) n_min++;
		    if (*out_extr_file != '\0' && mex[ii+j]) {
			double theta=acos(x1[i]), phi=j*dp, glat, glon;
			pol2gal(theta, phi, &glat, &glon);
			glat *= 180./M_PI;
			glon *= 180./M_PI;
			PR(fp_e, "%12.5ed %12.5ed %12.5e ",
			    glat, glon, f1[ii+j]);
			if (mex[ii+j] ==  1)
			    PR(fp_e, "max\n");
			else
			    PR(fp_e, "min\n");
		    }
		}
	    }
	    free((void *)0);
	    PR(ST,"N of minima = %d, N of maxima = %d\n", n_min, n_max);
	    if (fp_e != (FILE *)0) fclose(fp_e);
	}

	if(fp_hg != (FILE *)0)   fclose(fp_hg);
	if(hg    != (int *)0) free((void *)hg);
	if (fmask != (double *)0) free((void *)fmask);

	exit(0);
    } else if (estimator_type) {
	 extern void stat_estim(double *, double *, int, int, int *,
		char *, double *, int, int);
	 double *m_est = (double *)0;
	 /* maximum multipole for stat. analysis */
	 int lmax_stat = esl_lmax_stat < 1 ? 50 : esl_lmax_stat;

	 read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	 if ((m_est=(double*)calloc((unsigned)(nx1*np1), sizeof *m_est))==
	      (double *)0) {
		PR(ER, "%s: no memory for estimator map...\n", myname);
		exit(1);
	 }

if (glesp_comments)
PR(ER, "lmax_stat=%d for temporal arrays.\n", lmax_stat);
	 stat_estim(f1, x1, nx1, np1, nph1, map1_file, m_est, estimator_type,
		lmax_stat);

	 free((void *)f1);
	 f1 = m_est;

    } else if (corr_flag) {
	if (corr_flag == 1) {
	    double aver1=0.0, aver2=0.0, sigma1=0.0, sigma2=0.0, corr=0.0;
	    int ntot = 0;
	    double *fmask=(double *)0, *gmask=(double *)0;

	    /* calculation of correlation coefficient of 2 maps */
	    if (map_type == 't' || map_type == 'T') {
		read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
		read_glesp_tmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	    } else if (map_type == 'u') {
		read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
		read_glesp_umap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	    } else {
		read_glesp_qmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
		read_glesp_qmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	    }

	    if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
		   PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
			  myname, map2_file, map1_file);
		   PR(ER, "Sorry...\n");
		   exit(1);
	    }

	    if (zmask_flag) {
	      if ((zmask=(double*)calloc((unsigned)(nx1*np1), sizeof *zmask))
		== (double *)0) {
		  PR(ER,"%s: no memory for mask array... No mask.\n",myname);
		  zmask_flag = 0;
	      }
	    }
	    if (*hp_mask_file != '\0') {
	      if((fmask=(double*)calloc((unsigned)(nx1*np1),sizeof *fmask))
		  == (double *)0) {
		  PR(ER,"%s: no memory for mask array... No mask.\n",myname);
		  zmask_flag = 0;
	      }
	    }
	    if (*gl_mask_file != '\0') {
	      if((gmask=(double*)calloc((unsigned)(nx1*np1),sizeof *gmask))
		  == (double *)0) {
		  PR(ER,"%s: no memory for mask array... No mask.\n",myname);
		  zmask_flag = 0;
	      }
	    }

	    /* read region for masking */
	    /*-------------------------*/
	    if ( zmask_flag ) {
	       extern void map_zone(struct _zone_ *, int, int,
		       double *, double *, int, int, int *, double);
	       map_zone(zz,zns,function_ztype,zmask,x1,nx1,np1,nph1,
		    zmask_value);

	       /* set mask from HPix-file */
	       if (*hp_mask_file != '\0') {
		  extern
		    void mask_hp(double*,double*,int,int,int*,char*,double);
		  i = nx1;
		  mask_hp(fmask,x1,i,np1,nph1,hp_mask_file,zmask_value);
	       }
	       if (*gl_mask_file != '\0') {
		  extern
		    void mask_gl(double*,double*,int,int,int*,char*,double);
		  i = nx1;
		  mask_gl(gmask,x1,i,np1,nph1,gl_mask_file,zmask_value);
	       }
	    }

	    /* average */
	    for(i=0; i<nx1; i++) {
		int ii = i * np1;
		for(j=0; j<nph1[i]; j++) {
		    if (zmask_flag) {
			double zm = zmask[ii+j] - zmask_value;
			if (ABS(zm) > 1.e-6) {
			      double v1 = f1[ii+j];
			      double v2 = f2[ii+j];
			      aver1 += v1;
			      aver2 += v2;
			      ntot++;
			}
		    }  else {
			double v1 = f1[ii+j];
			double v2 = f2[ii+j];
			aver1 += v1;
			aver2 += v2;
			ntot++;
		    }
		}
	    }
	    aver1 /= (double)ntot;
	    aver2 /= (double)ntot;

	    /* sigma */
	    for(i=0; i<nx1; i++) {
		int ii = i * np1;
		for(j=0; j<nph1[i]; j++) {
		    if (zmask_flag) {
			double zm = zmask[ii+j] - zmask_value;
			if (ABS(zm) > 1.e-6) {
			    double v1 = f1[ii+j] - aver1;
			    double v2 = f2[ii+j] - aver2;
			    corr += v1 * v2;
			    sigma1 += v1 * v1;
			    sigma2 += v2 * v2;
			}
		    }  else {
			double v1 = f1[ii+j] - aver1;
			double v2 = f2[ii+j] - aver2;
			corr += v1 * v2;
			sigma1 += v1 * v1;
			sigma2 += v2 * v2;
		    }
		}
	    }
	    corr   /= (double)ntot;
	    sigma1 /= (double)ntot;
	    sigma2 /= (double)ntot;
	    corr /= sqrt(sigma1 * sigma2);

	    PR(stdout, "ntot = %d, nx = %d, np = %d\n", ntot, nx1, np1);
	    PR(stdout, "aver1 = %12.5e,  sigma1^2 = %12.5e\n", aver1, sigma1);
	    PR(stdout, "aver2 = %12.5e,  sigma2^2 = %12.5e\n", aver2, sigma2);
	    PR(stdout, "correlation coefficient = %12.5e\n", corr);
	    if (zmask) PR(stdout, "mask was used.\n");

	    if (fmask != (double *)0) free((void *)fmask);
	    if (gmask != (double *)0) free((void *)gmask);
	    exit(0);
	} else {
	    /* correlation for hemispheres */
	    double aver1=0.0, aver2=0.0, sigma1=0.0, sigma2=0.0, corr=0.0;
	    int ntot = 0;
	    int nn;

	    /* calculation of correlation coefficient of 2 maps */
	    read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    nn = (nx1 + 1)/2;

	    /* average */
	    for(i=0; i<nn; i++) {
		int ii = i * np1;
		int ii2 = (nx1-1-i) * np1;
		for(j=0; j<nph1[i]; j++) {
		    double v1 = f1[ii+j];
		    double v2 = f1[ii2+j];
		    aver1 += v1;
		    aver2 += v2;
		}
		ntot += nph1[i];
	    }
	    aver1 /= (double)ntot;
	    aver2 /= (double)ntot;

	    /* sigma */
	    for(i=0; i<nn; i++) {
		int ii = i * np1;
		int ii2 = (nx1-1-i) * np1;
		for(j=0; j<nph1[i]; j++) {
		    double v1 = f1[ii+j] - aver1;
		    double v2 = f1[ii2+j] - aver2;
		    corr += v1 * v2;
		    sigma1 += v1 * v1;
		    sigma2 += v2 * v2;
		}
	    }
	    corr   /= (double)ntot;
	    sigma1 /= (double)ntot;
	    sigma2 /= (double)ntot;

	    corr /= sqrt(sigma1 * sigma2);

	    PR(stdout, "ntot = %d, nx = %d, np = %d\n", ntot, nx1, np1);
	    PR(stdout, "aver1 = %12.5e,  sigma1^2 = %12.5e\n", aver1, sigma1);
	    PR(stdout, "aver2 = %12.5e,  sigma2^2 = %12.5e\n", aver2, sigma2);
	    PR(stdout, "correlation coefficient of 2 hemispheres = %12.5e\n",
		    corr);

	    exit(0);
	}
    } else if (do_stacking) {
	extern void area_stack(double *, double *, int, int, int *,
	    char *, double **, int *, int *, double *, double *, int);
	double *area = (double *)0;
	int nx=0, ny=0;
	double dx=0.0, dy = 0.0;
	int flag_coor = 0;
	register int i, j;
	double amin= 1e30, amax= -1e30, av=0.0, sig=0.0;

	if (do_stacking == 2) flag_coor = 1;

	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	area_stack(f1,x1, nx1,np1,nph1, file_stack,
		 &area, &nx, &ny, &dx, &dy, flag_coor);

	if ((fp_out = fopen(outfile, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, outfile);
		exit(1);
	}
	if (nx && ny && glesp_comments) {
	    for(i=0; i<ny; i++) {
		  int ii = i * nx;
		  for(j=0; j<nx; j++) {
		    double v = area[ii+j];
		    if (v < amin)  amin = v;
		    if (v > amax)  amax = v;
		    av += v;
		  }
	    }
	    av /= nx * ny;
	    for(i=0; i<ny; i++) {
		  int ii = i * nx;
		  for(j=0; j<nx; j++) {
		    double v = area[ii+j] - av;
		    sig += v * v;
		  }
	    }
	    sig /= nx * ny;
	    sig = sqrt(sig);
	    PR(ER, "aver = %12.5e, sig=%12.5e, min = %12.5e, max = %12.5e\n",
		av, sig, amin, amax);
	}

	if (flag_coor) { /* equatorial system */
		/* change variables for output */
		double d_ra = dx, d_dec = dy, ra0 = 0.0, dec0 = 0.0;
		d_ra *= 15.;  /* dRA,  deg */
		ra0  *= 15.;  /* RA,   deg */

		/* write FITS-map */
		write_2d_fits(fp_out, area, nx, ny, ra0, dec0, d_ra, d_dec,
		     "RA---TAN", "DEC--TAN", 1., "T");
	} else {
		double dp = dx,
		       dt = -dy; /* minus because from South to North */
		double lon0 = 0.0, lat0 = 0.0;

		/* write FITS-map */
		write_2d_fits(fp_out, area, nx,ny, lon0, lat0, dp, dt,
		   "GLON", "GLAT", 1., "T");
	}

	exit(0);

    } else if (mink_func) {
	/* calculation of Minkowski functionals 1 and 2 */
	extern void minkowski_func(double *, double *, double *, int, int,
	      int *, double *, double *, char *),
	   minkowski_func_sq(double *, double *, double *, int, int,
	      int *, double,double,double,double, double *,double *, char *);
	double *pw = (double *)0, *mf1 = (double *)0, *mf2 = (double *)0;
	double lat1 = zone_latR0, lon1 = zone_lonR0,
	       lat2 = zone_latR1, lon2 = zone_lonR1;
	int flag = (lat1 >-900 && lon1> -900 && lat2 >-900 && lon2>-900)? 1:0;

	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	if ((pw=(double *)calloc((unsigned)nx1, sizeof *pw))==(double *)0) {
		PR(ER, "%s: mo memory for temporal arrays.\n", myname);
		exit(1);
	}
	weights(1, -1., 1., 2*M_PI, x1, pw, nx1, nph1);

	if (!flag && (lat1 >-900 ||lon1 >-900 ||lat2 >-900 ||lon2 >-900)) {
	  PR(ER,"\nProblem with area size. Full map will be processed...\n");
	}

	if (flag)
	    minkowski_func_sq(f1, pw, x1, nx1, np1, nph1, lat1,lon1,lat2,lon2,
		mf1, mf2, NULL);
	else
	    minkowski_func(f1, pw, x1, nx1, np1, nph1, mf1, mf2, NULL);

	exit(0);
    } else if (*map2_file != '\0') {
	if (map_type == 't') {
	    /* check whether both files are 3-map files ? */
	    if (test_glesp_map(map1_file)==3 && test_glesp_map(map1_file)==3)
		map_type = 3;
	}

	/* read 2 mapes */
	if (map_type == 't' || map_type == 'T') {
	    read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	    read_glesp_tmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	} else if (map_type == 'u') {
	   read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	   read_glesp_umap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	} else if (map_type == 'q') {
	   read_glesp_qmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	   read_glesp_qmap(&f2, &x2, &nx2, &np2, &nph2, map2_file);
	} else { /* 3-map */
	   read_glesp_3map(&f1, &q1, &u1, &x1, &nx1, &np1, &nph1, map1_file);
	   read_glesp_3map(&f2, &q2, &u2, &x2, &nx2, &np2, &nph2, map2_file);
	   if (u1 == (double *)0 || q1 == (double *)0 ||
	       u2 == (double *)0 || q2 == (double *)0) {
		   PR(ER,
		      "%s:\nInput polarization data file is not full.\n",
		      myname);
		   PR(ER, "Only anisotropy data are used.\n");
		   map_type = 't';
	   }
	}
	/* compare maps resolution */
	if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
	       PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		      myname, map2_file, map1_file);
	       PR(ER, "Sorry...\n");
	       exit(1);
	}

	/*******************/
	/* find difference */
	/*******************/
	c1 *= cf;
	c2 *= cf;
	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		 switch (action) {
		 case 1 :  /* sum */
		      f1[ii+j] = c1 * f1[ii+j] + c2 * f2[ii+j] + sc;
		      break;
		 case 2 :  /* multiplication */
		      f1[ii+j] = c1 * f1[ii+j] * c2 * f2[ii+j] + sc;
		      break;
		 case 3 :  /* division */
		      f1[ii+j] = c1 * f1[ii+j] / c2 / f2[ii+j] + sc;
		      break;
		 case 0 :
		 default:
		      f1[ii+j] = c1 * f1[ii+j] - c2 * f2[ii+j] + sc;
		      break;
		 }
	    }
	}
	if (zmask_flag) {
	    if ((zmask=(double*)calloc((unsigned)(nx1*np1), sizeof *zmask))==
	      (double *)0) {
		PR(ER, "%s: no memory for mask array... No mask.\n", myname);
		zmask_flag = 0;
	    }
	}

	/* read region for masking */
	/*-------------------------*/
	if ( zmask_flag ) {
	    extern void map_zone(struct _zone_ *, int, int,
		   double *, double *, int, int, int *, double);
	    map_zone(zz,zns,function_ztype,zmask,x1,nx1,np1,nph1,zmask_value);
	    for(i=0; i<nx1; i++) {
	      int ii = i * np1;
	      for(j=0; j<nph1[i]; j++) {
		  int ij=ii+j;
		  double zm = zmask[ij] - zmask_value;
		  if(( (z_remove_flag && ABS(zm) >= 1.e-6)  ||
		       (z_remain_flag && ABS(zm) <  1.e-6) )) {
			f1[ij] = dummy;
		  }
	      }
	    }
	}

	if (map_type == 3) {
	    for(i=0; i<nx1; i++) {
		int ii = i * np1;
		for(j=0; j<nph1[i]; j++) {
		     switch (action) {
		     case 1 :  /* sum */
			  q1[ii+j] = c1 * q1[ii+j] + c2 * q2[ii+j];
			  u1[ii+j] = c1 * u1[ii+j] + c2 * u2[ii+j];
			  break;
		     case 2 :  /* multiplication */
			  q1[ii+j] = c1 * q1[ii+j] * c2 * q2[ii+j];
			  u1[ii+j] = c1 * u1[ii+j] * c2 * u2[ii+j];
			  break;
		     case 3 :  /* division */
			  q1[ii+j] = c1 * q1[ii+j] / c2 / q2[ii+j];
			  u1[ii+j] = c1 * u1[ii+j] / c2 / u2[ii+j];
			  break;
		     case 0 :
		     default:
			  q1[ii+j] = c1 * q1[ii+j] - c2 * q2[ii+j];
			  u1[ii+j] = c1 * u1[ii+j] - c2 * u2[ii+j];
			  break;
		     }
		}
	    }
	    /* read region for masking */
	    /*-------------------------*/
	    if ( zmask_flag ) {
		for(i=0; i<nx1; i++) {
		  int ii = i * np1;
		  for(j=0; j<nph1[i]; j++) {
		      int ij=ii+j;
		      double zm = zmask[ij] - zmask_value;
		      if(( (z_remove_flag && ABS(zm) >= 1.e-6)  ||
			   (z_remain_flag && ABS(zm) <  1.e-6) )) {
				   f1[ij] = dummy;
		      }
		  }
		}
	    }
	}

	/*************************************/
	/* and then write map to output file */
	/*************************************/

	/* free memory */
	free((void *)f2);
	if (map_type == 3) {
	    free((void *)q2);
	    free((void *)u2);
	}

    } else if (convert_coors) {
	if (map_type == 't' || map_type == 'T')
	    read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	else if (map_type == 'u')
	    read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	else
	    read_glesp_qmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	f2 = (double *)calloc((unsigned)(nx1*np1), sizeof *f2);
	if (f2 == (double *)0) {
		PR(ER, "%s: no space for second array...\n", myname);
		exit(1);
	}

	switch (convert_coors) {
	case 'q':
	    gal2eq_map(f1, x1, nx1, np1, nph1, f2);
	    break;
	case 'e':
	    gal2ec_map(f1, x1, nx1, np1, nph1, f2);
	    break;
	}

	/* free memory */
	if (f1 != (double *)0)  free((void *)f1);

	f1 = f2;
    } else if (mirror_flag) {
	read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	f2 = (double *)calloc((unsigned)(nx1*np1), sizeof *f2);
	if (f2 == (double *)0) {
		PR(ER, "%s: no space for second array...\n", myname);
		exit(1);
	}

	/* mirror reflection by equator, 0-meridian or center */
	switch (mirror_flag) {
	default:
	case 1:  /* equator */
	    for(i=0; i<nx1; i++) {
		int ii  = i * np1;
		int ii2 = (nx1-1-i)*np1;
		for(j=0; j<nph1[i]; j++) {
			  f2[ii2+j] = f1[ii+j];
		}
	    }
	    break;
	case 2:  /* meridian */
	    for(i=0; i<nx1; i++) {
		int ii  = i * np1;
		for(j=0; j<nph1[i]; j++) {
			  int jj2 = nph1[i] - 1 - j;
			  f2[ii+jj2] = f1[ii+j];
		}
	    }
	    break;
	case 3:  /* center */
	    for(i=0; i<nx1; i++) {
		int ii  = i * np1;
		int ii2 = (nx1-1-i)*np1;
		for(j=0; j<nph1[i]; j++) {
			  int jj2 = nph1[i] - 1 - j;
			  f2[ii2+jj2] = f1[ii+j];
		}
	    }
	    break;
	case 4:  /* center + PI-shift */
	    for(i=0; i<nx1; i++) {
		int ii  = i * np1;
		int ii2 = (nx1-1-i)*np1;
		for(j=0; j<nph1[i]; j++) {
			  int nn = (nph1[i]+1)/2;
			  int jj2 = j < nn ? nn + j : j - nn;
			  f2[ii2+jj2] = f1[ii+j];
		}
	    }
	    break;
	case 5: /* duplicate North at South */
	    {   int nx_2 = (nx1+1)/2;
		for(i=0; i<nx_2; i++) {
		    int ii  = i * np1;
		    int ii2 = (nx1-1-i)*np1;
		    for(j=0; j<nph1[i]; j++) {
			      f2[ii+j]  = f1[ii+j];
			      f2[ii2+j] = f1[ii+j];
		    }
		}
	    }
	    break;
	case 6: /* duplicate South at North */
	    {   int nx_2 = (nx1+1)/2;
		for(i=0; i<nx_2; i++) {
		    int ii  = i * np1;
		    int ii2 = (nx1-1-i)*np1;
		    for(j=0; j<nph1[i]; j++) {
			      f2[ii+j]  = f1[ii2+j];
			      f2[ii2+j] = f1[ii2+j];
		    }
		}
	    }
	    break;
	case 7: /* duplicate North at South with changing a sign */
	    {   int nx_2 = (nx1+1)/2;
		for(i=0; i<nx_2; i++) {
		    int ii  = i * np1;
		    int ii2 = (nx1-1-i)*np1;
		    for(j=0; j<nph1[i]; j++) {
			      f2[ii+j]  =  f1[ii+j];
			      f2[ii2+j] = -f1[ii+j];
		    }
		}
	    }
	    break;
	case 8: /* duplicate South at North with changing a sign */
	    {   int nx_2 = (nx1+1)/2;
		for(i=0; i<nx_2; i++) {
		    int ii  = i * np1;
		    int ii2 = (nx1-1-i)*np1;
		    for(j=0; j<nph1[i]; j++) {
			      f2[ii+j]  = -f1[ii2+j];
			      f2[ii2+j] =  f1[ii2+j];
		    }
		}
	    }
	    break;
	}

	/* free memory */
	if (f1 != (double *)0)  free((void *)f1);

	f1 = f2;
    } else if (rotate_flag) {
	if (map_type == 't' || map_type == 'T')
	    read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	else if (map_type == 'u')
	    read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	else
	    read_glesp_qmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	f2 = (double *)calloc((unsigned)(nx1*np1), sizeof *f2);
	if (f2 == (double *)0) {
		PR(ER, "%s: no memory for second array...\n", myname);
		exit(1);
	}

	rotate_map(f1, x1, nx1, np1, nph1, f2, dtheta, dphi);

	/* free memory */
	if (f1 != (double *)0)  free((void *)f1);

	f1 = f2;
    } else if (t_cos_th_flag) { /* flag to calc T(th) vs cos(th)    */
	double *tt = (double *)0;

	/* read map */
	if (map_type == 't' || map_type == 'T')
	    read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	else if (map_type == 'u')
	    read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	else
	    read_glesp_qmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	if ((tt=(double *)calloc((unsigned)nx1, sizeof *tt)) == (double *)0) {
	    PR(ER, "%s: no memory for T(theta) array...\n", myname);
	    exit(1);
	}

	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) tt[i] += f1[ii+j];
	    tt[i] /= nph1[i];
	    printf ("%12.5e %12.5e\n", x1[i], tt[i]);
	}
	free((void *)tt);
	exit(0);
    } else {
	double *fmask = (double *)0, *gmask = (double *)0;
	if (map_type == 't' || map_type == 'T')
	    read_glesp_tmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	else if (map_type == 'u')
	    read_glesp_umap(&f1, &x1, &nx1, &np1, &nph1, map1_file);
	else
	    read_glesp_qmap(&f1, &x1, &nx1, &np1, &nph1, map1_file);

	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		 f1[ii+j] = cf * f1[ii+j] + sc;
	    }
	}
	if (zmask_flag) {
	    if ((zmask=(double*)calloc((unsigned)(nx1*np1), sizeof *zmask))==
	      (double *)0) {
		PR(ER, "%s: no memory for mask array... No mask.\n", myname);
		zmask_flag = 0;
	    }

	    /* set mask from HPix-file */
	    if (*hp_mask_file != '\0') {
	       extern
		 void mask_hp(double*,double*,int,int,int*,char*,double);
	       i = nx1;
	       if((fmask=(double*)calloc((unsigned)(nx1*np1),sizeof *fmask))
		 == (double *)0) {
		   PR(ER, "%s: no memory for Hpix-mask array... No mask.\n",
			myname);
		   *hp_mask_file = '\0';
	       } else
		   mask_hp(fmask,x1,i,np1,nph1,hp_mask_file,zmask_value);
	    }
	    if (*gl_mask_file != '\0') {
	       extern
		 void mask_gl(double*,double*,int,int,int*,char*,double);
	       i = nx1;
	       if((gmask=(double*)calloc((unsigned)(nx1*np1),sizeof *gmask))
		 == (double *)0) {
		   PR(ER, "%s: no memory for file-mask array... No mask.\n",
			myname);
		   *gl_mask_file = '\0';
	       } else
		    mask_gl(gmask,x1,i,np1,nph1,gl_mask_file,zmask_value);
	    }
	}
	/* read region for masking */
	/*-------------------------*/
	if ( zmask_flag ) {
	    extern void map_zone(struct _zone_ *, int, int,
		   double *, double *, int, int, int *, double);
	    map_zone(zz,zns,function_ztype,zmask,x1,nx1,np1,nph1,zmask_value);
	    for(i=0; i<nx1; i++) {
	      int ii = i * np1;
	      for(j=0; j<nph1[i]; j++) {
		  int ij=ii+j;
		  double zm = zmask[ij] - zmask_value;
		  if (*hp_mask_file != '\0') {
			double fm = fmask[ii+j] - zmask_value;

			fm = ABS(fm);
			/* 1. remove data under mask */
			/* 2. remove data out of mask */
			if ( !((!f_mask_flag && fm <= 1.e-6 ) ||
			      (f_mask_flag && fm >= 1.e-6)))
				f1[ij] = dummy;
		  }
		  if (*gl_mask_file != '\0') {
			double fm = gmask[ii+j] - zmask_value;

			fm = ABS(fm);
			/* 1. remove data under mask */
			/* 2. remove data out of mask */
			if ( !((!f_mask_flag && fm <= 1.e-6 ) ||
			      (f_mask_flag && fm >= 1.e-6)))
				f1[ij] = dummy;
		  }
		  if ( (z_remove_flag && ABS(zm) <  1.e-6) ||
		       (z_remain_flag && ABS(zm) >= 1.e-6) ) {
			f1[ij] = dummy;
		  }
	      }
	    }
	}
	if (fmask != (double *)0) free ((void *)fmask);
	if (gmask != (double *)0) free ((void *)gmask);
    }
    if (abs_flag) {
	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		 if (f1[ii+j] < 0.) f1[ii+j] = -f1[ii+j];
		 if (map_type == 3) {
		    if (q1[ii+j] < 0.) q1[ii+j] = -q1[ii+j];
		    if (u1[ii+j] < 0.) u1[ii+j] = -u1[ii+j];
		 }
	    }
	}
    }
    if (inverse_flag) {
	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		 if (fabs(f1[ii+j]) > 1.e-30)  f1[ii+j] = 1. / f1[ii+j];
		 if (map_type == 3) {
		    if (fabs(q1[ii+j]) > 1.e-30) q1[ii+j] = 1. / q1[ii+j];
		    if (fabs(u1[ii+j]) > 1.e-30) u1[ii+j] = 1. / u1[ii+j];
		 }
	    }
	}
    }

    if (*outfile == '\0' || strcmp(outfile,"-") == 0
       || strcmp(outfile,"stdout") == 0) {
	   fp_out = stdout;
    } else {
	    /* open output file */
	    if((fp_out = fopen(outfile, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, outfile);
		exit(1);
	    }
    }

    /*  write map  */
    if (ffile_flag) {
	/* map output in F-format */
	ff_write_ext2_map(f1, x1, nx1,np1,nph1, fp_out, x_type, grid_type);
    } else if (print_flag) {
	char format[51];
	sprintf(format, "%s  %s  %s\n", format_out, format_out, format_out);
	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		double theta, phi;
		int ij=ii+j;
		double zm = zmask[ij] - zmask_value;
		if (zmask_flag) {
		  if(!((z_remove_flag && ABS(zm) >= 1.e-6) ||
		       (z_remain_flag && ABS(zm) <  1.e-6) )) {
			continue;
		  }
		}
		pixel2angle(x1, nx1, nph1, i, j, &theta, &phi);
		PR(fp_out, format, theta, phi, f1[ii+j]);
	    }
	}
    } else {
	/* map output in FITS-format */
	if (map_type == 3)
	    outp_glesp_3map(f1, q1, u1, x1, nx1, np1, nph1, fp_out,
	       x_type, grid_type);
	else
	    outp_glesp_map(f1, x1, nx1, np1, nph1, fp_out, x_type, grid_type);
    }

    fclose(fp_out);

    /* free memory */
    if (f1 != (double *)0)  free((void *)f1);
    if (map_type == 3)  {
	if (q1 != (double *)0)  free((void *)q1);
	if (u1 != (double *)0)  free((void *)u1);
    }
    if (robust_average) {
	for(i=1; i<nf_m; i++)  free((void *)mf[i]);
    }
    /* free memory from zone-array */
    if (zz != (struct _zone_ *)0)  free((void *)zz);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
	    switch(argv[i][1]) {
	    case 'a':
		    if (strncmp(argv[i]+1, "ab", 2) == 0) {
			abs_flag = 1;
		    } else if (strncmp(argv[i]+1, "apm", 3) == 0) {
			apply_mask = 1;
			mask_flag = 1; dif_mask = 7;
		    } else
			simple_average = 1;
		    break;
	    case 'c':
		    switch(argv[i][2]) {
		    case '1':
			    c1 = atof(argv[++i]);
			    break;
		    case '2':
			    c2 = atof(argv[++i]);
			    break;
		    case 'd':
			    cf = 1./atof(argv[++i]);
			    break;
		    case 'c': /* concentric figure */
		      { char vstr[30];
		      switch(argv[i][3]) {
		      case 's':
			 break;
		      case 'r':  /* rectangle */
			 ++i;
			 skip_wd(0, argv[i], vstr);
			 lat_c = deg2rad(vstr);
			 skip_wd(1, argv[i], vstr);
			 lon_c = deg2rad(vstr);
			 skip_wd(2, argv[i], vstr);
			 lat1_c = deg2rad(vstr);
			 skip_wd(3, argv[i], vstr);
			 lon1_c = deg2rad(vstr);
			 skip_wd(4, argv[i], vstr);
			 w_c = deg2rad(vstr); /* radius of ring width */
			 flag_conc_fig = 2;
			 break;
		      default :  /* circle */
			 ++i;
			 skip_wd(0, argv[i], vstr);
			 lat_c = deg2rad(vstr);
			 skip_wd(1, argv[i], vstr);
			 lon_c = deg2rad(vstr);
			 skip_wd(2, argv[i], vstr);
			 r_c = deg2rad(vstr);/* radius of inner circle */
			 skip_wd(3, argv[i], vstr);
			 w_c = deg2rad(vstr); /* radius of ring width */
			 flag_conc_fig = 1;
			 break;
		      }}
		      z_remain_flag = 1;
		      break;
		    case 'u':
			    dif_mask = 5;
			    mask_flag = 1;
			    cut_level = atof(argv[++i]);
			    break;
		    case 's': /* scalar sum */
			    sc = atof(argv[++i]);
			    break;
		    case 'm':
		    case 'f':
		    default :
			    cf = atof(argv[++i]);
			    break;
		    case 'o': /* correlation */
			    corr_flag = 1;
			    break;
		    case 'w': /* correlation window: side of new pixel */
			    corr_win = atof(argv[++i]);
			    break;
		    case 'r': /* correlation running window */
			    corr_rwin = atof(argv[++i]);
			    break;
		    }
		    break;
	    case 'C':
		    if (argv[i][2] == '7' || argv[i][3] == '7')
			stat_flag = 7;
		    else
			stat_flag = 6;
		    strcpy(file_stat_interv, argv[++i]);
		    break;
	    case 'd':
		    switch(argv[i][2]) {
		    case 'i':
			    if ( strncmp(argv[i]+1, "dip", 3) == 0 ) {
				dipole_fitting = 1;
			    } else {
				/* division */
				action = 3;
			    }
			    break;
		    case 'o':
			    if (strcmp(argv[i]+1, "double") == 0) {
				    bitpix_map = DOUBLEPIX;
			    } else {
				    PR(ER,"%s: illegal flag '%s'\n",
					myname, argv[i]);
				    usage();
			    }
			    break;
		    case 'p':
			    dphi = deg2rad(argv[++i]);
			    rotate_flag = 1;
			    if (rot_q < rot_max) {
				    rotate_queue[rot_q].cr = 'p';
				    rotate_queue[rot_q].ang = dphi;
				    rot_q++;
			    }
			    break;
		    case 't':
			    dtheta = -deg2rad(argv[++i]);
			    rotate_flag = 1;
			    if (rot_q < rot_max) {
				    rotate_queue[rot_q].cr = 't';
				    rotate_queue[rot_q].ang = dtheta;
				    rot_q++;
			    }
			    break;
		    case 'u':
			    dummy = atof(argv[++i]);
			    break;
		    case '2':
			    dif_mask = 2;
			    break;
		    case '3':
			    dif_mask = 3;
			    break;
		    case '4':
			    dif_mask = 4;
			    break;
		    default :
		    case '0':
			    dif_mask = 0;
			    break;
		    }
		    break;
	    case 'e':
		    if (strncmp(argv[i]+1, "eq", 2) == 0) {
		       /* transform from galactic coors to equatorial */
		       convert_coors='q';
		    } else if (strncmp(argv[i]+1, "ec", 2) == 0) {
		       /* transform from galactic coors to ecliptic */
		       convert_coors='e';
		    } else if (strncmp(argv[i]+1, "est", 3) == 0) {
		       /* select type of mapping estimator */
		       int est1=0, est2=0, est3=0, est4=0;
		       char *st = argv[i]+4;
		       /* -est[{CDEFPMG}{NS}o-] - calc. mapping estimator */
		       while (*st) {
			    switch(*st) {
			    case 'B': /* sign of rotation angle */
			    case 'b':
				 est3 += 200; break;
			    case 'C': /* Cl-sigma estimator   */
			    case 'c':
				 est1 = 1; break;
			    case 'D': /* Dl-sigma estimator   */
			    case 'd':
				 est1 = 2; break;
			    case 'E': /* Cl-sigma difference estimator */
			    case 'e':
				 est1 = 3; break;
			    case 'F': /* Dl-sigma difference estimator */
			    case 'f':
				 est1 = 4; break;
			    case 'P': /* parity estimator      */
			    case 'p':
				 est1 = 5; break;
			    case 'M': /* (MAX-MIN)_north / (MAX-MIN)_south */
			    case 'm':
				 est1 = 6; break;
			    case 'G': /* Sigma_north / Sigma_south */
			    case 'g':
				 est1 = 7; break;
			    case 'N':  /* copy Northen hemisphere to South */
			    case 'n':
				 est2 = 10; break;
			    case 'S':  /* copy Southen hemisphere to North */
			    case 's':
				 est2 = 20; break;
			    case 'O':  /* select rotation (theta, phi) */
			    case 'o':
				 est4 = 1000; break;
			    case '-':  /* change sign when reflecion */
				       /* N->S or S->N */
				 est3 += 100; break;
			    default:
				 PR(ER,
				  "%s: unknown flag '%c' in argument '%s'\n",
				    myname, *st, argv[i]);
				 exit(1);
			    }
			    st++;
		       }
		       estimator_type = est1 + est2 + est3 + est4;
		    } else if (strncmp(argv[i]+1, "esl", 3) == 0) {
		       esl_lmax_stat = atoi(argv[++i]);
		    } else if (strncmp(argv[i]+1, "es", 2) == 0) {
		       ex_stat = 1;
		    } else if (strncmp(argv[i]+1, "ex", 2) == 0) {
		       strcpy(out_extr_file, argv[++i]);
		       ex_stat = 1;
		       stat_flag = 'e';
		    } else if (strncmp(argv[i]+1, "er", 2) == 0) {
		       dp_ex = deg2rad(argv[++i]);
		       ex_stat = 1;
		    } else {
		       eps = atof(argv[++i]);
		    }
		    break;
	    case 'f':
		    if (strncmp(argv[i]+1, "flo", 3) == 0) {
			    bitpix_map = FLOATPIX;
		    } else if (strncmp(argv[i]+1, "fi", 2) == 0) {
			    strcpy(fit_map_patt_file, argv[++i]);;
		    } else {
			    PR(ER,"%s: illegal flag '%s'\n",
				myname, argv[i]);
			    usage();
		    }
		    break;
	    case 'g':
		    if (strncmp(argv[i]+1, "gp", 2) == 0) {
			grad_flag = 1;
		    } else if (strncmp(argv[i]+1, "gt", 2) == 0) {
			grad_flag = 2;
		    } else if (strncmp(argv[i]+1, "g2e", 3) == 0) {
			convert_coors='e';
		    } else if (strncmp(argv[i]+1, "g2q", 3) == 0) {
			convert_coors='q';
		    }
		    break;
	    case 'h':
		    if (strncmp(argv[i]+1, "hf", 2) == 0) {
			    /* output file for ASCII histogram data */
			    strcpy(hg_file, argv[++i]);
		    } else if (strncmp(argv[i]+1, "hn", 2) == 0) {
			    /* number of bins for histogram */
			    n_hg = atoi(argv[++i]);
		    } else if (strncmp(argv[i]+1, "h0", 2) == 0) {
			    /* start value for histogram */
			    s_hg0 = atof(argv[++i]);
		    } else if (strncmp(argv[i]+1, "he", 2) == 0) {
			    /* start value for histogram */
			    s_hgF = atof(argv[++i]);
		    } else if (strncmp("h2", argv[i]+1, 2) == 0)
			usage2();
		    else
			usage();
		    break;
	    case 'i':
		    if (strncmp("inv", argv[i]+1, 3) == 0) {
			inverse_flag = 1;
		    } else {
			PR(ER, "%s: illegal flag '%s'\n",  myname, argv[i]);
			usage();
		    }
		    break;
	    case 'k':
		    if (strncmp("ke", argv[i]+1, 2) == 0) {
			char s1[30], s2[30], s3[30], s4[30];
			get_4_str(argv[++i], s1, s2, s3, s4);
			if(*s1 !='\0')
			  zone_latR0=deg2rad(s1);else zone_latR0= -1001.;
			if(*s2 !='\0')
			  zone_lonR0=deg2rad(s2);else zone_lonR0= -1001.;
			if(*s3 !='\0')
			  zone_latR1=deg2rad(s3);else zone_latR1= -1001.;
			if(*s4 !='\0')
			zone_lonR1=deg2rad(s4);else zone_lonR1= -1001.;
			z_remain_flag = 1;
			z_remove_flag = 0;
			zmask_flag = 1;
		    } else if (strcmp(argv[i]+1, "ksz") == 0 ||
			       strcmp(argv[i]+1, "kzs") == 0)  {
			 char vstr[30];
			 flag_kZS = 1;
			 /* and concentric circle */
			 ++i;
			 skip_wd(0, argv[i], vstr);
			 lat_c = deg2rad(vstr);
			 skip_wd(1, argv[i], vstr);
			 lon_c = deg2rad(vstr);
			 skip_wd(2, argv[i], vstr);
			 r_c = deg2rad(vstr);/* radius of inner circle */
			 skip_wd(3, argv[i], vstr);
			 w_c = deg2rad(vstr); /* radius of ring width */
			 flag_conc_fig = 1;
		    } else {
			PR(ER, "%s: illegal flag '%s'\n", myname, argv[i]);
			usage();
		    }
		    break;
	    case 'K':  /* keep data in selected zones */
		    z_remain_flag = 1;
		    z_remove_flag = 0;
		    break;
	    case 'l':
		    level = atof(argv[++i]);
		    level_flag = 1;
		    break;
	    case 'm':
		    if (strncmp(argv[i]+1, "mas", 3) == 0) {
			   mask_flag = 1;
		    } else if(strncmp("mh", argv[i]+1, 2) == 0) {
			   strcpy(hp_mask_file, argv[++i]);
			   zmask_flag = 1;
			   if (strncmp("mh-", argv[i-1]+1, 3) == 0)
				/* remove hp_mask_zone:1, remain: 0 */
				f_mask_flag = 0;
		    } else if(strncmp("mg", argv[i]+1, 2) == 0) {
			   strcpy(gl_mask_file, argv[++i]);
			   zmask_flag = 1;
			   if (strncmp("mg-", argv[i-1]+1, 3) == 0)
				/* remove gl_mask_zone:1, remain: 0 */
				f_mask_flag = 0;
		    } else if (strncmp(argv[i]+1, "mf", 2) == 0) {
			   mask_file_flag = 1;
		    } else if (strncmp(argv[i]+1, "mir", 3) == 0) {
			    switch(argv[i][4]) {
			    case 'H':
			    case 'h':
			    case '\0':
			    default :
				mirror_flag = 1; /* reflection via equator */
				break;
			    case 'V':
			    case 'v':
			       mirror_flag = 2; /* reflection via meridian */
			       break;
			    case 'C':
			    case 'c':
			       mirror_flag = 3; /* reflection via center */
			       break;
			    case 'P':
			    case 'p':
			       /* reflection via equator wit pi-shift */
			       mirror_flag = 4;
			       break;
			    case 'N':
			    case 'n':
			       if (argv[i][5] == '-')
				 /* reflection -North via equator */
				 mirror_flag = 7;
			       else
				 /* reflection North via equator */
				 mirror_flag = 5;
			       break;
			    case 'S':
			    case 's':
			       if (argv[i][5] == '-')
				 /* reflection -South via equator */
				 mirror_flag = 8;
			       else
				 /* reflection South via equator */
				 mirror_flag = 6;
			       break;
			    }
		    } else if (strncmp(argv[i]+1, "mink", 4) == 0) {
			   mink_func = 1;
		    } else if (strncmp(argv[i]+1, "mi", 2) == 0) {
			   min_flag = 1;
		    } else if (strncmp(argv[i]+1, "ma", 2) == 0) {
			   max_flag = 1;
		    } else if (strncmp(argv[i]+1, "mul", 2) == 0) {
		       action = 2;
		    } else {
		       PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
		       usage();
		    }
		    break;
	    case 'n':
		    if (strncmp(argv[i]+1, "noc", 3) == 0) {
		       glesp_comments = 0;
		    } else {
		       PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
		       usage();
		    }
		    break;
	    case 'o':
		    if (strcmp("oq", argv[i]+1) == 0) {
			strcpy(q_out_file, argv[++i]);
			if (polariz_flag < 4) /* map of PA not in degs */
			    polariz_flag = 3;
		    } else if (strcmp("ou", argv[i]+1) == 0) {
			strcpy(u_out_file, argv[++i]);
			if (polariz_flag < 4) /* map of PA not in degs */
			    polariz_flag = 3;
		    } else if(strncmp("of", argv[i]+1, 2) == 0) {
			ffile_flag = 1;
			strcpy(outfile, argv[++i]);
		    } else {
			strcpy(outfile, argv[++i]);
		    }
		    break;
	    case 'p':
		    if (strncmp(argv[i]+1, "pg", 2) == 0) {
			grad_flag = 3;
			strcpy(outp_angle_file, argv[++i]);
		    } else if (strncmp(argv[i]+1, "pr", 2) == 0) {
			print_flag = 1;
		    } else if (strncmp(argv[i]+1, "pa", 2) == 0) {
			if (polariz_flag < 3) {
			    if (argv[i][3] == 'd')
				polariz_flag = 2;  /* values of map in degs */
			    else
				polariz_flag = 1;  /* values of map in rads */
				strcpy(outp_angle_file, argv[++i]);
			} else {
			    if (argv[i][3] == 'd')
				polariz_flag = 4;  /* values of map in degs */
			    strcpy(map2_file, argv[++i]);
			    map_type='p'; /* type of map: polarization pars */
			}
		    } else if (strncmp(argv[i]+1, "pp", 2) == 0) {
			if (polariz_flag < 4) {
				polariz_flag = 3;  /* values of map in degs */
			}
			map_type = 'P'; /* type of map: polarization pars */
			strcpy(map1_file, argv[++i]);
		    }
		    break;
	    case 'q':
		    strcpy(map1_file, argv[++i]);
		    polariz_flag = 1;
		    break;
	    case 'Q':
		    map_type = 'q'; /* type of map: T, Q or U-map */
		    break;
	    case 'r':
		    if (strcmp("rm", argv[i]+1) == 0 ||
			strncmp("re", argv[i]+1, 2) == 0) {
			 char s1[30], s2[30], s3[30], s4[30];
			 get_4_str(argv[++i], s1, s2, s3, s4);
			 if(*s1 !='\0')
			   zone_latR0=deg2rad(s1);else zone_latR0= -1001.;
			 if(*s2 !='\0')
			   zone_lonR0=deg2rad(s2);else zone_lonR0= -1001.;
			 if(*s3 !='\0')
			   zone_latR1=deg2rad(s3);else zone_latR1= -1001.;
			 if(*s4 !='\0')
			   zone_lonR1=deg2rad(s4);else zone_lonR1= -1001.;
			 z_remove_flag = 1;
			 z_remain_flag = 0;
			 zmask_flag = 1;
		    } else if (strncmp ("ra", argv[i]+1, 2) == 0) {
			 zone_radius = deg2rad(argv[++i]);
			 zone_circle = 1;
		    } else if (strncmp ("rwn", argv[i]+1, 3) == 0) {
			 repix_win = -atof(argv[++i]);
			 repix_flag = 1;
		    } else if (strncmp ("rw", argv[i]+1, 2) == 0) {
			 repix_win = deg2rad(argv[++i]);
			 repix_flag = 1;
		    } else {
			robust_average = 1;
		    }
		    break;
	    case 'R':  /* keep data outside selected zones */
		    z_remain_flag = 0;
		    z_remove_flag = 1;
		    break;
	    case 's':
		    if (strncmp(argv[i]+1, "su", 2) == 0)
			   action = 1;
		    else if (strncmp(argv[i]+1, "sta", 3) == 0) {
			   do_stacking = 1;
			   if (argv[i][strlen(argv[i])-1] == 'q')
				do_stacking = 2;
			   strcpy(file_stack, argv[++i]);
		    } else if (strncmp(argv[i]+1, "st", 2) == 0) {
			   if (strncmp(argv[i]+1, "str", 3) == 0)
				stat_flag = 2;
			    else if (strncmp(argv[i]+1, "stover",3)==0) {
				stat_flag = 3;
				level_low = atof(argv[++i]);
			    } else if(strncmp(argv[i]+1,"stunder",3)==0){
				/* under the level */
				stat_flag = 4;
				level_up = atof(argv[++i]);
			    } else
				stat_flag = 1;
		    } else  if (strncmp(argv[i]+1, "sc", 2) == 0) {
			    corr_flag = 2;
		    } else  if (strncmp(argv[i]+1, "sw", 2) == 0) {
			    if (strncmp(argv[i]+1, "swr", 3) == 0)
				sig_win_flag = 2; /* running window */
			    else
				sig_win_flag = 1; /* pixel of sigma */
			    sig_win = atof(argv[++i]);
		    } else {
		       PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
		       usage();
		    }
		    break;
	    case 't':
		    if (strcmp(argv[i]+1, "tt") == 0)  {
		       t_cos_th_flag = 1;
		    } else {
		       PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
		       usage();
		    }
		    break;
	    case 'T':
		    map_type = 'T'; /* type of map: T, Q or U-map */
		    break;
	    case 'u':
		    strcpy(map2_file, argv[++i]);
		    polariz_flag = 1;
		    break;
	    case 'U':
		    map_type = 'u'; /* type of map: T, Q or U-map */
		    break;
	    case 'v':
		    printf("%s: version 1.10 (10/01/2016), GLESP 2.0\n",
			myname);
		    exit(0);
	    case 'w':
		    if (strncmp(argv[i]+1, "wc", 2) == 0) {
			    /* correlation window */
			    corr_win = atof(argv[++i]);
		    } else  if (strncmp(argv[i]+1, "ws", 2) == 0) {
			    if (strncmp(argv[i]+1, "wsr", 3) == 0)
				sig_win_flag = 2;
			    else
				sig_win_flag = 1;
			    sig_win = atof(argv[++i]);
		    } else {
		       PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
		       usage();
		    }
	    case 'z':
		switch (argv[i][2]) {
		case 'c':
		    {  char s1[30], s2[30];
		      /* center of a circle to be masked */
		      get_2_str(argv[++i], s1, s2);
		      if(*s1 !='\0')
			zone_latR0=deg2rad(s1); else zone_latR0= -1001.;
		      if(*s2 !='\0')
			zone_lonR0=deg2rad(s2); else zone_lonR0= -1001.;
		      zone_circle = 1;
		    }
		    break;
		case 'e':
		    coor_type='e';
		    break;
		case 'n':
		    if (strncmp(argv[i], "-zna", 4) == 0)
			norm_flag=2;
		    else
			norm_flag=1;
		    break;
		case 'r':
		    /* radius of a circle to be masked */
		    zone_radius = deg2rad(argv[++i]);
		    zone_circle = 1;
		    break;
		case 'o':
		    /* select a zone to be masked */
		    { char s1[30], s2[30], s3[30], s4[30];
			get_4_str(argv[++i], s1, s2, s3, s4);
			if(*s1 !='\0') zone_latR0=deg2rad(s1);
			    else zone_latR0= -1001.;
			if(*s2 !='\0') zone_lonR0=deg2rad(s2);
			    else zone_lonR0= -1001.;
			if(*s3 !='\0') zone_latR1=deg2rad(s3);
			    else zone_latR1= -1001.;
			if(*s4 !='\0') zone_lonR1=deg2rad(s4);
			    else zone_lonR1= -1001.;
		    }
		    break;
		default:
		    /* file with list of zones to be processed */
		    strcpy(zone_file, argv[++i]);
		    break;
		}
		zmask_flag = 1;
		break;
	    default:
		    break;
	    }
	    break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		    if (strncmp("h2", argv[i]+1, 2) == 0)
			usage2();
		    else
			usage();
		    break;
		default:
			usage();
			break;
		}
		break;
	default:
		if (*map1_file == '\0')      strcpy(map1_file, argv[i]);
		else if (*map2_file == '\0') strcpy(map2_file, argv[i]);
		if(nf_m == NFF)  {
		    static int q=0;
		    if (!q) PR(ER, "only %d files...\n", NFF);
		    q=1;
		    break;
		}
		ind_nf[nf_m] = i;  nf_m++;
		break;
	}
    }

    /* if no zone is set, and '-K' or '-R' flags are set than nothing to do */
    if ( (zone_latR0 < -999 && zone_lonR0 < -999 &&
	 zone_latR1 < -999 && zone_lonR1 < -999 && mask_file_flag == 0 &&
	 (*hp_mask_file != '\0' || *gl_mask_file != '\0')) &&
       !(corr_win > 0. &&
//            ((zone_latR0 > -999 || zone_lonR0 > -999) ||
	    (*hp_mask_file != '\0' || *gl_mask_file != '\0'))
    ) {
	    if( z_remain_flag || z_remove_flag ) {
		z_remain_flag =  z_remove_flag = 0;
		zmask_flag = 0;
	    }
    }
    return;
}

void usage()
{
PR(ER, "usage: %s file1 file2 -o output_file\n", myname);
PR(ER, "[-c1 c1] [-c2 c2] [-c{f,d} cf] [-cs scalar] [-{sum,mul,div}] [-abs]\n");
PR(ER, "[-r file file2 file3...] [-st{r}(stat)] [-inv(iversion)] [-nocom]\n");
PR(ER, "[-hn n_hg] [-hf hg_file] [-h0 start_his.value] [-he stop_his.val.]\n");
PR(ER, "[-cor] [-lev level [-e eps](calc.N_pix)] [-eq(set eq.cr.)]\n");
PR(ER, "[-dt rotate_theta_angle] [-dp rotate_phi_angle] [-{double,float}]\n");
PR(ER, "[-mg file_with_mask] [-mh[-] hp_file_with_mask(+/-_in/out)]\n");
PR(ER, "[-q file -u file2 [-pa[d] pol_ang_file,rad[deg]]] [-g2e] [-g2q]\n");
PR(ER, "[-cut cut_level(cut 1st file by 2nd)] [-{Q,U}] [-st{over/under} level]\n");
PR(ER, "[-g{pt}(gradient_file{phi,theta})] [-pg pos_angle(for grad)] [-pr]\n");
PR(ER, "[-cw correlation_window(arcmin)] [-mir[{HVCPNS}](mirroring)] [-sc(aut_cor_flag)]\n");
PR(ER, "[-zf file_with_zones] [-zo lat1,lon1,lat2,lon2(mask sky zone to remove)]\n");
PR(ER, "[-zc lat,lon] [-zr radius_in_degs] [-zeq] [-{K,R}] [-rw[n] repix_win(arcmin)]\n");
PR(ER, "[-{remove,keep} lat1,lon1,lat2,lon2(mask sky zone to remove/keep)]\n");
PR(ER, "[-zn[a](norm)] [-mask] [-e eps] [-d{0123}(masks)] [-Cf lim_file]\n");
PR(ER, "[-cr cor_run_win] [-cc lat,lon,rad,w(ring mask in degs)]\n");
PR(ER, "[-ccr lat1,lon1,lat2,lon2,w(rectang.mask in degs)\n");
PR(ER, "[-es(calc. extrem)] [-ex extr_out] [-er extr.cir_radius]\n");
PR(ER, "[-mink(calc.Minkowski func.)] [-dipfit]\n");
exit(2);
}

void usage2()
{
PR(ST, "usage: %s file1 [file2] -o output_file [FLAGS]\n", myname);
PR(ST, "-----------------------------------------------------------------\n");
PR(ST, "[-cut cut_level(cut 1st file by 2nd)]\n");
PR(ST, "[-g{pt}(gradient_file{phi,theta})]\n");
PR(ST, "[-pg pos_angle(for grad)]\n");
PR(ST, "[-sc(aut_cor_flag)]\n");
PR(ST, "[-mask] [-e eps]\n");
PR(ST, "[-d{0123}(masks)]\n");
PR(ST, "[-Cf lim_file]\n");

PR(ST, "-abs - calculates absolute value: ABS(map.fts)\n");
PR(ST, "-apm - applys mask. Default value for mask is dummy (dflt=0)\n");
PR(ST, "-c1 k1  -c2 k2 - calc expression for files: k1*map1.fts + k2*map2.fts\n");
PR(ST, "-cc   lat,lon,rad,w - set ring mask for sky zone to be kept:\n");
PR(ST, "      in center (lat,lon) of radius 'rad' and width 'w' in degs.\n");
PR(ST, "-ccr  lat1,lon1,lat2,lon2,w - set rectang. mask for sky zone to be kept\n");
PR(ST, "      in rectangle (lat1,lon1,lat2,lon2) of width 'w' in degs.\n");
PR(ST, "-cf value - calc expression: value * map.fts (see '-c1', '-cs')\n");
PR(ST, "-cor - calculates correlation: map1.fts <*> map2.fts\n");
PR(ST, "-cr cor_win - calc. running correlation in circle of diameter, arcmin\n");
PR(ST, "-cs value - calc expression: value + map.fts (see '-c1', '-cf')\n");
PR(ST, "-cw cor_win - calculates correlation in pixels of size 'cor_win', arcmin\n");
PR(ST, "-div - calculates divison: map1.fts / map2.fts\n");
PR(ST, "-dipfit - fits dipole to map\n");
PR(ST, "-double - output file in 64-bit float format (default) (see '-float')\n");
PR(ST, "-dt theta_angle - rotates map in theta-direction for 'theta_angle', rad\n");
PR(ST, "-dp phi_angle - rotates map in phi-direction for 'phi_angle', rad\n");
PR(ST, "-du dummy - sets value for mask\n");
PR(ST, "-e eps - sets interval of level+/-eps to calc statistics (see '-st','-lev') \n");
PR(ST, "-es - caculates number of extrema (see '-ex')\n");
PR(ST, "-est[{CDEFPMG}{NS}bo-] - calculates mapping estimator:\n");
PR(ST, "   C|D|E|F|M|G|P - Cl|Dl|Cl-dif.sig||Dl-dif.sig|MiniMax|sigN/sigS|parity,\n");
PR(ST, "    N|S - with North|South reflection, '-' - change sign for N|S:odd/[even]\n");
PR(ST, "   'o' - rotation order: phi,theta|theta,phi, 'b'-back rotation\n");
PR(ST, "-esl lmax_stat - lmax for tempora arrats in '-est'-opertation (def:50)\n");
PR(ST, "-ex ex_file - caculates number of extrema and write them to file\n");
PR(ST, "-er radius - caculates number of extrema within radius, in degs (default:2pix)\n");
PR(ST, "-eq - converts map in galactic coordinates to equatorial\n");
PR(ST, "-fit pattern - fits pattern-map from file 'pattern' to the input map\n");
PR(ST, "-float - outputs file in 32-bit float format (see '-double')\n");
PR(ST, "-g2q - converts map in galactic coordinates to equatorial\n");
PR(ST, "-g2e - converts map in galactic coordinates to ecliptic\n");
PR(ST, "-h0 v1 - calc histogram from value v1 (see:'-st','-he','-hf','-hn')\n");
PR(ST, "-he v2 - calc histogram upto value v2 (see:'-st','-h0','-hf','-hn')\n");
PR(ST, "-hf hist.dat - outputs filename for histogram data (see:'-st','-h0','-he','-hn')\n");
PR(ST, "-hn num - sets number of histogram bins 'num' (see:'-st','-h0','-he','-hf')\n");
PR(ST, "-kzs lat,lon,rad,w - calc. stat. in circle using ring with flag '-cc'\n");
PR(ST, "-keep lat1,lon1,lat2,lon2 - set mask for sky zone to be kept\n");
PR(ST, "-inv - makes iversion of a map\n");
PR(ST, "-K - flag to keep selected zone on the figure (see '-zo','-zc','-zf')\n");
PR(ST, "-lev level - sets interval of level+/-eps to calc statistics (see '-st','-e')\n");
PR(ST, "-mf  gl_mask.fts - sets mask from glesp-file 'gl_mask.fts'(remove inside) \n");
PR(ST, "-mg  gl_mask.fts - sets mask from glesp-file 'gl_mask.fts'(remove inside) \n");
PR(ST, "-mg- gl_mask.fts - sets mask from glesp-file 'gl_mask.fts'(remove outside) \n");
PR(ST, "-mh  hp_mask.fts - sets mask from hpix-file 'hp_mask.fts' (remove inside)\n");
PR(ER, "-mh- hp_mask.fts - sets mask from hpix-file 'hp_mask.fts' (remove outside)\n");
PR(ST, "-mink - calculates 1st and 2nd Minkowski functionals\n");
PR(ST, "-mirC  - map reflection with the (0,0)-center\n");
PR(ST, "-mirH  - map reflection through equator\n");
PR(ST, "-mirN  - map reflection of North via equator\n");
PR(ST, "-mirN- - map reflection of North via equator changing a sign\n");
PR(ST, "-mirV  - map reflection through 0-meridian\n");
PR(ST, "-mirP  - map reflection with the (0,0)-center + pi-shift\n");
PR(ST, "-mirS  - map reflection of South via equator\n");
PR(ST, "-mirS- - map reflection of South via equator changing a sign\n");
PR(ST, "-mul - calculates multiplication: map1.fts * map2.fts\n");
PR(ST, "-nocom - no output comments\n");
PR(ST, "-o outp_file - output file name\n");
PR(ST, "-oq q_outp_file - outputs Q-map file name (see '-ou', '-pa', '-pp')\n");
PR(ST, "-ou u_outp_file - outputs U-map file name (see '-oq', '-pa', '-pp')\n");
PR(ST, "-pa[d] pol_ang_file - outputs map of pol-angle=0.5*atan(u/q) (see '-u','-q')\n");
PR(ST, "-pr - prints map data in ASCII format\n");
PR(ST, "-pp - pol_intensity file - calculates Q,U-maps (see '-pa', '-oq', '-ou')\n");
PR(ST, "-q file1 - inputs Q-pol file to calc intensity=sqrt(q^2+u^2) (see '-u','-pa')\n");
PR(ST, "-Q - selects Q-polarization map from 3-map for analysis (see '-U','-T')\n");
PR(ST, "-r file file2 file3... - calc robust average for files: file1,...,fileN\n");
PR(ST, "-removes lat1,lon1,lat2,lon2 - sets mask for sky zone to be removed\n");
PR(ST, "-rw repix_win - repixilizes for new pixe size 'repix_win', arcmin\n");
PR(ST, "-rwn  repix_factor - decreasing factor for pixel in repixilizing\n");
PR(ST, "-R - flag to remove selected zone on the figure (see '-zo','-zc','-zf')\n");
PR(ST, "-st  - calculates simple statistics for map\n");
PR(ST, "-stack[q] src_list - makes stacking for selected zones forn file 'src_list'\n");
PR(ST, "-stover  level - calc statistics over level (see 'stunder')\n");
PR(ST, "-str - calculates simple statistics for map and rings\n");
PR(ST, "-stunder level - calc statistics under level (see 'stover')\n");
PR(ST, "-sum - calculates sum: map1.fts + map2.fts\n");
PR(ST, "-sw  sig_win - calculates sigma in pixels of size 'sig_win', arcmin\n");
PR(ST, "-swr sig_win - calc sigma in pixels in running window  of size 'sig_win'\n");
PR(ST, "-T - selects temperature map [default] from 3-map file (see '-Q','-U')\n");
PR(ST, "-u file2 - inputs U-pol file to calc intensity=sqrt(q^2+u^2) (see '-q','-pa')\n");
PR(ST, "-U - selects U-polarization map from 3-map for analysis (see '-Q','-T')\n");
PR(ST, "-v - prints version of '%s'\n", myname);
PR(ST, "-wc  cor_win - calculates correlation in 'cor_win' (same as -cw)\n");
PR(ST, "-ws  sig_win - calculates sigma in 'sig_win' (same as -sw)\n");
PR(ST, "-wsr sig_win - calculates sigma in running 'sig_win' (same as -swr)\n");
PR(ST, "-zc lat,lon - selects zone center (see '-K', '-R', '-zr')\n");
PR(ST, "-zeq - sets equat.coordinates for circular zone centers (see '-zc','-zf')\n");
PR(ST, "-zf file_with_zones - reads zones from file for masking of format:\n");
PR(ST, "    lat1,lon1,lat1,lon1 from rectangular or lat,lon,radius for circle\n");
PR(ST, "-zn  - normalizes with the total number of map pixels\n");
PR(ST, "-zna - normalizes with the actual number of map pixels\n");
PR(ST, "-zo lat1,lon1,lat2,lon2 - selects zone by coordinates: default to remove\n");
PR(ST, "                          (see '-K','-R', '-remove', '-keep')\n");
PR(ST, "-zr radius_in_degs - selects zone radius (see '-K', '-R', '-zc')\n");
exit(0);
}
