
/*
 * functions to read and write environment's variables
 */

#include <stdio.h>
#include <string.h>

#ifdef MSDOS
#include <stdlib.h>
#else
char **environ;
#endif

#ifdef __GNUC__
#include <stdlib.h>
#include <unistd.h>
#else
#include <malloc.h>
#endif

#include "ff.h"

/*
 * This block  to escape the same descriptions in files of
 * one program, VO: 06/05/2002, TAC
 */
int    bitpix;
int    naxis;
long   naxis1;
double bscale;
double b_zero;
double cdelt1;
double crval1;
double crpix1;
double obsra, obsdec, timeesh=0.0, timewsh=0.0;
char   ctype1[30];
char   bunit[30];
char   object[30];
char   rcv_mode[30];
/* end general definition */

#define PR     fprintf
#define ER     stderr
#define ON     1
extern int sun_data_key; /* set from main function, =1, when sun */

extern void EV_fatal();  /* print error message and terminate */

/*
 *  basic structure to keep variables and their values
 */

#define MAXVAR 1024       /* maximum number of keywords  */

static int max_var = MAXVAR;

/*
 *  static struct varslot {
 *      char *name;             \* key word  *\
 *      char *val;              \* key value *\
 *      BOOLEAN exported;
 *  } sym[MAXVAR];
 */

static struct varslot {
    char *name;             /* key word  */
    char *val;              /* key value */
    BOOLEAN exported;
} *sym;

#define NULLS  (struct varslot  *)0

static int first_EVp_call = 1;


static struct varslot *find(name) /* find name of variable */
char *name;
{
    int i;
    struct varslot *v;

    v = NULLS;
    for (i = 0; i < max_var; i++) {
/*
	if (sym[i].name != NULL && sym[i].name[0] == '\0') {
		free(sym[i].name);
		sym[i].name = NULL;
	}
*/
	if (sym[i].name == NULL){
	    if (v == NULLS)
		v = &sym[i];
	} else if (strcmp(sym[i].name,name) == 0) {
	    v = &sym[i];
	    break;
	}
    }
    return(v);
}

static BOOLEAN assign(p, s) /* find memory for value */
char **p, *s;
{
    int size;

    size = strlen(s)+1;
/*
    if (*p != NULL && **p == '\0') {
	    free (*p);
	    *p = NULL;
    }
*/
    if (*p == NULL) {
	if ((*p = (char *)malloc(size)) == NULL)
	   return(FALSE);
    }
    else if ((*p = (char *)realloc(*p, size)) == NULL)
	    return(FALSE);
    strcpy(*p,s);
    return(TRUE);
}

void EVp_buf_size(n)
int n;
{
    if ( first_EVp_call )
	    max_var = n;

    return;
}

BOOLEAN EVset(name, val) /* appropriate value 'val' to variable 'name' */
char *name, *val;
{
     struct varslot *v;

     if ( first_EVp_call ) {
	    register int k;

	    if ( (sym = (struct varslot *)calloc(max_var,
		  sizeof (struct varslot))) == (struct varslot *)0 ) {
			PR(ER, "EVp: no memory for header's buffers\n");
			max_var = 0;
			return(FALSE);
	    }

	    for(k=0; k<max_var; k++) {
		    v = &sym[k];
		    v->name = NULL;
		    v->val  = NULL;
		    v->exported = 0;
	    }
	    first_EVp_call = 0;
     }

     if ( strcmp(name, "HISTORY") == 0 || strcmp(name, "COMMENT") == 0 ) {
	    if (strlen(val) > 60)   val[60] = '\0';
     }
     if ((v = find(name)) == NULLS)
	 return(FALSE);

     return(assign(&v->name,name) && assign(&v->val, val));
}

BOOLEAN EVexport (name)  /* find memory and set a variable */
char *name;
{
    struct varslot *v;
	  if ((v = find(name)) == NULLS)
	      return(FALSE);
	  if (v->name == NULL)
	      if (!assign(&v->name, name) || !assign(&v->val,""))
		  return(FALSE);
	  v->exported = TRUE;
	  return(TRUE);
}

char *EVget(name)   /* get value of variable */
char *name;
{
     struct varslot *v;

     if ((v = find(name)) == NULLS || v->name == NULL)
	 return(NULL);
     return(v->val);
}

BOOLEAN EVinit()
{
   int i, namelen;
   char name[20];

   for(i = 0; environ[i] != NULL; i++) {
       namelen = strcspn(environ[i], "=");
       strncpy(name, environ[i], namelen);
       name[namelen] = '\0';
       if (!EVset(name, &environ[i][namelen+1]) || !EVexport(name))
	   return(FALSE);
   }
   return(TRUE);
}

BOOLEAN EVupdate()
{
     int i, envi, nvlen;
     struct varslot *v;
     static BOOLEAN updated = FALSE;

     if (!updated)
	 if ((environ = (char **)malloc((max_var+1)*sizeof(char *)))
	    == (char **)0)
	     return(FALSE);
     envi = 0;
     for (i = 0; i <max_var; i++) {
	 v = &sym[i];
	 if (v->name == NULL || !v->exported)
	     continue;
	 nvlen = strlen(v->name) +strlen(v->val) +2;
	 if (!updated) {
	     if ((environ[envi] = (char *)malloc(nvlen)) == NULL)
		 return(FALSE);
	 }
	 else if ((environ[envi] = (char *)realloc(environ[envi], nvlen))
	    == NULL)
		 return(FALSE);
	 sprintf(environ[envi], "%s=%s", v->name, v->val);
	 envi++;
     }
     environ[envi] = NULL;
     updated = TRUE;
     return(TRUE);
}

void EVprint() /* print environment's variable and name */
{
     int i;

     for(i = 0; i<max_var; i++)
	if(sym[i].name != NULL)
	   printf("%3s %s=%s\n", sym[i].exported ? "[E]" : "",
	     sym[i].name, sym[i].val);
}

#define HSTRLEN 160  /* length of tmp bufer for for string with keyword */
static char hdrstr[HSTRLEN + 1];
static int  Ff_head_len;

int Hinit1(stream)  /* reads usual F-file head   */
FILE *stream;
{
     char *getenv();
/*
     int i, namelen, vallen;
*/
     char *pe, *pn;
/*   char hdrstr[HSTRLEN];
*/

     if ( getenv("SUNDATA") != NULL)
	       if( strcmp(getenv("SUNDATA"),"ON") == 0)
		       sun_data_key = ON;

     Ff_head_len = 0;

     while (fgets(hdrstr, HSTRLEN, stream) != NULL) {
	  Ff_head_len += strlen(hdrstr);
	  if (strncmp(hdrstr, "END", 3) == 0) break;
	  pe = strchr(hdrstr, '=');
	  *pe = '\0';
	  pn = strchr(++pe, '\n');
	  *pn = '\0';
	  if (!EVset(hdrstr, pe) || !EVexport(hdrstr))
	  return(0);
     }
     return(Ff_head_len);
}

int Hinit(stream)  /* reads common F-file head */
FILE *stream;
{
     register char *pe, *pn;
#ifndef __GNUC__
     char *getenv();
#endif

     if ( getenv("SUNDATA") != NULL)
	       if( strcmp(getenv("SUNDATA"),"ON") == 0)
		    sun_data_key = ON;

     if ( first_EVp_call ) {
	    register int k;
	    struct varslot *v;

	    if ( (sym = (struct varslot *)calloc(max_var,
		  sizeof (struct varslot))) == (struct varslot *)0 ) {
			PR(ER, "EVp: no memory for header's buffers\n");
			max_var = 0;
			return(FALSE);
	    }

	    for(k=0; k<max_var; k++) {
		    v = &sym[k];
		    v->name = NULL;
		    v->val  = NULL;
		    v->exported = 0;
	    }
	    first_EVp_call = 0;
     }

     Ff_head_len = 0;

     while (fgets(hdrstr, HSTRLEN, stream) != NULL) {
	  int key;
	  Ff_head_len += strlen(hdrstr);
	  if (strncmp(hdrstr, "END", 3) == 0) break;

	  /* missing empty string */
	  pe = hdrstr; key = 0;
	  while(*pe) {
		 if( *pe != '\t' && *pe != ' ' && *pe != '\n')  {
			key = 1;
			break;
		 } else
			pe++;
	  }
	  if(!key)  continue;

	  /* missing possible comments */
	  if (*pe == '#')
		continue;

	  /* shift a string to the begining */
	  pn = hdrstr;
	  if (pn != pe )
		while ( (*pn = *pe) ) { pn++; pe++; }

	  /* reading string till symbol '=' */
	  pe = strchr(hdrstr, '=');
	  if ( pe == NULL ) {
		    PR(ER,"EVp: bad string (no sign '='): %s\n", hdrstr);
		    continue;
	  }
	  *pe = '\0';

	  /* deleting '\t' and ' ' between keyword and '=' */
	  pn = hdrstr;
	  while(*pn && *pn != ' ' && *pn != '\t') pn++;
	  *pn = '\0';

	  /* missing '\t' and ' ' between '=' and value */
	  while(*++pe && (*pe == '\t' || *pe == ' ')) ;

	  /* reading value of keyword */
	  pn = strchr(pe, '\n');
	  if ( pn == NULL ) {
		    PR(ER,"EVp: bad string (may be too long): %s\n", pe);
		    continue;
	  }
	  *pn = '\0';

	  /* deleting always from ' \' or ' /' */
	  pn = pe; key = 0;
	  while(*pn) {
		 if(*pn == ' ')   key = 1;
		 pn++;
		 if(key && (*pn == '\\' || *pn == '/' || *pn == ' ') ) {
			pn--;
			break;
		 } else key = 0;
	  }
	  *pn = '\0';

	  if (!EVset(hdrstr, pe) || !EVexport(hdrstr))
		 return(0);
     }
     return(Ff_head_len);
}

int Hupdate(stream) /* output head of image */
FILE *stream;
{
     register int i;
     struct varslot *v;
#ifndef __GNUC__
     char *getenv();
#endif

     if ( getenv("SUNDATA") != NULL)
	       if( strcmp(getenv("SUNDATA"),"ON") == 0)
		       sun_data_key = ON;

     Ff_head_len = 0;

     for (i = 0; i < max_var; i++) {
	 v = &sym[i];
	 if(v->name==NULL ||*(v->name)=='\0' || !v->exported || v->val==NULL)
	       continue;
	 sprintf(hdrstr, "%s=%s\n", v->name, v->val);
	 fprintf(stream, "%s", hdrstr);
	 Ff_head_len += strlen(hdrstr);
     }
     i = (Ff_head_len + 4)%sizeof(int);
     fprintf(stream, "%-*s\n", (i+3), "END");
     Ff_head_len += i + 4;
     fflush(stream);
     return (Ff_head_len);
}

/* output head of image without keywords 'keyword, ...' */
/* added by VO, 25-June-1995                            */
#ifdef sun
#include <varargs.h>
int Hupdate_key_off(va_alist)
va_dcl
{
     va_list vap;
     register int i = 0, j, jn = 0;
     struct varslot *v;
#ifndef __GNUC__
     char *getenv();
#endif
     FILE *stream;
     char *args[MAXVAR];

     if ( getenv("SUNDATA") != NULL)
	       if( strcmp(getenv("SUNDATA"),"ON") == 0)
		       sun_data_key = ON;

     Ff_head_len = 0;

     /* arguments  reading */
     va_start (vap);
     stream = va_arg (vap, FILE *);
     while ((args[jn++] = va_arg(vap, char *)) != (char *)0) ;
     jn--; /* miss NULL at the end */

     for (i = 0; i < max_var; i++) {
	    int keyw = 0;
	    v = &sym[i];
	    if (v->name == NULL || !v->exported)
		  continue;
	    for (j = 0; j < jn; j++) {
		   if ( strcmp ( v->name, args[j] ) == 0 )
			    keyw = 1;
	    }
	    if (keyw) {
		    keyw = 0;
		    continue;
	    }
	    sprintf(hdrstr, "%s=%s\n", v->name, v->val);
	    fprintf(stream, "%s", hdrstr);

	    Ff_head_len += strlen(hdrstr);
     }
     i = (Ff_head_len + 4)%sizeof(int);
     fprintf(stream, "%-*s\n", (i+3), "END");
     Ff_head_len += i + 4;
     fflush(stream);

     va_end (vap);
     return (Ff_head_len);
}
#else
int Hupdate_key_off(stream, kw) /* output head of image */
FILE *stream;
char *kw;
{
     register int i = 0, j, jn = 0;
     struct varslot *v;
#ifndef __GNUC__
     char *getenv();
#endif
     char *args[MAXVAR];
     register char **ap;

     if ( getenv("SUNDATA") != NULL)
	       if( strcmp(getenv("SUNDATA"),"ON") == 0)
		       sun_data_key = ON;

     Ff_head_len = 0;
     /* arguments  reading */
     ap = &kw;
     while(*(ap++) ) jn++;

     ap = &kw;
     for (i = 0; i < jn; ++i) {
	     args[i] = *ap;
	     ap++;
     }

     for (i = 0; i < max_var; i++) {
	    int keyw = 0;

	    v = &sym[i];
	    if (v->name == NULL || !v->exported)
		  continue;

	    for (j = 0; j < jn; j++) {
		   if ( strcmp ( v->name, args[j] ) == 0 )
			    keyw = 1;
	    }
	    if (keyw) {
		    keyw = 0;
		    continue;
	    }

	    sprintf(hdrstr, "%s=%s\n", v->name, v->val);
	    fprintf(stream, "%s", hdrstr);
	    Ff_head_len += strlen(hdrstr);
     }
     i = (Ff_head_len + 4)%sizeof(int);
     fprintf(stream, "%-*s\n", (i+3), "END");
     Ff_head_len += i + 4;
     fflush(stream);
     return (Ff_head_len);
}
#endif

int Hlen()
{
    return Ff_head_len;
}

void sEVset(name,value)
char *name, *value;
{
 if(!EVset(name, value))
	EV_fatal("sEVset");
}

void iEVset(name,i)
char *name;
int i;
{
 char value[20];

 sprintf(value,"%d", i);
 if(!EVset(name, value))
	EV_fatal("iEVset");
}

void lEVset(name, l)
char *name;
long l;
{
 char value[20];

 sprintf(value,"%ld", l);
 if(!EVset(name, value))
	EV_fatal("lEVset");
}


#if !defined(_WIN32)
void llEVset(name, ll)
char *name;
long long ll;
{
 char value[30];

 sprintf(value,"%lld", ll);
 if(!EVset(name, value))
	EV_fatal("llEVset");
}
#else
void llEVset(name, ll)
char *name;
long ll;
{
 char value[30];

 sprintf(value,"%ld", ll);
 if(!EVset(name, value))
	EV_fatal("llEVset");
}
#endif


void eEVset(name, d)
char *name;
double d;
{
 char value[20];

 sprintf(value,"%12.6e", d);
 if(!EVset(name, value))
	EV_fatal("eEVset");
}

#ifdef __GNUC__
void fEVset(char *name, float f)
#else
void fEVset(name, f)
char *name;
float f;
#endif
{
	 int e_format = 0;
	 char value[20];

	 sprintf(value, "%.7g", f);
	 /* delete nulls at the end of string after '.' */
	 /* added by O.Verkhodanov ,  22-May-1995 */
	 {
		register char *cs;
		int k=0;
		/* Is '.' here ? */
		cs = value;
		while (*cs) {
			if( *cs == '.')  k = 1;
			if( *cs == 'e' || *cs == 'E'  )  e_format = 1;
			cs++;
		}
		/* delete nulls */
		if ( k && ! e_format ) {
			k = strlen (value);
			cs = value + k - 1;
			while (*cs != '.') {
				if ( *cs == '0' ) {
					*cs = '\0';
					cs--;
				} else
					break;
			}
		}
	 }

	 if(!EVset(name, value))
		EV_fatal("fEVset");
}

void dEVset(name, d)
char *name;
double d;
{
	 int e_format = 0;
	 char value[50];

	 sprintf(value, "%.15g", d);

	 /* delete nulls at the end of string after '.' */
	 /* added by O.Verkhodanov ,  22-May-1995 */
	 {
		register char *cs;
		int k=0;
		/* Is '.' here ? */
		cs = value;
		while (*cs) {
			if( *cs == '.')  k = 1;
			if( *cs == 'e' || *cs == 'E'  )  e_format = 1;
			cs++;
		}
		/* delete nulls */
		if ( k && ! e_format ) {
			k = strlen (value);
			cs = value + k - 1;
			while (*cs != '.') {
				if ( *cs == '0' ) {
					*cs = '\0';
					cs--;
				} else
					break;
			}
		}
	 }
	 if(!EVset(name, value))
	       EV_fatal("fEVset");
}

void EV_fatal(msg) /* print error message and terminate */
char *msg;
{
    fprintf(stderr, "EVp_error: %s\n", msg);
    exit(1);
}

void Hfree()  /* remove head */
{
     register int i;

     for (i = 0; i < max_var; i++) {
	if (sym[i].name != NULL) { free(sym[i].name);  sym[i].name = NULL; }
	if (sym[i].val  != NULL) { free(sym[i].val);   sym[i].val  = NULL; }
     }
}

void EVdel(name)  /* remove variable and free memory */
char *name;
{
     register int i,j;
     char bufval[HSTRLEN + 1];
/*
     BOOLEAN exported;
*/
     for (i = 0; i < max_var; i++)
	if (sym[i].name != NULL)
	    if (strcmp(sym[i].name, name) == 0) {
		free(sym[i].name);
		free(sym[i].val);
		sym[i].name = NULL;  sym[i].val = NULL;
		for (j = i; j < max_var-1; j++) {
		       if(sym[j+1].val  != NULL) {
			       strcpy(bufval,sym[j+1].val);
			       free(sym[j+1].val);
			       sym[j+1].val = NULL;
		       }
		       if(sym[j+1].name != NULL) {
			       strcpy(hdrstr,sym[j+1].name);
			       free(sym[j+1].name);
			       sym[j+1].name = NULL;
			       sym[j].exported = sym[j+1].exported;
			       EVset(hdrstr,bufval); EVexport(hdrstr);
		       }
		}
		break;
	    }
}

/* for FORTRAN use */
FILE *_fp_FORTRAN_;

void ckey_set__(name, c)
char *name;
char *c;
{
     Ckey_set(name, c);
}
void skey_set__(name, s)
char *name;
short  *s;
{
     Skey_set(name, *s);
}
void lkey_set__(name, l)
char *name;
long  *l;
{
     Skey_set(name, *l);
}

#if !defined (_WIN32)
void llkey_set__(name, ll)
char *name;
long long *ll;
{
     Skey_set(name, *ll);
}
#else
void llkey_set__(name, ll)
char *name;
long *ll;
{
     Skey_set(name, *ll);
}
#endif

#ifdef __GNUC__
void fkey_set__(char *name, float *f)
#else
void fkey_set__(name, f)
char *name;
float *f;
#endif
{
     Fkey_set(name, *f);
}
void dkey_set__(name, d)
char *name;
double *d;
{
     Dkey_set(name, *d);
}

short check_key__(name)
char *name;
{
     return ( EVget(name) != NULL ? 1 : 0);
}
void ckey_get__(name, c)
char *name;
char *c;
{
     if ( EVget(name) != NULL )
	    strcpy(c, EVget(name));
     else
	    c[0] = '\0';
}
short skey_get__(name)
char *name;
{
     return ( EVget(name) != NULL ? Skey_get(name) : 0);
}
long lkey_get__(name)
char *name;
{
     return ( EVget(name) != NULL ? Lkey_get(name) : 0);
}

#if !defined (_WIN32)
long long llkey_get__(name)
char *name;
{
     return ( EVget(name) != NULL ? Lkey_get(name) : 0);
}
#else
long llkey_get__(name)
char *name;
{
     return ( EVget(name) != NULL ? Lkey_get(name) : 0);
}
#endif

float fkey_get__(name)
char *name;
{
     return ( EVget(name) != NULL ? (float)(Fkey_get(name)) : 0.);
}
double dkey_get__(name)
char *name;
{
     return ( EVget(name) != NULL ? Dkey_get(name) : 0.);
}

#ifdef __STDC__
extern void set_sig_fatal(void);
#else
extern void set_sig_fatal();
#endif

#if defined __linux__ || defined sun
char *myname;
#else
extern char *myname;
#endif

short hopen_(file, key)
char *file, *key;
{
	char *evpname = "EVp - FORTRAN";
	myname = evpname;

	set_sig_fatal();

	_fp_FORTRAN_ = fopen(file, key);
	if ( _fp_FORTRAN_ == (FILE *)0 ) {
		fprintf(stderr, "%s: I can't open file %s\n",
			myname, file);
		return -1;
	}
	return 1;
}
void hclose_()
{
	fclose (_fp_FORTRAN_);
}
long hinit_()  /* reads common F-file head in FORTRAN */
{
	return (long) Hinit(_fp_FORTRAN_);
}
long hupdate_()  /* writes common F-file head in FORTRAN */
{
	return (long) Hupdate (_fp_FORTRAN_);
}


