/*
 *   processing signals
 *      created by Oleg Verkhodanov (vo@sao.ru), 2003-2004
 */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#ifdef  __DARWIN__
#define SIGCLD SIGCHLD  /* added by Tristan Beau <beau@in2p3.fr> */
#endif
#include <string.h>

#ifdef MSDOS
	#include <fcntl.h>
	#include <io.h>
	#include <float.h>
#endif

#if !defined (_WIN32)
#endif

extern char *myname;

static char tmp_bufer[10][81];  /* bufer for tmp-file  */
static int  n_tmp = 0;          /* number of tmp-files */

void abrt(), brpipe(), fperr(),sig_h(), fatal2(), fatal();
void set_sig_fatal()
{

#ifdef MSDOS
	setmode(0,O_BINARY);
#endif

	signal(SIGINT,  abrt);
	signal(SIGTERM, abrt);
#if defined (_WIN32) || defined(MSDOS)
	signal(SIGABRT, abrt);
#else
	signal(SIGQUIT, abrt);
	signal(SIGPIPE, brpipe);
#endif
	signal(SIGILL,  sig_h);
#if !defined (__GNUC__) && !defined (_WIN32)
	signal(SIGEMT,  sig_h);
#endif
#if !defined (_WIN32)
	signal(SIGBUS,  sig_h);
	signal(SIGHUP,  sig_h);
	signal(SIGIOT,  sig_h);
	signal(SIGKILL, sig_h);
	signal(SIGTRAP, sig_h);
#else
	signal(SIGBREAK, sig_h);
#endif
	signal(SIGSEGV, sig_h);
	signal(SIGFPE,  fperr);
}

/* save tmp-file to delete after signal */
int sig_tmp_save(name)
char *name;
{
	int flag = 0;
	if ( strlen(name) < 81 && n_tmp < 10 ) {
		strcpy(tmp_bufer[n_tmp], name);
		n_tmp++;
		flag = 1;
	}
	return flag;
}

void abrt()
{
	fatal2("unexpected interrupt",NULL);
}
void brpipe()
{
	fatal2("broken pipe",NULL);
}
void fperr()
{
	fatal2("floating point error",NULL);
}
#ifdef __GNUC__
int unlink(char *);
#else
int unlink();
#endif
void fatal2(s1,s2)
char *s1, *s2;
{
    signal(SIGINT, SIG_IGN);
    signal(SIGTERM,SIG_IGN);
#if defined(MSDOS) || defined(_WIN32)
	signal(SIGABRT,SIG_IGN);
#else
	signal(SIGQUIT,SIG_IGN);
	signal(SIGPIPE,SIG_IGN);
#endif
    signal(SIGFPE, SIG_IGN);
    if(s2 == NULL) fprintf(stderr, "\n%s: %s\n",myname,s1);
    else           fprintf(stderr, "\n%s: %s (%s)\n",myname,s1,s2);

    if ( n_tmp ) {
	    while (n_tmp >= 0 ) {
		    FILE *p;
		    if ((p = fopen(tmp_bufer[n_tmp], "r")) != (FILE *)0) {
			    fclose(p);
			    unlink(tmp_bufer[n_tmp]);
		    }
		    n_tmp--;
	    }
    }
    exit(1);
}
void fatal(s)
char *s;
{       fatal2(s,NULL);
}
void
sig_h(sig)
int sig;
{
	fflush(stdout);

	switch(sig) {
#if !defined (_WIN32)
	case   SIGHUP   :    /*     01    */
		fprintf(stderr,"\nHangup\n");
		break;
#endif
	case   SIGINT   :    /*     02    */
		fprintf(stderr,"\nInterrupt\n");
		break;
#if !defined (_WIN32)
	case   SIGQUIT  :    /*     03*   */
		fprintf(stderr,"\nQuit\n");
		break;
#endif
	case   SIGILL   :    /*     04*   */
		fprintf(stderr,"\nIllegal instruction\n");
		break;
#if !defined (_WIN32)
	case   SIGTRAP  :    /*     05*   */
		fprintf(stderr,"\nTrace trap\n");
		break;
	case   SIGIOT   :    /*     06*   */
		fprintf(stderr,"\nI/O trap instruction\n");
		break;
#ifndef __GNUC__
	case   SIGEMT   :    /*     07*   */
		fprintf(stderr,"\nEmulator trap instruction\n");
		break;
#endif
#endif
	case   SIGFPE   :    /*     08*   */
		fprintf(stderr,"\nFloating-point exception\n");
		break;
#if !defined (_WIN32)
	case   SIGKILL  :    /*     09    */
		fprintf(stderr,"\nKilled\n");
		break;
	case   SIGBUS   :    /*     10*   */
		fprintf(stderr,"\nBus error\n");
		break;
#endif
	case   SIGSEGV  :    /*     11*   */
		fprintf(stderr,"\nSegmentation violation\n");
		break;
#if !defined  (__GNUC__) && !defined (_WIN32)
	case   SIGSYS   :    /*     12*   */
		fprintf(stderr,"\nBad argument to system call\n");
		break;
#endif
#if !defined (_WIN32)
	case   SIGPIPE  :    /*     13    */
		fprintf(stderr,"\nWrite on a pipe with no one to read it\n");
		break;
	case   SIGALRM  :    /*     14    */
		fprintf(stderr,"\nAlarm clock\n");
		break;
#endif
	case   SIGTERM  :    /*     15    */
		fprintf(stderr,"\nSoftware termination signal\n");
		break;
#if !defined (_WIN32)
	case   SIGUSR1  :    /*     16    */
		fprintf(stderr,"\nUser-defined signal 1\n");
		break;
	case   SIGUSR2  :    /*     17    */
		fprintf(stderr,"\nUser-defined signal 2\n");
		break;
	case   SIGCLD   :    /*     18    */
		fprintf(stderr,"\nDeath of a child\n");
		break;
#if !defined (sun) && !defined (__DARWIN__)
	case   SIGPWR   :    /*     19    */
		fprintf(stderr,"\nPower fail\n");
		break;
#endif
#else
	/* Window system */
	case   SIGBREAK :    /*     21    */
		fprintf(stderr,"\nCtrl-Break sequence\n");
		break;
#endif
	}
	if ( n_tmp ) {
	    while (n_tmp >= 0 ) {
		FILE *p;
		if ((p = fopen(tmp_bufer[n_tmp], "r")) != (FILE *)0) {
			fclose(p);
			unlink(tmp_bufer[n_tmp]);
		}
		n_tmp--;
	    }
	}
	exit( sig );
}
