/*
 * some output functions,
 *      created by Oleg Verkhodanov (vo@sao.ru), 2003-2004
 *
 */
# include <stdio.h>
# include <math.h>
#ifndef __DARWIN__      /* by Tristan Beau <beau@in2p3.fr>, 17/06/2005 */
# include <values.h>
#endif

# include <ff.h>
# include "glesp.h"

int    naxis2;
double crval2;
double cdelt2;
double crpix2;
char   ctype2[61];

extern char *myname;
extern int  bitpix_map; /* default: Float Point Pixels for map */

void ff_write_ext2_map(f, x, nx, np, nph, fp_out, x_type, grid_type)
double *f;     /* 2-dim map                           */
double *x;     /* argument, x = cos theta             */
int nx;        /* number of elements in x-array       */
int np;        /* 2-dim factor: f(i*nbuf+j)<->f[i][j] */
int *nph;      /* array of elements in phi-arrays     */
FILE *fp_out;
char *x_type, *grid_type;
{
	register int i,j;
	char ttype1[30], tform1[30], tunit1[30], tdim1[30],
	     ttype2[30], tform2[30], tunit2[30], tdim2[30],
	     ttype3[30], tform3[30], tunit3[30], tdim3[30];
	int tfields = 3;
	long nphi=0;     /* total number of elements in nph-arrays */
	int bitp1 = FLOATPIX, bitp2 = LONGPIX, bitp3;
	long long long_naxis1;

	/*****************/
	/* basic keywods */
	/*****************/
	bitp3 = bitpix_map; /* default: Float Point Pixels for map */
	bscale = 1.0;
	b_zero = 0.0;
	naxis  = 2;
	naxis2 = 1;

	strcpy(ctype1, x_type);
	strcpy(bunit,  "K");
	strcpy(object, "CMB map");

	tfields = 3;

	sprintf(tform1, "%d", nx*(bitp1/8));
	sprintf(tdim1,  "(%d,1)", nx);

	strcpy(ttype1,  ctype1);
	if (strncmp(ctype1, "COS(", 4) == 0)
		strcpy(tunit1,  "COS(RAD)");
	else
		strcpy(tunit1,  "RAD");
	sprintf(tform2, "%d", nx*(bitp2/8));
	sprintf(tdim2,  "(%d,1)", nx);
	strcpy(ttype2,  "PHI");
	strcpy(tunit2,  "RAD");

	for (i=0; i<nx; i++)  nphi += nph[i];
#if defined (_OSF_SOURCE)
	sprintf(tform3, "%ld", (long)nphi*(bitp3/8));
#else
	sprintf(tform3, "%lld", (long long)nphi*(bitp3/8));
#endif
	sprintf(tdim3,  "(%ld,1)", nphi);
	strcpy(ttype3,  "TEMPERATURE");
	strcpy(tunit3,  "K");

	/* Number of entries in table */
	long_naxis1 = (long long)nphi * (ABS(bitp3) / 8);

	/************/
	/* set head */
	/************/
	Ckey_set("SIMPLE",    "T");
	Skey_set("BITPIX",    8);
	Skey_set("NAXIS",     0);
	Ckey_set("EXTEND",    "T");
	Ckey_set("COMMENTS",  "Map of non-uniform array");

	Hupdate(fp_out);
	Hfree();

	Ckey_set("XTENSION",  "BINTABLE");
	Skey_set("BITPIX",  8);
	Skey_set("NAXIS",   2);
	LLkey_set("NAXIS1", long_naxis1);
	Lkey_set("NAXIS2",  (long)naxis2);
	Lkey_set("PCOUNT",  0);
	Lkey_set("GCOUNT",  1);

	Lkey_set("TFIELDS", tfields);

	Ckey_set("TFORM1",  tform1); /* 1st field: vector of X */
	Ckey_set("TDIM1",   tdim1);
	Ckey_set("TTYPE1",  ttype1);
	Ckey_set("TUNIT1",  tunit1);

	Ckey_set("TFORM2",  tform2); /* 2nd field: vector nph(x) */
	Ckey_set("TDIM2",   tdim2);
	Ckey_set("TTYPE2",  ttype2);
	Ckey_set("TUNIT2",  tunit2);

	Ckey_set("TFORM3",  tform3); /* 3d field: vector of V(X,Y)  */
	Ckey_set("TDIM3",   tdim3);
	Ckey_set("TTYPE3",  ttype3);
	Ckey_set("TUNIT3",  tunit3);

	Ckey_set("OBJECT",  object);
	Ckey_set("ORIGIN",  "TAC");
	Ckey_set("PIXTYPE", "GLESP");
	Ckey_set("GRID",    grid_type);

	Hupdate(fp_out);

	/* recording 1-dim array of  X-argument vector */
	bitpix = bitp1;
	for(i=0; i<nx; i++) {
		float v = x[i];
		put_pix(fp_out, v);
	}

	/* recording 1-dim array of dimensions of \phi-vector */
	bitpix = bitp2;
	for(i=0; i<nx; i++) {
		float v = nph[i];
		put_pix(fp_out, v);
	}

	bitpix = bitp3;
	/* recording 2-dim array */
	for(i=0; i<nx; i++) {
		register int ii = i * np;
		for(j=0; j<nph[i]; j++) {
			double v = f[ii+j];
			put_pix_d(fp_out, v);
		}
	}
	/* free memory from head */
	Hfree();

	return;
}
void ff_write_map(f, nbuf, x, nx, phi, nphi, flag_grid, transp_flag, fp_out)
double *f;     /* 2-dim map                           */
int nbuf;      /* 2-dim factor: f(i*nbuf+j)<->f[i][j] */
double *x;     /* argument, x = cos theta             */
int nx;        /* number of elements in x-array       */
double *phi;   /* array of phi (Y-axis)               */
int nphi;      /* number of elements in phi-array     */
int flag_grid; /* <0 - dx=2/(Nx-1), >0 - special grid */
int transp_flag; /* transposing flag                  */
FILE *fp_out;
{
	register double *pd = f, *f_new = (double *)0;
	int  nx_new = nx;
	double x0=x[0], dx = (x[nx-1] - x[0]) / (double)(nx-1);
	register int i,j;
	int condition_flag = 1; /* for spline x0<x1<x2... */

PR(ER, "flag_grid = %d, theta_grid = %d, trnasp_flag=%d\n",
flag_grid, theta_grid, transp_flag);
	if (flag_grid > 0) {
	    /***************************************/
	    /* if special grid we do recalculation */
	    /***************************************/
	    if ((f_new = (double *)calloc(nphi*nx, sizeof *f_new)) ==
		(double *)0) {
nomem:
		    PR(ER,"%s: no memory for output arrays.\n", myname);
		    exit(1);
	    }

	    if (theta_grid == 2) {
		/* recalculate from cos(theta)-grid to theta-grid */
		x0 = acos(x[0]);
		dx = (acos(x[nx-1]) - x0) / (double)(nx-1);
		for(i=0; i<nx; i++)
			x[i] = acos(x[i]);
	    }
	    if (x[0] > x[1]) {
		condition_flag = 0;
		for(i=0; i<nx; i++)  x[i] = -x[i];
		x0 = -x0; dx = -dx;
	    }

	    /*-------------------------------------------*/
	    /* spline interpolation (modified input):    */
	    /*   non-uniform step by X ---> uniform step */
	    /*-------------------------------------------*/
	    intrpl2sm_nu_u(f, nx, nbuf, f_new,nx_new,nphi, x, x0, dx, nphi);

	    if (condition_flag == 0) {
		condition_flag = 1;
		for(i=0; i<nx; i++)  x[i] = -x[i];
		x0 = -x0; dx = -dx;
	    }

	    nbuf = nphi;
	    pd = f_new;
	}

	if (theta_grid == 1) {
	    /* recalculate from cos(theta)-grid to theta-grid */
	    x0 =  acos(x[0]);
	    dx = (acos(x[nx-1]) - x0) / (double)(nx-1);
	    for(i=0; i<nx; i++)
		    x[i] = acos(x[i]);

	    if (x[0] > x[1]) {
		condition_flag = 0;
		for(i=0; i<nx; i++)  x[i] = -x[i];
		x0 = -x0; dx = -dx;
	    }

	    i= nphi * (nx > nx_new ? nx : nx_new);
	    /* if special grid we do recalculation */
	    if ((f_new = (double *)calloc(i, sizeof *f_new)) == (double *)0)
		    goto nomem;

	    /*-------------------------------------------*/
	    /* spline interpolation (modified input):    */
	    /*   non-uniform step by X ---> uniform step */
	    /*-------------------------------------------*/
	    intrpl2sm_nu_u(f, nx, nx, f_new, nx_new, nx_new, x, x0, dx,nphi);

	    if (condition_flag == 0) {
		condition_flag = 1;
		for(i=0; i<nx; i++)  x[i] = -x[i];
		x0 = -x0; dx = -dx;
	    }

	    nbuf = nphi;
	    pd = f_new;
	    ph->grid = GRID_UNIFORM_THETA;
	} else
	    ph->grid = GRID_UNIFORM_COS;

	/* just to be sure there nothing in memory */
	Ckey_set("SIMPLE",  "T");
	Hfree();

	/*****************/
	/* basic keywods */
	/*****************/
	bitpix = FLOATPIX;
	bscale = 1.000000000000;
	b_zero = 0.000000000000;
	naxis  = 2;

	/* just to save temporal value */
	if (theta_grid) {
		strcpy(bunit, "THETA");
	} else {
		strcpy(bunit, "COS(THETA)");
	}
	/**************************************/
	/* change axis if we need transposing */
	/**************************************/
	if ( transp_flag ) {
		naxis1 = nx;
		naxis2 = nphi;

		crval1 = x0;
		cdelt1 = dx;
		crpix1 = 1.;
		strcpy(ctype1, bunit);

		crval2 = phi[0];
		cdelt2 = (phi[nphi-1] - phi[0]) / (double)(nphi-1);
		crpix2 = 1.;
		strcpy(ctype2, "PHI");
	} else {
		naxis2 = nx;
		naxis1 = nphi;

		crval1 = phi[0];
		cdelt1 = (phi[nphi-1] - phi[0]) / (double)(nphi-1);
		crpix1 = 1.;
		strcpy(ctype1, "PHI");

		crval2 = x0;
		cdelt2 = dx;
		crpix2 = 1.;
		strcpy(ctype2, bunit);
	}
	strcpy(bunit,  "mK");
	strcpy(object, "CMB map");

	/************/
	/* set head */
	/************/
	Ckey_set("SIMPLE",  "T");
	Skey_set("BITPIX",  bitpix);
	Skey_set("NAXIS",   naxis);
	Lkey_set("NAXIS1",  naxis1); /* transp_flag=0:|nx  |=1:|nphi */
	Lkey_set("NAXIS2",  naxis2); /*               |nphi|   |nx   */
	Ckey_set("BUNIT",   bunit);
	Dkey_set("BSCALE",  bscale);
	Dkey_set("BZERO",   b_zero);

	Ckey_set("CTYPE1",  ctype1);
	Dkey_set("CRVAL1",  crval1);
	Dkey_set("CRPIX1",  crpix1);
	Dkey_set("CDELT1",  cdelt1);

	Ckey_set("CTYPE2",  ctype2);
	Dkey_set("CRVAL2",  crval2);
	Dkey_set("CRPIX2",  crpix2);
	Dkey_set("CDELT2",  cdelt2);

	Ckey_set("OBJECT",  object);
	Ckey_set("ORIGIN",  "TAC");

	switch(ph->grid) {
	case  GRID_LEGENDRE:
		Ckey_set("GRID",    "LEGENDRE");
		break;
	case  GRID_UNIFORM_COS:
		Ckey_set("GRID",    "UNIFORM COS");
		break;
	case  GRID_UNIFORM_THETA:
		Ckey_set("GRID",    "THETA UNIFORM");
		break;
	default :
		Ckey_set("GRID",    "NOT SET");
		break;
	}

	Hupdate(fp_out);

	if (  transp_flag ) {
/*
Good figure (circles): what's going on ?
such output, but nax1 and nax2 change each other... (nax2 = 2*nax1)
*/
	    for(j=0; j<nphi; j++) {
		for(i=0; i<nx_new; i++) {
			register int k = i * nbuf;
			float v = (float)((pd[k+j] - b_zero)/bscale);
			put_pix(fp_out, v);
		}
	    }
	} else { /* write by standard way */
	    for(i=0; i<nx_new; i++) {
		register int k = i * nbuf;
		for(j=0; j<nphi; j++) {
			float v = (float)((pd[k+j] - b_zero)/bscale);
			put_pix(fp_out, v);
//if (i==nx_new/2-1 && j==0)
//PR(ER, "v=%12.5e\n", v);
		}
	    }
	}
	if ( f_new != (double *)0 )  free ((void *)f_new);

	return;
}

void ff_write_ext_map(f, nbuf, x, nx, phi, nphi, fp_out)
double *f;     /* 2-dim map                           */
int nbuf;      /* 2-dim factor: f(i*nbuf+j)<->f[i][j] */
double *x;     /* argument, x = cos theta             */
int nx;        /* number of elements in x-array       */
double *phi;   /* array of phi (Y-axis)               */
int nphi;      /* number of elements in phi-array     */
FILE *fp_out;
{
	double x0=x[0], dx = (x[nx-1] - x[0]) / (double)(nx-1);
	register int i,j;
	char ttype1[30], tform1[30], tunit1[30], tdim1[30],
	     ttype2[30], tform2[30], tunit2[30], tdim2[30];
	int tfields = 2;

	/*****************/
	/* basic keywods */
	/*****************/
	bitpix = FLOATPIX;
	bscale = 1.0;
	b_zero = 0.0;
	naxis  = 2;
	naxis1 = nx * nphi * ABS(bitpix);
	naxis2 = 1;
	crval1 = x0;
	crval2 = phi[0];
	cdelt1 = dx;
	cdelt2 = (phi[nphi-1] - phi[0]) / (double)(nphi-1);
	crpix1 = 1.;
	crpix2 = 1.;
	if (theta_grid) {
		strcpy(ctype1, "THETA");
	} else {
		strcpy(ctype1, "COS(THETA)");
	}
	strcpy(ctype2, "PHI");
	strcpy(bunit,  "mK");
	strcpy(object, "CMB map");

	tfields = 2;

	sprintf(tform1, "%d", nx*bitpix);
	sprintf(tdim1,  "(%d,1)", nx);
	strcpy(ttype1,  "COS(THETA)");
	strcpy(tunit1,  "COS(RAD)");

	sprintf(tform2, "%d", nx*nphi*bitpix);
	sprintf(tdim2,  "(%d,%d)", nx, nphi);
	strcpy(ttype2,  "PHI");
	strcpy(tunit2,  "RAD");

	/************/
	/* set head */
	/************/
	Ckey_set("SIMPLE",    "T");
	Skey_set("BITPIX",    8);
	Skey_set("NAXIS",     0);
	Ckey_set("EXTEND",    "T");
	Ckey_set("COMMENTS",  "Map of non-uniform array");

	Hupdate(fp_out);
	Hfree();

	Ckey_set("XTENSION",  "BINTABLE");
	Skey_set("BITPIX",  8);
	Skey_set("NAXIS",   2);
	Lkey_set("NAXIS1",  naxis1);
	Lkey_set("NAXIS2",  naxis2);
	Lkey_set("PCOUNT",  0);
	Lkey_set("GCOUNT",  1);

	Lkey_set("TFIELDS", tfields);

	Ckey_set("TFORM1",  tform1); /* 1st field: vector of X */
	Ckey_set("TDIM1",   tdim1);
	Ckey_set("TTYPE1",  ttype1);
	Ckey_set("TUNIT1",  tunit1);
	Ckey_set("TFORM2",  tform2); /* 2nd field: vector of V(X,Y) */
	Ckey_set("TDIM2",   tdim2);
	Ckey_set("TTYPE2",  ttype2);
	Ckey_set("TUNIT2",  tunit2);

	Dkey_set("CRVAL2",  crval2);
	Dkey_set("CDELT2",  cdelt2);
	Dkey_set("CRPIX2",  crpix2);
	Ckey_set("CTYPE2",  ctype2);

	Ckey_set("OBJECT",  object);
	Ckey_set("ORIGIN",  "TAC");
	Ckey_set("GRID",    "LEGENDRE");

	Hupdate(fp_out);

	/* recording 1-dim array of  X-argument vector */
	for(i=0; i<nx; i++) {
		float v = (float)x[i];
		put_pix(fp_out, v);
	}

	/* recording 2-dim array */
	for(i=0; i<nx; i++) {
		register int ii = i * nbuf;
		for(j=0; j<nphi; j++) {
			float v = (float)f[ii+j];
			put_pix(fp_out, v);
		}
	}
	/* free memory from head */
	Hfree();

	return;
}

void write_simple_map(fp, f, nax1, nax2, crv1, crv2, cdlt1, cdlt2, ctp1, ctp2,
		bsc, bunit)
FILE *fp;
double *f;
int nax1, nax2;
double crv1, crv2;
double cdlt1, cdlt2;
char *ctp1, *ctp2, *bunit;
double bsc;
{
	register int i, j;
	double epoch=2000.0;

	bitpix = FLOATPIX;
	Ckey_set("SIMPLE",    "T");
	Skey_set("BITPIX",    bitpix);
	Skey_set("NAXIS",     2);
	Lkey_set("NAXIS1",    (long)nax1);
	Lkey_set("NAXIS2",    (long)nax2);
	Dkey_set("CRVAL1",    (double)crv1);
	Dkey_set("CDELT1",    (double)cdlt1);
	Dkey_set("CRPIX1",    1.);
	Ckey_set("CTYPE1",    ctp1);
	Dkey_set("CRVAL2",    (double)crv2);
	Dkey_set("CDELT2",    (double)cdlt2);
	Dkey_set("CRPIX2",    1.);
	Ckey_set("CTYPE2",    ctp2);
	Ckey_set("BUNIT",     bunit);
	Dkey_set("BSCALE",    bsc);
	Dkey_set("EQUINOX",   epoch);
	Ckey_set("OBJECT",    "CMB cut");

	Hupdate(fp);

	for(i=0; i<nax1; i++) {
	    int ii = nax2 * i;
	    for(j=0; j<nax2; j++) {
		   put_pix(fp, f[ii+j]/bsc);
	    }
	}

	return;
}

void ff_write_map2d(f, x, nx, nphi, nph, fp_out)
double *f, *x;
int nx, nphi;
int *nph;
FILE *fp_out;
{
    register int i, j;
    double th0 = acos(x[0]);
    double dth = (acos(x[nx-1]) - th0) / nx;
    double dphi = 2. * M_PI / nphi;
    double rad2ang = 180. / M_PI;
    double bsc = 1.0000;

    bitpix = FLOATPIX;
    Ckey_set("SIMPLE",    "T");
    Skey_set("BITPIX",    bitpix);
    Skey_set("NAXIS",     2);
    Lkey_set("NAXIS1",    (long)nphi);
    Lkey_set("NAXIS2",    (long)nx);
    Dkey_set("CRVAL1",    th0*rad2ang);
    Dkey_set("CDELT1",    dth*rad2ang);
    Dkey_set("CRPIX1",    1.);
    Ckey_set("CTYPE1",    "GLON");
    Dkey_set("CRVAL2",    0.);
    Dkey_set("CDELT2",    dphi*rad2ang);
    Dkey_set("CRPIX2",    1.);
    Ckey_set("CTYPE2",    "GLAT");
    Ckey_set("BUNIT",     bunit);
    Dkey_set("BSCALE",    bsc);
    Ckey_set("OBJECT",    "CMB cut");

    Hupdate(fp_out);

    for(i=0; i<nx; i++) {
	double th = i * dth;
	for(j=0; j<nphi; j++) {
	       double phi = dphi * j;
	       register int nn = ang2n(th, phi, x, nx, nphi, nph);
	       float v = f[nn]/bsc;
	       put_pix(fp_out, v);
	}
    }

    return;
}
