#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"
#include "fits.h"

#define PR              fprintf
#define ER              stderr

extern void fhp_in(double **, int *, int *, FILE *);
extern void fhp2tac(double *, int, int, double **, int **, int *, int *,
		double **);
extern void read_hp_file(double **, int **, int *, int *, double **, char *);

int  test_flag = 0;

extern char *myname;

#define N_HP_FIELDS 16
void fhp_in(map_ring, nside, npixtot, fp_in)
double **map_ring;         /* map_ring(ntheta, nphi) */
int *nside, *npixtot;
FILE *fp_in;
{
	register int i, j, tf;
	char str[81];
	int tfields=0, first_pix=0, last_pix=0;
	char ttype1[30], tform1[30], tunit1[30];
	char ttype2[30], tform2[30], tunit2[30];
	char ttype3[30], tform3[30], tunit3[30];
	char ttype4[30], tform4[30], tunit4[30];
	int order[N_HP_FIELDS], bitp[N_HP_FIELDS];
	double bad_pix = 0., max_pix = -1e30, min_pix = 1e30;
	int n_bad_pix = 0, bad_pix_flag = 0;
	int block_size = 0, n_blocks;
	int nested_flag = 0;
	int npix=0;
	double bad_pix_default = 0.0;

	/************************************/
	/* input data to FITS binary Table */
	/************************************/
	/* 1st string analysis */
	for(i=0; i<80; i++) {
		if (feof(fp_in)) break;
		str[i] = fgetc(fp_in);
	}
	str[i] = '\0';
	if (*str == '\0') {
		PR(ER, "%s: empty file with a map.\n", myname);
		exit(1);
	} else {
		if(strncmp(str, "SIMPLE  =", 9) != 0) {
			PR(ER, "%s: a map file is not in FITS.\n", myname);
			exit(1);
		}
	}

	/*------------*/
	/* 1st header */
	/*------------*/

	FHinit_FITS(fp_in, 80);  /* input 1st header */

	/* FITS dataset may contain extensions */
	if (FEVget("EXTEND") == NULL) {
		PR(ER, "%s: a map file has not XTENSION.\n", myname);
		exit(1);
	}

	FHfree();        /* free memory after 1st header output */

	/*------------*/
	/* 2nd header */
	/*------------*/
	FHinit_FITS(fp_in, 0);   /* input 2nd header */

	/* binary table extension           */
	if (FEVget("XTENSION") == NULL) {
		PR(ER, "%s: a map file has not XTENSION.\n", myname);
		exit(1);
	}
	strcpy(str, FEVget("XTENSION"));
	if (strcmp(str, "BINTABLE") != 0) {
		PR(ER, "%s: a map file has not BINTABLE extension.\n",
			myname);
		exit(1);
	}

	/* number of fields in each row     */
	if (FEVget("TFIELDS") != NULL)
		tfields = atoi(FEVget("TFIELDS"));


	/*------------------------------------------*/
	/*  Sky Map Pixelisation Specific Keywords  */
	/*------------------------------------------*/
	if (FEVget("PIXTYPE") != NULL) {
	    char pixtype[20];
	    strcpy(pixtype, FEVget("PIXTYPE"));
	    if (strcmp(pixtype, "HEALPIX") == 0) {
		  /* Pixel ordering scheme (RING)     */
		  if (FEVget("ORDERING") != NULL)
			  strcpy(str, FEVget("ORDERING"));
		  else
			  str[0] = '\0';
		  if (strcmp(str, "RING") != 0) {
		      if (strcmp(str, "NESTED") != 0) {
			  PR(ER, "%s: unknown ordering of pixels\n",myname);
		      } else
			    nested_flag = 1; /* flag to re-order pixels */
		  }
	    } else {
		PR(ER,
		   "%s: procedure 'fhp_in' works only with HEALPix data.\n",
		   myname);
		exit(1);
	    }
	}  else {
	    PR(ER,
	       "%s: uknown pixel type of data (no keyword PIXTYPE).\n",
		   myname);
	    exit(1);
	}
	/* Resolution parameter for HEALPIX */
	if (FEVget("NSIDE") != NULL)   *nside = atoi(FEVget("NSIDE"));

	/* First pixel # (0 based)          */
	if (FEVget("FIRSTPIX") != NULL) first_pix = atoi(FEVget("FIRSTPIX"));
	/* Last pixel # (0 based)           */
	if (FEVget("LASTPIX") != NULL)  last_pix = atoi(FEVget("LASTPIX"));
	*npixtot = last_pix - first_pix + 1;
	/* N pixels total */
	if (FEVget("NPIX") != NULL)  npix = atoi(FEVget("NPIX"));
	if (npix > *npixtot) *npixtot = npix;
	if (FEVget("BAD_DATA") != NULL) {
		bad_pix = atof(FEVget("BAD_DATA"));
		bad_pix_flag = 1;
	}

	/* how many TFIELDS are possible to read ? */
	/* let it be N_HP_FIELDS */
	if (tfields >= N_HP_FIELDS) {
	    PR(ER, "%s: sorry, the TFIELDS keyword = %d, ", myname, tfields);
	    PR(ER, "the maximal TFIELDS is %d.\n", N_HP_FIELDS);
	    exit(1);
	}

	/*------------------------------------*/
	/* Data Description Specific Keywords */
	/*------------------------------------*/
	for(tf=0; tf<tfields; tf++) {
	    char ttype[50], tform[50], tunit[50];

	    sprintf(ttype, "TTYPE%d", tf+1);
	    sprintf(tform, "TFORM%d", tf+1);
	    sprintf(tunit, "TUNIT%d", tf+1);

	    if (strcmp(FEVget(ttype), "TEMPERATURE") == 0 ||
		strncmp(FEVget(ttype), "INT",3) == 0 ||  /* intensity */
		strncmp(FEVget(ttype), "I_STOKES",8) == 0 ||
		strncmp(FEVget(ttype), "I_Stokes",8) == 0) {
		/* Temperature map */
		if (FEVget(ttype) != NULL)
			strcpy(ttype1, FEVget(ttype));

		/* data format of field: */
		if (FEVget(tform) != NULL)
			strcpy(tform1, FEVget(tform));

		/* map unit */
		if (FEVget(tunit) != NULL)
			strcpy(tunit1, FEVget(tunit));

		order[tf] = 'T';
		strcpy(tform, tform1);
	    } else if (strncmp(FEVget(ttype), "Q-POL",5) == 0 ||
		       strncmp(FEVget(ttype), "Q_POL",5) == 0 ||
		       strncmp(FEVget(ttype), "Q_STOKES",8) == 0 ||
		       strncmp(FEVget(ttype), "Q_Stokes",8) == 0) {
		if (FEVget(ttype) != NULL)
			strcpy(ttype2, FEVget(ttype));

		/* data format of field: */
		if (FEVget(tform) != NULL)
			strcpy(tform2, FEVget(tform));

		if (FEVget(tunit) != NULL)
			strcpy(tunit2, FEVget(tunit));

		order[tf] = 'Q';
		strcpy(tform, tform2);
	    } else if (strncmp(FEVget(ttype), "U-POL",5) == 0 ||
		       strncmp(FEVget(ttype), "U_POL",5) == 0 ||
		       strncmp(FEVget(ttype), "U_STOKES",8) == 0 ||
		       strncmp(FEVget(ttype), "U_Stokes",8) == 0) {
		if (FEVget(ttype) != NULL)
			strcpy(ttype3, FEVget(ttype));

		/* data format of field: */
		if (FEVget(tform) != NULL)
			strcpy(tform3, FEVget(tform));

		if (FEVget(tunit) != NULL)
			strcpy(tunit3, FEVget(tunit));

		order[tf] = 'U';
		strcpy(tform, tform3);
	    } else if (strcmp(FEVget(ttype), "N_OBS") == 0) {
		if (FEVget(ttype) != NULL)
			strcpy(ttype4, FEVget(ttype));

		/* data format of field: */
		if (FEVget(tform) != NULL)
			strcpy(tform4, FEVget(tform));

		if (FEVget(tunit) != NULL)
			strcpy(tunit4, FEVget(tunit));

		order[tf] = 'N';
		strcpy(tform, tform4);
	    } else {
		/* data format of field: */
		if (FEVget(tform) != NULL)
			strcpy(tform, FEVget(tform));

		order[tf] = 'N';
	    }

	    bitpix = -32;
	    i = strlen(tform) - 1;
	    switch(tform[i]) {
	    case 'E':
	    case 'F':
		    bitpix = -32;
		    break;
	    case 'J':
		    bitpix = 32;
		    break;
	    case 'D':
		    bitpix = 64;
		    break;
	    case 'I':
		    bitpix = 16;
		    break;
	    case 'B':
		    bitpix = 8;
		    break;
	    }
	    bitp[tf] = bitpix;


	    if (strlen(tform) == 1) i = 1;
	    else  i = atoi(tform);
	    if (!tf) {
		block_size = i;
	    } else {
		if (block_size != i) {
		    PR(ER, "%s: I can not read these data because ", myname);
		    PR(ER, "of different blocks for fields...\n");
		}
	    }
if (glesp_comments) {
if(!tf)
PR(ER, "input data: bitpix=%d, npixtot=%d, nside=%d\n",
bitpix, *npixtot, *nside);
}
	}

	FHfree();  /* free memory after 2nd header */

	/* find memory */
	if ((*map_ring = (double *)calloc(*npixtot +1, sizeof **map_ring)) ==
	     (double *)0) {
		PR(ER,"%s: no memory for temporal map array...\n", myname);
		exit(1);
	}

	/*********************/
	/* input binary data */
	/*********************/
	n_blocks = *npixtot / block_size;

if (glesp_comments)
PR(ER,"%s: Read map...", myname);
	for(i=0; i < n_blocks; i++) {
	  int ib = i * block_size;

	  for(tf=0; tf<tfields; tf++) {
	    bitpix = bitp[tf];

	    for(j=0; j < block_size; j++) {
		int pix_number = ib + j;
		double pix_val;

		if (nested_flag) {
		    pix_number = convert_nested2ring(pix_number, *nside);
		}
		if (feof(fp_in)) {
		    PR(ER, "\n");
		    PR(ER, "%s: end of file (pos=%ld),", myname,ftell(fp_in));
		    PR(ER, " no data after element %d.\n", ib+j);
		    exit(2);
		}
		if (pix_number > *npixtot) {
		   PR(ER,
		   "\n%s: pixel number (%d) is greater than npixtot (%d)\n",
		   myname, pix_number, *npixtot);
		   exit(1);
		}

		pix_val = get_pix_fits(fp_in);

		if (bad_pix_flag && ABS(pix_val/bad_pix - 1.0) < 0.5) {
		    pix_val = bad_pix_default;
		    n_bad_pix++;
		} else {
		    switch(order[tf]) {
		    case 'T':
			if (pix_val > max_pix)  max_pix = pix_val;
			if (pix_val < min_pix)  min_pix = pix_val;
			if (test_flag)
			      pix_val=(fabs(pix_val)<1.e-15 ? 0.0 : 1.0);
			(*map_ring)[pix_number] = pix_val;
			break;
		    case 'Q':
		    case 'U':
		    case 'N':
		    default :
			break;
		    }
		}
	    } /* end of loop for block_size */
	} /* end of loop for tfields */
    }  /* end of loop for n_blocks */
if (glesp_comments) {
PR(ER,"     O'k\n");
if (n_bad_pix) {
if (n_bad_pix==1) PR(ER, "1 bad data pixel has been detected.\n");
else PR(ER, "%d bad data pixels have been detected.\n", n_bad_pix);
}
PR(ER, "minimum = %12.5e, maximum = %12.5e.\n", min_pix, max_pix);
}

	return;
}

// #define N_HP_FIELDS 16
void fhpol_in(map_ring, q, u, counts, nside, npixtot, fp_in)
double **map_ring, **u, **q;         /* map_ring(ntheta, nphi) */
int **counts;
int *nside, *npixtot;
FILE *fp_in;
{
	register int i, j;
	char str[81];
	int tfields=0, first_pix=0, last_pix=0;
	char ttype1[30], tform1[30], tunit1[30];
	char ttype2[30], tform2[30], tunit2[30];
	char ttype3[30], tform3[30], tunit3[30];
	char ttype4[30], tform4[30], tunit4[30];
	int order[N_HP_FIELDS], bitp[N_HP_FIELDS];
	int block_size=0, n_blocks;
	int nested_flag = 0;
	int npix=0;
	int tf;

	/************************************/
	/* input data to FITS binary Table */
	/************************************/
	/* 1st string analysis */
	for(i=0; i<80; i++) {
		if (feof(fp_in)) break;
		str[i] = fgetc(fp_in);
	}
	str[i] = '\0';
	if (*str == '\0') {
		PR(ER, "%s: empty file with a map.\n", myname);
		exit(1);
	} else {
		if(strncmp(str, "SIMPLE  =", 9) != 0) {
			PR(ER, "%s: a map file is not in FITS.\n", myname);
			exit(1);
		}
	}

	/*------------*/
	/* 1st header */
	/*------------*/

	FHinit_FITS(fp_in, 80);  /* input 1st header */

	/* FITS dataset may contain extensions */
	if (FEVget("EXTEND") == NULL) {
		PR(ER, "%s: a map file has not XTENSION.\n", myname);
		exit(1);
	}

	FHfree();        /* free memory after 1st header output */

	/*------------*/
	/* 2nd header */
	/*------------*/
	FHinit_FITS(fp_in, 0);   /* input 2nd header */

	/* binary table extension           */
	if (FEVget("XTENSION") == NULL) {
		PR(ER, "%s: a map file has not XTENSION.\n", myname);
		exit(1);
	}
	strcpy(str, FEVget("XTENSION"));
	if (strcmp(str, "BINTABLE") != 0) {
		PR(ER, "%s: a map file has not BINTABLE extension.\n",
			myname);
		exit(1);
	}

	/* number of fields in each row     */
	if (FEVget("TFIELDS") != NULL)
		tfields = atoi(FEVget("TFIELDS"));

	if ( tfields >= N_HP_FIELDS ) {
		PR(ER, "%s: maximum allowable value of TFIELDS is %d. ",
			myname, N_HP_FIELDS);
		PR(ER, "Here is TFIELDS=%d\n", tfields);
		exit(1);
	}

	/*------------------------------------------*/
	/*  Sky Map Pixelisation Specific Keywords  */
	/*------------------------------------------*/
	if (FEVget("PIXTYPE") != NULL) {
	    char pixtype[20];
	    strcpy(pixtype, FEVget("PIXTYPE"));
	    if (strcmp(pixtype, "HEALPIX") == 0) {
		  /* Pixel ordering scheme (RING)     */
		  if (FEVget("ORDERING") != NULL)
			  strcpy(str, FEVget("ORDERING"));
		  else
			  str[0] = '\0';
		  if (strcmp(str, "RING") != 0) {
		      if (strcmp(str, "NESTED") != 0) {
			  PR(ER, "%s: unknown ordering of pixels\n",myname);
		      } else
			    nested_flag = 1; /* flag to re-order pixels */
		  }
	    } else {
		PR(ER,
		   "%s: procedure 'fhp_in' works only with HEALPix data.\n",
		   myname);
		exit(1);
	    }
	}  else {
	    PR(ER,
	       "%s: uknown pixel type of data (no keyword PIXTYPE).\n",
		   myname);
	    exit(1);
	}
	/* Resolution parameter for HEALPIX */
	if (FEVget("NSIDE") != NULL)   *nside = atoi(FEVget("NSIDE"));

	/* First pixel # (0 based)          */
	if (FEVget("FIRSTPIX") != NULL) first_pix = atoi(FEVget("FIRSTPIX"));
	/* Last pixel # (0 based)           */
	if (FEVget("LASTPIX") != NULL)  last_pix = atoi(FEVget("LASTPIX"));
	*npixtot = last_pix - first_pix + 1;
	/* N pixels total */
	if (FEVget("NPIX") != NULL)  npix = atoi(FEVget("NPIX"));
	if (npix > *npixtot) *npixtot = npix;

	/*------------------------------------*/
	/* Data Description Specific Keywords */
	/*------------------------------------*/
	for(tf=0; tf<tfields; tf++) {
	    char ttype[50], tform[50], tunit[50];

	    sprintf(ttype, "TTYPE%d", tf+1);
	    sprintf(tform, "TFORM%d", tf+1);
	    sprintf(tunit, "TUNIT%d", tf+1);

	    if (strcmp(FEVget(ttype), "TEMPERATURE") == 0 ||
		strncmp(FEVget(ttype), "INT",3) == 0 ||  /* intensity */
		strncmp(FEVget(ttype), "I_STOKES",8) == 0 ||
		strncmp(FEVget(ttype), "I_Stokes",8) == 0) {
		/* Temperature map */
		if (FEVget(ttype) != NULL)
			strcpy(ttype1, FEVget(ttype));

		/* data format of field: */
		if (FEVget(tform) != NULL)
			strcpy(tform1, FEVget(tform));

		/* map unit */
		if (FEVget(tunit) != NULL)
			strcpy(tunit1, FEVget(tunit));

		order[tf] = 'T';
		strcpy(tform, tform1);
	    } else if (strncmp(FEVget(ttype), "Q-POL",5) == 0 ||
		       strncmp(FEVget(ttype), "Q_POL",5) == 0 ||
		       strncmp(FEVget(ttype), "Q_STOKES",8) == 0 ||
		       strncmp(FEVget(ttype), "Q_Stokes",8) == 0) {
		if (FEVget(ttype) != NULL)
			strcpy(ttype2, FEVget(ttype));

		/* data format of field: */
		if (FEVget(tform) != NULL)
			strcpy(tform2, FEVget(tform));

		if (FEVget(tunit) != NULL)
			strcpy(tunit2, FEVget(tunit));

		order[tf] = 'Q';
		strcpy(tform, tform2);
	    } else if (strncmp(FEVget(ttype), "U-POL",5) == 0 ||
		       strncmp(FEVget(ttype), "U_POL",5) == 0 ||
		       strncmp(FEVget(ttype), "U_STOKES",8) == 0 ||
		       strncmp(FEVget(ttype), "U_Stokes",8) == 0) {
		if (FEVget(ttype) != NULL)
			strcpy(ttype3, FEVget(ttype));

		/* data format of field: */
		if (FEVget(tform) != NULL)
			strcpy(tform3, FEVget(tform));

		if (FEVget(tunit) != NULL)
			strcpy(tunit3, FEVget(tunit));

		order[tf] = 'U';
		strcpy(tform, tform3);
	    } else if (strcmp(FEVget(ttype), "N_OBS") == 0) {
		if (FEVget(ttype) != NULL)
			strcpy(ttype4, FEVget(ttype));

		/* data format of field: */
		if (FEVget(tform) != NULL)
			strcpy(tform4, FEVget(tform));

		if (FEVget(tunit) != NULL)
			strcpy(tunit4, FEVget(tunit));

		order[tf] = 'N';
		strcpy(tform, tform4);
	    } else {
		/* data format of field: */
		if (FEVget(tform) != NULL)
			strcpy(tform, FEVget(tform));

		order[tf] = 'N';
	    }

	    bitpix = -32;
	    i = strlen(tform) - 1;
	    switch(tform[i]) {
	    case 'E':
	    case 'F':
		    bitpix = -32;
		    break;
	    case 'J':
		    bitpix = 32;
		    break;
	    case 'D':
		    bitpix = 64;
		    break;
	    case 'I':
		    bitpix = 16;
		    break;
	    case 'B':
		    bitpix = 8;
		    break;
	    }
	    bitp[tf] = bitpix;


	    if (strlen(tform) == 1) i = 1;
	    else  i = atoi(tform);
	    if (!tf) {
		block_size = i;
	    } else {
		if (block_size != i) {
		    PR(ER, "%s: I can not read these data because ", myname);
		    PR(ER, "of different blocks for fields...\n");
		}
	    }
if (glesp_comments) {
if(!tf)
PR(ER, "input data: bitpix=%d, npixtot=%d, nside=%d\n",
bitpix, *npixtot, *nside);
}
	}

	FHfree();  /* free memory after 2nd header */

	/* find memory */
	if ((*map_ring = (double *)calloc(*npixtot, sizeof **map_ring)) ==
	     (double *)0) {
nomem:
		PR(ER,"%s: no memory for temporal map array...\n", myname);
		exit(1);
	}
	if ((*q = (double *)calloc(*npixtot, sizeof **q)) ==
	     (double *)0)
		goto nomem;

	if ((*u = (double *)calloc(*npixtot, sizeof **u)) ==
	     (double *)0)
		goto nomem;

	if ((*counts = (int *)calloc(*npixtot, sizeof **counts)) == (int *)0)
		goto nomem;

	/*********************/
	/* input binary data */
	/*********************/
	n_blocks = *npixtot / block_size;

if (glesp_comments)
PR(ER,"%s: Read 3map...", myname);
	for(i=0; i < n_blocks; i++) {
	  int ib = i * block_size;

	  for(tf=0; tf<tfields; tf++) {
	    bitpix = bitp[tf];

	    for(j=0; j < block_size; j++) {
		int pix_number = ib + j;
		if (nested_flag) {
		    pix_number = convert_nested2ring(pix_number, *nside);
		}
		if (feof(fp_in)) {
		    PR(ER, "\n");
		    PR(ER, "%s: end of file (pos=%ld),", myname,ftell(fp_in));
		    PR(ER, " no data after element %d.\n", ib+j);
		    exit(2);
		}
		if (pix_number > *npixtot) {
		   PR(ER,
		   "\n%s: pixel number (%d) is greater than npixtot (%d)\n",
		   myname, pix_number, *npixtot);
		   exit(1);
		}

		switch(order[tf]) {
		case 'T':
		    (*map_ring)[pix_number] = get_pix_fits(fp_in);
		    break;
		case 'Q':
		    (*q)[pix_number] = get_pix_fits(fp_in);
		    break;
		case 'U':
		    (*u)[pix_number] = get_pix_fits(fp_in);
		    break;
		case 'N':
		    (*counts)[pix_number] = get_pix_fits(fp_in) + .5;
		    break;
		default :
		    get_pix_fits(fp_in);
		    break;
		}
	    }
	  }
	}
if (glesp_comments)
PR(ER,"     O'k\n");

	return;
}

void fhp2tac(map_ring, nside, npixtot, f, nph, nx, nphi, x)
double *map_ring;
int nside, npixtot;
double **f;
int **nph;
int *nx, *nphi;
double **x;
{
     register int i, j, k;
     int ns2 = nside * 2;
     double pi2 = 2 * M_PI;
     int l_n, l_s;
     double fc1, fc2;
     double *y = (double *)0, *a = (double *)0, *w = (double *)0;

     *nphi = nside * 4;
     *nx = *nphi - 1;

     if((*x = (double *)calloc((unsigned)*nx, sizeof **x)) == (double *)0) {
nomem:
	    PR(ER,"%s: no memory for arrays...\n", myname);
	    exit(1);
     }
     if((y = (double *)calloc((unsigned)*nphi, sizeof *y)) == (double *)0)
	    goto nomem;
     if((a = (double *)calloc((unsigned)*nphi, sizeof *a)) == (double *)0)
	    goto nomem;
     if((w = (double *)calloc((unsigned)*nphi, sizeof *w)) == (double *)0)
	    goto nomem;
     if((*nph = (int *)calloc((unsigned)*nx, sizeof **nph)) == (int *)0)
	    goto nomem;
     i = *nx * *nphi;
     if((*f = (double *)calloc((unsigned)i, sizeof **f)) == (double *)0)
	    goto nomem;

     fc1 = 1./3./nside/nside;
     fc2 = 2./3./nside;
     l_n = 0; l_s = npixtot;
     for (i=0; i<ns2; i++) {
	    int i1=i+1, ii, jj, np;
	    int cf = (i+1 < nside ? 1 : (i+2-nside)%2 );
	    double step;
	    j = *nx -1-i;
	    ii = i * *nphi;
	    jj = j * *nphi;

	    if ( i < nside-1 || j > *nx - nside ) {
		  (*x)[i] = 1. - fc1 * i1 * i1;
		  np = i1 * 4;
	    } else {
		  (*x)[i] = (ns2 - i1) * fc2;
		  np = nside * 4;
	    }
	    (*nph)[j] = (*nph)[i] = np;
	    step = pi2/np;

cf = 0;
	    /* coping a map */
	    for(k=0; k < np; k++)  {
		   y[k] = map_ring[l_n+k];        /* map */
		   a[k] = k * step + cf*step/2.;  /* phi */
	    }

//          /* symmetrically reflection (what for ?) */
//          np2 = np/2;
//          for(k=0; k<np2; k++)  {
//                double v = y[k]; y[k] = y[np-1-k]; y[np-1-k] = v;
//          }
	    if ( cf ) {
		    /* used for splint */
		    spline (a, y, np, 2.e30, 2.e30, w);

		    /* interpolation */
		    for(k=0; k<np; k++) {
			 double p = k * step, v;
			 /* used with spline */
			 splint (a, y, w, np, p, &v);

			 (*f)[ii+k] = v;
		    }
	    } else {
		    /* just copying */
		    for(k=0; k<np; k++) (*f)[ii+k] = y[k];
	    }
	    l_n += np;
	    l_s -= np;
/*
PR(ER,"i=%d, j=%d, nph=%d, x=%g, l_n=%d, l_s=%d\n",
i,j,(*nph)[i], (*x)[i], l_n, l_s);
*/
	    if (i < ns2-1) {
		(*x)[j] = -(*x)[i];

		/* coping a map */
		for(k=0; k < np; k++)  y[k] = map_ring[l_s+k];

//              /* symmetricaly reflection */
//              for(k=0; k<np2; k++)  {
//                    double v = y[k]; y[k] = y[np-1-k]; y[np-1-k] = v;
//              }
		if ( cf ) {
			spline (a, y, np, 2.e30, 2.e30, w);

			for(k=0; k<np; k++) {
			     double p = k * step, v;
			     /* used with spline */
			     splint (a, y, w, np, p, &v);

			     (*f)[jj+k] = v;
			}
		} else {
			/* just copying */
			for(k=0; k<np; k++) (*f)[jj+k] = y[k];
		}
	    }

/*
PR(ER,"i=%d, j=%d, x[i]=%g, x[j]=%g, np=%d\n", i, j, (*x)[i], (*x)[j], np);
*/
     }

     free((void *)w);
     free((void *)a);
     free((void *)y);

     return;
}

void read_hp_file(f, nph, nx, nphi, x, filename)
double **f;
int **nph;
int *nx, *nphi;
double **x;
char *filename;
{
     FILE *fp;
     double *map_ring = (double *)0;
     int nside, npixtot;

     *f = (double *)0,
     *x = (double *)0;
     *nph = (int *)0;

     if ((fp = fopen(filename, "r")) == (FILE *)0) {
	   PR(ER,"%s: can not open a file '%s'\n", myname, filename);
	   exit(1);
     }
     /* read mape from Healpix file */
     fhp_in(&map_ring, &nside, &npixtot, fp);

     /* close file */
     fclose(fp);

     /* convert to GLESP-file format */
     fhp2tac(map_ring, nside, npixtot, f, nph, nx, nphi, x);

     /* free memory after map_ring */
     if ( map_ring != (double *)0)   free((void *)map_ring);

     return;
}

void read_bhp_file(f, nph, nx, nphi, x, filename)
double **f;
int **nph;
int *nx, *nphi;
double **x;
char *filename;
{
     FILE *fp;
     double *map_ring = (double *)0;
     int nside, npixtot;
     extern void fh_in(double **, int *, int *, FILE *);

     *f = (double *)0,
     *x = (double *)0;
     *nph = (int *)0;

     if ((fp = fopen(filename, "r")) == (FILE *)0) {
	   PR(ER,"%s: can not open a file '%s'\n", myname, filename);
	   exit(1);
     }
     /* read mape from Healpix file */
     fh_in(&map_ring, &nside, &npixtot, fp);

     /* close file */
     fclose(fp);

     /* convert to GLESP-file format */
     fhp2tac(map_ring, nside, npixtot, f, nph, nx, nphi, x);

     /* free memory after map_ring */
     if ( map_ring != (double *)0)   free((void *)map_ring);

     return;
}

void read_hp_pol(f, q, u, nph, nx, nphi, x, filename)
double **f, **u, **q;
int **nph;
int *nx, *nphi;
double **x;
char *filename;
{
     FILE *fp;
     double *map_ring = (double *)0, *uh = (double *)0, *qh = (double *)0;
     int *ch = (int *)0;
     int nside, npixtot;

     *f = (double *)0,
     *u = (double *)0,
     *q = (double *)0,
     *x = (double *)0;
     *nph = (int *)0;

     if ((fp = fopen(filename, "r")) == (FILE *)0) {
	   PR(ER,"%s: can not open a file '%s'\n", myname, filename);
	   exit(1);
     }
     /* read mape from Healpix file */
     fhpol_in(&map_ring, &qh, &uh, &ch, &nside, &npixtot, fp);

     /* close file */
     fclose(fp);

     /* convert to GLESP-file format */
     fhp2tac(map_ring, nside, npixtot, f, nph, nx, nphi, x);
     fhp2tac(qh, nside, npixtot, q, nph, nx, nphi, x);
     fhp2tac(uh, nside, npixtot, u, nph, nx, nphi, x);

     /* free memory after map_ring */
     if ( map_ring != (double *)0)   free((void *)map_ring);
     if ( qh != (double *)0)   free((void *)qh);
     if ( uh != (double *)0)   free((void *)uh);
     if ( ch != (int *)0)      free((void *)ch);

     return;
}

void fhp_in_nobs(counts, nside, npixtot, fp_in)
int **counts;           /* fields  */
int *nside, *npixtot;
FILE *fp_in;
{
	register int i, j;
	char str[81];
	int tfields=0, first_pix=0, last_pix=0;
	char ttype1[30], tform1[30], tunit1[30];
	char ttype2[30], tform2[30], tunit2[30];
	int bitpix1=0, bitpix2=0;
	double *map_ring = (double *)0;
	int block_size, n_blocks;
	int nested_flag = 0;

	/************************************/
	/* input data to FITS binary Table */
	/************************************/
	/* 1st string analysis */
	for(i=0; i<80; i++) {
		if (feof(fp_in)) break;
		str[i] = fgetc(fp_in);
	}
	str[i] = '\0';
	if (*str == '\0') {
		PR(ER, "%s: empty file with a map.\n", myname);
		exit(1);
	} else {
		if(strncmp(str, "SIMPLE  =", 9) != 0) {
			PR(ER, "%s: a map file is not in FITS.\n", myname);
			exit(1);
		}
	}

	/*------------*/
	/* 1st header */
	/*------------*/

	FHinit_FITS(fp_in, 80);  /* input 1st header */

	/* FITS dataset may contain extensions */
	if (FEVget("EXTEND") == NULL) {
		PR(ER, "%s: a map file is has not XTENSION.\n", myname);
		exit(1);
	}

	FHfree();        /* free memory after 1st header output */

	/*------------*/
	/* 2nd header */
	/*------------*/
	FHinit_FITS(fp_in, 0);   /* input 2nd header */

	/* binary table extension           */
	if (FEVget("XTENSION") == NULL) {
		PR(ER, "%s: a map file has not XTENSION.\n", myname);
		exit(1);
	}
	strcpy(str, FEVget("XTENSION"));
	if (strcmp(str, "BINTABLE") != 0) {
		PR(ER, "%s: a map file has not BINTABLE extension.\n",
			myname);
		exit(1);
	}

	/* number of fields in each row     */
	if (FEVget("TFIELDS") != NULL)
		tfields = atoi(FEVget("TFIELDS"));


	/*------------------------------------------*/
	/*  Sky Map Pixelisation Specific Keywords  */
	/*------------------------------------------*/
	if (FEVget("PIXTYPE") != NULL) {
	    char pixtype[20];
	    strcpy(pixtype, FEVget("PIXTYPE"));
	    if (strcmp(pixtype, "HEALPIX") == 0) {
		  /* Pixel ordering scheme (RING)     */
		  if (FEVget("ORDERING") != NULL)
			  strcpy(str, FEVget("ORDERING"));
		  else
			  str[0] = '\0';
		  if (strcmp(str, "RING") != 0) {
		      if (strcmp(str, "NESTED") != 0) {
			  PR(ER, "%s: unknown ordering of pixels\n",myname);
		      }
		  }
		  nested_flag = 1; /* flag to re-order pixels */
	    } else {
		PR(ER,
		   "%s: procedure 'fhp_in' works only with HEALPix data.\n",
		   myname);
		exit(1);
	    }
	}  else {
	    PR(ER,
	       "%s: uknown pixel type of data (no keyword PIXTYPE).\n",
		   myname);
	    exit(1);
	}
	/* Resolution parameter for HEALPIX */
	if (FEVget("NSIDE") != NULL)   *nside = atoi(FEVget("NSIDE"));

	/* First pixel # (0 based)          */
	if (FEVget("FIRSTPIX") != NULL) first_pix = atoi(FEVget("FIRSTPIX"));
	/* Last pixel # (0 based)           */
	if (FEVget("LASTPIX") != NULL)  last_pix = atoi(FEVget("LASTPIX"));
	*npixtot = last_pix - first_pix + 1;

	/*------------------------------------*/
	/* Data Description Specific Keywords */
	/*------------------------------------*/
	/* Temperature map */
	if (FEVget("TTYPE1") != NULL)
		strcpy(ttype1, FEVget("TTYPE1"));

	/* data format of field: */
	if (FEVget("TFORM1") != NULL)
		strcpy(tform1, FEVget("TFORM1"));

	bitpix = -32;

	i = strlen(tform1) - 1;
	switch(tform1[i]) {
	case 'E':
	case 'F':
		bitpix = -32;
		break;
	case 'L':
	case 'J':
		bitpix = 32;
		break;
	case 'D':
		bitpix = 64;
		break;
	case 'I':
		bitpix = 16;
		break;
	}
	bitpix1 = bitpix;
	if (i == 0)
	    block_size =  1;
	else
	    block_size =  atoi(tform1);

if (glesp_comments) {
PR(ER, "input data: bitpix=%d, npixtot=%d, nside=%d, block_size=%d\n",
bitpix, *npixtot, *nside, block_size);
}
	/* map unit */
	if (FEVget("TUNIT1") != NULL)
		strcpy(tunit1, FEVget("TUNIT1"));

	if (tfields == 2) {
	    int k;
	    if((*counts = (int *)calloc(*npixtot, sizeof **counts)) ==
	       (int *)0)
		    goto nomem;

	    if (FEVget("TTYPE2") != NULL)
		    strcpy(ttype2, FEVget("TTYPE2"));

	    /* data format of field: */
	    if (FEVget("TFORM2") != NULL)
		    strcpy(tform2, FEVget("TFORM2"));

	    if (FEVget("TUNIT2") != NULL)
		    strcpy(tunit2, FEVget("TUNIT2"));

	    bitpix = -32;

	    i = strlen(tform2) - 1;
	    switch(tform2[i]) {
	    case 'E':
	    case 'F':
		    bitpix = -32;
		    break;
	    case 'L':
	    case 'J':
		    bitpix = 32;
		    break;
	    case 'D':
		    bitpix = 64;
		    break;
	    case 'I':
		    bitpix = 16;
		    break;
	    case 'B':
		    bitpix = 8;
		    break;
	    }
	    bitpix2 = bitpix;
	    k = atoi(tform2);
	    if (k == 0)  k = 1;
	    if (block_size !=  k) {
		PR(ER, "%s: I can not read these data because ", myname);
		PR(ER, "of different blocks for fields...\n");
	    }
	} else {
	    PR(ER, "%s: sorry, no second field in a data file\n", myname);
	    exit(1);
	}

	FHfree();  /* free memory after 2nd header */

	/* find memory */
	if ((map_ring = (double *)calloc(*npixtot, sizeof *map_ring)) ==
	     (double *)0) {
nomem:
		PR(ER,"%s: no memory for temporal map array...\n", myname);
		exit(1);
	}

	/*********************/
	/* input binary data */
	/*********************/
	n_blocks = *npixtot / block_size;

if (glesp_comments)
PR(ER,"%s: Read map...", myname);
	for(i=0; i < n_blocks; i++) {
	    int ib = i * block_size;
	    bitpix = bitpix1;
	    for(j=0; j < block_size; j++) {
		int pix_number = ib + j;
		if (nested_flag) {
			pix_number = convert_nested2ring(pix_number, *nside);
		}
		if (feof(fp_in)) {
			PR(ER, "%s: no data after %d-th element.\n",
				myname, i);
			break;
		}
		map_ring[pix_number] = get_pix_fits(fp_in);
if (glesp_comments)
if (!((ib+j)%5000)) PR(ER,"%8d\b\b\b\b\b\b\b\b", ib+j);
	    }
	    if (bitpix2) {
		bitpix = bitpix2;
		for(j=0; j < block_size; j++) {
		    int pix_number = ib + j;
		    if (nested_flag) {
			pix_number = convert_nested2ring(pix_number, *nside);
		    }
		    if (feof(fp_in)) {
			PR(ER, "%s: no data after %d-th element.\n",
				myname, i);
			break;
		    }
		    (*counts)[pix_number] = (int)(get_pix_fits(fp_in) + 0.5);
		}
	    }

	}
	if (map_ring != (double *)0)  free((void *)map_ring);
	return;
}

void read_hp_file_nobs(counts, nph, nx, nphi, x, filename)
int **counts;
int **nph;
int *nx, *nphi;
double **x;
char *filename;
{
     register int i, j, k;
     FILE *fp;
     int nside, npixtot;
     double *a = (double *)0;
     int *y = (int *)0, *cn = (int *)0;
     int l_n, l_s;
     double fc1, fc2, pi2 = 2 * M_PI;
     int ns2;

     *x = (double *)0;
     *nph = (int *)0;

     if ((fp = fopen(filename, "r")) == (FILE *)0) {
	   PR(ER,"%s: can not open a file '%s'\n", myname, filename);
	   exit(1);
     }
     /* read mape from Healpix file */
     fhp_in_nobs(counts, &nside, &npixtot, fp);
     ns2 = nside * 2;

     /* close file */
     fclose(fp);

     /* convert to tac2cmb-file format */
     *nphi = nside * 4;
     *nx = *nphi - 1;

     if((*x = (double *)calloc((unsigned)*nx, sizeof **x)) == (double *)0) {
nomem:
	    PR(ER,"%s: no memory for arrays...\n", myname);
	    exit(1);
     }
     if((*nph = (int *)calloc((unsigned)*nx, sizeof **nph)) == (int *)0)
	    goto nomem;
     if((cn = (int *)calloc((unsigned)*nx * *nphi,sizeof *cn))== (int *)0)
	    goto nomem;

     if((y = (int *)calloc((unsigned)*nphi, sizeof *y)) == (int *)0)
	    goto nomem;
     if((a = (double *)calloc((unsigned)*nphi, sizeof *a)) == (double *)0)
	    goto nomem;

     fc1 = 1./3./nside/nside;
     fc2 = 2./3./nside;
     l_n = 0;
     l_s = npixtot;

     for (i=0; i<ns2; i++) {
	    int i1=i+1, ii, jj, np;
	    int cf = (i+1 < nside  || i+1 > *nx-nside ? 1 : (i+2-nside)%2 );
	    double step;
	    j = *nx -1-i;
	    ii = i * *nphi;
	    jj = j * *nphi;

	    if ( i < nside-1 || j > *nx - nside ) {
		  (*x)[i] = 1. - fc1 * i1 * i1;
		  np = i1 * 4;
	    } else {
		  (*x)[i] = (ns2 - i1) * fc2;
		  np = nside * 4;
	    }
	    (*nph)[j] = (*nph)[i] = np;
	    step = pi2/np;

	    /* coping a map */
	    for(k=0; k < np; k++)  {
		   y[k] = (*counts)[l_n+k];       /* map */
		   a[k] = k * step + cf*step/2.;  /* phi */
	    }

//            /* symmetricaly reflection */
//            np2 = np/2;
//            for(k=0; k<np2; k++)  {
//                    int v = y[k]; y[k] = y[np-1-k]; y[np-1-k] = v;
//            }
	    /* just copying */
	    for(k=0; k<np; k++) cn[ii+k] = y[k];

	    l_n += np;
	    l_s -= np;
	    if (i < ns2-1) {
		(*x)[j] = -(*x)[i];

		/* coping a map */
		for(k=0; k < np; k++)  y[k] = (*counts)[l_s+k];

//                /* symmetrical reflection */
//                for(k=0; k<np2; k++)  {
//                      int v = y[k]; y[k] = y[np-1-k]; y[np-1-k] = v;
//                }

		/* just copying */
		for(k=0; k<np; k++) cn[jj+k] = y[k];
	    }
     }
     free((void *)y);
     free((void *)a);
     free((void *)(*counts));

     *counts = cn;

     return;
}
void hp_arrays(x, offset, nph, nside)
double *x, *offset;
int *nph;
int nside;
{
     register int i,j;
     int nx, nphi, ns2;
     double fc1, fc2, pi2 = 2 * M_PI;

     fc1 = 1./3./nside/nside;
     fc2 = 2./3./nside;

     ns2  = nside * 2;
     nphi = nside * 4;
     nx   = nphi - 1;

     for (i=0; i<ns2; i++) {
	    int i1=i+1, np;
	    int cf = (i+1 < nside  || i+1 > nx-nside ? 1 : (i+2-nside)%2 );
	    double step;

	    j = nx -1-i;

	    if ( i < nside-1 || j > nx - nside ) {
		  x[i] = 1. - fc1 * i1 * i1;
		  np = i1 * 4;
	    } else {
		  x[i] = (ns2 - i1) * fc2;
		  np = nside * 4;
	    }
	    nph[j] = nph[i] = np;
	    step = pi2/np;
	    offset[i] = offset[j] = cf*step/2.;  /* phi */

	    if (i < ns2-1)  x[j] = -x[i];
     }
     return;
}

extern int hang2n(double, double, double *, int, int, int *, double *);

void read_mask(counts, nph, nx, nphi, x, filename, nside)
int **counts;
int **nph;
int *nx, *nphi;
double **x;
char *filename;
int nside;
{
    register int counter = 0;
    int mask_ffile_flag = 0;
    FILE *fp;
    char str[161];
    double *offs = (double *)0;

    /* open file */
    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
    }

    /* strings count */
    while (fgets(str, 160, fp) != NULL) {
	double theta, phi;
	int cn, n;

	/* check if this file is F-file with masks */
	if (!counter && strncmp(str, "SIMPLE", 6) == 0) {
	    mask_ffile_flag = 1;
	    break;
	}
	/* else allocate memory */
	if (!counter) {
	    unsigned int nxp;

	    *nx = 4*nside - 1;
	    *nphi = nside * 4;
	    nxp = *nx * *nphi;
	    if((*counts = (int *)calloc(nxp, sizeof **counts)) == (int *)0) {
nomem:
		 PR(ER, "%s: no memory for array...\n", myname);
		 exit(1);
	    }
	    if((*nph = (int *)calloc(*nx, sizeof **nph)) == (int *)0)
		 goto nomem;
	    if((*x = (double *)calloc(*nx, sizeof **x)) == (double *)0)
		 goto nomem;
	    if((offs= (double *)calloc(*nx, sizeof *offs)) == (double *)0)
		 goto nomem;

	    /* set hp-arrays corresponding to nside */
	    hp_arrays(*x, offs, *nph, nside);
	}

	if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
	    continue;

	sscanf(str, "%lf%lf%d", &theta, &phi, &cn);

	/* convert (theta, phi) -> number */
	n = hang2n(theta, phi, *x, *nx, *nphi, *nph, offs);

	(*counts)[n] = cn;

	counter++;
    }
    if (offs != (double *)0)  free((void *)offs);
    fclose(fp);

    /* if FITS file */
    if ( mask_ffile_flag ) {
	int nside_mask;
	read_hp_file_nobs(counts, nph, nx, nphi, x, filename);
	nside_mask = (*nx+1) / 4;
	if(nside_mask != nside) {
	   PR(ER,
	    "%s: parameter NSIDE is different in datafile and mask-file...\n",
		myname);
	   exit(1);
	}
    }

    return;
}

void in_hpix_map(f, x, offs, nx, np, nph, filename)
double **f, **x, **offs;
int *nx, *np, **nph;
char *filename;
{
    register int i, j, n=0;
    int nside=0, npixtot=0;
    double *map_ring = (double *)0;
    FILE *fp = (FILE *)0;

    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open Hpix-file '%s'.\n", myname, filename);
	exit(1);
    }
    /* read Hpix-map with mask */
    fhp_in(&map_ring, &nside, &npixtot, fp);
    if (fp != (FILE *)0) fclose(fp);

    *np = nside * 4;
    *nx = *np - 1;

    if ((*offs=(double*)calloc((unsigned)*nx, sizeof **offs))==(double*)0) {
nomem_e:
	PR(ER, "%s: no memory for mask array...\n", myname);
	exit(1);
    }
    if ((*x = (double*)calloc((unsigned)*nx, sizeof **x)) == (double*)0)
	goto nomem_e;
    if ((*nph = (int *)calloc((unsigned)*nx, sizeof **nph)) == (int *)0)
	goto nomem_e;
    if ((*f =(double*)calloc((unsigned)*np * *nx, sizeof **f))==(double*)0)
	goto nomem_e;

    /* calculate arrays: x_hp & nph_hp (again), offs (new) */
    hp_arrays(*x, *offs, *nph, nside);

    for(i=0; i< *nx; i++) {
	int ii = i * *np;
	for(j=0; j < (*nph)[i]; j++) {
	    (*f)[ii+j] = map_ring[n];
	    n++;
	}
    }
    if (map_ring != (double *)0)  free((void *)map_ring);

    return;
}

void in_hpix_3map(f, q, u, x, offs, nx, np, nph, filename)
double **f, **q, **u, **x, **offs;
int *nx, *np, **nph;
char *filename;
{
    register int i, j, n=0;
    int nside=0, npixtot=0;
    double *map_ring = (double *)0, *qh = (double *)0, *uh = (double *)0;
    int *ch = (int *)0;
    FILE *fp = (FILE *)0;

    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open Hpix-file '%s'.\n", myname, filename);
	exit(1);
    }
    /* read Hpix-map with mask */
    // fhp_in(&map_ring, &nside, &npixtot, fp);
    fhpol_in(&map_ring, &qh, &uh, &ch, &nside, &npixtot, fp);
    if (fp != (FILE *)0) fclose(fp);

    *np = nside * 4;
    *nx = *np - 1;

    if ((*offs=(double*)calloc((unsigned)*nx, sizeof **offs))==(double*)0) {
nomem_e:
	PR(ER, "%s: no memory for mask array...\n", myname);
	exit(1);
    }
    if ((*x = (double*)calloc((unsigned)*nx, sizeof **x)) == (double*)0)
	goto nomem_e;
    if ((*nph = (int *)calloc((unsigned)*nx, sizeof **nph)) == (int *)0)
	goto nomem_e;
    if ((*f =(double*)calloc((unsigned)*np * *nx, sizeof **f))==(double*)0)
	goto nomem_e;
    if ((*q =(double*)calloc((unsigned)*np * *nx, sizeof **q))==(double*)0)
	goto nomem_e;
    if ((*u =(double*)calloc((unsigned)*np * *nx, sizeof **u))==(double*)0)
	goto nomem_e;

    /* calculate arrays: x_hp & nph_hp (again), offs (new) */
    hp_arrays(*x, *offs, *nph, nside);

    for(i=0; i < *nx; i++) {
	int ii = i * *np;
	for(j=0; j< (*nph)[i]; j++) {
	    (*f)[ii+j] = map_ring[n];
	    (*q)[ii+j] = qh[n];
	    (*u)[ii+j] = uh[n];
	    n++;
	}
    }
    if (map_ring != (double *)0)  free((void *)map_ring);
    if (qh != (double *)0)  free((void *)qh);
    if (uh != (double *)0)  free((void *)uh);
    if (ch != (int *)0)     free((void *)ch);
    return;
}

void fh_in(map_ring, nside, npixtot, fp_in)
double **map_ring;         /* map_ring(ntheta, nphi) */
int *nside, *npixtot;
FILE *fp_in;
{
	register int i;
	char str[81];
	int first_pix=0, last_pix=0;
	int nested_flag = 0;
	int npix=0;

	/***********************************/
	/* input data to FITS binary Table */
	/***********************************/
	/* 1st string analysis */
	for(i=0; i<80; i++) {
		if (feof(fp_in)) break;
		str[i] = fgetc(fp_in);
	}
	str[i] = '\0';
	if (*str == '\0') {
		PR(ER, "%s: empty file with a map.\n", myname);
		exit(1);
	} else {
		if(strncmp(str, "SIMPLE  =", 9) != 0) {
			PR(ER, "%s: a map file is not in FITS.\n", myname);
			exit(1);
		}
	}

	/*-----------------*/
	/* only one header */
	/*-----------------*/

	FHinit_FITS(fp_in, 80);  /* input one header */

	/*------------------------------------------*/
	/*  Sky Map Pixelisation Specific Keywords  */
	/*------------------------------------------*/
	if (FEVget("PIXTYPE") != NULL) {
	    char pixtype[20];
	    strcpy(pixtype, FEVget("PIXTYPE"));
	    if (strcmp(pixtype, "HEALPIX") == 0) {
		  /* Pixel ordering scheme (RING)     */
		  if (FEVget("ORDERING") != NULL)
			  strcpy(str, FEVget("ORDERING"));
		  else
			  str[0] = '\0';
		  if (strcmp(str, "RING") != 0) {
		      if (strcmp(str, "NESTED") != 0) {
			  PR(ER, "%s: unknown ordering of pixels\n",myname);
		      } else
			    nested_flag = 1; /* flag to re-order pixels */
		  }
	    } else {
		PR(ER,
		   "%s: procedure 'fh_in' works only with HEALPix data.\n",
		   myname);
		exit(1);
	    }
	}  else {
	    PR(ER,
	       "%s: uknown pixel type of data (no keyword PIXTYPE).\n",
		   myname);
	    exit(1);
	}
	/* Resolution parameter for HEALPIX */
	if (FEVget("NSIDE") != NULL)   *nside = atoi(FEVget("NSIDE"));

	/* First pixel # (0 based)          */
	if (FEVget("FIRSTPIX") != NULL) first_pix = atoi(FEVget("FIRSTPIX"));
	/* Last pixel # (0 based)           */
	if (FEVget("LASTPIX") != NULL)  last_pix = atoi(FEVget("LASTPIX"));
	*npixtot = last_pix - first_pix + 1;
	/* N pixels total */
	if (FEVget("NPIX") != NULL)  npix = atoi(FEVget("NPIX"));
	if (npix > *npixtot) *npixtot = npix;

	/*------------------------------------*/
	/* Data Description Specific Keywords */
	/*------------------------------------*/
	bitpix = atoi(FEVget("BITPIX"));

if (glesp_comments)
PR(ER, "input data: bitpix=%d, npixtot=%d, nside=%d\n",
bitpix, *npixtot, *nside);

	FHfree();  /* free memory after header */

	/* find memory */
	if ((*map_ring = (double *)calloc(*npixtot +1, sizeof **map_ring)) ==
	     (double *)0) {
		PR(ER,"%s: no memory for temporal map array...\n", myname);
		exit(1);
	}

	/*********************/
	/* input binary data */
	/*********************/

if (glesp_comments)
PR(ER,"%s: Read map...", myname);
	for(i=0; i < *npixtot; i++) {
		int pix_number = i;
		if (nested_flag) {
			pix_number = convert_nested2ring(pix_number, *nside);
		}
		if (feof(fp_in)) {
			PR(ER, "%s: no data after %d-th element.\n",
				myname, i);
			break;
		}
		if (pix_number > *npixtot) {
		   PR(ER,
		     "\n%s: pixel number (%d) is greater than npixtot (%d)\n",
		     myname, pix_number, *npixtot);
		   exit(1);
		}

		(*map_ring)[pix_number] = get_pix_fits(fp_in);
	}
if (glesp_comments)
PR(ER,"     O'k\n");

	return;
}

