#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "glesp.h"
#include "fits.h"

#define PR              fprintf
#define ER              stderr

extern char tmpcat[161];
extern int  debug_flag;
extern int  nside;
extern double  nside_factor;   /* factor to decrease nside   */

int Cl_max=0;         /* Max multipole l used in map synthesis      */
int alm_rand_seed= -32767; /* Random generator seed */
double fwhm= -1.;          /* FWHM in DEGREES       */

/* output data in HEALPix format */
void fhp_out2p(f, x, nx, nph, nphi, nside, fp_out)
double *f;             /* f(ntheta, nphi) */
int nx, nphi;
int *nph;               /* array of dimension of nphi */
double *x;
int nside;
FILE *fp_out;
{
	int ns2;
	register int i, j, l_n, l_s;
	unsigned int npixtot, nn=0;
	double *map_ring=(double *)0, /* HEALPix ring array       */
	    *fi = (double *)0;
	double t0, dt;
	double pi2 = 2 * M_PI;
	int nt;
	char sdate [50], tmpfile[161];
	char tform [20];
	int nv, nmap, nrows;
	FILE *fp_tmp = (FILE *)0;
	double *cos_theta_hp = (double *)0;
	double fc1, fc2;

	/* calculate 'nside' if it was not input */
	if (nside < 1)  {
		/* ln 2 = M_LN2 = 0.69314718055994530942 */
		i = (int) (log((double)(nphi))/M_LN2  + 0.5);
		for(l_n=1,j=0; j<i; j++)  l_n *= 2;
		nside = l_n / 4;
	}
i=(int)(nside/nside_factor+0.5);
PR(ER, "nside=%d, i=%d\n", nside, i);
if(i==0) {
PR(ER, "\nnside decreasing factor is too small. Nside will be =1.\n\n");
i=1;
}
PR(ER, "nphi=%d, nside_init=%d, lgNphi=%g, nside_decreased=%d, ",
nphi, nside, log((double)(nphi)),
(int)(pow(2., (int)( log(nside/nside_factor)/M_LN2)+1.)+.5));
	nside = nside / nside_factor + 0.5;
	i=pow(2., (int)(log(nside/M_LN2))+1.)+.5;
	nside=i;
if(nside<1) nside=1;
	nt = 4*nside - 1;
	ns2 = nside*2;
	t0 = acos(x[0]);
	dt = (acos(x[nx-1]) - acos(x[0])) / (double)(nt-1);

	npixtot = 12 * nside * nside;
PR(ER, "npixtot=%d\n", npixtot);

if(debug_flag) {
PR(ER, "ith=%g (%g), iith=%g\n",
acos(x[nx-1])-acos(x[0]), (acos(x[1])-acos(x[0]))*(double)(nx-1), dt*(nt-1));
PR(ER, "nside=%d, nt=%d, npixtot=%d, nphi=%d, nx=%d, dt=%g\n",
nside, nt, npixtot, nphi, nx, dt);
}
	/* Healpix cos(theta)-discretezation */
	if ((cos_theta_hp = (double *)calloc(nt, sizeof *cos_theta_hp)) ==
	   (double *)0) {
		PR(ER, "%s: no memory for array of arguments.\n", myname);
		exit(1);
	}

	i = (nt   > nx  ? nt : nx);
	j = (nphi > nside*4 ? nphi : nside*4)+1;
	if (i < 4)  i=4;
	if (j < 4)  j=4;
	nn = i*j;

	/* fill in 'theta_hp'-array */
	fc1 = 1./3./nside/nside;
	fc2 = 2./3./nside;
	for (i=0; i<ns2; i++) {
	    int i1 = i+1;
	    if ( i1 < nside )
		  cos_theta_hp[i] = 1. - fc1 * i1 * i1;
	    else
		  cos_theta_hp[i] = (ns2 - i1) * fc2;

	    j = nt-1-i;
	    if ( j > nt - nside - 1)
		  cos_theta_hp[j] = -cos_theta_hp[i];
	    else
		  cos_theta_hp[j] = -cos_theta_hp[i];
	}

	/* take memory for temporal array */
	j = nphi * (nt < nx ? nx : nt);
	if ((fi = (double *)calloc((unsigned)j, sizeof *fi)) ==
	     (double *)0) {
		PR(ER, "%s: no memory for temporal array (%d elements)\n",
		    myname, nphi*nt);
		exit(1);
	}

	/* transpose matrix:                   */
	/* f(nphi, ntheta) -> fi(ntheta, nphi) */
	for(j=0; j<nphi; j++) {
		register int k = j * nx;
		for(i=0; i<nx; i++) {
			  fi[i*nphi+j] = f[k+i];
		}
	}

	/* temporal array, nt>4 is awaited to have memory */
	i = (nt   > nx ? nt : nx);
	j = (nphi > nside*4 ? nphi : nside*4);
	if (i > j)  j = i;

	if ((map_ring = (double *)calloc(npixtot, sizeof *map_ring)) ==
	     (double *)0) {
		PR(ER,"%s: no memory for temporal map array...\n", myname);
		exit(1);
	}

	/* HEALPix ring pixelization scheme */
	nv = nphi/2;
	for(l_s=npixtot,l_n=i=0; i< ns2; i++) {
	    /* here nph is a number of intervals, like (nphi-1) */
	    register int np = 4 * (i+1 < nside ? i+1 : nside);
	    double dphi = pi2/np, th;
	    int cf = (i+1 < nside ? 1 : (i+2-nside)%2 );
	    double cff = cf * dphi / 2.;

	    /***************************/
	    /* for Northern hemisphere */
	    /***************************/
	    th = acos(cos_theta_hp[i]);
	    for(j=0; j<np; j++) {
		    double phi = j * dphi + cff;
		    int nn = ang2n(th, phi, x, nx, nphi, nph);
		    map_ring[l_n+j] = fi[nn];
	    }
	    l_n += np;

	    /***************************/
	    /* for Southern hemisphere */
	    /***************************/
	    l_s -= np;
	    th = acos(cos_theta_hp[nt-1-i]);
	    for(j=0; j<np; j++) {
		    double phi = j * dphi + cff;
		    int nn = ang2n(th, phi, x, nx, nphi, nph);
		    map_ring[l_s+j] = fi[nn];
	    }
	}
	/* free memory */
	if (fi != (double *)0)  free((void *)fi);

	if (fp_tmp != (FILE *)0) {
		fclose(fp_tmp);
		unlink(tmpfile);
		fp_tmp = (FILE *)0;
	}

	/************************************/
	/* output data to FITS binary Table */
	/************************************/

	/*********************/
	/* prepare FITS head */
	/*********************/
	/* 1st header */
	/*------------*/
	cFEVset("SIMPLE", "T"); FEVexport ("SIMPLE");
	iFEVset("BITPIX", 8);   FEVexport ("BITPIX");
	/* number of data axes */
	iFEVset("NAXIS",  0);   FEVexport ("NAXIS");
	/* FITS dataset may contain extensions */
	cFEVset("EXTEND", "T"); FEVexport ("EXTEND");
	/* read current date */
	take_cur_date(sdate);
	/* date: format 'yyyy-mm-dd' */
	sFEVset("DATE", sdate); FEVexport ("DATE");

	Fits_update(fp_out);    /* output 1st header                   */
	FHfree();               /* free memory after 1st header output */

	/*------------*/
	/* 2nd header */
	/*------------*/
	nv = 1024;
	nmap = 1;
	if (npixtot < nv) {
		nrows = npixtot;
		strcpy(tform, "1E");
		nv = 4; /* width of table in bytes */
	} else {
		nrows = npixtot/nv;
		strcpy(tform, "1024E");
		nv *= 4; /* width of table in bytes */
	}

	/* binary table extension           */
	sFEVset("XTENSION","BINTABLE");  FEVexport ("XTENSION");
	/* 8-bit bytes                      */
	iFEVset("BITPIX",  8);           FEVexport ("BITPIX");
	/* 2-dimensional binary table       */
	iFEVset("NAXIS",   2);           FEVexport ("NAXIS");
	/* width of table in bytes          */
	iFEVset("NAXIS1",  nv);          FEVexport ("NAXIS1");
	/* number of rows in table          */
	iFEVset("NAXIS2",  nrows);       FEVexport ("NAXIS2");
	/* size of special data area        */
	iFEVset("PCOUNT",  0);           FEVexport ("PCOUNT");
	/* one data group (required keyword)*/
	iFEVset("GCOUNT",  1);           FEVexport ("GCOUNT");
	/* number of fields in each row     */
	iFEVset("TFIELDS", nmap);        FEVexport ("TFIELDS");

	/*------------------------------------------*/
	/*  Sky Map Pixelisation Specific Keywords  */
	/*------------------------------------------*/
	sFEVset("PIXTYPE", "HEALPIX");   FEVexport ("PIXTYPE");
	/* Pixel ordering scheme (RING)     */
	sFEVset("ORDERING","RING");      FEVexport ("ORDERING");
	/* Resolution parameter for HEALPIX */
	iFEVset("NSIDE",   nside);       FEVexport ("NSIDE");
	/* First pixel # (0 based)          */
	iFEVset("FIRSTPIX",0);           FEVexport ("FIRSTPIX");
	/* Last pixel # (0 based)           */
	iFEVset("LASTPIX", npixtot-1);   FEVexport ("LASTPIX");

	/*--------------------------------------*/
	/*  Planck Simulation Specific Keywords */
	/*--------------------------------------*/
	sFEVset("EXTNAME","SIMULATION"); FEVexport ("EXTNAME");
	sFEVset("CREATOR","CL2MAP");     FEVexport ("CREATOR");
	/*----------------------------------*/
	/* Software creating the FITS file  */
	/*----------------------------------*/
	if ( Cl_max > 0) {
	     /* Max multipole l used in map synthesis */
	     iFEVset("MAX-LPOL",Cl_max);  FEVexport ("MAX-LPOL");
	}
	if ( alm_rand_seed > -32767) {
	     /* Random generator seed */
	     iFEVset("RANDSEED",alm_rand_seed); FEVexport ("RANDSEED");
	}
	if ( fwhm > 0. ) {
	     /* FWHM in DEGREES */
	     dFEVset("FWHM",    fwhm);    FEVexport ("FWHM");
	}
	/*------------------------------------*/
	/* Data Description Specific Keywords */
	/*------------------------------------*/
	/* Temperature map */
	sFEVset("TTYPE1", "TEMPERATURE"); FEVexport ("TTYPE1");
	/* data format of field: 4-byte REAL */
	sFEVset("TFORM1", tform);         FEVexport ("TFORM1");
	/* map unit */
	sFEVset("TUNIT1", "K");           FEVexport ("TUNIT1");

	Fits_update(fp_out);    /* output 2nd header */

	/**********************/
	/* output binary data */
	/**********************/
	bitpix = -32;
	put_dvec_fits(fp_out, map_ring, npixtot);

	/* free memory */
	free((void *)map_ring);

	return;
}

void fhp_out2p_pol(f, q, u, x, nx, nph, nphi, nside, fp_out)
double *f, *q, *u;      /* f(ntheta, nphi) */
int nx, nphi;
int *nph;               /* array of dimension of nphi */
double *x;
int nside;
FILE *fp_out;
{
	int ns2;
	register int i, j, l_n, l_s;
	unsigned int npixtot, nn=0;
	double *map_ring=(double *)0, /* HEALPix ring array  */
	       *q_buf = (double *)0,
	       *u_buf = (double *)0,
	       *fi_q = (double *)0,
	       *fi_u = (double *)0,
	       *fi = (double *)0;
	double t0, dt;
	double pi2 = 2 * M_PI;
	int nt;
	char sdate [50], tmpfile[161];
	char tform [20];
	int nv, nmap, nrows;
	FILE *fp_tmp = (FILE *)0;
	double *cos_theta_hp = (double *)0;
	double fc1, fc2;

	/* calculate 'nside' if it was not input */
	if (nside < 1)  {
		/* ln 2 = M_LN2 = 0.69314718055994530942 */
		i = (int) (log((double)(nphi))/M_LN2  + 0.5);
		for(l_n=1,j=0; j<i; j++)  l_n *= 2;
		nside = l_n / 4;
	}
i=(int)(nside/nside_factor+0.5);
if(i==0) {
PR(ER, "\nnside decreasing factor is too small. Nside will be =1.\n\n");
i=1;
}
PR(ER, "nphi=%d, nside=%d, lgNphi=%g, nside_decreased=%d, ",
nphi, nside, log((double)(nphi)),
(int)(nside/nside_factor+0.5));
	nside = nside / nside_factor + 0.5;
if(nside<1) nside=1;
	nt = 4*nside - 1;
	ns2 = nside*2;
	t0 = acos(x[0]);
	dt = (acos(x[nx-1]) - acos(x[0])) / (double)(nt-1);

	npixtot = 12 * nside * nside;
PR(ER, "npixtot=%d\n", npixtot);

if(debug_flag) {
PR(ER, "ith=%g (%g), iith=%g\n",
acos(x[nx-1])-acos(x[0]), (acos(x[1])-acos(x[0]))*(double)(nx-1), dt*(nt-1));
PR(ER, "nside=%d, nt=%d, npixtot=%d, nphi=%d, nx=%d, dt=%g\n",
nside, nt, npixtot, nphi, nx, dt);
}
	/* Healpix cos(theta)-discretezation */
	if ((cos_theta_hp = (double *)calloc(nt, sizeof *cos_theta_hp)) ==
	   (double *)0) {
		PR(ER, "%s: no memory for array of arguments.\n", myname);
		exit(1);
	}

	i = (nt   > nx  ? nt : nx);
	j = (nphi > nside*4 ? nphi : nside*4)+1;
	if (i < 4)  i=4;
	if (j < 4)  j=4;
	nn = i*j;

	/* fill in 'theta_hp'-array */
	fc1 = 1./3./nside/nside;
	fc2 = 2./3./nside;
	for (i=0; i<ns2; i++) {
	    int i1 = i+1;
	    if ( i1 < nside )
		  cos_theta_hp[i] = 1. - fc1 * i1 * i1;
	    else
		  cos_theta_hp[i] = (ns2 - i1) * fc2;

	    j = nt-1-i;
	    if ( j > nt - nside - 1)
		  cos_theta_hp[j] = -cos_theta_hp[i];
	    else
		  cos_theta_hp[j] = -cos_theta_hp[i];
	}

	/* take memory for temporal array */
	j = nphi * (nt < nx ? nx : nt);
	if ((fi = (double *)calloc((unsigned)j, sizeof *fi)) ==
	     (double *)0) {
nomem1:
		PR(ER, "%s: no memory for temporal array (%d elements)\n",
		    myname, nphi*nt);
		exit(1);
	}
	if ((fi_q =(double*)calloc((unsigned)j, sizeof *fi_q)) == (double*)0)
	    goto nomem1;
	if ((fi_u =(double*)calloc((unsigned)j, sizeof *fi_u)) == (double*)0)
	    goto nomem1;

	/* transpose matrix:                   */
	/* f(nphi, ntheta) -> fi(ntheta, nphi) */
	for(j=0; j<nphi; j++) {
		register int k = j * nx;
		for(i=0; i<nx; i++) {
			fi[i*nphi+j]   = f[k+i];
			fi_q[i*nphi+j] = q[k+i];
			fi_u[i*nphi+j] = u[k+i];
		}
	}

	/* temporal array, nt>4 is awaited to have memory */
	i = (nt   > nx ? nt : nx);
	j = (nphi > nside*4 ? nphi : nside*4);
	if (i > j)  j = i;

	if ((map_ring = (double *)calloc(npixtot, sizeof *map_ring)) ==
	     (double *)0) {
nomem2:
		PR(ER,"%s: no memory for temporal map array...\n", myname);
		exit(1);
	}
	if ((q_buf = (double *)calloc(npixtot, sizeof *q_buf))==(double *)0)
		goto nomem2;
	if ((u_buf = (double *)calloc(npixtot, sizeof *u_buf))==(double *)0)
		goto nomem2;

	/* HEALPix ring pixelization scheme */
	nv = nphi/2;
	for(l_s=npixtot,l_n=i=0; i< ns2; i++) {
	    /* here nph is a number of intervals, like (nphi-1) */
	    register int np = 4 * (i+1 < nside ? i+1 : nside);
	    double dphi = pi2/np, th;
	    int cf = (i+1 < nside ? 1 : (i+2-nside)%2 );
	    double cff = cf * dphi / 2.;

	    /***************************/
	    /* for Northern hemisphere */
	    /***************************/
	    th = acos(cos_theta_hp[i]);
	    for(j=0; j<np; j++) {
		    double phi = j * dphi + cff;
		    int nn = ang2n(th, phi, x, nx, nphi, nph);
		    map_ring[l_n+j] = fi[nn];
		    q_buf[l_n+j] = fi_q[nn];
		    u_buf[l_n+j] = fi_u[nn];
	    }
	    l_n += np;

	    /***************************/
	    /* for Southern hemisphere */
	    /***************************/
	    l_s -= np;
	    th = acos(cos_theta_hp[nt-1-i]);
	    for(j=0; j<np; j++) {
		    double phi = j * dphi + cff;
		    int nn = ang2n(th, phi, x, nx, nphi, nph);
		    map_ring[l_s+j] = fi[nn];
		    q_buf[l_s+j] = fi_q[nn];
		    u_buf[l_s+j] = fi_u[nn];
	    }
	}
	/* free memory */
	if (fi != (double *)0)    free((void *)fi);
	if (fi_q != (double *)0)  free((void *)fi_q);
	if (fi_u != (double *)0)  free((void *)fi_u);

	if (fp_tmp != (FILE *)0) {
		fclose(fp_tmp);
		unlink(tmpfile);
		fp_tmp = (FILE *)0;
	}

	/************************************/
	/* output data to FITS binary Table */
	/************************************/

	/*********************/
	/* prepare FITS head */
	/*********************/
	/* 1st header */
	/*------------*/
	cFEVset("SIMPLE", "T"); FEVexport ("SIMPLE");
	iFEVset("BITPIX", 8);   FEVexport ("BITPIX");
	/* number of data axes */
	iFEVset("NAXIS",  0);   FEVexport ("NAXIS");
	/* FITS dataset may contain extensions */
	cFEVset("EXTEND", "T"); FEVexport ("EXTEND");
	/* read current date */
	take_cur_date(sdate);
	/* date: format 'yyyy-mm-dd' */
	sFEVset("DATE", sdate); FEVexport ("DATE");

	Fits_update(fp_out);    /* output 1st header                   */
	FHfree();               /* free memory after 1st header output */

	/*------------*/
	/* 2nd header */
	/*------------*/
	nv = 1024;
	nmap = 3;
	if (npixtot < nv) {
		nrows = npixtot;
		strcpy(tform, "1E");
		nv = 4; /* width of table in bytes */
	} else {
		nrows = npixtot/nv;
		strcpy(tform, "1024E");
		nv *= 4; /* width of table in bytes */
	}

	/* binary table extension           */
	sFEVset("XTENSION","BINTABLE");  FEVexport ("XTENSION");
	/* 8-bit bytes                      */
	iFEVset("BITPIX",  8);           FEVexport ("BITPIX");
	/* 2-dimensional binary table       */
	iFEVset("NAXIS",   2);           FEVexport ("NAXIS");
	/* width of table in bytes          */
	iFEVset("NAXIS1",  nv * 3);      FEVexport ("NAXIS1");
	/* number of rows in table          */
	iFEVset("NAXIS2",  nrows);       FEVexport ("NAXIS2");
	/* size of special data area        */
	iFEVset("PCOUNT",  0);           FEVexport ("PCOUNT");
	/* one data group (required keyword)*/
	iFEVset("GCOUNT",  1);           FEVexport ("GCOUNT");
	/* number of fields in each row     */
	iFEVset("TFIELDS", nmap);        FEVexport ("TFIELDS");

	/*------------------------------------------*/
	/*  Sky Map Pixelisation Specific Keywords  */
	/*------------------------------------------*/
	sFEVset("PIXTYPE", "HEALPIX");   FEVexport ("PIXTYPE");
	/* Pixel ordering scheme (RING)     */
	sFEVset("ORDERING","RING");      FEVexport ("ORDERING");
	/* Resolution parameter for HEALPIX */
	iFEVset("NSIDE",   nside);       FEVexport ("NSIDE");
	/* First pixel # (0 based)          */
	iFEVset("FIRSTPIX",0);           FEVexport ("FIRSTPIX");
	/* Last pixel # (0 based)           */
	iFEVset("LASTPIX", npixtot-1);   FEVexport ("LASTPIX");

	/*--------------------------------------*/
	/*  Planck Simulation Specific Keywords */
	/*--------------------------------------*/
	sFEVset("EXTNAME","SIMULATION"); FEVexport ("EXTNAME");
	sFEVset("CREATOR","F2MAP");      FEVexport ("CREATOR");
	/*----------------------------------*/
	/* Software creating the FITS file  */
	/*----------------------------------*/
	if ( Cl_max > 0) {
	     /* Max multipole l used in map synthesis */
	     iFEVset("MAX-LPOL",Cl_max);  FEVexport ("MAX-LPOL");
	}
	if ( alm_rand_seed > -32767) {
	     /* Random generator seed */
	     iFEVset("RANDSEED",alm_rand_seed); FEVexport ("RANDSEED");
	}
	if ( fwhm > 0. ) {
	     /* FWHM in DEGREES */
	     dFEVset("FWHM",    fwhm);    FEVexport ("FWHM");
	}
	/*------------------------------------*/
	/* Data Description Specific Keywords */
	/*------------------------------------*/
	/* Temperature map */
	sFEVset("TTYPE1", "TEMPERATURE"); FEVexport ("TTYPE1");
	/* data format of field: 4-byte REAL */
	sFEVset("TFORM1", tform);         FEVexport ("TFORM1");
	/* map unit */
	sFEVset("TUNIT1", "K");           FEVexport ("TUNIT1");

	sFEVset("TTYPE2", "Q-POLARISATION"); FEVexport ("TTYPE2");
	/* data format of field: 4-byte REAL */
	sFEVset("TFORM2", tform);         FEVexport ("TFORM2");
	/* map unit */
	sFEVset("TUNIT2", "unknown");     FEVexport ("TUNIT2");

	sFEVset("TTYPE3", "U-POLARISATION"); FEVexport ("TTYPE3");
	/* data format of field: 4-byte REAL */
	sFEVset("TFORM3", tform);         FEVexport ("TFORM3");
	/* map unit */
	sFEVset("TUNIT3", "unknown");     FEVexport ("TUNIT3");

	Fits_update(fp_out);    /* output 2nd header */

	/**********************/
	/* output binary data */
	/**********************/
	bitpix = -32;
	put_dvec_fits_pol(fp_out, map_ring, q_buf, u_buf, npixtot, nv);

	/* free memory */
	free((void *)map_ring);
	free((void *)q_buf);
	free((void *)u_buf);

	return;
}

/* output data in HEALPix format               */
/* with interpolation over pixelization scheme */
void fhp_out2(f, ntheta, nph, nphi, theta, phi, nside, fp_out, flag_int)
double **f;             /* f(ntheta, nphi) */
int ntheta, nphi;
int *nph;               /* array of dimension of nphi */
double *theta, *phi;
int nside;
FILE *fp_out;
int flag_int;
{
	int ns2;
	register int i, j, l_n, l_s;
	unsigned int npixtot, nn=0;
	double *map_ring=(double *)0, /* HEALPix ring array       */
	    *fi = (double *)0,
	    *fi1= (double *)0; /* interpolated map with 'nside' phi steps */
	double t0, dt;
	double pi2 = 2 * M_PI;
	double *phi_w = phi;
	int nt;
	double *work = (double *)0;
	char sdate [50], tmpfile[161];
	char tform [20];
	int nv, nmap, nrows;
	int flag_phi_set = 0;
	FILE *fp_tmp = (FILE *)0;
	double p0 = 0.0, phi2 = M_PI * 2;
	double *theta_hp = (double *)0;
	double fc1, fc2;

	/* calculate 'nside' if it was not input */
	if (nside < 1)  {
		/* ln 2 = M_LN2 = 0.69314718055994530942 */
		i = (int) (log((double)(nphi))/M_LN2  + 0.5);
		for(l_n=1,j=0; j<i; j++)  l_n *= 2;
		nside = l_n / 4;
	}
PR(ER, "nphi=%d, nside=%d, lgNphi=%g, nside_decreased=%d, ",
nphi, nside, log((double)(nphi)), (int)(nside/nside_factor+0.5));
	nside = nside / nside_factor + 0.5;
	nt = 4*nside - 1;
	ns2 = nside*2;
	t0 = theta[0];
	dt = (theta[ntheta-1] - theta[0]) / (double)(nt-1);

	npixtot = 12 * nside * nside;
PR(ER, "npixtot=%d\n", npixtot);

if(debug_flag) {
PR(ER, "ith=%g (%g), iith=%g\n",
theta[ntheta-1]-theta[0], (theta[1]-theta[0])*(double)(ntheta-1), dt*(nt-1));
PR(ER, "nside=%d, nt=%d, npixtot=%d, nphi=%d, ntheta=%d, dt=%g\n",
nside, nt, npixtot, nphi, ntheta, dt);
}
	/* Healpix cos(theta)-discretezation */
	if ((theta_hp = (double *)calloc(nt, sizeof *theta_hp)) ==
	   (double *)0) {
		PR(ER, "%s: no memory for array of arguments.\n", myname);
		exit(1);
	}

	/* fill in 'theta_hp'-array */
	fc1 = 1./3./nside/nside;
	fc2 = 2./3./nside;
	for (i=0; i<ns2; i++) {
	    int i1 = i+1;
	    if ( i1 < nside )
		  theta_hp[i] = 1. - fc1 * i1 * i1;
	    else
		  theta_hp[i] = (ns2 - i1) * fc2;

	    j = nt-1-i;
	    if ( j > nt - nside - 1)
		  theta_hp[j] = -theta_hp[i];
	    else
		  theta_hp[j] = -theta_hp[i];
	}

	i = (nt   > ntheta  ? nt   : ntheta);
	j = (nphi > nside*4 ? nphi : nside*4)+1;
	if (i < 4)  i=4;
	if (j < 4)  j=4;
	nn = i*j;
	fi1 = (double *)calloc(nn, sizeof *fi1);

	/*--------------------------------------------*/
	/* spline interpolation (modified input):     */
	/*   non-uniform step by X ---> uniform step  */
	/*--------------------------------------------*/
	/* f(nphi, ntheta) -> fi1(nphi, nt=4*nside-1) */
	/* but here both steps are uniform            */

	if (dt < 0) {
	     /*   condition_flag = 0;   */
		for(i=0; i<ntheta; i++) theta[i] = -theta[i];
/*
		for(i=0; i<nt; i++)     theta_hp[i] = -theta_hp[i];
*/
		dt = -dt;
		t0 = -t0;
	}
       /*--------------------------------------------------
	*
	*    ntheta = n_x_in = nmx_in          nt = n_x_out = nmx_out
	*
	*        ******                           ********
	*        ******                           ********
	* nph(i) ******           --------->      ********
	* = nya  *****                            ********  nphi
	*        *****                            ********
	* < nphi  ***                             ********
	*          *                              ********
	*
	*         f()                               fi1()
	*
	*/
	if ( fi1 != (double *)0 ) {
	    if (flag_int) {
		copy2sm_Nnu_nu(*f, ntheta, ntheta, fi1, nt, nt,
			theta, theta_hp, p0, phi2, nph, nphi);
	    } else {
		intrpl2sm_Nnu_nu(*f, ntheta, ntheta, fi1, nt, nt,
			theta, theta_hp, p0, phi2, nph, nphi);
	    }
	    free((void *)*f);
	    *f = (double *)0;

	} else {
	    /* save array to file */
	    PR(ER,"%s: no memory (%d Mb) for map square interpolation...\n",
		    myname, nn*8/1024000);
	    PR(ER, "%s: trying to use disk space for temporal array ",
		    myname);
	    PR(ER, "of same size...");

	    if ((fi = (double *)calloc((unsigned)nphi, sizeof *fi))
		    == (double *)0) {
		    PR(ER,"\n%s: Sorry. No memory at all...\n",
			    myname);
		    fclose(fp_tmp);
		    unlink(tmpfile);
		    exit(1);
	    }
	    /* open temporal file and save array to disk */
	    sprintf(tmpfile, "%s/fmXXXXXX", tmpcat);
	    if((fp_tmp=fopen(mktemp(tmpfile), "w+")) == (FILE *)0) {
		    PR(ER,"%s: can not create temporal file '%s'\n",
			 myname, tmpfile);
		    exit(1);
	    }
	    /* transpose matrix when saving */
	    /* in_data(n_y , n_x_in) -> in_data(n_x_in, ny) */
	    /*--------------------------------------------------
	     *
	     *    ntheta = n_x_in = nmx_in
	     *                                    nphi(i) = nya(i) <= n_y
	     *       ******
	     *       ******                       ***
	     *  nphi ******           --------->  ******   ntheta
	     * = n_y *****                        *******
	     *       *****                        ******
	     *        ***                         *****
	     *         *
	     *        f()                         (FILE *)
	     */
	    for(i=0; i<ntheta; i++) {
	       for(j=0; j<nphi; j++) fi[j] = (*f)[j*ntheta+i];

	       if (fwrite(fi,  sizeof *fi,  nphi, fp_tmp) != nphi) {
		    PR(ER,"%s: error during writing...\n", myname);
		    fclose(fp_tmp);
		    unlink(tmpfile);
		    exit(1);
	       }
	    }
	    fseek(fp_tmp, 0l, 0);

	    /* free memory  */
	    free((void *)*f);
	    free((void *)fi);
	    *f = (double *)0;
	    fi = (double *)0;

	    /* find memory again */
	    fi1 = (double *)calloc(nn, sizeof *fi1);
	    if ( fi1 == (double *)0 ) {
		    fclose(fp_tmp);
		    unlink(tmpfile);
		    PR(ER, "\n");
		    goto nomem;
	    }

	    /* make interpolation */
/*
	    intrpl2Fsm_Nnu_u(fp_tmp, ntheta, ntheta, fi1, nt, nt,
		    theta, t0, dt, p0, phi2, nph, nphi);
*/
	    intrpl2Fsm_Nnu_nu(fp_tmp, ntheta, ntheta, fi1, nt, nt,
		    theta, theta_hp, p0, phi2, nph, nphi);

	    fclose(fp_tmp);
	    unlink(tmpfile);
	}

	if (theta_hp != (double *)0)   free((void *)theta_hp);

	/* set 'phi' array */
	if (phi_w == (double *)0) {
		double dphi = phi2 / nphi;
		flag_phi_set = 1;
		/* take memory for phi_w */
		phi_w = (double *)calloc(nphi, sizeof *phi_w);
		if (phi_w == (double *)0) {
			PR(ER,"%s: no memory for output arrays.\n", myname);
			exit(1);
		}
		for(i=0; i<nphi; i++)  phi_w[i] = p0 + i * dphi;
	}

if(debug_flag && fi1 != (double *)0)
{
int flag_grid = -1;  /* homogeneous grid */
int transposing = 0; /* no transposing   */
char tmpfile[161];
FILE *fp;
double *x=(double *)calloc(nt, sizeof *x);

strcpy(tmpfile, "er.f");
fp=fopen(tmpfile, "w");
PR(ER,"ATTENTION: temporal file '%s' is created in current directory.\n",
      tmpfile);
for(i=0; i<nt; i++) x[i] = t0 + dt * i;

if (fi == (double *)0) {
if ((fi = (double *)calloc(nn, sizeof *fi))==(double *)0)
	goto nomem;
}
j=nphi*nt;
for(i=0; i<j; i++) fi[i] = fi1[i];

/* output in basic F-format           */
ff_write_map(fi, nphi, x, nt, phi_w, nphi, flag_grid,
	transposing, fp);

fclose(fp);
free((void *)x);
free((void *)fi);
fi = (double *)0;
}

/*
	if (condition_flag == 0) {
		condition_flag = 1;
		for(i=0; i<nt; i++)     theta_hp[i] = -theta_hp[i];
//              for(i=0; i<ntheta; i++) theta[i] = -theta[i];
		dt = -dt;
		t0 = -t0;
	}
*/

	if ((fi = (double *)calloc(nn, sizeof *fi ))==(double *)0) {
	   PR(ER,"%s: no memory (%d Mb) for map recalculation...\n",
		   myname, nn*8/1024000);
	   PR(ER, "%s: trying to use disk space for temporal array ",
		   myname);
	   PR(ER, "of same size...");

	   /* open temporal file and save array to disk */
	   sprintf(tmpfile, "%s/fmXXXXXX", tmpcat);
	   if((fp_tmp=fopen(mktemp(tmpfile), "w+")) == (FILE *)0) {
		   PR(ER,"%s: can not create temporal file '%s'\n",
			myname, tmpfile);
		   exit(1);
	   }
	   if ((fi = (double *)calloc((unsigned)nphi, sizeof *fi))
		   == (double *)0) {
		   PR(ER,"\n%s: Sorry. No memory at all...\n",
			   myname);
		   fclose(fp_tmp);
		   unlink(tmpfile);
		   exit(1);
	   } else {
		/* transpose matrix:             */
		/* fi1(nphi, nt) -> fi(nt, nphi) */
		/* and write to disk             */
		for(j=0; j<nt; j++) {
		   for(i=0; i<nphi; i++) {
			register int ii = i * nt;
			fi[i] = fi1[ii+j];
		   }
		   if (fwrite(fi,  sizeof *fi,  nphi, fp_tmp) != nphi) {
			PR(ER,"%s: error during writing...\n", myname);
			fclose(fp_tmp);
			unlink(tmpfile);
			exit(1);
		   }
		}
		free((void *)fi);
		fi = (double *)0;
	   }
	   free((void *)fi1);

	   PR(ER, "Done.\n");

	   /* temporal array, nt>4 is awaited to have memory */
	   i = (nt   > ntheta  ? nt   : ntheta);
	   j = (nphi > nside*4 ? nphi : nside*4);
	   if (i > j)  j = i;

	   PR(ER, "Now try to find memory again for %d elements...", j*2);
	   if ((fi1 = (double *)calloc((unsigned)j, sizeof *fi1)) ==
		(double *)0) {
nomem_again:
		     PR(ER,"\n%s: no memory again...\n", myname);
		     exit(1);
	   }
	   if ((work = (double *)calloc((unsigned)j, sizeof *work)) ==
		(double *)0)
		     goto nomem_again;

	   fseek(fp_tmp, 0l, 0);

	   PR(ER, "O'k.\n");
	} else {
		/* transpose matrix:             */
		/* fi1(nphi, nt) -> fi(nt, nphi) */
		for(i=0; i<nphi; i++) {
			register int k = i * nt;
			for(j=0; j<nt; j++) {
				  fi[j*nphi+i] = fi1[k+j];
				 /* fi[k+j] = fi1[k+j]; */
			}
		}

		/* temporal array, nt>4 is awaited to have memory */
		i = (nt   > ntheta  ? nt   : ntheta);
		j = (nphi > nside*4 ? nphi : nside*4);
		if (i > j)  j = i;
		work = fi1 + j + 1;
if(debug_flag)
{
int flag_grid = -1;  /* homogeneous grid */
int transposing = 0; /* no transposing   */
char tmpfile[161];
int nmphi = nphi;
FILE *fp;
double *x=(double *)calloc(nt, sizeof *x);

strcpy(tmpfile, "er2.f");
fp=fopen(tmpfile, "w");
if(fp != (FILE *)0) {
PR(ER,"ATTENTION: temporal file '%s' is created in current directory.\n",
      tmpfile);
for(i=0; i<nt; i++) x[i] = -t0 + i * (-dt);

/* output in basic F-format           */
/* first nmx='nphi' for calculation of  */
/* position f[i*nphi+j]                */
ff_write_map(fi, nmphi, x, nt, phi_w, nphi, flag_grid, transposing, fp);
fclose(fp);
} else {
PR(ER,"no temporal file\n");
}
free((void *)x);
}
	}

	if ((map_ring = (double *)calloc(npixtot, sizeof *map_ring)) ==
	     (double *)0) {
		PR(ER,"%s: no memory for temporal map array...\n", myname);
		PR(ER,"%s: I shall try to save some memory...\n", myname);
		free((void *)fi1);

		/* temporal array, nt>4 is awaited to have memory */
		i = (nt   > ntheta  ? nt   : ntheta);
		j = (nphi > nside*4 ? nphi : nside*4);
		if (i > j)  j = i;

		PR(ER,"Now try to find memory for %d elements...",j*2);
		if ((fi1 = (double *)calloc((unsigned)j, sizeof *fi1)) ==
		     (double *)0)
			  goto nomem_again;
		if ((work = (double *)calloc((unsigned)j, sizeof *work)) ==
		     (double *)0)
			  goto nomem_again;

		PR(ER,"Now try to find memory again for %d elements...",
		      npixtot);

		if((map_ring=(double *)calloc(npixtot, sizeof *map_ring))==
		     (double *)0) {
nomem:
			PR(ER, "%s: Sorry, no memory at all...\n", myname);
			exit(1);
		}
	}

	/* HEALPix ring pixelization scheme */
	nv = nphi/2;
	for(l_s=npixtot,l_n=i=0; i< ns2; i++) {
	    /* here nph is a number of intervals, like (nphi-1) */
	    register int np = 4 * (i+1 < nside ? i+1 : nside);
	    double step = pi2/np;
	    int cf = (i+1 < nside ? 1 : (i+2-nside)%2 );

	    /***************************/
	    /* for Northern hemisphere */
	    /***************************/
	    if ( fp_tmp != (FILE *)0) {
		 fseek(fp_tmp, (long)i*(sizeof *fi1)*nphi, 0);
		 if (fread(fi1,  sizeof *fi1, nphi, fp_tmp) != nphi) {
		      PR(ER,"%s: error during reading...\n", myname);
		      exit(1);
		 }
	    } else {
		 for(j=0; j<nphi; j++)  fi1[j] = fi[i*nphi+j];
	    }

	    if (flag_int) {
		/* symmetrical copy */
/*
		for(j=0; j<nv; j++)  {
		     int k = j+nv;
		     if (k > nphi - 1)  k -= nphi-1;
		     p0 = fi1[j]; fi1[j] = fi1[k]; fi1[k] = p0;
		}
*/
		for(j=0; j<np; j++) {
			double x = j * step + cf*step/2.;
			double y;
			int in = indx(phi_w, nphi, x);
			if (in >= nphi) in = 0;
			y = fi1[in];
			map_ring[l_n+j] = y;
		}
	    } else {
		/* symmetricaly copy */
		for(j=0; j<nv; j++)  {
		     p0 = fi1[j]; fi1[j] = fi1[nphi-1-j]; fi1[nphi-1-j] = p0;
		}

		/* used for splint */
		spline (phi_w, fi1, nphi, 2.e30, 2.e30, work);

		/* interpolation */
		for(j=0; j<np; j++) {
			/* use middle point of intervals */
			double x = j * step + cf*step/2.;
			double y;

			/* used with spline */
			splint (phi_w, fi1, work, nphi, x, &y);

			if (flag_int)  y = (int)(y+0.5);
			map_ring[l_n+j] = y;
		}
	    }

/*
if(debug_flag) {
if(i<-1) {
FILE *fp;
char s[30];
sprintf(s, "fd%03d.nphi", i);
fp=fopen(s, "w");
for(j=0; j<nphi; j++) PR(fp, "%12g  %12g\n", phi_w[j], fi1[j]);
fclose(fp);

sprintf(s, "fd%03d.nph", i);
fp=fopen(s, "w");
for(j=0; j<np; j++) PR(fp, "%12g  %12g\n",
phi_w[0] + j * step + step/2., map_ring[l_n+j]);
fclose(fp);
}
}
*/
	    l_n += np;

	    /***************************/
	    /* for Southern hemisphere */
	    /***************************/
	    if ( i != ns2 ) {
	       if ( fp_tmp != (FILE *)0) {
		    fseek(fp_tmp, (long)(nt-1-i)*(sizeof *fi1)*nphi, 0);
		    if (fread(fi1,  sizeof *fi1, nphi, fp_tmp) != nphi) {
			 PR(ER,"%s: error during reading...\n", myname);
			 exit(1);
		    }
	       } else {
		    for(j=0; j<nphi; j++)  fi1[j] = fi[(nt-1-i)*nphi+j];
	       }

	       /* symmetricaly copy */
	       for(j=0; j<nv; j++)  {
		  p0 = fi1[j]; fi1[j] = fi1[nphi-1-j]; fi1[nphi-1-j] = p0;
	       }

	       l_s -= np;
	       if (flag_int) {
		   for(j=0; j<np; j++) {
			   double x = j * step + cf*step/2.;
			   double y = fi1[indx(phi_w, nphi, x)];
			   map_ring[l_s+j] = y;
		   }
	       } else {
		   /* used for splint */
		   spline (phi_w, fi1, nphi, 2.e30, 2.e30, work);

		   /* interpolation */
		   for(j=0; j<np; j++) {
			   /* use middle point of intervals */
			   double x = j * step + cf*step/2.;
			   double y;

			   /* used with spline */
			   splint (phi_w, fi1, work, nphi, x, &y);

			   if (flag_int)  y = (int)(y+0.5);
			   map_ring[l_s+j] = y;
		   }
	       }
	    }
	}
	/* free memory */
	if (flag_phi_set) free((void *)phi_w);
	if (fi != (double *)0)  free((void *)fi);
	free((void *)fi1);

	if (fp_tmp != (FILE *)0) {
		free((void *)work);
		fclose(fp_tmp);
		unlink(tmpfile);
		fp_tmp = (FILE *)0;
	}

	/************************************/
	/* output data to FITS binary Table */
	/************************************/

	/*********************/
	/* prepare FITS head */
	/*********************/
	/* 1st header */
	/*------------*/
	cFEVset("SIMPLE", "T"); FEVexport ("SIMPLE");
	iFEVset("BITPIX", 8);   FEVexport ("BITPIX");
	/* number of data axes */
	iFEVset("NAXIS",  0);   FEVexport ("NAXIS");
	/* FITS dataset may contain extensions */
	cFEVset("EXTEND", "T"); FEVexport ("EXTEND");
	/* read current date */
	take_cur_date(sdate);
	/* date: format 'yyyy-mm-dd' */
	sFEVset("DATE", sdate); FEVexport ("DATE");

	Fits_update(fp_out);    /* output 1st header                   */
	FHfree();               /* free memory after 1st header output */

	/*------------*/
	/* 2nd header */
	/*------------*/
	nv = 1024;
	nmap = 1;
	if (npixtot < nv) {
		nrows = npixtot;
		strcpy(tform, "1E");
		nv = 4; /* width of table in bytes */
	} else {
		nrows = npixtot/nv;
		strcpy(tform, "1024E");
		nv *= 4; /* width of table in bytes */
	}

	/* binary table extension           */
	sFEVset("XTENSION","BINTABLE");  FEVexport ("XTENSION");
	/* 8-bit bytes                      */
	iFEVset("BITPIX",  8);           FEVexport ("BITPIX");
	/* 2-dimensional binary table       */
	iFEVset("NAXIS",   2);           FEVexport ("NAXIS");
	/* width of table in bytes          */
	iFEVset("NAXIS1",  nv);          FEVexport ("NAXIS1");
	/* number of rows in table          */
	iFEVset("NAXIS2",  nrows);       FEVexport ("NAXIS2");
	/* size of special data area        */
	iFEVset("PCOUNT",  0);           FEVexport ("PCOUNT");
	/* one data group (required keyword)*/
	iFEVset("GCOUNT",  1);           FEVexport ("GCOUNT");
	/* number of fields in each row     */
	iFEVset("TFIELDS", nmap);        FEVexport ("TFIELDS");

	/*------------------------------------------*/
	/*  Sky Map Pixelisation Specific Keywords  */
	/*------------------------------------------*/
	sFEVset("PIXTYPE", "HEALPIX");   FEVexport ("PIXTYPE");
	/* Pixel ordering scheme (RING)     */
	sFEVset("ORDERING","RING");      FEVexport ("ORDERING");
	/* Resolution parameter for HEALPIX */
	iFEVset("NSIDE",   nside);       FEVexport ("NSIDE");
	/* First pixel # (0 based)          */
	iFEVset("FIRSTPIX",0);           FEVexport ("FIRSTPIX");
	/* Last pixel # (0 based)           */
	iFEVset("LASTPIX", npixtot-1);   FEVexport ("LASTPIX");

	/*--------------------------------------*/
	/*  Planck Simulation Specific Keywords */
	/*--------------------------------------*/
	sFEVset("EXTNAME","SIMULATION"); FEVexport ("EXTNAME");
	sFEVset("CREATOR","CL2MAP");     FEVexport ("CREATOR");
	/*----------------------------------*/
	/* Software creating the FITS file  */
	/*----------------------------------*/
	if ( Cl_max > 0) {
	     /* Max multipole l used in map synthesis */
	     iFEVset("MAX-LPOL",Cl_max);  FEVexport ("MAX-LPOL");
	}
	if ( alm_rand_seed > -32767) {
	     /* Random generator seed */
	     iFEVset("RANDSEED",alm_rand_seed); FEVexport ("RANDSEED");
	}
	if ( fwhm > 0. ) {
	     /* FWHM in DEGREES */
	     dFEVset("FWHM",    fwhm);    FEVexport ("FWHM");
	}
	/*------------------------------------*/
	/* Data Description Specific Keywords */
	/*------------------------------------*/
	/* Temperature map */
	sFEVset("TTYPE1", "TEMPERATURE"); FEVexport ("TTYPE1");
	/* data format of field: 4-byte REAL */
	sFEVset("TFORM1", tform);         FEVexport ("TFORM1");
	/* map unit */
	sFEVset("TUNIT1", "K");           FEVexport ("TUNIT1");

	Fits_update(fp_out);    /* output 2nd header */

	/**********************/
	/* output binary data */
	/**********************/
	bitpix = -32;
	put_dvec_fits(fp_out, map_ring, npixtot);

	/* free memory */
	free((void *)map_ring);

	return;
}

/* reads current system date and time          */
/* and write then in to character buffers like */
/* dd/mm/yyyy  and hh:mm:ss.sss                */
/* returns day of a week                       */
#include <time.h>
#include <sys/types.h>
char *take_cur_date(sdate)
char *sdate;
{
#ifdef __GNUC__
	time_t t;
#else
	long t;
#endif
	struct tm *qq;

	time ( &t );
	qq = localtime ( &t );
/*
	sprintf(stm,  "%02d:%02d:%05.2f",
		qq->tm_hour, qq->tm_min, (double)qq->tm_sec);
	sprintf(date, "%02d/%02d/%4d",
		qq->tm_mday, qq->tm_mon+1, qq->tm_year+1900);
*/
	/* return qq->tm_wday;  */
	sprintf(sdate, "%04d-%02d-%02d",
		qq->tm_year+1900, qq->tm_mon+1, qq->tm_mday);

	return sdate;
}
