#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define PR      fprintf
#define ER      stderr
struct _ph_st_ {
     int l, m;
     double ph;
};

extern void  setopt(int, char **), usage(void);
extern void  sort_phase(struct _ph_st_ *, int);
extern void  sort_phase2(double *, int);
extern void  read_ar(char *, double **, int *);

char ph_file[161], links_file[161];
int lmax = 2000;
double ph0 = 0.0, eps = 1.e-9;

char *myname="find_phase";
int main(argc, argv)
int argc;
char **argv;
{
	register int i, j, l, m;
	double *ph = (double *)0, *phl = (double *)0;
	struct _ph_st_ *ph_st;
	double pi2 = 2 *M_PI;
	unsigned int nl;
	FILE *fp_out = stdout;
	int n_ph=0;

	ph_file[0] = links_file[0] = '\0';

	/* loop over external parameters */
	setopt(argc, argv);

	read_ph_table(ph_file,  &ph,  &lmax);
	read_ar(links_file, &phl, &n_ph);

	nl = (lmax+1)*(lmax+1)/2;

	if ((ph_st=(struct _ph_st_ *)calloc(nl, sizeof *ph_st)) ==
	     (struct _ph_st_ *)0) {
		PR(ER, "%s: no memory for arrays...\n", myname);
		exit(1);
	}

	/* fill in array */
	nl=0;
	for(l=2; l<=lmax; l++)
	    for(m=1; m<=l; m++) {
		ph_st[nl].l  = l;
		ph_st[nl].m  = m;
		ph_st[nl].ph = ph[l*(lmax+1)+m];
		nl++;
	    }

	/* sort array */
	sort_phase(ph_st, nl);
	sort_phase2(phl, n_ph);

	/* select all phases near given one */
	for(i=0; i<n_ph; i++) {
	    for(j=0; j<nl; j++) {
		double dp = ph_st[j].ph - phl[i];

		if (fabs(dp) < eps || fabs(dp+pi2)<eps || fabs(dp-pi2)<eps) {
		    PR(fp_out, "%4d  %4d  %12.5e\n",
			    ph_st[j].l, ph_st[j].m, ph_st[j].ph);
		} else
		    continue;
	    }
	}

	exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'e':
			eps = atof(argv[++i]);
			break;
		case 'h':
			usage();
			break;
		case 'l':
			strcpy(links_file, argv[++i]);
			break;
		case 'p':
			ph0 = atof(argv[++i]);
			break;
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		strcpy(ph_file, argv[i]);
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s file -p phi0 -e eps -l links_file\n", myname);
exit(2);
}

/* "Quicker-sort algorithm" from system library         */
int i_cmp1(p1, p2)
void  *p1, *p2;
{
	struct _ph_st_ *ph1 = (struct _ph_st_ *)p1,
		       *ph2 = (struct _ph_st_ *)p2;
	return (ph1->ph < ph2->ph ? -1 : (ph1->ph > ph2->ph ? 1 : 0));
}
int i_cmp2(p1, p2)
void  *p1, *p2;
{
	double *ph1 = (double *)p1,
	       *ph2 = (double *)p2;
	return (ph1 < ph2 ? -1 : (ph1 > ph2 ? 1 : 0));
}
void sort_phase(ph, n)
struct _ph_st_ *ph;
int n;
{
	qsort(ph, n, sizeof(*ph), (int (*)())i_cmp1);
	return;
}

void sort_phase2(ph, n)
double *ph;
int n;
{
	qsort(ph, n, sizeof(*ph), (int (*)())i_cmp2);
	return;
}

void read_ar(filename, ar, n)
char *filename;
double **ar;
int *n;
{
    register int i=0;
    char str[161];
    FILE *fp;

    /* open file */
    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
    }

    /* strings count */
    while (fgets(str, 160, fp) != NULL) {
	    if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		    continue;
	    i++;
    }
    fseek(fp, 0l, 0);

    if (i <= 0) {
	    PR(ER, "%s: problem with content of file '%s'\n",
		    myname, filename);
	    exit(1);
    }
    *n = i;

    /* find memory for arrays */
    if ((*ar = (double *)calloc((unsigned)i, sizeof **ar)) ==
	(double *)0) {
	    PR(ER,"%s: can not find memory for array...\n", myname);
	    exit(1);
    }

    i = 0;
    /* strings reading */
    while (fgets(str, 160, fp) != NULL) {
	    if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		    continue;

	    (*ar)[i] = atof(str);
	    i++;
    }

    fclose(fp);
    return;
}
