#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PR         fprintf
#define ER         stderr
#define MAX(x,y)   ((x) > (y) ? (x) : (y))

double err(double), errcc(double);
double probks(double);
double kolmogorov_smirnov_gauss(int, double, double *);
double ks_arr2hist(double *, int);

void  alm_ks_test(alm, nalm, lmax, l_calc, m_calc, n_low, fp_out, p_flag)
double *alm;
int nalm, lmax;
int l_calc;     /* if l_calc > 0, then l=l_calc -mutipoles are calculated */
int m_calc;     /* if m_calc >=0, then m=m_calc -mutipoles are calculated */
int n_low;      /* upper limit of {l,m} of a_{lm}                         */
FILE *fp_out;
int p_flag;     /* if p_flag =0, then both parts participate in output    */
		/* if p_flga =1, then real one, if =2, then imagine one   */
{
	register int l, m, n;
	double *v = (double *)0;
	double *d;
	int flag_print = 0;
	double t;

	if (n_low < 3)  n_low = 100;

	n = (n_low+1) * (n_low+2);
	m = 2*lmax + 1;
	n = MAX(n, m);

	if((v=(double *)calloc((unsigned)n, sizeof *v)) == (double *)0) {
		PR(ER, "alm_ks_test: no memory for %d element array.\n", n);
		exit(1);
	}
	d = v;

	PR(fp_out, "#Realization    K-S Probability\n");

	if (l_calc > 0) {
		int ll = l_calc*nalm;
		n=0;
		if( !p_flag || p_flag == 1) {n++; *d++ = alm[ll];}
		for (m=1; m<=l_calc; m++) {
		    if( !p_flag || p_flag==1) {n++; *d++ = alm[ll+m];       }
		    if( !p_flag || p_flag==2) {n++; *d++ = alm[ll+m+lmax+1];}
		}
		/* n = 2*l_calc+1; */
	} else if (m_calc >= 0) {
		for(n=0,l=m_calc; l<=lmax; l++) {
		    int ll = l*nalm;
		    if( !p_flag || p_flag == 1)
			{ n++; *d++ = alm[ll+m_calc];       }
		    if(m_calc && (!p_flag || p_flag == 2))
			{ n++; *d++ = alm[ll+m_calc+lmax+1];}
		}
		/* n = 2 * (lmax - m_calc + 1); */
	} else if(lmax <= n) {
	    /* select all values (l,m) before 100 */
	    n=0;
	    for (l=2; l<=lmax; l++) {
		 int ll = l*nalm;
		 if( !p_flag || p_flag == 1) {*d++ = alm[ll]; n++;}
		 for (m=1; m<=l; m++) {
		    if( !p_flag || p_flag == 1)
			      { n++; *d++ = alm[ll+m];        }
		    if( !p_flag || p_flag == 2)
			      { n++; *d++ = alm[ll+m+lmax+1]; }
		 }
	    }
	} else {
		/*-------------------*/
		/*  if lmax > n=100  */
		/*-------------------*/
		/* 1) consider, l<=n */
		/* select all values (l,m) before 100 */
		n=0;
		for (l=2; l<=n_low; l++) {
		     int ll = l*nalm;
		     if( !p_flag || p_flag == 1) {*d++ = alm[ll]; n++;}
		     for (m=1; m<=l; m++) {
			if( !p_flag || p_flag == 1)
				{n++; *d++ = alm[ll+m];       }
			if( !p_flag || p_flag == 2)
				{n++; *d++ = alm[ll+m+lmax+1];}
		     }
		}
		t = ks_arr2hist(v, n);
		PR(fp_out, "     0             %5.3e\n", t);

		/* 2) consider, l>n */
		for (l=n_low+1; l<=lmax; l++) {
		     int ll = l*nalm;
		     n=0;
		     d=v;
		     if( !p_flag || p_flag == 1) {*d++ = alm[ll]; n++;}
		     for (m=1; m<=l; m++) {
			if( !p_flag || p_flag == 1)
			     { n++; *d++ = alm[ll+m];       }
			if( !p_flag || p_flag == 2)
			     { n++; *d++ = alm[ll+m+lmax+1];}
		     }
		     t = ks_arr2hist(v, n);
		     PR(fp_out, "   %4d            %5.3e\n", l, t);
		}
		flag_print = 1;
	}
	if (!flag_print) {
		t = ks_arr2hist(v, n);
		PR(fp_out, "     0             %5.3e\n", t);
	}

	if (v != (double *)0)  free((double *)v);
	return;
}

int dcmp(d1, d2)
const void *d1, *d2;
{
	int flag = 0;
	double d = *(double *)d1 - *(double *)d2;
	if (d < 0.) flag = -1;
	else if (d > 0.) flag = 1;
	return flag;
}

void histog(in,out,n,m,min,step)
register double *in;
register double *out;
double min, step;
unsigned n,m;
{
	register unsigned j;
	register int i;
	for(j=0; j < n; j++)  {
		i = (in[j] - min) / step + .5;
		if(i >= 0 && i < m)    out[i] += 1;
	}
	return ;
}


double ks_arr2hist(v, n)
double *v;
int n;
{
	register int i;
	int n_bin = n/10;
	double min = 1.e+30, max = -1.e+30, step;
	double *g = (double *)0;
	double t;
/*
	double sig=0., a=0.;
*/

	g = (double *)calloc(n, sizeof *g);
	if (g == (double *)0) {
		PR(ER, "arr2hist: no memory for temporal array.\n");
		exit(1);
	}

	/* sort data */
	/* qsort(v, n, sizeof *v, dcmp);
	 */

	/* normalize to dispersion */
/*
	for(i=0; i<n; i++)  a += v[i];
	a /= n;
	for(i=0; i<n; i++)  v[i] -= a;
	a=0.;
	for(i=0; i<n; i++)  sig += (t=v[i] - a)*t;
	sig = sqrt(sig/n);
	for(i=0; i<n; i++)  v[i] /= sig * sqrt(2.*M_PI);
*/
	/* find minimum/maximum */
	for(i=0; i<n; i++) {
		if (max < v[i])  max = v[i];
		if (min > v[i])  min = v[i];
	}
	step = (max-min) / n_bin;

	/* calc histogram */
	histog(v,g,n,n_bin,min, step);

	/* normalization */
	max = 0.0;
	for(i=0; i<n_bin; i++)  max  += g[i];
	for(i=0; i<n_bin; i++)  g[i] /= max;

	t = kolmogorov_smirnov_gauss(n_bin, step, g);

	if (g != (double *)0)  free((double *)g);
	return t;
}

double kolmogorov_smirnov_gauss(n, dx, g)
int n;
double dx;
double *g;
{
	register int i;
	double sq2 = M_SQRT1_2; /* 0.70710678118654752440 */
	double x = -0.5 * dx * n, y=0;
	double d = 0., dd, ff;
	double prb;

	for (i=0; i<n; i++) {
		y = sq2 * fabs(x);
		ff = 0.5 * (1. +  x>0.? err(y) : -err(y));
		dd = fabs(ff - g[i]);
		if(dd > d)  d = dd;
		x += dx;
	}
	ff = sqrt((double)n);

	y = (ff + 0.12 + 0.11/ff) * d;

	prb = probks(y);

	return prb;
}


double  err(x)
double x;
{
	double x2 = x*x;
	double y=1.;
	int n=15;
	double er;

	if(x2 < 4.)  {
		register int i;
		for(i=0; i<n; i++) {
			double xk = n - i;
			y = 1.-(xk-0.5)/(xk+0.5)*x2/xk*y;
		}
		er = 1.12837916 * y * fabs(x);
	} else
		er = 1. - errcc(x);

	er = fabs(er);

	return er;
}

double errcc(x)
double x;
{
	double x2 = x*x;
	double y=1.;
	int n=4;
	double er;

	if(x2 > 4.) {
		register int i;
		for (i=0; i<n; i++) {
		     double xk = n - i;
		     y = 1.-0.5*(2.*xk-1.)/x2*y;
		}
		er = 0.56418958/fabs(x)*y*exp(-x2);
	} else
		er = 1.- err(x);

	if(x < 0.)  er = 2. - er;

	return er;
}

/*
 *  Kolomogorov-Smirnov probability function
 */
#define EPS1 0.001
#define EPS2 1.0e-8

double probks(double alam)
{
    register int j;
    double  a2, fac=2.0, sum=0.0, term, termbf=0.0;

    a2 = -2.0*alam*alam;

    for (j=1;j<=100;j++) {
	term = fac*exp(a2*j*j);
	sum += term;
	if (fabs(term) <= EPS1*termbf || fabs(term) <= EPS2*sum)
		return sum;
	fac = -fac;
	termbf = fabs(term);
    }
    return 1.0;
}
