#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "gd/gd.h"
#include "glesp.h"

char *myname = "gif2fts";

int main(int argc, char **argv)
{
    gdImagePtr im;
    FILE *in, *out;
    double *f = (double *)0, *x = (double *)0;
    int nx=0, np=0, *nph = (int *)0;
    extern void convert_gd2map(gdImagePtr, double **, double **, int *,
	int*, int **);
    char x_type[40], grid_type[40];

    strcpy(x_type, "COS(THETA)");
    strcpy(grid_type, "PHI(X)+LEGENDRE");

    if (argc != 3) {
	    fprintf(stderr, "Usage: %s filename.gif filename.fts\n", myname);
	    exit(1);
    }
    in = fopen(argv[1], "r");
    if (!in) {
	    fprintf(stderr, "Input file does not exist!\n");
	    exit(1);
    }
    im = gdImageCreateFromGif(in);
    fclose(in);
    if (!im) {
	    fprintf(stderr, "Input is not in GIF format!\n");
	    exit(1);
    }
    convert_gd2map(im, &f, &x, &nx, &np, &nph);
    gdImageDestroy(im);

    out = fopen(argv[2], "w");
    if (!out) {
	    fprintf(stderr, "Output file cannot be written to!\n");
	    exit(1);
    }
    outp_glesp_map(f, x, nx, np, nph, out, x_type, grid_type);
    fclose(out);

    exit(0);
}


#define MAX(x,y) ((x)>(y)?(x):(y))
void convert_gd2map(im, f, x, nx, np, nph)
gdImagePtr im;
double **f, **x;
int *nx, *np, **nph;
{
    register int i, j;
    int ny_gd=im->sy, nx_gd=im->sx;
    int y_c = ny_gd / 2,
	x_c = nx_gd / 2,
	x_r = nx_gd / 2,
	y_r = ny_gd / 2;
    double lon_0 = 0.0;
    double x1=-1.0, x2=1.0, pi=M_PI;

    *nx = 401;
    *np = *nx * 2;
    if ((*f = (double *)calloc((unsigned)(*nx * *np), sizeof **f)) ==
	(double *)0) {
nomem:
	    fprintf(stderr, "no memory for a map...\n");
	    exit(1);
    }
    if ((*x = (double *)calloc((unsigned)(*nx), sizeof **x)) == (double *)0)
	goto nomem;
    if ((*nph = (int *)calloc((unsigned)(*nx), sizeof **nph)) == (int *)0)
	goto nomem;
    gauleg(x1, x2, *x, (double *)0, *nx);
    for (i=0; i < *nx; i++) (*nph)[i] = *np;

    for(i=0; i<ny_gd; i++) {
	double v, u, theta,phi;
//      v = (i - y_c) / (y_r/1.02);
	v = (i - y_c) /  y_r;
	for(j=0; j<nx_gd; j++) {
//          u = 2. * (j - x_c) / (x_r/1.02);
	    u = 2. * (j - x_c) / x_r;
	    theta = phi = 1.e+30;
	    if ( u*u/4. + v*v <= 1.) {
		double sqv = sqrt(1.-v*v);
		/* theta in [0, pi] */
		theta = pi/2. - asin(2./pi * (asin(v) + v*sqv));
		/* lon in [-pi,pi], the minus sign for astro convention */
		phi = -lon_0 - pi/2 * u / MAX(sqv, 1.e-6);
		/* phi in [0, 2pi] */
		if ( phi < 0.0 )    phi += 2. * pi;
	    }
	    /* keep only meaningful points */
	    if (ABS(theta) < 1.e5) {
		int n = ang2n(theta, phi, *x, *nx, *np, *nph);
		(*f)[n] = gdImageGetPixel(im, j, i);
	    }
	}
    }
    return;
}
