#include <stdio.h>
#include <stdlib.h>
#include "glesp.h"

#define PR      fprintf
#define ER      stderr
#define ABS(x)  ((x) < 0 ? -(x) : (x))
#define SMALL   1.e-15

extern void spline (double *, double *, int, double, double, double *);
extern void splint (double *, double *, double *, int, double, double *);

extern int intrpl2Fsm_nu_u(FILE *, int, double *, int, int, double *,
		double, double, int);
extern int intrpl2Fsm_nu_nu(FILE *, int, double *, int, int, double *,
		double *, int);
extern int  intrpl2sm_Nnu_u(double *, int, int, double *, int, int,
		double *, double, double,  double, double, int *, int);
extern int  intrpl2Fsm_Nnu_nu(FILE *, int, int, double *, int, int,
		double *, double *, double, double, int *, int);

extern char *myname;

/*-------------------------------------------*/
/* spline interpolation (modified input):    */
/*   non-uniform step by X ---> uniform step */
/*   using FILE *stream;                     */
/*-------------------------------------------*/
int intrpl2Fsm_nu_u(fp_in, n_x_in, out_data, n_x_out, nmx_out,
	arr, start, step, n_y)
FILE *fp_in;       /* file stream with 2-dim input array       */
double *out_data,  /* 2-dim output array                       */
       *arr;       /* arguments x for 'in_data'-array          */
int     n_x_in,    /* number of elements in 'in_data' by X     */
	n_x_out,   /* number of elements in 'out_data' by X    */
	nmx_out,   /* array for for output data: (i*nmx_out+j) */
	n_y;       /* number of strings in both arrays         */
double step,       /* step by X-axis for 'out_data'-array      */
       start;      /* start on X-axis for 'out_data'-array     */
{
       register int i,j;
       double *val,
	      *work;
       double x, y;

       if(( val = (double *)calloc(n_x_in,sizeof *val)) == (double *)0) {
nomem:
	       PR(ER,"%s:  I can't find memory....\n", myname);
	       exit(1);
       }
       if((work = (double *)calloc(n_x_in,sizeof *work)) == (double *)0)
	       goto nomem;

       for(i=0; i<n_y; i++) {
PR(ER,"%5d\b\b\b\b\b", i);
	       if (fread(val, sizeof *val, n_x_in, fp_in) != n_x_in) {
		    PR(ER,"%s: error during reading for interpolation...\n",
			myname);
		    exit(1);
	       }

	       /* used for splint */
	       spline (arr, val, n_x_in, 2.e30, 2.e30, work);
	       for(j=0; j<n_x_out; j++) {
		       x = start +  j * step;

		       /* used with spline */
		       splint (arr, val, work, n_x_in, x, &y);
		       out_data[i*nmx_out+j] = y;
	       }
       }

PR(ER,"\n");
       free((void *)val);
       free((void *)work);

       return 1;
}

/*-------------------------------------------*/
/* spline interpolation (modified input):    */
/*   non-uniform step by X ---> uniform step */
/*   and non-uniform step by Y               */
/* so matrix is not square                   */
/*-------------------------------------------*/
int intrpl2sm_Nnu_u(in_data, n_x_in, nmx_in, out_data, n_x_out, nmx_out,
	arr, start, step, s0, slen, nya, n_y)
double *in_data,   /* 2-dim input array                        */
       *out_data,  /* 2-dim output array                       */
       *arr;       /* arguments x for 'in_data'-array          */
int     n_x_in,    /* number of elements in 'in_data' by X     */
	n_x_out,   /* number of elements in 'out_data' by X    */
	nmx_in,    /* array for for input data: (i*nmx_in+j)   */
	nmx_out,   /* array for for output data: (i*nmx_out+j) */
	n_y;       /* max number of strings in both arrays     */
int    *nya;       /* array of ny(x)                           */
double s0, slen;
double step,       /* step by X-axis for 'out_data'-array      */
       start;      /* start on X-axis for 'out_data'-array     */
{
       register int i,j;
       double *val  = (double *)0,
	      *work = (double *)0,
	      *ar1  = (double *)0;
       double x, y, dy = slen/n_y;

       /* transpose matrix */
       /* in_data(n_y , n_x_in) -> in_data(n_x_in, ny) */
       /*--------------------------------------------------
	*
	*    ntheta = n_x_in = nmx_in
	*                                        nphi(i) = nya(i) <= n_y
	*       ******
	*       ******                           ***
	*  nphi ******           --------->      ******   ntheta
	* = n_y *****                            *******
	*       *****                            ******
	*        ***                             *****
	*         *
	*
	*/
       transpose_map(in_data, n_y, n_x_in);

       /* take temporal memory */
       i = n_y > n_x_in ? n_y : n_x_in;
       if(( val = (double *)calloc(i, sizeof *val)) == (double *)0) {
nomem:
	       PR(ER,"%s:  I can't find memory....\n", myname);
	       exit(1);
       }
       if((work = (double *)calloc(i, sizeof *work)) == (double *)0)
	       goto nomem;
       if((ar1  = (double *)calloc(i, sizeof *ar1)) == (double *)0)
	       goto nomem;

PR(ER, "interpolation: 1st level (%d steps)...\n", n_x_in);
       /* make array square by Y-axis */
       /*
	*
	*   nphi = nya(i)                  n_y = nphi
	*
	*   ***                          *******
	*   ******   ntheta  -------->   *******  ntheta = n_x_in
	*   *******    ||                *******
	*   ******    n_x_in             *******
	*   *****
	*
	*/
       for(i=0; i<n_x_in; i++) {
	       double dys = slen / nya[i];
PR(ER,"%5d\b\b\b\b\b", i);
	       for(j=0; j<nya[i]; j++) {
			ar1[j] = s0 + j * dys;
			val[j] = in_data[i*n_y+j];
	       }
	       /* used for splint */
	       spline (ar1, val, nya[i], 2.e30, 2.e30, work);

	       for(j=0; j<n_y; j++) {
		       x = s0 + j * dy;

		       /* used with spline */
		       splint (ar1, val, work, nya[i], x, &y);
		       in_data[i*n_y+j] = y;
	       }
	       nya[i] = n_y;
       }

PR(ER, "\ninterpolation: 2nd level (%d steps)...\n", n_y);
       /* now change step by X-axis */
       /*
	*                                n_x_out = nt = nmx_out
	*            n_y
	*                                  ****
	*          *******                 ****
	*  ntheta  *******    --------->   **** n_y = nphi
	*  =       *******                 ****
	*  n_x_in  *******                 ****
	*                                  ****
	*
	*/
       for(i=0; i<n_y; i++) {
	       register int ii = i * nmx_out;
PR(ER,"%5d\b\b\b\b\b", i);
	       for(j=0; j<n_x_in;  j++) val[j] = in_data[j*n_y+i];

	       /* used for splint */
	       spline (arr, val, n_x_in, 2.e30, 2.e30, work);
	       for(j=0; j<n_x_out; j++) {
		       x = start +  j * step;

		       /* used with spline */
		       splint (arr, val, work, n_x_in, x, &y);
		       out_data[ii+j] = y;
	       }
       }
PR(ER, "\n");
       if (val  != (double *)0) free((void *)val);
       if (work != (double *)0) free((void *)work);
       if (ar1  != (double *)0) free((void *)ar1);

       return 1;
}

int intrpl2sm_Nnu_nu(in_data, n_x_in, nmx_in, out_data, n_x_out, nmx_out,
	arr_in, arr_out, s0, slen, nya, n_y)
double *in_data,   /* 2-dim input array                        */
       *out_data,  /* 2-dim output array                       */
       *arr_in;    /* arguments x for 'in_data'-array          */
int     n_x_in,    /* number of elements in 'in_data' by X     */
	n_x_out,   /* number of elements in 'out_data' by X    */
	nmx_in,    /* array for for input data: (i*nmx_in+j)   */
	nmx_out,   /* array for for output data: (i*nmx_out+j) */
	n_y;       /* max number of strings in both arrays     */
int    *nya;       /* array of ny(x)                           */
double s0, slen;
double *arr_out;   /* arguments x for 'out_data'-array         */
{
       register int i,j;
       double *val  = (double *)0,
	      *work = (double *)0,
	      *ar1  = (double *)0;
       double x, y, dy = slen/n_y;

       /* transpose matrix */
       /* in_data(n_y , n_x_in) -> in_data(n_x_in, ny) */
       /*--------------------------------------------------
	*
	*    ntheta = n_x_in = nmx_in
	*                                        nphi(i) = nya(i) <= n_y
	*       ******
	*       ******                           ***
	*  nphi ******           --------->      ******   ntheta
	* = n_y *****                            *******
	*       *****                            ******
	*        ***                             *****
	*         *
	*
	*/
       transpose_map(in_data, n_y, n_x_in);

       /* take temporal memory */
       i = n_y > n_x_in ? n_y : n_x_in;
       if(( val = (double *)calloc(i, sizeof *val)) == (double *)0) {
nomem:
	       PR(ER,"%s:  I can't find memory....\n", myname);
	       exit(1);
       }
       if((work = (double *)calloc(i, sizeof *work)) == (double *)0)
	       goto nomem;
       if((ar1  = (double *)calloc(i, sizeof *ar1)) == (double *)0)
	       goto nomem;

PR(ER, "interpolation: 1st level (%d steps)...\n", n_x_in);
       /* make array square by Y-axis */
       /*
	*
	*   nphi = nya(i)                  n_y = nphi
	*
	*   ***                          *******
	*   ******   ntheta  -------->   *******  ntheta = n_x_in
	*   *******    ||                *******
	*   ******    n_x_in             *******
	*   *****
	*
	*/
       for(i=0; i<n_x_in; i++) {
	       double dys = slen / nya[i];
PR(ER,"%5d\b\b\b\b\b", i);
	       for(j=0; j<nya[i]; j++) {
			ar1[j] = s0 + j * dys;
			val[j] = in_data[i*n_y+j];
	       }
	       /* used for splint */
	       spline (ar1, val, nya[i], 2.e30, 2.e30, work);

	       for(j=0; j<n_y; j++) {
		       x = s0 + j * dy;

		       /* used with spline */
		       splint (ar1, val, work, nya[i], x, &y);
		       in_data[i*n_y+j] = y;
	       }
	       nya[i] = n_y;
       }

PR(ER, "\ninterpolation: 2nd level (%d steps)...\n", n_y);
       /* now change step by X-axis */
       /*
	*                                n_x_out = nt = nmx_out
	*            n_y
	*                                  ****
	*          *******                 ****
	*  ntheta  *******    --------->   **** n_y = nphi
	*  =       *******                 ****
	*  n_x_in  *******                 ****
	*                                  ****
	*
	*/
       for(i=0; i<n_y; i++) {
	       register int ii = i * nmx_out;
PR(ER,"%5d\b\b\b\b\b", i);
	       for(j=0; j<n_x_in;  j++) val[j] = in_data[j*n_y+i];

	       /* used for splint */
	       spline (arr_in, val, n_x_in, 2.e30, 2.e30, work);
	       for(j=0; j<n_x_out; j++) {
		       x = arr_out[j];

		       /* used with spline */
		       splint (arr_in, val, work, n_x_in, x, &y);
		       out_data[ii+j] = y;
	       }
       }
PR(ER, "\n");
       if (val  != (double *)0) free((void *)val);
       if (work != (double *)0) free((void *)work);
       if (ar1  != (double *)0) free((void *)ar1);

       return 1;
}

double fabs(double);
int indx(arr, n, x)
double *arr;
int n;
double x;
{
	register int i, imin=0;
	double ds = 1.e30, ds_min = 1.e30;
	for(i=0; i<n; i++) {
		ds = fabs(arr[i]-x);
		if (ds < ds_min)  {ds_min = ds; imin = i;}
	}
	return imin;
}

int copy2sm_Nnu_nu(in_data, n_x_in, nmx_in, out_data, n_x_out, nmx_out,
	arr_in, arr_out, s0, slen, nya, n_y)
double *in_data,   /* 2-dim input array                        */
       *out_data,  /* 2-dim output array                       */
       *arr_in;    /* arguments x for 'in_data'-array          */
int     n_x_in,    /* number of elements in 'in_data' by X     */
	n_x_out,   /* number of elements in 'out_data' by X    */
	nmx_in,    /* array for for input data: (i*nmx_in+j)   */
	nmx_out,   /* array for for output data: (i*nmx_out+j) */
	n_y;       /* max number of strings in both arrays     */
int    *nya;       /* array of ny(x)                           */
double s0, slen;
double *arr_out;   /* arguments x for 'out_data'-array         */
{
       register int i,j;
       double *val  = (double *)0,
	      *ar1  = (double *)0;
       double x, y, dy = slen/n_y;

       /* transpose matrix */
       /* in_data(n_y , n_x_in) -> in_data(n_x_in, ny) */
       /*--------------------------------------------------
	*
	*    ntheta = n_x_in = nmx_in
	*                                        nphi(i) = nya(i) <= n_y
	*       ******
	*       ******                           ***
	*  nphi ******           --------->      ******   ntheta
	* = n_y *****                            *******
	*       *****                            ******
	*        ***                             *****
	*         *
	*
	*/
       transpose_map(in_data, n_y, n_x_in);

       /* take temporal memory */
       i = n_y > n_x_in ? n_y : n_x_in;
       if(( val = (double *)calloc(i, sizeof *val)) == (double *)0) {
nomem:
	       PR(ER,"%s:  I can't find memory....\n", myname);
	       exit(1);
       }
       if((ar1  = (double *)calloc(i, sizeof *ar1)) == (double *)0)
	       goto nomem;

       /* make array square by Y-axis */
       /*
	*
	*   nphi = nya(i)                  n_y = nphi
	*
	*   ***                          *******
	*   ******   ntheta  -------->   *******  ntheta = n_x_in
	*   *******    ||                *******
	*   ******    n_x_in             *******
	*   *****
	*
	*/
       for(i=0; i<n_x_in; i++) {
	       double dys = slen / nya[i];
	       for(j=0; j<nya[i]; j++) {
			ar1[j] = s0 + j * dys;
			val[j] = in_data[i*n_y+j];
	       }
	       /* dy = slen/n_y; */
	       for(j=0; j<n_y; j++) {
		       int in = (s0 + j * dy  - s0) / dys + 0.5;
/*
		       int in = indx(ar1, nya[i], x);
		       double x = s0 + j * dy;
*/
		       double y;

		       if (in == nya[i]) in = 0;
		       else if (in == nya[i]+1) in = 1;
		       y = val [in];
		       in_data[i*n_y+j] = y;
	       }
	       nya[i] = n_y;
       }

       /* now change step by X-axis */
       /*
	*                                n_x_out = nt = nmx_out
	*            n_y
	*                                  ****
	*          *******                 ****
	*  ntheta  *******    --------->   **** n_y = nphi
	*  =       *******                 ****
	*  n_x_in  *******                 ****
	*                                  ****
	*
	*/
       for(i=0; i<n_y; i++) {
	       register int ii = i * nmx_out;
	       for(j=0; j<n_x_in;  j++) val[j] = in_data[j*n_y+i];

	       for(j=0; j<n_x_out; j++) {
		       x = arr_out[j];
		       y = val [indx(arr_in, n_x_in, x)];
		       out_data[ii+j] = y;
	       }
       }
       if (val  != (double *)0) free((void *)val);
       if (ar1  != (double *)0) free((void *)ar1);

       return 1;
}

/*-------------------------------------------*/
/* spline interpolation (modified input):    */
/*   non-uniform step by X ---> uniform step */
/*   and non-uniform step by Y               */
/* so matrix is not square                   */
/* input array from file using FILE *stream; */
/*  ===>  matrix is transposed <===          */
/*-------------------------------------------*/
int intrpl2Fsm_Nnu_u(fp_in, n_x_in, nmx_in, out_data, n_x_out, nmx_out,
	arr, start, step, s0, slen, nya, n_y)
FILE *fp_in;       /* file stream with 2-dim input array       */
double *out_data,  /* 2-dim output array                       */
       *arr;       /* arguments x for 'in_data'-array          */
int     n_x_in,    /* number of elements in 'in_data' by X     */
	n_x_out,   /* number of elements in 'out_data' by X    */
	nmx_in,    /* array for for input data: (i*nmx_in+j)   */
	nmx_out,   /* array for for output data: (i*nmx_out+j) */
	n_y;       /* max number of strings in both arrays     */
int    *nya;       /* array of ny(x)                           */
double s0, slen;
double step,       /* step by X-axis for 'out_data'-array      */
       start;      /* start on X-axis for 'out_data'-array     */
{
       register int i,j;
       double *val  = (double *)0,
	      *work = (double *)0,
	      *ar1  = (double *)0;
       double x, y, dy = slen/n_y;

       /* take temporal memory */
       i = n_y > n_x_in ? n_y : n_x_in;
       if(( val = (double *)calloc(i, sizeof *val)) == (double *)0) {
nomem:
	       PR(ER,"%s:  I can't find memory....\n", myname);
	       exit(1);
       }
       if((work = (double *)calloc(i, sizeof *work)) == (double *)0)
	       goto nomem;
       if((ar1  = (double *)calloc(i, sizeof *ar1)) == (double *)0)
	       goto nomem;

PR(ER, "\ninterpolation: 1st level (%d steps)...\n", n_x_in);
       /* make array square by Y-axis */
       /*
	*
	*   nphi = nya(i)                  n_y = nphi
	*
	*   ***                          *******
	*   ******   ntheta  -------->   *******  ntheta = n_x_in
	*   *******    ||                *******
	*   ******    n_x_in             *******
	*   *****
	*
	*/
       for(i=0; i<n_x_in; i++) {
	       double dys = slen / nya[i];
PR(ER,"%5d\b\b\b\b\b", i);
	       for(j=0; j<nya[i]; j++) ar1[j] = s0 + j * dys;
	       if (fread(val, sizeof *val, n_y, fp_in) != n_y) {
		    PR(ER,"%s: error during reading for interpolation...\n",
			myname);
		    exit(1);
	       }

	       /* used for splint */
	       spline (ar1, val, nya[i], 2.e30, 2.e30, work);

	       for(j=0; j<n_y; j++) {
		       x = s0 + j * dy;

		       /* used with spline */
		       splint (ar1, val, work, nya[i], x, &y);
		       out_data[j*nmx_in+i] = y;
	       }
	       nya[i] = n_y;
       }
PR(ER, "\ninterpolation: 2nd level (%d steps)...\n", n_y);

       /* now change step by X-axis */
       /*
	*                                n_x_out = nt = nmx_out
	*            n_y
	*                                  ****
	*          *******                 ****
	*  ntheta  *******    --------->   **** n_y = nphi
	*  =       *******                 ****
	*  n_x_in  *******                 ****
	*                                  ****
	*
	*/

       /* write file in transposed mode */
       fseek(fp_in, 0l, 0);
       for(i=0; i<n_y; i++) {
	   if(fwrite(out_data+i*n_x_in, sizeof *out_data, n_x_in, fp_in) !=
	      n_x_in) {
		PR(ER,"%s: error during writing for interpolation...\n",
		    myname);
		exit(1);
	   }
       }
       fseek(fp_in, 0l, 0);


       for(i=0; i<n_y; i++) {
	       register int ii = i * nmx_out;
PR(ER,"%5d\b\b\b\b\b", i);
	       if (fread(val, sizeof *val, n_x_in, fp_in) != n_x_in) {
		    PR(ER,"%s: error during reading for interpolation...\n",
			myname);
		    exit(1);
	       }

	       /* used for splint */
	       spline (arr, val, n_x_in, 2.e30, 2.e30, work);

	       for(j=0; j<n_x_out; j++) {
		       x = start +  j * step;

		       /* used with spline */
		       splint (arr, val, work, n_x_in, x, &y);
		       out_data[ii+j] = y;
	       }
       }
       if (val  != (double *)0) free((void *)val);
       if (work != (double *)0) free((void *)work);
       if (ar1  != (double *)0) free((void *)ar1);

       return 1;
}

int intrpl2Fsm_Nnu_nu(fp_in, n_x_in, nmx_in, out_data, n_x_out, nmx_out,
	arr_in, arr_out, s0, slen, nya, n_y)
FILE *fp_in;       /* file stream with 2-dim input array       */
double *out_data,  /* 2-dim output array                       */
       *arr_in;    /* arguments x for 'in_data'-array          */
int     n_x_in,    /* number of elements in 'in_data' by X     */
	n_x_out,   /* number of elements in 'out_data' by X    */
	nmx_in,    /* array for for input data: (i*nmx_in+j)   */
	nmx_out,   /* array for for output data: (i*nmx_out+j) */
	n_y;       /* max number of strings in both arrays     */
int    *nya;       /* array of ny(x)                           */
double s0, slen;
double *arr_out;   /* arguments x for 'out_data'-array         */
{
       register int i,j;
       double *val  = (double *)0,
	      *work = (double *)0,
	      *ar1  = (double *)0;
       double x, y, dy = slen/n_y;

       /* take temporal memory */
       i = n_y > n_x_in ? n_y : n_x_in;
       if(( val = (double *)calloc(i, sizeof *val)) == (double *)0) {
nomem:
	       PR(ER,"%s:  I can't find memory....\n", myname);
	       exit(1);
       }
       if((work = (double *)calloc(i, sizeof *work)) == (double *)0)
	       goto nomem;
       if((ar1  = (double *)calloc(i, sizeof *ar1)) == (double *)0)
	       goto nomem;

PR(ER, "\ninterpolation: 1st level (%d steps)...\n", n_x_in);
       /* make array square by Y-axis */
       /*
	*
	*   nphi = nya(i)                  n_y = nphi
	*
	*   ***                          *******
	*   ******   ntheta  -------->   *******  ntheta = n_x_in
	*   *******    ||                *******
	*   ******    n_x_in             *******
	*   *****
	*
	*/
       for(i=0; i<n_x_in; i++) {
	       double dys = slen / nya[i];
PR(ER,"%5d\b\b\b\b\b", i);
	       for(j=0; j<nya[i]; j++) ar1[j] = s0 + j * dys;
	       if (fread(val, sizeof *val, n_y, fp_in) != n_y) {
		    PR(ER,"%s: error during reading for interpolation...\n",
			myname);
		    exit(1);
	       }

	       /* used for splint */
	       spline (ar1, val, nya[i], 2.e30, 2.e30, work);

	       for(j=0; j<n_y; j++) {
		       x = s0 + j * dy;

		       /* used with spline */
		       splint (ar1, val, work, nya[i], x, &y);
		       out_data[j*nmx_in+i] = y;
	       }
	       nya[i] = n_y;
       }
PR(ER, "\ninterpolation: 2nd level (%d steps)...\n", n_y);

       /* now change step by X-axis */
       /*
	*                                n_x_out = nt = nmx_out
	*            n_y
	*                                  ****
	*          *******                 ****
	*  ntheta  *******    --------->   **** n_y = nphi
	*  =       *******                 ****
	*  n_x_in  *******                 ****
	*                                  ****
	*
	*/

       /* write file in transposed mode */
       fseek(fp_in, 0l, 0);
       for(i=0; i<n_y; i++) {
	   if(fwrite(out_data+i*n_x_in, sizeof *out_data, n_x_in, fp_in) !=
	      n_x_in) {
		PR(ER,"%s: error during writing for interpolation...\n",
		    myname);
		exit(1);
	   }
       }
       fseek(fp_in, 0l, 0);


       for(i=0; i<n_y; i++) {
	       register int ii = i * nmx_out;
PR(ER,"%5d\b\b\b\b\b", i);
	       if (fread(val, sizeof *val, n_x_in, fp_in) != n_x_in) {
		    PR(ER,"%s: error during reading for interpolation...\n",
			myname);
		    exit(1);
	       }

	       /* used for splint */
	       spline (arr_in, val, n_x_in, 2.e30, 2.e30, work);

	       for(j=0; j<n_x_out; j++) {
		       x = arr_out[j];

		       /* used with spline */
		       splint (arr_in, val, work, n_x_in, x, &y);
		       out_data[ii+j] = y;
	       }
       }
       if (val  != (double *)0) free((void *)val);
       if (work != (double *)0) free((void *)work);
       if (ar1  != (double *)0) free((void *)ar1);

       return 1;
}

/*-------------------------------------------*/
/* spline interpolation (modified input):    */
/*   non-uniform step by X ---> uniform step */
/*-------------------------------------------*/
int intrpl2sm_nu_u(in_data, n_x_in, nmx_in, out_data, n_x_out, nmx_out,
	arr, start, step, n_y)
double *in_data,   /* 2-dim input array                        */
       *out_data,  /* 2-dim output array                       */
       *arr;       /* arguments x for 'in_data'-array          */
int     n_x_in,    /* number of elements in 'in_data' by X     */
	n_x_out,   /* number of elements in 'out_data' by X    */
	nmx_in,    /* array for for input data: (i*nmx_in+j)   */
	nmx_out,   /* array for for output data: (i*nmx_out+j) */
	n_y;       /* number of strings in both arrays         */
double step,       /* step by X-axis for 'out_data'-array      */
       start;      /* start on X-axis for 'out_data'-array     */
{
       register int i,j;
       double *val,
	      *work;
       double x, y;

       if(( val = (double *)calloc(n_x_in,sizeof *val)) == (double *)0) {
nomem:
	       PR(ER,"%s:  I can't find memory....\n", myname);
	       exit(1);
       }
       if((work = (double *)calloc(n_x_in,sizeof *work)) == (double *)0)
	       goto nomem;

       for(i=0; i<n_y; i++) {

PR(ER,"%5d\b\b\b\b\b", i);
	       for(j=0; j<n_x_in;  j++) val[j] = in_data[i*nmx_in+j];

	       /* used for splint */
	       spline (arr, val, n_x_in, 2.e30, 2.e30, work);
	       for(j=0; j<n_x_out; j++) {
		       x = start +  j * step;

		       /* used with spline */
		       splint (arr, val, work, n_x_in, x, &y);
		       out_data[i*nmx_out+j] = y;
	       }
       }

       free((void *)val);
       free((void *)work);

       return 1;
}

/*-----------------------------------------------*/
/* spline interpolation (modified input):        */
/*   non-uniform step by X ---> non-uniform step */
/*   using FILE stream                           */
/*-----------------------------------------------*/
int intrpl2Fsm_nu_nu(fp_in, n_x_in, out_data, n_x_out, nmx_out,
	x_in, x_out, n_y)
FILE *fp_in;       /* file stream with 2-dim input array       */
double *out_data,  /* 2-dim output array                       */
       *x_in,      /* arguments x for 'in_data'-array          */
       *x_out;     /* arguments x for 'out_data'-array         */
int     n_x_in,    /* number of elements in 'in_data' by X     */
	n_x_out,   /* number of elements in 'out_data' by X    */
	nmx_out,   /* array for for output data: (i*nmx_out+j) */
	n_y;       /* number of strings in both arrays         */
{
       register int i,j;
       double *val, *work;
       double x, y;

       if(( val = (double *)calloc(n_x_in,sizeof *val)) == (double *)0) {
nomem:
	       PR(ER,"%s:  I can't find memory....\n", myname);
	       exit(1);
       }
       if((work = (double *)calloc(n_x_in,sizeof *work)) == (double *)0)
	       goto nomem;

       for(i=0; i<n_y; i++) {
PR(ER,"%5d\b\b\b\b\b", i);
	       if (fread(val, sizeof *val, n_x_in, fp_in) != n_x_in) {
		    PR(ER,"%s: error during reading for interpolation...\n",
			myname);
		    exit(1);
	       }
	       /* used for splint */
	       spline (x_in, val, n_x_in, 2.e30, 2.e30, work);
	       for(j=0; j<n_x_out; j++) {
		       x = x_out[j];
		       /* used with spline */
		       splint (x_in, val, work, n_x_in, x, &y);
		       out_data[i*nmx_out+j] = y;
	       }
       }

PR(ER,"\n");
       free((void *)val);
       free((void *)work);

       return 1;
}

/*-----------------------------------------------*/
/* spline interpolation (modified input):        */
/*   non-uniform step by X ---> non-uniform step */
/*-----------------------------------------------*/
int intrpl2sm_nu_nu(in_data, n_x_in, nmx_in, out_data, n_x_out, nmx_out,
	x_in, x_out, n_y)
double *in_data,   /* 2-dim input array                        */
       *out_data,  /* 2-dim output array                       */
       *x_in,      /* arguments x for 'in_data'-array          */
       *x_out;     /* arguments x for 'out_data'-array         */
int     n_x_in,    /* number of elements in 'in_data' by X     */
	n_x_out,   /* number of elements in 'out_data' by X    */
	nmx_in,    /* array for for input data: (i*nmx_in+j)   */
	nmx_out,   /* array for for output data: (i*nmx_out+j) */
	n_y;       /* number of strings in both arrays         */
{
       register int i,j;
       double *val, *work;
       double x, y;

       if(( val = (double *)calloc(n_x_in,sizeof *val)) == (double *)0) {
nomem:
	       PR(ER,"%s:  I can't find memory....\n", myname);
	       exit(1);
       }
       if((work = (double *)calloc(n_x_in,sizeof *work)) == (double *)0)
	       goto nomem;

       for(i=0; i<n_y; i++) {
PR(ER,"%5d\b\b\b\b\b", i);
	       for(j=0; j<n_x_in;  j++) val[j] = in_data[i*nmx_in+j];
	       /* used for splint */
	       spline (x_in, val, n_x_in, 2.e30, 2.e30, work);
	       for(j=0; j<n_x_out; j++) {
		       x = x_out[j];
		       /* used with spline */
		       splint (x_in, val, work, n_x_in, x, &y);
		       out_data[i*nmx_out+j] = y;
	       }
       }

PR(ER,"\n");
       free((void *)val);
       free((void *)work);

       return 1;
}

/*
  Given arrays x(1:n) and y(1:n) containing a tabulated function,
  i.e. y_i = f(x_i), with x_1<x_2<...<x_N, and given values yp_1 and yp_n
  for the interpolating at points 1 and n, respectively, this routine
  returns an arrray y2(1:n) of length n which contains the second derivatives
  of the interpolating function at the tabulated points x_i. If yp_1
  and/or yp_n are equal to 1x10^30 or large, the routine is signaled
  to set the corresponding boundary condition for a natural spline, with
  zero second derivative on that boundary.
  Parameter: NMAX is the largest anticipaped value of n
 *-----------------------------------------------------------------
 *       Numerical Recipies in Fortran 77
 *       Second Edition.
 *       The Art of Scientific Computing
 *       W.H.Press, S.A.Teukoisky, W.T.Vetterling, Brian P.Flannery
 *       Cambridge University Press. 1992,
 *       V.1, Chapter 3.3 "Cubic spline interpolation"
 *
 */
void spline (x, y, n, yp1, ypn, y2)
double *x, *y;
int n;
double yp1, ypn;
double *y2;
{
	register int i;
	double p, qn, sig, un, *u;

	/* here (n+1) instead of NMAX  */
	if((u=(double*)calloc(n+1, sizeof *u)) == (double *)0) {
		PR(ER, "spline interpolation: no memory...\n");
		exit(1);
	}

	if (yp1 > .99e30) {  /* The lower boundary condition is set */
		y2[0] = 0.;  /* either to be "natural"              */
		u[0] = 0.;
	} else {  /* or else to have a specified first derivative */
		y2[0] = -0.5;
		u[0] = (3/(x[1]-x[0])) * ((y[1]-y[0])/(x[2]-x[1])-yp1);
	}

	/*
	   This is the decomposition loop of the tridiagonal algorithm.
	   y2 and u are used for temporary storage of the decomposed
	   factors.
	 */
	for (i=1; i<n-1; i++) {
		sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
		p = sig * y2[i-1] + 2.;
		y2[i] = (sig-1.) / p;
		u[i] = (6.*((y[i+1]-y[i]) / (x[i+1]-x[i]) -(y[i]-y[i-1]) /
			(x[i] - x[i-1]))/(x[i+1]-x[i-1])-sig*u[i-1]) / p;
	}

	/* The upper boundary condition is set either to be "natural" */
	if (ypn > .99e30) {
		qn = 0.;
		un = 0.;
	} else {  /* or else to have a specified first derivative */
		qn = 0.5;
		un=(3./(x[n-1]-x[n-2]))*(ypn-(y[n-1]-y[n-2])/(x[n-1]-x[n-2]));
	}
	y2[n-1] = (un-qn*u[n-2]) / (qn*y2[n-2]+1.);

	/* This is the backsubstitution loop of the tridiagonal algorithm */
	for (i=n-2; i>=0; i--)
		y2[i] = y2[i]*y2[i+1] + u[i];

	free((void *)u);

	return;
}

/*
    Given the arrays xa(1:n) and ya(1:n) of lenth n, which tabulate
    a function (with the za_i's in order), and given the array y2a (1:n),
    which is the output from spline abouve and given a value of x, this
    routine returns a cubic spline interpolated value y.
 */
void splint (xa, ya, y2a, n, x, y)
double x, *y;
double *xa, *ya, *y2a;
int n;
{
	int k, khi, klo;
	double a,b,h;

	/*
	  We will find the right place in the table by means of bisection.
	  This is optinal if sequential calls are in order, and closely
	  spaced, one would do better to store previous values of
	  klo and khi and test if they remain appropriate on the next call.
	 */
	klo = 0;
	khi = n-1;
label_1:
	if (khi - klo > 1 ) {
		k = (khi + klo) / 2;
		if (xa[k] > x) {
			khi = k;
		} else {
			klo = k;
		}
		goto label_1;
	}   /* klo and khi now bracket the input value of x */

	h = xa[khi] - xa[klo];
	if ( ABS(h) < 1.e-16 ) {
	    PR(ER, "bad xa(%g for %d<n<%d) input in splint. ", x, khi,klo);
	    PR(ER, "The xa's must be distinct.\n");
	    *y = ya[klo];
	}  else {
	    /* Cubic spline polynomial is now evaluated */
	    a = (xa[khi] - x) /h;
	    b = (x - xa[klo]) /h;
	    *y = a*ya[klo] + b*ya[khi] +
		    ((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.;
	}
	return;
}
