# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>
# include <math.h>
# include "glesp.h"

#define PR      fprintf
#define ER      stderr
#define MOD(a,p)   ((a) - (p) * (int)((a)/(p)))
#define ABS(x)     ((x) < 0 ? -(x) : (x))

extern int max_double_(void);
/*
	SUBROUTINE lgndr_M(Lmx,m,cmm,c,x,glm)
C   Legandre functions F^m_k(x), k=m,m+1,...,Kmx
      DIMENSION glm(Lmx+1),c(2*Lmx+1)
c   Lim=Log(eps)< 0, ~ -20
*/
void lgndr_m_(Lmx,m_i,cmm,c,x_i,glm)
int *Lmx, *m_i;
double *cmm;
double *c, *x_i, *glm;
{
	register int i, k, Imax = 0;
	register int m = *m_i, m1 = m+1, m2 = m+2;
	register double x = *x_i;
	double vl = (double)1.0 - (double)(x)*(x) + 1.e-50;
	register double dfm=0.50*m*log10((double)vl);
	int Lmin = _lmin_gl;
	register double Enorm = _enorm_gl;    /* 10^Lmin */
	register double *gm1 = glm + m1, *c2 = c + *Lmx;

	if(ABS(dfm) > Lmin) Imax = -dfm/Lmin;
	dfm += Imax*Lmin;
	// glm[m1] = *cmm * pow(10., dfm);    /* P_m^m    */
	*gm1 = *cmm * pow(10., dfm);    /* P_m^m    */

	if(m == *Lmx || m1 == *Lmx) {
	    if(m == *Lmx) {
		while(Imax > 0) { *gm1 /= Enorm; Imax--; }
	    } else
		if(m1 == *Lmx) {
		    glm[m2] = x * c[m1] * *gm1;    /* P_m+1^m  */
		    while(Imax>0) {*gm1 /=Enorm; glm[m2]/=Enorm; Imax--;}
		}
	    return;
	}
	glm[m2] = x * c[m1] * *gm1;    /* P_m+1^m  */

	for(k=m2; k <= *Lmx; k++)  { /* P_m+l^m */
	    register int kk, k1=k+1;
	    register double *ck = c+k, *c2k = c2 + k, *g = glm + k;

	    /* recursional relation */
	    // glm[k1] = x * c[k]*glm[k] - c[*Lmx+k]*glm[k-1];
	    *(g+1) = x *  *ck  *  *g - *c2k  *  *(g-1);

	    if(ABS(glm[k1]) > Enorm) {
	       for(kk=m1; kk <= k1; kk++) {
			glm[kk] /= Enorm;
			if(Imax >= 2) glm[kk] /= Enorm;
	       }
	       if(Imax >= 2) Imax -= 2; else Imax--;
	    }
	}
	if(Imax >= 1) {
		register int l1 = *Lmx + 1;
		for(i=1; i<=Imax; i++) {
			for(k=m+1; k <= l1; k++)
				glm[k] /= Enorm;
		}
	}

	return;
}
/*
 *    dimension c(2*Lmx+1)
 *    preparation of coefficients for recurrent
 *    relation for P_l^m
 */
void start_clm_(Lmx,m_i,cmm,c)
int *Lmx;
int *m_i;
double *cmm,
       *c;
{
	register int k, k1, k2, km2;
	register int m = *m_i, mm = m+m, m2=m*m;

	*cmm = 0.5;
	if(m > 0) {
		for(k=2; k<=mm; k+=2)  *cmm *= (double)(k+1)/k;
	}
	*cmm = (1-2*MOD(m,2)) * sqrt(*cmm);

	k = m+1;
	k2 = k*k;
	c[k] = sqrt((double)(4*k2-1)/(k2-m2));
	c[*Lmx+k] = 0.;

	if(m != *Lmx) {
	   register double *c2 = c + *Lmx;
	   for(k=m+2; k <= *Lmx; k++) {
	      k1 = k-1; k2 = k*k; km2 = k2-m2;
	      c[k] = sqrt((double)(4*k2-1)/km2);
	      c2[k]= sqrt((k+k+1.)*(k1*k1-m2)/(double)(k+k-3)/km2);
	   }
	}
	return;
}

/*
C   Legendre polynomials F^0_l(x)
      DIMENSION glm0(2*Lmx+1)
*/
void lgndr_0_(Lmx,x,glm0)
int *Lmx;
double *x,
       *glm0;
{
	register int i;
	double x2 = *x * *x;

	glm0[*Lmx+1] = sqrt(0.50);                /* P_0 */
	glm0[*Lmx+2] = sqrt(1.50) * *x;           /* P_1 */
	glm0[*Lmx+3] = sqrt(2.50)*(1.50*x2-0.50); /* P_2 */

	if(*Lmx == 2)
		return;

	glm0[*Lmx+4] = sqrt(3.50) * *x *(2.50*x2-1.50); /* P_3 */
	if(*Lmx == 3)
		return;

	for(i=4; i <= *Lmx; i++) {
	   double eps  = 1.-1./(double)i;         /* i=l */
	   double eps1 = i+0.50;
	   double eps2 = sqrt(eps1/(eps1-1.));
	   int j = *Lmx+i+1;
	   eps1 = sqrt(eps1/(eps1-2.));
	   glm0[j] = *x*(1.+eps)*eps2*glm0[j-1] - eps*eps1*glm0[j-2];
	}

	return;
}

/*
 *       Legendre function F_m^m
 */
double f_mm_(m,x)
int *m;
double *x;
{
	register int k, mm=2 * *m;
	double cmm = 0.50;
	double xs = sqrt(1.0 - *x * *x);

	for(k=2; k <= mm; k+=2)  cmm *= (k+1.0)/k;
	cmm = (1-2*MOD(*m,2)) * sqrt(cmm);


	return cmm * pow(xs, (double)*m);
}

/*
	dimension abc(4*n),u(n)
*/
void  tridag_(abc,u,n,r1,rn)
double *abc,    /* numerated from 0 */
       *u;      /* here 'u' is 'glm' numerated from 1 to Lmax */
int *n;
double *r1, *rn;
{
/*
ccc    gam=abc(1:n),a=abc(n+1:2n),b=abc(2n+1:3n),c=abc(3n+1:4n)
*/
	register int j;
	double bet;
	int na = *n,
	    nb = *n + *n,
	    nc = *n + *n + *n;

	if ( fabs(abc[nb]) < 1.e-50)  {
		PR(ER, "tridag: rewrite equations or ");
		PR(ER, "use even number of rings...\n");
		exit(1);
	}
	bet = abc[nb];
	u[1] = *r1 / bet;
	for(j=2; j <= *n; j++) {
		abc[j-1] = abc[nc+j-2] / bet;
		bet = abc[nb+j-1] - abc[na+j-1] * abc[j-1];
		if(fabs(bet) < 1.e-50) {
			PR(ER, "tridag: failed...\n");
			exit(1);
		}
		u[j] = -abc[na+j-1] * u[j-1] / bet;
	}
	u[*n] = (*rn - abc[na+ *n -1] * u[*n-1]) / bet;
	for(j= *n-1; j>=1; j--)  u[j] -= abc[j] * u[j+1];

	return;
}
/*
 *  Legandre functions F^m_L(x), m=1,2,...,L
      DIMENSION glm(2*Lmx+1),abc(4*Lmx)
 */
void lgndr_l_(Lmx,L,x,abc,glm)
int *Lmx;
int *L;
double *x;
double *abc, *glm;
{
	double x2 = *x * *x;
	if(*L == 2) {
		glm[1] = -sqrt(3.75) * *x * sqrt(1.0-x2);
		glm[2] =  sqrt(0.9375)*(1.-x2);
		return;
	}

	if( *L > 2) {
	    register int m;
	    int n = *L - 1;
	    int nb = 2*n;
	    int nc = 3*n;
	    double xs = 2.* *x / sqrt(1.0 - x2);
	    /*
	    C     abc(n+i)=a,abc(2n+i)=b,abc(3n+i)=c,
	    C   parameters of the matrix
	    */
	    double r1 = -glm[*Lmx+ *L+1]/sqrt((1.-1./ *L)*(1.+1./(1.+ *L)));
	    double rf = -f_mm_(L, x);
	    abc[n]=0.;                       /* a(1)=0 */
	    abc[nb]=xs/sqrt(*L * *L+ *L-2.); /* b(1)=0 */
	    abc[nc]=1.;                      /* c(1)=1 */
/*
PR(ER, "n=%d, nb=%d, L=%d, xs=%g, abc[nb]=%g\n",
n, nb, *L, xs, abc[nb]);
*/
	    for(m=1; m<=n-1; m++) {
	       double e1=sqrt((*L-m-1.)*(*L+m+2.));
	       abc[n+m]=sqrt((*L-m)*(*L+m+1.))/e1; /* a(m)   */
	       abc[nb+m]=xs*(m+1.)/e1;             /* b(m)   */
	       abc[nc+m]=1.;                       /* c(m)=1 */
	    }
	    abc[nc+n-1]=0.;                        /* c(L)=0 */

	    tridag_(abc,glm,&n,&r1,&rf);

	    glm[*L] = -rf;
	}
	return;
}
        
