#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"
#define PR  fprintf
#define ER  stderr

/*
 *   procedure of selection best fitting parameter alpha of two maps.
 *
 *   Oleg Verkhodanov, 29/11/2011, vo@sao.ru
 */

void patt_coef(f, pat, x, nx, np, nph, sigma, alpha)
double *f, *pat, *x;
int nx, np, *nph;
double sigma;
double *alpha;
{
    register int i, j, k;
    int k_steps = 1000, n_tot =0, k_min=0;
    double f_mean=0.0, f_var=0.0, pat_mean=0.0, pat_var=0.0;
    double alp, dalp, alpha_min = -1.0, alpha_max = 1.0, min = 1.e+30;
    double *sig = (double *)0;

    if ((sig = (double *)calloc(k_steps, sizeof *sig)) == (double *)0) {
	PR(ER, "%s: no memory for temporal arrays.\n", myname);
	exit(1);
    }

    /* find mean and variace */
    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
		f_mean += f[ii+j];
		pat_mean += pat[ii+j];
	}
	n_tot += nph[i];
    }
    f_mean /= n_tot;
    pat_mean /= n_tot;
    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
		double ff = f[ii+j] - f_mean, pp = pat[ii+j] - pat_mean;
		f_var += ff*ff;
		pat_var += pp*pp;
	}
    }
    f_var /= n_tot;
    pat_var /= n_tot;
    f_var = sqrt(f_var);
    pat_var = sqrt(pat_var);


    alpha_max =  alpha_max  *  10. * f_var/pat_var;
    alpha_min = -alpha_max;
    dalp = (alpha_max - alpha_min) / k_steps;
PR(ER, "alpha_max = %g, dalp = %g\n", alpha_max, dalp);
PR(ER, "aa = %g\n", f_mean/pat_mean);

    /* find alpha */
    for(k=0; k< k_steps; k++) {
	alp = alpha_min + dalp * k;
	for(i=0; i<nx; i++) {
	    int ii = i * np;
	    for(j=0; j<nph[i]; j++) {
		    double a = f[ii+j] - alp * pat[ii+j];
		    sig[k] += a * a;
	    }
	}
//printf ("%d %12.5e\n", k, sig[k]);
	if (sig[k] < min) {min = sig[k]; k_min = k;}
    }
PR(ER, "k_min = %d\n", k_min);
    if (k_min == 0 || k_min == k_steps) {
	PR(ER, "the searched alpha-parameter is out of checking range.\n");
	exit (1);
    }

    *alpha = alpha_min + dalp * k_min;

    free((void *)sig);
    return;
}


#define SMALL   1.e-30
#define PR      fprintf
#define ER      stderr

#define Q(x)    ((x)*(x))
#define DET(a1,a2,a3,a4,a5,a6,a7,a8,a9)  ((a1)*(a5)*(a9) + (a2)*(a6)*(a7) + \
    (a3)*(a4)*(a8) - (a3)*(a5)*(a7) - (a6)*(a8)*(a1) - (a9)*(a2)*(a4))

#define N_N 3
/*
 *  compact method of equation sets decision
 */
#ifdef __GNUC__
int equat(double a[N_N][N_N], double *b, double *x, int n)
#else
int equat(a,b,x,n)
double  a[N_N][N_N], /*    square matrix of coefficients in [n]x[n]      */
	*b,          /*    array of equations meanings(free parameters)  */
	*x;          /*    matrix of variables                           */
int      n;          /*    dimension of arrays                           */
#endif
{
    register int i,j;
    int k;
    double   d;
    for(k=0; k < n-1; k++) {
	/* to be sure that a_kk != 0 */
	for(i=k; i < n; i++)
	      if(ABS(a[i][k]) > SMALL)
		break;
	if ( i == n ) {
	    PR(ER, "equat: can't diagonalize matrix...\n");
	    return -1;
	} else
	    if ( ABS(a[k][k]) < SMALL) {
		for(j=k; j < n; j++)
		    a[k][j] += a[i][j];
		b[k] += b[i];
	    }

	/* make a_ik = 1 */
	for(i=k; i<n; i++) {
	    d = a[i][k];
	    if(ABS(d) > SMALL) {
		for(j=k; j<n; j++)
		       a[i][j] /= d;
		b[i] /= d;
	    }
	}

	/* make matrix triangular */
	for(i=k+1; i<n; i++) {
	    if (ABS(a[i][k]) > SMALL ) {
		for(j=k; j<n; j++)
		       a[i][j] -= a[k][j] ;
		b[i] -= b[k];
	    }
	}
    }

    x[n-1] = b[n-1] / a[n-1][n-1];
	for(i=n-2; i>=0; i--) {
	d = 0.;
	for(j=i+1; j<n; j++)
	    d += x[j] * a[i][j];
	x[i] = b[i] - d;
    }
    return 1;
}


void dipole_fit(f,x,nx,np,nph,mask,
    tx,ty,tz, dtx,dty,dtz, t_ex,b_ex,l_ex, dt_ex,db_ex,dl_ex)
double *f, *x;
int nx, np,*nph;
int *mask;
double *tx,*ty,*tz, *dtx,*dty,*dtz,
       *t_ex, *b_ex, *l_ex, *dt_ex, *db_ex, *dl_ex;
{
    register int i, j, k=0;
    int n_tot = 0;
    double lon, lat;
    double (*xa)[3], *ya, a[3][3], b[3], z[3];
    double pi = M_PI, ss=0., dta = 0.;

    for(i=0; i<nx; i++) {
	int ii = i* np;
	for(j=0; j<nph[i]; j++)
	    if (!mask[ii+j])  n_tot++;
    }

    if ((xa=(double (*)[3])calloc(n_tot, sizeof (*xa))) ==(double (*)[3])0) {
nomem:
	PR(ER, "%s: no memory for temporal array for equations.\n", myname);
	exit(1);
    }
    if ((ya = (double *)calloc(n_tot, sizeof (*ya))) == (double *)0)
	goto nomem;

    for(i=0; i<nx; i++) {
	int ii = i* np;
	double dphi = 2*pi / nph[i];
	lat = pi/2. - acos(x[i]);
	for(j=0; j<nph[i]; j++) {
	    if (mask[ii+j]) continue;
	    lon = dphi * j;
	    /* T(l,b) = Tx cos(l)cos(b) + Ty sin(l)cos(b) + Tz sin(b) */
	    xa[k][0] = cos (lon) * cos(lat);
	    xa[k][1] = sin (lon) * cos(lat);
	    xa[k][2] = sin (lat);
	    ya[k] = f[ii+j];
	    k++;
	}
    }
    n_tot = k;

    for(i=0; i<3; i++) {
	for(j=0; j<3; j++) {
	    a[i][j] = 0.;
	    for(k=0; k<n_tot; k++) a[i][j] += xa[k][i]*xa[k][j];
	}
	b[i] = 0.;
	for(k=0; k<n_tot; k++)  b[i] += ya[k] * xa[k][i];
    }

    equat(a,b,z,3);

    *tx = z[0];
    *ty = z[1];
    *tz = z[2];

    for(k=0;k<n_tot;k++) {
	double d = ya[k] - z[0]*xa[k][0] - z[1]*xa[k][1] - z[2]*xa[k][2];
	ss += Q(d);
    }
    ss = sqrt(ss / (n_tot-3));
    dta=DET(a[0][0],a[0][1],a[0][2],a[1][0],a[1][1],a[1][2],
	a[2][0],a[2][1],a[2][2]);
    *dtx = ss * sqrt(ABS((a[1][1]*a[2][2]-a[1][2]*a[2][1])/dta));
    *dty = ss * sqrt(ABS((a[0][0]*a[2][2]-a[0][2]*a[2][0])/dta));
    *dtz = ss * sqrt(ABS((a[0][0]*a[1][1]-a[1][0]*a[0][1])/dta));

    *t_ex = sqrt (Q(*tx) + Q(*ty) + Q(*tz));
    /* +-atan(Tz / sqrt(Tx*Tx + Ty*Ty)) */
    *b_ex = atan(*tz / sqrt(Q(*tx) + Q(*ty)));
    *l_ex = atan2(*ty, *tx);
    if (*l_ex < 0) *l_ex += 2*M_PI;

    /* differentials */
    *dt_ex = (*tx * *dtx + *ty * *dty + *tz * *dtz) / *t_ex;
    ss = Q(*tx) + Q(*ty);
    dta = 1. / (1. + Q(*tz)/ss);
    *db_ex = dta * sqrt(ss) * (*dtz - (*dtx + *dty) * *tz / ss);
    *dl_ex = (*tx * *dty - *ty * *dtx) / ss;

    for(i=0; i<nx; i++) {
	int ii = i* np;
	double dphi = 2*pi / nph[i];
	lat = pi/2. - acos(x[i]);
	for(j=0; j<nph[i]; j++) {
	    lon = dphi * j;
	    f[ii+j] = *tx * cos(lon) * cos(lat) +
		    *ty * sin(lon) * cos(lat) + *tz * sin(lat);
	}
    }
    free((void *)xa);
    free((void *)ya);
    return;
}
