#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"
#define PR  fprintf
#define ER  stderr

/*
 *   procedure of mapping of re-pixelization
 *
 *   Oleg Verkhodanov, 13/07/2009, vo@sao.ru
 */

void map_repix(f,x,nx,np,nph, repix_win, rmap, xc, nxc, npc, nphc)
double *f, *x;
int nx, np, *nph;
double repix_win;        /* re-pixelization window, min */
double **rmap,           /* output array            */
       **xc;             /* output X-array          */
int *nxc, *npc, **nphc;  /* output nx,np,nph        */
{
    register int i, j;
    int n;
    double x1, x2;
    int pin_flag = nph[0] == np ? 0 : 1;

    /* when repix_win is the multiplication factor */
    if ( repix_win < 0. && fabs(repix_win+1.) > 1.e-5 ) {
	double dp_min = (2.*M_PI / np) * 180/M_PI * 60.;
	repix_win = dp_min / fabs(repix_win);
    }

    *nxc = 180 * 60 / repix_win + 0.5;
    repix_win = 180. * 60. / *nxc;       /* in minutes */

    PR(ER, "\nactual re-pixelization window of %gx%g (%g) sq.arcmin.\n",
	    repix_win, repix_win, repix_win*repix_win);
    *npc = 2 * *nxc;
    PR(ER, "new map resolution: nx=%d, np=%d, lmax=%d.\n",
	    *nxc, *npc, (*nxc-1)/2);

    n = *nxc * *npc;
    if ((*rmap = (double *)calloc((unsigned)n, sizeof **rmap))==(double *)0){
nomem:
	PR(ER, "%s: no memory for output arrays...\n", myname);
	exit(1);
    }
    if ((*xc = (double *)calloc((unsigned)(*nxc), sizeof **xc))==(double *)0)
	goto nomem;
    if ((*nphc = (int *)calloc((unsigned)(*nxc), sizeof **nphc))==(int *)0)
	goto nomem;

    x1 = -1.0; x2 = 1.0;
    gauleg(x1, x2, *xc, (double *)0, *nxc);
    if (pin_flag) {
	/* GLESP pixelization */
//      for (i=0; i < *nxc; i++)  (*nphc)[i] = get_nphi_by_x(*xc, *nxc, i);
	for (i=0; i < *nxc; i++)
	    (*nphc)[i] = get_nphi_by_x2(*xc, *nxc, *npc, i);
    } else {
	/* pin = 0 */
	for (i=0; i < *nxc; i++)  (*nphc)[i] = *npc;
    }


    n = *nxc - 1;
    for (i=0; i < *nxc; i++) {
	register int ii = i * *npc;
	double dphi = 2*M_PI / (*nphc)[i];
	double theta = acos((*xc)[i]);
	int pi, pj;

	for (j=0; j < (*nphc)[i]; j++) {
	    double phi = j * dphi;

	    angle2pixel(x, nx, nph, &pi, &pj, theta, phi);
	    (*rmap)[ii+j] = f[pi*np+pj];
	}
    }

    return;
}


