#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

/*
 * Output vector graphics on a sphere
 *    Oleg Verkhodanov, 05/2009, 03/2012
 *
 */

#define PR            fprintf
#define ER            stderr
#define MAX(x,y)      ((x) > (y) ?  (x) : (y))
#define ABS(x)        ((x) < 0   ? -(x) : (x))


extern void P_bcolor(int), P_color(int), P_fcircle(int, int, int),
       P_marker(int, int), P_p_line(int *, int), P_f_pol(int *, int),
       P_circle(int, int, int),
       P_arc(double,double,double,double,double),
       P_linemod(char *), P_line(int, int, int, int),
       P_move(int,int), P_cont(int,int), P_point(int,int),
       P_text(int,int,int,int,char *), P_text_label(int,int,int,int,char *),
       P_openpl(), P_erase();
extern void Pg_bcolor(int), Pg_color(int), Pg_fcircle(double, double, double),
       Pg_marker(double,int), Pg_p_line(double *,int), Pg_f_pol(double *,int),
       Pg_circle(double,double,double),
       Pg_arc(double,double,double,double,double),
       Pg_linemod(char *), Pg_line(double,double,double,double),
       Pg_move(double,double), Pg_cont(double,double),Pg_point(double,double),
       Pg_text(int,int,int,int,char *), Pg_text_label(int,int,int,int,char *),
       Pg_openpl(), Pg_erase();
extern int symb(FILE *, int, int *);
extern double str2deg(char *);

static int tsize = 15, italic = 0, angle=0, font=1;
static int n_l = 2000, n_b=1000;
static int *mpv =  (int *)0;
struct _plain_cart_ {
	float theta,phi;
	int col;
};
static struct _plain_cart_ *mpv_cart =  (struct _plain_cart_ *)0;
static int *mpv_glob =  (int *)0;
static int i_t0=0,  j_p0=0;
static int ig_t0=0, jg_p0=0;
static double curr_lat = 0.0, curr_lon = 0.0;
static int cur_linemode = 0;
# define L_SOLID        0
# define L_DOTTED       1
# define L_DOTDASHED    2
# define L_SHORTDASHED  2
# define L_LONGDASHED   3
static int nx_gd, ny_gd;
static int cur_x = 0, cur_y = 0;

double getl(FILE *, char *, int, int *);


/* put plot-command on a sphere */
void vect2mol(plotfile, lon0)
char *plotfile;
double lon0;
{
    extern int get_nx_gd(), get_ny_gd();
    extern void read_vect(char *);
    extern void move_gd(int, int);

    int y_c, x_c, x_r, y_r;
    register int i, j;
    double pi= M_PI;
    extern void point_gd (int, int);

    nx_gd = get_nx_gd();
    ny_gd = get_ny_gd();

    mpv = (int *)calloc((unsigned)(n_l*n_b), sizeof *mpv);
    if (mpv == (int *)0) {
nomem:
	PR(ER, "%s: no memory.\n", myname);
	exit(1);
    }
    i = nx_gd*ny_gd;
    mpv_glob = (int *)calloc((unsigned)i, sizeof *mpv_glob);
    if (mpv_glob == (int *)0)  goto nomem;
    mpv_cart=(struct _plain_cart_ *)calloc((unsigned)i,sizeof *mpv_cart);
    if (mpv_cart == (struct _plain_cart_ *)0)  goto nomem;

    y_c = ny_gd / 2;
    x_c = nx_gd / 2;
    y_r = ny_gd / 2;
    x_r = nx_gd / 2;

    /* put copy of base map to memory, select Molweide projection */
    for(i=0; i<ny_gd; i++) {
	double u, v = (i - y_c) / (y_r/1.02);
	int ii = nx_gd * i;

	for(j=0; j<nx_gd; j++) {
	    double phi = 1e20, theta=1e20;
	    u = 2. * (j - x_c) / (x_r/1.02);
	    if ( u*u/4. + v*v <= 1.) {
		double sqv = sqrt(1.-v*v);
		/* theta in [0, pi] */
		theta = pi/2. - asin(2./pi * (asin(v) + v*sqv));
		/* lon in [-pi,pi], the minus sign for astro convention */
		phi = -lon0 - pi/2 * u / MAX(sqv, 1.e-6);
		/* phi in [0, 2pi] */
		if ( phi < 0.0 )    phi += 2. * pi;
	    }
	    mpv_cart[ii+j].theta = theta;
	    mpv_cart[ii+j].phi = phi;
	}
    }

    /* read file with vector graphics and copy it to mpv-array */
    read_vect(plotfile);

    /* put copy of base map to memory */
    y_c = n_b * n_l;
    x_c = nx_gd * ny_gd;
    for(i=0; i<ny_gd; i++) {
	double phi, theta;
	int ii = nx_gd * i;
	for(j=0; j<nx_gd; j++) {
	    int nr = ii+j;
	    theta = mpv_cart[nr].theta;
	    phi   = mpv_cart[nr].phi;

	    /* keep only meaningful points of 'mpv' */
	    if (ABS(theta) < 1.e5) {
		double db = pi/n_b, dl = 2*pi/n_l;
		int ib = theta / db,
		    jl = (phi < pi ? pi - phi : 3*pi - phi) / dl,
		    n = ib * n_l + jl,
		    /* some neibouring pixels to n-th pixel */
		    n1 = (ib-1) * n_l + jl,
		    n2 = (ib+1) * n_l + jl,
		    n3 =  ib    * n_l + jl-1,
		    n4 =  ib    * n_l + jl+1;

		if (n1 < 0 || n1 >= y_c)  n1 = -1;
		if (n2 < 0 || n2 >= y_c)  n2 = -1;
		if (n3 < 0 || n3 >= y_c)  n3 = -1;
		if (n4 < 0 || n4 >= y_c)  n4 = -1;
		if (mpv[n] ||
		    ( n1 >= 0 && mpv[n1] ) ||
		    ( n2 >= 0 && mpv[n2] ) ||
		    ( n3 >= 0 && mpv[n3] ) ||
		    ( n4 >= 0 && mpv[n4] )
		) {
		      mpv_glob[nr] = 1;  /* locate in general map */
		      mpv[n] = 0;
		}
	    }
	    /* keep only meaningful points of 'mpv_cart' */
	    {
		int n1 = (i-1) * nx_gd + j,
		    n2 = (i+1) * nx_gd + j,
		    n3 =  i    * nx_gd + j-1,
		    n4 =  i    * nx_gd + j+1;

		if (n1 < 0 || n1 >= x_c)  n1 = -1;
		if (n2 < 0 || n2 >= x_c)  n2 = -1;
		if (n3 < 0 || n3 >= x_c)  n3 = -1;
		if (n4 < 0 || n4 >= x_c)  n4 = -1;
		if (mpv_cart[nr].col
//               || ( n1 >= 0 && mpv_cart[n1].col ) ||
//                  ( n2 >= 0 && mpv_cart[n2].col ) ||
//                  ( n3 >= 0 && mpv_cart[n3].col ) ||
//                  ( n4 >= 0 && mpv_cart[n4].col )
		) {
		    /* locate in general map */
		    mpv_glob[nr] = mpv_cart[nr].col;
		    mpv_cart[nr].col = 0;
		}
	    }
	}
    }

    /* put map with vector graphics on the image */
    for(i=0; i<ny_gd; i++) {
	for(j=0; j<nx_gd; j++) {
	    if (mpv_glob[i*nx_gd+j] > 1e-29) {
		move_gd(j,i);
		point_gd(j, i);
	    }
	}
    }

    free((void *)mpv);
    free((void *)mpv_cart);
    free((void *)mpv_glob);
    return;
}
void read_vect(plotfile)
char *plotfile;
{
    register int i;
    char str[81], la[81];
    FILE *fp = (FILE *)0;
    int a1, a2, a3, a4;
    double b1, b2, b3, b4;
    int c, n;
    int end_flag=0;
    int nc=0;
    double *coor = (double *)0;

    if ((fp = fopen(plotfile, "r")) == NULL) {
	PR(ER, "%s: I can not open file '%s'.\n", myname, plotfile);
	exit(1);
    }

    do {
	c = fgetc(fp);

	while(c ==' ' || c=='\n' || c=='\t' || c == '\r')  c = fgetc(fp);

	if ( feof(fp) )  { /* end of file */
		if ( end_flag == 0 ) {
			P_erase();
			end_flag = 1;
		}
		break;
	}

	if ( feof(fp) )  { /* end of file */
		if ( end_flag == 0 ) {
			P_erase();
			end_flag = 1;
		}
		break;
	}

	if (end_flag) {
		P_openpl(NULL);
		end_flag = 0;
	}

	switch(c) {
	case '#':  /* comment */
		c = fgetc(fp);
		if ( c == '@' ) {
		      register char *cs=NULL;
		      *cs++ = '@';
		      while((c=fgetc(fp)) !='@' && c!=EOF)
			    { *cs++=c; }
		      *cs++ = '@';
		      *cs = '\0';
		} else {
		      while((c=fgetc(fp)) !='\n' && c!=EOF) ;
		}
		break;
	case 'F':  /* X font */
		for(i=0; (str[i]=fgetc(fp)) != '\n'; i++) ;
		if(str[i-1] == '\r')  str[i-1] = '\0';
		str[i] = '\0';
		break;
	case 'B':
		a1 = getl(fp,la,0, &nc) + 0.5;
		P_bcolor(a1);
		break;
	case 'C': /* color of text */
		a1 = getl(fp,la,0, &nc) + 0.5;
		P_color(a1);
		break;
	case 'I': /* filled circle */
		getl(fp,la,0, &nc);
		b1 = str2deg(la);    /* g.lat. */
		getl(fp,la,0, &nc);
		b2 = str2deg(la);    /* g.lon. */
		getl(fp,la,0, &nc);
		b3 = str2deg(la);    /* radius */
		Pg_fcircle(b1,b2,b3);
		break;
	case 'M': /* marker */
		getl(fp,la,0, &nc);
		b1 = str2deg(la);               /* radius, degs */
		a2 = getl(fp,la,0, &nc) + 0.5;  /* marker type  */
		Pg_marker(b1, a2);
		break;
	case 'P':   /* zigzag */
		n = getl(fp,la,0, &nc) + 0.5;
		if((coor = (double *)calloc( (unsigned)n * 2,
			sizeof *coor )) == (double *)0) {
			PR(ER,"%s: I can't find memory...\n",
				myname);
			exit(1);
		}
		getl(fp,la,0, &nc);
		b1 = str2deg(la);    /* g.lat. */
		getl(fp,la,0, &nc);
		b2 = str2deg(la);    /* g.lon. */
		coor[0] = b1; coor[1] = b2;
		for(i=1; i<n; i++) {
		    getl(fp,la,0, &nc);
		    b1 = str2deg(la);    /* g.lat. */
		    /* unexpected finish of command */
		    if(symb(fp, 0, &nc)) {
			   n = i;
			   break;
		    }

		    getl(fp,la,0, &nc);
		    b2 = str2deg(la);    /* g.lon. */

		    /* unexpected finish of command */
		    if(symb(fp, 0, &nc) && i < n-1) {
			   n = i;
			   break;
		    }

		    coor[2*i]   = b1;
		    coor[2*i+1] = b2;
		}
		Pg_p_line(coor, n);
		free((char *)coor);
		break;
	case 'N':  /* filled polygon */
		n = getl(fp,la,0, &nc) + 0.5;
		if((coor = (double *)calloc( (unsigned)n * 2,
			sizeof *coor )) == (double *)0) {
			PR(ER,"%s: I can't find memory...\n", myname);
			exit(1);
		}
		getl(fp,la,0, &nc);
		b1 = str2deg(la);    /* g.lat. */
		getl(fp,la,0, &nc);
		b2 = str2deg(la);    /* g.lon. */
		coor[0] = b1; coor[1] = b2;
		for(i=1; i<n; i++) {
			getl(fp,la,1, &nc);
			b1 = str2deg(la);    /* g.lat. */
			/* unexpected finish of command */
			if(symb(fp, 0, &nc)) {
			       n = i;
			       break;
			}

			getl(fp,la,1, &nc);
			b2 = str2deg(la);    /* g.lon. */

			/* unexpected finish of command */
			if(symb(fp, 0, &nc) && i < n-1) {
			       n = i;
			       break;
			}

			coor[2*i]   = b1;
			coor[2*i+1] = b2;
		}
		Pg_f_pol(coor, n);
		free((char *)coor);
		break;
	case 'E':
	case 'e':
		P_erase();
		if((c = fgetc(fp)) != EOF) {
		    /* P_closepl(); */
		    ungetc(c,fp);
		}
		end_flag = 1;
		break;
	case 'c':
		getl(fp,la,0, &nc);
		b1 = str2deg(la);    /* g.lat. */
		getl(fp,la,0, &nc);
		b2 = str2deg(la);    /* g.lon. */
		getl(fp,la,0, &nc);
		b3 = str2deg(la);    /* radius */
		Pg_circle(b1,b2,b3);
		break;
	case 'a':
	      {
		double alpha;
		b1 = getl(fp,la,0, &nc);
		b2 = getl(fp,la,0, &nc);
		b3 = getl(fp,la,0, &nc);
		b4 = getl(fp,la,0, &nc);
		alpha = getl(fp,la,0, &nc);
		Pg_arc(b1,b2,b3,b4,alpha);
		break;
	      }
	case 'f':
		getl(fp,la,1,&nc);
		P_linemod(str);
		break;
	case 'l':
		getl(fp,la,0, &nc);
		b1 = str2deg(la);
		getl(fp,la,0, &nc);
		b2 = str2deg(la);
		getl(fp,la,0, &nc);
		b3 = str2deg(la);
		getl(fp,la,0, &nc);
		b4 = str2deg(la);
		Pg_line(b1,b2,b3,b4);
		break;
	case 'm':
		getl(fp,la,0, &nc);
		b1 = str2deg(la);    /* g.lat. */
		getl(fp,la,0, &nc);
		b2 = str2deg(la);    /* g.lon. */
		Pg_move(b1,b2);
		break;
	case 'n':
		getl(fp,la,0, &nc);
		b1 = str2deg(la);    /* g.lat. */
		getl(fp,la,0, &nc);
		b2 = str2deg(la);    /* g.lon. */
		Pg_cont(b1,b2);
		break;
	case 'p':
		getl(fp,la,0, &nc);
		b1 = str2deg(la);    /* g.lat. */
		getl(fp,la,0, &nc);
		b2 = str2deg(la);    /* g.lon. */
		P_point(b1,b2);
		break;
	case 'T':
		a1 = getl(fp,la,0, &nc) + 0.5;
		a2 = getl(fp,la,0, &nc) + 0.5;
		a3 = getl(fp,la,0, &nc) + 0.5;
		a4 = getl(fp,la,0, &nc) + 0.5;
		getl(fp,la,1, &nc);

		tsize = a1; italic = a2; angle=a3; font=a4;
		P_text(a1,a2,a3,a4,str);
		break;
	case 't':
		getl(fp,la,1,&nc);
		P_text_label(tsize,italic,angle,font,la);
		break;
	default:
		break;
	}
    } while (!feof(fp)) ;

    if ( feof(fp) && !end_flag ) {
	    P_erase();
    }

    fclose(fp);
    return;
}

void P_openpl()
{
}

void P_erase()
{
}

double getl(fps, str,key, num)
register FILE *fps;  /* input file stream    */
register char *str;  /* output string        */
int key;             /* key=1 to read string, key=0 to read number */
int *num;            /* number of read bytes */
{
	register int c, i;

	*num = 0;
	while((c=fgetc(fps)) ==' ' || c=='\n' || c=='\t' || c=='\r')
	       (*num)++;
	if ( c == '#' ) {
error_label:
	     PR(ER,"\n%s: comment symbol '#' in a command of PLOT-file...\n",
		   myname);
	     exit(1);
	}
	str[0] = c; i=1; (*num)++;
	if(key) /* read string */
		while((c=fgetc(fps)) != '\n' ) {
			str[i] = c;
			i++; (*num)++;
		}
	else   /* read number */
		while((c=fgetc(fps)) !=' ' && c!='\n' && c!='\t') {
			if ( c == '#' )  goto error_label;
			str[i] = c;
			i++; (*num)++;
		}
	(*num)++;
	if(str[i-1] == '\r')  str[i-1] = '\0';
	str[i] = '\0';

	return(atof(str));
}

int symb(fp, key, num)
register FILE *fp;
int key;
int *num;
{
       register int c;
       register char *s;
       char *special_symb="BCEFHMNPTaceflmnpst";

       *num = 0;
       switch(key) {
       case 0:
	      return 0;
       case 1:
	      key = 0;
	      while((c=fgetc(fp)) ==' ' || c=='\n' || c=='\t' || c=='\r')
		      (*num)++;
	      s = special_symb;
	      while(c != *s && *s )  s++;
	      if(*s)   key = 1;
	      ungetc(c, fp);
	      break;
       }
       return key;
}


void P_bcolor(c)
int c;
{
    return;
}
void P_color(c)
int c;
{
    return;
}
void Pg_fcircle(b,l,r)
double b,l,r;
{
    register int i, j;
    int l_count = 0, l_count_lim =0, l_count_space = 0;
    int j1, j2, i1, i2;
    double dt = M_PI/n_b;
    double dp = 2*M_PI/n_l;
    double theta0= (90 - b) * M_PI/180., phi0 = l * M_PI/180.;
    double cth0, sth0, c = 1.5;
    double radius = r * M_PI/180.;
    int cur_color = 1;
    int io = -1;

    switch (cur_linemode ) {
    case L_DOTTED:
	l_count_lim=4;
	l_count_space=l_count_lim+4;
	break;
    case L_SHORTDASHED:
	l_count_lim=32;
	l_count_space=l_count_lim+16;
	break;
    case L_LONGDASHED:
	l_count_lim=64;
	l_count_space=l_count_lim+32;
	break;
    }

    if (phi0 < M_PI) {
	j_p0 = (M_PI-phi0) / dp;
    } else { /* if (phi > M_PI) { */
	j_p0 = (3*M_PI-phi0) / dp;
    }
    i_t0 = theta0 / dt;

    if (theta0 - c*radius < 0 )  {
	i1 = 0;
	i2 = (theta0 + c*radius) / dt;
	j1 = 0;
	j2 = n_l-1;
    } else if (theta0 + c*radius > 90 )  {
	i1 = (theta0 + c*radius) / dt;
	i2 = n_b-1;
	j1 = 0;
	j2 = n_l;
    } else {
	j1 = (M_PI-phi0+c*radius) / dp;
	j2 = (M_PI-phi0-c*radius) / dp;
	i1 = (theta0 - c*radius) / dt;
	i2 = (theta0 + c*radius) / dt;
    }

    cth0 = cos(theta0);
    sth0 = sqrt(1-cth0*cth0);

    for(i=i1; i<=i2; i++) {
	double b_i = M_PI/2. - i * dt, theta = (M_PI/2 - b_i);
	double cth = cos(theta), sth = sqrt(1-cth*cth);
	int ii = i * n_l;
	int ja = j1 < j2 ? j1 : j2, jb = j1 > j2 ? j1 : j2;
	for(j=ja; j<=jb; j++) {
	    double l_j = M_PI - j * dt, phi = l_j, v;

	    /* distance 'v' between two pixels on a sphre */
	    dp = phi0 - phi;
	    v = acos(cos(dp)*sth0*sth + cth0*cth);

	    if (v <= radius) {
		  if(io<0) io=i;
		  if (cur_linemode) {
		      if (l_count < l_count_lim) {
			  mpv[ii+j] = cur_color;
			  l_count++;
		      } else if (l_count < l_count_space) {
			  l_count++;
		      } else l_count = 0;
		  } else
		      mpv[ii+j] = cur_color;
		  if (ABS(io-i) > 1) {
		    int is, iia = io < i ? io : i, iib = io > i ? io : i;
		    for(is=iia; is<=iib; is++) {
			if (cur_linemode) {
			    if (l_count < l_count_lim) {
				mpv[is*n_l+j] = cur_color;
				l_count++;
			    } else if (l_count < l_count_space) {
				l_count++;
			    } else l_count = 0;
			} else
			    mpv[is*n_l+j] = cur_color;
		    }
		  }
		  io=i;
	    }
	}
    }
    curr_lat = b;
    curr_lon = l;

    return;
}

void P_fcircle(x,y,r)
int x,y,r;
{
    register int i, j;
    int l_count = 0, l_count_lim =0, l_count_space = 0;
    int j1, j2, i1, i2;
    double c = 1.5;
    double v;
    int cur_color = 1;
    int io = -1;

    switch (cur_linemode ) {
    case L_DOTTED:
	l_count_lim=4;
	l_count_space=l_count_lim+4;
	break;
    case L_SHORTDASHED:
	l_count_lim=32;
	l_count_space=l_count_lim+16;
	break;
    case L_LONGDASHED:
	l_count_lim=64;
	l_count_space=l_count_lim+32;
	break;
    }

    j_p0 = x;
    i_t0 = y;

    j1 = x - c*r ;
    j2 = x + c*r ;
    i1 = y - c*r ;
    i2 = y + c*r ;

    for(i=i1; i<=i2; i++) {
	int ii = i * nx_gd;
	int ja = j1 < j2 ? j1 : j2, jb = j1 > j2 ? j1 : j2;
	double yi2 = (y-i)*(y-i);
	for(j=ja; j<=jb; j++) {
	    /* distance 'v' between two pixels on a sphre */
	    v = sqrt((x-j)*(x-j) + yi2);

	    if (v <= r) {
		  if(io<0) io=i;
		  if (cur_linemode) {
		      if (l_count < l_count_lim) {
			  mpv_cart[ii+j].col = cur_color;
			  l_count++;
		      } else if (l_count < l_count_space) {
			  l_count++;
		      } else l_count = 0;
		  } else
		      mpv_cart[ii+j].col = cur_color;
		  if (ABS(io-i) > 1) {
		    int is, iia = io < i ? io : i, iib = io > i ? io : i;
		    for(is=iia; is<=iib; is++) {
			if (cur_linemode) {
			    if (l_count < l_count_lim) {
				mpv_cart[is*nx_gd+j].col = cur_color;
				l_count++;
			    } else if (l_count < l_count_space) {
				l_count++;
			    } else l_count = 0;
			} else
			    mpv_cart[is*nx_gd+j].col = cur_color;
		    }
		  }
		  io=i;
	    }
	}
    }

    return;
}

void Pg_marker(dsize, markertype)
double dsize;     /* size in degs     */
int markertype;   /* number of marker */
{
    register int i;
    int s = dsize * nx_gd/180.;
    int x, y, v;
    double min;
    extern void P_pointm(int, int, int, int);

    /* convert (b,l) to pixels (i,j) */
    x = nx_gd / 2;
    min = 1e30;
    for (y=i=0; i<ny_gd; i++) {
	double lat = 90. - mpv_cart[i*nx_gd + x].theta * 180./M_PI;
	if ( ABS(lat - curr_lat) < min ) {y = i; min = ABS(lat - curr_lat);}
    }
    min = 1e30;
    v = y * nx_gd;
    for (x=i=0; i<nx_gd; i++) {
	double lon = mpv_cart[v + i].phi * 180./M_PI, curr_b = curr_lon;
	if (curr_b < 0.) curr_b += 360.;
	if ( ABS(lon - curr_b) < min ) {x = i; min = ABS(lon - curr_b);}
    }

    P_pointm(x, y, s, markertype);
    return;
}
void P_marker(size, markertype)
int size,         /* size of marker by X */
    markertype;   /* number of marker    */
{
    int x = cur_x, y = cur_y;
    extern void P_pointm(int, int, int, int);

    P_pointm(x, y, size, markertype);
    return;
}
void P_p_line(coor, n)
int *coor;
int n;
{
    register int i;
    P_move(coor[0], coor[1]);
    for(i=1; i<n; i++) {
	int k = i*2;
	P_cont(coor[k], coor[k+1]);
    }
    return;
}
void Pg_p_line(coor, n)
double *coor;
int n;
{
    register int i;
    Pg_move(coor[0], coor[1]);
    for(i=1; i<n; i++) {
	int k = i*2;
	Pg_cont(coor[k], coor[k+1]);
    }
    return;
}
void P_f_pol(coor, n)
int *coor;
int n;
{
    register int i, j;
    int xmin=nx_gd, ymin=ny_gd, xmax=0, ymax=0;
    int *buf = (int *)0, x_n, y_n;
    int cur_col = 1;

    for(i=0; i<n; i++) {
	int k = i*2;
	x_n = coor[k], y_n = coor[k+1];
	if (x_n < xmin) xmin = x_n;
	if (x_n > xmax) xmax = x_n;
	if (y_n < ymin) ymin = y_n;
	if (y_n > ymax) ymax = y_n;
    }

    /* plot polygon */
    P_move(coor[0], coor[1]);
    for(i=1; i<n; i++) {
	int k = i*2;
	P_cont(coor[k], coor[k+1]);
    }
    P_cont(coor[0], coor[1]);

    x_n = xmax - xmin + 1;
    y_n = ymax - ymin + 1;
    if ((buf = (int *)calloc((unsigned)(x_n*y_n), sizeof *buf)) == (int*)0){
	PR(ER, "%s, P_f_pol: problem to find temporal memory.\n", myname);
	return;
    }
    /* copy array of polygon */
    for(i=0; i<y_n; i++) {
	int ii = i * y_n, ig = nx_gd * (i+ymin);
	for(j=0; j<x_n; j++) {
	    buf[ii+j] = mpv[ig+j+xmin];
	}
    }

    /* fill in lines between colr points */
    for(i=0; i<y_n; i++) {
	int ii = i * y_n, x_a = -1, x_b=0;
	for(j=0; j<x_n; j++) {
	    if ( buf[ii+j] != 0) {
		if (x_a < 0) {
		    x_a = j;
		} else {
		    int k;
		    x_b = j;
		    for (k=x_a; k<=x_b; k++) buf[ii+j] = cur_col;
		}   x_a = -1;
	    }
	}
    }

    /* copy array of polygon to general map back*/
    for(i=0; i<y_n; i++) {
	int ii = i * y_n, ig = nx_gd * (i+ymin);
	for(j=0; j<x_n; j++) {
	    mpv[ig+j+xmin] = buf[ii+j];
	}
    }

    if (buf != (int *)0) free((void *)buf);
    return;
}
void Pg_f_pol(coor, n)
double *coor;
int n;
{
    return;
}
void P_circle(x,y,r)
int x,y,r;
{
    register int i, j;
    int l_count = 0, l_count_lim =0, l_count_space = 0;
    int j1, j2, i1, i2;
    double c = 1.5;
    double dpx = 1.5, v;
    int cur_color = 1;
    int io = -1;

    switch (cur_linemode ) {
    case L_DOTTED:
	l_count_lim=4;
	l_count_space=l_count_lim+4;
	break;
    case L_SHORTDASHED:
	l_count_lim=32;
	l_count_space=l_count_lim+16;
	break;
    case L_LONGDASHED:
	l_count_lim=64;
	l_count_space=l_count_lim+32;
	break;
    }

    j_p0 = x;
    i_t0 = y;

    j1 = x - c*r ;
    j2 = x + c*r ;
    i1 = y - c*r ;
    i2 = y + c*r ;

    if (j1 < 0)      j1 =0;
    if (j2 >= nx_gd) j2 =nx_gd;
    if (i1 < 0)      i1 =0;
    if (i2 >= ny_gd) i2 =ny_gd;

    for(i=i1; i<=i2; i++) {
	int ii = i * nx_gd;
	int ja = j1 < j2 ? j1 : j2, jb = j1 > j2 ? j1 : j2;
	double yi2 = (y-i)*(y-i);
	for(j=ja; j<=jb; j++) {
	    /* distance 'v' between two pixels */
	    v = sqrt((x-j)*(x-j) + yi2);

	    if (ABS(v-r) <= dpx) {
//          if (v <= r) {
		  if(io<0) io=i;
		  if (cur_linemode) {
		      if (l_count < l_count_lim) {
			  mpv_cart[ii+j].col = cur_color;
			  l_count++;
		      } else if (l_count < l_count_space) {
			  l_count++;
		      } else l_count = 0;
		  } else
		      mpv_cart[ii+j].col = cur_color;
		  if (ABS(io-i) > 1) {
		    int is, iia = io < i ? io : i, iib = io > i ? io : i;
		    for(is=iia; is<=iib; is++) {
			if (cur_linemode) {
			    if (l_count < l_count_lim) {
				mpv_cart[is*nx_gd+j].col = cur_color;
				l_count++;
			    } else if (l_count < l_count_space) {
				l_count++;
			    } else l_count = 0;
			} else
			    mpv_cart[is*nx_gd+j].col = cur_color;
		    }
		  }
		  io=i;
	    }
	}
    }

    return;
}
void Pg_circle(b,l,r)
double b,l,r;
{
    register int i, j;
    int l_count = 0, l_count_lim =0, l_count_space = 0;
    int j1, j2, i1, i2;
    double dt = M_PI/n_b;
    double dp = 2*M_PI/n_l;
    double theta0= (90 - b) * M_PI/180., phi0 = l * M_PI/180.;
    double cth0, sth0, c = 1.5;
    double dpx = dt > dp ? dt : dp,  radius = r * M_PI/180.;
    int cur_color = 1;
    int io = -1;

    switch (cur_linemode ) {
    case L_DOTTED:
	l_count_lim=4;
	l_count_space=l_count_lim+4;
	break;
    case L_SHORTDASHED:
	l_count_lim=32;
	l_count_space=l_count_lim+16;
	break;
    case L_LONGDASHED:
	l_count_lim=64;
	l_count_space=l_count_lim+32;
	break;
    }

    if (phi0 < M_PI) {
	jg_p0 = (M_PI-phi0) / dp;
    } else { /* if (phi > M_PI) { */
	jg_p0 = (3*M_PI-phi0) / dp;
    }
    ig_t0 = theta0 / dt;

    if (theta0 - c*radius < 0 )  {
	i1 = 0;
	i2 = (theta0 + c*radius) / dt;
	j1 = 0;
	j2 = n_l-1;
    } else if (theta0 + c*radius > 90 )  {
	i1 = (theta0 + c*radius) / dt;
	i2 = n_b-1;
	j1 = 0;
	j2 = n_l;
    } else {
	j1 = (M_PI-phi0+c*radius) / dp;
	j2 = (M_PI-phi0-c*radius) / dp;
	i1 = (theta0 - c*radius) / dt;
	i2 = (theta0 + c*radius) / dt;
    }

    cth0 = cos(theta0);
    sth0 = sqrt(1-cth0*cth0);

    for(i=i1; i<=i2; i++) {
	double b_i = M_PI/2. - i * dt, theta = (M_PI/2 - b_i);
	double cth = cos(theta), sth = sqrt(1-cth*cth);
	int ii = i * n_l;
	int ja = j1 < j2 ? j1 : j2, jb = j1 > j2 ? j1 : j2;
	for(j=ja; j<=jb; j++) {
	    double l_j = M_PI - j * dt, phi = l_j, v;

	    /* distance 'v' between two pixels on a spehre */
	    dp = phi0 - phi;
	    v = acos(cos(dp)*sth0*sth + cth0*cth);

	   if (ABS(v-radius) <= dpx) {
//         if (v <= radius) {
		  if(io<0) io=i;
		  if (cur_linemode) {
		      if (l_count < l_count_lim) {
			  mpv[ii+j] = cur_color;
			  l_count++;
		      } else if (l_count < l_count_space) {
			  l_count++;
		      } else l_count = 0;
		  } else
		      mpv[ii+j] = cur_color;
		  if (ABS(io-i) > 1) {
		    int is, iia = io < i ? io : i, iib = io > i ? io : i;
		    for(is=iia; is<=iib; is++) {
			if (cur_linemode) {
			    if (l_count < l_count_lim) {
				mpv[is*n_l+j] = cur_color;
				l_count++;
			    } else if (l_count < l_count_space) {
				l_count++;
			    } else l_count = 0;
			} else
			    mpv[is*n_l+j] = cur_color;
		    }
		  }
		  io=i;
	    }
	}
    }
    curr_lat = b;
    curr_lon = l;

    return;
}
void P_arc(bc,lc, bp,lp, alpha)
double bc,lc, bp,lp, alpha;
{
    register int i, j;
    int l_count = 0, l_count_lim =0, l_count_space = 0;
    int j1, j2, i1, i2;
    double dt = M_PI/n_b;
    double dp = 2*M_PI/n_l;
    double theta0= (90 - bc) * M_PI/180., phi0 = lc * M_PI/180.;
    double thetaP= (90 - bp) * M_PI/180., phiP = lp * M_PI/180.;
    double cth0 = cos(theta0), sth0 = sqrt(1-cth0*cth0), c = 1.5;
    double cthP = cos(thetaP), sthP = sqrt(1-cthP*cthP);
    double alp = alpha * M_PI/180.;
    double dpx = dt > dp ? dt : dp,
	   radius;
    int cur_color = 1;
    int io = -1;

    switch (cur_linemode ) {
    case L_DOTTED:
	l_count_lim=4;
	l_count_space=l_count_lim+4;
	break;
    case L_SHORTDASHED:
	l_count_lim=32;
	l_count_space=l_count_lim+16;
	break;
    case L_LONGDASHED:
	l_count_lim=64;
	l_count_space=l_count_lim+32;
	break;
    default :
	break;
    }

    radius = acos(cos(phi0-phiP)*sth0*sthP + cth0*cthP);

    if (phi0 < M_PI) {
	j_p0 = (M_PI-phi0) / dp;
    } else { /* if (phi > M_PI) { */
	j_p0 = (3*M_PI-phi0) / dp;
    }
    i_t0 = theta0 / dt;

    if (theta0 - c*radius < 0 )  {
	i1 = 0;
	i2 = (theta0 + c*radius) / dt;
	j1 = 0;
	j2 = n_l-1;
    } else if (theta0 + c*radius > 90 )  {
	i1 = (theta0 + c*radius) / dt;
	i2 = n_b-1;
	j1 = 0;
	j2 = n_l;
    } else {
	j1 = (M_PI-phi0+c*radius) / dp;
	j2 = (M_PI-phi0-c*radius) / dp;
	i1 = (theta0 - c*radius) / dt;
	i2 = (theta0 + c*radius) / dt;
    }

    for(i=i1; i<=i2; i++) {
	double b_i = M_PI/2. - i * dt, theta = (M_PI/2 - b_i);
	double cth = cos(theta), sth = sqrt(1-cth*cth);
	int ii = i * n_l;
	int ja = j1 < j2 ? j1 : j2, jb = j1 > j2 ? j1 : j2;
	for(j=ja; j<=jb; j++) {
	    double l_j = M_PI - j * dt, phi = l_j, v, ang;

	    /* distance 'v' between two pixels on a sphere */
	    dp = phi0 - phi;
	    v = acos(cos(dp)*sth0*sth + cth0*cth);

	    ang = acos(cos(phiP-phi)*sthP*sth + cthP*cth);
	    if ((alp > 0 && ang > alp) || (alp < 0 && ang < alp)) continue;

	    if ( ! (alp > 0 && phiP > phi) )
		    continue;
	    if (ABS(v-radius) <= dpx) {
//PR(ER, "alp=%g, ang=%g\n", alp*180/M_PI, ang*180/M_PI);
		if(io<0) io=i;
		if (cur_linemode) {
		    if (l_count < l_count_lim) {
			mpv[ii+j] = cur_color;
			l_count++;
		    } else if (l_count < l_count_space) {
			l_count++;
		    } else l_count = 0;
		} else
		    mpv[ii+j] = cur_color;
		if (ABS(io-i) > 1) {
		    int is, iia = io < i ? io : i, iib = io > i ? io : i;
		    for(is=iia; is<=iib; is++) {
			if (cur_linemode) {
			    if (l_count < l_count_lim) {
				mpv[is*n_l+j] = cur_color;
				l_count++;
			    } else if (l_count < l_count_space) {
				l_count++;
			    } else l_count = 0;
			} else
			    mpv[is*n_l+j] = cur_color;
		    }
		}
		io=i;
	    }
	}
    }
    curr_lat = bc;
    curr_lon = lc;

    return;
}
void Pg_arc(bc,lc, bp,lp, alpha)
double bc,lc, bp,lp, alpha;
{
    register int i, j;
    int l_count = 0, l_count_lim =0, l_count_space = 0;
    int j1, j2, i1, i2;
    double dt = M_PI/n_b;
    double dp = 2*M_PI/n_l;
    double theta0= (90 - bc) * M_PI/180., phi0 = lc * M_PI/180.;
    double thetaP= (90 - bp) * M_PI/180., phiP = lp * M_PI/180.;
    double cth0 = cos(theta0), sth0 = sqrt(1-cth0*cth0), c = 1.5;
    double cthP = cos(thetaP), sthP = sqrt(1-cthP*cthP);
    double alp = alpha * M_PI/180.;
    double dpx = dt > dp ? dt : dp,
	   radius;
    int cur_color = 1;
    int io = -1;

    switch (cur_linemode ) {
    case L_DOTTED:
	l_count_lim=4;
	l_count_space=l_count_lim+4;
	break;
    case L_SHORTDASHED:
	l_count_lim=32;
	l_count_space=l_count_lim+16;
	break;
    case L_LONGDASHED:
	l_count_lim=64;
	l_count_space=l_count_lim+32;
	break;
    default :
	break;
    }

    radius = acos(cos(phi0-phiP)*sth0*sthP + cth0*cthP);

    if (phi0 < M_PI) {
	jg_p0 = (M_PI-phi0) / dp;
    } else { /* if (phi > M_PI) { */
	jg_p0 = (3*M_PI-phi0) / dp;
    }
    ig_t0 = theta0 / dt;

    if (theta0 - c*radius < 0 )  {
	i1 = 0;
	i2 = (theta0 + c*radius) / dt;
	j1 = 0;
	j2 = n_l-1;
    } else if (theta0 + c*radius > 90 )  {
	i1 = (theta0 + c*radius) / dt;
	i2 = n_b-1;
	j1 = 0;
	j2 = n_l;
    } else {
	j1 = (M_PI-phi0+c*radius) / dp;
	j2 = (M_PI-phi0-c*radius) / dp;
	i1 = (theta0 - c*radius) / dt;
	i2 = (theta0 + c*radius) / dt;
    }

    for(i=i1; i<=i2; i++) {
	double b_i = M_PI/2. - i * dt, theta = (M_PI/2 - b_i);
	double cth = cos(theta), sth = sqrt(1-cth*cth);
	int ii = i * n_l;
	int ja = j1 < j2 ? j1 : j2, jb = j1 > j2 ? j1 : j2;
	for(j=ja; j<=jb; j++) {
	    double l_j = M_PI - j * dt, phi = l_j, v, ang;

	    /* distance 'v' between two pixels on a sphere */
	    dp = phi0 - phi;
	    v = acos(cos(dp)*sth0*sth + cth0*cth);

	    ang = acos(cos(phiP-phi)*sthP*sth + cthP*cth);
	    if ((alp > 0 && ang > alp) || (alp < 0 && ang < alp)) continue;

	    if ( ! (alp > 0 && phiP > phi) )
		    continue;
	    if (ABS(v-radius) <= dpx) {
//PR(ER, "alp=%g, ang=%g\n", alp*180/M_PI, ang*180/M_PI);
		if(io<0) io=i;
		if (cur_linemode) {
		    if (l_count < l_count_lim) {
			mpv[ii+j] = cur_color;
			l_count++;
		    } else if (l_count < l_count_space) {
			l_count++;
		    } else l_count = 0;
		} else
		    mpv[ii+j] = cur_color;
		if (ABS(io-i) > 1) {
		    int is, iia = io < i ? io : i, iib = io > i ? io : i;
		    for(is=iia; is<=iib; is++) {
			if (cur_linemode) {
			    if (l_count < l_count_lim) {
				mpv[is*n_l+j] = cur_color;
				l_count++;
			    } else if (l_count < l_count_space) {
				l_count++;
			    } else l_count = 0;
			} else
			    mpv[is*n_l+j] = cur_color;
		    }
		}
		io=i;
	    }
	}
    }
    curr_lat = bc;
    curr_lon = lc;

    return;
}
void P_linemod(str)
char *str;
{
    if (strcmp (str, "solid") == 0 || strcmp (str, "0") == 0)
	cur_linemode = L_SOLID;
    else if (strcmp (str, "dotted") == 0 || strcmp (str, "1") == 0)
	cur_linemode = L_DOTTED;
    else if (strcmp (str, "dotdashed") == 0 || strcmp (str, "2") == 0)
	cur_linemode = L_DOTDASHED;
    else if (strcmp (str, "shortdashed") == 0 || strcmp (str, "2") == 0)
	cur_linemode = L_SHORTDASHED;
    else if (strcmp (str, "longdashed") == 0 || strcmp (str, "2") == 0)
	cur_linemode = L_LONGDASHED;

    return;
}
void P_line(x1,y1, x2,y2)
int x1,y1, x2,y2;
{
    P_move(x1,y1);
    P_cont(x2,y2);
    return;
}
void Pg_line(b1,l1,b2,l2)
double b1,l1, b2,l2;
{
    Pg_move(b1,l1);
    Pg_cont(b2,l2);
    return;
}
void P_move(x,y)
int x,y;
{
    if (y < 0 || y > ny_gd) {
	PR(ER, "%s: bad parameter for 'P_move' command: y = %d\n", myname, y);
	exit(1);
    }
    if (x < 0 || x > nx_gd) {
	PR(ER, "%s: bad parameter for 'P_move' command: x = %d\n", myname, x);
	exit(1);
    }

    cur_x = x;
    cur_y = y;

    return;
}
void Pg_move(b,l)
double b,l;
{
    double dt = M_PI/n_b;
    double dp = 2*M_PI/n_l;
    double theta= (90 - b) * M_PI/180., phi = l * M_PI/180.;

    if (phi < 0) phi += 2.*M_PI;
    if (phi < M_PI) {
	jg_p0 = (M_PI-phi) / dp;
    } else { /* if (phi > M_PI) { */
	jg_p0 = (3*M_PI-phi) / dp;
    }
    ig_t0 = theta / dt;

    if (b < -90 || b > 90) {
	PR(ER, "%s: bad parameter for 'Pg_move' command: b = %g\n", myname,b);
	exit(1);
    }

    curr_lat = b;
    curr_lon = l;

    return;
}
void P_cont(x,y)
int x,y;
{
    register int i=0, j=0, i_old=0;
    int l_count = 0, l_count_lim =0, l_count_space = 0;
    int flag=0;
    double r = 0.0;
    int cur_color = 1;

    switch (cur_linemode ) {
    case L_DOTTED:
	l_count_lim=4;
	l_count_space=l_count_lim+4;
	break;
    case L_SHORTDASHED:
	l_count_lim=16;
	l_count_space=l_count_lim+8;
	break;
    case L_LONGDASHED:
	l_count_lim=32;
	l_count_space=l_count_lim+16;
	break;
    }

    if (y < 0 || y > ny_gd) {
	PR(ER, "%s: bad parameter for 'P_cont' command: y = %d\n", myname, y);
	exit(1);
    }

    if (x != cur_x)  {
	r = (double)(y - cur_y) / (x - cur_x);
    }
    flag = cur_x < x ? 1 : (cur_x > x ? -1 : 0);

    if (flag) {
	int beg = cur_x, end = x;
	if (flag < 0) { beg = x, end = cur_x; }
	i_old = cur_y + r * (beg - cur_x);
	for(j=beg; j<end; j++)  {
	    i = cur_y + r * (x - cur_x);
	    if (cur_linemode) {
		if (l_count < l_count_lim) {
		    mpv_cart[i*nx_gd+j].col = cur_color;
		    l_count++;
		} else if (l_count < l_count_space) {
		    l_count++;
		} else l_count = 0;
	    } else
		mpv_cart[i*nx_gd+j].col = cur_color;
	    if (ABS(i-i_old) > 1) {
		int k, ka = i > i_old? i_old:i, kb = i >i_old? i:i_old;
		for(k=ka; k<kb; k++)  {
		    if (cur_linemode) {
			if (l_count < l_count_lim) {
			    mpv_cart[k*nx_gd+j].col = cur_color;
			    l_count++;
			} else if (l_count < l_count_space) {
			    l_count++;
			} else l_count = 0;
		    } else
			mpv_cart[k*nx_gd+j].col = cur_color;
		}
	    }
	    i_old = i;
	}
    } else {
	int beg = cur_y, end = y;
	if (cur_y > y) { beg = y, end = cur_y; }
	/* same phi, different theta */
	for(i=beg; i<end; i++)  {
	   if (cur_linemode) {
	       if (l_count < l_count_lim) {
		   mpv_cart[i*nx_gd + x].col = cur_color;
		   l_count++;
	       } else if (l_count < l_count_space) {
		   l_count++;
	       } else l_count = 0;
	   } else
	       mpv_cart[i*nx_gd + x].col = cur_color;
	}
    }

    cur_y = y;
    cur_x = x;

    return;
}
void Pg_cont(b,l)
double b,l;
{
    register int i=0, j=0, i_old=0;
    int l_count = 0, l_count_lim =0, l_count_space = 0;
    double dt = M_PI / n_b;
    double dp = 2*M_PI / n_l;
    double theta= (90 - b) * M_PI/180., phi = l * M_PI/180.;
    int i_t=0, j_p=0, flag=0, wrap_flag = 0;
    double r = 0.0;
    int cur_color = 1;

    switch (cur_linemode ) {
    case L_DOTTED:
	l_count_lim=4;
	l_count_space=l_count_lim+4;
	break;
    case L_SHORTDASHED:
	l_count_lim=16;
	l_count_space=l_count_lim+8;
	break;
    case L_LONGDASHED:
	l_count_lim=32;
	l_count_space=l_count_lim+16;
	break;
    }

    if (b < -90 || b > 90) {
	PR(ER, "%s: bad parameter for 'PG_cont' command: b = %g\n",myname,b);
	exit(1);
    }
    if ( (-180. <= curr_lon && curr_lon <= 180.) &&  l > 180.)  wrap_flag = 1;

    if ( wrap_flag ) {
	if ( (-180. <= curr_lon && curr_lon <= 180.) &&  l > 180.) {
	    int j_p1, j_p2, i_t1;
	    i_t = theta/dt;
	    j_p = (M_PI-phi) / dp;
	    r = (double)(i_t-ig_t0)/(j_p-jg_p0);
	    j_p1 = 0;
	    i_t1 = ig_t0 + r * (j_p1-jg_p0);

	    i_old = ig_t0;
	    for(j=jg_p0; j>j_p1; j--)  {
		i = ig_t0 + r * (j-jg_p0);
		if (cur_linemode) {
		    if (l_count < l_count_lim) {
			mpv[i*n_l+j] = cur_color;
			l_count++;
		    } else if (l_count < l_count_space) {
			l_count++;
		    } else l_count = 0;
		} else
		    mpv[i*n_l+j] = cur_color;
		if (ABS(i-i_old) > 1) {
		    int k, ka = i >i_old? i_old:i, kb = i >i_old? i:i_old;
		    for(k=ka; k<kb; k++)  {
			if (cur_linemode) {
			    if (l_count < l_count_lim) {
				mpv[k*n_l+j] = cur_color;
				l_count++;
			    } else if (l_count < l_count_space) {
				l_count++;
			    } else l_count = 0;
			} else
			    mpv[k*n_l+j] = cur_color;
		    }
		}
		i_old = i;
	    }

	    /* wrap and continue */
	    j_p1 = n_l-1;
	    j_p2 = j_p + (n_l-1);
	    i_old = i_t1;
	    for(j=j_p1; j>j_p2; j--)  {
		i = i_t1 + r * (j-j_p1);
		if (cur_linemode) {
		    if (l_count < l_count_lim) {
			mpv[i*n_l+j] = cur_color;
			l_count++;
		    } else if (l_count < l_count_space) {
			l_count++;
		    } else l_count = 0;
		} else
		    mpv[i*n_l+j] = cur_color;

		if (ABS(i-i_old) > 1) {
		    int k, ka = i >i_old? i_old:i, kb = i >i_old? i:i_old;
		    for(k=ka; k<kb; k++)  {
			if (cur_linemode) {
			    if (l_count < l_count_lim) {
				mpv[k*n_l+j] = cur_color;
				l_count++;
			    } else if (l_count < l_count_space) {
				l_count++;
			    } else l_count = 0;
			} else
			    mpv[k*n_l+j] = cur_color;
		    }
		}
		i_old = i;
	    }
	    j_p = j_p2;
	}
    } else {
	if (phi < M_PI) {
	    j_p = (M_PI-phi) / dp;
	} else {  /* if (phi > M_PI) {  */
	    j_p = (3*M_PI-phi) / dp;
	}
	i_t = theta/dt;

	if (j_p != jg_p0)  {
	    r = (double)(i_t-ig_t0)/(j_p-jg_p0);
	}
	flag = jg_p0 < j_p ? 1 : (jg_p0 > j_p ? -1 : 0);

	if (flag) {
	    int beg = jg_p0, end = j_p;
	    if (flag < 0) { beg = j_p, end = jg_p0; }
	    i_old = ig_t0 + r * (beg-jg_p0);
	    for(j=beg; j<end; j++)  {
		i = ig_t0 + r * (j-jg_p0);
		if (cur_linemode) {
		    if (l_count < l_count_lim) {
			mpv[i*n_l+j] = cur_color;
			l_count++;
		    } else if (l_count < l_count_space) {
			l_count++;
		    } else l_count = 0;
		} else
		    mpv[i*n_l+j] = cur_color;
		if (ABS(i-i_old) > 1) {
		    int k, ka = i >i_old? i_old:i, kb = i >i_old? i:i_old;
		    for(k=ka; k<kb; k++)  {
			if (cur_linemode) {
			    if (l_count < l_count_lim) {
				mpv[k*n_l+j] = cur_color;
				l_count++;
			    } else if (l_count < l_count_space) {
				l_count++;
			    } else l_count = 0;
			} else
			    mpv[k*n_l+j] = cur_color;
		    }
		}
		i_old = i;
	    }
	} else {
	    int beg = ig_t0, end = i_t;
	    if (ig_t0 > i_t) { beg = i_t, end = ig_t0; }
	    /* same phi, different theta */
	    for(i=beg; i<end; i++)  {
	       if (cur_linemode) {
		   if (l_count < l_count_lim) {
		       mpv[i*n_l+j_p] = cur_color;
		       l_count++;
		   } else if (l_count < l_count_space) {
		       l_count++;
		   } else l_count = 0;
	       } else
		   mpv[i*n_l+j_p] = cur_color;
	    }
	}
    }
    ig_t0 = i_t;
    jg_p0 = j_p;

    curr_lat = b;
    curr_lon = l;

    return;
}
void P_point(x, y)
int x, y;
{
    return;
}
void P_text(size,italic,angle,font,str)
int size,italic,angle,font;
char *str;
{
    int page_length = n_l;
    int page_height = n_b;
    int y = i_t0, x=j_p0;
    float ang = angle;
    int set=0;
    extern void text_f_out(int,int,int, char*, float, int, int, int, int);

    text_f_out(x,y,size, str, ang, italic, set, page_length, page_height);
    return;
}
void Pg_text(size,italic,angle,font,str)
int size,italic,angle,font;
char *str;
{
    int page_length = n_l;
    int page_height = n_b;
    int y = i_t0, x=j_p0;
    float ang = angle;
    int set=0;
    extern void text_f_out(int,int,int, char*, float, int, int, int, int);
    text_f_out(x,y,size, str, ang, italic, set, page_length, page_height);
    return;
}
void P_text_label(size,italic,angle,font,str)
int size,italic,angle,font;
char *str;
{
    register int i;
    int page_length = nx_gd;
    int page_height = ny_gd;
    int y, x, v;
    float ang = angle;
    int set=0;
    double min;
    extern void text_f_out(int,int,int, char*, float, int, int, int, int);

    /* convert (b,l) to pixels (i,j) */
    x = nx_gd / 2;
    min = 1e30;
    for (y=i=0; i<ny_gd; i++) {
	double lat = 90. - mpv_cart[i*nx_gd + x].theta * 180./M_PI;
	if ( ABS(lat - curr_lat) < min ) {y = i; min = ABS(lat - curr_lat);}
    }
    min = 1e30;
    v = y * nx_gd;
    for (x=i=0; i<nx_gd; i++) {
	double lon = mpv_cart[v + i].phi * 180./M_PI;
	if ( ABS(lon - curr_lon) < min ) {x = i; min = ABS(lon - curr_lon);}
    }

    P_move(x, y);
    text_f_out(x,y,size, str, ang, italic, set, page_length, page_height);
    return;
}
void Pg_text_label(size,italic,angle,font,str)
int size,italic,angle,font;
char *str;
{
    return;
}

#ident  "#(@)symbol.c,V(C) O.Verkhodanov (SAO RAS), 10/05/1994"

static int afin_key = 0;
static float aa1=1.0, aa2=0.0, bb1=0.0, bb2=1.0, cc1=0.0, cc2=0.0;

#ifdef __GNUC__
extern void symtab_f(char *, int, int *, int *, int);
#else
extern void symtab_f();
#endif

/*
*=====================================================
*  Output to the plotter the symbol information
*  Variant: afin. transforms inside GRAFOR
*
*   Input parameres:
*     x     coordinate x of the text begining
*     y                y
*     size  - size of font
*     str   - string fro output
*     theta - rotation angle
*
*   this program for relative coordinates.
*   it uses move_f(int x, int y),  cont_f (int x, int y)
*
*   translated from GRAFOR-FORTRAN symbol()
*   to "C" by O.Verkhodanov, 1/7/1993
*
*=====================================================
*/

#define  TIP_ANGLE  0.4



#ifdef __GNUC__
void text_f_out (int x, int y, int size, char *str, float theta,
	 int italic, int set, int page_length, int page_height)
#else
void text_f_out(x,y,size, str, theta, italic, set, page_length, page_height)
int x,y;                              /* coordinates of begining */
int size;                             /* size of letter          */
char *str;                            /* string for output       */
float theta;                          /* angle, degs             */
int set;                              /* No of symbol set        */
int italic;                           /* type of symbol tipping  */
int page_length,                      /* length of page by 'X'   */
    page_height;                      /* length of page by 'Y'   */
#endif
{
      int j, kk=0;
      int nplot[16][2];          /* matrix of symbols */
      float xn=0., yn=0.;
      float tip_angle;           /* tip of symbols   */
      int len = strlen(str);     /* number of symbols in a string */
      /* cell's size, in which a symbol is drawn */
      float fact = (float)size / 9.0;
      float t  = (theta/180.) * 3.14159;   /* angle of tip in rad */
      float ct, st, fx, fy, xlp, ylp, xt, yt, xn1, yn1;

      if ( !len )   return;
      switch (italic) {
      case -1:
	       tip_angle = -TIP_ANGLE;
	       break;
      case  1:
	       tip_angle =  TIP_ANGLE;
	       break;
      case  0:
      default:
	       tip_angle = 0.0;
	       break;
      }

      ct = fact * cos(t);
      st = fact * cos(1.5708-t);
      fx = ct * 6. ;                    /* step by 'x'             */
      fy = st * 6. ;                    /* step by 'y'             */
      xlp= page_length;                 /* page length by 'x'      */
      ylp= page_height;                 /* page length by 'y'      */
      xt = (float)x;
      yt = (float)y;
      xn = xt;
      yn = yt;
      j =  0;                           /* n% of drawn symbol      */

      while( len-- )  {
	       if (afin_key)  {    /*  is affin     ? */

			/* execution affin   */
			xn1 = aa1 * xn + bb1 * yn + cc1;
			yn1 = aa2 * xn + bb2 * yn + cc2;

			/* Is a symbol on a page ? */
			if(xn1 < 0. || yn1 < 0. || xn1-xlp-0.01 > 0. ||
			       yn1-ylp-0.01 > 0.)  {
			       xn += fx; /* new current point in a symbol */
			       yn += fy;
			       /* goto the begining of symbol */
			       P_move((int)xn, (int)yn);
			       continue;
			}
	       }  else {
			/* Is a symbol on a page ? */
		       if(xn+0.01 < 0. || yn+0.01 < 0. || xn-xlp-0.01 > 0. ||
			    yn-ylp-0.01 > 0.)  {
			       xn += fx; /* new current point in a symbol */
			       yn += fy;
			       /* goto the begining of symbol */
			       P_move((int)xn, (int)yn);
			       continue;
		       }
	       }

	       symtab_f(str+j++, 0, nplot[0], &kk, set);

	       if(kk > 0) {
		  int key=0; /* move to new position */
		  register int k;

		  /* kk - amount of steps for symbols drawing */
		  for(k=0; k<kk; k++) {
			int tnx, tny;
			if ( nplot[k][0] - 7 >= 0 ) {
				key = 0;
				continue;
			}
			yt = nplot[k][1];
			xt = nplot[k][0] + yt * tip_angle;
			tnx = (int) (xn + xt * ct - yt * st);
			tny = (int) (yn + xt * st + yt * ct);
			if(key)
				P_cont(tnx, tny);
			else {
				P_move(tnx, tny);
			}
			key = 1;  /* continue to new position */
		  }
	      }

	      xn += fx; /* new current point in a symbol */
	      yn += fy;
	      /* goto the begining of symbol */
	      P_move((int)xn, (int)yn);
	}
	return;
}

#ifdef __GNUC__
void text_fg_out (int x, int y, int size, char *str, float theta,
	 int italic, int set, int page_length, int page_height)
#else
void text_fg_out(x,y,size, str, theta, italic, set, page_length, page_height)
int x,y;                              /* coordinates of begining */
int size;                             /* size of letter          */
char *str;                            /* string for output       */
float theta;                          /* angle, degs             */
int set;                              /* No of symbol set        */
int italic;                           /* type of symbol tipping  */
int page_length,                      /* length of page by 'X'   */
    page_height;                      /* length of page by 'Y'   */
#endif
{
      int j, kk=0;
      int nplot[16][2];          /* matrix of symbols */
      float xn=0., yn=0.;
      float tip_angle;           /* tip of symbols   */
      int len = strlen(str);     /* number of symbols in a string */
      /* cell's size, in which a symbol is drawn */
      float fact = (float)size / 9.0;
      float t  = (theta/180.) * 3.14159;   /* angle of tip in rad */
      float ct, st, fx, fy, xlp, ylp, xt, yt, xn1, yn1;

      if ( !len )   return;
      switch (italic) {
      case -1:
	       tip_angle = -TIP_ANGLE;
	       break;
      case  1:
	       tip_angle =  TIP_ANGLE;
	       break;
      case  0:
      default:
	       tip_angle = 0.0;
	       break;
      }

      ct = fact * cos(t);
      st = fact * cos(1.5708-t);
      fx = ct * 6. ;                    /* step by 'x'             */
      fy = st * 6. ;                    /* step by 'y'             */
      xlp= page_length;                 /* page length by 'x'      */
      ylp= page_height;                 /* page length by 'y'      */
      xt = (float)x;
      yt = (float)y;
      xn = xt;
      yn = yt;
      j =  0;                           /* n% of drawn symbol      */

      while( len-- )  {
	       if (afin_key)  {    /*  is affin     ? */

			/* execution affin   */
			xn1 = aa1 * xn + bb1 * yn + cc1;
			yn1 = aa2 * xn + bb2 * yn + cc2;

			/* Is a symbol on a page ? */
			if(xn1 < 0. || yn1 < 0. || xn1-xlp-0.01 > 0. ||
			       yn1-ylp-0.01 > 0.)  {
			       xn += fx; /* new current point in a symbol */
			       yn += fy;
			       /* goto the begining of symbol */
			       Pg_move((int)xn, (int)yn);
			       continue;
			}
	       }  else {
			/* Is a symbol on a page ? */
		       if(xn+0.01 < 0. || yn+0.01 < 0. || xn-xlp-0.01 > 0. ||
			    yn-ylp-0.01 > 0.)  {
			       xn += fx; /* new current point in a symbol */
			       yn += fy;
			       /* goto the begining of symbol */
			       Pg_move((int)xn, (int)yn);
			       continue;
		       }
	       }

	       symtab_f(str+j++, 0, nplot[0], &kk, set);

	       if(kk > 0) {
		  int key=0; /* move to new position */
		  register int k;

		  /* kk - amount of steps for symbols drawing */
		  for(k=0; k<kk; k++) {
			int tnx, tny;
			if ( nplot[k][0] - 7 >= 0 ) {
				key = 0;
				continue;
			}
			yt = nplot[k][1];
			xt = nplot[k][0] + yt * tip_angle;
			tnx = (int) (xn + xt * ct - yt * st);
			tny = (int) (yn + xt * st + yt * ct);
			if(key)
				Pg_cont(tnx, tny);
			else {
				Pg_move(tnx, tny);
			}
			key = 1;  /* continue to new position */
		  }
	      }

	      xn += fx; /* new current point in a symbol */
	      yn += fy;
	      /* goto the begining of symbol */
	      Pg_move((int)xn, (int)yn);
	}
	return;
}

/* Symbol table GRAFOR/DEMOS */

/*
 *    subroutine symtab(str,i,nplot,kk,is1)
 *    byte str                  ! str[i] - symbol for coding
 *                              ! by the font set is1
 *    dimension nplot(2,16)     ! output array of strokes of symbol
 *                              ! kk - number of strokes
 *
 * translated to "C" and added by V.Shergin, 1987.
 */

int  symcnf_f(c, nplot, nset)
int  c, nset;
int *nplot;
{      register int jj, mxy, j;
	int ir, lb, le,  k=c&0177;

	/* table of coding configuration of symbols */
	/* code[i] = x|(y<<3); if(x==7) PENUP */
	static char code[1763] = {
  7, 26,18,19,27,26,7,34,74,75,35,34,                           /* !     */
  73,66,58,57,65,66,7,75,68,60,59,67,68,                        /* "     */
  32,36,7,52,48,7,66,17,7,18,67,                                /* #     */
  24,17,19,28,44,51,49,56,64,73,75,68,23,74,18,                 /* $     */
  65,64,72,73,65,7,76,16,7,27,28,20,19,27,                      /* %     */
  20,27,65,74,67,40,32,25,26,36, 73,66,58,57,65,66, 73,64,24,17,/* & ' ( */
  19,28,68,75, 24,60,7,56,28,7,44,40, 40,44,42,58,26,           /* ) * + */
  17,26,34,33,25,26, 40,44, 17,25,26,18,17, 16,76,              /* , - . / */
  17,32,56,73,75,60,36,19,17, 17,19,18,74,65,                   /* 0 1   */
  64,73,75,68,52,24,16,20, 64,73,75,68,60,51,49,51,44,28,19,17,24,/* 2 3 */
  72,48,41,44,76,20, 76,72,48,51,44,28,19,17,24,                /* 4 5   */
  68,75,73,64,24,17,19,28,44,51,49,40, 64,72,76,68,25,17,       /* 6 7   */
  49,56,64,73,75,68,60,51,49,40,24,17,19,28,44,51,              /* 8     */
  24,17,19,28,68,75,73,64,48,41,43,52,                          /* 9     */
  25,33,34,26,25,7,49,57,58,50,49,                              /* : ; < = */
  49,57,58,50,49,7,26,25,33,34,26,17, 12,40,76, 48,52,23,36,32,
  8,44,72, 17,26,19,17,7,34,50,51,60,68,75,73,64,               /* > ? */
  36,27,25,32,56,65,67,60,44,35,34,50,59,                       /* @   */

  16,64,73,75,68,20,44,40,                                  /* A    */
  16,48,51,48,72,75,68,60,51,44,28,19,16,                   /* B    */
  68,75,73,64,24,17,19,28, 16,72,75,68,28,19,16,            /* C D */
  20,16,40,43,40,72,76, 16,48,51,48,72,75,                  /* E F */
  68,75,73,64,24,17,19,28,44,43, 16,72,40,44,76,20,         /* G H */
  17,19,18,74,73,75, 76,28,19,17,24,32,                     /* I J */
  16,72,40,76,49,20, 20,16,72, 16,72,50,76,20,              /* K L M */
  16,72,20,76, 73,64,24,17,19,28,68,75,73,                  /* N O */
  16,48,51,48,72,75,68,60,51,                               /* P    */
  73,64,24,17,19,28,68,75,73,23,27,20,                      /* Q    */
  16,48,51,48,72,75,68,60,51,50,20,                         /* R    */
  24,17,19,28,44,51,49,56,64,73,75,68,                      /* S    */
  72,76,74,18, 76,28,19,17,24,32,72, 72,18,76,              /* T U V */
  72,16,42,20,76, 16,76,23,72,20, 72,50,18,50,76,           /* W X Y */
  72,76,16,20,23,49,51, 18,16,72,74, 72,20,                 /* Z [ \  */
  18,20,76,74, 56,64,73,75,68,60, 16,20,                    /* ] ^ - */

    /* Russian big letters */
  16,72,48,49,65,74,75,68,28,19,18,25,49,                   /*      */
  16,64,73,75,68,20,44,40, 48,51,44,28,19,16,72,76,         /*     */
  72,16,20,76,20,21,12, 8,16,21,13,21,20,76,73,74,17,       /*     */
  20,16,40,43,40,72,76,                                     /*      */
  17,19,18,66,65,56,32,25,27,36,60,67,66,74,73,75,          /*      */
  16,72,76,68, 16,76,23,72,20, 72,16,76,20,                 /*    */
  72,16,76,20,7,73,75, 16,72,40,76,49,20, 16,74,20,         /*    */
  16,72,50,76,20, 16,72,40,44,76,20,                        /*     */
  73,64,24,17,19,28,68,75,73, 16,72,76,68,20,               /*     */
  20,76,73,64,48,41,44,41,16, 16,48,51,48,72,75,68,60,51,   /*     */
  68,75,73,64,24,17,19,28, 72,76,74,18,                     /*     */
  72,40,33,35,44,76,28,19,17,24,                            /*      */
  16,49,72,49,50,74,18,50,51,76,51,20,                      /*      */
  16,48,51,48,72,75,68,60,51,44,28,19,16,                   /*      */
  48,51,44,28,19,16,72, 72,24,17,18,27,43,50,49,40,7,76,20, /*     */
  64,73,75,68,60,51,49,51,44,28,19,17,24,                   /*      */
  72,16,18,58,18,20,76,20, 24,17,19,28,52,49,52,68,75,73,64,/*     */
  72,16,18,58,18,20,76,20,21,12, 72,48,41,44,76,20,         /*     */

			      /*  english little letters   */
  40,49,51,44,28,19,17,24,33,36,20,                         /* a     */
  19,28,36,43,41,32,58,65,64,24,17,19,                      /* b     */
  28,19,17,24,40,49,51,44, 44,51,49,40,24,17,19,28,20,68,   /* c d >e*/
  28,19,17,24,40,49,51,44,36,32, 17,49,48,51,49,57,66,67,
  8,1,3,12,44,51,49,40,24,17,19,28,52, 72,16,40,49,51,44,20,/* <f g h*/
  18,50,7,58,66, 9,18,50,7,58,66, 16,64,32,51,32,19,        /* i j k */
  19,18,25,65, 16,48,26,52,20, 16,48,40,49,51,44,20,        /*l m n  */
  40,49,51,44,28,19,17,24,40,                               /*o      */
  40,49,51,44,28,19,17,24,40,48,00,                         /*p      */
  44,51,49,40,24,17,19,28,52,4,13, 16,48,40,49,51,44,       /*q r    */
  24,17,19,20,35,33,40,49,51,44, 19,18,25,65,49,48,50,      /*s t    */
  48,24,17,19,28,20,52, 48,18,52, 48,17,34,19,52,           /*u v w  */
  16,52,34,48,20, 48,18,52,18,2, 48,52,16,20,               /*x y z  */

				       /*  russian little letters */
  16,48,32,33,25,18,19,28,44,51,50,41,33,                       /*    */
  40,49,51,44,28,19,17,24,33,36,20,                             /*     */
  51,44,28,19,17,24,40,49,51,56,65,68, 48,24,17,18,27,51,19,20,11,  /* */
  44,51,49,40,24,17,19,28,60,67,65,56,                          /*     */
  28,19,17,24,40,49,51,44,36,32,                                /*     */
  50,49,40,24,17,19,28,44,51,50,58,2,                           /*     */
  40,49,51,44,35,33,24,17,19,28, 16,52,34,48,20,                /*    */
  48,24,17,19,28,20,52, 48,24,17,19,28,20,52,7,51,59,           /*    */
  16,64,32,51,32,19, 16,41,50,51,19, 16,48,26,52,20,            /*   */
  16,48,32,35,51,19, 40,49,51,44,28,19,17,24,40,                /*    */
  16,48,40,49,51,44,20, 16,25,35,51,49,40,32,25,35,19,          /*    */
  40,49,51,44,28,19,17,24,40,48,00, 28,19,17,24,40,49,51,44,    /*    */
  18,50,48,52, 48,24,17,19,28,52,12,3,1,8,                      /*    */
  48,33,16,33,34,50,18,34,35,52,35,20,                          /*     */
  19,28,36,43,41,32,58,65,64,24,17,19, 40,42,35,27,18,16,48,    /*    */
  48,16,18,27,35,42,40,7,52,20, 40,49,50,33,34,27,11,2,1,8,26,  /*    */
  48,24,17,26,50,26,19,28,52,20, 24,17,19,28,36,34,44,51,49,40, /*    */
  48,24,17,26,50,26,19,28,52,20,21,12, 48,40,33,35,51,19,       /*    */

					/* special symbols  */
  7, 24,33,35,28,35,51,60,51,49,56,49,33,35,51,
  16,9,10,19,74,83,84,77, 40,41,10,83,85,
  24,33,1,7,2,34,36,4,2,
  24,28,7,44,40,7,56,60, 32,36,7,52,48,7,25,59,
  24,56,65,67,60,28,
  64,32,25,27,36,68, 28,25,32,48,57,60,
  24,27,36,52,59,56,
  48,52,50,66,34,7,24,28, 40,49,36,45,
  32,36,26,46,58,52,48,
  38,34,28,40,60,50,54, 58,18,20,16,
  40,50,34,40,45, 40,45,51,35,45,
  18,74,57,59,4, 18,33,35,18,74,
  18,74, 57,17,17,19,59, 72,76,

			    /* greek big letters */
  16,74,43,41,43,20, 16,48,51,48,72,75,68,60,51,44,28,19,16,/*A B */
  16,72,76,68, 16,74,20,16, 20,16,40,43,40,72,76,           /*G D E  */
  17,19,18,66,65,56,32,25,27,36,60,67,66,74,73,75,          /*F        */
  16,72,76,68, 16,72,40,44,76,20, 17,19,18,74,73,75,        /*G H I  */
  56,32,17,19,36,60,75,73,56,7,57,41,49,51,59,43,           /*TETA  */
  16,72,40,76,49,20, 16,74,20, 16,72,50,76,20,              /*K L MU*/
  16,72,20,76, 73,64,24,17,19,28,68,75,73,                  /*N O */
  17,73,72,76,75,19,                                        /*P    */
  24,16,20,28,7,43,59,51,49,41,57,7,64,72,76,68,            /*Q    */
  16,48,51,48,72,75,68,60,51, 28,20,16,51,72,76,68,         /*R S */
  72,76,74,18, 72,50,18,50,76,                              /*T Y */
  64,40,33,34,18,17,19,18,74,73,75,74,34,35,44,68,          /*FI  */
  16,17,32,56,73,75,60,36,19,20,                            /*OMEGA */
  16,76,23,72,20, 72,50,18,50,76, 72,76,16,20,23,49,51,     /*X Y Z */

				       /* little greek letters  */
  20,43,50,49,40,24,17,18,19,52,
  0,56,65,67,60,52,43,41,43,36,28,19,17,24,                 /*beta */
  48,41,18,2,18,43,52, 51,44,28,19,17,24,40,49,51,56,65,68, /*g d  */
  44,51,49,40,33,35,33,24,17,19,28,                         /*e    */
  49,40,24,17,19,28,44,51,50,2, 48,41,18,2,18,43,52,        /*f g  */
  16,48,40,49,51,44,4, 50,18,19,                            /*nu l */
  40,32,17,19,36,44,40,48,65,67,52,44,                      /*teta */
  48,25,16,52,43,20, 16,50,64,50,20,
  0,25,49,25,18,19,28,20,52, 48,42,19,52,                   /*mu u */
  40,49,51,44,28,19,17,24,40, 17,49,48,52,51,19,            /*o pi */
  65,59,60,59,48,40,35,36,35,24,16,9,12,4,
  0,40,49,51,44,28,19,17,24, 52,49,40,24,17,19,28,44,50,    /*r s  */
  40,49,52,50,18, 48,24,17,19,28,44,51,                     /*t u  */
  48,24,17,18,2,58,18,19,28,52,                             /*psi  */
  49,40,24,17,26,42,26,19,28,44,51,                         /*omega*/
  48,49,26,0,52,26,3,4, 48,24,17,19,28,44,51,               /*x y  */
  64,57,59,68,32,24,17,20,12,                               /*z    */
     };

     /* table of pointers onto coding table */
     static int m[227] = {
  1,12,25,36,51,65,75,81,85,89,97,102,108,110,115,117, /* numbers */
  126,131,139,152,158,167,179,185,201,213,224,236,239,244,
  247,260,273,             /* english upper case letters  */
  281,294,302,309,316,322,332,338,344,350,356,359,364,368,
  377,386,398,409,421,425,432,435,440,445,450,457, /* symbols */
  461,463,467,473,475,     /* russian upper case letters  */
  488,496,504,511,521,528,544,548,553,557,564,570,573,
  578,584,593,598,607,616,624,628,638,650,663,670,682,
  695,703,714,724,730,     /* english lower case letters  */
  741,753,761,771,781,789,802,809,814,820,826,
  830,835,842,851,862,873,879,889,896,903,906,
  911,916,921,925,         /* russian lower case letters  */
  938,949,961,970,982,992,1004,1014,1019,1026,1036,
  1042,1047,1052,1058,1067,1074,1084,1095,1103,1107,
  1117,1129,1141,1148,1158,1169,1179,1189,
  1201,1207,               /* spec.letters   */
  1208,1222,1230,1235,1244,1252,1260,1266,1272,1278,
  1284,1292,1296,1303,1310,1314,1319,1324,1329,1334,
  1336,1341,1343,          /* greek upper and lower letters  */
  1349,1362,1366,1370,1377,1393,1397,1403,
  1409,1425,1431,1434,1439,1443,1452,1458,
  1474,1483,1490,1494,1499,1515,1525,1530,
  1535,1542,1552,1566,1573,1585,1596,1606,1613,1620,
  1623,1635,1641,1646,1655,1659,1668,1674,1688,
  1697,1706,1711,1718,1728,1739,1747,1754,1763
	};

	/* translation from koi-8 to combining set + nset */
	if( c&0200 ) {
	    if( k>=0100 && k<=0136 ) {
		k |= 040; nset |= 1;
	    }
	} else {
	    if( k>=0141 && k<=0172 ) {
		k &= 0137; nset |= 1;
	    }
	}
	/* access to table of pointers */
	switch (nset) {
	case 0:
	    if     ( k>=32 && k<=126) ir = k-32;
	    else ir=0;
	    break;
	case 1:
	    if     ( k>=32 && k<=64 ) ir = k-32;
	    else if( k>=65 && k<=90 ) ir = k+30;
	    else if( k>=91 && k<=95 ) ir = k-32;
	    else if( k>=96 && k<=126) ir = k+25;
	    else ir=0;
	    break;
	case 2:
	case 3:
	    if     ( k>=32 && k<=47 ) ir=k+120;
	    else if( k>=48 && k<=57 ) ir = k-32;
	    else if( k>=58 && k<=64 ) ir = k+110;
	    else if( k>=65 && k<=90 ) ir = (nset==2)? k+110 : k+136;
	    else ir=0;
	    break;
	default:
	    ir=0;
	    break;
	}
	if(ir == 0) lb = le = 0;
	else  lb=m[ir-1], le=m[ir]-1;

	/* selection from symbol coding table */
	for( j=lb,jj=0; j<=le; j++) {
	    mxy = code[j];
	    nplot[jj++] = mxy&07;
	    nplot[jj++] = mxy>>3;
	}

	return( ((ir==0)&&(k!=32))? 0 : (jj>>1) );
}

void symtab_f(str,i,nplot,kk, set)
char *str;
int  i, *nplot, *kk, set;
{
      *kk =  symcnf_f( (int)(*(str+i)), nplot, set);
}

/*
 *  draw marker:
 */
#define XPM_POINT                   0  /*  0   -point;                      */
#define XPM_X                       1  /*  1,x -X;                          */
#define XPM_CROSS                   2  /*  2,+ -plus '+';                   */
#define XPM_CIRCLE                  3  /*  3,o -empty  circle;              */
#define XPM_F_CIRCLE                4  /*  4,O -filled circle;              */
#define XPM_RHOMB                   5  /*  5   -empty  rhomb;               */
#define XPM_F_RHOMB                 6  /*  6   -filled rhomb;               */
#define XPM_SQUARE                  7  /*  7,s -empty  square;              */
#define XPM_F_SQUARE                8  /*  8,S -filled square;              */
#define XPM_TRIANG_DOWN             9  /*  9,v -empty  triangular down;     */
#define XPM_F_TRIANG_DOWN          10  /* 10,V -filled triangular down;     */
#define XPM_TRIANG_UP              11  /* 11,^ -empty  triangular up;       */
#define XPM_F_TRIANG_UP            12  /* 12,  -filled triangular up;       */
#define XPM_TRIANG_LEFT            13  /* 13,> -empty  triangular left;     */
#define XPM_F_TRIANG_LEFT          14  /* 14,  -filled triangular left;     */
#define XPM_TRIANG_RIGHT           15  /* 15,< -empty  triangular right;    */
#define XPM_F_TRIANG_RIGHT         16  /* 16,  -filled triangular right;    */
#define XPM_ARROW_DOWN             17  /* 17   -arrow down;                 */
#define XPM_ARROW_UP               18  /* 18   -arrow up;                   */
#define XPM_ARROW_LEFT             19  /* 19   -arrow left;                 */
#define XPM_ARROW_RIGHT            20  /* 20   -arrow right;                */

#define XPM_PENT_STAR              21  /* 21   -pentagon star               */
#define XPM_F_PENT_STAR            22  /* 22   -filled pentagon star        */
#define XPM_ASTERISK               23  /* 23,* -asterisk                    */
#define XPM_SOLOMON_STAR           24  /* 24   -Solomon's star              */
#define XPM_F_SOLOMON_STAR         25  /* 25   -filled Solomon's star       */

#define XPM_FAT_RIGHT_DOWN_ARROW   35  /* 35   -fat right-down arrow        */
#define XPM_F_FAT_RIGHT_DOWN_ARROW 36  /* 36   -filled fat right-down arrow */
#define XPM_FAT_LEFT_DOWN_ARROW    37  /* 37   -fat left-down arrow         */
#define XPM_F_FAT_LEFT_DOWN_ARROW  38  /* 38   -filled fat left-down arrow  */
#define XPM_FAT_DOWN_ARROW         39  /* 39   -fat down arrow              */
#define XPM_F_FAT_DOWN_ARROW       40  /* 40   -filled fat down arrow       */

void P_pointm(x, y, r, mr)
register int x,y;  /* coordinates of central point of marker */
register int r;    /* size of marker by X                    */
int mr;            /* number of marker                       */
{
	register int rr = r; /* * aspect_ratio_XP; */
	int xy[20];

	r  /= 2;
	rr /= 2;

	switch (mr) {
	case XPM_POINT:
		P_point(x,y);
		if (r > 1) {
		    P_fcircle(x, y, r);
		}
		break;
	case 'x': /* cross 'x' */
	case XPM_X:
		P_move(x-r, y+rr);
		P_cont(x+r, y-rr);
		P_move(x-r, y-rr);
		P_cont(x+r, y+rr);
		break;
	case XPM_CROSS:  /* plus '+' */
		P_move(x, y-rr);
		P_cont(x, y+rr);
		P_move(x-r, y);
		P_cont(x+r, y);
		break;
	case XPM_CIRCLE:  /* empty  circle */
		P_circle(x, y, r);
		break;
	case XPM_F_CIRCLE:  /* filled circle */
		P_fcircle(x, y, r);
		break;
	case XPM_RHOMB:    /* empty  rhomb */
	case XPM_F_RHOMB:  /* filled rhomb */
		xy[0] = x;   xy[1] = y-rr;
		xy[2] = x-r; xy[3] = y;
		xy[4] = x;   xy[5] = y+rr;
		xy[6] = x+r; xy[7] = y;
		xy[8] = x;   xy[9] = y-rr;
		if (mr == XPM_F_RHOMB)
			P_f_pol (xy, 5);
		else
			P_p_line (xy, 5);
		break;
	case XPM_SQUARE:    /* empty  square */
	case XPM_F_SQUARE:  /* filled square */
	      { int r1 = r, rr1 = rr;
		xy[0] = x-r1; xy[1] = y-rr1;
		xy[2] = x-r1; xy[3] = y+rr1;
		xy[4] = x+r1; xy[5] = y+rr1;
		xy[6] = x+r1; xy[7] = y-rr1;
		xy[8] = x-r1; xy[9] = y-rr1;
		if (mr == XPM_F_SQUARE)
			P_f_pol (xy, 5);
		else
			P_p_line (xy, 5);
	      }
		break;
	case XPM_TRIANG_DOWN:   /* empty  triangle down */
	case XPM_F_TRIANG_DOWN: /* filled triangle down */
		xy[0] = x-r; xy[1] = y+rr;
		xy[2] = x+r; xy[3] = y+rr;
		xy[4] = x;   xy[5] = y-rr;
		xy[6] = x-r; xy[7] = y+rr;
		if (mr == XPM_F_TRIANG_DOWN)
			P_f_pol (xy, 4);
		else
			P_p_line (xy, 4);
		break;
	case XPM_TRIANG_UP:   /* empty  triangle up */
	case XPM_F_TRIANG_UP: /* filled triangle up */
		xy[0] = x-r; xy[1] = y-rr;
		xy[2] = x+r; xy[3] = y-rr;
		xy[4] = x;   xy[5] = y+rr;
		xy[6] = x-r; xy[7] = y-rr;
		if (mr == XPM_F_TRIANG_UP)
			P_f_pol (xy, 4);
		else
			P_p_line (xy, 4);
		break;
	case XPM_TRIANG_RIGHT:  /* empty  triangle right */
	case XPM_F_TRIANG_RIGHT:/* filled triangle right */
		xy[0] = x-r; xy[1] = y;
		xy[2] = x+r; xy[3] = y+rr;
		xy[4] = x+r; xy[5] = y-rr;
		xy[6] = x-r; xy[7] = y;
		if (mr == XPM_F_TRIANG_RIGHT)
			P_f_pol (xy, 4);
		else
			P_p_line (xy, 4);
		break;
	case XPM_TRIANG_LEFT:    /* empty  triangle left  */
	case XPM_F_TRIANG_LEFT:  /* filled triangle left  */
		xy[0] = x+r; xy[1] = y;
		xy[2] = x-r; xy[3] = y+rr;
		xy[4] = x-r; xy[5] = y-rr;
		xy[6] = x+r; xy[7] = y;
		if (mr == XPM_F_TRIANG_LEFT)
			P_f_pol(xy, 4);
		else
			P_p_line(xy, 4);
		break;
	case XPM_ARROW_DOWN: /* arrow down */
		xy[0] = r/2;
		xy[1] = y - 2 * rr;
		xy[2] = y - 3 * rr;
		P_move(x-r, y);    P_cont(x+r, y);
		P_move(x,   y);    P_cont(x, xy[2]);
		P_move(x-xy[0], xy[1]); P_cont(x, xy[2]);
		P_move(x+xy[0], xy[1]); P_cont(x, xy[2]);
		break;
	case XPM_ARROW_UP: /* arrow   up */
		xy[0] = r/2;
		xy[1] = y + 2 * rr;
		xy[2] = y + 3 * rr;
		P_move(x-r, y);    P_cont(x+r, y);
		P_move(x,   y);    P_cont(x, xy[2]);
		P_move(x-xy[0], xy[1]); P_cont(x, xy[2]);
		P_move(x+xy[0], xy[1]); P_cont(x, xy[2]);
		break;
	case XPM_ARROW_LEFT: /* arrow left */
		xy[0] = rr/2;
		xy[1] = x - 2 * r;
		xy[2] = x - 3 * r;
		P_move(x, y-rr);   P_cont(x, y+rr);
		P_move(x,   y);    P_cont(xy[2], y);
		P_move(xy[1], y-xy[0]); P_cont(xy[2], y);
		P_move(xy[1], y+xy[0]); P_cont(xy[2], y);
		break;
	case XPM_ARROW_RIGHT: /* arrow right */
		xy[0] = rr/2;
		xy[1] = x + 2 * r;
		xy[2] = x + 3 * r;
		P_move(x, y-rr);   P_cont(x, y+rr);
		P_move(x,   y);    P_cont(xy[2], y);
		P_move(xy[1], y-xy[0]); P_cont(xy[2], y);
		P_move(xy[1], y+xy[0]); P_cont(xy[2], y);
		break;
	case XPM_PENT_STAR:   /* pentagon star        */
	case XPM_F_PENT_STAR: /* filled pentagon star */
	       {              /* sin */           /* cos */
		int r72 = r *   .951,   rr72 = rr *   .309 / 2,
		   r144 = r *   .588,  rr144 = rr * (-.809),
		   r216 = r * (-.588), rr216 = rr * (-.809),
		   r288 = r * (-.951), rr288 = rr *   .309 / 2;

		xy[0] = x;        xy[1] = y+r;
		xy[2] = x + r144; xy[3] = y + rr144;
		xy[4] = x + r288; xy[5] = y + rr288;
		xy[6] = x + r72;  xy[7] = y + rr72;
		xy[8] = x + r216; xy[9] = y + rr216;
		xy[10] = x;       xy[11] = y+r;
	       }
		if (mr == XPM_F_PENT_STAR)
			P_f_pol (xy, 6);
		else
			P_p_line (xy, 6);
		break;
	case XPM_ASTERISK: /* asterisk */
	      { int r2 = r*0.707, rr2= rr*0.707;
		P_move(x-r2, y+rr2);
		P_cont(x+r2, y-rr2);
		P_move(x-r2, y-rr2);
		P_cont(x+r2, y+rr2);
		P_move(x, y-rr);
		P_cont(x, y+rr);
		P_move(x-r, y);
		P_cont(x+r, y);
	      }
		break;
	case XPM_SOLOMON_STAR:   /* Solomon's star        */
	case XPM_F_SOLOMON_STAR: /* filled Solomon's star */
	    {
		int rr2 = rr / 2.5;

		xy[0] = x-r; xy[1] = y+rr-rr2;
		xy[2] = x+r; xy[3] = y+rr-rr2;
		xy[4] = x;   xy[5] = y-rr-rr2;
		xy[6] = x-r; xy[7] = y+rr-rr2;
		if (mr == XPM_F_SOLOMON_STAR)
			P_f_pol (xy, 4);
		else
			P_p_line (xy, 4);
		xy[0] = x-r; xy[1] = y-rr+rr2;
		xy[2] = x+r; xy[3] = y-rr+rr2;
		xy[4] = x;   xy[5] = y+rr+rr2;
		xy[6] = x-r; xy[7] = y-rr+rr2;
		if (mr == XPM_F_SOLOMON_STAR)
			P_f_pol (xy, 4);
		else
			P_p_line (xy, 4);
	     }
	     break;
	case XPM_FAT_RIGHT_DOWN_ARROW:   /* fat left-down arrow */
	case XPM_F_FAT_RIGHT_DOWN_ARROW: /* filled fat left-down arrow */
	    {
		int r3 = r/3, rr3 = rr/3;
		xy[0]  = x;         xy[1]  = y;        /*            */
		xy[2]  = x-r;       xy[3]  = y;        /*            */
		xy[4]  = x-r+r3;    xy[5]  = y+rr3;    /*    / \     */
		xy[6]  = x-r-r3;    xy[7]  = y+rr;     /*    \   \/| */
		xy[8]  = x-r;       xy[9]  = y+rr+rr3; /*      \   | */
		xy[10] = x-r3;      xy[11] = y+rr-rr3; /*     /____| */
		xy[12] = x;         xy[13] = y+rr;     /*            */
		xy[14] = x;         xy[15] = y;        /*            */

		if (mr == XPM_F_FAT_RIGHT_DOWN_ARROW)
			P_f_pol  (xy, 8);
		else
			P_p_line (xy, 8);
	     }
	     break;
	case XPM_FAT_LEFT_DOWN_ARROW:   /* fat left-down arrow */
	case XPM_F_FAT_LEFT_DOWN_ARROW: /* filled fat left-down arrow */
	    {
		int r3 = r/3, rr3 = rr/3;
		xy[0]  = x;         xy[1]  = y;        /*            */
		xy[2]  = x+r;       xy[3]  = y;        /*            */
		xy[4]  = x+r-r3;    xy[5]  = y+rr3;    /*      / \   */
		xy[6]  = x+r+r3;    xy[7]  = y+rr;     /*  |\/   /   */
		xy[8]  = x+r;       xy[9]  = y+rr+rr3; /*  |   /     */
		xy[10] = x+r3;      xy[11] = y+rr-rr3; /*  |____\    */
		xy[12] = x;         xy[13] = y+rr;     /*            */
		xy[14] = x;         xy[15] = y;        /*            */

		if (mr == XPM_F_FAT_LEFT_DOWN_ARROW)
			P_f_pol (xy, 8);
		else
			P_p_line (xy, 8);
	     }
	     break;
	case XPM_FAT_DOWN_ARROW:   /* fat down arrow */
	case XPM_F_FAT_DOWN_ARROW: /* filled fat down arrow */
	    {
		int r2 = r/1.5;
		xy[0]  = x;         xy[1]  = y;        /*            */
		xy[2]  = x-r;       xy[3]  = y+rr;     /*     __     */
		xy[4]  = x-r+r2;    xy[5]  = y+rr;     /*    |  |    */
		xy[6]  = x-r+r2;    xy[7]  = y+rr+rr;  /*   _|  |_   */
		xy[8]  = x+r-r2;    xy[9]  = y+rr+rr;  /*   \    /   */
		xy[10] = x+r-r2;    xy[11] = y+rr;     /*     \/     */
		xy[12] = x+r;       xy[13] = y+rr;     /*            */
		xy[14] = x;         xy[15] = y;        /*            */

		if (mr == XPM_F_FAT_DOWN_ARROW)
			P_f_pol (xy, 8);
		else
			P_p_line (xy, 8);
	     }
	     break;
	default:
		P_point(x, y);
		break;
	}
	P_move(x,y);
	return;
}


/* put plot-command on a sphere */
void patt2mol(ov_flag, lon0)
int ov_flag;
double lon0;
{
    extern int get_nx_gd(), get_ny_gd();
    extern void plot_spur(), plot_milky_way(), plot_src();
    extern void move_gd(int, int);

    int y_c, x_c, x_r, y_r;
    register int i, j;
    double pi= M_PI;
    extern void point_gd (int, int);

    nx_gd = get_nx_gd();
    ny_gd = get_ny_gd();

    mpv = (int *)calloc((unsigned)(n_l*n_b), sizeof *mpv);
    if (mpv == (int *)0) {
nomem:
	PR(ER, "%s: no memory.\n", myname);
	exit(1);
    }
    i = nx_gd*ny_gd;
    mpv_glob = (int *)calloc((unsigned)i, sizeof *mpv_glob);
    if (mpv_glob == (int *)0)  goto nomem;
    mpv_cart=(struct _plain_cart_ *)calloc((unsigned)i,sizeof *mpv_cart);
    if (mpv_cart == (struct _plain_cart_ *)0)  goto nomem;

    y_c = ny_gd / 2;
    x_c = nx_gd / 2;
    y_r = ny_gd / 2;
    x_r = nx_gd / 2;

    /* put copy of base map to memory, select Molweide projection */
    for(i=0; i<ny_gd; i++) {
	double u, v = (i - y_c) / (y_r/1.02);
	int ii = nx_gd * i;

	for(j=0; j<nx_gd; j++) {
	    double phi = 1e20, theta=1e20;
	    u = 2. * (j - x_c) / (x_r/1.02);
	    if ( u*u/4. + v*v <= 1.) {
		double sqv = sqrt(1.-v*v);
		/* theta in [0, pi] */
		theta = pi/2. - asin(2./pi * (asin(v) + v*sqv));
		/* lon in [-pi,pi], the minus sign for astro convention */
		phi = -lon0 - pi/2 * u / MAX(sqv, 1.e-6);
		/* phi in [0, 2pi] */
		if ( phi < 0.0 )    phi += 2. * pi;
	    }
	    mpv_cart[ii+j].theta = theta;
	    mpv_cart[ii+j].phi = phi;
	}
    }

    /* read file with vector graphics and copy it to mpv-array */
    if (ov_flag == 1 || ov_flag == 12 || ov_flag == 13 || ov_flag == 123) {
	 plot_spur();
    }
    if (ov_flag == 2 || ov_flag == 12 || ov_flag == 23 || ov_flag == 123) {
	 plot_milky_way();
    }
    if (ov_flag == 3 || ov_flag == 13 || ov_flag == 23 || ov_flag == 123) {
	 plot_src();
    }


    /* put copy of base map to memory */
    y_c = n_b * n_l;
    x_c = nx_gd * ny_gd;
    for(i=0; i<ny_gd; i++) {
	double phi, theta;
	int ii = nx_gd * i;
	for(j=0; j<nx_gd; j++) {
	    int nr = ii+j;
	    theta = mpv_cart[nr].theta;
	    phi   = mpv_cart[nr].phi;

	    /* keep only meaningful points of 'mpv' */
	    if (ABS(theta) < 1.e5) {
		double db = pi/n_b, dl = 2*pi/n_l;
		int ib = theta / db,
		    jl = (phi < pi ? pi - phi : 3*pi - phi) / dl,
		    n = ib * n_l + jl,
		    /* some neibouring pixels to n-th pixel */
		    n1 = (ib-1) * n_l + jl,
		    n2 = (ib+1) * n_l + jl,
		    n3 =  ib    * n_l + jl-1,
		    n4 =  ib    * n_l + jl+1;

		if (n1 < 0 || n1 >= y_c)  n1 = -1;
		if (n2 < 0 || n2 >= y_c)  n2 = -1;
		if (n3 < 0 || n3 >= y_c)  n3 = -1;
		if (n4 < 0 || n4 >= y_c)  n4 = -1;
		if (mpv[n] ||
		    ( n1 >= 0 && mpv[n1] ) ||
		    ( n2 >= 0 && mpv[n2] ) ||
		    ( n3 >= 0 && mpv[n3] ) ||
		    ( n4 >= 0 && mpv[n4] )
		) {
		      mpv_glob[nr] = 1;  /* locate in general map */
		      mpv[n] = 0;
		}
	    }
	    /* keep only meaningful points of 'mpv_cart' */
	    {
		int n1 = (i-1) * nx_gd + j,
		    n2 = (i+1) * nx_gd + j,
		    n3 =  i    * nx_gd + j-1,
		    n4 =  i    * nx_gd + j+1;

		if (n1 < 0 || n1 >= x_c)  n1 = -1;
		if (n2 < 0 || n2 >= x_c)  n2 = -1;
		if (n3 < 0 || n3 >= x_c)  n3 = -1;
		if (n4 < 0 || n4 >= x_c)  n4 = -1;
		if (mpv_cart[nr].col
//               || ( n1 >= 0 && mpv_cart[n1].col ) ||
//                  ( n2 >= 0 && mpv_cart[n2].col ) ||
//                  ( n3 >= 0 && mpv_cart[n3].col ) ||
//                  ( n4 >= 0 && mpv_cart[n4].col )
		) {
		    /* locate in general map */
		    mpv_glob[nr] = mpv_cart[nr].col;
		    mpv_cart[nr].col = 0;
		}
	    }
	}
    }

    /* put map with vector graphics on the image */
    for(i=0; i<ny_gd; i++) {
	for(j=0; j<nx_gd; j++) {
	    if (mpv_glob[i*nx_gd+j] > 1e-29) {
		move_gd(j,i);
		point_gd(j, i);
	    }
	}
    }

    free((void *)mpv);
    free((void *)mpv_cart);
    free((void *)mpv_glob);
    return;
}

/* Galactic Spur */
void plot_spur()
{
    /* move(b,l), cont(b,l): b,l in degs */
    Pg_move(10., 90.);
    Pg_cont(16., 83.);
    Pg_cont(13., 75.);
    Pg_cont(9.,  68.5);
    Pg_cont(15., 64.);
    Pg_cont(21., 60.);
    Pg_cont(26., 53.);
    Pg_cont(32., 45.);
    Pg_cont(41., 40.);
    Pg_cont(47., 38.);
    Pg_cont(60., 37.);
    Pg_cont(74., 22.);
    Pg_cont(77.,-30.);
    Pg_cont(70.,  0.);
    Pg_cont(60., 15.);
    Pg_cont(53., 21.);
    Pg_cont(51., 10.);
    Pg_cont(45., 15.);
    Pg_cont(35., 18.);
    Pg_cont(29., 17.);
    Pg_cont(22., 15.);
    Pg_cont(21.,  7.);
    return;
}
void plot_milky_way()
{
    return;
}
void plot_src()
{
// THE ASTROPHYSICAL JOURNAL, 498:689-703, 1998 May 10, CARL HEILES 1
//The Gum Nebula is large, with diameter ~35deg, (l,b)=(258,-2)
//Eridanus Loop b...  l=205     b=-19     radius=20
//...
// Rho Ophiuchi cloud complex,  R.A. 16h 28m 06s, Dec=-24d 32.5m,d=4.5x6.5d
    return;
}
