#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>

#include "glesp.h"
#include "ff.h"

#define PR      fprintf
#define ER      stderr
#define ST      stdout

/*
 *
 *  Oleg Verkhodanov (TAC;SAO), vo@sao.ru, 2002
 *   last modification, 1/10/2004; 2/12/2004, 06/09/2009, 23/01/2012
 *   23/04/2012, 2013, 2015, 2016
 */
extern void   setopt(int, char **), usage(void), usage2(void);

extern int sun_data_key;
extern int swap_write_key;
char zone_file[81];    /* file with list of zones to be processed */
char map_file[161], outfile[161];
char dir_out[161];          /* output directory */
char cr_file[161];
double cut_low = -1.e-30, cut_up = 1.e+30;
double lat0 = -1000., lon0 = -1000., lat1 = -1000., lon1 = -1000.;
double latR0 = -1000., lonR0 = -1000., latR1 = -1000., lonR1 = -1000.;
int cut_values = 0;
int remove_flag=0, remain_flag=0;
int print_flag=0;
double dummy_value = 0.0;
double zone_radius = -1.;   /* radius of removed circle in degs */
double delta_dec = 0.0;
int zone_circle = 0;
int ffile_flag = 0;         /* output to FITS-file              */
int inp_ff_flag = 0;        /* input F-file                     */
double conv_factor = 1.0;   /* conversion factor of temperature */
int coors_type = 0;         /* input table in polar coordinates. 1=gal,2=eq*/
int ext_pol_map = 0;        /* to extract polarization maps     */
char format_out[81];        /* ASCII output format              */
char hp_mask_file[161];     /* HPix mask file to overlay mask   */
char gl_mask_file[161];     /* GLESP mask file to overlay mask  */
double ring_glatitude  = -1001.,
       ring_glongitude = -1001.,
       ring_decl = -1001.,
       ring_ra   = -1001.;
int par_def = 0;            /* source parameter definition     */
int flag_slat=0;            /* option for '-sq' step by LAT    */
double dLAT=0.;             /* new step by LAT                 */
int flag_slon=0;            /* option for '-sq' step by LON    */
double dLON=0.;             /* new step by LON                 */
int flag_equal_dRA_dDec=0;  /* when =1 set same angular step   */
double factor_nx = 1.0,     /* proprotional size factors       */
       factor_ny = 1.0;     /*     of extracted image          */
int flag_gnomo = 0;         /* if flag=1, proj. without cos d  */

int format_ext_areas = 0;   /* flag to extract areas form list */
char file_ext_areas[80];    /* file with extracted areas       */
double b_size = 0.5;        /* default box size for areas: .5d */

double lat_c,  /* galactic coordinate of ring center or rectang. corner */
       lon_c,  /* galactic coordinate of ring center or rectang. corner*/
       lat1_c, /* galactic coordinate of another rectang. corner */
       lon1_c, /* galactic coordinate of another rectang. corner */
       r_c,    /* radius of inner circle */
       w_c;    /* radius of ring width   */
int flag_conc_fig = 0;  /* 1-circle, 2 - rectangle */
int area_coordinate_system = 'g';    /* cut area in galactic coordinates */
int rand_nlines = 0;     /* selection of N=rand_nlines random lines   */
int rand_seed  = -1;     /* random seed fo selection of N=rand_nlines */
int do_stacking = 0;  /* stacking procedure for zones    */
char file_stack[81];  /* list of stacking zones          */
int  stat_area = 0;   /* calculate simple statistics in area */


int eq_input = 0;
struct _cross_sect_ {       /* input cross section by RA        */
	double ra, dec;
	double st, st2;
	int n, n2;
} cs;
#define NV_INTR 10
struct _vv_  { double low, high; } v_intr [NV_INTR];
int nv_intr=0;

char *myname="mapcut";

int main(argc, argv)
int argc;
char **argv;
{
    register int i=0, j=0;
    double *f = (double *)0;
    double *x = (double *)0;
    int *nph = (int *)0;
    int nx, np;
    double *r = (double *)0;
    FILE *fp_out = (FILE *)0;
    char x_type[40], grid_type[40];
    int map_out_flag = 1;
    char output_name [301];

    map_file[0] = outfile[0] = cr_file[0] = '\0';
    zone_file[0] = '\0';
    cs.n = cs.n2 = -1;

    /* select grid output keyword values */
    strcpy(x_type, "COS(THETA)");
    strcpy(grid_type, "PHI(X)+LEGENDRE");
    strcpy(tmpcat, "/tmp");
    dir_out[0] = hp_mask_file[0] = gl_mask_file[0] = '\0';
    output_name[0] = '\0';

    strcpy(format_out, "%15.8e");

    /* loop over external parameters */
    setopt(argc, argv);

    /* create directory if required */
    if ( *dir_out != '\0' ) {
	 DIR *dr;
	 if ((dr = opendir(dir_out)) == NULL ) {
	     if (mkdir(dir_out, 0755) != 0) {
		PR(ER, "%s: problem with opening diretory '%s'...\n",
		   myname, dir_out);
		PR(ER, "output directory will be './'\n");
		strcpy(dir_out, ".");
	     }
	 } else
	     closedir (dr);
    }

    /* extract polarization maps */
    if ( ext_pol_map ) {
	double *q = (double *)0, *u = (double *)0;
	char p_outfile[81];
	int test_flag = test_glesp_map(map_file);

	if (test_flag != 3) { /* not  FITS-file 3 maps (T,Q,U) */
	    PR(ER,
	       "%s: file '%s' does not contain 3 polarization GLESP maps.\n",
	       myname, map_file);
	    exit(1);
	}
	in_glesp_3map(&f, &q, &u, &x, &nx, &np, &nph, map_file);

	if (ext_pol_map != 1) {
		if (strcmp(outfile, "stdout") == 0)  strcpy(outfile, "-");
		if (*outfile == '\0' || strcmp(outfile, "-") == 0)
		    fp_out = stdout;
		else {
		    if ( *dir_out == '\0' ) {
			if((fp_out = fopen(outfile, "w")) == (FILE *)0) {
			    PR(ER, "%s: can not create file '%s'\n",
				myname, outfile);
			    exit(1);
			}
		    } else {
			sprintf(output_name, "%s/%s", dir_out, outfile);
			if((fp_out = fopen(output_name, "w")) == (FILE *)0) {
			    PR(ER, "%s: can not create file '%s'\n",
				myname, output_name);
			    exit(1);
			}
		    }
		}
	}
	/*  extract polarization maps */
	switch (ext_pol_map) {
	case 'q':
		/* Q-map output in FITS-format */
		outp_glesp_map(q, x, nx, np,nph, fp_out, x_type, grid_type);
		break;
	case 'u':
		/* U-map output in FITS-format */
		outp_glesp_map(u, x, nx, np,nph, fp_out, x_type, grid_type);
		break;
	case 't':
		/* T-map output in FITS-format */
		outp_glesp_map(f, x, nx, np,nph, fp_out, x_type, grid_type);
		break;
	case 1: /* output 3 maps into separate file */
		if ( *dir_out == '\0' )
		    sprintf(p_outfile, "%s_t",  outfile);
		else
		    sprintf(p_outfile, "%s/%s_t",  dir_out, outfile);
		if((fp_out = fopen(p_outfile, "w")) == (FILE *)0) {
badfile:
		  PR(ER,"%s: can not create file '%s'.\n",myname,p_outfile);
		  exit(1);
		}
		outp_glesp_map(f, x, nx, np,nph, fp_out, x_type, grid_type);
		fclose(fp_out);

		/* Q */
		sprintf(p_outfile, "%s_q",  outfile);
		if((fp_out=fopen(p_outfile, "w")) == (FILE*)0) goto badfile;
		outp_glesp_map(q, x, nx, np,nph, fp_out, x_type, grid_type);
		fclose(fp_out);

		/* U */
		sprintf(p_outfile, "%s_u",  outfile);
		if((fp_out=fopen(p_outfile, "w")) == (FILE*)0) goto badfile;
		outp_glesp_map(u, x, nx, np,nph, fp_out, x_type, grid_type);
		fclose(fp_out);

		break;
	}
	if (ext_pol_map != 1 && fp_out != stdout)  fclose(fp_out);

	exit(0);
    }

    /**************************/
    /* simple anisotropy maps */
    /**************************/

    read_glesp_map(&f, &x, &nx, &np, &nph, map_file);

    /*--------------------------------*/
    /* set limits of values to be cut */
    /*--------------------------------*/
    if (nv_intr) {
	for(i=0; i<nx; i++) {
	    int ii = i * np;
	    for(j=0; j<nph[i]; j++) {
		int n = ii+j;
		int flag = 0, k;
		for (k=0; k<nv_intr; k++) {
		    if (v_intr[k].low < f[n] && f[n] < v_intr[k].high){
			flag = 1;
			break;
		    }
		}
		if ( !flag ) f[n] = dummy_value;
	    }
	}
    }

    if (cut_values) {
	  for(i=0; i<nx; i++) {
	      int ii = i * np;
	      for(j=0; j<nph[i]; j++) {
		    double *ff = f+ii+j;
		    if (cut_values == -1) {
			if(cut_low < *ff && *ff < cut_up)  *ff = dummy_value;
		    } else if (cut_values == 10) {
			if(cut_low < *ff)  *ff = dummy_value;
			else *ff = cut_low;
		    } else if (cut_values == 11) {
			if(cut_low > *ff)  *ff = dummy_value;
			else *ff = cut_low;
		    } else {
			if(*ff > cut_up )  *ff = cut_up ;
			if(*ff < cut_low)  *ff = cut_low ;
		    }
	      }
	  }
    }
    if ( remove_flag || remain_flag ) {
       register int is=0;
       int ns = 1;
       struct _zone_ {
	    double latR0, lonR0;
	    double latR1, lonR1;
	    double radius;
	    double dr;
	    int type;               /* zone type */
       } *zz = (struct _zone_ *)0;
       extern void map_zone(struct _zone_ *, int, int,
	       double *, double *, int, int, int *, double);
       int function_type = remove_flag ? 1 : 0;
       int zone_type = zone_circle ? 0 : 1;

       if (zone_file[0] != '\0') {
	    FILE *fz;
	    char str[81];
	    int rk=0, rko=0, *rand_arr = (int *)0;

	    /* read data from file with zones */
	    if((fz = fopen(zone_file, "r")) == (FILE *)0) {
		PR(ER, "%s: I can not open file '%s'...\n",
		    myname, zone_file);
		exit(1);
	    }
	    /* get number of zones */
	    is =0;
	    while(fgets(str, 80, fz) != NULL) {
		if (*str == '#' || *str == '\n')
		    continue;
		is++;
	    }
	    /* allocate memory */
	    i = ns = is;
	    zz = (struct _zone_ *)calloc(i, sizeof *zz);
	    if (zz == (struct _zone_ *)0) {
nomem:
		PR(ER, "%s: no memory for temporal array of '%d' lements.\n",
		    myname, i);
		exit(1);
	    }

	    /* set random lines with zone to be read */
	    /* for estimation of scatter             */
	    if (rand_nlines > 0) {
		extern void random_numbers(int *, int, int, int);
		if((rand_arr = (int *)calloc((unsigned)rand_nlines,
			   sizeof *rand_arr)) == (int *)0) {
		    PR(ER, "%s: no memory for temporal array.\n", myname);
		    exit(1);
		}
		/* generate rand_lines numbers in rand_arr */
		random_numbers(rand_arr, rand_nlines, ns, rand_seed);
	    }

	    /* read zones  */
	    is=0; /* number of input line */
	    rk=0; /* number of random line in array */
	    fseek(fz, 0l, 0);
	    while(fgets(str, 80, fz) != NULL) {
		char s1[30], s2[30], s3[30], s4[30];
		char *cs = str;
		int n_par;  /* number of parameters per string */
		if (*str == '#' || *str == '\n')
			continue;
		while(*cs) if (*cs == '#') {*cs = '\0'; break;} else cs++;

		/* select random line if one needs */
		if (rand_nlines) {
			int flag=0;
			/* for sorted rand_arr[] */
			for(rk=rko; rk<rand_nlines; rk++)
			      if (is == rand_arr[rk])
				    {flag=1; break;}
			if (!flag) {
			    is++;
			    continue;
			}
			rko = rk;
// PR(ER, "%d,%s", is,str);
		}

		n_par = n_skip_wd(str);
		switch (n_par) {
		case 1:
		    PR(ER,"%s: there is only 1 parameter in the string:\n",
			    myname);
		    PR(ER,"%s", str);
		    PR(ER, "\n%s: I've skipped it...\n", myname);
		    break;
		case 2:
		    zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		    zz[is].radius = zone_radius;
		    zz[is].type = 0;
		    break;
		case 3:
		    zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		    if (zone_radius > 0.)
			zz[is].radius = zone_radius;
		    else
			zz[is].radius = deg2rad((skip_wd(2,str,s3),s3));
		    zz[is].type = 0;
		    break;
		case 4:
		    zz[is].latR0 = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0 = deg2rad((skip_wd(1,str,s2),s2));
		    zz[is].latR1 = deg2rad((skip_wd(2,str,s3),s3));
		    zz[is].lonR1 = deg2rad((skip_wd(3,str,s4),s4));
		    if (zone_type == 0) {
			    zz[is].radius = zz[is].latR1;
			    zz[is].type = 0;
		    } else  zz[is].type = 1;
		    break;
		}
		is++;
	    }

	    fclose(fz);
	    if (rand_arr != (int *)0)  free((void *)rand_arr);

       } else {
	 if ( flag_conc_fig ) {  /* 1-circle, 2 - rectangle */
	    int fun_type = remove_flag ? 1 : 0; /* remain=0 or remove=1 */
	    extern void map_ring_zone();

	    i = ns = 1;
	    zz = (struct _zone_ *)calloc(i, sizeof *zz);
	    if (zz == (struct _zone_ *)0) goto nomem;
	    if ( flag_conc_fig == 1) {
		zz[0].latR0 = lat_c;  zz[0].lonR0 = lon_c;
		zz[0].radius = r_c;
		zz[0].dr = w_c;
		zz[0].type = 0; /* circle */
	    } else {
		zz[0].latR0 = lat_c;   zz[0].lonR0 = lon_c;
		zz[0].latR1 = lat1_c;  zz[0].lonR1 = lon1_c;
		zz[0].dr = w_c;
		zz[0].type = 1; /* rectangular */
	    }
	    map_ring_zone(zz, ns, fun_type,f,x,nx,np,nph, dummy_value);
	    free((void *)zz);
	 } else {
	   /* checking single zone */

	  if (latR0 < -900 && latR1< -900 && lonR0 < -900 &&
	     lonR1 < -900 && zone_radius < 0) {
		/* noze zone selected */
		ns = 0;
	  } else {
	    /* allocate memory for 1 zone */
	    i = ns = 1;
	    zz = (struct _zone_ *)calloc(i, sizeof *zz);
	    if (zz == (struct _zone_ *)0) goto nomem;
	    zz[0].latR0 = latR0;  zz[0].lonR0 = lonR0;
	    zz[0].latR1 = latR1;  zz[0].lonR1 = lonR1;
	    zz[0].radius = zone_radius;
	    if (latR1 < -900.) {
		    zz[0].type = 0;
	    } if (lonR1 < -900.) {
		    if (zone_radius > 0.)
			zz[is].radius = zone_radius;
		    else
			zz[is].radius = latR1;
		    zz[0].type = 0;  /* circle */
	    } else {
		    zz[0].type = 1;  /* rectangular */
	    }
	  }
	 }
       }

       if ( !flag_conc_fig && ns) {  /* 1-circle, 2 - rectanhular */
	   map_zone(zz, ns, function_type, f, x, nx, np, nph, dummy_value);

	   /* free memory from zone-array */
	   free((void *)zz);
       }
    }

    /* set mask from HPix-file */
    if (*hp_mask_file != '\0') {
	    extern void mask_hp(double*,double*,int,int,int*, char*, double);
	    if ( remain_flag ) nx = -nx;
	    mask_hp(f,x,nx,np,nph,hp_mask_file,dummy_value);
	    if ( remain_flag ) nx = -nx;
    }
    /* set mask from GLESP-file */
    if (*gl_mask_file != '\0') {
	    extern void mask_gl(double*,double*,int,int,int*, char*, double);
	    if ( remain_flag ) nx = -nx;
	    mask_gl(f,x,nx,np,nph,gl_mask_file,dummy_value);
	    if ( remain_flag ) nx = -nx;
    }


    /********************/
    /* open output file */
    /********************/
    if (strcmp(outfile, "stdout") == 0)  strcpy(outfile, "-");
    if (*outfile == '\0' || strcmp(outfile, "-") == 0)
	fp_out = stdout;
    else {
	if ( *dir_out == '\0' ) {
	    if((fp_out = fopen(outfile, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, outfile);
		exit(1);
	    }
	} else {
	    sprintf(output_name, "%s/%s", dir_out, outfile);
	    if((fp_out = fopen(output_name, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, output_name);
		exit(1);
	    }
	}
    }

    if (*cr_file != '\0') {
	/***************************************/
	/* extract vector using data from file */
	/***************************************/
	struct _coor_  *cr = (struct _coor_ *)0;
	int n_cr;
	double p = M_PI/180.;

	if (inp_ff_flag) {
	    /* read (lat,lon) from F-file */
	    read_ff(&cr, &n_cr, cr_file, 1, delta_dec);

	    i = n_cr;
	    if ((r = (double*)calloc((unsigned)i,sizeof *r)) == (double *)0)
		goto nomem;

	    /* read vector */
	    for(i=0; i<n_cr; i++) {
		 int k;
		 double alp=cr[i].lat,
			del=cr[i].lon,
			lat=cr[i].lat, lon=cr[i].lon;
		 if (eq_input)  {
			alp /= 15.*p;
			del /= p;
			equat2gal(alp, del, &lat, &lon);
			lat *= p;
			lon *= p;
		 }
		 double theta, phi;
		 gal2pol(lat, lon, &theta, &phi);
		 k = ang2n(theta, phi, x, nx, np, nph);
		 r[i] = f[k] * conv_factor;
	    }

	    /* write to F-file */
	    write_ff(r, n_cr, fp_out);
	} else {
	    int n_tp = nx * np;
	    if (coors_type == 2) {
		    /* read equatorial (alp,del) in radians from file    */
		    /* and converse them to (lat,lon) radians            */
		    read_data_eq2gal(&cr, &n_cr, cr_file);
	    } else if (coors_type == 1) {
		    /* read galactic (lat,lon) in radians from file */
		    /* and converse them to radians        */
		    read_data_gal(&cr, &n_cr, cr_file);
	    } else {
		    /* read (theta,phi) in radians from file */
		    read_data_pol(&cr, &n_cr, cr_file);
	    }

	    i = n_cr;
	    if ((r =(double *)calloc((unsigned)i, sizeof *r))==(double *)0)
		goto nomem;

	    /* read vector and print */
	    for(i=0; i<n_cr; i++) {
	       int k;
	       double theta, phi;
	       if (coors_type == 1) {
		      /* for galactic coordinates */
		      gal2pol(cr[i].lat, cr[i].lon, &theta, &phi);
	       }else if (coors_type == 2) {
		      /* for equatorial coordinates */
		      /* same as galactic 'cause conversion was earlier */
		      gal2pol(cr[i].lat, cr[i].lon, &theta, &phi);
	       } else {
		      theta = cr[i].lat;
		      phi = cr[i].lon;
	       }
	       if (theta < 0 || M_PI < theta) {
		    PR(ER,
		     "%s:\n theta-coordinate (theta=%g) for string %d",
		     myname, theta, i+1);
		    PR(ER, " of input file is out of limits [0;pi]\n");
		    exit(1);
	       }
	       k = ang2n(theta, phi, x, nx, np, nph);

	       if (k < 0 || k >= n_tp) {
		    PR(ER,
		    "%s: something wrong with coordinates in the string %d\n",
			  myname, i);
		    exit(1);
	       }
	       r[i] = f[k] * conv_factor;
	       if (coors_type) {
		  double lat = cr[i].lat/M_PI*180,
			 lon = cr[i].lon/M_PI*180;
		  int dd1, dd2, mm1, mm2, c1='+';
		  double ss1, ss2;

		  if (coors_type == 2) {
		      double alp, del;
		      /* convert galactic to equatorial */
		      gal2equat(lat, lon, &alp, &del);
		      if (del < 0.0) { c1 = '-'; del = -del; }
		      dd1 = (int)del;
		      mm1 = (int)((del - dd1)*60.);
		      ss1 = (del - dd1 - mm1/60.) * 3600.;
		      dd2 = (int)alp;
		      mm2 = (int)((alp - dd2)*60.);
		      ss2 = (alp - dd2 - mm2/60.) * 3600.;

		      PR(fp_out,
			     "%02d:%02d:%05.2f  %c%02d:%02d:%04.1f %12.5e\n",
			     dd2,mm2,ss2, c1,dd1,mm1,ss1, r[i]);
		  } else {
		      if (lat < 0.0) { c1 = '-'; lat = -lat; }
		      dd1 = (int)lat;
		      mm1 = (int)((lat - dd1)*60.);
		      ss1 = (lat - dd1 - mm1/60.) * 3600.;
		      if (lon < 0.)    lon += 360.;
		      if (lon > 360.)  lon -= 360.;
		      dd2 = (int)lon;
		      mm2 = (int)((lon - dd2)*60.);
		      ss2 = (lon - dd2 - mm2/60.) * 3600.;

		      PR(fp_out,
			     "%c%02d:%02d:%04.1f  %03d:%02d:%04.1f %12.5e\n",
			     c1,dd1,mm1,ss1, dd2,mm2,ss2, r[i]);
		  }
	       }
	       else
		     PR(fp_out, "%12.5e  %12.5e  %12.5e\n",
			    theta, phi, r[i]);
	    }
	}
	if (cr != (struct _coor_ *)0) free((void *)cr);
	map_out_flag = 0;
    } if (ring_glatitude > -1000. || ring_glongitude > -1000. ) {
	/********************************/
	/* extract ring by 1 coordinate */
	/********************************/
	if (ring_glatitude > -1000.) {
	    /* fixed galactic latitude: 2pi */
	    /*------------------------------*/
	    double theta, phi;
	    int k;
	    if((r = (double*)calloc((unsigned)np, sizeof *r))==(double *)0)
		goto nomem;

	    gal2pol(ring_glatitude, 0.0, &theta, &phi);
	    /* get pixel position number of a ring from phi=0 */
	    k = ang2n(theta, phi, x, nx, np, nph);
	    i = k/np;   /* start position of i-th ring */
	    for(j=0; j<nph[i]; j++) r[j] = f[k+j];

	    free((void *)r);
	} else if ( ring_glongitude > -1000. ) {
	    /* fixed galactic longitude: 2pi */
	    /*-------------------------------*/
	    int k, nx2 = (nx+1)/2, nt = nx2;
	    double *tmp, *t_o, *t_i, *work,
		   dt = acos(x[nx2])-acos(x[nx2-1]);
	    if((r = (double*)calloc((unsigned)np, sizeof *r))==(double *)0)
		goto nomem;
	    if((tmp=(double*)calloc((unsigned)np, sizeof *tmp))==(double *)0)
		goto nomem;
	    if((t_o=(double*)calloc((unsigned)np, sizeof *t_o))==(double*)0)
		goto nomem;
	    if((t_i=(double*)calloc((unsigned)np, sizeof *t_i))==(double*)0)
		goto nomem;
	    if((work=(double*)calloc((unsigned)np,sizeof *work))==(double*)0)
		goto nomem;

	    /* real data distributed by theta */
	    for (i=0; i<nx; i++) {
		double theta, phi;
		gal2pol(acos(x[i]), ring_glongitude, &theta, &phi);
		k = ang2n(theta, phi, x, nx, np, nph);
		r[i] = f[k];
	    }
	    /* go to equidistant points */
	    /* interpolation from the center: nx/2 to both Poles  */
	    /* with a step dt = acos(x[nx2])-acos(x[nx2-1])     */
	    j = M_PI/2 / dt + 0.5;  /* number of intervals */
	    if ( dt < 0 ) {
		for (i=0; i<nx2; i++) t_i[i] = acos(x[nx2-1-i]);
		for (i=0; i<nx2; i++) {
		    t_o[i] = (nx2-1-i) * dt;
		    if (t_o[i] < 0.0) { /* only positive angles */
			nt = i;
			break;
		    }
		}
	    } else {
		for (i=0; i<nx2; i++) t_i[i] = acos(x[nx2-1-i]);
		for (i=0; i<nx2; i++) {
		    t_o[i] = (nx2-1-i) * dt;
		    if (t_o[i] < 0.0) { /* only positive angles */
			nt = i;
			break;
		    }
		}
	    }
	    for (i=0; i<nx2; i++) {
		spline (t_i, r, nx2, 2.e30, 2.e30, work);
		/* interpolate from my points to nx */
		for(j=0; j<nt; j++) {
		    double v;
		    /* used with spline */
		    splint (t_i, r, work, nx2, t_o[j], &v);
		    tmp[j] = v;           /* final array */
		}
	    }

	    free((void *)r);
	    free((void *)tmp);
	    free((void *)t_o);
	    free((void *)t_i);
	    free((void *)work);
	}
    } else if (do_stacking) {
	extern void area_stack(double *, double *, int, int, int *,
	    char *, double **, int *, int *, double *, double *, double, int);
	double *area = (double *)0;
	int nzx=0, nzy=0;
	double dx=0.0, dy = 0.0;
	double box_size = 0.0;
	int flag_coor = 0;

	/* box size in degs */
	if (b_size > 0.) box_size = b_size;

	if (do_stacking == 2) flag_coor = 1;

	area_stack(f,x, nx,np,nph, file_stack,
		 &area, &nzx, &nzy, &dx, &dy, box_size, flag_coor);

	if ( *dir_out == '\0' ) {
	    if((fp_out = fopen(outfile, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, outfile);
		exit(1);
	    }
	} else {
	    sprintf(output_name, "%s/%s", dir_out, outfile);
	    if((fp_out = fopen(output_name, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, output_name);
		exit(1);
	    }
	}

	if (flag_coor) { /* equatorial system */
		/* change variables for output */
		double d_ra = dx, d_dec = dy, ra0 = 0.0, dec0 = 0.0;
		d_ra *= 15.;  /* dRA,  deg */
		ra0  *= 15.;  /* RA,   deg */

		/* write FITS-map */
		write_2d_fits(fp_out, area, nzx, nzy, ra0, dec0, d_ra, d_dec,
		     "RA---TAN", "DEC--TAN", 1., "T");
	} else {
		double dp = dx,
		       dt = -dy; /* minus because from South to North */
		double lon0 = 0.0, lat0 = 0.0;

		/* write FITS-map */
		write_2d_fits(fp_out, area, nzx,nzy, lon0, lat0, dp, dt,
		   "GLON", "GLAT", 1., "T");
	}

	if (stat_area) {
	    int i,j;
	    double aver = 0.0, sig = 0.0;
	    double min = 1e30, max=-1e30, med;
	    int i_min=0, i_max=0, j_min=0, j_max=0;
	    extern int i_cmp_double(void *, void *);

	    for (i=0; i<nzy; i++) {
		for (j=0; j<nzx; j++) {
		    double v = area[i*nzx+j];
		    aver += v;
		    if (v > max) { max = v; i_max = i; j_max = j; }
		    if (v < min) { min = v; i_min = i; j_min = j; }
		}
	    }
	    aver /= nzx * nzy;
	    for (i=0; i<nzy; i++) {
		for (j=0; j<nzx; j++) {
		    double v = area[i*nzx+j] - aver;
		    sig += v*v;
		}
	    }
	    sig /= nzx * nzy;
	    sig = sqrt(sig);

	    qsort(area, nzx*nzy, sizeof(*area), (int (*)())i_cmp_double);
	    med = area [ nzx*nzy/2 ];

	    PR(ST, "staked area: dim=(%d,%d), aver=%g, med=%g, sig=%g\n",
		nzx, nzy, aver, med, sig);

	    if (area != (double *)0)  free((void *)area);
	}

	exit(0);

    } if (cs.n > 0) {
	/****************************************************/
	/* extract vector using parameters of cross-section */
	/****************************************************/
	int n_cr = cs.n;
	double p = M_PI/180;
	double lat = eq_input ? cs.ra  : cs.ra * 15. * p,
	       lon = eq_input ? cs.dec : cs.dec * p,
	       st  = eq_input ? cs.st  : cs.st * 15. * p;
	double st2 = eq_input ? cs.st2 : cs.st2 * p;

	if (cs.n2 > 0)  i = cs.n * cs.n2; else i = n_cr;
	if ((r = (double*)calloc((unsigned)i, sizeof *r)) == (double *)0)
	    goto nomem;

	if (cs.n2 > 0) {
	    /* read vector */
	    for(i=0; i<cs.n2; i++) {
		int ii = n_cr * i;
		double del = lon + st2 * j;
		for(j=0; j<n_cr; j++) {
		    int k;
		    double lat1, lon1;
		    double theta, phi;
		    if (eq_input)  {
			   double alp=lat+st*j;
			   equat2gal(alp, del, &lat1, &lon1);
			   lat1 *= p;
			   lon1 *= p;
		    } else {
			   lat1=lat+st*j; lon1=del;
		    }
		    gal2pol(lat1, lon1, &theta, &phi);
		    k = ang2n(theta, phi, x, nx, np, nph);
		    r[ii+j] = f[k] * conv_factor;
		}
	    }

	    if (ffile_flag) {
	       /* write ff-map */
	       write_simple_map(fp_out, r, n_cr, cs.n2, lon, lat, st, st2,
		    "GLON", "GLAT", 1., "T");
	    } else {
	       /* write ff-map */
	       write_2d_fits(fp_out, r, n_cr, cs.n2, lon, lat, st, st2,
		    "GLON", "GLAT", 1., "T");
	    }
	} else {
	    /* read vector */
	    for(i=0; i<n_cr; i++) {
		 int k;
		 double lat1, lon1;
		 double theta, phi;
		 if (eq_input)  {
			double alp=lat+st*i, del =lon;
			equat2gal(alp, del, &lat1, &lon1);
			lat1 *= p;
			lon1 *= p;
		 } else {
			lat1=lat+st*i; lon1=lon;
		 }
		 gal2pol(lat1, lon1, &theta, &phi);
		 k = ang2n(theta, phi, x, nx, np, nph);
		 r[i] = f[k] * conv_factor;
	    }

	    /* prepare head of F-file */
	    create_ff(FLOATPIX,(long)n_cr,"K", 1.0,0.0,
		   cs.ra, cs.st * 3600., 1.0, "TIME");

	    /* write to F-file */
	    write_ff(r, n_cr, fp_out);
	}
	if (r != (double *)0) { free((void *)r);  r = (double *)0; }
	map_out_flag = 0;
    } else if (lat0 > -1000. ) {
	if (lat1 > -1000.) {
	    /****************************/
	    /* extract rectangular area */
	    /****************************/
	    int nn=0, nm=0;
	    double p=180/M_PI;
	    double dt=0.0, dp=0.0;

	    if (area_coordinate_system == 'q') {
		/* extract rectangular area in equatorial coordinates */
		/******************************************************/
		extern void extract_area_equ(double **,
		  int*,int*,double*,double*,   double,double,
		  double *,double*,int,int,int*, double,double,double,double);
		double ra0=lat0 *p, dec0=lon0 * p, /* rad_r -> RA in hours */
		       ra1=lat1 *p, dec1=lon1 * p, /* rad   -> Dec in degs */
		       d_ra, d_dec;
		double cos_d = cos((lon0+lon1)/2);
		double box_size_ra = (ra1 - ra0);
		double ra = (ra0 + ra1) / 2.;

		/* if equidistant RA and Dec */
		if (flag_equal_dRA_dDec) nm = -1000;

		/* coding values of 'flag_slat' and 'flag_slon'          */
		/* in steps 'd_ra' and 'd_dec' to use in 'extract area'  */
		if (flag_slat==1)  d_ra = GLESPMASK;
		else if(flag_slat==2)  d_ra = dLAT * p/15;  /* in hours */
		if (flag_slon==1)  d_dec = GLESPMASK;
		else if(flag_slon==2)  d_dec = dLON *p;     /* in degs  */

		/* RA limits in hours /*/
		if ( flag_gnomo ) {
		    ra0 = ra - .5 * box_size_ra;
		    ra1 = ra0 + box_size_ra;
		} else {
		    ra0 = ra - .5 * box_size_ra;
		    ra1 = ra0 + box_size_ra;
		}
		/* here (ra1,ra2) in hours, (dec1,dec2) in degs */
		extract_area_equ(&r,&nm,&nn, &d_ra,&d_dec,
			  factor_nx,factor_ny,
			  f, x, nx, np, nph,  ra0,dec0,ra1,dec1);

		/* change variables for output */
		if ( flag_gnomo ) {
		    d_ra *= 15.;  /* dRA,  deg */
		    ra0  *= 15.;  /* RA,   deg */
		} else {
		    d_ra *= 15.*cos_d;  /* dRA,  deg */
		    ra0  *= 15.;  /* RA,   deg */
		}
		/* write FITS-map */
		write_2d_fits(fp_out, r, nm, nn, ra0, dec0, d_ra, d_dec,
		     "RA---TAN", "DEC--TAN", 1., "T");

		/* renaming parameters for source detection procedure */
		lon0 = ra0;  lat0 = dec0;  dp = d_ra;  dt = d_dec;
	    } else {
		/* extract rectangular area in galactic coordinates */
		/****************************************************/
		extern void extract_area_gal(double **,
		  int*,int*,double*,double*,   double,double,
		  double *,double*,int,int,int*, double,double,double,double);

		/* coding values of 'flag_slat' and 'flag_slon' */
		/* in steps 'dt' and 'dp' to use in 'extract area'  */
		if (flag_slat==1)  dt = GLESPMASK;
		else if(flag_slat==2)  dt = dLAT;
		if (flag_slon==1)  dp = GLESPMASK;
		else if(flag_slon==2)  dp = dLON;

		/* output array 'r' and steps 'dp','dr' in radians */
		/* lat0,lon0,lat1,lon1 - coordinates of requested area, rad */
		extract_area_gal(&r,&nm,&nn,&dp,&dt,
			  factor_nx,factor_ny,
			  f, x, nx, np, nph,  lat0,lon0,lat1,lon1);

		/* output */
		/*--------*/
		lon0 *= p;
		lat0 *= p;
		dt *= -p; /* minus because of map from South to North */
		dp *=  p;
		if (ffile_flag) {
		    /* write ff-map */
		    write_simple_map(fp_out, r, nm, nn, lon0, lat0, dp, dt,
		       "GLON", "GLAT", 1., "T");
		} else {
		    /* write FITS-map */
		    write_2d_fits(fp_out, r, nm, nn, lon0, lat0, dp, dt,
		       "GLON", "GLAT", 1., "T");
		}
	    }

	    /* definition of source parameters */
	    /*---------------------------------*/
	    if (par_def) {
		  extern void source_pars();
		  source_pars(r, nm, nn, lon0,lat0,dp,dt);
	    }

	    map_out_flag = 0;
	} else {
	    if (!zone_circle) {
		/*********************/
		/* print pixel value */
		/*********************/
		double theta, phi;
		gal2pol(lat0, lon0, &theta, &phi);
		i = ang2n(theta, phi, x, nx, np, nph);

		if (print_flag)
		   PR(fp_out, "%15.8e  %15.8e  %15.8e\n",
		       theta, phi, f[i] * conv_factor);
		else
		   PR(fp_out, "theta = %g, phi = %g, t = %g\n",
		       theta, phi, f[i] * conv_factor);

		map_out_flag = 0;
	    }
	}
    } else if (format_ext_areas) {   /* flag to extract areas form list */
	char str[81];
	char name[51], sra[30], sdec[30], ssize[20];
	double box_size = 0.0, bb=0.0, ra=0.0, dec=0.0, glat=0.0, glon=0.0,
	       dt, dp;
	double p = 180./M_PI;
	int nm=0, nn=0;
	FILE *fpl = (FILE *)0, *fpo = (FILE *)0;

	if ((fpl = fopen(file_ext_areas, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n",
		    myname, file_ext_areas);
		exit(1);
	}
	while(fgets(str, 80, fpl) != NULL) {
	    name[0] = sra[0] = sdec[0] = ssize[0] = '\0';

	    if (*str == '#' || *str == '\n' || *str == '\r')
		continue;

	    switch(format_ext_areas) {
	    case 'q': /* equatorial coordinates */
		  {
		    double s_ra; int h_ra, m_ra;
		    double ds, s_dec; int d_dec, m_dec, c;
		    sscanf(str, "%s%s%s", sra, sdec, ssize);
		    ra = str2deg(sra);
		    dec = ds = str2deg(sdec);
		    bb = str2deg(ssize);
		    h_ra = (int)ra;
		    m_ra = (int)((ra - h_ra) * 60.);
		    s_ra = (ra - h_ra - m_ra/60.) * 3600;
		    if ( ds < 0. ) {ds = ABS(dec); c='-';} else c='+';
		    d_dec = (int)ds;
		    m_dec = (int)((ds - d_dec) * 60.);
		    s_dec = (ds - d_dec - m_dec/60.) * 3600;
		    sprintf(name, "%02d%02d%02.0f%c%02d%02d%02.0f",
			h_ra, m_ra, s_ra, c, d_dec, m_dec, s_dec);
		  }
		  break;
	    case 'Q': /* equatorial coordinates */
		    sscanf(str, "%s%s%s%s", name, sra, sdec, ssize);
		    ra = str2deg(sra);
		    dec = str2deg(sdec);
		    bb = str2deg(ssize);
		    break;
	    case 'G': /* galactic coordinates */
		    sscanf(str, "%s%s%s%s", name, sra, sdec, ssize);
		    glat = str2deg(sra);
		    glon = str2deg(sdec);
		    bb = str2deg(ssize);
		    break;
	    case 'g': /* galactic coordinates */
	    default :
		  {
		    double ds=0.0, s_lat; int d_lat, m_lat, c;
		    double s_lon; int d_lon, m_lon;
		    sscanf(str, "%s%s%s", sra, sdec, ssize);
		    glat = ds = str2deg(sra);
		    glon = str2deg(sdec);
		    bb = str2deg(ssize);
		    d_lon = (int)glon;
		    m_lon = (int)((glon - d_lon) * 60.);
		    s_lon = (glon - d_lon - m_lon/60.) * 3600;
		    if ( ds < 0. ) {ds = ABS(glat); c='-';} else c='+';
		    d_lat = (int)ds;
		    m_lat = (int)((ds - d_lat) * 60.);
		    s_lat = (ds - d_lat - m_lat/60.) * 3600;
		    sprintf(name, "G%03d%02d%02.0f%c%02d%02d%02.0f",
			 d_lon, m_lon, s_lon, c, d_lat, m_lat, s_lat);
		  }
		  break;
	    }
	    /* box size in degs */
	    if (bb > 0.) box_size = bb;
	    else if (b_size > 0.) box_size = b_size;
	    nm=0, nn=0;

	    /* create output file for Basic FITS data of area */
	    if ( *dir_out == '\0' )
		sprintf (str, "%s.fts", name);
	    else
		sprintf (str, "%s/%s.fts", dir_out, name);
	    if ((fpo = fopen(str, "w")) == (FILE *)0) {
		    PR(ER, "%s: can not open file '%s'\n",  myname, str);
		    exit(1);
	    }

	    /* extract rectangular areas according to list */
	    /*---------------------------------------------*/
	    if (format_ext_areas == 'q' || format_ext_areas == 'Q' ) {
		/* extract rectangular area in equatorial coordinates */
		/******************************************************/
		extern void extract_area_equ(double **,
		  int*,int*,double*,double*,   double,double,
		  double *,double*,int,int,int*, double,double,double,double);
		double ra0, dec0, ra1, dec1,  d_ra=0.0, d_dec=0.0;
		double cos_d = cos (dec / p);

		/* RA limits in hours /*/
		if ( flag_gnomo ) {
		    ra0 = ra - .5 * box_size / 15;
		    ra1 = ra0 + box_size / 15;
		} else {
		    ra0 = ra - .5 * box_size / (15. * cos_d);
		    ra1 = ra0 + box_size / (15. * cos_d) ;
		}
		/* Dec limits in degs /*/
		dec0 = dec - .5 * box_size;
		dec1 = dec0 + box_size;


		/* if equidistant RA and Dec */
		if (flag_equal_dRA_dDec) nm = -1000;

		/* coding values of 'flag_slat' and 'flag_slon'          */
		/* in steps 'd_ra' and 'd_dec' to use in 'extract area'  */
		if (flag_slat==1)  d_ra = GLESPMASK;
		else if(flag_slat==2)  d_ra = dLAT * p/15;  /* in hours */
		if (flag_slon==1)  d_dec = GLESPMASK;
		else if(flag_slon==2)  d_dec = dLON *p;     /* in degs  */

		/* here (ra1,ra2) in hours, (dec1,dec2) in degs */
		extract_area_equ(&r,&nm,&nn, &d_ra,&d_dec,
			  factor_nx,factor_ny,
			  f, x, nx, np, nph,  ra0,dec0,ra1,dec1);

		/* change variables for output */
		if ( flag_gnomo ) {
		    d_ra *= 15.;  /* dRA,  deg */
		    ra0  *= 15.;  /* RA,   deg */
		} else {
		    d_ra *= 15.*cos_d;  /* dRA,  deg */
		    ra0  *= 15.;        /* RA,   deg */
		}

		/* write FITS-map */
		write_2d_fits(fpo, r, nm, nn, ra0, dec0, d_ra, d_dec,
		     "RA---TAN", "DEC--TAN", 1., "T");
	    } else {
		/* extract rectangular area in galactic coordinates */
		/****************************************************/
		extern void extract_area_gal(double **,
		  int*,int*,double*,double*,   double,double,
		  double *,double*,int,int,int*, double,double,double,double);

		/* GLON in degs /*/
		lon0 = glon - .5 * box_size / cos (glat / p);
		lon1 = lon0 + box_size / cos (glat / p);
		/* GLAT limits in degs /*/
		lat0 = glat - .5 * box_size;
		lat1 = lat0 + box_size;

		/* coding values of 'flag_slat' and 'flag_slon' */
		/* in steps 'dt' and 'dp' to use in 'extract area'  */
		dt = dp = 0.0;
		if (flag_slat==1)  dt = GLESPMASK;
		else if(flag_slat==2)  dt = dLAT;
		if (flag_slon==1)  dp = GLESPMASK;
		else if(flag_slon==2)  dp = dLON;

		/* output array 'r' and steps 'dp','dr' in radians */
		/* lat0,lon0,lat1,lon1 - coordinates of requested area, rad */
		extract_area_gal(&r,&nm,&nn,&dp,&dt,
			  factor_nx,factor_ny,
			  f, x, nx, np, nph,  lat0,lon0,lat1,lon1);

		/* output */
		/*--------*/
		lon0 *= p;
		lat0 *= p;
		dt *= -p; /* minus because of map from South to North */
		dp *=  p;
		if (ffile_flag) {
		    /* write ff-map */
		    write_simple_map(fpo, r, nm, nn, lon0, lat0, dp, dt,
		       "GLON", "GLAT", 1., "T");
		} else {
		    /* write FITS-map */
		    write_2d_fits(fpo, r, nm, nn, lon0, lat0, dp, dt,
		       "GLON", "GLAT", 1., "T");
		}
	    }

	    if (fpo != (FILE *)0) fclose(fpo);

	}
	if (fpl != (FILE *)0) fclose(fpl);
	exit(0);
    }

    /*  write map  */
    if (print_flag) {
	  /* ASCII   */
	  double theta, phi;
	  char format[51];
	  sprintf(format, "%s  %s  %s\n", format_out, format_out, format_out);

	  for(i=0; i<nx; i++) {
	    int ii = i * np;
	    for(j=0; j<nph[i]; j++) {
		pixel2angle(x, nx, nph, i, j, &theta, &phi);
		PR(fp_out, format,
		    theta, phi, f[ii+j] * conv_factor);
	    }
	  }
    } else {
	/*  binary */
	if (map_out_flag)  {
	    if (ffile_flag) {
		/* map output in F-format */
		ff_write_ext2_map(f,x,nx,np,nph, fp_out, x_type,grid_type);
	    } else {
		/* map output in FITS-format */
		outp_glesp_map(f, x, nx, np,nph, fp_out, x_type, grid_type);
	    }
	}
    }
    fclose(fp_out);

    /* free memory */
    if (r != (double *)0)  free((void *)r);
    if (f != (double *)0)  free((void *)f);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    char s1[50], s2[50], s3[50], s4[50];
    int flag_cr = 0;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
	    switch(argv[i][1]) {
	    case 'c':
		switch(argv[i][2]) {
		case 'c': /* concentric figure */
		  { char vstr[30];
		  switch(argv[i][3]) {
		  case 'r':  /* rectangle */
		     ++i;
		     skip_wd(0, argv[i], vstr);
		     lat_c = deg2rad(vstr);
		     skip_wd(1, argv[i], vstr);
		     lon_c = deg2rad(vstr);
		     skip_wd(2, argv[i], vstr);
		     lat1_c = deg2rad(vstr);
		     skip_wd(3, argv[i], vstr);
		     lon1_c = deg2rad(vstr);
		     skip_wd(4, argv[i], vstr);
		     w_c = deg2rad(vstr); /* radius of ring width */
		     flag_conc_fig = 2;
		     break;
		  default :  /* circle */
		     ++i;
		     skip_wd(0, argv[i], vstr);
		     lat_c = deg2rad(vstr);
		     skip_wd(1, argv[i], vstr);
		     lon_c = deg2rad(vstr);
		     skip_wd(2, argv[i], vstr);
		     r_c = deg2rad(vstr);/* radius of inner circle */
		     skip_wd(3, argv[i], vstr);
		     w_c = deg2rad(vstr); /* radius of ring width */
		     flag_conc_fig = 1;
		     break;
		  }}
		  remain_flag = 1;
		  break;
		case 'i':
		     zone_circle = 1;
		     break;
		case 'r': /* get value by coordinates */
		     get_2_str(argv[++i], s1, s2);
		     /* read coordinate (deflt: galactic) */
		     /* and transfer them to radians      */
		     if(*s1 !='\0') lat0=deg2rad(s1); else lat0= -1001.0;
		     if(*s2 !='\0') lon0=deg2rad(s2); else lon0= -1001.0;
		     if (argv[i-1][3] == 'q') {
			    /* if input equatorial coordinates */
			    double ra = lat0 * 180/M_PI, dec = lon0*180/M_PI;
			    /* ra in hours, dec in degrees */
			    equat2gal(ra, dec, &lat0, &lon0);
			    lat0 *=  M_PI/180.;
			    lon0 *=  M_PI/180.;
		     } else if (argv[i-1][3] == 'p') {
			    /* if input polar coordinates, */
			    /* then polar -> galactic      */
			    lat0 = M_PI/2 - lat0;
		     }
		     flag_cr = 1;
		     break;
		case 's':  /* section */
		     /* section in galactic & equatorial coordinates */
		     cs.ra  = str2deg((skip_wd(0, argv[++i], s1),s1));
		     cs.dec = str2deg((skip_wd(1, argv[i], s1),s1));
		     cs.st  = str2deg((skip_wd(2, argv[i], s1),s1));
		     cs.n   = atoi((skip_wd(3, argv[i], s1),s1));
		     *s1 = '\0';
		     cs.st2 = str2deg((skip_wd(4, argv[i], s1),s1));
		     *s1 = '\0';
		     cs.n2  = atoi((skip_wd(5, argv[i], s1),s1));
		     break;
		case 'u': /* get value by galactic   coordinates */
		case 'g': /* get value by galactic   coordinates */
		case 'q': /* get value by equatorial coordinates */
		      get_4_str(argv[++i], s1, s2, s3, s4);
		      if(*s1 !='\0') lat0=deg2rad(s1); else lat0= -1001.;
		      if(*s2 !='\0') lon0=deg2rad(s2); else lon0= -1001.;
		      if(*s3 !='\0') lat1=deg2rad(s3); else lat1= -1001.;
		      if(*s4 !='\0') lon1=deg2rad(s4); else lon1= -1001.;
		      break;
		default :
		     PR(ER,"%s: illegal flag '%s'\n",myname,argv[i]);
		     usage();
		     break;
		}
		break;
	    case 'C':
		    if (argv[i][2] == 'i') {
			cut_values = -1;
			get_val_str(argv[++i], &cut_low, &cut_up);
		    } else if (argv[i][2] == 'm') {
			cut_values = 10;
			cut_low = atof(argv[++i]);
		    } else  {
			cut_values =  1;
			get_val_str(argv[++i], &cut_low, &cut_up);
		    }
		    break;
	    case 'd':
		    switch (argv[i][2]) {
		    case 'd':
			    if (strncmp(argv[i]+1, "dde", 3) == 0) {
				dLON = deg2rad(argv[++i]);
				flag_slon = 2;
			    } else { /* '-dd' */
				delta_dec = str2deg(argv[++i]);
			    }
			    break;
		    case 'i':  /* '-dir ' */
			    strcpy(dir_out, argv[++i]);
			    break;
		    case 'l':
			    if (strncmp(argv[i]+1, "dla", 3) == 0) {
			       dLAT = deg2rad(argv[++i]);
			       flag_slat = 2;
			    } else if (strncmp(argv[i]+1, "dlo", 3) == 0) {
			       dLON = deg2rad(argv[++i]);
			       flag_slon = 2;
			    }
			    break;
		    case 'o':
			    /* if (strcmp("double", argv[i]+1) == 0) */
			    bitpix_map = DOUBLEPIX;
			    break;
		    case 'r':
			    /* if (strncmp(argv[i]+1, "dra", 3) == 0) */
			    dLAT = deg2rad(argv[++i]);
			    flag_slat = 2;
			    break;
		    default: /* '-d' */
			    i++;
			    if (argv[i][0] == 'M' || argv[i][0] == 'm')
				dummy_value = GLESPMASK;
			    else
				dummy_value = atof(argv[i]);
			    break;
		    }
		    break;
	    case 'e':
		    if (strncmp("eq", argv[i]+1, 2) == 0) {
			    eq_input = 1;
		    } else if (strncmp("ex", argv[i]+1, 2) == 0) {
			    /* extract polarization map */
			    int p = argv[++i][0];
			    if (p == 't' || p == 'T')
				ext_pol_map = 't';
			    else if (p == 'q' || p == 'Q')
				ext_pol_map = 'q';
			    else if (p == 'u' || p == 'U')
				ext_pol_map = 'u';
			    else
				ext_pol_map = 1;
		    } else {
			 PR(ER,"%s: illegal flag '%s'\n",myname,argv[i]);
			 usage();
		    }
		    break;
	    case 'f':
		    if (strncmp("fc", argv[i]+1, 2) == 0) {
			    conv_factor = atof(argv[++i]);
		    } else if (strcmp("float", argv[i]+1) == 0) {
			    bitpix_map = FLOATPIX;
		    } else if (strncmp("form", argv[i]+1, 4) == 0) {
			    strcpy(format_out, argv[++i]);
		    } else if (strcmp("fcr", argv[i]+1) == 0) {
			    /* extract values of map pixels from list */
			    /* in galacti coordinates          */
			    coors_type = 1; /* galactic    */
			    strcpy(cr_file, argv[++i]);
		    } else if (strncmp("fcrq", argv[i]+1, 4) == 0) {
			    /* extract values of map pixels from list */
			    /* in equatorial coordinates       */
				coors_type = 1; /* galactic    */
			    coors_type = 2; /* equatorial */
			    strcpy(cr_file, argv[++i]);
		    } else if (strncmp("fz", argv[i]+1, 2) == 0) {
			format_ext_areas = 'g';
			if (argv[i][3] == 'q') {
			    if (argv[i][4] == '1')
				format_ext_areas = 'Q';
			    else
				format_ext_areas = 'q';
			} else if (argv[i][3] == 'g') {
			    if (argv[i][4] == '1')
				format_ext_areas = 'G';
			    else
				format_ext_areas = 'g';
			}
			strcpy(file_ext_areas, argv[++i]);
			flag_equal_dRA_dDec = 1;
		    } else {
			 PR(ER, "%s: illegal flag '%s'\n",
			      myname, argv[i]);
			 usage();
		    }
		    break;
	    case 'h':
		    if (strncmp("h2", argv[i]+1, 2) == 0)
			usage2();
		    else
			usage();
		    break;
	    case 'i':
		    /* input f-file for cut */
		    inp_ff_flag = 1;
		    strcpy(cr_file, argv[++i]);
		    break;
	    case 'k':
		    if (strncmp("ke", argv[i]+1, 2) == 0) {
			 get_4_str(argv[++i], s1, s2, s3, s4);
			 if(*s1 !='\0') latR0=deg2rad(s1);else latR0= -1001.;
			 if(*s2 !='\0') lonR0=deg2rad(s2);else lonR0= -1001.;
			 if(*s3 !='\0') latR1=deg2rad(s3);else latR1= -1001.;
			 if(*s4 !='\0') lonR1=deg2rad(s4);else lonR1= -1001.;
			 remain_flag = 1;
		    } else {
			 PR(ER, "%s: illegal flag '%s'\n",
			      myname, argv[i]);
			 usage();
		    }
		    break;
	    case 'K':
		    remain_flag = 1;
		    remove_flag = 0;
		    break;
	    case 'l':
		    if (strncmp(argv[i]+1, "li", 2) == 0) {
		      /* set limits to be kept */
		      char s[50], *cs=s, *ps=argv[i+1], *ss;
		      while (*ps) {
			while ((*cs = *ps) != '/' && *cs) {cs++;ps++;}
			*cs = '\0';
			if (*ps == '/') ps++;
			v_intr[nv_intr].low = *s!='\0' ? atof(s):dummy_value;
			if ( index(s,',') == NULL)
			    ss = NULL;
			else
			    ss = index(s,',') + 1;
			v_intr[nv_intr].high =
			    (ss!=NULL && *ss != '\0') ? atof(ss):dummy_value;
			if (fabs(v_intr[nv_intr].low -dummy_value) > 1.e-10 ||
			    fabs(v_intr[nv_intr].high-dummy_value) > 1.e-10 )
				nv_intr++;
			cs=s;
			if (nv_intr == NV_INTR)
			    break;
		      }
		    } else if (strncmp("lat", argv[i]+1,3) == 0) {
			switch(argv[i][4]) {
			case 'm': /* keep: lat < value */
			    if (argv[i][5] == 'e')
				latR0 = deg2rad(argv[++i]);
			    else
				latR0 = deg2rad(argv[++i]) + 1.e-10;
			    latR1 = M_PI/2;
			    lonR0 = 0.;
			    lonR1 = 2*M_PI;
			    break;
			case 'l': /* keep: lat > value */
			    latR0 = -M_PI/2;
			    if (argv[i][5] == 'e')
				latR1 = deg2rad(argv[++i]);
			    else
				latR1 = deg2rad(argv[++i]) - 1.e-10;
			    lonR0 = 0.;
			    lonR1 = 2*M_PI;
			    break;
			}
			remove_flag = 1;
		    } else if (strncmp("lon", argv[i]+1,3) == 0) {
			switch(argv[i][4]) {
			case 'm': /* keep: lon > value */
			    if (argv[i][5] == 'e')
				lonR0 = deg2rad(argv[++i]);
			    else
				lonR0 = deg2rad(argv[++i]) + 1.e-10;
			    lonR1 = 2*M_PI;
			    latR0 = -M_PI/2;
			    latR1 = M_PI/2;
			    break;
			case 'l': /* keep: lon < value */
			    lonR0 = 0.;
			    if (argv[i][5] == 'e')
				lonR1 = deg2rad(argv[++i]);
			    else
				lonR1 = deg2rad(argv[++i]) - 1.e-10;
			    latR0 = -M_PI/2;
			    latR1 = M_PI/2;
			    break;
			}
			remove_flag = 1;
		    } else {
			PR(ER, "%s: illegal flag '%s'\n",
			     myname, argv[i]);
			usage();
		    }
		    break;
	    case 'm':
		    /* mask for data from HP */
		    if(strncmp("mh", argv[i]+1, 2) == 0) {
			 strcpy(hp_mask_file, argv[++i]);
		    } else if(strncmp("mg", argv[i]+1, 2) == 0) {
			 strcpy(gl_mask_file, argv[++i]);
		    } else {
			 PR(ER, "%s: illegal flag '%s'\n", myname, argv[i]);
			 usage();
		    }
		    break;
	    case 'M':
		    /* mask for data from HP */
		    if(strncmp("Mh", argv[i]+1, 2) == 0) {
			 strcpy(hp_mask_file, argv[++i]);
			 dummy_value = GLESPMASK;
		    } else if(strncmp("Mg", argv[i]+1, 2) == 0) {
			 strcpy(gl_mask_file, argv[++i]);
			 dummy_value = GLESPMASK;
		    } else {
			 PR(ER, "%s: illegal flag '%s'\n", myname, argv[i]);
			 usage();
		    }
		    break;
	    case 'n':
		    if(strncmp("nf", argv[i]+1, 2) == 0) {
			 /* change size of extracted image proportionally */
			 /* to (NX x NY) */
			 char st[30];
			 strcpy(st, argv[++i]);
			 if(strchr(st,'x') != NULL) {
			    factor_ny=atof(strchr(st,'x')+1);
			    *(strchr(st,'x')) = '\0';
			    factor_nx=atof(st);
			 } else {
			    factor_nx = factor_ny = atof(st);
			 }
		    } else {
			 PR(ER, "%s: illegal flag '%s'\n", myname, argv[i]);
			 usage();
		    }
		    break;
	    case 'o':
		    if(strncmp("of", argv[i]+1, 2) == 0) {
			    ffile_flag = 1;
		    }
		    strcpy(outfile, argv[++i]);
		    break;
	    case 'p':
		    print_flag = 1;
		    break;
	    case 'r':
		    if (strcmp("rm", argv[i]+1) == 0 ||
			strncmp("re", argv[i]+1, 2) == 0) {
			 get_4_str(argv[++i], s1, s2, s3, s4);
			 if(*s1 !='\0') latR0=deg2rad(s1);else latR0= -1001.;
			 if(*s2 !='\0') lonR0=deg2rad(s2);else lonR0= -1001.;
			 if(*s3 !='\0') latR1=deg2rad(s3);else latR1= -1001.;
			 if(*s4 !='\0') lonR1=deg2rad(s4);else lonR1= -1001.;
			 remove_flag = 1;
		    } else if (strncmp ("ra", argv[i]+1, 2) == 0) {
			 zone_radius = deg2rad(argv[++i]);
			 zone_circle = 1;
		    } else if (strncmp ("rinl", argv[i]+1, 4) == 0) {
			 ring_glongitude = deg2rad(argv[++i]);
		    } else if (strncmp ("rinb", argv[i]+1, 4) == 0) {
			 ring_glatitude  = deg2rad(argv[++i]);
		    } else if (strncmp ("rn", argv[i]+1, 2) == 0) {
			 /* selection of N=rand_nlines random lines */
			 /* from file of zones of sources           */
			 rand_nlines =  atoi(argv[++i]);
		    } else if (strncmp ("rs", argv[i]+1, 2) == 0) {
			 /* random seed for selection of N=rand_nlines */
			 /* random lines from file of zones of sources */
			 rand_seed =  atoi(argv[++i]);
		    } else {
			 PR(ER, "%s: illegal flag '%s'\n", myname, argv[i]);
			 usage();
		    }
		    break;
	    case 'R':
		    remove_flag = 1;
		    remain_flag = 0;
		    break;
	    case 's':
		    switch(argv[i][2]) {
		    case 'a': /* get value by coordinates to cut area */
			 if (strncmp(argv[i]+1, "saq", 3) == 0)
			    area_coordinate_system = 'q'; /* equatorial */
			 get_4_str(argv[++i], s1, s2, s3, s4);
			 /* RA */
			 if(*s1 !='\0') lat0=deg2rad(s1); else lat0= -1001.;
			 /* Dec */
			 if(*s2 !='\0') lon0=deg2rad(s2); else lon0= -1001.;
			 /* size of box, degs */
			 if(*s3 !='\0') lat1=deg2rad(s3); else lat1= -1001.;

			 /* set area bounaries */
			 if (lon0 > -1001.) lon0 -= .5 * lat1; /* in rad */
			 lon1 = lon0 + lat1;
			 if (area_coordinate_system == 'q') {
			    /* equatorial coordinate limits */
			    /* RA */
			    double cos_d = 1.0;
			    double box_size=lat1,  /* in radians */
				   ra=0.0, box_r=0.0;

			    /* set same step by RA and Dec in radians */
			    if (strncmp(argv[i-1]+1, "saq-", 4) != 0) {
				cos_d = cos ((lon0+lon1)/2);
				if (cos_d < 0.) {
				   PR(ER, "%s: cos(delta) < 0... ", myname);
				   PR(ER, "Wrong input coordinates ?\n");
				}
				flag_equal_dRA_dDec = 1;
			    }
			    /* convert RA to houres, and then to h_rads */
			    ra = lat0 * 180/M_PI;  /* rad -> hours */
			    box_r = box_size * 180/M_PI / (15.*cos_d);
			    lat0 = ra - 0.5*box_r;
			    lat1 = ra + 0.5*box_r;

			    lat0 *= M_PI/180.;
			    lat1 *= M_PI/180.;

			 } else {
			    /* galactic coordinate limits */
			    if (lat0 > -1001.)  lat0 -= .5 * lat1;
			    lat1 = lat0 + lat1;
			 }
			 break;
		    case 'q': /* get value by coordinates to cut area */
			 if (strncmp(argv[i]+1, "sqq", 3) == 0)
			    area_coordinate_system = 'q'; /* equatorial */
			 get_4_str(argv[++i], s1, s2, s3, s4);
			 if(*s1 !='\0') lat0=deg2rad(s1); else lat0= -1001.;
			 if(*s2 !='\0') lon0=deg2rad(s2); else lon0= -1001.;
			 if(*s3 !='\0') lat1=deg2rad(s3); else lat1= -1001.;
			 if(*s4 !='\0') lon1=deg2rad(s4); else lon1= -1001.;
			 /* set same step by RA and Dec in radians */
			 if (strncmp(argv[i-1]+1, "sqq+", 4) == 0)
			     flag_equal_dRA_dDec = 1;
			 break;
		    case 'f': /* name of file with coordinates list */
			 if (strncmp(argv[i]+1, "sfg", 3) == 0)
				coors_type = 1; /* galactic    */
			 else if (strncmp(argv[i]+1, "sfq", 3) == 0)
				coors_type = 2; /* equatorial */
			 else if (strncmp(argv[i]+1, "sfe", 3) == 0)
				coors_type = 2; /* equatorial */
			 strcpy(cr_file, argv[++i]);
			 break;
		    case 'd':
			 if (strncmp(argv[i]+1, "sde", 3) == 0)
			    flag_slon = 1;
			 break;
		    case 'r':
			 if (strncmp(argv[i]+1, "sra", 3) == 0)
			    flag_slat = 1;
			 else
			    /* source parameter definition */
			    par_def = 1;
			 break;
		    case 'l':
			 if (strncmp(argv[i]+1, "sla", 3) == 0)
				 flag_slat = 1;
			 else if (strncmp(argv[i]+1, "slo", 3) == 0)
				 flag_slon = 1;
			 break;
		    case 't':
			 if (strcmp(argv[i]+1, "stat") == 0) {
			     stat_area = 1;
			 } else if (strncmp(argv[i]+1, "sta", 3) == 0) {
			     do_stacking = 1;
			     if (argv[i][strlen(argv[i])-1] == 'q')
				  do_stacking = 2;
			     strcpy(file_stack, argv[++i]);
			 }
			 break;
		    case 'u':
			 sun_data_key = 1;
			 break;
		    case 'w':
			 swap_write_key = 1;
			 break;
		    default :
			 break;
		    }
		    break;
	    case 'q':
		    ext_pol_map = 'q';
		    break;
	    case 't':
		    ext_pol_map = 't';
		    break;
	    case 'u':
		    ext_pol_map = 'u';
		    break;
	    case 'v': /* version */
		    PR(ER, "%s: GLESP 2.0, version 1.94 (16/03/2016)\n",
			myname);
		    exit(0);
	    case 'z':
		    switch (argv[i][2]) {
		    case 'c':
			get_2_str(argv[++i], s1, s2);
			if(*s1 !='\0') latR0=deg2rad(s1); else latR0= -1001.;
			if(*s2 !='\0') lonR0=deg2rad(s2); else lonR0= -1001.;
			zone_circle = 1;
			break;
		    case 'r':
			/* radius of a circle to be masked */
			zone_radius = deg2rad(argv[++i]);
			zone_circle = 1;
			/* if box_size selected for extr. areas from list */
			b_size = 2. * zone_radius * 180./M_PI;
			break;
		    case 'd':
			/* if box_size selected for extr. areas from list */
			b_size = str2deg(argv[++i]);
			break;
		    case 'o':
			/* select a zone to be masked */
			{ char s1[30], s2[30], s3[30], s4[30];
			    get_4_str(argv[++i], s1, s2, s3, s4);
			    if(*s1 !='\0') latR0=deg2rad(s1);
				else latR0= -1001.;
			    if(*s2 !='\0') lonR0=deg2rad(s2);
				else lonR0= -1001.;
			    if(*s3 !='\0') latR1=deg2rad(s3);
				else latR1= -1001.;
			    if(*s4 !='\0') lonR1=deg2rad(s4);
				else lonR1= -1001.;
			}
			break;
		    default:
			/* file with list of zones to be processed */
			strcpy(zone_file, argv[++i]);
			break;
		    }
		    break;
	    default:
		    break;
	    }
	    break;
	case '+':
	    switch(argv[i][1]) {
	    case 'C':
		    if (argv[i][2] == 'i') {
			cut_values = -1;
			get_val_str(argv[++i], &cut_low, &cut_up);
		    } else if (argv[i][2] == 'm') {
			cut_values = 11;
			cut_low = atof(argv[++i]);
		    } else {
			cut_values =  1;
			get_val_str(argv[++i], &cut_low, &cut_up);
		    }
		    break;
	    case 'h':
		    if (strncmp("h2", argv[i]+1, 2) == 0)
			usage2();
		    else
			usage();
		    break;
	    default:
		    usage();
		    break;
	    }
	    break;
	default:
		strcpy(map_file, argv[i]);
		break;
	}
    }

    if (flag_cr && zone_circle) {
	latR0 = lat0;
	lonR0 = lon0;
    }
    if ( lat0 > -1000 && lat1 > -1000 && eq_input == 1) {
	/* extract rectangular in equatorial coordinates */
	double p = 180/M_PI;
	double ra0  = lat0*p, ra1  = lat1*p,    /* in hours */
	       dec0 = lon0*p, dec1 = lon1*p;    /* in degs  */
	equat2gal(ra0, dec0, &lat0, &lon0);
	equat2gal(ra1, dec1, &lat1, &lon1);
	lat0 /= p; lat1 /= p;
	lon0 /= p; lon1 /= p;
	if (lat0 > lat1) { ra0=lat0; lat0=lat1; lat1=ra0; }
	if (lon0 > lon1) { ra0=lon0; lon0=lat1; lon1=ra0; }
    }
    return;
}

void usage()
{
PR(ER, "usage: %s file  -o output_file [-C[{im}] cut_low,cut_up]\n", myname);
PR(ER, "[-sq lat1,lon1,lat2,lon2] [-sf[{g,e}] file_to_select_by_theta,phi(lat,lon)]\n");
PR(ER, "[-{rm,keep} lat1,lon1,lat2,lon2(remove/keep sky zone)] [-d dummy]\n");
PR(ER, "[-pr(print mapdata in ASCII)] [-circ] [-ra radius_in_degs] [-zc lat,lon]\n");
PR(ER, "[-cr[{p,q}] lat,lon(get value in coors.)]  [-src(define src.pars)]\n");
PR(ER, "[-fc factor] [-zf file_with_zones] [-{R,K}] [-dd ddec] [-rs seed] [-rn N]\n");
PR(ER, "[-cs lat,lon,dlat,n1(cross-section by LAT),dlon,n2]\n");
PR(ER, "[-eq -cs ra,dec,dra,n1(cross-section by RA),ddec,n2]\n");
PR(ER, "[-eq(inp.coors)] [-{q,u,t} or '-ex {q,u,t,a}' (extract pol-maps)]\n");
PR(ER, "[-{lat,lon}m[e] keep:<l[i]] [-{lat,lon}l[e] keep:>l[i]]\n");
PR(ER, "[-mh hp_mask_file] [-mg gl_mask_file] [-stack[q] file_coor] [-stat(for area)\n");
PR(ER, "[-li  a1,a2/b1,b2/c1,c2/... (set lim-s to)]\n");
PR(ER, "[-rinl longitude] [-rinb latitude] [-slat] [-slon] [-dlat dLAT] [-dlon dLON]\n");
PR(ER, "[-cc lat,lon,rad,w(ring mask in degs)] [-nf NXxNY(image size factors)]\n");
PR(ER, "[-ccr lat1,lon1,lat2,lon2,w(rectang.mask in degs)] [-dir out_directory]\n");
PR(ER, "[-fzq[1] list_source_equat.][-fzq[1] list_src_gal.] [-zd box_size_d]\n");
PR(ER, "[-{double,float}] [-if F-file]\n");
exit(2);
}


void usage2()
{
PR(ST, "usage: %s file -o output_file [FLAGS]\n", myname);
PR(ST, "-----------------------------------------------------------------\n");
PR(ST, "-cc lat,lon,rad,w  - sets ring mask for sky zone to be kept:\n");
PR(ST, "      in center (lat,lon) of radius 'rad' and width 'w' in degs.\n");
PR(ST, "-ccr lat1,lon1,lat2,lon2,w - set rectang. mask for sky zone to be kept\n");
PR(ST, "      in rectangle (lat1,lon1,lat2,lon2) of width 'w' in degs.\n");
PR(ST, "-circ - selects circle to remove or remain (see '-R','-K','-zc','-zr')\n");
PR(ST, "-C  v_min,v_max - truncates values of map in range [v_min,v_max]\n");
PR(ST, "-Ci v_min,v_max - truncates values of map out range [v_min,v_max]\n");
PR(ST, "-Cm v_min - truncates values: if (v>v_min) v=0; else v=v_min\n");
PR(ST, "-cr lat,lon - displays the value at position (lat,lon),degs\n");
PR(ST, "-crp theta,phi - displays the value at position (theta,phi),degs\n");
PR(ST, "-crq ra,dec - displays the value at position (ra2000,dec2000)\n");
PR(ST, "-cs lat,lon,dlat,n1,dlon,n2 - (see '-eq')\n");
PR(ST, "    (GLat,Glon) - start point, dlat - step by glat, n - naxis1,\n");
PR(ST, "    dlon - step by glon, n2 - naxis2, to be recorded in F-file\n");
PR(ST, "-d dummy - sets default value 'dummy' into masked pixels (M - GLESPMASK)\n");
PR(ST, "-dd ddec - shift for section by declination in degs (see flag '-if')\n");
PR(ST, "-dir output_directory\n");
PR(ST, "-dlat dLAT - sets step dLAT for rectangular to extract (see '-dla','-sq')\n");
PR(ST, "-dlon dLON - sets step dLON for rectangular to extract (see '-dlo','-sq')\n");
PR(ST, "-dra dRA  - sets step dRA for rectangular to extract (see '-dlo','-sqq')\n");
PR(ST, "-dde dDec - sets step dDec for rectangular to extract (see '-dlo','-sqq')\n");
PR(ST, "-double - sets output map data format as double [default] (see '-float')\n");
PR(ST, "-ex {t,q,u} - extracts temperature, Q- or U-pol. map from 3-map \n");
PR(ST, "-eq -cs ra,dec,dra,n1,ddec,n2 - cross-sections of a map: (see '-cs'\n");
PR(ST, "     (RA,Dec) - start point, dra - step by RA, n - naxis1,\n");
PR(ST, "     ddec - step by Dec, n2 - naxis2, to be recorded in F-file\n");
PR(ST, "-fc factor - multiplication factor\n");
PR(ST, "-fcrq - same '-sfe' - get values at eq. coordinates stored in file \n");
PR(ST, "-fcr  - same '-sf'  - get values at gal.coordinates stored in file \n");
PR(ST, "-fzg file - gets areas in FITS-files from a source list in gal.coors\n");
PR(ST, "            file format:  lat_DDD.DDDd lon_DDD.DDDd [DD.DDd]\n");
PR(ST, "-fzg1 file - gets areas in FITS-files from a source list in gal.coors\n");
PR(ST, "            file format:  NAME lat_DDD.DDDd lon_DDD.DDDd [DD.DDd]\n");
PR(ST, "-fzq file - gets areas in FITS-files from a source list in eq.coors\n");
PR(ST, "            [see also '-fzq','-zd']\n");
PR(ST, "            file format:  hh:mm:ss.ss ddd:mm:ss.ss [DD.DDd]\n");
PR(ST, "-fzq1 file - gets areas in FITS-files from a source list in eq.coors\n");
PR(ST, "            [see also '-fzq','-zd']\n");
PR(ST, "            file format:  NAME hh:mm:ss.ss ddd:mm:ss.ss [DD.DDd]\n");
PR(ST, "-float - sets output map data format as float (see '-double'\n");
PR(ST, "-if F-file - extracts 1-pixel width ring corresponding to F-file\n");
PR(ST, "-keep lat1,lon1,lat2,lon2 - sets mask for sky zone to be kept\n");
PR(ST, "-K - flag to keeps selected zone on the figure (see '-zo','-zc','-zf')\n");
PR(ST, "-latl[e] lat - keeps value v >='lat' or, if 'e', v>'lat'\n");
PR(ST, "-latm[e] lat - keeps value v <='lat' or, if 'e', v<'lat'\n");
PR(ST, "-li  a1,a2/b1,b2/c1,c2/... - sets output limit in ranges\n");
PR(ST,"      [a1,a2],[b1,b2],[c1,c2],...\n");
PR(ST, "-lonl[e] lon - keeps value v >='lon' or, if 'e', v>'lon'\n");
PR(ST, "-lonm[e] lon - keeps value v <='lon' or, if 'e', v<'lon'\n");
PR(ST, "-mg gl_mask_file  - sets mask=dummy from glesp-file 'gl_mask_file'\n");
PR(ST, "-Mh hp_mask_file  - sets mask=GLESPMASK from hpix-file 'hp_mask_file'\n");
PR(ST, "-mg gl_mask_file  - sets mask=dummy from glesp-file 'gl_mask_file'\n");
PR(ST, "-Mg gl_mask_file  - sets mask=GLESPMASK from glesp-file 'gl_mask_file'\n");
PR(ST, "-nf NXxNY - sets of extrated image size multiplied by factors (see '-sa')\n");
PR(ST, "-nf N - same to '-nf NxN (see '-nf','-sa','-saq')\n");
PR(ST, "-pr - prints map data in ASCII format\n");
PR(ST, "-q - extracts Q-polarisation map from 3-map file (see '-u','-t','-ex')\n");
PR(ST, "-ra radius_in_degs - selects zone radius (see '-K', '-R', '-zc')\n");
PR(ST, "-remove lat1,lon1,lat2,lon2 - sets mask for sky zone to be removed\n");
PR(ST, "-rinb lat - extracts 1-pixel width ring at latitude=lat\n");
PR(ST, "-rinl lon - extracts 1-pixel width ring at longitude=lon\n");
PR(ST, "-R - flag to remove selected zone on the figure (see '-zo','-zc',-zf')\n");
PR(ST, "-rs seed - random selection from a sample (see also '-rn','-zf')\n");
PR(ST, "-rn N - random selection on N lines from a sample (see also '-rs','-zf')\n");
PR(ST, "-sf file  - print values of 'file' data by select by theta & phi\n");
PR(ST, "-sfe file - print values of 'file' data by select by ra & dec\n");
PR(ST, "-sfg file - print values of 'file' data by select by g.lat & g.lon\n");
PR(ST, "-slat - sets step dLON=dLAT for rectangular area to extract (see '-dla')\n");
PR(ST, "-slon - sets step dLAT=dLON for rectangular area to extract (see '-dlo')\n");
PR(ST, "-sa  lat0,lon0,size_deg - selects galactic   rectangular area to extract\n");
PR(ST, "-saq  ra0,dec0,size_deg - selects equatorial rectangular area to extract\n");
PR(ST, "-saq- ra0,dec0,size_deg - selects equatorial rect. area to extract (dRA!=dDec)\n");
PR(ST, "-sq lat1,lon1,lat2,lon2 - selects galactic   rectangular area to extract\n");
PR(ST, "-sqq  ra1,dec1,ra2,dec2 - selects equatorial rectangular area to extract\n");
PR(ST, "-sqq+ ra1,dec1,ra2,dec2 - selects equatorial rect. area to extract (dRA=dDec)\n");
PR(ST, "-src - defines source parameters if it exists in area selected by '-sq'\n");
PR(ST, "-stack[q] src_list - makes stacking for selected zones from file 'src_list'\n");
PR(ST, "-stat - calculates & prints simple statistics for area (see '-stack')\n");
PR(ST, "     file format: GLat GLon box_size(degs) [or RA Dec box_size(degs)]\n");
PR(ST, "-t - extracts temperature map from 3-map file (see '-q','-u','-ex')\n");
PR(ST, "-u - extracts U-polarisation map from 3-map file (see '-q','-t','-ex')\n");
PR(ST, "-v - prints version of '%s'\n", myname);
PR(ST, "-zc lat,lon - selects zone center (see '-K', '-R', '-zr')\n");
PR(ST, "-zf file_with_zones - reads zones from file for masking of format:\n");
PR(ST, "    lat1,lon1,lat1,lon1 for rectangle or lat,lon,radius for circle\n");
//PR(ST, "-zfg file_with_zones - reads zones in Gal.coor-sfrom file for masking\n");
PR(ST, "-zo lat1,lon1,lat2,lon2 - selects zone by coordinates: default to remove\n");
PR(ST, "                          (see '-K','-R', '-remove', '-keep')\n");
PR(ST, "-zr radius_in_degs - selects zone radius (see '-K', '-R', '-zc;)\n");
PR(ST, "-zd box_size_in_degs - selects zone box side (see '-fzg','-fzq';)\n");

exit(0);
}


