#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"

#define PR  fprintf
#define ER  stderr

//struct _zone_ {
//     double latR0, lonR0;
//     double latR1, lonR1;
//     double radius;
//     double dr;
//     int type;
//};

void map_zone(zz, nz, function_type, f, x, nx, np, nph, dummy_value)
struct _zone_ *zz;  /* zone coordinates          */
int nz;             /* number of zones           */
int function_type;  /* remain=0 or remove=1      */
double *f, *x;
int nx, np, *nph;
double dummy_value;
{
    register int iz, i, j;
    char *m = NULL;
    double lat0, lon0, lat1, lon1, radius;
    double theta0, theta1, phi0, phi1;
    double pi2 = 2. * M_PI;
    int nx0, nx1, np0=0, np1=0;
    int zone_type; /* circle=0 or rectangular=1 */
    double dpx = pi2/np;

    /***********************************************/
    /* if save several zones,  we should set masks */
    /***********************************************/
    /* Basic formula for distance between points is
     *     d = arccos(cos(lon2-lon1)cos(lat1)cos(lat2)+sin(lat1)sin(lat2))
     * for polar coordinates"
     *     d = acos(cos(dphi)sin(th1)sin(th2)+cos(th1)*cos(th2))
     */

    if ( function_type == 0 ) { /* remain_flag */
	 m = (char *)calloc(nx*np, sizeof *m);
	 if (m == NULL) {
	     PR(ER, "%s: no memory for masking removed zones.", myname);
	     exit(1);
	 }
	 for(iz=0; iz<nz; iz++) {
	     double cth0, sth0, v;
	     zone_type = zz[iz].type; /* circle=0 or rectangular=1 */
	     if (zone_type == 0) {  /* circle */
	       int nx_c, np_c;   /* center of circle */
	       lat0  = zz[iz].latR0;
	       lon0  = zz[iz].lonR0;
	       radius = zz[iz].radius;
	       if (radius < -1.e+6)  radius = fabs(radius + 1.e+6) * dpx;

	       /* extract circle */
	       gal2pol(lat0, lon0, &theta0, &phi0);
	       cth0 = cos(theta0);
	       sth0 = sqrt(1-cth0*cth0);
	       nx_c = ind_x(nx, x, cth0);
	       np_c = phi0 / pi2 * (double)nph[nx_c];

	       nx0 = ind_x(nx, x, cos(theta0-radius));
	       nx1 = ind_x(nx, x, cos(theta0+radius));
	       if ( nx0 > nx1 ) { j=nx0; nx0=nx1; nx1=j; }

	       if (theta0-radius < 0. || theta0+radius > M_PI ) {
		 /* process pixels near pole */
		 for(i=0; i<nx; i++) {
		   int ii = i * np;
		   double cth = x[i], sth = sqrt(1.-cth*cth);
		   double dphi = pi2 / (double)nph[i];
		   double dp;   /* dt = theta0 - acos(cth), */

		   if (!((theta0+radius > M_PI &&
		     ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))) ||
		     (theta0-radius < 0.   &&
		     ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))))) {
			for(j=0; j<nph[i]; j++) {
			    dp = phi0 - j*dphi;
			    v = acos(cos(dp)*sth0*sth + cth0*cth);
			    if (v <= radius || v < dpx)  {
				    m[ii+j] = 1;
				    continue;
			    }
			}
		   }
		 }
	       } else {
		  for(i=0; i<nx; i++) {
		    int ii = i * np;
		    double cth = x[i],  sth = sqrt(1-cth*cth);
		    double dphi = pi2 / (double)nph[i];
		    double dp; /* dt = theta0 - acos(cth), */
		    /* make checking interval a little larger */
		    double rl = radius/sth0/sth;

		    if (i >= nx0 && i <= nx1) {
			np0 = (phi0-rl) / dphi;
			np1 = (phi0+rl) / dphi;
			if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

			if (np0 < 0) {
			     int npR = nph[i] + np0;
			     for(j=0; j<nph[i]; j++) {
			       if (npR < j || j < np1) {
				 dp = (2*M_PI+phi0) - j*dphi;
				 v = acos(cos(dp)*sth0*sth + cth0*cth);
				 if (v <= radius || v < dpx) {
					 m[ii+j] = 1;
					 continue;
				 }
			       }
			     }
			}
			if (np1 > nph[i]) {
			    int npR = np1 - nph[i];
			    for(j=0; j<nph[i]; j++) {
			      if (npR < j || j < np0) {
				dp = -phi0 + j*dphi + 2*M_PI;
				v = acos(cos(dp)*sth0*sth + cth0*cth);
				if (v <= radius || v < dpx) {
					m[ii+j] = 1;
					continue;
				}
			      }
			    }
			}
			for(j=0; j<nph[i]; j++) {
			    if (j >= np0 && j <= np1) {
				  dp = phi0 - j*dphi;
				  v = acos(cos(dp)*sth0*sth + cth0*cth);
				  if (v <= radius || v < dpx) {
					  m[ii+j] = 1;
					  continue;
				  }
			    }
			}
		    }
		  }
	       }
	     } else {
	       lat0  = zz[iz].latR0;
	       lon0  = zz[iz].lonR0;
	       lat1  = zz[iz].latR1;
	       lon1  = zz[iz].lonR1;

	       /* set mask for rectangular */
	       gal2pol(lat0, lon0, &theta0, &phi0);
	       gal2pol(lat1, lon1, &theta1, &phi1);
	       nx0 = ind_x(nx, x, cos(theta0));
	       nx1 = ind_x(nx, x, cos(theta1));

	       if (nx0 > nx1)  {i=nx0; nx0=nx1; nx1=i;}

	       for(i=0; i<nx; i++) {
		    int ii = i * np;
		    if (i >= nx0 && i < nx1) {
			np0 = phi0 / pi2 * (double)nph[i];
			np1 = phi1 / pi2 * (double)nph[i];
			if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }
			/* process pixels near center */
			if (np0 < 0) {
			      int npR = nph[i] + np0;
			      for(j=0; j<nph[i]; j++) {
				    if (j <= np1 || j >= npR) {
					    m[ii+j] = 1;
					    continue;
				    }
			      }
			}
			if (np1 >= nph[i]) {
			    int npR = np1 - nph[i];
			    for(j=0; j<nph[i]; j++) {
				    if (j >= np0 || j <= npR) {
					    m[ii+j] = 1;
					    continue;
				    }
			    }
			}
			for(j=0; j<nph[i]; j++) {
			    if (j >= np0 && j <= np1) {
				m[ii+j] = 1;
				continue;
			    }
			}
		    }
	       }
	     }
	 }
    }

    /******************/
    /* process zones  */
    /******************/
    /* Basic formula for distance between points is
     *     d = arccos(cos(lon2-lon1)cos(lat1)cos(lat2)+sin(lat1)sin(lat2))
     * for polar coordinates"
     *     d = acos(cos(dphi)sin(th1)sin(th2)+cos(th1)*cos(th2))
     */

    for(iz=0; iz<nz; iz++) {
     zone_type = zz[iz].type; /* circle=0 or rectangular=1 */
     if (zone_type == 0) { /* circle */
       int nx_c, np_c;   /* center of circle */
       double cth0, sth0, v;

       lat0  = zz[iz].latR0;
       lon0  = zz[iz].lonR0;
       radius = zz[iz].radius;
       if (radius < -1.e+6)  radius = fabs(radius + 1.e+6) * dpx;

       /* extract circle */
       gal2pol(lat0, lon0, &theta0, &phi0);
       cth0 = cos(theta0);
       sth0 = sqrt(1.-cth0*cth0);
       nx_c = ind_x(nx, x, cth0);
       np_c = phi0 / pi2 * (double)nph[nx_c];

       nx0 = ind_x(nx, x, cos(theta0-radius));
       nx1 = ind_x(nx, x, cos(theta0+radius));
       if ( nx0 > nx1 ) { j=nx0; nx0=nx1; nx1=j; }

       if (theta0-radius < 0. || theta0+radius > M_PI ) {
	 /* process pixels near pole */
	 for(i=0; i<nx; i++) {
	    int ii = i * np;
	    double cth = x[i], sth = sqrt(1.-cth*cth);
	    double dphi = pi2 / (double)nph[i];
	    double dp;  /* dt = theta0 - acos(cth), */

	    if ( function_type == 1) {  /* remove_flag */
		if (theta0+radius > M_PI) {
		    if ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))
			  continue;
		} else {
		    if ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))
			  continue;
		}

		for(j=0; j<nph[i]; j++) {
		    dp = phi0 - j*dphi;
		    v = acos(cos(dp)*sth0*sth + cth0*cth);
		    if (v > radius && v > dpx)
			    continue;
		    f[ii+j] = dummy_value;
		}
	    }

	    if ( function_type == 0) {  /* remain_flag */
		if (!((theta0+radius > M_PI &&
		    ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))) ||
		    (theta0-radius < 0.   &&
		    ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))))) {
			for(j=0; j<nph[i]; j++) {
			    dp = phi0 - j*dphi;
			    v = acos(cos(dp)*sth0*sth + cth0*cth);
			    if (v <= radius || v < dpx)
				    continue;
			    if (m[ii+j] != 1)  f[ii+j] = dummy_value;
			}
		} else {
		     for(j=0; j<nph[i]; j++)
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		}
	    }
	 }
       } else {
	 for(i=0; i<nx; i++) {
	      int ii = i * np;
	      double cth = x[i], sth = sqrt(1-cth*cth);
	      double dphi = pi2 / (double)nph[i];
	      double dp;  /* dt = theta0 - acos(cth),  */
	      /* make checking interval a little larger */
	      double rl = radius/sth0/sth;

	      if ( function_type == 1 ) {  /* remove_flag */
		     if (i < nx0 || i > nx1)
			     continue;
		     np0 = (phi0-rl) / dphi;
		     np1 = (phi0+rl) / dphi;
		     if ( np1 < np0 ) { j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			 int npR = nph[i] + np0;
			 for(j=0; j<nph[i]; j++) {
			     if (np1<j  && j < npR)
				     continue;
			     dp = (2*M_PI+phi0) - j*dphi;
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (v > radius && v > dpx)
				     continue;
			     f[ii+j] = dummy_value;
			 }
		     }
		     if (np1 >= nph[i]) {
			 int npR = np1 - nph[i];
			 for(j=0; j<nph[i]; j++) {
			     if (npR < j && j < np0)
				     continue;
			     dp = -phi0 + j*dphi + 2*M_PI;
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (v > radius && v > dpx) {
				     continue;
			     }
			     f[ii+j] =dummy_value;
			 }
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if (j < np0 || j > np1)
				 continue;

			 dp = phi0 - j*dphi;
			 v = acos(cos(dp)*sth0*sth + cth0*cth);
			 if (v > radius && v > dpx)
				 continue;
			 f[ii+j] = dummy_value;
		     }
	      }
	      if ( function_type == 0 ) { /* remain_flag */
		 if (i >= nx0 && i <= nx1) {
		     np0 = (phi0-rl) / dphi;
		     np1 = (phi0+rl) / dphi;
		     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			 int npR = nph[i] + np0;
			 for(j=0; j<nph[i]; j++) {
			   if (npR < j || j < np1) {
			     dp = 2*M_PI-phi0 + j*dphi;
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (v <= radius || v < dpx)  {
				     continue;
			     }
			   }
			   if (m[ii+j] !=1) f[ii+j] =dummy_value;
			 }
		     }
		     if (np1 >= nph[i]) {
			 int npR = np1 - nph[i];
			 for(j=0; j<nph[i]; j++) {
			   if (npR < j || j < np0) {
			     dp = phi0 + j*dphi - 2*M_PI;
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (v <= radius || v < dpx)  {
				     continue;
			     }
			   }
			   if (m[ii+j] !=1) f[ii+j] =dummy_value;
			 }
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if (j >= np0 && j <= np1) {
			       dp = phi0 - j*dphi;
			       v = acos(cos(dp)*sth0*sth + cth0*cth);
			       if (v <= radius || v < dpx)  {
				       continue;
			       }
			 }
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		     }
		 } else {
		     for(j=0; j<nph[i]; j++)
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		 }
	      }
	 }
       }
     } else {
	 lat0  = zz[iz].latR0;
	 lon0  = zz[iz].lonR0;
	 lat1  = zz[iz].latR1;
	 lon1  = zz[iz].lonR1;

	 /* extract rectangular */
	 gal2pol(lat0, lon0, &theta0, &phi0);
	 gal2pol(lat1, lon1, &theta1, &phi1);

	 nx0 = ind_x(nx, x, cos(theta0));
	 nx1 = ind_x(nx, x, cos(theta1));
	 if (nx0 > nx1)  {i=nx0; nx0=nx1; nx1=i;}

	 for(i=0; i<nx; i++) {
	      int ii = i * np;


	      if ( function_type == 1 ) {  /* remove_flag */
		     if (i < nx0 || i >= nx1)
			    continue;
		     np0 = phi0 / pi2 * (double)nph[i];
		     np1 = phi1 / pi2 * (double)nph[i];

		     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			 int npR = nph[i] + np0;
			 for(j=0; j<nph[i]; j++) {
				 if (j > np1  && j < npR)
					 continue;
				 f[ii+j] = dummy_value;
			 }
		     }
		     if (np1 >= nph[i]) {
			 int npR = np1 - nph[i];
			 for(j=0; j<nph[i]; j++) {
				 if (j < np0 && j > npR)
					 continue;
				 f[ii+j] = dummy_value;
			 }
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if (j < np0 || j > np1)
				 continue;
			 f[ii+j] = dummy_value;
		     }
	      }
	      if ( function_type == 0 ) { /* remain_flag */
		 if (i >= nx0 && i < nx1) {
		     np0 = phi0 / pi2 * (double)nph[i];
		     np1 = phi1 / pi2 * (double)nph[i];
		     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			   int npR = nph[i] + np0;
			   for(j=0; j<nph[i]; j++) {
				 if (j <= np1 || j >= npR)
				    continue;
				 if (m[ii+j] !=1) f[ii+j] =dummy_value;
			   }
		     }
		     if (np1 >= nph[i]) {
			 int npR = np1 - nph[i];
			 for(j=0; j<nph[i]; j++) {
				 if (j >= np0 || j <= npR)
				    continue;
				 if (m[ii+j] !=1) f[ii+j] =dummy_value;
			 }
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if (j >= np0 && j <= np1)
				 continue;
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		     }
		 } else {
		     for(j=0; j<nph[i]; j++)
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		 }
	      }
	 }
       }
    }

    if (m != NULL) free((void *)m);
    return;
}


/*
 * struct _zone_ {
 *      double latR0, lonR0;
 *      double latR1, lonR1;
 *      double radius;
 *      double dr;
 *      int type;
 * };
*/
void map_ring_zone(zz, nz, fun_type, f, x, nx, np, nph, dummy_value)
struct _zone_ *zz;  /* zone coordinates      */
int nz;             /* number of zones       */
int fun_type;       /* remain=0 or remove=1  */
double *f, *x;
int nx, np, *nph;
double dummy_value;
{
    register int iz, i, j;
    char *m = NULL;
    double lat0, lon0, lat1, lon1, radius, dr;
    double theta0, theta1, phi0, phi1;
    double pi2 = 2. * M_PI;
    int nx0, nx1, np0=0, np1=0,
	nx0_in, nx1_in, np0_in=0, np1_in=0;
    int zone_type; /* circle=0 or rectangular=1 */
    double radius1, radius2, dpx = 2*M_PI/np;
//    int ztype;
    int function_type = fun_type;

    /***********************************************/
    /* if save several zones,  we should set masks */
    /***********************************************/
    /* Basic formula for distance between points is
     *     d = arccos(cos(lon2-lon1)cos(lat1)cos(lat2)+sin(lat1)sin(lat2))
     * for polar coordinates"
     *     d = acos(cos(dphi)sin(th1)sin(th2)+cos(th1)*cos(th2))
     */
    if ( function_type == 0 ) { /* remain_flag */
	 m = (char *)calloc(nx*np, sizeof *m);
	 if (m == NULL) {
	     PR(ER, "%s: no memory for masking removed zones.", myname);
	     exit(1);
	 }
	 for(iz=0; iz<nz; iz++) {
	     double cth0, sth0, v;
	     zone_type = zz[iz].type; /* circle=0 or rectangular=1 */
	     dr = zz[iz].dr;

	     if (zone_type == 0) {  /* circle */
	       int nx_c, np_c;   /* center of circle */

	       lat0 = zz[iz].latR0;
	       lon0 = zz[iz].lonR0;
	       radius = zz[iz].radius;
	       if (radius < -1.e+6)  radius = fabs(radius + 1.e+6) * dpx;
	       radius1 = radius;
	       radius2 = radius1 + dr;

	       /* extract circle */
	       gal2pol(lat0, lon0, &theta0, &phi0);
	       cth0 = cos(theta0);
	       sth0 = sqrt(1-cth0*cth0);
	       nx_c = ind_x(nx, x, cth0);
	       np_c = phi0 / pi2 * (double)nph[nx_c];

	       nx0 = ind_x(nx, x, cos(theta0-radius2));
	       nx1 = ind_x(nx, x, cos(theta0+radius2));
	       if ( nx0 > nx1 ) { j=nx0; nx0=nx1; nx1=j; }

	       if (theta0-radius2 < 0. || theta0+radius2 > M_PI ) {
		 /* process pixels near pole */
		 for(i=0; i<nx; i++) {
		   int ii = i * np;
		   double cth = x[i], sth = sqrt(1.-cth*cth);
		   double dphi = pi2 / (double)nph[i];
		   double dp;   /* dt = theta0 - acos(cth), */

		   if (!((theta0+radius2 > M_PI &&
		     ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))) ||
		     (theta0-radius2 < 0.   &&
		     ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))))) {
			for(j=0; j<nph[i]; j++) {
			    dp = phi0 - j*dphi;
			    v = acos(cos(dp)*sth0*sth + cth0*cth);
			    if (radius1 <= v && v <= radius2)  {
				    m[ii+j] = 1;
				    continue;
			    }
			}
		   }
		 }
	       } else {
		  for(i=0; i<nx; i++) {
		    int ii = i * np;
		    double cth = x[i],  sth = sqrt(1-cth*cth);
		    double dphi = pi2 / (double)nph[i];
		    double dp; /* dt = theta0 - acos(cth), */
		    /* make checking interval a little larger */
		    double rl = radius2/sth0/sth;

		    if (i >= nx0 && i <= nx1) {
			np0 = (phi0-rl) / dphi;
			np1 = (phi0+rl) / dphi;
			if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

			if (np0 < 0) {
			     int npR = nph[i] + np0;
			     for(j=0; j<nph[i]; j++) {
			       if (npR < j || j < np1) {
				 dp = (2*M_PI+phi0) - j*dphi;
				 v = acos(cos(dp)*sth0*sth + cth0*cth);
				 if (radius1 <= v && v <= radius2) {
					 m[ii+j] = 1;
					 continue;
				 }
			       }
			     }
			}
			if (np1 > nph[i]) {
			    int npR = np1 - nph[i];
			    for(j=0; j<nph[i]; j++) {
			      if (npR < j || j < np0) {
				dp = -phi0 + j*dphi + 2*M_PI;
				v = acos(cos(dp)*sth0*sth + cth0*cth);
				if (radius1 <= v && v <= radius2) {
					m[ii+j] = 1;
					continue;
				}
			      }
			    }
			}
			for(j=0; j<nph[i]; j++) {
			    if (j >= np0 && j <= np1) {
				  dp = phi0 - j*dphi;
				  v = acos(cos(dp)*sth0*sth + cth0*cth);
				  if (radius1 <= v && v <= radius2) {
					  m[ii+j] = 1;
					  continue;
				  }
			    }
			}
		    }
		  }
	       }
	     } else {
	       /* rectangle */
	       lat0 = zz[iz].latR0;
	       lon0 = zz[iz].lonR0;
	       lat1 = zz[iz].latR1;
	       lon1 = zz[iz].lonR1;

	       /* set mask for rectangular */
	       gal2pol(lat0, lon0, &theta0, &phi0);
	       gal2pol(lat1, lon1, &theta1, &phi1);
	       nx0 = ind_x(nx, x, cos(theta0));
	       nx1 = ind_x(nx, x, cos(theta1));
	       nx0_in = ind_x(nx, x, cos(theta0-dr));
	       nx1_in = ind_x(nx, x, cos(theta1+dr));
	       if (nx0 > nx1)  {i=nx0; nx0=nx1; nx1=i;}

	       if (phi0 < 0. && phi1 < 0.) {
		      phi0 += pi2;
		      phi1 += pi2;
	       }
	       if (phi0 > pi2 && phi1 > pi2) {
		      phi0 -= pi2;
		      phi1 -= pi2;
	       }

	       for(i=0; i<nx; i++) {
		 int ii = i * np;
		 if (i >= nx0 && i < nx1) {
		   np0 = phi0 / pi2 * (double)nph[i];
		   np1 = phi1 / pi2 * (double)nph[i];
		   np0_in = (phi0+dr) / pi2 * (double)nph[i];
		   np1_in = (phi1-dr) / pi2 * (double)nph[i];
		   if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }
		   /* process pixels near center */
		   if (np0 < 0) {
		     int npR = nph[i] + np0;
		     int npR0_in = np0_in<0 ? nph[i]+np0_in : np0_in;
		     int npR1_in = np1_in<0 ? nph[i]+np1_in : np1_in;

		     for(j=0; j<nph[i]; j++) {

			if( (npR <= j || j <= np1) &&
			(
			  /*  |***|  |  |***|  */
			((np0_in < 0 && np1_in >= 0) &&
			   ((npR<=j&&j<=npR0_in)||(npR1_in<=j&&j<=np1)))||
			  /*  |***|    |*|**|  */
			((np0_in >= 0 && np1_in >= 0) &&
			   ((npR<=j&&j<=npR0_in)||(npR1_in<=j&&j<=np1)))||
			  /*  |*|**|    |***|  */
			((np0_in < 0 && np1_in <= 0) &&
			    ((0<=j&&j<=np1)|| j >= npR1_in || j <= npR0_in))||
			(i <= nx0_in || nx1_in <= i)
			))  {

			      m[ii+j] = 1;
			      continue;
			}
		     }
		   }
		   if (np1 >= nph[i]) {
		       int npR = np1 - nph[i];
		       int npR0_in = np0_in > nph[i] ? np0_in-nph[i]:np0_in;
		       int npR1_in = np1_in > nph[i] ? np1_in-nph[i]:np1_in;
		       for(j=0; j<nph[i]; j++) {
			 if( (npR <= j || j <= np1)
			   &&
			   (
			     /*  |***|  |  |***|  */
			   ((np0_in < 0 && np1_in >= 0) &&
			    ((npR<=j&&j<=npR0_in)||(npR1_in<=j&&j<=np1)))||
			     /*  |***|    |*|**|  */
			   ((np0_in >= 0 && np1_in >= 0) &&
			     (npR<=j || j<=npR0_in ||(npR1_in<=j&&j<=np1)))||
			     /*  |*|**|    |***|  */
			   ((np0_in < 0 && np1_in <= 0) &&
			     ((0<=j&&j<=np1)|| j>= npR1_in || j<= npR0_in))||
			   (i <= nx0_in || nx1_in <= i)
			   ))
			   {
				m[ii+j] = 1;
				continue;
			   }
		       }
		   }
		   for(j=0; j<nph[i]; j++) {
			 if (
			    ((np0 <=j&&j<= np0_in) ||
				   (np1_in <=j&&j<= np1)) ||
			     ((np0<=j&&j<=np1)&&
				   (i<= nx0_in||nx1_in <= i))
			 ) {
			   m[ii+j] = 1;
			   continue;
		       }
		   }
		 }
	       }
	     }
	 }
    }

    /******************/
    /* process zones  */
    /******************/
    /* Basic formula for distance between points is
     *     d = arccos(cos(lon2-lon1)cos(lat1)cos(lat2)+sin(lat1)sin(lat2))
     * for polar coordinates
     *     d = acos(cos(dphi)sin(th1)sin(th2)+cos(th1)*cos(th2))
     */

    for(iz=0; iz<nz; iz++) {
     zone_type = zz[iz].type; /* circle=0 or rectangular=1 */
     dr = zz[iz].dr;

     if (zone_type == 0) { /* circle */
       int nx_c, np_c;   /* center of circle */
       double cth0, sth0, v;

       lat0 = zz[iz].latR0;
       lon0 = zz[iz].lonR0;
       radius = zz[iz].radius;
       if (radius < -1.e+6)  radius = fabs(radius + 1.e+6) * dpx;
       radius1 = radius;
       radius2 = radius1 + dr;

       /* extract circle */
       gal2pol(lat0, lon0, &theta0, &phi0);
       cth0 = cos(theta0);
       sth0 = sqrt(1.-cth0*cth0);
       nx_c = ind_x(nx, x, cth0);
       np_c = phi0 / pi2 * (double)nph[nx_c];

       nx0 = ind_x(nx, x, cos(theta0-radius2));
       nx1 = ind_x(nx, x, cos(theta0+radius2));
       if ( nx0 > nx1 ) { j=nx0; nx0=nx1; nx1=j; }

       if (theta0-radius2 < 0. || theta0+radius2 > M_PI ) {
	 /* process pixels near pole */
	 for(i=0; i<nx; i++) {
	    int ii = i * np;
	    double cth = x[i], sth = sqrt(1.-cth*cth);
	    double dphi = pi2 / (double)nph[i];
	    double dp;  /* dt = theta0 - acos(cth), */

	    if ( function_type == 1) {  /* remove_flag */
		if (theta0+radius2 > M_PI) {
		    if ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))
			  continue;
		} else {
		    if ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))
			  continue;
		}

		for(j=0; j<nph[i]; j++) {
		    dp = phi0 - j*dphi;
		    v = acos(cos(dp)*sth0*sth + cth0*cth);
		    if (v > radius2)
			    continue;
		    f[ii+j] = dummy_value;
		}
	    }

	    if ( function_type == 0) {  /* remain_flag */
		if (!((theta0+radius2 > M_PI &&
		    ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))) ||
		    (theta0-radius2 < 0.   &&
		    ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))))) {
			for(j=0; j<nph[i]; j++) {
			    dp = phi0 - j*dphi;
			    v = acos(cos(dp)*sth0*sth + cth0*cth);
			    if (v <= radius1)
				    continue;
			    if (m[ii+j] != 1)  f[ii+j] = dummy_value;
			}
		} else {
		     for(j=0; j<nph[i]; j++)
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		}
	    }
	 }
       } else {
	 for(i=0; i<nx; i++) {
	      int ii = i * np;
	      double cth = x[i], sth = sqrt(1-cth*cth);
	      double dphi = pi2 / (double)nph[i];
	      double dp;  /* dt = theta0 - acos(cth),  */
	      /* make checking interval a little larger */
	      double rl = radius/sth0/sth;

	      if ( function_type == 1 ) {  /* remove_flag */
		     if (i < nx0 || i > nx1)
			     continue;
		     np0 = (phi0-rl) / dphi;
		     np1 = (phi0+rl) / dphi;
		     if ( np1 < np0 ) { j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			 int npR = nph[i] + np0;
			 for(j=0; j<nph[i]; j++) {
			     if (np1<j  && j < npR)
				     continue;
			     dp = (2*M_PI+phi0) - j*dphi;
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (v > radius2 || v < radius1)
				     continue;
			     f[ii+j] = dummy_value;
			 }
		     }
		     if (np1 >= nph[i]) {
			 int npR = np1 - nph[i];
			 for(j=0; j<nph[i]; j++) {
			     if (npR < j && j < np0)
				     continue;
			     dp = -phi0 + j*dphi + 2*M_PI;
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (v > radius2 || v < radius1) {
				     continue;
			     }
			     f[ii+j] = dummy_value;
			 }
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if (j < np0 || j > np1)
				 continue;

			 dp = phi0 - j*dphi;
			 v = acos(cos(dp)*sth0*sth + cth0*cth);
			 if ((v > radius2 || v < radius1))
				 continue;
			 f[ii+j] = dummy_value;
		     }
	      }
	      if ( function_type == 0 ) { /* remain_flag */
		 if (i >= nx0 && i <= nx1) {
		     np0 = (phi0-rl) / dphi;
		     np1 = (phi0+rl) / dphi;
		     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			 int npR = nph[i] + np0;
			 for(j=0; j<nph[i]; j++) {
			   if (npR < j || j < np1) {
			     dp = 2*M_PI-phi0 + j*dphi;
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (radius1 <= v && v <= radius2) {
				     continue;
			     }
			   }
			   if (m[ii+j] !=1) f[ii+j] =dummy_value;
			 }
		     }
		     if (np1 >= nph[i]) {
			 int npR = np1 - nph[i];
			 for(j=0; j<nph[i]; j++) {
			   if (npR < j || j < np0) {
			     dp = phi0 + j*dphi - 2*M_PI;
			     v = acos(cos(dp)*sth0*sth + cth0*cth);
			     if (radius1 <= v && v <= radius2) {
				     continue;
			     }
			   }
			   if (m[ii+j] !=1) f[ii+j] =dummy_value;
			 }
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if (j >= np0 && j <= np1) {
			       dp = phi0 - j*dphi;
			       v = acos(cos(dp)*sth0*sth + cth0*cth);
			       if (radius1 <= v && v <= radius2) {
				       continue;
			       }
			 }
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		     }
		 } else {
		     for(j=0; j<nph[i]; j++)
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		 }
	      }
	 }
       }
     } else {
	 /* rectangular */
	 lat0 = zz[iz].latR0;
	 lon0 = zz[iz].lonR0;
	 lat1 = zz[iz].latR1;
	 lon1 = zz[iz].lonR1;

	 /* extract rectangular */
	 gal2pol(lat0, lon0, &theta0, &phi0);
	 gal2pol(lat1, lon1, &theta1, &phi1);

	 nx0 = ind_x(nx, x, cos(theta0));
	 nx1 = ind_x(nx, x, cos(theta1));
	 nx0_in = ind_x(nx, x, cos(theta0-dr));
	 nx1_in = ind_x(nx, x, cos(theta1+dr));
	 if (nx0 > nx1)  {i=nx0; nx0=nx1; nx1=i;}

	 if (phi0 < 0. && phi1 < 0.) {
		phi0 += pi2;
		phi1 += pi2;
	 }
	 if (phi0 > pi2 && phi1 > pi2) {
		phi0 -= pi2;
		phi1 -= pi2;
	 }

	 for(i=0; i<nx; i++) {
	      int ii = i * np;

	      if ( function_type == 1 ) {  /* remove_flag */
		     if (i < nx0 || i >= nx1)
			    continue;

		     np0 = phi0 / pi2 * (double)nph[i];
		     np1 = phi1 / pi2 * (double)nph[i];
		     np0_in = (phi0+dr) / pi2 * (double)nph[i];
		     np1_in = (phi1-dr) / pi2 * (double)nph[i];

		     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			 int npR = nph[i] + np0;
			 int npR0_in = np0_in < 0 ? np0_in+nph[i] : np0_in;
			 int npR1_in = np1_in < 0 ? np1_in+nph[i] : np1_in;
			 for(j=0; j<nph[i]; j++) {
			 if(!( (npR <= j || j <= np1)
			   &&
			   (
			     /*  |***|  |  |***|  */
			   ((np0_in < 0 && np1_in >= 0) &&
			    ((npR<=j&&j<=npR0_in)||(npR1_in<=j&&j<=np1)))||
			     /*  |***|    |*|**|  */
			   ((np0_in >= 0 && np1_in >= 0) &&
			     (npR<=j || j<=npR0_in ||(npR1_in<=j&&j<=np1)))||
			     /*  |*|**|    |***|  */
			   ((np0_in < 0 && np1_in <= 0) &&
			     ((0<=j&&j<=np1)|| j>= npR1_in || j<= npR0_in))||
			   (i <= nx0_in || nx1_in <= i)
			   )))
				continue;
			   f[ii+j] = dummy_value;
			 }
		     }
		     if (np1 >= nph[i]) {
		       int npR = np1 - nph[i];
		       int npR0_in = np0_in > nph[i] ? np0_in-nph[i]:np0_in;
		       int npR1_in = np1_in > nph[i] ? np1_in-nph[i]:np1_in;
		       for(j=0; j<nph[i]; j++) {
			  if(! ((np0 <= j || j <= npR)
			   &&
			   (
			     /*  |***|  |  |***|  */
			   ((np1_in > nph[i] && np0_in < nph[i]) &&
			    ((npR<=j&&j<=npR0_in)||(npR1_in<=j&&j<=np1)))||
			     /*  |*|**|    |***|  */
			   ((np0_in <= nph[i] && np1_in <= nph[i]) &&
			     (j<=npR || npR1_in<=j ||(np0<=j&&j<=npR0_in)))||
			     /*  |***|    |*|**|  */
			   ((np0_in >= nph[i] && np1_in >= nph[i]) &&
			     ((np0<=j&&j<=nph[i])|| j<=npR0_in||j>=npR1_in))||
			   (i <= nx0_in || nx1_in <= i)
			   )))
				continue;
			   f[ii+j] = dummy_value;
		       }
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if(!(
			    ((np0 <=j&&j<= np0_in)||(np1_in <=j&&j<= np1)) ||
			     ((np0<=j&&j<=np1)&&(i<= nx0_in||nx1_in <= i))
			 ))
				 continue;
			 f[ii+j] = dummy_value;
		     }
	      }
	      if ( function_type == 0 ) { /* remain_flag */
		 if (nx0 <= i && i <= nx1) {
		     np0 = phi0 / pi2 * (double)nph[i];
		     np1 = phi1 / pi2 * (double)nph[i];
		     np0_in = (phi0+dr) / pi2 * (double)nph[i];
		     np1_in = (phi1-dr) / pi2 * (double)nph[i];
		     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

		     /* process pixels near center */
		     if (np0 < 0) {
			int npR = nph[i] + np0;
			int npR0_in = np0_in<0 ? nph[i]+np0_in : np0_in;
			int npR1_in = np1_in<0 ? nph[i]+np1_in : np1_in;

			for(j=0; j<nph[i]; j++) {
			 if( (npR <= j || j <= np1)
			   &&
			   (
			     /*  |***|  |  |***|  */
			   ((np0_in < 0 && np1_in >= 0) &&
			    ((npR<=j&&j<=npR0_in)||(npR1_in<=j&&j<=np1)))||
			     /*  |***|    |*|**|  */
			   ((np0_in >= 0 && np1_in >= 0) &&
			     (npR<=j || j<=npR0_in ||(npR1_in<=j&&j<=np1)))||
			     /*  |*|**|    |***|  */
			   ((np0_in < 0 && np1_in <= 0) &&
			     ((0<=j&&j<=np1)|| j>= npR1_in || j<= npR0_in))||
			   (i <= nx0_in || nx1_in <= i)
			   ))
			       continue;
			   if (m[ii+j] !=1) f[ii+j] =dummy_value;
			}
		     }
		     if (np1 >= nph[i]) {
			int npR = np1 - nph[i];
			int npR0_in = np0_in > nph[i] ? np0_in-nph[i]:np0_in;
			int npR1_in = np1_in > nph[i] ? np1_in-nph[i]:np1_in;
			for(j=0; j<nph[i]; j++) {
			  if( (np0 <= j || j <= npR)
			   &&
			   (
			     /*  |***|  |  |***|  */
			   ((np1_in > nph[i] && np0_in < nph[i]) &&
			    ((npR<=j&&j<=npR0_in)||(npR1_in<=j&&j<=np1)))||
			     /*  |*|**|    |***|  */
			   ((np0_in <= nph[i] && np1_in <= nph[i]) &&
			     (j<=npR || npR1_in<=j ||(np0<=j&&j<=npR0_in)))||
			     /*  |***|    |*|**|  */
			   ((np0_in >= nph[i] && np1_in >= nph[i]) &&
			     ((np0<=j&&j<=nph[i])|| j<=npR0_in||j>=npR1_in))||
			   (i <= nx0_in || nx1_in <= i)
			   ))
				 continue;
			    if (m[ii+j] !=1) f[ii+j] =dummy_value;
			}
		     }

		     /* process pixels in other zone */
		     for(j=0; j<nph[i]; j++) {
			 if (
			    ((np0 <=j&&j<= np0_in)||(np1_in <=j&&j<= np1)) ||
			     ((np0<=j&&j<=np1)&&(i<= nx0_in||nx1_in <= i))
			 )
				 continue;
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		     }
		 } else {
		     for(j=0; j<nph[i]; j++)
			 if (m[ii+j] != 1)  f[ii+j] = dummy_value;
		 }
	      }
	 }
       }
    }

    if (m != NULL) free((void *)m);
    return;
}


/* extract rectangular area in equatorial coordinates */
void  extract_area_equ(r, nm, nn, d_r, d_d, fnx, fny,
	  f, x, nx, np, nph,  ra1,dec1,ra2,dec2)
double **r, *d_r, *d_d;    /* output array '**r' and steps '*d_r','*d_d' */
int *nm, *nn;              /* dimensions of '**r' */
double *f, *x;             /* input map and knots */
int nx, np, *nph;          /* dimensions of input map        */
double fnx,fny;            /* increasing factors for nx & ny */
double ra1,dec1,ra2,dec2;  /* coordinates of requested area  */
{
    double d_ra= *d_r, d_dec= *d_d, r0,d0, b0,l0,th0,ph0, dRA=0.0,dDec=0.0;
    double p = 180/M_PI;
    int nx0, nx1;
    int flag_sra = 0, flag_sdec = 0;
    register int i,j;

    /* restoration of values of 'flag_sra' and 'flag_sdec' */
    /* from steps 'd_ra' and 'd_dec'                       */
    if (ABS(d_ra-GLESPMASK) < 500) { flag_sra=1; }
    else if (ABS(d_ra) > 1.e-6) { dRA=d_ra; flag_sra=2; }
    if (ABS(d_dec-GLESPMASK) < 500) { flag_sdec=1; }
    else if (ABS(d_dec) > 1.e-6)  { dDec=d_dec; flag_sdec=2; }

    /* center of area */
    r0 = ra1  + .5 * (ra2 -ra1) ;      /* RA, hours */
    d0 = dec1 + .5 * (dec2-dec1);      /* Dec, degs */
    equat2gal(r0, d0, &b0, &l0);       /* b,l   in degs */
    gal2pol(b0/p, l0/p, &th0, &ph0);   /* th,ph in rad  */

    /* find approximate number of pixels by latitude/declination */
    d_dec = .5 * (dec2-dec1) / p;      /* halfsize in rad */
    nx0 = ind_x(nx, x, cos(th0-d_dec));   /* No of ring */
    nx1 = ind_x(nx, x, cos(th0+d_dec));
    *nn = (ABS(nx1-nx0)+1) * 2;        /* number of pixels by Dec */

    d_dec = (dec2-dec1) / (double)*nn; /* default step by Dec, degs */

    /* find approx. number of pixels by RA */
    {
	double cos_d = 1.0;
	/* if (*nm == -1000) then use cos(delta) to set step by RA */
	if (*nm == -1000) cos_d = cos(d0/p);
	d_ra = (ra2-ra1) * 15. * cos_d / d_dec;    /* RA in degs --> nm */
    }
    *nm = ABS(d_ra) + .5;              /* number of pixels  */
    d_ra = (ra2-ra1) / (double)*nm;    /* step by RA, hours */

    /* if step by RA is given by user in hours */
    if (flag_sra) {
	  /* 'iy' is a number by Dec, 'ix' is a number by Ra */
	  int ix = *nm, iy = *nn;
	  double dr0, dd0;
	  /* same pixels equal to external dRA */
	  if (flag_sra == 2) {
		if (d_ra < 0) dr0 = -fabs(dRA);
		else dr0 = fabs(dRA);
		*nm = d_ra * ix / dr0 + 0.5;   /* new number */
		d_ra = dr0;
	  }
	  /* we have 'd_ra' and then calculte 'd_dec' */
	  if (d_dec < 0) dd0 = -fabs(d_ra*15.);
	  else dd0 = fabs(d_ra*15.);           /* step by Dec, in degs */
	  *nn = d_dec * iy / dd0 + 0.5;
	  d_dec = dd0;
    } else  if (flag_sdec) {
	  /* if step by declination is given */
	  int ix = *nm, iy = *nn;
	  double dr0, dd0;
	  /* same pixels equal to dDec */
	  if (flag_sdec == 2) {
		if (d_dec < 0) dd0 = -fabs(dDec);
		else dd0 = fabs(dDec);
		*nn = d_dec * iy / dd0 + 0.5;
		d_dec = dd0;
	  }
	  if (d_ra < 0) dr0 = -fabs(d_dec/15);
	  else dr0 = fabs(d_dec/15);
	  *nm = d_ra * ix / dr0 + 0.5;
	  d_ra = dr0;
    }

    /* change size of pixels if fnx!=1 and fny!=1 */
    *nm = *nm * fnx + .5;
    if (*nm < 2)  *nm = 2;
    d_ra = (ra2 - ra1) / (double)*nm;
    *nn = *nn * fny + .5;
    if (*nn < 2)  *nn = 2;
    d_dec = (dec2 - dec1) / (double)*nn;

    /* find memory */
    i = *nn * *nm;
    if ((*r = (double *)calloc((unsigned)i, sizeof **r)) == (double *)0) {
	PR(ER, "%s: no memory for temporal array.\n", myname);
	exit(1);
    }

    r0 = ra1;   /* in hours */
    d0 = dec1;  /* in degs  */

    /* fill in arrays */
    for(i=0; i < *nn; i++) {  /* loop over Dec */
       int ii = i * *nm, k;
       double tn = d0 + i * d_dec, pn;
       for(j=0; j< *nm; j++) {   /* loop over RA */
	   pn = r0 + j * d_ra;  /* in hours */
	   /* (pn; tn) = (RA,hours; Dec,degs) -> (b,l) */
	   equat2gal(pn, tn, &b0, &l0);
	   /* (b0/p; l0/p) = (glat,rad; glon,rad) -> (th,phi) */
	   gal2pol(b0/p, l0/p, &th0, &ph0);
	   k = ang2n(th0, ph0, x, nx, np, nph);
	   if (k<0) continue;
	   (*r)[ii+j] = f[k];
       }
    }

    *d_r = d_ra;
    *d_d = d_dec;
    return;
}

/* extract rectangular area in galctic coordinates */
void  extract_area_gal(r, nm, nn, dt_r, dp_r,  fnx, fny,
	  f, x, nx, np, nph,  lat0,lon0,lat1,lon1)
double **r, *dt_r, *dp_r;    /* output array '**r' and steps '*dt','*dp' */
int *nm, *nn;                /* dimensions of '**r' */
double *f, *x;               /* input map and knots */
int nx, np, *nph;            /* dimensions of input map        */
double fnx,fny;              /* increasing factors for nx & ny */
double lat0,lon0,lat1,lon1;  /* coordinates of requested area  */
{
    register int i, j;
    int flag_lon=lon0 < 0.0 ? 1 : 0;
    double theta0, theta1, phi0, phi1;
    double dt = *dt_r, dp = *dp_r;
    int nx0, nx1, np0, np1, nb;
    int flag_slat=0, flag_slon=0;
    double dLAT=0.0, dLON=0.0;
    double pi2 = M_PI * 2;

    /* restoration of values of 'flag_slat' and 'flag_slon' */
    /* form steps 'dt' and 'dp                              */
    if (ABS(dt-GLESPMASK) < 500) { flag_slat=1; }
    else if (ABS(dt) > 1.e-6) { dLAT=dt; flag_slat=2; }
    if (ABS(dp-GLESPMASK) < 500) { flag_slon=1; }
    else if (ABS(dp) > 1.e-6)  { dLON=dp; flag_slon=2; }

    if (lon0 < 0.0) lon0 += 2*M_PI;
    if (lon1 < 0.0) { lon1 += 2*M_PI; flag_lon = 0; }

    /* limits of array by theta */
    gal2pol(lat0, lon0, &theta0, &phi0);
    gal2pol(lat1, lon1, &theta1, &phi1);
    nx0 = ind_x(nx, x, cos(theta0));
    nx1 = ind_x(nx, x, cos(theta1));
    *nn = ABS(nx1-nx0+1);
    dt = (theta1 - theta0) / (double)*nn;

    /* limits of array by phi */
    /* pixels in nx0-ring */
    np0 = phi0 / pi2 * (double)nph[nx0];
    np1 = phi1 / pi2 * (double)nph[nx0];

    /* pixels in nx1-ring */
    i   = phi0 / pi2 * (double)nph[nx1];
    j   = phi1 / pi2 * (double)nph[nx1];

    if (flag_lon) {
	   /* when negative g_lon0 is input */
	   if ( ABS(np1+(nph[nx0]-np0)) > ABS(j+(nph[nx1]-i)) ) {
		 nb = nph[nx0]-np0+1;
		 *nm = ABS(np1+nb);
	   } else {
		 nb = nph[nx1]-i+1;
		 *nm = ABS(j+nb);
	   }
	   dp = (phi1 + (pi2 - phi0)) / (double)*nm;
	   phi0 -= pi2;
    } else {
	   if ( ABS(np1-np0) > ABS(j-i) )
		 *nm = ABS(np1-np0+1);
	   else
		 *nm = ABS(j-i+1);
	   dp = (phi1 - phi0) / (double)*nm;
    }

    /* if step by latitude is given by user */
    if (flag_slat) {
	  int ix = *nm, iy = *nn;
	  double dt0, dp0;
	  /* same pixels equal to dLAT */
	  if (flag_slat == 2) {
		if (dt < 0) dt0 = -fabs(dLAT);
		else dt0 = fabs(dLAT);
		*nn = dt * iy / dt0 + 0.5;
		dt = dt0;
	  }
	  if (dp < 0) dp0 = -fabs(dt);
	  else dp0 = fabs(dt);
	  *nm = dp * ix / dp0 + 0.5;
	  dp = dp0;
    } else  if (flag_slon) {
	  /* if step by longitude is given */

	  int ix = *nm, iy = *nn;
	  double dt0, dp0;
	  /* same pixels equal to dLON */
	  if (flag_slon == 2) {
		if (dp < 0) dp0 = -fabs(dLON);
		else dp0 = fabs(dLON);
		*nm = dp * ix / dp0 + 0.5;
		dp = dp0;
	  }
	  if (dt < 0) dt0 = -fabs(dp);
	  else dt0 = fabs(dp);
	  *nn = dt * iy / dt0 + 0.5;
	  dt = dt0;
    }

    /* find memory */
    i = *nn * *nm;
    if ((*r = (double *)calloc((unsigned)i, sizeof *r)) == (double *)0) {
	 PR(ER, "%s: no memory for arrays of '%d' elements...\n",
	     myname, i);
	 exit(1);
    }
    /* fill in arrays */
    for(i=0; i< *nn; i++) {
       int ii = i * *nm, k;
       double tn = theta0 + i * dt, pn;
       for(j=0; j< *nm; j++) {
	   pn = phi0 + j * dp;
	   if (pn < 0.0) pn += pi2;
	   k = ang2n(tn, pn, x, nx, np, nph);
	   if (k<0) continue;
	   (*r)[ii+j] = f[k];
       }
    }

    *dt_r = dt;
    *dp_r = dp;
    return;
}


/* extract pixels from in radius r_c */
void get_circle_array (lat_c, lon_c, r_c, zmask, f, x, nx, np, nph, arr, nar)
double lat_c, lon_c, r_c;   /* in radian */
double *zmask, *f, *x;
int nx, np, *nph;
double **arr;
int *nar;
{
    register int i, j;
    double lat0  = lat_c, lon0  = lon_c;     /* in radian */
    double radius = r_c,  dpx = 2*M_PI/np;
    int nx_c, np_c;   /* center of circle */
    int nx0, nx1, np0, np1;
    double theta0, phi0, cth0, sth0;
    int n_pixels = 0, n_pix = 0;
    double v, pi2 = 2.*M_PI;

    if (radius < -1.e+6)  radius = fabs(radius + 1.e+6) * dpx;

    /* extract circle */
    gal2pol(lat0, lon0, &theta0, &phi0);
    cth0 = cos(theta0);
    sth0 = sqrt(1-cth0*cth0);
    nx_c = ind_x(nx, x, cth0);
    np_c = phi0 / pi2 * (double)nph[nx_c];

    nx0 = ind_x(nx, x, cos(theta0-radius));
    nx1 = ind_x(nx, x, cos(theta0+radius));
    if ( nx0 > nx1 ) { j=nx0; nx0=nx1; nx1=j; }


    /* culculation of number of pixels */
    /*---------------------------------*/
    if (theta0-radius < 0. || theta0+radius > M_PI ) {
      /* process pixels near pole */
      for(i=0; i<nx; i++) {
	int ii = i * np;
	double cth = x[i], sth = sqrt(1.-cth*cth);
	double dphi = pi2 / (double)nph[i];
	double dp;   /* dt = theta0 - acos(cth), */

	if (!((theta0+radius > M_PI &&
	  ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))) ||
	  (theta0-radius < 0.   &&
	  ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))))) {
	     for(j=0; j<nph[i]; j++) {
		 if (zmask[ii+j] > 0)   /* if the area is masked */
			continue;
		 dp = phi0 - j*dphi;
		 v = acos(cos(dp)*sth0*sth + cth0*cth);
		 if (v <= radius || v < dpx)  {
			 n_pixels ++;
			 continue;
		 }
	     }
	}
      }
    } else {
       for(i=0; i<nx; i++) {
	 int ii = i * np;
	 double cth = x[i],  sth = sqrt(1-cth*cth);
	 double dphi = pi2 / (double)nph[i];
	 double dp; /* dt = theta0 - acos(cth), */
	 /* make checking interval a little larger */
	 double rl = radius/sth0/sth;

	 if (i >= nx0 && i <= nx1) {
	     np0 = (phi0-rl) / dphi;
	     np1 = (phi0+rl) / dphi;
	     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

	     if (np0 < 0) {
		  int npR = nph[i] + np0;
		  for(j=0; j<nph[i]; j++) {
		    if (zmask[ii+j] > 0)   /* if the area is masked */
			   continue;
		    if (npR < j || j < np1) {
		      dp = (2*M_PI+phi0) - j*dphi;
		      v = acos(cos(dp)*sth0*sth + cth0*cth);
		      if (v <= radius || v < dpx) {
			      n_pixels ++;
			      continue;
		      }
		    }
		  }
	     }
	     if (np1 > nph[i]) {
		 int npR = np1 - nph[i];
		 for(j=0; j<nph[i]; j++) {
		   if (zmask[ii+j] > 0)   /* if the area is masked */
			continue;
		   if (npR < j || j < np0) {
		     dp = -phi0 + j*dphi + 2*M_PI;
		     v = acos(cos(dp)*sth0*sth + cth0*cth);
		     if (v <= radius || v < dpx) {
			     n_pixels ++;
			     continue;
		     }
		   }
		 }
	     }
	     for(j=0; j<nph[i]; j++) {
		 if (zmask[ii+j] > 0)   /* if the area is masked */
			continue;
		 if (j >= np0 && j <= np1) {
		       dp = phi0 - j*dphi;
		       v = acos(cos(dp)*sth0*sth + cth0*cth);
		       if (v <= radius || v < dpx) {
			       n_pixels ++;
			       continue;
		       }
		 }
	     }
	 }
       }
    }

    /* find memory  for array of pixels */
    *nar = n_pixels;
    if ((*arr = (double *)calloc(n_pixels, sizeof **arr)) == (double *)0){
	PR(ER, "%s: no memory for array of pixels from circle.\n", myname);
	exit(1);
    }

    /* filling in array with selected pixels */
    /*--------------------------------------*/
    if (theta0-radius < 0. || theta0+radius > M_PI ) {
      /* process pixels near pole */
      for(i=0; i<nx; i++) {
	int ii = i * np;
	double cth = x[i], sth = sqrt(1.-cth*cth);
	double dphi = pi2 / (double)nph[i];
	double dp;   /* dt = theta0 - acos(cth), */

	if (!((theta0+radius > M_PI &&
	  ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))) ||
	  (theta0-radius < 0.   &&
	  ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))))) {
	     for(j=0; j<nph[i]; j++) {
		 if (zmask[ii+j] > 0)   /* if the area is masked */
			continue;
		 dp = phi0 - j*dphi;
		 v = acos(cos(dp)*sth0*sth + cth0*cth);
		 if (v <= radius || v < dpx)  {
			 (*arr)[n_pix++] = f[ii+j];
			 continue;
		 }
	     }
	}
      }
    } else {
       for(i=0; i<nx; i++) {
	 int ii = i * np;
	 double cth = x[i],  sth = sqrt(1-cth*cth);
	 double dphi = pi2 / (double)nph[i];
	 double dp; /* dt = theta0 - acos(cth), */
	 /* make checking interval a little larger */
	 double rl = radius/sth0/sth;

	 if (i >= nx0 && i <= nx1) {
	     np0 = (phi0-rl) / dphi;
	     np1 = (phi0+rl) / dphi;
	     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

	     if (np0 < 0) {
		  int npR = nph[i] + np0;
		  for(j=0; j<nph[i]; j++) {
		    if (zmask[ii+j] > 0)   /* if the area is masked */
			   continue;
		    if (npR < j || j < np1) {
		      dp = (2*M_PI+phi0) - j*dphi;
		      v = acos(cos(dp)*sth0*sth + cth0*cth);
		      if (v <= radius || v < dpx) {
			      (*arr)[n_pix++] = f[ii+j];
			      continue;
		      }
		    }
		  }
	     }
	     if (np1 > nph[i]) {
		 int npR = np1 - nph[i];
		 for(j=0; j<nph[i]; j++) {
		   if (zmask[ii+j] > 0)   /* if the area is masked */
			continue;
		   if (npR < j || j < np0) {
		     dp = -phi0 + j*dphi + 2*M_PI;
		     v = acos(cos(dp)*sth0*sth + cth0*cth);
		     if (v <= radius || v < dpx) {
			     (*arr)[n_pix++] = f[ii+j];
			     continue;
		     }
		   }
		 }
	     }
	     for(j=0; j<nph[i]; j++) {
		 if (zmask[ii+j] > 0)   /* if the area is masked */
			continue;
		 if (j >= np0 && j <= np1) {
		       dp = phi0 - j*dphi;
		       v = acos(cos(dp)*sth0*sth + cth0*cth);
		       if (v <= radius || v < dpx) {
			       (*arr)[n_pix++] = f[ii+j];
			       continue;
		       }
		 }
	     }
	 }
       }
    }

    return;
}
