#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr
#define ST      stdout

/*
 * 'mappat' generates standard patterns of spherical and analytical functions
 *  and converts ASCII data to GLESP-format and back
 *      By Oleg Verkhodanov (2003, TAC, Copenhagen)
 *      last modification 29/10/2005 (SAO, vo@sao.ru), 23/07/2008
 *      04/08/2010, 26/10/2010 (NBI, development for Alisa)
 *      09/2015, 05/2016
 */
extern void setopt(int, char **), usage(void), usage2(void);
extern double y20(double),
	      y21(double, double, int), y22(double, double, int),
	      y21_cos(double, double),  y21_sin(double, double),
	      y22_cos(double, double),  y22_sin(double, double),
	      y30(double),
	      y31(double, double, int), y32(double, double, int),
	      y33(double, double, int),
	      y31_cos(double, double),  y31_sin(double, double),
	      y32_cos(double, double),  y32_sin(double, double),
	      y33_cos(double, double),  y33_sin(double, double),
	      y10(double), y11(double, double, int),
	      y11_cos(double, double),  y11_sin(double, double);
extern double yf(double, double, int);
extern void   gauleg (double, double, double *, double *, int);
extern int    map_formu(double *, double *, int, int, int *, char *);

int  set_Legendre_grid = 0;
int lx_add = 0;     /* addition to Nphi numbers for 'coord'-function */
		    /*   case 'a': case 'A': lx_add = -10;           */
		    /*   case 'n': case 'N': lx_add =   0;           */
		    /*   case 's': case 'S': lx_add =   4;           */

int ffile_flag = 0;   /* output to FITS file                           */

char inpfile_ps[81];      /* ASCII file with sources   */
char inpfile_map[81];     /* ASCII file with map       */
int pattern_type = 0;
int counts_flag = 0;      /* count number points per pixel */
int cats_flag = 0;        /* if =1, than read str from http://cats.sao.ru */


int debug_flag = 0;       /* debug flag                */
int  set_hp_grid = 0;     /* set Healpix grid          */
int nside = -1;
double nside_factor = 1.0;
int nphi_flag=1;          /* select phi-size of pixel  */
int  y_flag = 21, ft_flag = 0;
char out_map_file[161], x_output_file[161], w_output_file[161];
char formu_str[513];

char logN_logS_file[161];
char inp_file[161];
int nx=0, np=0;
int coor_flag = 0;  /* 0 - polar coor-s,  1 - Galactic, 2 - equatorial */
int print_flag=0;
double conv_factor = 1.0; /* conversion factor of temperature       */
int rescale_flag = 0,     /* rescaling to covert any coor.sys. to polar */
	cr1_sg = 0,       /* flag to change sign by the 1st coordinate */
	cr2_sg = 0;       /* flag to change sign by the 2nd coordinate */
double  phi0 = 0.0;       /* shift of phi0 */
int phi0_flag = 0;        /* flag for phi0 */
int square_area_input=0;  /* use interpolation to calc corresp. map    */
int z_mask_flag = 0;      /* set mask                                  */
double mask_value = M_PI * 1.e+30;
int zone_circle = 0;
char zone_file[81];       /* file with list of zones to be processed   */
double zone_radius = -1.; /* radius of removed circle in degs          */
double zone_latR0 = -1000., zone_lonR0 = -1000.,
       zone_latR1 = -1000., zone_lonR1 = -1000.;
int    z_remove_flag = 1,
       z_remain_flag = 0;
/* incorporate  Basic FITS to GLESP FITS map */
char bfits_file[161];     /* input Basic FITS file with a map          */
int flag_bf = 0;/* if flag=1,then converse FITS map to the full sphere */
		/* if flag=0,then converse FITS map as it is (part)    */
int layout = 0; /* layout in basic fits to be read                     */
/* limits for map values */
#define NV_INTR 10
struct _vv_  { double low, high; } v_intr [NV_INTR];
int nv_intr=0;
int n_rand_points = 0;    /* put N random ponits on a sphere           */
int n_rand = -1;          /* start random seed                         */
int noise_level = 0.;     /* noise level                               */
int flag_rand = 0;        /* 'g' - gaussian, 'u' - uniform             */
int flag_nn = 0;          /* convert coorinates to pixel numbers       */
char nn_file[0];          /* name of file with input coodinates        */
double glon0=-1001,glat0=-1001;
char bfits_list[161];

int Alisa_format_input=0; /* LHC Alisa format input flag               */
int Alisa_event_num= -1;  /* select event with corresponding number    */
int Alisa_event_type= -1; /* select corr. type event: 1-prim., 0-2nd.  */
int Alisa_event_infty= -1;/* flag to select infinty events             */

char *myname="mappat";

int main(argc, argv)
int argc;
char **argv;
{
    register int i, j;
    FILE *fp_in = (FILE *)0, *fp_out = (FILE *)0;
    char str[81];
    double *f = (double *)0, *x = (double *)0, *mask = (double *)0;
    int *nph = (int *)0,  *cn = (int *)0;
    double dx, x0, pi2 = 2. * M_PI, *pw=(double *)0;
    double *phi = (double *)0;
    char grid_type[30], x_type[30];
    int  grid_flag = 0;

    w_output_file[0] = x_output_file[0] = out_map_file[0] =
	    inp_file[0] = str[0] = '\0';
    inpfile_ps[0] = inpfile_map[0] = nn_file[0] = '\0';
    logN_logS_file[0] = '\0';

    strcpy(tmpcat, ".");

    /* loop over external parameters */
    setopt(argc, argv);

    /* flag values to set default Legendre grid */
    if (lx_add == -10 || lx_add == 0 || lx_add == 4 || lx_add == 'P') {
	  set_Legendre_grid = 1;
    }

    /* if no flags, then read a map from file */
    if ( *inp_file != '\0' && print_flag) {
	  read_glesp_map(&f, &x, &nx, &np, &nph, inp_file);
    }

    /* input a number of rows by theta */
    if (nx == 0) {
read_nx:
	    PR(ER, "%s: input a number of rows by cos(theta) => ", myname);
	    fgets(str, 60, stdin);
	    if (*str == '\0')  /* ^D */
		    exit(0);
	    if (*str == '\n') {
		    str[0] = '\0';
		    goto read_nx;
	    }
	    nx = atoi(str);
	    nx = ABS(nx);
    }
    /* input a number of columns by phi */
    if (np == 0) {
read_nphi:
	    PR(ER, "%s: input a number of columns by phi => ", myname);
	    fgets(str, 60, stdin);
	    if (*str == '\0')  /* ^D */
		    exit(0);
	    if (*str == '\n') {
		    str[0] = '\0';
		    goto read_nphi;
	    }
	    np = atoi(str);
	    np = ABS(np);
    }

    /* find memory for arrays */
    if (f == (double *)0) {
	if ((x = (double *)calloc((unsigned)nx, sizeof *x)) == (double *)0) {
nomem:
	    PR(ER, "%s: no memory for arrays...\n", myname);
	    exit(1);
	}
	if ((nph = (int *)calloc((unsigned)nx, sizeof *nph)) == (int *)0)
		goto nomem;
	if ((f=(double *)calloc((unsigned)(nx*np), sizeof *f))==(double *)0)
		goto nomem;
    }
    if (z_mask_flag) {
	if ((mask=(double *)calloc((unsigned)(nx*np), sizeof *mask))==
	    (double *)0)
		goto nomem;
    }
    if (pattern_type == 'a') {
	if ((cn = (int *)calloc((unsigned)(nx*np), sizeof *cn)) == (int *)0)
		goto nomem;
    }

    /* get output file name */
    if (*out_map_file == '\0') {
read_filename:
	    PR(ER, "%s: Could you give a name of output file ? => ",
		    myname);
	    fgets(out_map_file, 160, stdin);
	    if (*out_map_file == '\0')  /* ^D */
		    exit(0);
	    if (*out_map_file == '\n') {
		    out_map_file[0] = '\0';
		    goto read_filename;
	    }
    }

    if (strcmp(out_map_file, "-") == 0 ||
	strcmp(out_map_file, "stdout") == 0 ) {
	    fp_out = stdout;
    }
      else
	    if ((fp_out = fopen(out_map_file, "w")) == (FILE *)0) {
		 PR(ER, "%s: can not create file '%s'\n",
			myname, out_map_file);
		 exit(1);
	    }

    if ( *inp_file == '\0' ) {
	/* copy array of cos-theta */
	/*-------------------------*/
	/* recalc to Legendre_grid */
	if ( set_Legendre_grid ) {
		double x1= -1.0, x2 = 1.0;
		grid_flag = GRID_PHI_X_LEGENDRE;
		gauleg(x1, x2, x, (double *)0, nx);
	} else if ( set_hp_grid ) {
		/* HEALPix grid input */
		double fc1, fc2;
		int nt, ns2;
		if (nside < 1)  {
			int l_n;
			/* ln 2 = M_LN2 = 0.69314718055994530942 */
			i = (int) (log((double)(np))/M_LN2  + 0.5);
			for(l_n=1,j=0; j<i; j++)  l_n *= 2;
			nside = l_n / 4;
		}
		nside = nside/nside_factor +0.5; /* resolution parameter */
		nt = 4*nside - 1;                /* X dimension=nx       */
		ns2 = nside*2;                   /* half of nt           */

		if (nt > nx) {
		     /* free old memory and take new one */
		     if(x != (double *)0) free((void *)x);
		     nx = nt;
		     if ((x = (double *)calloc((unsigned)nx, sizeof *x)) ==
			(double *)0)
				goto nomem;
		     if(nph != (int *)0) free((void *)nph);
		     if ((nph = (int *)calloc((unsigned)nx, sizeof *nph)) ==
			(int *)0)
				goto nomem;
		     if (f != (double *)0) free((void *)f);
		     f = (double *)0;
		}
		i = nt + 1;
		if (i > np) {
		     if (f != (double *)0) free((void *)f);
		     f = (double *)0;
		     np = i;
		}
		if (f == (double *)0)
		     if ((f = (double *)calloc((unsigned)(nx*np), sizeof *f))
			== (double *)0)
				goto nomem;

		/* fill in 'theta_hp'-array */
		fc1 = 1./3./nside/nside;
		fc2 = 2./3./nside;
		for (i=0; i<ns2; i++) {
		    int i1 = i+1;

		    /* size by phi for North pole */
		    nph[i] = 4 * (i1 < nside ? i1 : nside);
		    /* value of cos(theta) for North pole */
		    if ( i1 < nside )
			  x[i] = 1. - fc1 * i1 * i1;
		    else
			  x[i] = (ns2 - i1) * fc2;

		    j = nt-i1; /* symmetrical index for South */
		    /* size by phi for South pole */
		    nph[j] = nph[i];
		    /* value of cos(theta) for South pole */
		    if ( j > nt - nside - 1)
			  x[j] = -x[i];
		    else
			  x[j] = -x[i];
		}

		/* set 'phi' array */
		if (phi == (double *)0) {
		    double dphi = 2*M_PI / np, p0 = 0.0;

		    /* take memory for phi */
		    phi = (double *)calloc(np, sizeof *phi);
		    if (phi == (double *)0) {
			PR(ER,"%s: no memory for output arrays.\n", myname);
			exit(1);
		    }
		    for(i=0; i<np; i++)  phi[i] = p0 + i * dphi;
		}

		PR(ER, "ntotal=%d\n", 12*nside*nside);
	} else {
		/* uniform grid input */
		dx = 2. / (nx - 1);
		x0 = -1.;
		for (i=0; i<nx; i++)  {
			x[i] = x0 + i * dx;
		}
	}

	/* copy array of phi-dimensions */
	if ( !set_hp_grid ) {
	    if (nphi_flag) {
		/* Legendre grid by phi */
		int nx2 = (nx+1)/2;

		int n0 = 0;

		for (i=0; i<nx2; i++) {
		    switch (lx_add) {
		    case 0 :  /* pin = 0 => square matrix */
			nph[i] = np;
			break;
		    case 4 :
			if ( i==0 )  nph[i] = 10;
			else         nph[i] = nph[0] + i * lx_add;
			if (nph[i] > np) nph[i] = np;
			break;
		    case 'P':   /* polarization */
			nph[i] = get_pol_nphi_by_x(x, nx, np, i);
			break;
		    case -10:
		    default :
			nph[i] = get_nphi_by_x2(x, nx, np, i);
			break;
		    }

		    n0 += nph[i];

		    j = nx - 1 - i;
		    if (j > i) {
			     nph[j] = nph[i];
			     n0 += nph[i];
		    }
		}
	    } else {
		/* uniform grid by phi */
		for (i=0; i<nx; i++)  nph[i] = np;
	    }
	}
    }

    /*------------------------------------*/
    /* set limits of values to be plotted */
    /*------------------------------------*/
    if (nv_intr) {
	for(i=0; i<nx; i++) {
	    int ii = i * np;
	    for(j=0; j<nph[i]; j++) {
		int n = ii+j;
		int flag = 0, k;
		for (k=0; k<nv_intr; k++) {
		    if (v_intr[k].low < f[n] && f[n] < v_intr[k].high){
			flag = 1;
			break;
		    }
		}
		if ( !flag ) f[n] = mask_value;
	    }
	}
    }

    /* read region for masking */
    /*-------------------------*/
    if ( z_mask_flag ) {
       register int is=0;
       int ns = 1;
       struct _zone_ {
	    double latR0, lonR0;
	    double latR1, lonR1;
	    double radius;
	    int type;               /* zone type */
       } *zz;
       extern void map_zone(struct _zone_ *, int, int,
	       double *, double *, int, int, int *, double);
       int function_type = 1; /* remove_flag */
       int zone_type = zone_circle ? 0 : 1;

       /* if to read zone-file */
       if (zone_file[0] != '\0') {
	    FILE *fz;
	    char str[81];
	    /* read data from file with zones */
	    if((fz = fopen(zone_file, "r")) == (FILE *)0) {
		PR(ER, "%s: I can not open file '%s'...\n",
		    myname, zone_file);
		exit(1);
	    }
	    /* get number of zones */
	    is =0;
	    while(fgets(str, 80, fz) != NULL) {
		if (*str == '#' || *str == '\n')
		    continue;
		is++;
	    }
	    /* allocate memory */
	    i = ns = is;
	    zz = (struct _zone_ *)calloc(i, sizeof *zz);
	    if (zz == (struct _zone_ *)0) goto nomem;

	    /* read zones  */
	    is=0;
	    fseek(fz, 0l, 0);
	    while(fgets(str, 80, fz) != NULL) {
		char s1[30], s2[30], s3[30], s4[30];
		char *cs = str;
		int n_par;  /* number of parameters per string */
		if (*str == '#' || *str == '\n')
			continue;
		while(*cs) if (*cs == '#') {*cs = '\0'; break;} else cs++;

		n_par = n_skip_wd(str);
		switch (n_par) {
		case 1:
		    PR(ER,"%s: there is only 1 parameter in the string:\n",
			    myname);
		    PR(ER,"%s", str);
		    PR(ER, "\n%s: I've skipped it...\n", myname);
		    break;
		case 2:
		    zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		    zz[is].radius = zone_radius;
		    zz[is].type = 0;
		    break;
		case 3:
		    zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		    if (zone_radius > 0.)
			zz[is].radius = zone_radius;
		    else
			zz[is].radius = deg2rad((skip_wd(2,str,s3),s3));
		    zz[is].type = 0;
		    break;
		case 4:
		    zz[is].latR0 = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0 = deg2rad((skip_wd(1,str,s2),s2));
		    zz[is].latR1 = deg2rad((skip_wd(2,str,s3),s3));
		    zz[is].lonR1 = deg2rad((skip_wd(3,str,s4),s4));
		    if (zone_type == 0) {
			    zz[is].radius = zz[is].latR1;
			    zz[is].type = 0;
		    } else  zz[is].type = 1;
		    break;
		}
		is++;
	    }

	    fclose(fz);
       } else {
	    /* allocate memory for 1 zone */
	    /* from command string */
	    i = ns = 1;
	    zz = (struct _zone_ *)calloc(i, sizeof *zz);
	    if (zz == (struct _zone_ *)0) goto nomem;
	    zz[0].latR0 = zone_latR0;  zz[0].lonR0 = zone_lonR0;
	    zz[0].latR1 = zone_latR1;  zz[0].lonR1 = zone_lonR1;
	    zz[0].radius = zone_radius;
	    if (zone_latR1 < -900.) {
		    zz[0].type = 0;
	    } if (zone_lonR1 < -900.) {
		    if (zone_radius > 0.)
			zz[is].radius = zone_radius;
		    else
			zz[is].radius = zone_latR1;
		    zz[0].type = 0;  /* circle */
	    } else {
		    zz[0].type = 1;  /* rectangular */
	    }
       }

       map_zone(zz, ns, function_type, mask, x, nx, np, nph, mask_value);

       /* free memory from zone-array */
       free((void *)zz);
    }


    if (*x_output_file != 0) {
	    FILE *fp_x;
	    if ((fp_x = fopen(x_output_file, "w")) == (FILE *)0) {
		 PR(ER, "%s: can not create file '%s'\n",
			myname, x_output_file);
		 exit(1);
	    }
	    for (i=0; i<nx; i++)
	       if (i==0)
		    PR(fp_x, "%12.5e %12.5e 0\n",
			    x[i], acos(x[i]));
	       else
		    PR(fp_x, "%12.5e %12.5e %12.5e\n",
			    x[i], acos(x[i]), acos(x[i])-acos(x[i-1]));
	    fclose(fp_x);
    }

    if (*w_output_file != 0) {
	    double x1= -1.0, x2 = 1.0, *pw0;
	    FILE *fp_w;
	    if ((pw = (double *)calloc((unsigned)nx, sizeof *pw)) ==
	       (double *)0)
		       goto nomem;
	    if ((pw0= (double *)calloc((unsigned)nx, sizeof *pw0)) ==
	       (double *)0)
		       goto nomem;
	    if ((fp_w = fopen(w_output_file, "w")) == (FILE *)0) {
		 PR(ER, "%s: can not create file '%s'\n",
			myname, w_output_file);
		 exit(1);
	    }
	    weights(1, x1, x2, 2.*M_PI, x, pw, nx, nph);
	    gauleg(x1, x2, x, pw0, nx);
	    for (i=0; i<nx; i++)
		    PR(fp_w, "%12.5e %12.5e %12.5e %12.5e\n",
			    x[i], acos(x[i]), pw[i], pw0[i]);
	    fclose(fp_w);
	    free((void *)pw);
	    free((void *)pw0);
    }

    if (flag_nn) {
	    int actual_number_flag = flag_nn == 2 ? 0 : 1;
	    int n_pix;
	    double theta_p, phi_p;

	    if (*out_map_file == '\0')
		    fp_out = stdout;


	    /* print pixel numbers */
	    if (nn_file[0] != 0) {
		FILE *fp_nn = (FILE *)0;
		double theta, phi, dist;

		fp_nn = fopen(nn_file, "r");
		if (fp_nn == (FILE *)0) {
		    PR(ER, "%s: I can not open file '%s'\n", myname, nn_file);
		    exit(1);
		}

		/* read coordinate from file */
		while (fgets(str, 80, fp_nn) != NULL) {
		   char st1[30], st2[30];
		   double lat, lon, alp, del;
		   double c = 180./M_PI;
		   register char *cs;

		   if (*str == '#' || *str == '\n' || *str == '\r')
			   continue;

		   cs = str;
		   while (*cs !='\0' && *cs != '#') cs++;
		   if (*cs == '#')  *cs='\0';

		   /* simple map */
		   sscanf(str, "%s %s", st1, st2);

		   switch (coor_flag) {
		   case 0 :  /* polar */
			   theta = deg2rad(st1);
			   phi   = deg2rad(st2);
			   break;
		   case 1 :  /* Galactic */
			   lat = atof(st1) / c;
			   lon = atof(st2) / c;
			   gal2pol(lat, lon, &theta, &phi);
			   break;
		   case 2 :  /* equatorial */
			   alp = str2deg(st1);
			   del = str2deg(st2);
			   equat2gal(alp, del, &lat, &lon);
			   lat  /= c;
			   lon  /= c;
			   gal2pol(lat, lon, &theta, &phi);
			   break;
		   }
		   if (actual_number_flag) {
		       /* angle to actual number of pixel in spherical grid */
		       n_pix = ang2na(theta, phi, x, nx, np, nph);
		       na2ang(n_pix, &theta_p, &phi_p, x, nx, np, nph);
		   } else {
		       /* convert angle to pixel number in rectangular array*/
		       n_pix = ang2n(theta, phi, x, nx, np, nph);
		       n2ang(n_pix, &theta_p, &phi_p, x, nx, np, nph);
		   }

		   dist = acos(cos(phi-phi_p) * sin(theta) * sin(theta_p)
				   + cos(theta) * cos(theta_p));
		   PR(fp_out, "%12.5e %12.5e %6d %12.5e %12.5e %12.5e\n",
		       theta, phi, n_pix+1, theta_p, phi_p, dist);
		}
		fclose(fp_nn);
	    } else {
		int ns = 0, n_pix = 0;
		/* printing pixels number */
		for(i=0; i<nx; i++) {
		    double dphi = pi2 / nph[i];
		    theta_p = acos(x[i]);
		    for(j=0; j<nph[i]; j++) {
			if (actual_number_flag) {
			    /* angle to actual number of pixel */
			    n_pix = ns+j;
			} else {
			    /* convert angle to pixel number in rect.grid */
			    n_pix = i*np+j;
			    n2ang(n_pix, &theta_p, &phi_p, x, nx, np, nph);
			}
			phi_p = j * dphi;
			PR(fp_out, "%6d %12.5e %12.5e\n",
					n_pix+1, theta_p, phi_p);
		    }
		    ns += nph[i];
		}
	    }
	    exit(0);
    }

    if (n_rand_points) pattern_type = 'r'; /* random points on a sphere */
    if (flag_rand)  pattern_type = 'R';   /* random noise in pixels */

    /******************************/
    /* calculation of map pattern */
    /******************************/
    switch(pattern_type) {
    case 'p': /* read ASCII file with sources  */
    case 'm': /* read ASCII file with map      */
    case 'a': /* read ASCII file with map and average */
	if ( set_hp_grid ) {
	    /* HEALPix grid input */
	    /*--------------------*/
	    double fc1, fc2;
	    int nt, ns2;
	    double *p_offset = (double *)0;
	    double cr1_min = 1.e+30, cr2_min = 1.e+30,
		   cr1_cf = 1.0, cr2_cf = 1.0;
	    double c = 180./M_PI;

	    if (nside < 1)  {
		    int l_n;
		    /* ln 2 = M_LN2 = 0.69314718055994530942 */
		    i = (int) (log((double)(np))/M_LN2  + 0.5);
		    for(l_n=1,j=0; j<i; j++)  l_n *= 2;
		    nside = l_n / 4;
	    }
	    nside = nside/nside_factor +0.5; /* resolution parameter */
	    nt = 4*nside - 1;                /* X dimension=nx       */
	    ns2 = nside*2;                   /* half of nt           */

	    if (nt > nx) {
		 /* free old memory and take new one */
		 if(x != (double *)0) free((void *)x);
		 nx = nt;
		 if ((x = (double *)calloc((unsigned)nx, sizeof *x)) ==
		    (double *)0)
			    goto nomem;
		 if ((p_offset = (double *)calloc((unsigned)nx,
		    sizeof *p_offset)) == (double *)0)
			    goto nomem;
		 if(nph != (int *)0) free((void *)nph);
		 if ((nph = (int *)calloc((unsigned)nx, sizeof *nph)) ==
		    (int *)0)
			    goto nomem;
		 if (f != (double *)0) free((void *)f);
		 f = (double *)0;
	    }
	    i = nt + 1;
	    if (i > np) {
		 if (f != (double *)0) free((void *)f);
		 f = (double *)0;
		 np = i;
	    }
	    if (f == (double *)0)
		 if ((f = (double *)calloc((unsigned)(nx*np), sizeof *f))
		    == (double *)0)
			    goto nomem;

	    /* fill in 'theta_hp'-array */
	    fc1 = 1./3./nside/nside;
	    fc2 = 2./3./nside;
	    for (i=0; i<ns2; i++) {
		int i1 = i+1;
		int cf = (i1 < nside || i1 > nx-nside ? 1 : (i+2-nside) % 2);
		double step;

		/* size by phi for North pole */
		nph[i] = 4 * (i1 < nside ? i1 : nside);

		/* value of cos(theta) for North pole */

		if ( i1 < nside ) {
		      x[i] = 1. - fc1 * i1 * i1;
		      np = i1 * 4;
		} else {
		      x[i] = (ns2 - i1) * fc2;
		      np = nside * 4;
		}

		j = nt-i1; /* symmetrical index for South */

		/* size by phi for South pole */
		nph[j] = nph[i];

		/* value of cos(theta) for South pole */
		if ( j > nt - nside - 1)
		      x[j] = -x[i];
		else
		      x[j] = -x[i];

		step = pi2/np;
		p_offset[i] = p_offset[j] = cf*step/2.;  /* shift by phi */
	    }

	    PR(ER, "hpix-grid: nside=%d, ntotal=%d\n", nside, 12*nside*nside);

	    /* read input data in HPix coordinate system */
	    while (fgets(str, 160, fp_in) != NULL) {
		   char st1[30], st2[30];
		   double theta, phi, t=0, lat, lon, alp, del;
		   int n_pix;
		   register char *cs;

		   if (*str == '#' || *str == '\n' || *str == '\r')
			   continue;

		   cs = str;
		   while (*cs !='\0' && *cs != '#') cs++;
		   if (*cs == '#')  *cs='\0';

		   /* simple map */
		   sscanf(str, "%s %s %lf", st1, st2, &t);

		   switch (coor_flag) {
		   case 0 :  /* polar */
			   if ( rescale_flag )  {
			       theta = (atof(st1)-cr1_min)*cr1_cf;
			       phi   = (atof(st2)-cr2_min)*cr2_cf;
			       if ( cr1_sg ) theta = M_PI - theta;
			       if ( cr2_sg ) phi   = 2*M_PI - phi;
			       if ( phi0_flag ) {
				   phi += phi0;
				   if ( phi > 2*M_PI )  phi -= 2*M_PI;
				   if ( phi < 0.0 )     phi += 2*M_PI;
			       }
			   } else {
			       theta = deg2rad(st1);
			       phi   = deg2rad(st2);
			   }
			   break;
		   case 1 :  /* Galactic */
			   lat = atof(st1) / c;
			   lon = atof(st2) / c;
			   gal2pol(lat, lon, &theta, &phi);
			   break;
		   case 2 :  /* equatorial */
			   alp = str2deg(st1);
			   del = str2deg(st2);
			   equat2gal(alp, del, &lat, &lon);
			   lat  /= c;
			   lon  /= c;
			   gal2pol(lat, lon, &theta, &phi);
			   break;
		   }

		   n_pix = hang2n(theta, phi, x, nx, np, nph, p_offset);

		   f[n_pix] += t;
	    }

	    if (p_offset != (double *)0)
		free ((void *)p_offset);
       } else {
	/* operations with GLESP */
	/*-----------------------*/
	int kx = 1;
	double x1 = -1., x2 = 1., phi2 = M_PI * 2.;

	if ((pw = (double *)calloc((unsigned)nx, sizeof *pw)) ==
	    (double *)0)
		    goto nomem;

	if (lx_add < -1000) { /* construct GLESP grid from polar cap */
	     kx = 0; j = -(lx_add + 1000);  /* number of pixels near pole */
	} else j = lx_add;

	coord(kx, j, x1, x2, phi2, x, pw, nx, nph, np);

	if (*inpfile_map != '\0') {
	    double cr1_min = 1.e+30, cr1_max = -1.e+30,
		   cr2_min = 1.e+30, cr2_max = -1.e+30,
		   cr1_cf = 1.0, cr2_cf = 1.0;
	    double c = 180./M_PI;
	    if ((fp_in = fopen(inpfile_map, "r")) == (FILE *)0) {
		    PR(ER, "%s: I can not open file '%s'.\n",
			myname, inpfile_map);
		    exit(1);
	    }
	    /* convert any coordinate system to polar one */
	    /* by simpliest way                           */

	    if ( rescale_flag) {
		double cr1, cr2, t;

		/* read data and find limits of coordinates */
		while (fgets(str, 160, fp_in) != NULL) {
		    char st1[30], st2[30], *cs;

		    if (*str == '#' || *str == '\n' || *str == '\r')
			    continue;

		    cs = str;
		    while (*cs !='\0' && *cs != '#') cs++;
		    if (*cs == '#')  *cs='\0';

		    st1[0] = st2[0] ='\0';

		    /* simple map */
		    sscanf(str, "%s %s %lf", st1, st2, &t);
		    cr1 = atof(st1);
		    cr2 = atof(st2);
		    if (cr1_min > cr1)  cr1_min = cr1;
		    if (cr1_max < cr1)  cr1_max = cr1;
		    if (cr2_min > cr2)  cr2_min = cr2;
		    if (cr2_max < cr2)  cr2_max = cr2;
		}
		cr1_cf = -M_PI / (cr1_max - cr1_min);
		cr2_cf =  M_PI*2/(cr2_max - cr2_min);
		if ( cr1_sg )  cr1_cf = -cr1_cf;
if (!glesp_comments) {
PR(ER, "coordinates rescaling: \n");
PR(ER, "min1=%12.5e -> 0, max1=%12.5e -> 3.14159\n", cr1_min, cr1_max);
PR(ER, "min2=%12.5e -> 0, max2=%12.5e -> 6.28318\n", cr2_min, cr2_max);
}
		/* return file pointer back */
		fseek(fp_in, 0l, 0);
	    }

	    if (square_area_input) {
		extern void read_uniform_area(FILE *, double **,int*,int*),
			    sq_area2glesp(double *, int, int,
				     double *,double *,int,int,int *, int);
		double *sq_area = (double *)0;
		int mx, my; /* dimensions on glesp */
		/* by polar direction coors.: 1 - theta, 2 - cos(theta) */
		int flag = square_area_input;
		/* read ASCII input data for rectang. coord. system */
		read_uniform_area(fp_in, &sq_area, &mx, &my);
		sq_area2glesp(sq_area, mx, my, f,x,nx,np,nph, flag);
		if (sq_area != (double *)0) free ((void *)sq_area);
	    } else {
		/* read input data just for GLESP coordinate system */
		while (fgets(str, 160, fp_in) != NULL) {
		   char st1[30], st2[30], *cs;
		   double theta, phi, t, lat, lon, alp, del;
		   int n_pix;

		   if (*str == '#' || *str == '\n' || *str == '\r')
			   continue;

		   cs = str;
		   while (*cs !='\0' && *cs != '#') cs++;
		   if (*cs == '#')  *cs='\0';

		   /* simple map */
		   sscanf(str, "%s %s %lf", st1, st2, &t);

		   switch (coor_flag) {
		   case 0 :  /* polar */
			   if ( rescale_flag )  {
			       theta = (atof(st1)-cr1_min)*cr1_cf;
			       phi   = (atof(st2)-cr2_min)*cr2_cf;
			       if ( cr1_sg ) theta = M_PI - theta;
			       if ( cr2_sg ) phi   = 2*M_PI - phi;
			       if ( phi0_flag ) {
				   phi += phi0;
				   if ( phi > 2*M_PI )  phi -= 2*M_PI;
				   if ( phi < 0.0 )     phi += 2*M_PI;
			       }
			   } else {
			       theta = deg2rad(st1);
			       phi   = deg2rad(st2);
			   }
			   break;
		   case 1 :  /* Galactic */
			   lat = atof(st1) / c;
			   lon = atof(st2) / c;
			   gal2pol(lat, lon, &theta, &phi);
			   break;
		   case 2 :  /* equatorial */
			   alp = str2deg(st1);
			   del = str2deg(st2);
			   equat2gal(alp, del, &lat, &lon);
			   lat  /= c;
			   lon  /= c;
			   gal2pol(lat, lon, &theta, &phi);
			   break;
		   }

		   n_pix = ang2n(theta, phi, x, nx, np, nph);

		   f[n_pix] += t;
		   if (pattern_type == 'a')
			cn[n_pix]++;
		}
	    }
	    fclose(fp_in);
	}

	/* point sources */
	if (*inpfile_ps != '\0') {
		int nxnp = nx*np;

		if ((fp_in = fopen(inpfile_ps, "r")) == (FILE *)0) {
			PR(ER, "%s: I can not open file '%s'.\n",
			    myname, inpfile_ps);
			exit(1);
		}

		while (fgets(str, 160, fp_in) != NULL) {
			double theta, phi, t;
			int n_pix;
			char *cs;
			if (*str == '#' || *str == '\n' || *str == '\r')
				continue;

			cs = str;
			while (*cs !='\0' && *cs != '#') cs++;
			if (*cs == '#')  *cs='\0';

			if (cats_flag) {
			    /* if coor_flag = 0 here, than              */
			    /* CATS has default coordinates: equatorial */
			    /* else consider coordinates as galactic    */
			    cats_str2sph(str, &theta, &phi, &t, coor_flag);
			} else
			    str2sph(str, &theta, &phi, &t, coor_flag);

			n_pix = ang2n(theta, phi, x, nx, np, nph);

			if (n_pix < 0 || n_pix > nxnp) {
			    if (str[strlen(str)-1] == '\n')
				str[strlen(str)-1] = '\0';
			    PR(ER, "%s:\n", myname);
			    PR(ER, "check coordinates (format or order) ");
			    PR(ER, "in a source data string:\n");
			    PR(ER, "'%s'\n", str);
			    exit(1);
			}

			if (counts_flag)
			    f[n_pix] += 1.0;
			else
			    f[n_pix] += t;
			if (pattern_type == 'a')
			     cn[n_pix]++;
		}
	}
	if (pattern_type == 'a') {
	     for (i=0; i<nx; i++) {
		int ii=i*np;
		for (j=0; j<nph[i]; j++) {
		   int k = ii+j;
		   if (cn[k])   f[k] /= (double)cn[k];
		}
	     }
	}
       }
       break;
    case 'A': /* read ALICE format file  */
       {
	extern void map_alisa();
	int kx = 1;
	double x1 = -1., x2 = 1., phi2 = M_PI * 2.;

	if ((pw = (double *)calloc((unsigned)nx, sizeof *pw)) ==
	    (double *)0)
		    goto nomem;

	if (lx_add < -1000) { /* construct GLESP grid from polar cap */
	     kx = 0; j = -(lx_add + 1000);  /* number of pixels near pole */
	} else j = lx_add;
	coord(kx, j, x1, x2, phi2, x, pw, nx, nph, np);
	map_alisa(f, x, nx, np, nph, inp_file,
	    Alisa_event_num, Alisa_event_type, Alisa_event_infty);
       }
       break;
    case 'b': /* read Basic FITS file with map  */
       {
	extern void map_basic_fits();
	/* incorporate  Basic FITS to GLESP FITS map */
	map_basic_fits(f, x, nx, np, nph, bfits_file, flag_bf, layout,
	    glon0,glat0);
       }
       break;
    case 'l': /* read Basic FITS files with map  */
       {
	extern void map_basic_files();
//        int kx = 1;
//        double x1 = -1., x2 = 1., phi2 = M_PI * 2.;
//
//        if ((pw = (double *)calloc((unsigned)nx, sizeof *pw)) ==
//            (double *)0)
//                    goto nomem;
//
//        if (lx_add < -1000) { /* construct GLESP grid from polar cap */
//             kx = 0; j = -(lx_add + 1000);  /* number of pixels near pole */
//        } else j = lx_add;
//        coord(kx, j, x1, x2, phi2, x, pw, nx, nph, np);
	map_basic_files(f,x,nx,np,nph, bfits_list, layout);
       }
       break;
    case 'd':  /* input form 'log N - log S' */
       break;
    case 'r': /* put N random points on a sphere */
       {
	extern int init_rand();
	extern double norm_rand();
	extern void rn2ij(int, int *, int, int *, int *);
	int n_tot = 0, r;

	/* random number */
	if (n_rand > -1)
	    srand(n_rand);
	else
	    init_rand(); /* take random seed from system time */

	for (i=0; i<nx; i++)  n_tot += nph[i]; /* total number of pixels */

	for (r=0; r<n_rand_points; r++) {
	    int ir,jr;
	    /* pixel random number */
	    int nr = n_tot * (double)rand() / MAX_RAND;

	    /* convert n-pixel to i,j */
	    rn2ij(nx,nph,nr,&ir,&jr);

	    if (ir < 0 || jr < 0) {
		    r--;
		    continue;
	    }
	    f[ir*np+jr] += 1.0;
	}
       }
       break;
    case 'R': /* random noise in pixels  */
       {
	extern int init_rand();
	extern void u_rand(double *, int, double, double);
	extern void g_rand(double *, int, double);
	extern void poi_rand(double*, int, double);
	double shift = 0.0;

	/* random number */
	if (n_rand > -1)
	    srand(n_rand);
	else
	    init_rand(); /* take random seed from system time */

	if (flag_rand == 'u')       /* uniform */
	    u_rand(f, nx*np, noise_level, shift);
	else if (flag_rand == 's')  /* poisson */
	    poi_rand(f, nx*np, noise_level);
	else                        /* gauss   */
	    g_rand(f, nx*np, noise_level);
       }
       break;
    case 'e': /* read ASCII file with extended sources  */
       {
	void map_ellip_src();
	map_ellip_src(f, x, nx, np, nph, inpfile_ps);
       }
       break;
    case 'c': /* show Gauss-Legendre grid  */
       {
	int nx2 = (nx+1)/2;
	extern int init_rand();
	extern double norm_rand();

	init_rand();
	for (i=0; i<=nx2; i++) {
		register int ii_n = i * np;
		register int ii_s = (nx-1-i) * np;
		int v_old=0;
		for (j=0; j<nph[i]; j++)  {
			int v = 15. * (double)rand() / MAX_RAND;

			/* neighbours have different values */
			while (v_old == v)
			     v = 15. * (double)rand() / MAX_RAND;

			f[ii_n+j] = v;
			if (ii_s > ii_n)
				 f[ii_s+j] = f[ii_n+j];
			v_old = v;
		}
	}
	break;
       }
    default :
	if ( *inp_file != '\0' )  break;
    case 'y':
	for (i=0; i<nx; i++) {
		register int ii = i * np;
		double dp = pi2 / nph[i];
		for (j=0; j<nph[i]; j++) {
			double phi = j * dp;
			f[ii+j] += yf(x[i], phi, ft_flag);
		}
	}
	break;
    case 'f':
	if ( map_formu(f, x, nx, np, nph, formu_str) < 0 )
	     exit(1);
	break;
    }

    if ( set_hp_grid ) {
	/* output HEALPix map */
	/*--------------------*/
	int flag_int=0; /* input of int values (in some simulations) */

	/* output data in HEALPix format */
	/* input array: f(nx, np)        */
	fhp_out2(&f, nx, nph, np, x, phi, -1, fp_out, flag_int);

    } else {
	/* output GLESP map */
	/*------------------*/

	/* select grid output keyword values */
	strcpy(x_type, "COS(THETA)");
	switch(grid_flag) {
	case GRID_PHI_X_LEGENDRE :
		strcpy(grid_type, "PHI(X)+LEGENDRE");
		break;
	default:
	case GRID_PHI_X:
		strcpy(grid_type, "PHI(X)+X");
		break;
	}

	if ( fabs(conv_factor-1.0) > 1.e-16) {
	    for (i=0; i<nx; i++) {
		register int ii = i * np;
		for (j=0; j<nph[i]; j++)  f[ii+j] *= conv_factor;
	    }
	}

	if (print_flag) {
	    double c = 180./M_PI;
	    double lim1 = mask_value * (1. - 1e-6),
		   lim2 = mask_value * (1. + 1e-6);

	    for (i=0; i<nx; i++) {
		register int ii, iw;

		/* print data in order of growing theta */
		if (x[0] < 0.)
		    { iw = nx - 1 - i; }
		else
		    {  iw = i; }
		ii = iw * np;

		for (j=0; j<nph[iw]; j++) {
		    double theta, phi, lat, lon, a, b;

		    if (nv_intr) {
			if ( lim1 < f[ii+j] && f[ii+j] < lim2)
			    continue;
		    }
		    if (z_mask_flag) {
		      if (z_remove_flag) {
			if ( lim1 < mask[ii+j] && mask[ii+j] < lim2)
			     continue;
		      }
		      if (z_remain_flag) {
			if ( mask[ii+j] < lim1 || lim2 < mask[ii+j])
			     continue;
		      }
		    }
		    pixel2angle(x, nx, nph, iw, j, &theta, &phi);

		    switch (coor_flag) {
		    case 0 :  /* polar */
			    a = theta;
			    b = phi;
			    break;
		    case 1 :  /* Galactic */
			    pol2gal(theta, phi, &lat, &lon);
			    a = lat * c;
			    b = lon * c;
			    break;
		    case 2 :  /* equatorial */
			    pol2gal(theta, phi, &lat, &lon);
			    lat  *= c;
			    lon  *= c;
			    gal2equat(lat, lon, &a, &b);
			    break;
		    }
		    if (coor_flag == 2) {
		      int hh, mm, dd, dm;
		      float ss, ds;
		      int c;
		      hh = (int)a; mm = (a - hh) * 60.;
		      ss = (a - hh - mm/60.) * 3600.;
		      c = b < 0. ? '-' : '+';
		      if (b < 0.)  b = -b;
		      dd = (int)b; dm = (b - dd) * 60.;
		      ds = (b - dd - dm/60.) * 3600.;
		      PR(fp_out,
			 "%02d:%02d:%06.3f  %c%02d:%02d:%05.2f  %15.8e\n",
			   hh,mm,ss, c,dd,dm,ds, f[ii+j]);
		    } else
			PR(fp_out, "%15.8e  %15.8e  %15.8e\n", a, b, f[ii+j]);
		}
	    }
	} else {
	    if (ffile_flag) {
	      /* map output in F-format */
	      ff_write_ext2_map(f, x, nx, np, nph, fp_out, x_type, grid_type);
	    } else {
	      /* map output in FITS-format */
	      outp_glesp_map(f, x, nx, np, nph, fp_out, x_type, grid_type);
	    }
	}
    }

    /* close files */
    if (fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);

    /* free memory */
    if (f != (double *)0)  free((void *)f);
    if (x != (double *)0)  free((void *)x);
    if (nph !=  (int *)0)  free((void *)nph);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case '0':
			y_flag = 0;
			pattern_type = 'y';
			break;
		case '1':
			y_flag = 1;
			pattern_type = 'y';
			break;
		case 'a': /* average for input data on a map */
			pattern_type = 'a';
			if (argv[i][2] == 'g')
			    /* input in galactic coordinates */
			    coor_flag = 1;
			else if (argv[i][2] == 'p')
			    /* input in polar coordinates */
			    coor_flag = 0;
			else   /* = q */
			    /* input in equatorial coordinates */
			    coor_flag = 2;
			strcpy(inpfile_map, argv[++i]);
			break;
		case 'A':
			/* LHC Alice format input flag */
			Alisa_format_input = 1;
			pattern_type = 'A';
			switch (argv[i][2]) {
			case 'c':
			    /* count of events in a pixel of selected type */
			    if (Alisa_event_type < 0) Alisa_event_type = 0;
			    Alisa_event_type += 1000;
			    break;
			case 'e':
			    /* select event with corresponding number    */
			    Alisa_event_num = atoi(argv[i]+3);
			    break;
			case 'i':
			    Alisa_event_infty = atoi(argv[i]+3);
			    if (!Alisa_event_infty) Alisa_event_infty=1;
			    break;
			case 't':
			    /* select corr. type event: 1-prim., 0-2nd.  */
			    if (argv[i][3] != '\0') {
				if (Alisa_event_type < 0) Alisa_event_type=0;
				Alisa_event_type += atoi(argv[i]+3);
			    }
			    break;
			}
			break;
		case 'c':
			if (strncmp(argv[i]+1, "co", 2) == 0) {
				/* count number points per pixel */
				counts_flag = 1;
			} else
				ft_flag = 1;
			break;
		case 'd':
			if (strcmp(argv[i]+1, "double") == 0) {
				bitpix_map = DOUBLEPIX;
			} else {
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			}
			break;
		case 'e':
			coor_flag = 2;  /* equatorial coordinates */
			break;
		case 'f':
			switch(argv[i][2]) {
			case 'b': /* read Basic FITS file */
				strcpy(bfits_file, argv[++i]);
				pattern_type = 'b';
				break;
			case 'c': /* conversion factor */
				conv_factor = atof(argv[++i]);
				break;
			case 'e': /* coordinates with extended objects */
				strcpy(inpfile_ps, argv[++i]);
				pattern_type = 'e';
				break;
			case 'f': /* coordinates with extended objects */
				strcpy(bfits_list, argv[++i]);
				pattern_type = 'l';
				break;
			case 'l': /* bitpix */
				bitpix_map = FLOATPIX;
				break;
			case 'm': /* read ASCII file with map */
				strcpy(inpfile_map, argv[++i]);
				pattern_type = 'm';
				break;
			case 'p':
				/* read ASCII file with sources  */
				if ( strncmp(argv[i]+1, "fpc", 3) == 0 )
				    cats_flag = 1;
				strcpy(inpfile_ps, argv[++i]);
				pattern_type = 'p';
				break;
			case 's': /* read ASCII string with formula */
				strcpy(formu_str, argv[++i]);
				pattern_type = 'f';
				break;
			default :
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
				break;
			}
			break;
		case 'g':
		case 'G':
			if (strncmp(argv[i]+1, "gr", 2) == 0) {
			  if (argv[i][3] != '\0') {
			    int c = argv[i][3];
			    switch (c) {
			    case 'a': case 'A': lx_add = -10; break;
			    case 'n': case 'N': lx_add =   0; break;
			    case 's': case 'S': lx_add =   4; break;
			    case 'i': goto lab1;
			    default :
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			    }
			  } else {
lab1:
			    i++;
			    if (argv[i][0] == 'a' || argv[i][0] == 'A') {
				/* quasi-equal area pixel */
				lx_add = -10;  /* old scheme */
			    } else if (argv[i][0]=='n' || argv[i][0]=='N') {
				/* equal pixel number in rings */
				lx_add = 0; /* pin 0 */
			    } else if (argv[i][0]=='s' || argv[i][0]=='S') {
				/* equal pixel number in rings */
				lx_add = 4; /* pin 4 */
			    }
			  }
			} else if (strncmp(argv[i]+1, "gla", 3) == 0) {
			    glat0 = atof(argv[++i]);
			} else if (strncmp(argv[i]+1, "glo", 3) == 0) {
			    glon0 = atof(argv[++i]);
			} else
			    coor_flag = 1;  /* galactic coordinates */
			break;
		case 'h':
			if (strcmp(argv[i]+1, "hp") == 0) {
				set_hp_grid = 1;
			} else if (strncmp (argv[i]+1, "h2", 2) == 0) {
			    usage2();
			} else
				usage();
			break;
		case 'i':
			strcpy(inp_file, argv[++i]);
			break;
		case 'k':
			if (strncmp("ke", argv[i]+1, 2) == 0) {
			    char s1[30], s2[30], s3[30], s4[30];
			    get_4_str(argv[++i], s1, s2, s3, s4);
			    if(*s1 !='\0')
			      zone_latR0=deg2rad(s1);else zone_latR0= -1001.;
			    if(*s2 !='\0')
			      zone_lonR0=deg2rad(s2);else zone_lonR0= -1001.;
			    if(*s3 !='\0')
			      zone_latR1=deg2rad(s3);else zone_latR1= -1001.;
			    if(*s4 !='\0')
			      zone_lonR1=deg2rad(s4);else zone_lonR1= -1001.;
			    z_remain_flag = 1;
			    z_remove_flag = 0;
			    z_mask_flag = 1;
			} else {
			    PR(ER, "%s: illegal flag '%s'\n",
				 myname, argv[i]);
			    usage();
			}
			break;
		case 'K':  /* keep data in selected zones */
			z_remain_flag = 1;
			z_remove_flag = 0;
			break;
		case 'l':
			if (strncmp(argv[i]+1, "li", 2) == 0) {
			  /* set limits to be plotted */
			  char s[50], *cs=s, *ps=argv[i+1], *ss;
			  while (*ps) {
			    while ((*cs = *ps) != '/' && *cs) {cs++;ps++;}
			    *cs = '\0';
			    if (*ps == '/') ps++;
			    v_intr[nv_intr].low = *s!='\0'?atof(s):mask_value;
			    if ( index(s,',') == NULL)
				ss = NULL;
			    else
				ss = index(s,',') + 1;
			    v_intr[nv_intr].high =
				(ss!=NULL&& *ss !='\0')? atof(ss):mask_value;
			    if (fabs(v_intr[nv_intr].low -mask_value)>1.e-10||
				fabs(v_intr[nv_intr].high-mask_value)>1.e-10)
				    nv_intr++;
			    cs=s;
			    if (nv_intr == NV_INTR)
				break;
			  }
			} else if (strncmp("la", argv[i]+1, 2) == 0) {
			    layout = atoi(argv[++i]);
			} else if (strncmp("logN", argv[i]+1, 4) == 0) {
			    /* input with 'log N - log S' */
			    strcpy(logN_logS_file, argv[++i]);
			} else
			    set_Legendre_grid = 1;
			break;
		case 'n':
			switch(argv[i][2]) {
			case 'f': /* nside factor */
				nside_factor = atof(argv[++i]);
				break;
			case 'g': /* generates gaussian noise at level: */
				noise_level = atof(argv[++i]); /* in K */
				flag_rand = 'g';
				break;
			case 'i': /* generates Poissonian noise at level: */
				noise_level = atof(argv[++i]); /* in K */
				flag_rand = 's';
				break;
			case 'n': /* print pixel numbers, actual number */
				if (argv[i][3] == 'g') coor_flag = 1;
				else if (argv[i][3] == 'q') coor_flag = 2;
				flag_nn = 1;
				if (i < argc-1 && argv[i+1][0] != '-')
				    strcpy(nn_file, argv[++i]);
				break;
			case 'N': /* print pixel numbers, numbe in rect.grid*/
				if (argv[i][3] == 'g') coor_flag = 1;
				else if (argv[i][3] == 'q') coor_flag = 2;
				flag_nn = 2;
				if (i < argc-1 && argv[i+1][0] != '-')
				    strcpy(nn_file, argv[++i]);
				break;
			case 's': /* Hpix nside */
				nside = atoi(argv[++i]);
				break;
			case 'x': /* number of points along cos(theta) */
				nx = atoi(argv[++i]);
				break;
			case 'p': /* number of points along phi */
				np = atoi(argv[++i]);
				break;
			case 'u': /* generates uniform noise at level: */
				noise_level = atof(argv[++i]); /* in K */
				flag_rand = 'u';
				break;
			default :
				PR(ER, "%s: illegal flag '%s'\n",
					myname, argv[i]);
				usage();
				break;
			}
			break;
		case 'N': /* put N random point on a sphere */
			n_rand_points = atoi(argv[++i]);
			break;
		case 'o':
			if (strcmp("ox", argv[i]+1) == 0) {
				strcpy(x_output_file, argv[++i]);
			} else if(strcmp("ow", argv[i]+1) == 0) {
				strcpy(w_output_file, argv[++i]);
			} else if(strncmp("of", argv[i]+1, 2) == 0) {
				strcpy(out_map_file, argv[++i]);
				ffile_flag = 1;
			} else
				strcpy(out_map_file, argv[++i]);
			break;
		case 'p':
			if (strncmp("pr", argv[i]+1, 2) == 0) {
				print_flag = 1;
			} else if (strncmp(argv[i]+1, "ph", 2) == 0) {
				phi0_flag = 1;
				phi0 = atof(argv[++i]) * M_PI/180;
			} else if (strncmp(argv[i]+1, "pin", 3) == 0) {
				/* phi increment number :       */
				/* addition to Nphi numbers for */
				/* 'coord'-function             */
				int v = atoi(argv[++i]);
				if (argv[i][0] == 'o')  {
				     v = atoi(argv[i]+1);
				     if ( v > 0 )
					lx_add = -10000 - v;
				     else
					lx_add = -10;
				} else if (argv[i][0] == 'P')
				     /* polarization */
				     lx_add = v = 'P';
				else if (argv[i][0] == 'r')
				     lx_add = v = -11;
				else if (v < -1)
				     lx_add = -1000 + v;
				else lx_add = v;
			} else {
				nphi_flag = 0;
			}
			break;
		case 'r':
			if (strcmp("rm", argv[i]+1) == 0 ||
			    strncmp("re", argv[i]+1, 2) == 0) {
			     char s1[30], s2[30], s3[30], s4[30];
			     get_4_str(argv[++i], s1, s2, s3, s4);
			     if(*s1 !='\0')
			       zone_latR0=deg2rad(s1);else zone_latR0= -1001.;
			     if(*s2 !='\0')
			       zone_lonR0=deg2rad(s2);else zone_lonR0= -1001.;
			     if(*s3 !='\0')
			       zone_latR1=deg2rad(s3);else zone_latR1= -1001.;
			     if(*s4 !='\0')
			       zone_lonR1=deg2rad(s4);else zone_lonR1= -1001.;
			     z_remove_flag = 1;
			     z_remain_flag = 0;
			     z_mask_flag = 1;
			} else if (strncmp ("ra", argv[i]+1, 2) == 0) {
			     zone_radius = deg2rad(argv[++i]);
			     zone_circle = 1;
			} else if (strncmp ("rs", argv[i]+1, 2) == 0) {
			     /* random seed */
			     n_rand  = atoi(argv[++i]);
			} else {
			    /* rescaling the ASCII map coor.sys. to polar */
			    rescale_flag = 1;
			    switch(argv[i][strlen(argv[i])-1]) {
			    case '1':
				    cr1_sg = 1;
				    if (argv[i][strlen(argv[i])-2] == '2')
					 cr2_sg = 1;
				    break;
			    case '2':
				    cr2_sg = 1;
				    if (argv[i][strlen(argv[i])-2] == '1')
					 cr1_sg = 1;
				    break;

			    }
			}
			break;
		case 'R':  /* keep data outside selected zones */
			z_remain_flag = 0;
			z_remove_flag = 1;
			break;
		case 's':
			if (strncmp(argv[i], "sq", 2) == 0) {
			      if (argv[i][strlen(argv[i])-1] == 'x')
				    square_area_input=2;
			      else
				    square_area_input=1;
			} else
			    ft_flag = -1;
			break;
		case 'v':
			printf("%s: version 0.8 (22/05/2016), GLESP 2.0\n",
			    myname);
			exit (0);
		case 'y':
			switch(argv[i][2]) {
			case '2':
				if (strcmp("y2", argv[i]+1) == 0) {
					y_flag = 20;
				} else if (strcmp("y21", argv[i]+1) == 0) {
					y_flag = 21;
				} else if (strcmp("y22", argv[i]+1) == 0) {
					y_flag = 22;
				} else if (strcmp("y20", argv[i]+1) == 0) {
					y_flag = 20;
				} else if (strcmp("y2-1", argv[i]+1) == 0) {
					y_flag = 21;
					ft_flag = 'c';
				} else if (strcmp("y2-2", argv[i]+1) == 0) {
					y_flag = 22;
					ft_flag = 0;
				}
				break;
			case '1':
				if (strcmp("y10", argv[i]+1) == 0) {
					y_flag = 10;
				} else if (strcmp("y11", argv[i]+1) == 0) {
					y_flag = 11;
				} else if (strcmp("y1-1", argv[i]+1) == 0) {
					y_flag = 11;
					ft_flag = 'c';
				}
				break;
			case '3':
				if (strcmp("y3", argv[i]+1) == 0) {
					y_flag = 30;
				} else if (strcmp("y30", argv[i]+1) == 0) {
					y_flag = 30;
				} else if (strcmp("y31", argv[i]+1) == 0) {
					y_flag = 31;
				} else if (strcmp("y3-1", argv[i]+1) == 0) {
					y_flag = 31;
					ft_flag = 'c';
				} else if (strcmp("y32", argv[i]+1) == 0) {
					y_flag = 32;
				} else if (strcmp("y3-2", argv[i]+1) == 0) {
					y_flag = 32;
					ft_flag = 'c';
				} else if (strcmp("y33", argv[i]+1) == 0) {
					y_flag = 33;
				} else if (strcmp("y3-3", argv[i]+1) == 0) {
					y_flag = 33;
					ft_flag = 'c';
				}
				break;
			default:
				break;
			}
			if (argv[i][strlen(argv[i])-1] == 'c')
				ft_flag = 1;
			else if (argv[i][strlen(argv[i])-1] == 's')
				ft_flag = -1;
			pattern_type = 'y';
			break;
		case 'x': /* random pixel value pattern */
			pattern_type = 'c';
			break;
		case 'z':
		    switch (argv[i][2]) {
		    case 'c':
			{  char s1[30], s2[30];
			  /* center of a circle to be masked */
			  get_2_str(argv[++i], s1, s2);
			  if(*s1 !='\0')
			    zone_latR0=deg2rad(s1); else zone_latR0= -1001.;
			  if(*s2 !='\0')
			    zone_lonR0=deg2rad(s2); else zone_lonR0= -1001.;
			  zone_circle = 1;
			}
			break;
		    case 'r':
			/* radius of a circle to be masked */
			zone_radius = deg2rad(argv[++i]);
			zone_circle = 1;
			break;
		    case 'o':
			/* select a zone to be masked */
			{ char s1[30], s2[30], s3[30], s4[30];
			    get_4_str(argv[++i], s1, s2, s3, s4);
			    if(*s1 !='\0') zone_latR0=deg2rad(s1);
				else zone_latR0= -1001.;
			    if(*s2 !='\0') zone_lonR0=deg2rad(s2);
				else zone_lonR0= -1001.;
			    if(*s3 !='\0') zone_latR1=deg2rad(s3);
				else zone_latR1= -1001.;
			    if(*s4 !='\0') zone_lonR1=deg2rad(s4);
				else zone_lonR1= -1001.;
			}
			break;
		    default:
			/* file with list of zones to be processed */
			strcpy(zone_file, argv[++i]);
			break;
		    }
		    z_mask_flag = 1;
		    break;
		default:
			PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
			usage();
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
			PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
			usage();
		default:
			PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
			usage();
			break;
		}
		break;
	default:
		strcpy(inp_file, argv[i]);
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s [-o out_map_file] [-nx nx] [-np nphi]\n",
myname);
PR(ER, "[-fp[c] [cats]ASCII_file_ps] [-count] [-fm ASCII_file_map] [-{gal,eq}(set.coor)]\n");
PR(ER, "[-zf file_with_zones] [-zo lat1,lon1,lat2,lon2(mask sky zone to remove)]\n");
PR(ER, "[-zc lat,lon] [-zr radius_in_degs] [-{K,R}] [-fb basic_fits-file]\n");
PR(ER, "[-{remove,keep} lat1,lon1,lat2,lon2(mask sky zone to remove/keep)]\n");
PR(ER, "[-pr(print ASCII)] [-fs formula] [-fe src_file] [-lay basic_fits_layout]\n");
PR(ER, "[-c (cos/sin-flag,cos default)] [-x(col.rand.pix.pat.)] [-l(set Legendre grid)] \n");
PR(ER, "[-ox x_out_file] [-ns nside] [-nf nside_factor] [-p(set equal dpi)]\n");
PR(ER, "[-N n_rand_points] [-rs random_seed] [-ng gaus_noise,K] [-nu uni_noise,K]\n");
PR(ER, "[-ni Poissonian_noise,K]\n");
PR(ER, "[-{double,float}] [-fc factor] [-glon glon0] [-glat glat0] [-ffb bfiles_list]\n");
PR(ER, "[-{double,float}] [-fc factor] [-glon glon0] [-glat glat0] [-ffb bfiles_list]\n");
PR(ER, "[-pin phi_incr] [-rsc[{12}](rescaling)] [-ph phi0(degs)] [-sq[x]]\n");
PR(ER, "[-y{123}{0123}{cs}] [-gr {A,N,S}] [-0(zero map)] [-1(unity map)]\n");
PR(ER, "[-li  a1,a2/b1,b2/c1,c2/... (set lim-s to)]\n");
PR(ST, "[-AtFlag] [-AeNumber] [-AiFlag]\n");
PR(ER, "---------------------------------------------------------------------\n");
PR(ER, "formats:\n");
PR(ER, "-fe:   ra_hh:mm:ss dec_ddd:mm:ss amp_mk min_ax_' max_ax_' x_ang_deg\n");
PR(ST, "-A:    eta theta[degs] phi[degs] primary delta_eta[degs] delta_phi[degs] comment\n");
exit(2);
}

void usage2()
{
PR(ST, "usage: %s [input_file] -o out_map_file FLAGS\n", myname);
PR(ST, "-----------------------------------------------------------------\n");
PR(ST, "-0 - creates zero map\n");
PR(ST, "-1 - creates unity map\n");
PR(ST, "-a[q] - reads data in equatorial coordinates (ra,dec,intensity) and\n");
PR(ST, "        puts on them sphere using averaging (see '-ag','-fm')\n");
PR(ST, "-ag - reads data in galactic coordinates (glat,glon,intensity) and\n");
PR(ST, "        puts on them sphere using averaging (see '-aq','-fm')\n");
PR(ST, "-ap - reads data in polar coordinates (p_ang,lon,intensity) and\n");
PR(ST, "        puts on them sphere using averaging (see '-aq','ag','-fm')\n");
PR(ST, "-count - when reading data from tables: use counts in pixels instead\n");
PR(ST, "         of intensity summing\n");
PR(ST, "-double - output data in double presion (default)\n");
PR(ST, "-eq  - set equatorial coordinates system for input table (see '-fp')\n");
PR(ST, "-fb basic_fits-file - read Basic FITS data and put them on a spere\n");
PR(ST, "                      (see '-glat', '-glon')\n");
PR(ST, "-fe src_file - read data from extened sour format containing:\n");
PR(ST, "               hh:mm:ss.s dd:mm:ss amp min_ax['] maj_ax['] pa[degs]\n");
PR(ST, "-ffb bfiles_list - list of Basic FITS file names like for '-fb' of format:  \n");
PR(ST, "                   filename  glon0  glat0. Overlaid data are averaged\n");
PR(ST, "-float  - output data in float presion\n");
PR(ST, "-fm ASCII_file_map - get map from table: theta, phi, temperature\n");
PR(ST, "-fp ASCII_file_ps - create map using format: xx:mm:ss xx:mm:ss amp\n");
PR(ST, "                    (see '-gal','-eq', default: -eq)\n");
PR(ST, "-fpc cats_ASCII_file_ps - put data on the from db CATS-format list\n");
PR(ST, "-fs formula  - create map from 'formula' containing expression of\n");
PR(ST, "               x=cos(theta), theta and phi\n");
PR(ST, "-gal - set Galactic coordinates system for input table (see '-fp')\n");
PR(ST, "-glat glat0 - galactic latitude  'glat0' to be centered the image\n");
PR(ST, "              from '-fb' (default it's taken from FITS-file data)\n");
PR(ST, "-glon glon0 - galactic longitude 'glon0' to be centered the image\n");
PR(ST, "              from '-fb' (default is's taken from FITS-file data)\n");
PR(ST, "-gr {A,N,S} - grid: quasiArea, same_Number_per_ring, Special_for_polar\n");
PR(ST, "-keep   lat1,lon1,lat2,lon2 - set mask for sky zone to be kept\n");
PR(ST, "-K - flag to keep selected zone on the figure (see '-zo','-zc','-zf')\n");
PR(ST, "-lay basic_fits_layout - select layout from 3D-FITS data for '-fb'\n");
PR(ST, "                         and '-ffb' (default lay=1)\n");
PR(ST, "-li  a1,a2/b1,b2/c1,c2/... - keep pixels with with values from given intervals\n");
PR(ST, "-ng noise_level,K - generates Gaussian noise in K\n");
PR(ST, "-ni noise_level,K - generates  Poissonian noise in K\n");
PR(ST, "-n[n,N]{gqp} - converts coor-s to pixel numbers or prints number(pa,l)\n");
PR(ST, "-np nphi - number of pixels in central ring for output map\n");
PR(ST, "-nu noise_level,K - generates uniform noise in K\n");
PR(ST, "-nx nx   - number of rings on a sphere for output map\n");
PR(ST, "-pr  - print map in ASCII format: pixel coordinates and value\n");
PR(ST, "-remove lat1,lon1,lat2,lon2 - set mask for sky zone to be removed\n");
PR(ST, "-R - flag to remove selected zone on the figure (see '-zo','-zc',-zf')\n");
PR(ST, "-v - print version of '%s'\n", myname);
PR(ST, "-y{123}{0123}{cs} - standard multipoles 1,2,3 {with modes 0,1,2,3} {cosine, sine}\n");
PR(ST, "-zc lat,lon - select zone center (see '-K', '-R', '-zr')\n");
PR(ST, "-zf file_with_zones - read zones from file for masking of format:\n");
PR(ST, "    lat1,lon1,lat1,lon1 fro rectangular or lat,lon,radius for circle\n");
PR(ST, "-zo lat1,lon1,lat2,lon2 - select zone by coordinates: default to remove\n");
PR(ST, "                          (see '-K','-R', '-remove', '-keep')\n");
PR(ST, "-zr radius_in_degs - select zone radius (see '-K', '-R', '-zc;)\n");
PR(ST, "-c  - cos/sin-flag (cos default) for '-y21', '-y22', '-y31' etc.\n");

PR(ST, "-x(col.rand.pix.pat.)\n");
PR(ST, "-l(set Legendre grid)\n");
PR(ST, "-ox x_out_file\n");
PR(ST, "-ns nside\n");
PR(ST, "-nf nside_factor\n");
PR(ST, "-p(set equal dpi)\n");
PR(ST, "-N n_rand_points\n");
PR(ST, "-rs random_seed\n");
PR(ST, "-fc factor\n");
PR(ST, "-pin phi_incr\n");
PR(ST, "-rsc[{12}](rescaling)\n");
PR(ST, "-ph phi0(degs)\n");
PR(ST, "-sq[x]\n");
PR(ST, "---------------------------------------------------------------------\n");
PR(ST, "-A        - just put all ALISE data on a map\n");
PR(ST, "-AtFlag   - set flag for ALISE data: Flag=1 -primary, =0 -secondary\n");
PR(ST, "-AeNumber - set Number of ALISE event to be mapped\n");
PR(ST, "-AiFlag   - set Flag to read data for infinity: 0-read without inf.,1-only inf.\n");
PR(ST, "-Ac       - set Flag to count number of events in pixel.\n");
PR(ST, "---------------------------------------------------------------------\n");
PR(ST, "formats:\n");
PR(ST, "-fe:   ra_hh:mm:ss dec_ddd:mm:ss amp_mk min_ax_' max_ax_' x_ang_deg\n");
PR(ST, "-A:    eta theta[degs] phi[degs] primary delta_eta[degs] delta_phi[degs] comment\n");
exit(0);
}

double yf(x, phi, s_flag)
double x, phi;
int s_flag;             /* if =1, cos(phi), else (=0) sin(phi) */
{
	double y;
	int sg = s_flag == 'c' ? -1 : 1;

	s_flag = s_flag == 'c' ? 0 : s_flag;
	switch (y_flag) {
	case 0:
		y = 0.0;
		break;
	case 1:
		y = 1.0;
		break;
	case 10:
		y = y10(x);
		break;
	case 11:
		y = !s_flag ? y11(x, phi, sg) :
		      (s_flag > 0 ? y11_cos(x, phi) : y11_sin(x, phi));
		break;
	case 20:
		y = y20(x);
		break;
	default :
	case 21 :
		y = !s_flag ? y21(x, phi, sg) :
		      (s_flag > 0 ? y21_cos(x, phi) : y21_sin(x, phi));
		break;
	case 22:
		y = !s_flag ? y22(x, phi, sg) :
		      (s_flag > 0 ?  y22_cos(x, phi) : y22_sin(x, phi));
		break;
	case 31 :
		y = !s_flag ? y31(x, phi, sg) :
		      (s_flag > 0 ? y31_cos(x, phi) : y31_sin(x, phi));
		break;
	case 32:
		y = !s_flag ? y32(x, phi, sg) :
		      (s_flag > 0 ?  y32_cos(x, phi) : y32_sin(x, phi));
		break;
	case 33:
		y = !s_flag ? y33(x, phi, sg) :
		      (s_flag > 0 ?  y33_cos(x, phi) : y33_sin(x, phi));
		break;
	}
	return y;
}

double y21(x, phi, sg)
double x, phi;
int sg;
{
	double y, c = 0.7725484040463791; /* c = sqrt(15/8/pi) */
	y = -c * x * sqrt(1-x*x)* (cos (phi) + sg*sin(phi));
	return y;
}
double y21_cos(x, phi)
double x, phi;
{
	double y, c = 0.7725484040463791; /* c = sqrt(15/8/pi) */
	y = -c * x * sqrt(1-x*x) * cos (phi);
	return y;
}
double y21_sin(x, phi)
double x, phi;
{
	double y, c = 0.7725484040463791; /* c = sqrt(15/8/pi) */
	y = -c * x * sqrt(1-x*x) * sin (phi);
	return y;
}

double y22(x, phi, sg)
double x, phi;
int sg;
{
	double y, c = 0.3862742020231896; /* c = 0.25 * sqrt (15/2/pi) */
	y = c * (1 - x*x) * (cos (2.*phi) + sg* sin (2.*phi));
	return y;
}
double y22_cos(x, phi)
double x, phi;
{
	double y, c = 0.3862742020231896; /* c = 0.25 * sqrt (15/2/pi) */
	y = c * (1 - x*x) * cos (2.*phi);
	return y;
}
double y22_sin(x, phi)
double x, phi;
{
	double y, c = 0.3862742020231896; /* c = 0.25 * sqrt (15/2/pi) */
	y = c * (1 - x*x) * sin (2.*phi);
	return y;
}
double y20(x)
double x;
{
	double y, c = 0.6307831305050401; /* c = sqrt (5/4/pi) */
	y = c * (1.5 *x*x - 0.5);
	return y;
}
double y10(x)
double x;
{
	double y, c = 0.4886025119029199; /* c = sqrt (3/4/pi) */
	y = c * x;
	return y;
}
double y11(x, phi, sg)
double x, phi;
int sg;
{
	double y, c = 0.3454941494713355; /* c = sqrt (3/8/pi) */
	y = -c * sqrt(1 - x*x) * (cos (phi) + sg*sin (phi));
	return y;
}
double y11_cos(x, phi)
double x, phi;
{
	double y, c = 0.3454941494713355; /* c = sqrt (3/8/pi) */
	y = -c * sqrt(1 - x*x) * cos (phi);
	return y;
}
double y11_sin(x, phi)
double x, phi;
{
	double y, c = 0.3454941494713355; /* c = sqrt (3/8/pi) */
	y = -c * sqrt(1 - x*x) * sin (phi);
	return y;
}

double y30(x)
double x;
{
	double y, c = 0.3731763325901154; /* c = sqrt(7/16/pi) */
	y = c * x * (5*x*x-3);
	return y;
}
double y31(x, phi, sg)
double x, phi;
int sg;
{
	double y, c = 0.3231801841141507; /* c = 3/8*sqrt(7/3/pi) */
	double x2=x*x, s = sqrt(1-x2);
	y = -c * s * (5*x2-1) * (cos(phi)+sg*sin(phi));
	return y;
}
double y31_sin(x, phi)
double x, phi;
{
	double y, c = 0.3231801841141507; /* c = 3/8*sqrt(7/3/pi) */
	double x2=x*x, s = sqrt(1-x2);
	y = -c * s * (5*x2-1) * sin(phi);
	return y;
}
double y31_cos(x, phi)
double x, phi;
{
	double y, c = 0.3231801841141507; /* c = 3/8*sqrt(7/3/pi) */
	double x2=x*x, s = sqrt(1-x2);
	y = -c * s * (5*x2-1) * cos(phi);
	return y;
}
double y32(x, phi, sg)
double x, phi;
int sg;
{
	double y, c = 1.0219854764332825; /* c = 15*sqrt(7/4/pi/120) */
	double phi2 = phi*2;
	y = c * (1-x*x) * x * (cos(phi2)+sg*sin(phi2));
	return y;
}
double y32_sin(x, phi)
double x, phi;
{
	double y, c = 1.0219854764332825; /* c = 15*sqrt(7/4/pi/120) */
	double phi2 = phi*2;
	y = c * (1-x*x) * x * sin(phi2);
	return y;
}
double y32_cos(x, phi)
double x, phi;
{
	double y, c = 1.0219854764332825; /* c = 15*sqrt(7/4/pi/120) */
	double phi2 = phi*2;
	y = c * (1-x*x) * x * cos(phi2);
	return y;
}
double y33(x, phi, sg)
double x, phi;
int sg;
{
	double y, c = 0.4172238236327840; /* c = 15*sqrt(7/4/pi/720) */
	double phi3 = phi*3, x2=x*x;
	y = -c * (1-x2) * sqrt(1-x2) * (cos(phi3)+sg*sin(phi3));
	return y;
}
double y33_sin(x, phi)
double x, phi;
{
	double y, c = 0.4172238236327840; /* c = 15*sqrt(7/4/pi/720) */
	double phi3 = phi*3, x2=x*x;
	y = -c * (1-x2) * sqrt(1-x2) * sin(phi3);
	return y;
}
double y33_cos(x, phi)
double x, phi;
{
	double y, c = 0.4172238236327840; /* c = 15*sqrt(7/4/pi/720) */
	double phi3 = phi*3, x2=x*x;
	y = -c * (1-x2) * sqrt(1-x2) * cos(phi3);
	return y;
}


/* read rectangular area of dadat */
void read_uniform_area(fp_in, sq_area, mx, my)
FILE *fp_in;
double **sq_area;
int *mx, *my;
{
    char str[161], st1[50], st2[50];
    int i = 0, j = 0, n=0;
    double y_prev = 0.0;

    /* detect: how many strings and rows are */
    while (fgets(str, 160, fp_in) != NULL) {
	    double x, y, t;
	    if (*str == '#' || *str == '\n' || *str == '\r')
		    continue;

	    /* simple map */
	    sscanf(str, "%s %s %lf", st1, st2, &t);
	    y = atof(st1);
	    x = atof(st2);

	    /* when unchangeble y */
	    if ( n == 0 )
		{ y_prev = y; j++; }
	    else {
		double vv;
		if (fabs(y) > 1.e-20)  vv = fabs((y-y_prev)/y);
		else vv = fabs(y-y_prev);

		if (vv > 1.e-16) { /* y != 0 */
		    /* relative value is not 0, then get next string */
		    if (fabs((y-y_prev)/y) > 1.e-16 ) { /* y was changed */
			*mx = j; /* so, this is a number rows */
			i++;     /* go to next string, and change a number */
			j = 0;   /* calc from begining N of rows */
		    } else {
			j++;     /* change number of row */
		    }
		}
		y_prev = y;  /* remember previos state of y-value */
	    }
	    n++; /* total number of strings */
    }
    *my = i; /* number of strings */

    /* contrlolling multiplication */
    if (*mx * *my != n) {
	PR(ER, "%s: different number of rows in strings,\n", myname);
	PR(ER, "because (nx=%d) * (ny=%d) != (n_total=%d).\n", *mx, *my, n);
	exit(1);
    }

    /* return FILE pointer back */
    fseek(fp_in, 0l, 0);

    /* take memory for array */
    if((*sq_area = (double*)calloc((unsigned)(*mx * *my), sizeof **sq_area))
	== (double *)0) {
	    PR(ER, "%s: no memory for temporal array...\n", myname);
	    exit(1);
    }

    i = j = n = 0;
    /* and now read data to array */
    while (fgets(str, 160, fp_in) != NULL) {
	    double t;
	    if (*str == '#' || *str == '\n' || *str == '\r')
		    continue;

	    /* simple map */
	    sscanf(str, "%s %s %lf", st1, st2, &t);

	    n = i * *mx + j;
	    (*sq_area)[n] = t;

	    j++;
	    if (j == *mx) {j=0; i++;}
    }

    return;
}

/* recalculate rectangular area to the GLESP map using interpolation */
void sq_area2glesp(sq_area, mx, my, f, x, nx,np,nph, flag)
double *sq_area;
int mx, my;
double *f, *x;
int nx, np, *nph;
int flag;  /* when =1, coor is theta; when =2, coor is cos(theta) */
{
    register int i, j;
    double *tt = (double *)0, *aa = (double *)0;  /* temporal arrays     */
    double *tx = (double *)0, *ax = (double *)0,  /* and temporal arrays */
	   *work = (double *)0;
    double th1 = flag-1 ? acos(x[0]) : x[0],
	   th2 = flag-1 ? acos(x[nx-1]) : x[nx-1],
	   phi1=0.0, phi2=2.*M_PI, dt;

    i = np * my;
    if ((tt = (double *)calloc((unsigned)i, sizeof *tt)) == (double *)0) {
nomem:
	PR(ER, "%s: no memory for temporal array...\n", myname);
	exit(1);
    }
    i = np;
    if ((tx =(double *)calloc((unsigned)i, sizeof *tx)) == (double *)0)
	goto nomem;
    if((work = (double *)calloc(i, sizeof *work)) == (double *)0)
	goto nomem;

    i = mx > my ? mx : my;
    if ((aa = (double *)calloc((unsigned)i, sizeof *aa)) == (double *)0)
	goto nomem;
    if ((ax = (double *)calloc((unsigned)i, sizeof *ax)) == (double *)0)
	goto nomem;

    /* prepare arrays with coordinates for intrepolation */
    dt = (th2-th1)/my;
    if (flag == 1) {
	for (i=0; i<nx; i++)    tx[i] = acos(x[i]);
    } else {
	for (i=0; i<nx; i++)    tx[i] = x[i];
    }
    for (i=0; i<my; i++)    ax[i] = th1 + i * dt;

    /* spline interpolation within rows */
    for (j=0; j<mx; j++) {
	for (i=0; i<my; i++)    aa[i] = sq_area[i*mx+j];   /* real column */

	/* used for splint */
	spline (ax, aa, my, 2.e30, 2.e30, work);
	for(i=0; i<nx; i++) {         /* interpolate from my points to nx */
	    double v;
	    /* used with spline */
	    splint (ax, aa, work, my, tx[i], &v);
	    tt[i*mx+j] = v;
	}
    }

    dt = (phi2-phi1)/mx;
    for (j=0; j<mx; j++)  ax[i] = phi1 + i * dt;
    /* interpolate within strings */
    for(i=0; i<nx; i++) {
	double tp = 2*M_PI / nph[i];
	for (j=0; j<mx; j++) aa[i] = tt[i*mx+j];   /* column values */
	/* used for splint */
	spline (ax, aa, my, 2.e30, 2.e30, work);
	for(j=0; j<nph[i]; j++) {   /* interpolate from my points to nx */
	    double v;
	    /* used with spline */
	    splint (ax, aa, work, my, tp*j, &v);
	    f[i*np+j] = v;           /* final array */
	}
    }

    /* free memory */
    if (tt != (double *)0)    free((void *)tt);
    if (tx != (double *)0)    free((void *)tx);
    if (aa != (double *)0)    free((void *)aa);
    if (ax != (double *)0)    free((void *)ax);
    if (work != (double *)0)  free((void *)work);

    return;
}
