#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PR  fprintf
#define ER  stderr

extern char *myname;

/*
 * subroutine for calculation Minkowski functionals of the I and II types
 * reproduced from the F77 program of Dmitrij I. Novikov
 *
 * called from 'difmap'-program
 * for GLESP maps: VO (SAO), 09/2011
 *
 */

void minkowski_func(ff, pw, x, nx, np0, nph, mf1, mf2, mf_name)
double *ff, /* GLESP map    */
       *pw, /* weights      */
       *x;  /* x=cos(theta) */
int nx, np0, *nph;  /* resolution parameters         */
double *mf1, *mf2;  /* Minkowski functionals 1 & 2, here are empty */
char *mf_name;      /* pattern name for output files */
{
    register int i, j, ia, ib;
    double *f = (double *)0, *fm = (double *)0, *fp = (double *)0,
	   *dlen = (double *)0;
    int np = np0 + 1, n1 = 1001, n2 = 201;
    int k;
    double pi = M_PI;
    double xx[5+1], yy[5+1];
    double dmean=0.0, disp=0.0, area=0.0;
    FILE *fd = (FILE *)0;
    char str[81], *pat = mf_name;
    double ddlen, dlength, hb, hh, ffm;

    pat = mf_name;
    if ((f = (double*)calloc((unsigned)(np*nx), sizeof *f)) == (double*)0) {
nomem:
	PR(ER, "%s: no memory for temporal array.\n", myname);
	exit(1);
    }
    if ((fm = (double*)calloc((unsigned)n1, sizeof *fm)) == (double*)0)
	    goto nomem;
    if ((fp = (double*)calloc((unsigned)n1, sizeof *fp)) == (double*)0)
	    goto nomem;
    if ((dlen = (double*)calloc((unsigned)n2, sizeof *dlen)) == (double*)0)
	    goto nomem;

    for (i=0; i<nx; i++) {
	for (j=0; j<nph[i]; j++) {
	       disp  += ff[i*np0+j]*ff[i*np0+j]*pw[i];
	       dmean += ff[i*np0+j]*pw[i];
	       area += pw[i];
	}
    }
    dmean /= area;
    disp = sqrt(disp/area);

    for(i=0; i<nx; i++) {
	for(j=0; j<nph[i]; j++)  f[i*np+j] = ff[i*np0+j] / disp;
	f[i*np+nph[i]] = f[i*np+0];
    }

    /*----- minkowski 2---------*/


    /* open output file */

    if (pat != NULL && *pat != '\0')
	sprintf(str, "%s_mk2", pat);
    else
	sprintf(str, "%s_mk2.dat", myname);

    if ((fd = fopen(str, "w")) == (FILE *)0) {
file_problem:
	PR(ER, "%s: problem to create file '%s'.\n", myname, str);
	exit(1);
    }


    /*  -100, 100 : level, aver^2 = 1 */

    ddlen = 0.0;
    hb = 0.07;

    for (ib=-100; ib<=100; ib++) {
	double b=ib*0.1;   /* step by level: from -10\sigma up to 10\sigma */
	dlength = 0.;
	int jl, iend, istep;

	for (jl=-1; jl<=1; jl+=2) {
	    if (jl == 1) {
		iend  = nx-2;
		istep = 1;
	    } else {
		iend  = 2;
		istep = -1;
	    }
	    for (i=(nx-1)/2; istep>0 ? i<=iend : i>=iend; i+=istep) {
		double hj=2. * pi / nph[i],
		      hj1=2.*pi / nph[i+jl],
		      y1 = acos(x[i]),
		      y2 = acos(x[i+jl]),
		      hv = y2 - y1,
		      ds = sqrt(1.-x[i]*x[i]);
		int ial = (i+jl)*np;
		ia  = i*np;

		for(j=0; j<nph[i]; j++) {
		    double ddx, phi, phi1, phi2, fij1, fij2;
		    int kn=0, kv=0, kl=0, kp=0, jj, jj1;
		    k=0;

		    /* ----(i,j):(i,j+1)----*/
		    if ((f[ia+j] >= b && f[ia+j+1] < b) ||
		      (f[ia+j] <= b && f[ia+j+1] > b)) {
			k++;   kn++;
			ddx = hj*(f[ia+j]-b)/(f[ia+j]-f[ia+j+1]);
			yy[k] = y1;
			xx[k] = hj*j+ddx;
		    }
		    /* ---(i+1,j):(i+1,j+1)-*/
		    phi = hj*j;
		    jj = (phi+0.0000000001)/hj1;
		    phi1 = hj1*jj;
		    phi2 = hj1*(jj+1);
		    fij1 = (phi-phi1) *(f[ial+jj+1]-f[ial+jj])/hj1+f[ial+jj];

		    if ((f[ial+jj] >= b && f[ial+jj+1] < b) ||
		      (f[ial+jj] <= b && f[ial+jj+1] > b)){
			  double pp1;
			  ddx=hj1*(f[ial+jj]-b)/(f[ial+jj]-f[ial+jj+1]);
			  pp1=hj1*jj+ddx;

			  if (pp1 >= hj*j && pp1 < hj*(j+1)) {
			     k++;  kv++;
			     yy[k] = y2;
			     xx[k] = pp1;
			  }
		    }
		    jj1 = jj;
		    /*------------------------------*/
		    phi = hj*(j+1);
		    jj = (phi+0.0000000001)/hj1;

		    phi1 = hj1*jj;
		    phi2 = hj1*(jj+1);
		    fij2=(phi-phi1)*(f[ial+jj+1]-f[ial+jj])/hj1+f[ial+jj];
		    if (jj1 != jj) {
			if ((f[ial+jj] >= b && f[ial+jj+1] < b) ||
			  (f[ial+jj] <= b && f[ial+jj+1] > b)) {
			    double pp2;
			    ddx = hj1*(f[ial+jj]-b)/(f[ial+jj]- f[ial+jj+1]);
			    pp2 = hj1*jj+ddx;

			    if (pp2 >= hj*j && pp2 < hj*(j+1)) {
				k++;  kv++;
				yy[k] = y2;
				xx[k] = pp2;
			    }
			}
		    }
		    /* --end of(i+1,j):(i+1,j+1)----*/

		    /*----------i,j+1:i,j-----------*/
		    if ((f[ia+j] >= b && fij1<b) || (f[ia+j] <=b && fij1 >b)){
			double ddy;
			k++; kl++;
			ddy = hv*(f[ia+j]-b)/(f[ia+j]-fij1);
			yy[k] = y1+ddy;
			xx[k] = hj*j;
		    }
		    /*------------------------------*/

		    /*----------i,j+1:i+1,j+1-------*/
		    if ((f[ia+j+1] >=b && fij2 <b)||(f[ia+j+1]<=b && fij2>b)){
			double ddy;
			k++;   kp++;
			ddy=hv*(f[ia+j+1]-b)/(f[ia+j+1]-fij2);
			yy[k] =y1+ddy;
			xx[k] =hj*(j+1);
		    }

		    /*------------------------------*/
		    /*------------k=2---------------*/
		    /*------------------------------*/
		    if (k == 2) {
			double xd = (xx[2]-xx[1])*ds,
			       yd =  yy[2]-yy[1];
			double ddlength = sqrt(xd*xd + yd*yd);
			dlength += ddlength;
		    }
		    /*----------end of k=2----------*/

		    /*------------------------------*/
		    /*------------k=4---------------*/
		    /*------------------------------*/
		    if (k == 4) {
			/*-----------kn=0-----------*/
			if (kn == 0) {
			    double xd = (xx[3]-xx[1])*ds,
				   yd =  yy[3]-yy[1];
			    double ddlength = sqrt(xd*xd + yd*yd);
			    dlength += ddlength;
			    xd = (xx[4]-xx[2])*ds,
			    yd =  yy[4]-yy[2];
			    ddlength = sqrt(xd*xd + yd*yd);
			    dlength += ddlength;
			}
			/*-----------kn=1-----------*/
			if (kn == 1) {
			    /*-------kn=1,kv=2------*/
			    if (kv == 2) {
				if (kl == 1) {
				    double xd = (xx[3]-xx[1])*ds,
					   yd =  yy[3]-yy[1];
				    double ddlength = sqrt(xd*xd + yd*yd);
				    dlength += ddlength;
				    xd = (xx[4]-xx[2])*ds,
				    yd =  yy[4]-yy[2];
				    ddlength = sqrt(xd*xd + yd*yd);
				    dlength += ddlength;
				} else {
				    double xd = (xx[2]-xx[1])*ds,
					   yd =  yy[2]-yy[1];
				    double ddlength = sqrt(xd*xd + yd*yd);
				    dlength += ddlength;
				    xd = (xx[4]-xx[3])*ds,
				    yd =  yy[4]-yy[3];
				    ddlength = sqrt(xd*xd + yd*yd);
				    dlength += ddlength;
				}
			    }
			    /*------kn=1,kv=1-------*/
			    if (kv == 1) {
				double xd = (xx[3]-xx[2])*ds,
				       yd =  yy[3]-yy[2];
				double ddlength = sqrt(xd*xd + yd*yd);
				dlength += ddlength;
				xd = (xx[4]-xx[1])*ds,
				yd =  yy[4]-yy[1];
				ddlength = sqrt(xd*xd + yd*yd);
				dlength += ddlength;
			    }
			}
		    }
		    /*---------end of k=4-----------*/
		}
	    }
	}
	dlen[ib+100] = dlength;
	ddlen += dlen[ib+100];
    }

    ddlen *= hb;
    /*  write: level, length in ...? */
    for(ib=-100; ib<=100; ib++) {
	double b=ib*hb;
	dlen[ib+100] /= ddlen;
	PR(fd, "%12.7e %12.7e\n", b, dlen[ib+100]);
    }
    fclose(fd);
    /*---------------end of minkowski 2-------------*/


    /*---------------minkowski 1--------------------*/

    /* open output file */

PR(ER, "\n%s: files '%s' and ", myname, str);
    if (pat != NULL && *pat != '\0')
	sprintf(str, "%s_mk1", pat);
    else
	sprintf(str, "%s_mk1.dat", myname);

    if ((fd = fopen(str, "w")) == (FILE *)0)  goto file_problem;
PR(ER, "'%s' have been created.\n\n", str);

    hh = 0.01;

    for(i=0; i<nx; i++) {
	for(j=0; j<nph[i]; j++) {
	      double ffp = f[i*np+j];
	      int ii = fabs(f[i*np+j])/hh;

	      if (ii <= 1000 && ffp <= 0)    fm[ii] += pw[i]/area;

	      if (ii <= 1000 && ffp > 0 )    fp[ii] += pw[i]/area;
	}
    }
    ffm=0.;
    for (i=1000; i>=0; i--) {
	double dnu = -(i+0.5)*hh;
	ffm += fm[i];

	/*  write: level, I Minkowski funcional, part1 */
	PR(fd, "%12.7e %12.7e\n", dnu, ffm);
    }
    for (i=0; i<=1000; i++) {
	    double dnu = (i+0.5)*hh;
	    ffm += fp[i];
	    /*  write: level, I Minkowski funcional, part2 */
	    PR(fd, "%12.7e %12.7e\n", dnu, ffm);
    }
    fclose(fd);

    free((void *)f);
    free((void *)fm);
    free((void *)fp);
    free((void *)dlen);

    return;
}



void minkowski_func_sq(ff, pw, x, nx, np0, nph,
	lat1,lon1,lat2,lon2,   mf1, mf2, mf_name)
double *ff, /* GLESP map    */
       *pw, /* weights      */
       *x;  /* x=cos(theta) */
int nx, np0, *nph;  /* resolution parameters         */
double lat1,lon1,lat2,lon2;
double *mf1, *mf2;  /* Minkowski functionals 1 & 2, here are empty */
char *mf_name;      /* pattern name for output files */
{
    register int i, j, ia, ib;
    double *f = (double *)0, *fm = (double *)0, *fp = (double *)0,
	   *dlen = (double *)0;
    int np = np0 + 1, n1 = 1001, n2 = 201;
    int k;
    double pi = M_PI, pi2 = 2.*pi;
    double xx[5+1], yy[5+1];
    double dmean=0.0, disp=0.0, area=0.0;
    FILE *fd = (FILE *)0;
    char str[81], *pat = mf_name;
    double ddlen, dlength, hb, hh, ffm;
    int i_lat1, i_lat2;
    extern int ind_x(int, double *, double);

    /* limits in index by latitude */
    i_lat1 = ind_x(nx, x, cos(lat1));
    i_lat2 = ind_x(nx, x, cos(lat2));
    if (i_lat1 > i_lat2) {i = i_lat2; i_lat2 = i_lat1; i_lat1 = i;}

    /* check and correct longitude values */
    if (lon1 < 0. && lon2 < 0.) {lon1 += pi2; lon1 += pi2;}
    if (lon1 > lon2) {hb = lon2; lon2 = lon1; lon1 = hb;}

    pat = mf_name;
    if ((f = (double*)calloc((unsigned)(np*nx), sizeof *f)) == (double*)0) {
nomem:
	PR(ER, "%s: no memory for temporal array.\n", myname);
	exit(1);
    }
    if ((fm = (double*)calloc((unsigned)n1, sizeof *fm)) == (double*)0)
	    goto nomem;
    if ((fp = (double*)calloc((unsigned)n1, sizeof *fp)) == (double*)0)
	    goto nomem;
    if ((dlen = (double*)calloc((unsigned)n2, sizeof *dlen)) == (double*)0)
	    goto nomem;

    for (i=0; i<nx; i++) {
	double dphi = pi2/nph[i];
	if (i < i_lat1) continue;
	if (i > i_lat2) break;

	for (j=0; j<nph[i]; j++) {
	    double aphi = j*dphi;
	    if (lon1 >= 0.) {
		if (aphi < lon1) continue;
		if (aphi > lon2) break;
	    } else {
		/* lon1 < 0. && lon2 > 0. */
		if (!(aphi > lon1+pi2 || aphi < lon2)) continue;
	    }
	    disp  += ff[i*np0+j]*ff[i*np0+j]*pw[i];
	    dmean += ff[i*np0+j]*pw[i];
	    area += pw[i];
	}
    }
    dmean /= area;
    disp = sqrt(disp/area);

    for(i=0; i<nx; i++) {
	for(j=0; j<nph[i]; j++)  f[i*np+j] = ff[i*np0+j] / disp;
	f[i*np+nph[i]] = f[i*np+0];
    }

    /*----- minkowski 2---------*/


    /* open output file */

    if (pat != NULL && *pat != '\0')
	sprintf(str, "%s_mk2", pat);
    else
	sprintf(str, "%s_mk2.dat", myname);

    if ((fd = fopen(str, "w")) == (FILE *)0) {
file_problem:
	PR(ER, "%s: problem to create file '%s'.\n", myname, str);
	exit(1);
    }


    /*  -100, 100 : level, aver^2 = 1 */

    ddlen = 0.0;
    hb = 0.07;

    for (ib=-100; ib<=100; ib++) {
	double b=ib*0.1;   /* step by level: from -10\sigma up to 10\sigma */
	dlength = 0.;
	int jl, iend, istep;

	for (jl=-1; jl<=1; jl+=2) {
	    if (jl == 1) {
		iend  = nx-2;
		istep = 1;
	    } else {
		iend  = 2;
		istep = -1;
	    }
	    for (i=(nx-1)/2; istep>0 ? i<=iend : i>=iend; i+=istep) {
		double dphi = pi2/nph[i];
		double hj=2. * pi / nph[i],
		      hj1=2.*pi / nph[i+jl],
		      y1 = acos(x[i]),
		      y2 = acos(x[i+jl]),
		      hv = y2 - y1,
		      ds = sqrt(1.-x[i]*x[i]);
		int ial = (i+jl)*np;
		ia  = i*np;

		if (i < i_lat1) continue;
		if (i > i_lat2) break;

		for(j=0; j<nph[i]; j++) {
		    double ddx, phi, phi1, phi2, fij1, fij2;
		    int kn=0, kv=0, kl=0, kp=0, jj, jj1;
		    double aphi = j*dphi;
		    k=0;

		    if (lon1 >= 0.) {
			if (aphi < lon1) continue;
			if (aphi > lon2) break;
		    } else {
			/* lon1 < 0. && lon2 > 0 */
			if (!(aphi > lon1+pi2 || aphi < lon2)) continue;
		    }

		    /* ----(i,j):(i,j+1)----*/
		    if ((f[ia+j] >= b && f[ia+j+1] < b) ||
		      (f[ia+j] <= b && f[ia+j+1] > b)) {
			k++;   kn++;
			ddx = hj*(f[ia+j]-b)/(f[ia+j]-f[ia+j+1]);
			yy[k] = y1;
			xx[k] = hj*j+ddx;
		    }
		    /* ---(i+1,j):(i+1,j+1)-*/
		    phi = hj*j;
		    jj = (phi+0.0000000001)/hj1;
		    phi1 = hj1*jj;
		    phi2 = hj1*(jj+1);
		    fij1 = (phi-phi1) *(f[ial+jj+1]-f[ial+jj])/hj1+f[ial+jj];

		    if ((f[ial+jj] >= b && f[ial+jj+1] < b) ||
		      (f[ial+jj] <= b && f[ial+jj+1] > b)){
			  double pp1;
			  ddx=hj1*(f[ial+jj]-b)/(f[ial+jj]-f[ial+jj+1]);
			  pp1=hj1*jj+ddx;

			  if (pp1 >= hj*j && pp1 < hj*(j+1)) {
			     k++;  kv++;
			     yy[k] = y2;
			     xx[k] = pp1;
			  }
		    }
		    jj1 = jj;
		    /*------------------------------*/
		    phi = hj*(j+1);
		    jj = (phi+0.0000000001)/hj1;

		    phi1 = hj1*jj;
		    phi2 = hj1*(jj+1);
		    fij2=(phi-phi1)*(f[ial+jj+1]-f[ial+jj])/hj1+f[ial+jj];
		    if (jj1 != jj) {
			if ((f[ial+jj] >= b && f[ial+jj+1] < b) ||
			  (f[ial+jj] <= b && f[ial+jj+1] > b)) {
			    double pp2;
			    ddx = hj1*(f[ial+jj]-b)/(f[ial+jj]- f[ial+jj+1]);
			    pp2 = hj1*jj+ddx;

			    if (pp2 >= hj*j && pp2 < hj*(j+1)) {
				k++;  kv++;
				yy[k] = y2;
				xx[k] = pp2;
			    }
			}
		    }
		    /* --end of(i+1,j):(i+1,j+1)----*/

		    /*----------i,j+1:i,j-----------*/
		    if ((f[ia+j] >= b && fij1<b) || (f[ia+j] <=b && fij1 >b)){
			double ddy;
			k++; kl++;
			ddy = hv*(f[ia+j]-b)/(f[ia+j]-fij1);
			yy[k] = y1+ddy;
			xx[k] = hj*j;
		    }
		    /*------------------------------*/

		    /*----------i,j+1:i+1,j+1-------*/
		    if ((f[ia+j+1] >=b && fij2 <b)||(f[ia+j+1]<=b && fij2>b)){
			double ddy;
			k++;   kp++;
			ddy=hv*(f[ia+j+1]-b)/(f[ia+j+1]-fij2);
			yy[k] =y1+ddy;
			xx[k] =hj*(j+1);
		    }

		    /*------------------------------*/
		    /*------------k=2---------------*/
		    /*------------------------------*/
		    if (k == 2) {
			double xd = (xx[2]-xx[1])*ds,
			       yd =  yy[2]-yy[1];
			double ddlength = sqrt(xd*xd + yd*yd);
			dlength += ddlength;
		    }
		    /*----------end of k=2----------*/

		    /*------------------------------*/
		    /*------------k=4---------------*/
		    /*------------------------------*/
		    if (k == 4) {
			/*-----------kn=0-----------*/
			if (kn == 0) {
			    double xd = (xx[3]-xx[1])*ds,
				   yd =  yy[3]-yy[1];
			    double ddlength = sqrt(xd*xd + yd*yd);
			    dlength += ddlength;
			    xd = (xx[4]-xx[2])*ds,
			    yd =  yy[4]-yy[2];
			    ddlength = sqrt(xd*xd + yd*yd);
			    dlength += ddlength;
			}
			/*-----------kn=1-----------*/
			if (kn == 1) {
			    /*-------kn=1,kv=2------*/
			    if (kv == 2) {
				if (kl == 1) {
				    double xd = (xx[3]-xx[1])*ds,
					   yd =  yy[3]-yy[1];
				    double ddlength = sqrt(xd*xd + yd*yd);
				    dlength += ddlength;
				    xd = (xx[4]-xx[2])*ds,
				    yd =  yy[4]-yy[2];
				    ddlength = sqrt(xd*xd + yd*yd);
				    dlength += ddlength;
				} else {
				    double xd = (xx[2]-xx[1])*ds,
					   yd =  yy[2]-yy[1];
				    double ddlength = sqrt(xd*xd + yd*yd);
				    dlength += ddlength;
				    xd = (xx[4]-xx[3])*ds,
				    yd =  yy[4]-yy[3];
				    ddlength = sqrt(xd*xd + yd*yd);
				    dlength += ddlength;
				}
			    }
			    /*------kn=1,kv=1-------*/
			    if (kv == 1) {
				double xd = (xx[3]-xx[2])*ds,
				       yd =  yy[3]-yy[2];
				double ddlength = sqrt(xd*xd + yd*yd);
				dlength += ddlength;
				xd = (xx[4]-xx[1])*ds,
				yd =  yy[4]-yy[1];
				ddlength = sqrt(xd*xd + yd*yd);
				dlength += ddlength;
			    }
			}
		    }
		    /*---------end of k=4-----------*/
		}
	    }
	}
	dlen[ib+100] = dlength;
	ddlen += dlen[ib+100];
    }

    ddlen *= hb;
    /*  write: level, length in ...? */
    for(ib=-100; ib<=100; ib++) {
	double b=ib*hb;
	dlen[ib+100] /= ddlen;
	PR(fd, "%12.7e %12.7e\n", b, dlen[ib+100]);
    }
    fclose(fd);
    /*---------------end of minkowski 2-------------*/


    /*---------------minkowski 1--------------------*/

    /* open output file */

PR(ER, "\n%s: files '%s' and ", myname, str);
    if (pat != NULL && *pat != '\0')
	sprintf(str, "%s_mk1", pat);
    else
	sprintf(str, "%s_mk1.dat", myname);

    if ((fd = fopen(str, "w")) == (FILE *)0)  goto file_problem;
PR(ER, "'%s' have been created.\n\n", str);

    hh = 0.01;

    for(i=0; i<nx; i++) {
	double dphi = pi2/nph[i];
	if (i < i_lat1) continue;
	if (i > i_lat2) break;

	for(j=0; j<nph[i]; j++) {
	    double ffp = f[i*np+j];
	    int ii = fabs(f[i*np+j])/hh;

	    double aphi = j*dphi;
	    if (lon1 >= 0.) {
		if (aphi < lon1) continue;
		if (aphi > lon2) break;
	    } else {
		/* lon1 < 0. && lon2 > 0 */
		if (!(aphi > lon1+pi2 || aphi < lon2)) continue;
	    }

	    if (ii <= 1000 && ffp <= 0)    fm[ii] += pw[i]/area;

	    if (ii <= 1000 && ffp > 0 )    fp[ii] += pw[i]/area;
	}
    }
    ffm=0.;
    for (i=1000; i>=0; i--) {
	double dnu = -(i+0.5)*hh;
	ffm += fm[i];

	/*  write: level, I Minkowski functional, part1 */
	PR(fd, "%12.7e %12.7e\n", dnu, ffm);
    }
    for (i=0; i<=1000; i++) {
	    double dnu = (i+0.5)*hh;
	    ffm += fp[i];
	    /*  write: level, I Minkowski functional, part2 */
	    PR(fd, "%12.7e %12.7e\n", dnu, ffm);
    }
    fclose(fd);

    free((void *)f);
    free((void *)fm);
    free((void *)fp);
    free((void *)dlen);

    return;
}

