#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr

extern void setopt(int, char **), usage(void);

int nside = 0, ntot=0;
int nx=0, np = 0;
double pixel_size = -1.;
char out_x_file[161];
char out_w_file[161];
char out_r_file[161];
char out_n_file[161];
char out_t_file[161];
int layer = 0;
int p_in = 'o';

char *myname="ntot";

int main(argc, argv)
int argc;
char **argv;
{
	register int i;
	int *nph = (int *)0;
	double *x = (double *)0, x1 = -1.0, x2 = 1.0;
	double *w;
	int nx2;
	double dt = 0., dp = 0.;
	double pi = M_PI;
	FILE *fp_out;
	extern int set_nphi_by_x2pin4(double *, int, int, int *);

	out_n_file[0] = out_x_file[0] = out_w_file[0] =
	     out_r_file[0] = out_t_file[0] = '\0';

	/* loop over external parameters */
	setopt(argc, argv);

	if (pixel_size > 0.) {
	    int nt;
	    dt = sqrt(pixel_size);
	    nx = pi / dt;
	    if ((nph = (int *)calloc(nx, sizeof *nph)) == (int *)0) {
nomem:
		    PR(ER, "%s: no memory for array...\n", myname);
		    exit(1);
	    }
	    if ((x = (double *)calloc(nx, sizeof *x)) == (double *)0)
		    goto nomem;

	    gauleg(x1, x2, x, (double *)0, nx);
	    nx2 = (nx+1)/2;
	    dt = fabs(x[nx2+1] - x[nx2-1]) / 2.;
	    if (!np)  np = 2*pi / dt + 0.5;
	    dp = 2*pi / np;
	    pixel_size = dt * dp;

	    switch (p_in) {
	    case 'o':
	    default :
		    nt = ntot = set_nphi_by_x2(x, nx, np, nph);
		    break;
	    case 'P':
		    nt = ntot = set_nphi_by_x2p(x, nx, np, nph);
		    break;
	    case  4 :
		    ntot = set_nphi_by_x2pin4(x, nx, np, nph);
		    nt = set_nphi_by_x2(x, nx, np, nph);
		    break;
	    case  0 :
		    ntot = nx * np;
		    nt = set_nphi_by_x2(x, nx, np, nph);
		    break;
	    }

	    nside = pow(2.,
		   (double)(int)(log(sqrt(nt/12.))/log(2.)+0.5))+0.5;
	} else if (nside > 0) {
	    ntot = nside * nside * 12;
	    pixel_size = 4*pi / ntot;
	    dt = sqrt(pixel_size);
	    nx = pi / dt;
	    if(!(nx%2))  nx++;
	    if ((nph = (int *)calloc(nx, sizeof *nph)) == (int *)0)
		    goto nomem;
	    if ((x = (double *)calloc(nx, sizeof *x)) == (double *)0)
		    goto nomem;

	    gauleg(x1, x2, x, (double *)0, nx);
	    nx2 = nx/2;
	    dt = fabs(x[nx2+1] - x[nx2-1]) / 2.;
//          if (!np) np = 2*pi / dt + 0.5;
	    if (!np) np = 2 * nx;
	    dp = 2*pi / np;
	    pixel_size = dt * dp;
	    ntot = set_nphi_by_x2(x, nx, np, nph);
	} else {
	    int nt;
	    if (nx == 0 && np  > 0)  nx = np / 2;
	    else if (nx >  0 && np == 0)  np = nx * 2;
	    else if (nx == 0 && np == 0)  {nx = 127; np = nx * 2;}

	    if ((nph = (int *)calloc(nx, sizeof *nph)) == (int *)0)
		    goto nomem;
	    if ((x = (double *)calloc(nx, sizeof *x)) == (double *)0)
		    goto nomem;

	    gauleg(x1, x2, x, (double *)0, nx);
	    nx2 = (nx+1)/2;
	    dt = fabs(x[nx2+1] - x[nx2-1]) / 2.;
	    if (!np) np = 2*pi / dt + 0.5;
	    dp = 2*pi / np;
	    pixel_size = dt * dp;

	    switch (p_in) {
	    case 'o':
	    default :
		    nt = ntot = set_nphi_by_x2(x, nx, np, nph);
		    break;
	    case 'P':
		    nt = ntot = set_nphi_by_x2p(x, nx, np, nph);
		    break;
	    case  4 :
		    ntot = set_nphi_by_x2pin4(x, nx, np, nph);
		    nt = set_nphi_by_x2(x, nx, np, nph);
		    break;
	    case  0 :
		    ntot = nx * np;
		    nt = set_nphi_by_x2(x, nx, np, nph);
		    break;
	    }

	    nside = pow(2.,
		   (double)(int)(log(sqrt(nt/12.))/log(2.)+0.5))+0.5;
	}

	if (p_in == 'o')
	    PR(ER,"GLESP:   quasi-equal pixel area is used.\n");
	else if (p_in == 0)
	    PR(ER,"GLESP:   equal pixel number in rings is used.\n");

	PR(ER,"GLESP:   pixel size=%g sq.min, side=%g, nx=%d, np=%d, ntot=%d\n",
		pixel_size * 60.*60.*180.*180./M_PI/M_PI,
		sqrt(pixel_size * 60.*60.*180.*180./M_PI/M_PI),
		nx, np, ntot);
	PR(ER,"HEALPix: pixel size=%g sq.min, side=%g, nside=%d, ntot=%d\n",
		4.*pi/(12.*nside*nside)  * 60.*60*180*180/M_PI/M_PI,
		sqrt(4.*pi/(12.*nside*nside)  * 60.*60*180*180/M_PI/M_PI),
		nside, 12*nside*nside);

	if (*out_n_file != '\0') {
	   if (strcmp(out_n_file, "-") == 0 ||
	       strcmp(out_n_file, "stdout") == 0 ) {
		   fp_out = stdout;
	   } else if ((fp_out = fopen(out_n_file, "w")) == (FILE *)0) {
		   PR(ER, "%s: can not create file '%s'\n",
			  myname, out_n_file);
		   exit(1);
	   }
	   for(i=0; i<nx; i++)
		PR(fp_out, "%d  %d\n", layer ? i-(nx+1)/2 : i, nph[i]);
	   if (fp_out != stdout)  fclose(fp_out);
	}
	if (*out_r_file != '\0') {
	   if (strcmp(out_r_file, "-") == 0 ||
	       strcmp(out_r_file, "stdout") == 0 ) {
		   fp_out = stdout;
	   } else if ((fp_out = fopen(out_r_file, "w")) == (FILE *)0) {
		   PR(ER, "%s: can not create file '%s'\n",
			  myname, out_r_file);
		   exit(1);
	   }
	   for(i=0; i<nx; i++) {
		  double pix_size_ratio;
		  double dth, pix_s, sth=sqrt(1.0-x[i]*x[i]);

		  if ( x[0] > 0. && i==0 ) {
		      dth =  (acos(x[i+1]) + acos(x[i])) / 2;
		  } else if ( x[0] < 0. && i==0)  {
		      dth = pi - (acos(x[i]) + acos(x[i+1]))/2;
		  } else if ( x[0] < 0. && i==nx-1 ) {
		      dth =  (acos(x[i]) + acos(x[i-1])) / 2;
		  } else if ( x[0] > 0. && i==nx-1 ) {
		      dth = pi - (acos(x[i]) + acos(x[i-1]))/2;
		  } else {
		      dth = (acos(x[i+1]) - acos(x[i-1])) / 2;
		  }

		  if (i==0 || i==nx-1) {
		      pix_s =  pi*2 * sth / nph[i] * fabs(dth)/2.;
		  } else {
		      pix_s =  pi*2 * sth / nph[i] * fabs(dth);
		  }
		  pix_size_ratio = pix_s / pixel_size;
		  PR(fp_out, "%d  %12.5e\n",
			layer ? i-(nx+1)/2 : i, pix_size_ratio);
	   }
	   if (fp_out != stdout)  fclose(fp_out);
	}
	if (*out_w_file != '\0') {
	   double *w0;
	   if (strcmp(out_w_file, "-") == 0 ||
	       strcmp(out_w_file, "stdout") == 0 ) {
		   fp_out = stdout;
	   } else if ((fp_out = fopen(out_w_file, "w")) == (FILE *)0) {
		   PR(ER, "%s: can not create file '%s'\n",
			  myname, out_w_file);
		   exit(1);
	   }
	   if ((w = (double *)calloc(nx, sizeof *w)) == (double *)0)
		   goto nomem;
	   if ((w0 = (double *)calloc(nx, sizeof *w0)) == (double *)0)
		   goto nomem;

	   /* calculate weights 'pw' */
	   weights(1, x1, x2, 2.*pi, x, w, nx, nph);

	   gauleg(x1, x2, x, w0, nx);

	   for(i=0; i<nx; i++)
		PR(fp_out, "%12.5e %12.5e %12.5e\n",
			x[i], w[i], w0[i]);

	   free((void *)w0);
	   free((void *)w);
	   if (fp_out != stdout)  fclose(fp_out);
	}
	if (*out_x_file != '\0') {
	   if (strcmp(out_x_file, "-") == 0 ||
	       strcmp(out_x_file, "stdout") == 0 ) {
		   fp_out = stdout;
	   } else if ((fp_out = fopen(out_x_file, "w")) == (FILE *)0) {
		   PR(ER, "%s: can not create file '%s'\n",
			  myname, out_x_file);
		   exit(1);
	   }
	   for(i=0; i<nx; i++)
		 PR(fp_out, "%d  %12.5e\n",  layer ? i-(nx+1)/2 : i, x[i]);
	   if (fp_out != stdout)  fclose(fp_out);
	}
	if (*out_t_file != '\0') {
	   if (strcmp(out_t_file, "-") == 0 ||
	       strcmp(out_t_file, "stdout") == 0 ) {
		   fp_out = stdout;
	   } else if ((fp_out = fopen(out_t_file, "w")) == (FILE *)0) {
		   PR(ER, "%s: can not create file '%s'\n",
			  myname, out_x_file);
		   exit(1);
	   }
	   for(i=0; i<nx; i++)
	      PR(fp_out, "%d  %12.5e  %12.5e\n",
		 layer ? i-(nx+1)/2 + 1 : i, acos(x[i]),
		 i==0 ? 0. : 60*180./M_PI*fabs(acos(x[i+1]) - acos(x[i])));
	   if (fp_out != stdout)  fclose(fp_out);
	}

	exit(0);
}

extern double str2deg(char *);

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'e':
			/* pixel side in arcmin */
			pixel_size = str2deg(argv[++i])*60;
			/* pixel size in sq.arcmin */
			pixel_size *= pixel_size;
			/* sq.arcmin -> sterad */
			pixel_size /= (60.*60*180*180/M_PI/M_PI);
			break;
		case 'g':
			if (strncmp(argv[i]+1, "gr", 2) == 0) {
			    i++;
			    if (argv[i][0] == 'a') {
				/* quasi-equal area pixel */
				p_in = 'o';  /* old scheme */
			    } else if (argv[i][0] == 'n') {
				/* equal pixel number in rings */
				p_in = 0; /* pin 0 */
			    }
			}
			if (strncmp(argv[i]+1, "gr", 2) == 0) {
			  if (argv[i][3] != '\0') {
			    int c = argv[i][3];
			    switch (c) {
			    case 'a': case 'A': p_in = 'o'; break;
			    case 'n': case 'N': p_in =   0; break;
			    case 's': case 'S': p_in =   4; break;
			    case 'i': goto lab1;
			    default :
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			    }
			  } else {
lab1:
			    i++;
			    if (argv[i][0] == 'a' || argv[i][0] == 'A') {
				/* quasi-equal area pixel */
				p_in = 'o';  /* old scheme */
			    } else if (argv[i][0]=='n' || argv[i][0]=='N') {
				/* equal pixel number in rings */
				p_in = 0; /* pin 0 */
			    } else if (argv[i][0]=='s' || argv[i][0]=='S') {
				/* equal pixel number in rings */
				p_in = 4; /* pin 4 */
			    }
			  }
			}

			break;
		case 'h':
			usage();
			break;
		case 'l':
			if (strncmp(argv[i]+1, "lma", 3) == 0) {
				nx = 2*atoi(argv[++i])+1;
			} else if (strncmp(argv[i]+1, "la", 2) == 0) {
				layer = 1;
			} else {
			    PR(ER, "%s: illegal flag '%s'\n",
				myname, argv[i]);
			    usage();
			}
			break;
		case 'n':
			switch(argv[i][2]) {
			case 'x':
				nx = atoi(argv[++i]);
				break;
			case 'p':
				np = atoi(argv[++i]);
				break;
			case 's':
				nside = atoi(argv[++i]);
				break;
			default:
				usage();
				break;
			}
			break;
		case 'o':
			switch(argv[i][2]) {
			case 'n':
				strcpy(out_n_file, argv[++i]);
				break;
			case 'r':
				strcpy(out_r_file, argv[++i]);
				break;
			case 't':
				strcpy(out_t_file, argv[++i]);
				break;
			case 'w':
				strcpy(out_w_file, argv[++i]);
				break;
			case 'x':
				strcpy(out_x_file, argv[++i]);
				break;
			default:
				usage();
				break;
			}
			break;
		case 'p':
			if (strncmp(argv[i]+1, "pi", 2) == 0) {
			    switch (argv [++i][0]) {
			    case 'o':
			    default :
				    p_in = 'o';
				    break;
			    case 'P':
				    p_in = 'P';
				    break;
			    case '0':
				    p_in = 0;
				    break;
			    case '4':
				    p_in = 4;
				    break;
			    }
			} else {
			    pixel_size = fabs(atof(argv[++i]));
			    /* sq.arcmin -> sterad */
			    pixel_size /= (60.*60*180*180/M_PI/M_PI);
			}
			break;
		case 'v':
			printf("%s: version 0.5 (31/08/2010), GLESP 2.0\n",
			    myname);
			exit(0);
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		usage();
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s [-nx nx] [-np np] [-ns nside] [-pin {0,P,o,4}]\n",
    myname);
PR(ER, "[-ox x_vector] [-ow weight_vect] [-on np_dims] [-or pix_ratio_vec]\n");
PR(ER, "[-la(layer instead of number)] [-lmax lmax(nx=lmax*2+1)]\n");
PR(ER, "[-p pixel_size(sq.min)] [-edge edge_{deg,min,sec}]\n");
exit(2);
}


/* subroutine of search for ':'-symbol in a string */
int semicol(s)
char *s;
{
   register char *cs,c;
   cs = s;
   while((c = *cs++))   if(c == ':')   return 1;
   return 0;
}

/* convert one of strings: hh:mm:ss.s, ss.s, Nm (N minutes),       */
/* mN (N min), hN,Nh (N hours), gN,Ng (defs), N' (N arcmin) to deg */
double str2deg(st)
char *st;
{
	double ang = 0.0;
	int sig=1;
	char *cs=st;
again:
	switch(*cs) {
	case '\0':
	case '\033':
		break;
	case ' ':
	case '+':
		cs++;
		goto again;
	case '-':
		sig = -1;
		cs++;
		goto again;
	case '\'':
	case 'm':
		ang = atof(cs+1) / 60.;
		break;
	case '\"':
	case 's':
		ang = atof(cs+1) / 3600. ;
		break;
	case 'h':
	case 'd':
	case 'g':
		ang = atof(cs+1);
		break;
	default :
		if(semicol(cs)) {  /* if symbol ':' */
		     long h=0, m=0;
		     float tang = 0.;

		     sscanf(cs,"%ld%*c%ld%*c%f",&h,&m,&tang);
		     ang = h + m / 60. + tang / 3600.;
		} else {
		      if( (cs[2]== 'h'&& (cs[3] != ' '&& cs[3] != '\0'))
			|| (cs[1]== 'h'&& (cs[2] != ' '&& cs[2] != '\0'))
			|| (cs[1]== 'd'&& (cs[2] != ' '&& cs[2] != '\0'))
			|| (cs[2]== 'd'&& (cs[3] != ' '&& cs[3] != '\0'))
			|| (cs[3]== 'd'&& (cs[4] != ' '&& cs[4] != '\0'))){
			    /* format:  HHhMMmSS.SS */
			    long h=0, m=0;
			    float ti = 0.;
			    sscanf(cs,"%ld%*c%ld%*c%f",&h,&m,&ti);
			    ang = (float)h + (float)m / 60. + ti / 3600.;
		      }else if((cs[2]== ' '&&(cs[3] !=' '&& cs[3] !='\0'))
			|| (cs[1]== ' '&& (cs[2] != ' '&& cs[2] != '\0'))
			|| (cs[3]== ' '&& (cs[4] != ' '&& cs[4] != '\0'))){
			    /* format:  "HH MM SS.SS" */
			    long h=0, m=0;
			    float ti = 0.;
			    sscanf(cs,"%ld%ld%f",&h,&m,&ti);
			    ang = (float)h + (float)m / 60. + ti / 3600.;
		      } else { /* if one of symbols at the end */
			      /* format: SS.SS{hdms'"} */
			      switch ( cs[strlen(cs)-1] )  {
			      case 'd':
			      case 'g':
			      case 'h':
				      ang = atof(cs);
				      break;
			      case '\'':
			      case 'm':
				      ang = atof(cs) / 60.;
				      break;
			      case '\"':
			      case 's':
				    {
				      char *qq;
				      int qq_key = 0;
				      qq = cs;
				      while(*qq) {
					    if (*qq == 'm' ||
						*qq == 'h' || *qq == 'd' ||
						*qq == '\'')
						    qq_key++;
					    qq++;
				      }
				      if (qq_key == 2) {
					    /* format:  "HHhMMmSS.SSs" */
					    long h=0, m=0;
					    float ti = 0.;
					    sscanf(cs,"%2ld%*c%2ld%*c%f",
						    &h,&m,&ti);
					    ang = (float)h + (float)m / 60.
						      + ti / 3600.;
				      } else {
					    ang = atof(cs) / 3600.;
				      }
				      break;
				    }
			      default:
				      ang = atof(cs) / 3600.;
				      break;
			      }
		      }
		}
	}
	return (sig*ang);
}
