#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#include "glesp.h"
#include "glespol.h"

#define PR      fprintf
#define ER      stderr
#define ST      stdout

extern void setopt(int, char **), usage(void), usage2(void);
extern int   qu2sp_n(double *, double *, double *, int, int,
	      double *, double *, double *, double *, double *, double *,
	      int, int, int *, double, double, int *);


char map_qfile[161],  map_ufile[161],
     outp_bfile[161], outp_efile[161],
     map_tfile[161],  outp_tfile[161];
char alm_hp_file[161]; /* file of alm, E- & B-modes in HEALPix format  */
char map3_file[161];   /* file with 3 maps          */
int  cut_lmin = -1;    /* cut value for a_lm        */
int  cut_lmax = 0;     /* cut value for a_lm        */
char cl_out_file[81], dl_out_file[81];
int  phi_deriv = 0;
double eps = 1.e-20;
double eps1 = 1.e-20,
       eps2 = 1.e-20;
//double eps1 = 1.e-12,
//       eps2 = 1.e-6;
int ffile_flag = 0;   /* default output in FITS-format */
int no_pol = 0,                /* if flag = 1, no polarization calc-s   */
    no_ani = 0,                /* if flag = 1, no anisotropy calc-s     */
    memory_separation=0;       /* separate calculattions of pol. & anis.*/
char tmp_dir[100];             /* directory for temporal files          */

int  sym_grid=1;    /* symmetric grid over x */
int  fill_0=0;      /* fill arrays by 0      */
double vnyq = 0.5;
int legendre_calc_scheme = 0;  /* if flag=-1, then direct calculations, */
			       /* else from largest M to lower          */
int N_pol1 = -1, N_pol2 = -1, N_pol3 = -1; /* number of pixels in P1,2,3 */

int naxis2=0;
int deriv = 1;
int lmax = -1;
int lmin = -1;
HEAD *ph, hd;
char *myname="polalm";

int main(argc, argv)
int argc;
char **argv;
{
    double *alm = (double *)0, *elm = (double *)0, *blm = (double *)0;
    double *f = (double *)0, *q = (double *)0, *u = (double *)0;
    int nx =0, np =0, *nph  = (int *)0;
    int nx2=0, np2=0, *nph2 = (int *)0;
    double *x = (double *)0, *x2 = (double *)0,
	   *w = (double *)0, *wqu = (double *)0;
    int *mx = (int *)0;
    int nbuf;
    FILE *fp_out = (FILE *)0;
    char tmp_a[100], tmp_e[100], tmp_b[100];
    char x_type[40], grid_type[40];
    int mem_sep_flag = 0;
    int flag_qu2sp = 0;
    extern void set_sig_fatal();
    extern void out_glesp_alm_T();

    map_tfile[0] = outp_tfile[0] = '\0';
    map_qfile[0] = map_ufile[0] = outp_bfile[0] = outp_efile[0] = '\0';
    alm_hp_file[0] = map3_file[0] = '\0';

    /* select grid output keyword values */
    strcpy(x_type, "COS(THETA)");
    strcpy(grid_type, "PHI(X)+LEGENDRE");
    strcpy(tmp_dir, "/tmp"); /* default directory for temporal files */
    cl_out_file[0] = dl_out_file[0] = '\0';

    /* control signal interruption */
    set_sig_fatal();

    /* loop over external parameters */
    setopt(argc, argv);

    if (glesp_comments < 0) {
	if (getenv("GLESP_COMMENTS") != NULL) {
	    char str[21];
	    strcpy(str,getenv("GLESP_COMMENTS"));
	    glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
	    if (glesp_comments) {
		 if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
		     glesp_comments = 0;
	    }
	} else
	    glesp_comments = 1;
    }

    /******************************/
    /* catch memory and read maps */
    /******************************/
    if (map3_file[0] != '\0') {
	in_glesp_3map(&f, &q, &u, &x,  &nx,  &np,  &nph,  map3_file);
mem_separ:
	if (memory_separation) { /* if no memory */
	    FILE *fp_tmp;
	    if (mem_sep_flag == 0) {  /* first stage */
		sprintf(tmp_a, "%s/polalm_aXXXXXX", tmp_dir);
		mktemp(tmp_a);

		if ((fp_tmp = fopen(tmp_a, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create temoral file '%s'\n",
			myname, tmp_a);
		    exit(1);
		}

		/* save alm-s to disk */
		outp_glesp_map(f, x, nx, np, nph, fp_tmp, x_type, grid_type);
		fclose(fp_tmp);

		free((void *)f);
		f = (double *)0;
		mem_sep_flag = 1;
	    } else if (mem_sep_flag == 1) { /* second stage */
		/* save B,E data after calculations */

		/* E-coeffcients temporal output in FITS-format */
		sprintf(tmp_e, "%s/polalm_eXXXXXX", tmp_dir);
		mktemp(tmp_e);
		if ((fp_tmp = fopen(tmp_e, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create temporal file '%s'\n",
			myname, tmp_e);
		    unlink(tmp_a);
		    exit(1);
		}
		out_glesp_alm_T(elm, lmax, nbuf, lmax, fp_tmp);
		fclose(fp_tmp);

		/* B-coeffcicients temporal output in FITS-format */
		sprintf(tmp_b, "%s/polalm_bXXXXXX", tmp_dir);
		mktemp(tmp_b);
		if ((fp_tmp = fopen(tmp_b, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create temporal file '%s'\n",
			myname, tmp_b);
		    unlink(tmp_a);
		    unlink(tmp_e);
		    exit(1);
		}
		out_glesp_alm_T(blm, lmax, nbuf, lmax, fp_tmp);
		fclose(fp_tmp);

		free((void *)q);
		free((void *)u);
		q = (double *)0;
		u = (double *)0;

		/* read saved T-map from disk */
		if((f = (double *)calloc((unsigned)(np*nx), sizeof *f))
		       == (double *)0)
			    goto nomem;
		read_glesp_map(&f, &x, &nx, &np, &nph, tmp_a);
		mem_sep_flag = 2;
	    }
	}
	/* calculation without anisotropy */
	if (no_ani) {
	    free((void *)f);
	    f = (double *)0;
	}
	/* calculation without polarization */
	if (no_pol) {
	    free((void *)q);
	    free((void *)u);
	    q = (double *)0;
	    u = (double *)0;
	}
    } else {
	if (map_qfile[0] != '\0' && map_ufile[0] != '\0') {
	    read_glesp_map(&q, &x,  &nx,  &np,  &nph,  map_qfile);
	    read_glesp_map(&u, &x2, &nx2, &np2, &nph2, map_ufile);

	    if (nx2 != nx  || np2 != np  || nph [nx/2-1] != nph2[nx/2-1]) {
dif_dims:
	       PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		      myname, map_qfile, map_ufile);
	       PR(ER, "Sorry...\n");
	       exit(1);
	    }
	    free((void *)x2);    x2 = (double *)0;
	    free((void *)nph2);  nph2 = (int *)0;
	}

	if (map_tfile[0] != '\0') {
	    read_glesp_map(&f, &x2, &nx2, &np2, &nph2, map_tfile);
	    if (q == (double *)0) {
		nx = nx2;
		np = np2;
		nph = nph2;
		x = x2;
	    }
	    if (nx2 != nx  || np2 != np  || nph [nx/2-1] != nph2[nx/2-1])
		goto dif_dims;
	    if (x2 != x)     free((void *)x2);
	    x2 = (double *)0;
	    if (nph2 != nph) free((void *)nph2);
	    nph2 = (int *)0;
	}

    }

    if (q == (double *)0 && u == (double *)0 && f == (double *)0) {
	PR(ER, "%s: no input files...\n", myname);
	PR(ER, "---------------------------------\n");
	usage();;
    }

    if (alm_hp_file[0] != '\0') {
	if (f == (double *)0 && !memory_separation) {
	  PR(ER,
	  "%s: can not contsruct file with 3 alms: no temperature data...\n",
	    myname);
	  PR(ER, "Use '-e' and '-b' flags for output\n");
	  exit(1);
	}
    }
    if ((w = (double *)calloc((unsigned)nx, sizeof *w)) == (double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays...\n", myname);
	    exit(1);
    }
    if ((wqu = (double *)calloc((unsigned)nx, sizeof *wqu)) == (double *)0)
		goto nomem;

    if (lmax < 0)
	lmax = nx / 2;

    /*********************/
    /* calculate weights */
    /*********************/
    weights_pol(0, 0, -1., 1., 2*M_PI, x, w, wqu, nx, nph, np);
    {
	register int i;
	if ((mx = (int *)calloc((unsigned)nx, sizeof *mx)) == (int *)0)
		goto nomem;
	for(i=0; i<nx; i++)  mx[i] = nph[i] * vnyq + 0.5;
    }

    nbuf = 2*(lmax+1);
    elm = q;
    blm = u;
    alm = f;

    /*************************************/
    /* calculation of E & B coefficients */
    /*************************************/
    flag_qu2sp = qu2sp_n(alm,elm,blm, lmax, nbuf, f,q,u,x,w,wqu,
	      nx,np,nph, eps1,eps2,mx);

    /* check free memory code */
    if (flag_qu2sp < 0) {
	if ( flag_qu2sp == -1 && f != (double *)0 ) {
	    PR(ER,
		"\n%s: I can not find enough memory for temporal arrays,\n",
		 myname);
	    PR(ER, "and continue calculation separately ");
	    PR(ER, "for anisotropy and Q,U-maps.\n");
	    memory_separation = 1;
	    goto mem_separ;
	} else {
	    PR(ER, "\n%s: sorry, ", myname);
	    PR(ER, " actually no memory for temporal arrays...\n");
	    unlink(tmp_a);
	    exit(1);
	}
    }

    /* return back and                         */
    /* calculate separately anisotropy and Q&U */
    if (mem_sep_flag == 1)  {
	if (glesp_comments)
	    PR(ER, "And anisotropy calculation now\n");
	goto mem_separ;
    }

    /* read Q,U-maps calculated earlier to unify in output file */
    if (mem_sep_flag == 2) {
	/* read polarization data from saved arrays from disk */
	extern int read_alm_table_T ();
	int nalm = 2 * (lmax+1);
	read_alm_table_T (tmp_e, &elm, &nalm, &nbuf, &lmax);
	read_alm_table_T (tmp_b, &blm, &nalm, &nbuf, &lmax);

    }


    /***************************/
    /*  output e- and b- alm   */
    /***************************/
    if (alm_hp_file[0] != '\0') {
	/* open output 3alm-file */
	if (alm_hp_file[0] == '\0' || strcmp(alm_hp_file, "stdout") == 0)
	    fp_out = stdout;
	else
	    if ((fp_out = fopen(alm_hp_file, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n", myname, alm_hp_file);
		exit(1);
	    }
	if (cut_lmax <= 1) cut_lmax = lmax;
	write_ieb_alm_T(alm, elm, blm, lmax, nbuf, cut_lmax, fp_out);

	/* close file */
	if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
    } else {
      if (elm != (double *)0) {
	cut_lmax = lmax;
	/* open output E-file */
	if (outp_efile[0] == '\0' || strcmp(outp_efile, "stdout") == 0)
	    fp_out = stdout;
	else
	    if ((fp_out = fopen(outp_efile, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n", myname, outp_efile);
		exit(1);
	    }
	/* write new a_{lm}-s */
	if (ffile_flag)
	    ff_write_alm(elm, nbuf, lmax, cut_lmax, fp_out);
	else
	    out_glesp_alm_T(elm, lmax, nbuf, cut_lmax, fp_out);
	/* close file */
	if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
	/*--------------------------------------------*/
      }
      if (blm != (double *)0) {
	cut_lmax = lmax;
	/* open output B-file */
	if (outp_bfile[0] == '\0' || strcmp(outp_bfile, "stdout") == 0)
	    fp_out = stdout;
	else
	    if ((fp_out = fopen(outp_bfile, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n", myname, outp_bfile);
		exit(1);
	    }
	/* write new a_{lm}-s */
	if (ffile_flag)
	    ff_write_alm(blm, nbuf, lmax, cut_lmax, fp_out);
	else
	    out_glesp_alm_T(blm, lmax, nbuf, cut_lmax, fp_out);
	/* close file */
	if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
      }

      /*****************/
      /*  output alm   */
      /*****************/
      if (map_tfile[0] != '\0' || outp_tfile[0] != '\0') {
	cut_lmax = lmax;
	  /* open output alm-file */
	  if (outp_tfile[0] == '\0' || strcmp(outp_tfile, "stdout") == 0)
	      fp_out = stdout;
	  else
	      if ((fp_out = fopen(outp_tfile, "w")) == (FILE *)0) {
		  PR(ER, "%s: can not create file '%s'\n",
		      myname, outp_tfile);
		  exit(1);
	      }
	  /* write new a_{lm}-s */
	  if (ffile_flag)
	      ff_write_alm(alm, nbuf, lmax, cut_lmax, fp_out);
	  else
	      out_glesp_alm_T(alm, lmax, nbuf, cut_lmax, fp_out);
	  /* close file */
	  if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);

	  /* output Cl-file */
	  if (*cl_out_file != '\0') {
		double *cl = (double *)0, *clp = (double *)0;
		/* allocate memory */
		if ((cl = (double*)calloc(lmax, sizeof *cl)) == (double*)0) {
		    PR(ER, "%s: no memory for spectrum - no spectrum.\n",
			myname);
		} else {
		    extern void alm2cl_T(double *, int, double *, int);
		    alm2cl_T(cl,lmax,alm,nbuf);
		    if ((fp_out = fopen(cl_out_file, "w")) == (FILE *)0) {
			PR(ER, "%s: can not ope file '%s - no spectrum.\n",
			    myname, cl_out_file);
		    } else {
			print_cl_table (fp_out, cl, clp, 0, lmax);
			fclose(fp_out);
		    }
		    free((void *)cl);
		}
	  }
	  /* output Dl-file */
	  if (*dl_out_file != '\0') {
		double *dl = (double *)0, *clp = (double *)0;
		/* allocate memory */
		if ((dl = (double*)calloc(lmax, sizeof *dl)) == (double*)0) {
		    PR(ER, "%s: no memory for spectrum - no spectrum.\n",
			myname);
		} else {
		    extern void alm2dl_T(double *, int, double *, int);
		    alm2dl_T(dl,lmax,alm,nbuf);
		    if ((fp_out = fopen(dl_out_file, "w")) == (FILE *)0) {
			PR(ER, "%s: can not ope file '%s - no spectrum.\n",
			    myname, dl_out_file);
		    } else {
			print_cl_table (fp_out, dl, clp, 0, lmax);
			fclose(fp_out);
		    }
		    free((void *)dl);
		}
	  }

	  free((void *)f);
      }
    }
    /*--------------------------------------------*/

    free((void *)q);
    free((void *)u);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    int eps_flag = 0;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'a':
			strcpy(outp_tfile, argv[++i]);
			break;
		case 'b':
			strcpy(outp_bfile, argv[++i]);
			break;
		case 'c':
		case 'C':
			strcpy(cl_out_file, argv[++i]);
			break;
		case 'd':
			if (strncmp(argv[i]+1, "do", 2) == 0) {
				bitpix_map = DOUBLEPIX;
			} else if (strncmp(argv[i]+1, "di", 2) == 0) {
				lmax = lmin = 1;
			} else if (strncmp(argv[i]+1, "dl", 2) == 0) {
				strcpy(dl_out_file, argv[++i]);
			} else {
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			}
			break;
		case 'D':
			strcpy(dl_out_file, argv[++i]);
			break;
		case 'f':
			if (strncmp(argv[i]+1, "fl", 2) == 0) {
				bitpix_map = FLOATPIX;
			} else {
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			}
			break;
		case 'e':
			if (strncmp(argv[i]+1, "ep1", 3) == 0) {
			      eps1 = atof(argv[++i]);
			} else if (strncmp(argv[i]+1, "ep2", 3) == 0) {
			      eps2 = atof(argv[++i]);
			} else if (strncmp(argv[i]+1, "ep", 2) == 0) {
			      /* accuracy for Legendre plynomials */
			      eps = atof(argv[++i]);
			      eps_flag = 1;
			} else {
			      strcpy(outp_efile, argv[++i]);
			}
			break;
		case 'h':
			if (strncmp(argv[i]+1,"h2", 2) == 0)
			    usage2();
			else
			    usage();
			break;
		case 'l':
			lmax = atoi (argv[++i]);
			break;
		case 'L':
			lmin = lmax = atoi (argv[++i]);
			break;
		case 'm':
			if (strncmp(argv[i]+1,"ms", 2) == 0) {
			    /* separation of calculation of anisotropy */
			    /* and polarization */
			    memory_separation = 1;
			} else {
			    PR(ER,"%s: illegal flag '%s'\n",
				myname, argv[i]);
			    usage();
			}
			break;
		case 'n':
			switch(argv[i][2]) {
			case 'o':
				if (strncmp(argv[i]+1, "nop", 3) == 0) {
				    no_pol = 1; /* no polarization calc */
				} else if (strncmp(argv[i]+1, "noa", 3)==0){
				    no_ani = 1; /* no anisotropy calc */
				} else {
				    PR(ER,"%s: illegal flag '%s'\n",
					myname, argv[i]);
				    usage();
				}
			case 'q':
				vnyq = atof(argv[++i]);
				break;
			default:
				usage();
				break;
			}
			break;
		case 'o':
			switch(argv[i][2]) {
			case 'e':
				if (strncmp(argv[i]+1, "oef", 3) == 0) {
				    ffile_flag = 1; /* output in F-format */
				}
				strcpy(outp_efile, argv[++i]);
				break;
			case 'b':
				if (strncmp(argv[i]+1, "oeb", 3) == 0) {
				    ffile_flag = 1; /* output in F-format */
				}
				strcpy(outp_bfile, argv[++i]);
				break;
			default:
				strcpy(alm_hp_file, argv[++i]);
				break;
			}
			break;
		case 'P':
			switch (argv[i][1]) {
			case 'P':
			default :
			    switch (argv[i][2]) {
			    default :
			    case '1':
				      N_pol1 = atoi(argv[++i]);
				      break;
			    case '2':
				      N_pol2 = atoi(argv[++i]);
				      break;
			    case '3':
				      N_pol3 = atoi(argv[++i]);
				      break;
			    }
			}

			break;
		case 'q':
			strcpy(map_qfile, argv[++i]);
			break;
		case 's':
			/* if flag=1, then direct calculations, */
			/* else from largest M to lower         */
			if (i < argc-1) {
			    if (argv[i+1][0] != '-') {
				i++;
				if(argv[i][0] == 'L' || argv[i][0] == '1') {
					legendre_calc_scheme = 1;
				} else
					legendre_calc_scheme = -1;
			    } else {
				legendre_calc_scheme = !legendre_calc_scheme;
			    }
			} else /* flag '-s' is the last parameter */
				legendre_calc_scheme = !legendre_calc_scheme;
			break;
		case 't':
			if (strncmp(argv[i]+1, "tmp", 3) == 0) {
			    strcpy(tmp_dir, argv[++i]); /* tmp-directory */
			} else
			    strcpy(map_tfile, argv[++i]);
			break;
		case 'u':
			strcpy(map_ufile, argv[++i]);
			break;
		case 'v':
			PR(ER, "%s: version 0.4 (22.03.2009), GLESP 2.0.\n",
			    myname);
			exit(0);
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		strcpy(map3_file, argv[i]);
		break;
	}
    }

    if (eps_flag) {
	eps1 = eps2 = eps;
    }
    return;
}

void usage()
{
PR(ER, "usage: %s [3map_file] [-o out_3alm_file] [-lmax lmax]\n", myname);
PR(ER, "[-q map_qfile -u map_ufile] [-e out_alm_efile -b out_alm_bfile]\n");
PR(ER, "[-t map_anis_file] [-a out_alm] [-Cl/Dl sp_file_out]\n");
PR(ER, "[-nop(no polarization)] [-noa(no anisotropy)] [-ms(memory_sep)]\n");
PR(ER, "[-nq nyq_coeff] [-ep eps] [-ep1 eps1] [-ep2 eps2] [-tmp tmp_dir]\n");
PR(ER, "[-{float,double}] [-s {M,L}] [-P{123} N{123}]\n");
exit(2);
}


void usage2()
{
PR(ST,"usage: %s [FLAGS]\n", myname);
PR(ST,"-----------------------------------------------------------------\n");
//PR(ST,"[-P{123} N{123}]\n");
PR(ST,"[3map_file] - input 3map-file (T,Q,U-maps)\n");
PR(ST,"-a out_alm - output file with a_lm coefficients (see '-t','-u','-e','-b')\n");
PR(ST,"-b out_alm_bfile - output file with B-mode coefficients (see '-e','-t')\n");
PR(ST,"-Cl sp_file_out - file for Cl-spectrum output (see '-Dl')\n");
PR(ST,"-Dl sp_file_out - file for Dl-spectrum output (see '-Cl')\n");
PR(ST,"-double - output file in 64-bit float format (default) (see '-float')\n");
PR(ST,"-e out_alm_efile - output file with E-mode coefficients (see '-b','-t')\n");
PR(ST,"-ep eps   - sets accuracy for temperature calculations (dflt: 1.e-20)\n");
PR(ST,"-ep1 eps1 - sets accuracy for E-mode calculations (dflt: 1.e-20)\n");
PR(ST,"-ep2 eps2 - sets accuracy for B-mode calculations (dflt: 1.e-20)\n");
PR(ST,"-float - output file in 32-bit float format (see '-double')\n");
PR(ST,"-lmax lmax - sets maximum l-multipole number to calculate\n");
PR(ST,"-ms - separates memory for temperature and polarization calculation\n");
PR(ST,"-noa - no calculation of temperature anisotropy for 3map-file\n");
PR(ST,"-nop - no calculation of polarization for 3map-file\n");
PR(ST,"-nq nyq_coeff - sets Nyquist frequency coefficient (default: 0.5)\n");
PR(ST,"-o out_3alm_file output 3alm-file (T-alm and E,B-modes)\n");
PR(ST,"-q map_qfile - input Q-pol. map file (see '-t','-u','-e','-b')\n");
PR(ST,"-s {M,L} - selects Legendre polynomial calculation scheme (dflt: M)\n");
PR(ST,"-t map_anis_file - input temperature anisotropy map file\n");
PR(ST,"-tmp tmpcat - changes temporal directory (default: /tmp)\n");
PR(ST,"-u map_ufile - input U-pol. map file (see '-t','-q','-e','-b')\n");
PR(ST,"-v - prints version of '%s'\n", myname);
exit(0);
}


