#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <string.h>

#include "glesp.h"
#include "glespol.h"
#define PR      fprintf
#define ER      stderr

#define MOD(a,p)   ((a) - (p) * (int)((a)/(p)))
#define ABS(x)     ((x) < 0   ? -(x) : (x))
#define MAX(x,y)   ((x) > (y) ?  (x) : (y))
#define MIN(x,y)   ((x) < (y) ?  (x) : (y))


extern int lmin_ylm_hpix(int, double);
extern void start_time(void), print_time(void);
extern double return_time(void);
int _lmin_;
double _enorm_;
extern int max_double_(void);

extern void a_lm_Pol(int *, int *, int *, int *, int *, int *, int *, int *,
      int *,int *,double *,double *,double *,double *,double *, int *,int *,
      double *,double *,double *,double *,double *,double *,
      double *,double *, int *, int *,
      double *,double *,double *,double *,
      double *,double *,double *,double *,double *,double *);


extern void alm_fast3_Pol(int *, int *, int *, int *, int *, int *, int *,
	double *, double *, double *, double *,double *, int *, int *,
	double *,double *,double *,double *,double *,double *);
extern void start_clm(int *, int *, double *, double *);
extern void lgndr_Mn(int *, double *, int *, int *, double *,
		     double *, double *, double *, int *);
extern void Temp_Mn(int, int, int, int, int *,
	   int,int,int,int,int, double,double * ,double * ,double *,double *);
extern void Pol_Mn(int, int, int, int, int, int,int, int *,
	       int,int,double,double *,double *,double *,double *,double *,
	       double *,double *);
extern void start_L0(int, int, int, double *);
extern void start_ll(int, int, int, double *, double *, double *, double *);
extern void lgndr_0(int, int, double, double *);
extern void lgndr_Ln(int,int,int,double,double,double *,double *,double *,
		double *, int *);
extern void Pll_n(int, int, int, double *,int,double,double,
		double *,int,double *,int *);
extern void Temp_Ln(int,int,int,int,int,int, int *,
		double,int,int,double *,double *,double *,double *,int *);
extern void Pol_Ln(int,int,int,int,int,int,int,int,double,int *,
       int *,double *,double *,double *,double *,double *,double *,double *);


/*
   input parameters: x(Nx) - coordinates on
   the sphere, f(Nx,Nph(ix)) - signal.
   Q(Nx,Nph(ix)) & U(Nx,Nph(ix)) - plarizations,
   Nph - number of cells vs. phi for a given x
   output: - f(l,m+1) - a_lm_c, f(l,Lmx+m+2) - a_lm_s,
   output: - Q&U(l,m+1) - real a_lm, Q&U(l,Lmx+m+2) - imag a_lm,
   a((l,-m)=(-1)^m a(l,m)*
 ------------------------------------------------------------ *
 totaly zero, because of memory allocation with 'calloc' in C
 */


void a_lm_Pol(Nmx,Nmp,Jmx,Lmn,Lmx,Nx,Ksym,Kaml,Ktst,
      K_temp,eps1, eps2, x, sqx, f,Nph, Mmax,
      wsave,cf,glm, c,abc,cc,
      apm,bpm, Mlim,
      K_pol, cpL, cpM, acL,
      xlbd, Q,U,qc,uc,qs,us)
int *Nmx, *Nmp, *Jmx, *Lmn, *Lmx, *Nx, *Nph;
int *Ksym, *Kaml, *Ktst, *K_temp;
double *eps1, *eps2;
register double *x, *sqx, *f;
int *Mmax;
double *wsave, *cf;
double *glm, *c, *abc, *cc;
double *apm, *bpm;
int *Mlim;
int *K_pol;
double *cpL, *cpM, *acL;
register double *Q, *U, *qc, *uc, *qs, *us;
double *xlbd;
{
    register int i, ii, j;
    register int nx = *Nmx, nb = *Nmp;
    register double nul = 0.0;
    _lmin_ = max_double_();
    _enorm_ = pow(10., (double)_lmin_);   /* 10^Lmin */
    int NTlm=0, NTeps=0, NPlm=0, NPeps=0;
    int lmin = *Lmn == 1 ? 2 : *Lmn;


    if (glesp_comments) {
	start_time();
	PR(ER, "Fourier transform:    ");
    }
//  PR(ER, "alm-fast3\n");
    alm_fast3_Pol(Nmx,Nmp,Nx,Jmx,Lmx,Mmax,Nph,f,
	  cf,apm,bpm,wsave,K_temp,K_pol,Q,U,qc,uc,qs,us);

    if (glesp_comments)
	PR(ER, "%gs\n", return_time());
//  print_time();

    for(i=ii=0; i< nx; i++, ii+=nb) {
	for(j=0; j< nb; j++) {
	    register int ij = ii+j;
	    if (f != (double *)0)  f[ij]=nul;
	    if (Q != (double *)0)
		    Q[ij]=U[ij]=nul;
	}
    }

    /* include dipole */
    if(*Lmn == 1 && f != (double *)0) {
	int nbuf = 2 * (*Lmx + 1);
	extern void alm_dipole_T(int, int, int, int, int, double *,
		 double *, double *, double *);
	alm_dipole_T(*Nmx, nbuf, *Lmx, *Ksym, *Nx, x,apm,bpm,f);

//        f[1]       *= -sqrt(0.5);
//        f[nbuf+1]          *= -0.5;
//        f[nbuf*(*Lmx+2)+1] *= -0.5;
	f[1]               *= -1.;
	f[nbuf+1]          *= -1.;
	f[nbuf*(*Lmx+2)+1] *= -1.;
    }


    if (glesp_comments)
	PR(ER, "Legendre polynomials: ");
    if(*Kaml > 0) {
//      PR(ER, "alm-lgndr-L, Kaml=%d\n", *Kaml);
	extern void alm_lgndr_LL();
	alm_lgndr_LL(Nmx,Nmp,Jmx,&lmin,Lmx,Ksym,Ktst,K_temp,
	     eps1,eps2,Nx,x,sqx,apm,bpm,cc,abc,glm,f,Mmax,
	     &NTlm,&NTeps,&NPlm,&NPeps,cpL,acL,
	     K_pol,Q,U,qc,uc,qs,us,xlbd);
    }
    if(*Kaml < 0) {
	extern void alm_lgndr_Mu(), alm_lgndr_M();
//        PR(ER, "alm-lgndr-M, Kaml=%d\n", *Kaml);

	alm_lgndr_M(Nmx,Nmp,Jmx,&lmin,Lmx,Ksym,Ktst,K_temp,
	    eps1,eps2,Nx,x,sqx,apm,bpm,glm,c,f,Mmax,
	    &NTlm,&NTeps,&NPlm,&NPeps,cpM,
	    Mlim,K_pol,Q,U,qc,uc,qs,us,xlbd);

    }  /*  Kml<0  */

    if (glesp_comments)
	PR(ER, "%gs\n", return_time());
//  print_time();

// PR(ER, "NTlm=%d, NTeps=%d, NPlm=%d, NPeps=%d\n", NTlm, NTeps, NPlm, NPeps);

    return;
}

/*
C   loop over l - SUM_l[a_{lm}P_l^m(x_p)]
C   input: x(Nx),Nph(Jmx),apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
C   ouput: f(Nmx,Nmp)

      IMPLICIT REAL*8(a-h,o-y)  
      DIMENSION f(Nmx,Nmp),x(Nmx),sqx(Nmx) 
      DIMENSION Mmax(Nmx)
      DIMENSION apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
      DIMENSION Q(Nmx,Nmp),U(Nmx,Nmp)
      DIMENSION qc(0:Jmx,Nmx),qs(0:Jmx,Nmx)
      DIMENSION uc(0:Jmx,Nmx),us(0:Jmx,Nmx)
      DIMENSION glm(Jmx+1),xlbd(2*Jmx+3),c(2*Jmx+2)
      DIMENSION cpM(4*Jmx,Jmx+2)


*/
void alm_lgndr_M(Nmx,Nmp,Jmx,Lmn,Lmx,Ksym,Ktst,K_temp,
	      eps1,eps2,Nx,x,sqx,apm,bpm,glm,c,f,Mmax,
	      NTlm,NTeps,NPlm,NPeps,cpM,
	      Mlim,K_pol,Q,U,qc,uc,qs,us,xlbd)
int *Nmx, *Nmp, *Jmx, *Lmn, *Lmx, *Ksym, *Ktst, *K_temp;
double *eps1, *eps2, *x, *sqx, *apm, *bpm, *glm, *c, *f;
int *Nx, *Mmax;
int *NTlm, *NTeps, *NPlm, *NPeps;
double *cpM;
int *Mlim, *K_pol;
double *Q, *U, *qc, *uc, *qs, *us, *xlbd;
{
    register int m, mb, ix, jx, L0, lmx1 = *Lmx+1, nx1 = *Nx - 1;
    double pi=M_PI;
    double cnorm=1. / sqrt(2.*pi);
    double Enrm= pow(10., (double)*Mlim);
    int Kx = *Nx;
    int ax = 1;
    int nbuf = (*Lmx + 1) * 2;
    extern void Temp_Mn_T(), Plm_n1b(), Pol_Mn_T(), renorm_T();

    if(*Ksym > 0) Kx=(*Nx+1)/2;

    for (m=0; m<= *Lmx; m++) {   /* for all m */
	 double xc, cmm;
	 double eps = *eps2;
	 int mv =m;

	 if(m < 100) eps = *eps1;
	 mb = lmx1 + m;
	 L0=m;
	 if(L0 < *Lmn) L0 = *Lmn;

	 start_clm(Lmx, &mv, &cmm, c);

//         if(*K_pol > 0)
//               ax=1-2*MOD((L0+m),2);
	       ax = 1;

	 for (ix=0; ix<Kx; ix++) {  /* for all x(ix) */
	    if(m > Mmax[ix])  continue;

	    xc=x[ix];
	    if(*Ksym > 0) {
		jx = nx1 - ix;
		if(jx <= ix) jx = -100;
	    } else
		jx = -100;

	    lgndr_Mn(Mlim, &Enrm, Lmx, &mv, &cmm, c, &xc, glm, NTlm);

	    if(*K_temp > 0)
//                  Temp_Mn(*Nmx, *Nmp, *Jmx,L0,NTeps,
//                           *Lmx,m,mb,ix,jx, eps,glm,apm,bpm,f);
		    Temp_Mn_T(*Nmx, nbuf, *Jmx,L0,NTeps,
			     *Lmx,m,mb,ix,jx, eps,glm,apm,bpm,f);

	    if(*K_pol > 0) {
	       int aax=ax;      /* for L>2 and all m */
	       extern void Plm_n1();
//             if(m == 1)  aax = -ax;
//             Plm_n(*Jmx, *Lmx,L0,m, cpM, &aax,xc,glm,Kxf,xlbd,NPlm);
//             Plm_n1(*Jmx, *Lmx,L0,m, cpM, &aax,xc,glm, xlbd,NPlm);
	       Plm_n1b(*Jmx, *Lmx,L0,m, cpM, &aax,xc,glm, xlbd,NPlm);

//             Pol_Mn(*Nmx, *Nmp, *Jmx,L0,*Lmx,m,mb,NPeps,
//                       ix,jx,eps,xlbd,qc,qs,Q,uc,us,U);
	       Pol_Mn_T(*Nmx, nbuf, *Jmx,L0,*Lmx,m,mb,NPeps,
			 ix,jx,eps,xlbd,qc,qs,Q,uc,us,U);
	    }        /*   for K_pol>0 */
	 }            /* for all x(ix) */
	 renorm_T(*Nmx,nbuf,L0,*Lmx,m,mb,cnorm,*K_pol, f,Q,U, *Ktst);

    }       /* over all m */

    return;
}

/*
 normalization and record of alm
 */
void renorm_T(Nmx,nbuf,L0,Lmx,m,mb,cnorm,K_pol, f,Q,U,Ktst)
register int Nmx, nbuf, Lmx;
int L0, m, mb;
double cnorm;
int K_pol;
double *f, *Q, *U;
int Ktst;
{
    register int l;
    register int sign = 1;
    double *alm_i = f + (m+Lmx+1)*nbuf,
	   *alm_r = f + m*nbuf;
    double *Qi = Q + (m+Lmx+1)*nbuf,
	   *Qr = Q + m*nbuf;
    double *Ui = U + (m+Lmx+1)*nbuf,
	   *Ur = U + m*nbuf;

    for(l=L0; l <= Lmx;  l++) {
	if (f != (double *)0) {
	    alm_i[l] *= -cnorm;
	    alm_r[l] *=  cnorm;
	}
	if(K_pol > 0 && l > 1) {
	    double q_lr = Qr[l], q_li = Qi[l],
		   u_lr = Ur[l], u_li = Ui[l];
	    double xsc, xps, xss, xpc;
	    if (sign > 0)  {
		xsc = (q_lr + q_li);
		xps = (q_li - q_lr);
		xss = (u_lr - u_li);
		xpc = (u_lr + u_li);
	    } else {
		xsc =  (q_lr - q_li);
		xps = -(q_li + q_lr);
		xss =  (u_lr + u_li);
		xpc =  (u_lr - u_li);
	    }
	    Qr[l] = xsc * cnorm;
	    Qi[l] = xss * cnorm;
	    Ur[l] = xpc * cnorm;
	    Ui[l] = xps * cnorm;

	    sign = -sign;
	}    /*  K_pol>0 */
    }   /*  over L   */

    return;
}

void renorm(Nmx,Nmp,L0,Lmx,m,mb,cnorm,K_pol, f,Q,U,Ktst)
register int *Nmx, *Nmp, *Lmx;
int L0, m, mb;
register double cnorm;
int *K_pol;
register double *f, *Q, *U;
int *Ktst;
{
    register int l, ll;
    register int sign = 1;

    for(l=L0,ll=*Nmp*L0;  l <= *Lmx;  l++,ll+=*Nmp) {
	register int lr=ll+m, li=ll+mb;
	if (f != (double *)0) {
	    register double *f_li = f + li, *f_lr = f + lr;
	    *f_li *= -cnorm;
	    *f_lr *=  cnorm;
	}
	if(*K_pol > 0 && l > 1) {
	    register double *q_lr = Q + lr, *q_li = Q + li,
			    *u_lr = U + lr, *u_li = U + li;
	    register double xsc, xps, xss, xpc;
	    if (sign > 0)  {
		xsc = (*q_lr + *q_li);
		xps = (*q_li - *q_lr);
		xss = (*u_lr - *u_li);
		xpc = (*u_lr + *u_li);
	    } else {
		xsc =  (*q_lr - *q_li);
		xps = -(*q_li + *q_lr);
		xss =  (*u_lr + *u_li);
		xpc =  (*u_lr - *u_li);
	    }
//            double xsc = (Q[lr]+sign*Q[li]);
//            double xps = (sign*Q[li]-Q[lr]);
//            double xss = (U[lr]-sign*U[li]);
//            double xpc = (U[lr]+sign*U[li]);
//            Q[lr] = xsc*cnorm;
//            Q[li] = xss*cnorm;
//            U[lr] = xpc*cnorm;
//            U[li] = xps*cnorm;

	      *q_lr = xsc * cnorm;
	      *q_li = xss * cnorm;
	      *u_lr = xpc * cnorm;
	      *u_li = xps * cnorm;

	    sign = -sign;
	}    /*  K_pol>0 */
    }   /*  over L   */

    return;
}

void Temp_Mn_T(Nmx,nbuf,Jmx,L0,Nt,Lmx,m,mb,
		    ix,jx,eps,glm,apm,bpm,f)
int Nmx, nbuf, Jmx, Lmx, *Nt;
register int  L0, m, mb, ix, jx;
double eps;
double *glm, *apm, *bpm, *f;
{
    register int l;
    int l0m = L0 + m;
    int ax=1-2*MOD(l0m,2);
    register int mm = m * Nmx, mi=mm+ix, mj=mm+jx;
    register double *f_r = f + m  * nbuf,
		    *f_i = f + mb * nbuf;
    register double a_n = apm[mi], a_s = apm[mj],
		    b_n = bpm[mi], b_s = bpm[mj];

    for(l=L0; l <= Lmx; l++) {    /* for all m<L<Lmx */
	register double gl1 = glm[l+1];

	if(ABS(gl1) > eps) {
	  (*Nt)++;

	    f_r[l] += gl1 * a_n;
	    f_i[l] += gl1 * b_n;
	    if(jx > 0)  {     /* for symmetric x */
		if (ax < 0) {
		    f_r[l] -= gl1 * a_s;
		    f_i[l] -= gl1 * b_s;
		} else {
		    f_r[l] += gl1 * a_s;
		    f_i[l] += gl1 * b_s;
		}
	    }      /* for symmetric x   */
	 }         /* for glm< eps      */
	 ax = -ax;
    }            /* for all m<L< Lmx  */
    return;
}
void Temp_Mn(Nmx,Nmp,Jmx,L0,Nt,Lmx,m,mb,
		    ix,jx,eps,glm,apm,bpm,f)
int Nmx, Nmp, Jmx, Lmx, *Nt;
register int  L0, m, mb, ix, jx;
double eps;
double *glm, *apm, *bpm, *f;
{
    register int l, ll=Nmp*L0;
    int l0m = L0 + m;
    int ax=1-2*MOD(l0m,2);
    register int mm = m * Nmx, mi=mm+ix, mj=mm+jx;
    register double *f_r = f + m, *f_i = f + mb;
    register double a_n = apm[mi], a_s = apm[mj],
		    b_n = bpm[mi], b_s = bpm[mj];

    for(l=L0; l <= Lmx; l++, ll += Nmp) {    /* for all m<L<Lmx */
	register double gl1 = glm[l+1];

	if(ABS(gl1) > eps) {
	    double *ff_r =  f_r + ll, *ff_i = f_i + ll;
	  (*Nt)++;

	    *ff_r += gl1 * a_n;
	    *ff_i += gl1 * b_n;
	    if(jx > 0)  {     /* for symmetric x */
		if (ax < 0) {
		    *ff_r -= gl1 * a_s;
		    *ff_i -= gl1 * b_s;
		} else {
		    *ff_r += gl1 * a_s;
		    *ff_i += gl1 * b_s;
		}
	    }      /* for symmetric x   */
	 }         /* for glm< eps      */
	 ax = -ax;
    }            /* for all m<L< Lmx  */
    return;
}

void Pol_Mn_T(Nmx,nbuf,Jmx,L0,Lmx,m,mb, Np,
	    ix,jx,eps,xlbd,qc,qs,Q,uc,us,U)
int Nmx, nbuf, Jmx, Lmx, *Np;
int  L0, m, mb, ix, jx;
double eps;
register double *xlbd, *qc, *qs, *uc, *us, *Q, *U;
{
    register int l;
    int ax=1-2*MOD((L0+m),2);
    register int mm = m * Nmx, mi = mm + ix, mj = mm + jx;
    double qci = qc[mi], qsi = qs[mi], uci = uc[mi], usi = us[mi],
	   qcj=0., qsj=0., ucj=0., usj=0.;
    register double *xr = xlbd,
		    *xi = xlbd + Jmx+1;
    double *Qr = Q + m * nbuf, *Qi = Q + mb * nbuf;
    double *Ur = U + m * nbuf, *Ui = U + mb * nbuf;

    if(jx > 0) {
	   qcj = qc[mj], qsj = qs[mj], ucj = uc[mj], usj = us[mj];
    }

    for (l=L0; l<=Lmx; l++) {  /* for all m<L<Lmx */
	double xll= xr[l], xlb= xi[l];

	if(ABS(xll) > eps || ABS(xlb) > eps) {
	  (*Np)++;
	    Qr[l] += xll * (qci + usi);  /* sc-ps */
	    Ur[l] += xll * (uci - qsi);  /* ss+pc */
	    Qi[l] += xlb * (qci - usi);  /* sc+ps */
	    Ui[l] += xlb * (qsi + uci);  /* pc-ss */
	    if(jx > 0)  {    /* for symmetric x */
		if (ax < 0) {
		      xll = -xll;
		      xlb = -xlb;
		}
		Qr[l] += xlb * (qcj + usj);   /* sc-ps */
		Ur[l] += xlb * (ucj - qsj);   /* ss+pc */
		Qi[l] += xll * (qcj - usj);   /* sc+ps */
		Ui[l] += xll * (qsj + ucj);   /* pc-ss */
	    }
	}           /* for xlbd< eps  */
    }            /* for all m<L< Lmx */

    return;
}

void Pol_Mn(Nmx,Nmp,Jmx,L0,Lmx,m,mb, Np,
	    ix,jx,eps,xlbd,qc,qs,Q,uc,us,U)
int Nmx, Nmp, Jmx, Lmx, *Np;
int  L0, m, mb, ix, jx;
double eps;
register double *xlbd, *qc, *qs, *uc, *us, *Q, *U;
{
    register int l;
    int ax=1-2*MOD((L0+m),2);
    int llp= Nmp*L0;
    register int mm = m * Nmx, mi = mm + ix, mj = mm + jx;
    register
	     double qci = qc[mi], qsi = qs[mi], uci = uc[mi], usi = us[mi],
		    qcj = qc[mj], qsj = qs[mj], ucj = uc[mj], usj = us[mj];
    register double *xr = xlbd + L0,
		    *xi = xlbd + L0 + Jmx+1;

    for (l=L0; l<=Lmx; l++) {  /* for all m<L<Lmx */
	register double xll= *xr;
	register double xlb= *xi;
	register int lr = llp+m, li = llp+mb;

	if(ABS(xll) > eps || ABS(xlb) > eps) {
	  (*Np)++;
	    Q[lr] += xll * (qci + usi);  /* sc-ps */
	    U[lr] += xll * (uci - qsi);  /* ss+pc */
	    Q[li] += xlb * (qci - usi);  /* sc+ps */
	    U[li] += xlb * (qsi + uci);  /* pc-ss */
	    if(jx > 0)  {    /* for symmetric x */
		if (ax < 0) {
		      xll = -xll;
		      xlb = -xlb;
		}
		Q[lr] += xlb * (qcj + usj);   /* sc-ps */
		U[lr] += xlb * (ucj - qsj);   /* ss+pc */
		Q[li] += xll * (qcj - usj);   /* sc+ps */
		U[li] += xll * (qsj + ucj);   /* pc-ss */
	    }
	}           /* for xlbd< eps  */
	llp += Nmp;
	xr++; xi++;
    }            /* for all m<L< Lmx */

    return;
}

/*
C   loop over l - SUM_l[a_{lm}P_l^m(x_p)]
C   input: x(Nx),Nph(Jmx),apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
C   ouput: f(Nmx,Nmp)
      IMPLICIT REAL*8(a-h,o-z)  
      DIMENSION f(Nmx,Nmp),x(Nmx),sqx(Nmx),xlbd(2*Jmx+3)
      DIMENSION Mmax(Nmx)
      DIMENSION apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
      DIMENSION Q(Nmx,Nmp),U(Nmx,Nmp)
      DIMENSION qc(0:Jmx,Nmx),qs(0:Jmx,Nmx)
      DIMENSION uc(0:Jmx,Nmx),us(0:Jmx,Nmx)

C ------------------------------------------
      DIMENSION abc(4*Jmx),cc(4*Jmx+4)
C         abc=wsave(2*Nmp+15),cc=cf(Nmp)
      DIMENSION glm(Jmx+1)  
c      DIMENSION e(3*Jmx+3),glm(Jmx+1)  
c      EQUIVALENCE (ee,e),(ee(3*Jmx+4),glm)
      DIMENSION cpL(3*Jmx*(Jmx+1)),acL(2*Jmx*(Jmx+1))
 */
void alm_lgndr_LL(Nmx,Nmp,Jmx,Lmn,Lmx,Ksym,Ktst,K_temp,
	  eps1,eps2,Nx,x,sqx,apm,bpm,cc,abc,glm,f,Mmax,
	  NTlm,NTeps,NPlm,NPeps,cpL,acL,
	  K_pol,Q,U,qc,uc,qs,us,xlbd)
int *Nmx, *Nmp, *Jmx, *Lmn, *Lmx, *Ksym, *Ktst, *K_temp;
int *Nx, *Mmax;
double *eps1, *eps2, *x, *sqx, *apm, *bpm, *cc, *abc, *glm, *f;
int *NTlm, *NTeps, *NPlm, *NPeps, *K_pol;
double *cpL, *acL,
	  *Q, *U, *qc, *uc, *qs, *us, *xlbd;
{
    register int l, m, ix, jx;
    int Kx = *Nx, Kxf, nx1 = *Nx - 1;
    int mb;
    double pi=M_PI, eps, cll;
    double cnorm=1./sqrt(2.*pi);

    if(*Ksym > 0)  Kx=(*Nx+1)/2;

    mb = *Lmx + 1;
    jx = -1;
    Kxf = 1;

    start_L0(*Jmx, *Lmn, *Lmx, cc);

    for (l= *Lmn; l<= *Lmx; l++) {  /* for all L */
      int ax;

      if(l < 100)
	    eps = *eps1;
      else
	    eps = *eps2;

      ax = 1-2*MOD(l,2);  /* ax(L=2)=1, ax(L=3)=-1,... */

      start_ll(*Jmx, *Lmx,l, abc, cc, acL, &cll);

      for (ix=0; ix<Kx; ix++) { /* for all x(ix) */
	    int Km, aax;
	    double xc = x[ix];
	    double sqxc = sqx[ix];

	    if(*Ksym > 0) jx = nx1-ix;
	    if(jx < ix)  jx = -100;

	    lgndr_0(*Jmx,*Lmx,xc,cc);

	    /*P_L^m,0<m<Km-1*/
	    lgndr_Ln(*Jmx,*Lmx,l,xc,sqxc,abc,cc,glm, &cll,NTlm);

	    Km=Mmax[ix]+1;
	    if(l < Mmax[ix]) Km=l+1;

	    aax = -ax;
	    if(*K_temp > 0)
		   Temp_Ln(*Nmx,*Nmp,*Jmx,mb,l,Km,
			&aax,eps,ix,jx,glm,apm,bpm,f,NTeps);

	    if(*K_pol > 0) {
	       aax = -ax;
	       Pll_n(*Jmx,Km,l,cpL,aax,xc,sqxc,glm,Kxf,xlbd,NPlm);
	       aax = -1;
	       Pol_Ln(*Nmx,*Nmp,*Jmx,l,Km,mb,ix,jx,eps,&aax,
		       NPeps,xlbd,qc,qs,Q,uc,us,U);
	    }                   /* for K_pol */
      }                         /* for x(ix) */
    }                           /* over  L=Lmn:Lmx */

    /* normalization and record of alm */
    for(m=0; m<=l; m++) { /* for m=0:L */
       int mm = mb+m;
       renorm(Nmx,Nmp,*Lmn,Lmx,m,mm,cnorm,K_pol,
		     f,Q,U,Ktst);
    }   /* over m=0:L */
    return;
}

/*
   loop over l - SUM_l[a_{lm}P_l^m(x_p)]
   input: x(Nx),Nph(Jmx),apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
   ouput: f(Nmx,Nmp)
*/
void Temp_Ln(Nmx,Nmp,Jmx,mb,l,Km,ax,eps,
			 ix,jx,glm,apm,bpm,f, Nt)
int Nmx, Nmp, Jmx;
int mb, l, ix,jx, Km;
double eps;
int *ax;
register double *glm, *apm, *bpm, *f;
int *Nt;
{
    register int m1;
    register int ll = l * Nmp;
    register double *fl = f + ll;

    for (m1=1; m1<=Km; m1++) {
	register int m=m1-1, mix = m * Nmx + ix, mjx = m * Nmx + jx;
	register double gl = glm[m1];
	if(ABS(gl) > eps) {
	    register double *f_r = fl + m,
			    *f_i = fl + m + mb;
//           (*Nt)++;
	    *f_r += gl * apm[mix];
	    *f_i += gl * bpm[mix];
	    if(jx > 0) {
	      if (*ax < 0) {
		    *f_r += gl * apm[mjx];
		    *f_i += gl * bpm[mjx];
	      } else {
		    *f_r -= gl * apm[mjx];
		    *f_i -= gl * bpm[mjx];
	      }
	      *ax = -(*ax);
	    }
	}       /* for glm< eps */
    }               /* over all m    */
    return;
}

void Pol_Ln(Nmx,Nmp,Jmx,l,Lm, mb,ix,jx,eps, ax,
		      Np, xlbd,qc,qs,Q,uc,us,U)
int Nmx, Nmp, Jmx, Lm;
int l,mb,ix,jx;
double eps;
int *ax;
double *xlbd, *qc, *qs, *Q, *uc, *us, *U;
int *Np;
{
    register int m, jmx1 = Jmx+1;
    int ll = l * Nmp;

    for(m=0; m< Lm; m++) {   /* for all m<L<Lmx */
	register int mx= jmx1+m;
	register double xr =  xlbd[m+1];
	register double xi =  xlbd[mx];
	int m1=m; /* +1 */
	int mm=mb+m;
	int mi = m * Nmx + ix;

	if(ABS(xr) > eps) {
//          (*Np)++;
	    Q[ll+m1] += xr  * (qc[mi] + us[mi]);   /* sc-ps */
	    U[ll+m1] += xr  * (uc[mi] - qs[mi]);   /* ss+pc */
	    if(jx > 0)  {              /* for symmetric x */
		int mj = m * Nmx + jx;
		xr *= *(ax);
		Q[ll+mm] -= xr * (qc[mj]-us[mj]);  /* sc+ps */
		U[ll+mm] -= xr * (qs[mj]+uc[mj]);  /* pc-ss */
	    }
	}      /* for xlbd< eps     */
	xi = (*ax) * xi;
	if(ABS(xi) > eps) {
	    Q[ll+mm] += xi * (qc[mi] - us[mi]);   /* sc+ps */
	    U[ll+mm] += xi * (qs[mi] + uc[mi]);   /* pc-ss */
	    if(jx > 0)  {              /* for symmetric x */
		int mj = m * Nmx + jx;
		xi =  xlbd[mx];
		Q[ll+m1] -= xi * (qc[mj]+us[mj]);  /* sc-ps */
		U[ll+m1] -= xi * (uc[mj]-qs[mj]);  /* ss+pc */
	    }
	}      /* for xlbd< eps     */
	*ax = -(*ax);
    }         /*  for all m<L< Lmx */
    return;
}

extern void fftw_b(int, double *, double *);
extern void fftw_f(int, double *, double *);
/*
C   fast loop over m - SUM_m[a_{lm}e^{i m phi}]
*/
extern int sym_grid;


void alm_fast3_Pol(Nmx,Nmp,Nx,Jmx,Lmx,
	       Mmax,Nph,f,c,apm,bpm,wsave,
	       K_temp,K_pol,Q,U,qc,uc,qs,us)
int *Nmx, *Nmp, *Nx, *Jmx, *Lmx;
int *Mmax;
int *Nph;
double *f, *c, *apm, *bpm, *wsave;
int *K_temp, *K_pol;
register double *Q, *U, *qc, *uc, *qs, *us;
{
    register int ix, ip, iz, iix, iiz, ix_s, ii, jj;
    register int Mx, nx=*Nmx, nb=*Nmp;
    int Kx = sym_grid > 0 ? (*Nx + 1)/2 :  *Nx;
    register double nul = 0.0;

    /*  loop over x  */
    ix_s = *Nx-1;
    jj = ix_s * nb;
    for(ix=ii=0;    ix< Kx;   ix++,ii+=nb,ix_s--,jj-=nb) {
	int nphi_work = Nph[ix];

//      RFFTI(Nph[ix], wsave);
	Mx = Mmax[ix];
	if(Mx > *Lmx)  Mx = *Lmx;
	if(2*Mx+1 > nphi_work)  Mx = (nphi_work-1)/2;

	if(*K_temp > 0) {
	    register double *fh = f + ii;
	    for(ip=0; ip<nphi_work; ip++)  c[ip] = fh[ip];

	    //RFFTF(Nph[ix],c,wsave);
	    fftw_b(nphi_work, c, wsave);
	    apm[ix] = -c[0];
	    bpm[ix] = nul;
	    for(iz=1,iix=nx+ix; iz<=Mx; iz++,iix+=nx) {
		register double *ci = c + iz + iz;
		apm[iix] = - *(ci-1);
		bpm[iix] =   *ci;
	    }
	    if (sym_grid > 0) { /* then calculation for south pole */
		if (ix_s != ix) {
		    fh = f + jj;
		    for(ip=0; ip<nphi_work; ip++)  c[ip] = *(fh+ip);

		    fftw_b(nphi_work, c, wsave);

		    apm[ix_s] = -c[0];
		    for(iz=1,iiz=nx+ix_s; iz <= Mx; iz++,iiz+=nx) {
			    register double *ci = c + iz + iz;
			    apm[iiz] = - *(ci-1);
			    bpm[iiz] =   *ci;
		    }
		}
	    }
	}

	if(*K_pol > 0) {
	    register double *h = Q + ii;
	    for(ip=0; ip<nphi_work; ip++)   c[ip] = *(h+ip);
	    //RFFTF(Nph[ix],c,wsave);
	    fftw_b(nphi_work, c, wsave);
	    qc[ix] = c[0];
	    qs[ix] = nul;
	    for(iz=1,iix=nx+ix; iz<=Mx; iz++,iix+=nx) {
		register double *ci = c + iz + iz;
		qc[iix] =   *(ci-1);
		qs[iix] = - *ci;
	    }
	    if (sym_grid > 0) { /* then calculation for south pole */
		if (ix_s != ix) {
		    h = Q + jj;
		    for(ip=0; ip<nphi_work; ip++)  c[ip] = *(h+ip);

		    fftw_b(nphi_work, c, wsave);

		    qc[ix_s] = c[0];
		    qs[ix_s] = nul;
		    for(iz=1,iiz=nx+ix_s; iz <= Mx; iz++,iiz+=nx) {
			    register double *ci = c + iz + iz;
			    qc[iiz] =   *(ci-1);
			    qs[iiz] = - *ci;
		    }
		}
	    }

	    h = U + ii;
	    for (ip=0; ip<nphi_work; ip++)  c[ip] = *(h+ip);
	    //RFFTF(Nph[ix],c,wsave);
	    fftw_b(nphi_work, c, wsave);
	    uc[ix] = c[0];
	    us[ix] = 0;
	    for(iz=1,iix=nx+ix; iz<=Mx; iz++,iix+=nx) {
		    //register int iz2 = iz + iz;
		    register double *ci = c + iz + iz;
		    uc[iix] =   *(ci-1);
		    us[iix] = - *ci;
	    }
	    if (sym_grid > 0) { /* then calculation for south pole */
		if (ix_s != ix) {
		    h = U + jj;
		    for(ip=0; ip<nphi_work; ip++)  c[ip] = *(h+ip);

		    fftw_b(nphi_work, c, wsave);

		    uc[ix_s] = c[0];
		    us[ix_s] = nul;
		    for(iz=1,iiz=nx+ix_s; iz<=Mx; iz++,iiz+=nx) {
			    //register int iz2 = iz + iz;
			    register double *ci = c + iz + iz;
			    uc[iiz] =   *(ci-1);
			    us[iiz] = - *ci;
		    }
		}
	    }
	}
    }  /*  END over ix  */
    return;
}


void alm_lgndr_Mu(Nmx,Nmp,Jmx,Lmn,Lmx,Ksym,Ktst,K_temp,
	      eps1,eps2,Nx,x,sqx,apm,bpm,glm,c,f,Mmax,
	      NTlm,NTeps,NPlm,NPeps,cpM,
	      Mlim,K_pol,Q,U,qc,uc,qs,us,xlbd)
int *Nmx, *Nmp, *Jmx, *Lmn, *Lmx, *Ksym, *Ktst, *K_temp;
double *eps1, *eps2, *x, *sqx, *apm, *bpm, *glm, *c, *f;
int *Nx, *Mmax;
int *NTlm, *NTeps, *NPlm, *NPeps;
double *cpM;
int *Mlim, *K_pol;
double *Q, *U, *qc, *uc, *qs, *us, *xlbd;
{
    register int m, ix, jx, l, nx1 = *Nx - 1;
    double pi=M_PI;
    double cnorm=1. / sqrt(2.*pi);
    double Enrm= pow(10., (double)*Mlim);
    int Kx = *Nx, lmx1 = *Lmx+1, L0, np = *Nmp;
    int ax = 1;
    register int mmb = *Jmx+2, mmb2 = mmb+mmb;
    register double  *cpmm = cpM + mmb;


    if(*Ksym > 0) Kx=(*Nx+1)/2;

    for (m=0; m<= *Lmx; m++) {   /* for all m */
	 register int mm = m * *Nmx, m1=m+1, mb = lmx1+m;
	 double xc, cmm;
	 double eps = *eps2;
	 int mv =m;
	 register double *f_r = f + m, *f_i = f + mb;

	 if(m < 100) eps = *eps1;
	 L0 = m;
	 if(L0 < *Lmn)   L0 = *Lmn;

	 start_clm(Lmx, &mv, &cmm, c);

//         ax=1-2*MOD((L0+m),2);
	 ax = 1;

	 for (ix=0; ix<Kx; ix++) {  /* for all x(ix) */
	    if(m > Mmax[ix])  continue;

	    xc=x[ix];
//            if(*Ksym > 0) {
		jx = nx1 - ix;
		if(jx <= ix) jx = -100;
//            } else
//                jx = -100;

	    lgndr_Mn(Mlim, &Enrm, Lmx, &mv, &cmm, c, &xc, glm, NTlm);

	    {
	       int aax=ax, aax1=ax, aax2=ax;      /* for L>2 and all m */
	       register double dfx=1./(1.-xc*xc + 1.e-20);
	       register double sx, sy, mdfx = m * dfx;
	       register int llp = np * L0;

	       register int mi = mm + ix, mj = mm + jx;
	       register
	       double qci = qc[mi], qsi = qs[mi], uci = uc[mi], usi = us[mi],
		      qcj = qc[mj], qsj = qs[mj], ucj = uc[mj], usj = us[mj];
	       register double *xr = xlbd + L0,
			       *xi = xlbd + L0 + *Jmx+1 ;

	       register double a_n = apm[mi], a_s = apm[mj],
			       b_n = bpm[mi], b_s = bpm[mj];


//               if(m == 1)  aax1 = -ax;
	       for (l=L0; l <= *Lmx; l++) {
		    register int i=4*l-3, imm = i * mmb + m1;
		    register double gl1 = glm[l+1], gl = glm[l],
				    cgl = cpM[imm]*gl;
		    register double *cpi = cpmm + imm;

		    register int lr = llp+m, li = llp+mb;
		    register double x_r, x_i;

		    sx = cgl*xc  +  *cpi        * gl1;
		    sx = sx* dfx -  *(cpi+mmb2) * gl1;      /* F_lm */
		    sy = cgl  - xc* *(cpi+mmb)  * gl1;
		    sy = mdfx * sy;                         /* Phi_lm */

		    /* if(Kxf > 0) {  */
		    *xr = sx+sy;             /* F_lm+Phi_lm=lbd^+ */
		    if (aax1 > 0)
			  *xi = sx-sy;           /* F_lm-Phi_lm=lbd^- */
		    else
			  *xi = sy-sx;           /* F_lm-Phi_lm=lbd^- */

		    x_r = *xr, x_i = *xi;
		    if(ABS(x_r) > eps || ABS(x_i) > eps) {
			Q[lr] += x_r * (qci + usi);  /* sc-ps */
			U[lr] += x_r * (uci - qsi);  /* ss+pc */
			Q[li] += x_i * (qci - usi);  /* sc+ps */
			U[li] += x_i * (qsi + uci);  /* pc-ss */
			if(jx > 0)  {    /* for symmetric x */
			    if (aax2 < 0) {
				x_r = -x_r;
				x_i = -x_i;
			    }
			    Q[lr] += x_i * (qcj + usj);   /* sc-ps */
			    U[lr] += x_i * (ucj - qsj);   /* ss+pc */
			    Q[li] += x_r * (qcj - usj);   /* sc+ps */
			    U[li] += x_r * (qsj + ucj);   /* pc-ss */
			 }
		    }           /* for xlbd< eps  */

		    if(ABS(gl1) > eps) {
			double *ff_r =  f_r + llp, *ff_i = f_i + llp;

			*ff_r += gl1 * a_n;
			*ff_i += gl1 * b_n;
			if(jx > 0)  {     /* for symmetric x */
			    if (aax < 0) {
				*ff_r -= gl1 * a_s;
				*ff_i -= gl1 * b_s;
			    } else {
				*ff_r += gl1 * a_s;
				*ff_i += gl1 * b_s;
			    }
			}      /* for symmetric x   */
		    }         /* for glm< eps      */
		    aax = -aax;

		    xr++; xi++;
		    llp += np;
		    aax1 = -aax1;
	       }
	    }        /*   for K_pol>0 */
	 }            /* for all x(ix) */
	 renorm(Nmx,Nmp,L0,Lmx,m,mb,cnorm,K_pol, f,Q,U,Ktst);

    }       /* over all m */
    return;
}

