#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PR      fprintf
#define ER      stderr
#define ABS(x)  ((x) < 0 ? -(x) : (x))


/* prints C_l, l, phase form file filename */
/* by VO                                   */
/* last modification 10/05/2007            */

void print_cl_table (fp_out, cl, clp, lmin, lmax)
FILE *fp_out;
double *cl;
double *clp;
int lmin, lmax;
{
    register int l;

    for (l=lmin; l<=lmax; l++) {
	PR(fp_out, "%5d %17.10e\n", l, cl[l]);
    }

    return;
}

void print_alm_table (fp_out, alm, lmax, nbuf, phase, lmin_print, lmax_print)
FILE *fp_out;
double *alm;
int nbuf, lmax, lmin_print, lmax_print;
double *phase;
{
    register int l, m;
    int flag = lmax_print < 0 ? 1 : 0; /* if flag=1; do not print zeroes */

    if (lmax_print < 0) lmax_print = -lmax_print;

    for (l=lmin_print; l<=lmax_print; l++) {
	register int ll=l*l+l+1,
		     bb = l*nbuf,
		     lp=(lmax+1)*l;
	for (m=0; m<=l; m++) {
	    int nlm = ll+m;
	    double re = alm [bb+m]       ;
	    double im = alm [bb+m+lmax+1];
	    double ph = phase[lp+m];
	    if (!flag || (flag && (ABS(re) > 1.e-30 || ABS(im) > 1.e-30))) {
		PR(fp_out, "nlm=%d, l=%d, m=%d, ", nlm, l, m);
		PR(fp_out, "%16.9e, %16.9e %5.3e\n", re, im, ph);
	    }
	}
    }
    return;
}

void  print_alm_plot (fp_out, alm, lmax, nbuf, lmin_print, lmax_print)
FILE *fp_out;
double *alm;
int nbuf, lmax, lmin_print, lmax_print;
{
    register int l, m;
    int flag = lmax_print < 0 ? 1 : 0; /* if flag=1; do not print zeroes */

    if (lmax_print < 0) lmax_print = -lmax_print;


    for (l=lmin_print; l<=lmax_print; l++) {
	register int bb = l*nbuf;
	for (m=0; m<=l; m++) {
	    double re = alm [bb+m]       ;
	    double im = alm [bb+m+lmax+1];
	    if (!flag || (flag && (ABS(re) > 1.e-30 || ABS(im) > 1.e-30)))
		PR(fp_out, "%5d  %5d  %16.9e %16.9e\n", l, m, re, im);
	}
    }
    return;
}

void  alm2sp(alm, nalm, lmax, l_calc, m_calc, fp_out)
double *alm;
int nalm, lmax;
int l_calc, m_calc;
FILE *fp_out;
{
	register int l=0, m=0;
	int l_start = 2;

	if (l_calc > 0 || m_calc >= 0) {
	    /* one-dimensional file */
	    if (l_calc > 0) {
		int ll = l_calc*nalm;
		for (m=0; m<=l_calc; m++) {
		    double re = alm[ll+m];
		    double im = alm[ll+m+lmax+1];
		    double v = sqrt (re*re + im*im);
		    PR(fp_out, "%5d  %17.10e\n", m, v);
		}
	    } else if (m_calc >= 0) {
		for(l= m_calc<l_start?l_start:m_calc ; l<=lmax; l++) {
		    int ll = l*nalm;
		    double re = alm[ll+m];
		    double im = alm[ll+m+lmax+1];
		    double v = sqrt (re*re + im*im);
		    PR(fp_out, "%5d  %17.10e\n", m, v);
		}
	    }
	} else {
		/* two-dimensional file */
	       ;
	}

	return;
}

void print_2eb_table (fp_out, cle, clb, lmin, lmax)
FILE *fp_out;
double *cle;
double *clb;
int lmin, lmax;
{
    register int l;

    for (l=lmin; l<=lmax; l++) {
	PR(fp_out, "%5d %17.10e %17.10e\n", l, cle[l], clb[l]);
    }

    return;
}
