#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define PR      fprintf
#define ER      stderr

#include "fits.h"
#include "glesp.h"

extern char *myname;


/* reads C_l, Cp_l, l form file filename */
/* creates arrays 'cl' and 'clp'         */
/* lmax is returned size of arrays       */

int read_cl_table (filename, cl, clp, lmax)
char *filename;
double **cl;
double **clp;
int *lmax;
{
	register int i=0;
	int l, lmax_set = 0;
	char str[161];
	FILE *fp;
	double cr;

	if (*lmax > 0)  lmax_set = *lmax;
	*lmax = -1;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}

	if (fgets(str, 160, fp) != NULL) {
	    if (strncmp (str, "SIMPLE  =", 9) == 0) {
		extern int read_cl_fits (char *, double **, int *);
		fclose(fp);
		read_cl_fits (filename, cl, lmax);
		FHfree();
		if (lmax_set == 0 )
			lmax_set = *lmax;
		else
			*lmax = lmax_set;

		return *lmax;
	    } else {
		fseek(fp, 0l, 0);
	    }
	} else {
		PR(ER, "%s: can not read file '%s'\n", myname, filename);
		exit(1);
	}


	/* strings count */
	while (fgets(str, 160, fp) != NULL) {
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;

		sscanf(str, "%d", &l);
		if (l > *lmax)   *lmax = l;
		i++;
	}
	fseek(fp, 0l, 0);

	if (*lmax < 0) {
		PR(ER, "%s: problem with content of file '%s'\n",
			myname, filename);
		exit(1);
	}

	if (lmax_set == 0 )
		lmax_set = *lmax;
	else
		*lmax = lmax_set;

	/* find memory for arrays */
	if ((*cl = (double *)calloc((unsigned)(lmax_set + 1), sizeof **cl)) ==
	    (double *)0) {
nomem:
		PR(ER,"%s: can not find memory for array...\n", myname);
		exit(1);
	}
	if ((*clp =(double *)calloc((unsigned)(lmax_set + 1), sizeof **clp))==
	    (double *)0)
		goto nomem;

	/* strings reading */
	while (fgets(str, 160, fp) != NULL) {
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;

		sscanf(str, "%d", &l);
		if (l > lmax_set)  continue;
		i = l;
		sscanf(str, "%d%lf%lf%lf", &l, (*cl)+i, (*clp)+i, &cr);
	}

	fclose(fp);

	return *lmax;
}

int read_cl (filename, cl, lmax)
char *filename;
double **cl;
int *lmax;
{
	register int i=0;
	int l, lmax_set = 0;
	char str[161];
	FILE *fp;

	if (*lmax > 0)  lmax_set = *lmax;
	*lmax = -1;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}

	if (fgets(str, 160, fp) != NULL) {
	    if (strncmp (str, "SIMPLE  =", 9) == 0) {
		extern int read_cl_fits (char *, double **, int *);
		fclose(fp);
		read_cl_fits (filename, cl, lmax);
		FHfree();
		if (lmax_set == 0 )
			lmax_set = *lmax;
		else
			*lmax = lmax_set;

		return *lmax;
	    } else {
		fseek(fp, 0l, 0);
	    }
	} else {
		PR(ER, "%s: can not read file '%s'\n", myname, filename);
		exit(1);
	}


	/* strings count */
	while (fgets(str, 160, fp) != NULL) {
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;

		sscanf(str, "%d", &l);
		if (l > *lmax)   *lmax = l;
		i++;
	}
	fseek(fp, 0l, 0);

	if (*lmax < 0) {
		PR(ER, "%s: problem with content of file '%s'\n",
			myname, filename);
		exit(1);
	}

	if (lmax_set == 0 )
		lmax_set = *lmax;
	else
		*lmax = lmax_set;

	/* find memory for arrays */
	if ((*cl = (double *)calloc((unsigned)(lmax_set + 1), sizeof **cl)) ==
	    (double *)0) {
		PR(ER,"%s: can not find memory for array...\n", myname);
		exit(1);
	}

	/* strings reading */
	while (fgets(str, 160, fp) != NULL) {
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;

		sscanf(str, "%d", &l);
		if (l > lmax_set)  continue;
		i = l;
		sscanf(str, "%d%lf", &l, (*cl)+i);
	}

	fclose(fp);

	return *lmax;
}

/* reads C_l, Cp_l, l form file filename */
/* creates arrays 'cl' and 'clp'         */
/* lmax is returned size of arrays       */
/*---------------------------------------*/
/* reads from CMBfast output             */
/*---------------------------------------*/
int read_dl_table (filename, cl, clp, lmax)
char *filename;
double **cl;
double **clp;
int *lmax;
{
	register int i=0;
	int l, lmax_set = 0;
	char str[161];
	FILE *fp;
	double T_CMB = 2.726;
	double cr;

	if (*lmax > 0)  lmax_set = *lmax;
	*lmax = -1;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}

	if (fgets(str, 160, fp) != NULL) {
	    if (strncmp (str, "SIMPLE  =", 9) == 0) {
		extern int read_cl_fits (char *, double **, int *);
		fclose(fp);
		read_cl_fits (filename, cl, lmax);
		return *lmax;
	    } else {
		fseek(fp, 0l, 0);
	    }
	} else {
		PR(ER, "%s: can not read file '%s'\n", myname, filename);
		exit(1);
	}

	/* strings count */
	while (fgets(str, 160, fp) != NULL) {
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;

		sscanf(str, "%d", &l);
		if (l > *lmax)   *lmax = l;
		i++;
	}
	fseek(fp, 0l, 0);

	if (*lmax < 0) {
		PR(ER, "%s: problem with content of file '%s'\n",
			myname, filename);
		exit(1);
	}

	if (lmax_set == 0 )
		lmax_set = *lmax;
	else
		*lmax = lmax_set;

	/* find memory for arrays */
	if ((*cl = (double *)calloc((unsigned)(lmax_set + 1), sizeof **cl)) ==
	    (double *)0) {
nomem:
		PR(ER,"%s: can not find memory for array...\n", myname);
		exit(1);
	}
	if ((*clp =(double *)calloc((unsigned)(lmax_set + 1), sizeof **clp))==
	    (double *)0)
		goto nomem;

	/* strings reading */
	while (fgets(str, 160, fp) != NULL) {
	    double C_l;
	    if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		    continue;

	    sscanf(str, "%d", &l);
	    if (l > lmax_set)  continue;
	    i = l;
	    sscanf(str, "%d%lf%lf%lf", &l, &C_l, (*clp)+i, &cr);

	    /* conversion: C_l from CMBfast to actual C_l */
	    (*cl)[l] = C_l * 2*M_PI * T_CMB * T_CMB / (double)(l*(l+1));
	}

	fclose(fp);

	return *lmax;
}



int read_cl_fits (filename, cl, lmax)
char *filename;
double **cl;
int *lmax;
{
	register int i=0;
	int l=0, lma=0, tf=0;
	char str[161];
	double *ar_ee = (double *)0, *ar_bb = (double *)0,
	       *ar_te = (double *)0, *ar_tb = (double *)0,
	       *ar_eb = (double *)0;
	int naxis=0, naxis1=0, naxis2=0;
	char xtens[40], extend[40], tform[40], tbcol[40], format[40];
	int flag_table = 0;
	FILE *fp_in;
	int block_size=0, n_blocks;
	int bitpix1=0, bitpix2=0, bitpix3=0, bitpix4=0;
	char tform1[30], tform2[30], tform3[30], tform4[30];

	if ((fp_in = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
	}

	/****************************************/
	/* read basic keywords from the 1st head */
	/****************************************/
	/* Hinit_fits(fp_in); */
	FHinit_FITS(fp_in, 0);

	if (FEVget("EXTEND") != NULL)
	    strcpy(extend, FEVget("EXTEND"));

	if (strcmp(extend, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	FHfree();

	/****************************************/
	/* read basic keywords from the 2nd head */
	/****************************************/
	/* Hinit(fp_in); */
	FHinit_FITS(fp_in, 0);

	if (FEVget("XTENSION") != NULL)
	    strcpy(xtens, FEVget("XTENSION"));

	if (strcmp(xtens, "TABLE") != 0 && strcmp(xtens, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with TABLE or BINTABLE extension. ",
			myname);
	    PR(ER, "File '%s' has another format.\n", filename);
	    exit(1);
	}
	if (strcmp(xtens, "BINTABLE") == 0)
	    flag_table = 1;

	if (FEVget("NAXIS") != NULL)
		naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)   /* number of rows in a table */
		naxis2  = FLkey_get("NAXIS2");
	if (FEVget("L_MAX") != NULL)   /* number of rows in a table */
	    lma  = FLkey_get("L_MAX");
	else
	    lma = naxis2 - 1;

	if ((*cl = (double*)calloc((unsigned)lma+1, sizeof **cl))==(double*)0){
nomem:
		PR(ER, "%s: no memory for arrays...\n", myname);
		exit(1);
	}

	if (FEVget("TFIELDS") != NULL)
		tf = FLkey_get("TFIELDS");

	switch  (tf) {
	case 1:
		break;
	case 4:
	    if ((ar_ee = (double*)calloc((unsigned)lma+2, sizeof *ar_ee))
		 ==(double*)0)
		    goto nomem;
	    if ((ar_bb = (double*)calloc((unsigned)lma+2, sizeof *ar_bb))
		 ==(double*)0)
		    goto nomem;
	    if ((ar_te = (double*)calloc((unsigned)lma+2, sizeof *ar_te))
		 ==(double*)0)
		    goto nomem;

	    if (flag_table) {
		/* XENSION = BINTABLE */
		/* data format of field: */
		if (FEVget("TFORM1") != NULL)
			strcpy(tform1, FEVget("TFORM1"));

		bitpix = -32;
		i = strlen(tform1) - 1;
		switch(tform1[i]) {
		case 'E':
		case 'F':
			bitpix = -32;
			break;
		case 'J':
			bitpix = 32;
			break;
		case 'D':
			bitpix = 64;
			break;
		case 'I':
			bitpix = 16;
			break;
		case 'B':
			bitpix = 8;
			break;
		}
		bitpix1 = bitpix;
		if (strlen(tform1) == 1) block_size = 1;
		else  block_size = atoi(tform1);

		/* data format of field: */
		if (FEVget("TFORM2") != NULL)
			strcpy(tform2, FEVget("TFORM2"));
		bitpix = -32;
		i = strlen(tform2) - 1;
		switch(tform2[i]) {
		case 'E':
		case 'F':
			bitpix = -32;
			break;
		case 'L':
		case 'J':
			bitpix = 32;
			break;
		case 'D':
			bitpix = 64;
			break;
		case 'I':
			bitpix = 16;
			break;
		case 'B':
			bitpix = 8;
			break;
		}
		bitpix2 = bitpix;
		if (strlen(tform2) == 1) i = 1;
		else  i = atoi(tform2);
		if (block_size !=  i) {
		    PR(ER, "%s: I can not read these data because ", myname);
		    PR(ER, "of different blocks for fields...\n");
		}

		if (FEVget("TFORM3") != NULL)
			strcpy(tform3, FEVget("TFORM3"));
		bitpix = -32;
		i = strlen(tform3) - 1;
		switch(tform3[i]) {
		case 'E':
		case 'F':
			bitpix = -32;
			break;
		case 'L':
		case 'J':
			bitpix = 32;
			break;
		case 'D':
			bitpix = 64;
			break;
		case 'I':
			bitpix = 16;
			break;
		case 'B':
			bitpix = 8;
			break;
		}
		bitpix3 = bitpix;
		if (strlen(tform3) == 1) i = 1;
		else  i = atoi(tform3);
		if (block_size !=  i) {
		    PR(ER, "%s: I can not read these data because ", myname);
		    PR(ER, "of different blocks for fields...\n");
		}

		if (FEVget("TFORM4") != NULL)
			strcpy(tform4, FEVget("TFORM4"));

		bitpix = -32;
		i = strlen(tform4) - 1;
		switch(tform4[i]) {
		case 'E':
		case 'F':
			bitpix = -32;
			break;
		case 'L':
		case 'J':
			bitpix = 32;
			break;
		case 'D':
			bitpix = 64;
			break;
		case 'I':
			bitpix = 16;
			break;
		case 'B':
			bitpix = 8;
			break;
		}
		bitpix4 = bitpix;
		if (strlen(tform4) == 1) i = 1;
		else  i = atoi(tform4);
		if (block_size !=  i) {
		    PR(ER, "%s: I can not read these data because ", myname);
		    PR(ER, "of different blocks for fields...\n");
		}
	    }
	    break;
	case 6:
		if ((ar_ee = (double*)calloc((unsigned)lma+1, sizeof *ar_ee))
		     ==(double*)0)
			goto nomem;
		if ((ar_bb = (double*)calloc((unsigned)lma+1, sizeof *ar_bb))
		     ==(double*)0)
			goto nomem;
		if ((ar_te = (double*)calloc((unsigned)lma+1, sizeof *ar_te))
		     ==(double*)0)
			goto nomem;
		if ((ar_tb = (double*)calloc((unsigned)lma+1, sizeof *ar_tb))
		     ==(double*)0)
			goto nomem;
		if ((ar_eb = (double*)calloc((unsigned)lma+1, sizeof *ar_eb))
		     ==(double*)0)
			goto nomem;
		break;
	default:
	case 7 :
		PR(ER, "%s: bad number of fields (%d) for function ",
			myname, ph->tfields);
		PR(ER, "'read_cl_fits'. There must be 1, 4 or 6.\n");
		exit(1);
	}

	if (flag_table) {
	    int *a_l = (int *)0;
	    /* input binary data (as in Planck) */
	    n_blocks = naxis2 / block_size;
	    if (block_size > 1) {
		a_l = (int *)calloc(block_size, sizeof *a_l);
	    }

	    for(i=0; i < n_blocks; i++) {
		int j;
		bitpix = bitpix1;
		for(j=0; j < block_size; j++) {
		    double pix_val;
		    if (feof(fp_in)) {
			    PR(ER, "%s: no data after %d-th element.\n",
				    myname, i);
			    break;
		    }
		    pix_val = get_pix_fits(fp_in);
		    l = pix_val + 0.5;
		    if (block_size > 1) a_l[j] = l;
		}
		if (bitpix2) {
		    bitpix = bitpix2;
		    for(j=0; j < block_size; j++) {
			double v;
			if (feof(fp_in)) {
			    PR(ER, "%s: no data after %d-th element.\n",
				    myname, i);
			    break;
			}
			v = get_pix_fits(fp_in);
			if (block_size > 1)  l = a_l[j];
			(*cl)[l] = v;  /* actually D_l */
		    }
		}
		if (bitpix3) {
		    bitpix = bitpix3;
		    for(j=0; j < block_size; j++) {
			if (feof(fp_in)) {
			    PR(ER, "%s: no data after %d-th element.\n",
				    myname, i);
			    break;
			}
			/* errors up  in Planck data */
			get_pix_fits(fp_in);
		    }
		}
		if (bitpix4) {
		    bitpix = bitpix4;
		    for(j=0; j < block_size; j++) {
			if (feof(fp_in)) {
			    PR(ER, "%s: no data after %d-th element.\n",
				    myname, i);
			    break;
			}
			/* errors down  in Planck data */
			get_pix_fits(fp_in);
		    }
		}
	    }
	    if (a_l != (int *)0) free((void *)a_l);
	} else {
	    for (l=0; l<=lma; l++) {
		int col_tbr = 1, col_tb;
		for (i=0; i<tf; i++) {
		    int n_digits;
		    sprintf(tform, "TFORM%i", i+1);
		    if (FEVget(tform) != NULL)
			strcpy(format, FEVget(tform));
		    n_digits = atoi(format+1);

		    sprintf(tbcol, "TBCOL%i", i+1);
		    if (FEVget(tbcol) != NULL)
			col_tb = atoi(FEVget(tbcol));
		    else
			col_tb = 0;

		    if (col_tbr < col_tb) {
			int ct = col_tb - col_tbr;
			col_tbr += ct;
			while (ct--)   fgetc(fp_in);
		    }
		    fread(str, n_digits, 1, fp_in);
		    str[n_digits] = '\0';

		    switch (i) {
		    case 0:
			(*cl)[l] = atof(str);
			break;
		    case 1:
			ar_ee[l] = atof(str);
			break;
		    case 2:
			ar_bb[l] = atof(str);
			break;
		    case 3:
			ar_te[l] = atof(str);
			break;
		    case 4:
			ar_tb[l] = atof(str);
			break;
		    case 5:
			ar_eb[l] = atof(str);
			break;
		    }
		    col_tbr += n_digits;
		}
	    }
	}
	*lmax = lma;

	if (*lmax <= 0) {
		PR(ER, "%s: problem with content of file '%s'\n",
			myname, filename);
		exit(1);
	}

	if (ar_ee != (double *)0) free ((void *)ar_ee);
	if (ar_bb != (double *)0) free ((void *)ar_bb);
	if (ar_te != (double *)0) free ((void *)ar_te);
	if (ar_tb != (double *)0) free ((void *)ar_tb);

	return *lmax;
}
