#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <ff.h>
#include "glesp.h"

#define  PR     fprintf
#define  ER     stderr

extern char tmpcat[161];
extern char *myname;

HEAD hd, *ph;


/* if we use Legendre grid, or uniform by cos theta */
/* then all other grids must be recalculated        */
void recalc_map2(f, Nx,Nphi, Nmx,Nmphi, Nph,x_in,x_out,
					       input_grid,output_grid)
double **f;            /* input map (transposed) */
int Nx, Nmx, Nphi;
int *Nph;
double *x_in, *x_out;
int input_grid, output_grid;
{
    register int i, j;
    double *fn = (double *)0;
    FILE *fp_tmp;
    char tmpfile[161];

    /* No conversion when input grid is output grid, */
    /* i.e. Legendre grid -> Legendre grid or        */
    /*      cos theta grid -> cos theta grid.        */
    /* Else conversion with interpolation            */
    if( !((input_grid == GRID_LEGENDRE || input_grid == GRID_PHI_X_LEGENDRE)
	 && output_grid ==  1)  &&
	!(input_grid == GRID_UNIFORM_COS && output_grid == -1)  ) {
	double *f_new = (double *)0;

	PR(ER, "%s: grid recalculation...\n", myname);
	if ((f_new = (double *)calloc((unsigned)(Nx*Nphi), sizeof *f_new)) ==
	   (double *)0) {
	   PR(ER,"%s: no memory (%d kb) for map recalculation...\n",
		   myname, Nx*Nphi*8);
	   exit(1);
	}

	if (input_grid == GRID_UNIFORM_THETA) {
	       /* transition to COS grid */
	       for(i=0; i<Nx; i++)  x_in[i] = cos(x_in[i]);
	}

	/*-----------------------------------------------*/
	/* spline interpolation (modified input):        */
	/*   non-uniform step by X ---> non-uniform step */
	/*-----------------------------------------------*/
	intrpl2sm_nu_nu(*f, Nx, Nx, f_new, Nx, Nx, x_in, x_out, Nphi);

	for(j=0; j<Nphi; j++)  {
		register int jj = Nx * j;
		for(i=0; i<Nx; i++)
			(*f)[jj+i] = f_new[jj+i];
	}
	free ((void *)f_new);
    }

    /* find memory */
    fn = (double *)calloc((unsigned)(Nmx*Nmphi), sizeof *fn);
    if (fn == (double *)0) {
	   PR(ER,"%s: no memory (%d kb) for map recalculation...\n",
		   myname, Nmx*Nmphi*16/1024);
	   PR(ER, "%s: trying to use disk space for temporal array ",
		   myname);
	   PR(ER, "of same size...");

	   /* open temporal file and save array to disk */
	   sprintf(tmpfile, "%s/clXXXXXX", tmpcat);
	   if((fp_tmp=fopen(mktemp(tmpfile), "w+")) == (FILE *)0) {
		   PR(ER,"%s: can not create temporal file '%s'\n",
			myname, tmpfile);
		   exit(1);
	   }
	   if ((fn = (double *)calloc((unsigned)Nphi, sizeof *fn))
		   == (double *)0) {
		   PR(ER,"\n%s: Sorry. No memory at all...\n",
			   myname);
		   fclose(fp_tmp);
		   unlink(tmpfile);
		   exit(1);
	   } else {
		for(i=0; i<Nx; i++) {
		    for(j=0; j<Nphi; j++)
			     fn[j] = (*f)[j*Nx+i];

		    if (fwrite(fn,  sizeof *fn,  Nphi, fp_tmp) != Nphi) {
			 PR(ER,"%s: error during writing...\n", myname);
			 fclose(fp_tmp);
			 unlink(tmpfile);
			 exit(1);
		    }
		}
		free((void *)fn);

	   }

	   /* free old array */
	   free((void *)*f);

	   PR(ER, "Done.\n");
	   PR(ER, "Now try to find memory again...");
	   if ((fn = (double *)calloc((unsigned)(Nmx*Nmphi), sizeof *fn)) ==
		(double *)0) {
		     PR(ER,"%s: no memory again...\n", myname);
		     exit(1);
	   }
	   fseek(fp_tmp, 0l, 0);
	   for(i=0; i<Nx; i++) {
		if (fread(fn+i*Nphi,  sizeof *fn,  Nphi, fp_tmp) != Nphi) {
		     PR(ER,"%s: error during reading...\n", myname);
		     exit(1);
		}
	   }
	   unlink(tmpfile);
	   PR(ER, "O'k.\n");
    } else {

	  /* because the map is in transposed form */
	  /* we have to make transposing back      */
	  for(i=0; i<Nx; i++) {
		  register int kk = i * Nmphi;
		  for(j=0; j<Nph[i]; j++) {
			   fn[kk+j] = (*f)[j*Nx+i];
		  }
	  }

	  /* free old array */
	  free((void *)*f);

    }
    /* appropriate new array addres to old pointer */
    *f = fn;

    return;
}
/* if we use Legendre grid, or uniform by cos theta */
/* then all other grids must be recalculated        */
void recalc_map(f, Nx, Nphi, Nbuf, x_in, x_out, input_grid, output_grid)
double **f;            /* input map (transposed) */
int Nx, Nphi, Nbuf;
double *x_in, *x_out;
int input_grid, output_grid;
{
    register int i, j;
    double *fn = (double *)0;
    int nb2 = Nbuf * 2;
    FILE *fp_tmp;
    char tmpfile[161];

    /* NO conversion when input grid is output grid, */
    /* i.e. Legendre grid -> Legendre grid or        */
    /*      cos theta grid -> cos theta grid.        */
    /* Else conversion with interpolation            */
    if( !(input_grid == GRID_LEGENDRE    && output_grid ==  1)  &&
	!(input_grid == GRID_UNIFORM_COS && output_grid == -1)  ) {
	double *f_new = (double *)0;

	PR(ER, "%s: grid recalculation...\n", myname);
	if ((f_new = (double *)calloc((unsigned)(Nx*Nphi), sizeof *f_new)) ==
	   (double *)0) {
	   PR(ER,"%s: no memory (%d kb) for map recalculation...\n",
		   myname, Nx*Nphi*8);
	   exit(1);
	}

	if (input_grid == GRID_UNIFORM_THETA) {
	       /* transition to COS grid */
	       for(i=0; i<Nx; i++)  x_in[i] = cos(x_in[i]);
	}

	/*-----------------------------------------------*/
	/* spline interpolation (modified input):        */
	/*   non-uniform step by X ---> non-uniform step */
	/*-----------------------------------------------*/
	intrpl2sm_nu_nu(*f, Nx, Nx, f_new, Nx, Nx, x_in, x_out, Nphi);

	for(j=0; j<Nphi; j++)  {
		register int jj = Nx * j;
		for(i=0; i<Nx; i++)
			(*f)[jj+i] = f_new[jj+i];
	}
	free ((void *)f_new);
    }

    /* find memory */
    fn = (double *)calloc((unsigned)(Nbuf*2*Nbuf), sizeof *fn);
    if (fn == (double *)0) {
	   PR(ER,"%s: no memory (%d kb) for map recalculation...\n",
		   myname, Nbuf*Nbuf*16/1024);
	   PR(ER, "%s: trying to use disk space for temporal array ",
		   myname);
	   PR(ER, "of same size...");

	   /* open temporal file and save array to disk */
	   sprintf(tmpfile, "%s/clXXXXXX", tmpcat);
	   if((fp_tmp=fopen(mktemp(tmpfile), "w+")) == (FILE *)0) {
		   PR(ER,"%s: can not create temporal file '%s'\n",
			myname, tmpfile);
		   exit(1);
	   }
	   if ((fn = (double *)calloc((unsigned)(Nphi), sizeof *fn))
		   == (double *)0) {
		   PR(ER,"\n%s: Sorry. No memory at all...\n",
			   myname);
		   fclose(fp_tmp);
		   unlink(tmpfile);
		   exit(1);
	   } else {
		for(i=0; i<Nx; i++) {
		    for(j=0; j<Nphi; j++)
			     fn[j] = (*f)[j*Nx+i];
/*
if(i>Nx-5)
PR(ER,"\n%12.5e %12.5e %12.5e %12.5e %12.5e\n",
fn[0], fn[1], fn[2], fn[3], fn[4]);
*/
		    if (fwrite(fn,  sizeof *fn,  Nphi, fp_tmp) != Nphi) {
			 PR(ER,"%s: error during writing...\n", myname);
			 fclose(fp_tmp);
			 unlink(tmpfile);
			 exit(1);
		    }
		}
		free((void *)fn);

	   }

	   /* free old array */
	   free((void *)*f);

	   PR(ER, "Done.\n");
	   PR(ER, "Now try to find memory again...");
	   if ((fn = (double *)calloc((unsigned)(Nbuf*2*Nbuf), sizeof *fn)) ==
		(double *)0) {
		     PR(ER,"%s: no memory again...\n", myname);
		     exit(1);
	   }
	   fseek(fp_tmp, 0l, 0);
	   for(i=0; i<Nx; i++) {
		if (fread(fn+i*nb2,  sizeof *fn,  Nphi, fp_tmp) != Nphi) {
		     PR(ER,"%s: error during reading...\n", myname);
		     exit(1);
		}
/*
if(i>Nx-5) {
double *p=fn+i*nb2;
PR(ER,"%12.5e %12.5e %12.5e %12.5e %12.5e\n",
p[0], p[1], p[2], p[3], p[4]);
}
*/
	   }
	   unlink(tmpfile);
	   PR(ER, "O'k.\n");
    } else {

	  /* because the map is in transposed form */
	  /* we have to make transposing back      */
	  for(i=0; i<Nx; i++) {
		  register int kk = i * nb2;  /* nb2 = 2* Nbuf */
		  for(j=0; j<Nphi; j++) {
			   fn[kk+j] = (*f)[j*Nx+i];
		  }
	  }

	  /* free old array */
	  free((void *)*f);

    }
    /* appropriate new array addres to old pointer */
    *f = fn;

/*
i=1; if(i) exit(0);
*/
    return;
}

