#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr

extern void euler_matrix(double, double, double, double [3][3]);
extern void xyz_euler_transform(double, double, double,
	double *, double *, double *, double [3][3]);

void rotate_map(f1, x, nx, np, nph, f2, dtheta, dphi)
double *f1, *f2;
double *x;
int nx, np, *nph;
double dtheta, dphi;     /* rotation angle */
{
   register int i,j;
   double euler_m[3][3];

   /* sign '-' because we use new pixel to find correspondig old one */
   /* i.e. from new position, we obtain old one                      */
   euler_matrix(-dphi, -dtheta, 0.0, euler_m);

//PR(ER, "dt=%g, dp=%g\n", dtheta, dphi);

   /* rotate sphere: copy pixel to pixel */
   for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	   int n_old, n_new;
	   double x_new, y_new, z_new, x_old, y_old, z_old;

	   /* get coordinates for new map (rotated) */
	   n_new = ii + j;
	   pix2sphere(n_new, nx,np,nph,x, &x_new, &y_new, &z_new);

	   /* get initial coordinates with Euler matrix */
	   /*    (rotate new pixel back ) */
	   xyz_euler_transform(x_new, y_new, z_new,
		&x_old, &y_old, &z_old, euler_m);

//if (i==10 && j== 3)
//PR(ER, "xo=%g, yo=%g, zo=%g,\nxn=%g, yn=%g, zn=%g\n",
//x_old, y_old, z_old, x_new, y_new, z_new);

	   /* corresponding pixel value */
	   n_old = sphere2pix(x_old, y_old, z_old, nx,np,nph,x);
//if (i==10 && j== 3)
//PR(ER, "n_new=%d, n_old=%d, nn=%d\n", n_new, n_old, nph[i]);

	   f2[n_new] = f1[n_old];
	}
   }
   
}

void xyz_euler_transform(x1, y1, z1, x2, y2, z2, euler_m)
double x1, y1, z1, *x2, *y2, *z2;
double euler_m[3][3];
{
    *x2 = euler_m[0][0] * x1  + euler_m[0][1] * y1 + euler_m[0][2] * z1;
    *y2 = euler_m[1][0] * x1  + euler_m[1][1] * y1 + euler_m[1][2] * z1;
    *z2 = euler_m[2][0] * x1  + euler_m[2][1] * y1 + euler_m[2][2] * z1;

    return;
}

/*
 * az, ay, ax = Euler angles, in radians
 *
 */
extern void matrix_multp33(double [3][3], double [3][3], double [3][3]);

void euler_matrix(az, ay, ax, euler_m)
double az, ay, ax;
double euler_m[3][3];
{
    double r1[3][3], r2[3][3], r3[3][3], buf[3][3];
    double cz = cos(az);
    double sz = sin(az);
    double cy = cos(ay);
    double sy = sin(ay);
    double cx = cos(ax);
    double sx = sin(ax);

    /* rotation around z-axis */
    r1[0][0] =   cz; r1[1][0] =  sz; r1[2][0] = 0.0;
    r1[0][1] =  -sz; r1[1][1] =  cz; r1[2][1] = 0.0;
    r1[0][2] =  0.0; r1[1][2] = 0.0; r1[2][2] = 1.0;

    /* rotation around y-axis */
    r2[0][0] =   cy; r2[1][0] = 0.0; r2[2][0] =  sy;
    r2[0][1] =  0.0; r2[1][1] = 1.0; r2[2][1] = 0.0;
    r2[0][2] =  -sy; r2[1][2] = 0.0; r2[2][2] =  cy;

    /* rotation around x-axis */
    r3[0][0] = 1.0;  r3[1][0] = 0.0; r3[2][0] = 0.0;
    r3[0][1] = 0.0;  r3[1][1] =  cx; r3[2][1] =  sx;
    r3[0][2] = 0.0;  r3[1][2] = -sx; r3[2][2] =  cx;

    matrix_multp33(r2, r3, buf);
    matrix_multp33(r1, buf, euler_m);
/*
PR(ER, "%12.5e  %12.5e  %12.5e\n",euler_m[0][0],euler_m[0][1],euler_m[0][2]);
PR(ER, "%12.5e  %12.5e  %12.5e\n",euler_m[1][0],euler_m[1][1],euler_m[1][2]);
PR(ER, "%12.5e  %12.5e  %12.5e\n",euler_m[2][0],euler_m[2][1],euler_m[2][2]);
*/
    return;
}

void matrix_multp33(a, b, c)
double a[3][3], b[3][3], c[3][3];
{
    register int i, j, k, n=3;
    for(i=0; i<n; i++)
	for(j=0; j<n; j++) {
	    c[i][j] = 0.0;
	    for(k=0; k<n; k++) c[i][j] += a[i][k] * b[k][j];
	}

    return;
}
