# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>
# include <strings.h>
# include <math.h>

# include "glesp.h"
# include "ff.h"

static struct _ff_ {
  int    bitpix;
  char   bunit[50];
  double bscale,
	 bzero;
  int    naxis;
  long   naxis1;
  double crval1,
	 cdelt1,
	 crpix1;
  double ra,
	 dec;
  double obsra,
	 obsdec;
  char   dateobs[50];
  double epoch;
  char   ctype1[50];
  char   origin[50];
  char   object[50];
  char   history[50];
} fh;

extern char *myname;
extern char tmpcat[161];

/* read equatorial coordinates from F-file and convert them */
void  read_ff(cr, n, filename, flag_coor, delta_dec)
struct _coor_  **cr;
int *n;
char *filename;
int flag_coor;   /* 0 - equatorial, 1 - galactic */
double delta_dec;
{
    register int i;
    char *tmp_pat = "mapcutXXXXXX";
    char tmp_file1[161], tmp_file2[161];
    FILE *fp;
    char command[161];
    char str[81];
    double pi2 = 2 * M_PI;

    fh.ra = fh.dec = 0.0;

    if((fp  = fopen(filename, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'...\n",  myname, filename);
	exit(1);
    }
    /* read head of F-file */
    Hinit(fp);

    /* close file */
    fclose(fp);

    /* read parameters from the head */
    fh.bitpix = Skey_get("BITPIX");        /* bits per pixel               */
    fh.naxis  = Skey_get("NAXIS");         /* number of axes; here=1       */
    fh.naxis1 = Lkey_get("NAXIS1");        /* number of points in 1st axis */
    fh.crval1 = Dkey_get("CRVAL1");        /* reference point of 1st axis  */
    fh.crpix1 = Dkey_get("CRPIX1");        /* number of pixels corres-     */
					   /* ponded to CRVAL1             */
    fh.cdelt1 = Dkey_get("CDELT1");        /* value of step between pixels */
    strcpy(ctype1, EVget("CTYPE1"));       /* type of scale in 1st axis    */

    fh.bscale = Dkey_get("BSCALE");        /* scale factor                 */
    if (EVget("BZERO") != NULL)
	    fh.bzero = Dkey_get("BZERO");  /* shift from zero level        */
    else
	    fh.bzero = 0.0;
    strcpy(fh.bunit,   EVget("BUNIT"));    /* type of units of data        */
    if (EVget("OBJECT") != NULL)
	strcpy(fh.object,  EVget("OBJECT"));/* name of object              */
    else
	strcpy(fh.object,  "0");            /* name of object              */
    if (EVget("DATE-OBS") != NULL)
	strcpy(fh.dateobs, EVget("DATE-OBS")); /* observation date         */
    else
	strcpy(fh.dateobs, "0");           /* observation date             */

    if (EVget("RA") != NULL)
	fh.ra  = Dkey_get("RA");
    else {
	fh.ra = fh.crval1 + (1 - fh.crpix1) * fh.cdelt1;
    }
    fh.dec = Dkey_get("DEC") + delta_dec;

    /* create temporal file name */
    sprintf(tmp_file1, "%s/%s", tmpcat, tmp_pat);
    mktemp(tmp_file1);

    if (EVget("EPOCH") != NULL) {
	fh.epoch  = Dkey_get("EPOCH");
	if (EVget("OBSRA")  != NULL) {
	    fh.obsra  = Dkey_get("OBSRA");
	    fh.obsdec = Dkey_get("OBSDEC") + delta_dec;
	} else {
	    sprintf(command, "epoch -eo %s -ei %g -a %gh -d %gd > %s",
	      fh.dateobs, fh.epoch, fh.ra, fh.dec, tmp_file1);
	    if (system(command) != 0) {
	      PR(ER,
		"%s: problems with execution of the following command:\n%s\n",
		   myname, command);
	      exit(1);
	    }
	    /* read galactic coordinates to array */
	    if((fp  = fopen(tmp_file1, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n",  myname, filename);
		exit(1);
	    }
	    /* strings count */
	    while (fgets(str, 160, fp) != NULL) {
		char s1[30], s2[30];
		if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
			continue;
		sscanf(str, "%s%s", s1, s2);
		fh.obsra  = str2deg(s1);
		fh.obsdec = str2deg(s2);
	    }
	    fclose(fp);
	}
    } else {
	fh.epoch  = 2000.;
	fh.obsra  = fh.ra ;
	fh.obsdec = fh.dec;
    }

    /* free memory from F-file */
    Hfree();

    *n = naxis1;

    /* create table of equatorial coordinates */
    if((fp  = fopen(tmp_file1, "w")) == (FILE *)0) {
	PR(ER, "%s: can not create temporal file '%s'\n",  myname, tmp_file1);
	exit(1);
    }
    for(i=0; i<fh.naxis1; i++)
	    PR(fp, "%12.5eh %12.5ed\n",
		 fh.obsra+((i-fh.crpix1)*fh.cdelt1)/3600., fh.obsdec);
    fclose(fp);

    /* create temporal file name */
    sprintf(tmp_file2, "%s/%s", tmpcat, tmp_pat);
    mktemp(tmp_file2);

    /**********************************************/
    /* convert observational coordinates to epoch */
    /**********************************************/
    /* command to be executed */
    sprintf(command, "epoch -ei %s -eo %g -f - < %s > %s",
	  fh.dateobs, fh.epoch, tmp_file1, tmp_file2);
    if (system(command) != 0) {
	PR(ER, "%s: problems with execution of the following command:\n%s\n",
	    myname, command);
	exit(1);
    }

    /* command to be executed */
    switch (flag_coor) {
    case 0:
    default:
      sprintf(command, "cp %s %s", tmp_file2, tmp_file1);
      break;
    case 1:
      /***************************************************/
      /* convert equatorial coordinates to galactic ones */
      /***************************************************/
      sprintf(command, "epoch -G -f - < %s > %s", tmp_file2, tmp_file1);
      break;
    }
    if (system(command) != 0) {
      PR(ER, "%s: problems with execution of the following command:\n%s\n",
	  myname, command);
      exit(1);
    }

    /* read coordinates to array */
    if((fp  = fopen(tmp_file1, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'\n",  myname, tmp_file1);
	exit(1);
    }
    /* strings count */
    i=0;
    while (fgets(str, 160, fp) != NULL) {
	    if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		    continue;
	    i++;
    }
    fseek(fp, 0l, 0);
    if (i == 0) {
	    PR(ER, "%s: problem with content of file '%s'\n",
		    myname, tmp_file1);
	    exit(1);
    }

    *n = i;
    /* find memory for arrays */
    if ((*cr = (struct _coor_ *)calloc((unsigned)i, sizeof **cr)) ==
	(struct _coor_ *)0) {
	    PR(ER,"%s: can not find memory for array...\n", myname);
	    exit(1);
    }

    /* strings reading */
    i=0;
    while (fgets(str, 160, fp) != NULL) {
	    char s1[30], s2[30];
	    double p = M_PI/180.;

	    if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		    continue;
	    sscanf(str, "%s%s", s1, s2);

	    switch(flag_coor) {
	    case 0:
	    default :
		/* convert (alp(hrs), del(degs)) -> (alpha, delta in rad) */
		(*cr)[i].lon = str2deg(s1) * 15. * p;   /* alpha */
		if ((*cr)[i].lon < 0.)  (*cr)[i].lon += pi2;
		if ((*cr)[i].lon > pi2) (*cr)[i].lon -= pi2;
		(*cr)[i].lat = str2deg(s2) * p;         /* delta */
		break;
	    case 1:
		strcat(s1, "d");
		strcat(s2, "d");
		(*cr)[i].lat = deg2rad(s2);
		(*cr)[i].lon = deg2rad(s1);
		break;
	    }
	    i++;
    }
    fclose(fp);

    /* remove temporal files */

    unlink(tmp_file1);
    unlink(tmp_file2);
/*
PR(ER, "file1='%s', file2='%s'\n", tmp_file1, tmp_file2);
*/
    return;
}

int swap_write_key = 0;

union _sw_ {
char c[8];
double d;
float  v;
long   l;
short  s;
};
void swap_bytes(union _sw_ *, int);

void  create_ff(bitp, nax1, bunit, bsc, bz, crv1, cdlt1, crp1, ctp1)
int bitp;
long nax1;
char *bunit, *ctp1;
double bsc, bz, crv1, cdlt1, crp1;
{
    /* create F-head */
    Ckey_set("SIMPLE", "T");
    fh.bitpix = bitp;
    strcpy(fh.bunit, bunit);
    fh.naxis  = 1;
    fh.bscale = bsc;
    fh.bzero  = bz;
    fh.crval1 = crv1;
    fh.cdelt1 = cdlt1;
    fh.crpix1 = crp1;
    strcpy(fh.ctype1, ctp1);
    return;
}

void  write_ff(r, n, fp_out)
double *r;
int n;
FILE *fp_out;
{
    register int i;

    /* create F-head */
    Ckey_set("SIMPLE", "T");
    Skey_set("BITPIX", fh.bitpix);
    Skey_set("NAXIS",  fh.naxis);
    Lkey_set("NAXIS1", (long)n); /* fh.naxis1 */
    Ckey_set("BUNIT",  fh.bunit);
    Dkey_set("BSCALE", fh.bscale);
    Dkey_set("BZERO",  fh.bzero);
    Dkey_set("CRVAL1", fh.crval1);
    Dkey_set("CDELT1", fh.cdelt1);
    Dkey_set("CRPIX1", fh.crpix1);
    Ckey_set("CTYPE1", fh.ctype1);
    Ckey_set("ORIGIN", "SAO RAS: GLESP");
    Ckey_set("OBJECT", "Sky Strip");

    /* write head to file */
    Hupdate(fp_out);

    /* write array to file */
    bitpix = fh.bitpix;
    for (i=0; i<n; i++) {
	    float v = (float)((r[i] - fh.bzero) / fh.bscale);
	    if (swap_write_key)
		    swap_bytes((union _sw_ *)&v, sizeof v);
	    put_pix(fp_out, v);
    }

    return;
}

void swap_bytes(sw, n)
union _sw_ *sw;
int n;
{
    int c;
    switch(n){
    case '8':
	    c = sw->c[7]; sw->c[7] = sw->c[0]; sw->c[0] = c;
	    c = sw->c[6]; sw->c[6] = sw->c[1]; sw->c[1] = c;
	    c = sw->c[5]; sw->c[5] = sw->c[2]; sw->c[2] = c;
	    c = sw->c[4]; sw->c[4] = sw->c[3]; sw->c[3] = c;
	    break;
    case '4':
	    c = sw->c[3]; sw->c[3] = sw->c[0]; sw->c[0] = c;
	    c = sw->c[2]; sw->c[2] = sw->c[1]; sw->c[1] = c;
	    break;
    case '2':
	    c = sw->c[1]; sw->c[1] = sw->c[0]; sw->c[0] = c;
	    break;
    }
    return;
}
