#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr
#define ABS(x)  ((x) < 0 ? -(x) : (x))

/* procedures of mapping estimators */
/* V.O., 01/2014                    */

#define  EST_SIG_CL       1  /* Cl-sigma estimator                */
#define  EST_SIG_DL       2  /* Dl-sigma estimator                */
#define  EST_CL_SIG_DIF   3  /* Cl-sigma difference estimator     */
#define  EST_DL_SIG_DIF   4  /* Dl-sigma difference estimator     */
#define  EST_PARITY       5  /* parity estimator                  */
#define  EST_PIX_MINIMAX  6  /* (MAX-MIN)_north / (MAX-MIN)_south */
#define  EST_PIX_DISP     7  /* Sigma_north / Sigma_south         */


char glesp_tmp_cat[51];

/* mapping estimator */
void stat_estim(f, x, nx, np, nph, mapfile, m_est, estimator_type, lmax_stat)
double *f, *x;
int nx, np, *nph;
char *mapfile;              /* initial file with map                */
double *m_est;              /* output map after estimator operation */
int estimator_type;         /* type of estimator                    */
int lmax_stat;              /* used lmax of stat. estimator         */
{
    register int i=0, j=0;
    double theta, phi;
    double *est_arr = (double *)0;
    int l, nxr, npr, na, nb, u, v, flag;
    char tmp_alm[161], tmpr_alm[161], tmprt_alm[161], tmpr_map[161],
	 tmpr_cl[161], tmprt_map[161];
    char comm[161];
    int order_rotation = 0,    /* (phi, theta)                 */
	reflection_type = 0,   /* no change sign in reflection */
	hemisphere_type = 1,   /* reflection: North -> South   */
	rot_angle_sign = 0;    /* sign of rotation angles      */
    extern int read_cl (char *, double **, int *);

    strcpy(glesp_tmp_cat, "/tmp");

    sprintf(tmp_alm,  "%s/almXXXXXX",   glesp_tmp_cat);
    sprintf(tmpr_alm, "%s/almrXXXXXX",  glesp_tmp_cat);
    sprintf(tmprt_alm,"%s/almtrXXXXXX", glesp_tmp_cat);
    sprintf(tmpr_map, "%s/maprXXXXXX",  glesp_tmp_cat);
    sprintf(tmprt_map,"%s/maprtXXXXXX", glesp_tmp_cat);
    sprintf(tmpr_cl,  "%s/clrXXXXXX",   glesp_tmp_cat);

    if( mktemp(tmp_alm) == NULL) {
	PR(ER,"%s: I can't create unique name ...\n", myname);
    }
    if( mktemp(tmpr_alm) == NULL) {
	PR(ER,"%s: I can't create unique name ...\n", myname);
    }
    if( mktemp(tmprt_alm) == NULL) {
	PR(ER,"%s: I can't create unique name ...\n", myname);
    }
    if( mktemp(tmpr_map) == NULL) {
	PR(ER,"%s: I can't create unique name ...\n", myname);
    }
    if( mktemp(tmprt_map) == NULL) {
	PR(ER,"%s: I can't create unique name ...\n", myname);
    }
    if( mktemp(tmpr_cl) == NULL) {
	PR(ER,"%s: I can't create unique name ...\n", myname);
    }

    /* estimator type */
    if (estimator_type >= 1000)       order_rotation = 1;   /* theta, phi  */
    if (estimator_type % 1000 > 100) {
	int k = (estimator_type % 1000) / 100;
	if ( k%2 )
	     reflection_type = 1; /* change sign in N->S & S->N reflection */
	if ( k>1 )
	     rot_angle_sign = 1;    /* sign of rotation angles      */
    }
    if (estimator_type % 100 >=  20)  hemisphere_type = 2;  /* South       */
//PR(ER, "order_rot=%d, rot_ang_s=%d\n", order_rotation, rot_angle_sign);
//PR(ER, "estimator_type=%d, case=%d\n", estimator_type, estimator_type%10);
    switch (estimator_type%10) {
    default:
    case  1:
	    estimator_type = EST_SIG_CL;
	    break;
    case  2:
	    estimator_type = EST_SIG_DL;
	    break;
    case  3:
	    estimator_type = EST_CL_SIG_DIF;
	    break;
    case  4:
	    estimator_type = EST_DL_SIG_DIF;
	    break;
    case  5:
	    estimator_type = EST_PARITY;
	    break;
    case  6:   /* (MAX-MIN)_north / (MAX-MIN)_south */
	    estimator_type = EST_PIX_MINIMAX;
	    glesp_comments = 0;
	    break;
    case  7:   /* Sigma_north / Sigma_south */
	    estimator_type = EST_PIX_DISP;
	    glesp_comments = 0;
	    break;
    }

    if (lmax_stat > 2*nx+1)  lmax_stat = 2*nx+1;
    nxr = 2 * lmax_stat + 1;
    npr = 2 * nxr;

    /* calculate temporal array of alm-s */
    sprintf (comm, "cl2map -lmax %d -ao %s -map %s -o /dev/null -nocom",
	    lmax_stat, tmp_alm, mapfile);
    if ((flag=system(comm)) != 0) {
err_syst:
	    PR(ER, "%s: can not execute the command:\n", myname);
	    PR(ER, "%s\n", comm);
	    exit(1);
    }

    if (order_rotation) {
	    /* theta, phi */
	    na = nx;
	    nb = np;
    } else {
	    /* phi, theta */
	    na = np;
	    nb = nx;
    }

    /* filling in estimating map */
    for(u=0; u<na; u++) {
	for(v=0; v<nb; v++) {
	    double estim = 0.0;
	    int ij;

	    if (order_rotation) {
		    /* theta, phi */
		    i = u; j = v;
		    if (nb >= nph[i])  continue;
	    } else {
		    /* phi, theta */
		    i = v; j = u;
		    if (j >= nph[i])  continue;
	    }

	    ij = i * np + j;

if (glesp_comments) {
if (order_rotation) {
if (j==0) PR(ER, "i=%-6d\b\b\b\b\b\b\b\b",i);
} else { if (i==0)  PR(ER, "j=%-6d\b\b\b\b\b\b\b\b",j); }
}

	    /* number of pixel -> rotation angles */
	    pixel2angle(x, nx, nph, i, j, &theta, &phi);

	    /* rotation in opposit direction */
	    if (rot_angle_sign) {
		theta = -theta;
		phi = -phi;
	    }

	    /* rotation by polar coordinates */
	    if (order_rotation) { /* (theta, phi) */
		if (j == 0) {
		    /* rotate alm by theta */
		    sprintf (comm, "difalm -dt %g %s -o %s -nocom",
			    theta, tmp_alm, tmprt_alm);
		    if ((flag=system(comm)) != 0)
			    goto err_syst;
		}
		/* rotate alm by phi */
		sprintf (comm, "difalm -dp %g %s -o %s -nocom",
			phi, tmprt_alm, tmpr_alm);
		if ((flag=system(comm)) != 0)
			goto err_syst;
	    } else {              /* (phi, theta) */
		if (i == 0) {
		    /* rotate alm by phi */
		    sprintf (comm, "difalm -dp %g %s -o %s -nocom",
			phi, tmp_alm, tmprt_alm);
		    if ((flag=system(comm)) != 0)
			    goto err_syst;
		}
		/* rotate alm by theta */
		sprintf (comm, "difalm -dt %g %s -o %s -nocom",
			theta, tmprt_alm, tmpr_alm);
		if ((flag=system(comm)) != 0)
			goto err_syst;
	    }

	    /* synthesize a map */
	    sprintf (comm,
	      "cl2map -lmax %d -nx %d -np %d -fa %s -o %s -nocom",
		    lmax_stat, nxr, npr, tmpr_alm, tmpr_map);
	    if ((flag=system(comm)) != 0)
		    goto err_syst;

	    /* reflection of hemisphere */
	    if (estimator_type != EST_PIX_MINIMAX &&
		estimator_type != EST_PIX_DISP &&
		estimator_type != EST_CL_SIG_DIF &&
		estimator_type != EST_DL_SIG_DIF ) {
		if (hemisphere_type == 2) {  /* South to North  */
		    if (reflection_type )   /* change sign */
			sprintf (comm, "difmap -mirS- %s -o %s -nocom",
			   tmpr_map, tmpr_cl);
		    else
			sprintf (comm, "difmap -mirS %s -o %s -nocom",
			   tmpr_map, tmpr_cl);

		} else  {  /* North to South */
		    if (reflection_type )  /* change sign */
			sprintf (comm, "difmap -mirN- %s -o %s -nocom",
			   tmpr_map, tmpr_cl);
		    else
			sprintf (comm, "difmap -mirN %s -o %s -nocom",
			   tmpr_map, tmpr_cl);
		}
		if ((flag=system(comm)) != 0)  goto err_syst;

		/* making a map actual after reflection */
		sprintf (comm, "mv %s %s", tmpr_cl, tmpr_map);
		system(comm);

	    } else if ( estimator_type == EST_CL_SIG_DIF ||
	      estimator_type == EST_DL_SIG_DIF ) {
		/* we need 2 maps */
		/* Cl|Dl -sigma difference estimating */
		if (hemisphere_type == 2) {
		    /* create first map:  South to North  */
		    if (reflection_type )   /* change sign */
			sprintf (comm, "difmap -mirS- %s -o %s -nocom",
			   tmpr_map, tmpr_cl);
		    else
			sprintf (comm, "difmap -mirS %s -o %s -nocom",
			   tmpr_map, tmpr_cl);

		    if ((flag=system(comm)) != 0)  goto err_syst;

		    /* create second map:  North to South */
		    if (reflection_type )   /* change sign */
			sprintf (comm, "difmap -mirN- %s -o %s -nocom",
			   tmpr_map, tmprt_map);
		    else
			sprintf (comm, "difmap -mirN %s -o %s -nocom",
			   tmpr_map, tmprt_map);

		    if ((flag=system(comm)) != 0)  goto err_syst;

		} else  {  /* North to South */

		    /* create first map:  North to South */
		    if (reflection_type )  /* change sign */
			sprintf (comm, "difmap -mirN- %s -o %s -nocom",
			   tmpr_map, tmpr_cl);
		    else
			sprintf (comm, "difmap -mirN %s -o %s -nocom",
			   tmpr_map, tmpr_cl);

		    if ((flag=system(comm)) != 0)  goto err_syst;

		    /* create the second map:  South to North */
		    if (reflection_type )   /* change sign */
			sprintf (comm, "difmap -mirS- %s -o %s -nocom",
			   tmpr_map, tmprt_map);
		    else
			sprintf (comm, "difmap -mirS %s -o %s -nocom",
			   tmpr_map, tmprt_map);

		    if ((flag=system(comm)) != 0)  goto err_syst;
		}

		/* making a map actual after reflection */
		sprintf (comm, "mv %s %s", tmpr_cl, tmpr_map);
		system(comm);

	    }

	    switch (estimator_type) {
	    default:
	    case EST_SIG_CL:
	    case EST_SIG_DL:
		/* spectrum of the map */
		sprintf (comm,
		   "cl2map -lmax %d -map %s -ao %s -o /dev/null -nocom",
		     lmax_stat, tmpr_map, tmpr_alm);
		if ((flag=system(comm)) != 0)  goto err_syst;

		if(estimator_type == EST_SIG_DL)
		    sprintf (comm, "alm2dl -Dl -lmax %d %s -o %s -nocom",
			 lmax_stat, tmpr_alm, tmpr_cl);
		else
		    sprintf (comm, "alm2dl -Cl -lmax %d %s -o %s -nocom",
			 lmax_stat, tmpr_alm, tmpr_cl);
		if ((flag=system(comm)) != 0)  goto err_syst;

		/* read with spectrum */
		read_cl (tmpr_cl, &est_arr, &l);

		/* estimator with spectrum */
		{  double aver=0.0, sig=0.0;
		   for (l=1; l<=lmax_stat; l++) aver += est_arr[l];
		   aver /= lmax_stat;
		   for (l=1; l<=lmax_stat; l++)
			sig += (est_arr[l]-aver) * (est_arr[l]-aver);
		   estim = sqrt(sig/lmax_stat);
		}
		if (est_arr != (double *)0) {
		    free((void *)est_arr);
		    est_arr = (double *)0;
		}
		break;
	    case EST_CL_SIG_DIF : /* Cl-sigma difference estimating */
	    case EST_DL_SIG_DIF : /* Dl-sigma difference estimating */
		{
		    double estim1=0.0, estim2 = 0.0;

		    /* process 1st map */
		    /* spectrum of the map */
		    sprintf (comm,
		       "cl2map -lmax %d -map %s -ao %s -o /dev/null -nocom",
			 lmax_stat, tmpr_map, tmpr_alm);
		    if ((flag=system(comm)) != 0)  goto err_syst;

		    if(estimator_type == EST_DL_SIG_DIF)
			sprintf (comm, "alm2dl -Dl -lmax %d %s -o %s -nocom",
			     lmax_stat, tmpr_alm, tmpr_cl);
		    else
			sprintf (comm, "alm2dl -Cl -lmax %d %s -o %s -nocom",
			     lmax_stat, tmpr_alm, tmpr_cl);
		    if ((flag=system(comm)) != 0)  goto err_syst;

		    /* read with spectrum */
		    read_cl (tmpr_cl, &est_arr, &l);

		    /* estimator with spectrum */
		    {  double aver=0.0, sig=0.0;
		       for (l=1; l<=lmax_stat; l++) aver += est_arr[l];
		       aver /= lmax_stat;
		       for (l=1; l<=lmax_stat; l++)
			    sig += (est_arr[l]-aver) * (est_arr[l]-aver);
		       estim1 = sqrt(sig/lmax_stat);
		    }
		    if (est_arr != (double *)0) {
			free((void *)est_arr);
			est_arr = (double *)0;
		    }

		    /* process 2nd map */
		    /* spectrum of the map */
		    sprintf (comm,
		       "cl2map -lmax %d -map %s -ao %s -o /dev/null -nocom",
			 lmax_stat, tmprt_map, tmpr_alm);
		    if ((flag=system(comm)) != 0)  goto err_syst;

		    if(estimator_type == EST_DL_SIG_DIF)
			sprintf (comm, "alm2dl -Dl -lmax %d %s -o %s -nocom",
			     lmax_stat, tmpr_alm, tmpr_cl);
		    else
			sprintf (comm, "alm2dl -Cl -lmax %d %s -o %s -nocom",
			     lmax_stat, tmpr_alm, tmpr_cl);

		    if ((flag=system(comm)) != 0)  goto err_syst;

		    /* read with spectrum */
		    read_cl (tmpr_cl, &est_arr, &l);

		    /* estimator with spectrum */
		    {  double aver=0.0, sig=0.0;
		       for (l=1; l<=lmax_stat; l++) aver += est_arr[l];
		       aver /= lmax_stat;
		       for (l=1; l<=lmax_stat; l++)
			    sig += (est_arr[l]-aver) * (est_arr[l]-aver);
		       estim2 = sqrt(sig/lmax_stat);
		    }
		    if (est_arr != (double *)0) {
			free((void *)est_arr);
			est_arr = (double *)0;
		    }

		    estim = estim1 - estim2;
		}
		break;
	    case EST_PARITY :     /* parity estimating */
		 /* spectrum of the map */
		 sprintf (comm,
		    "cl2map -lmax %d -map %s -ao %s -o /dev/null -nocom",
		      lmax_stat, tmpr_map, tmpr_alm);
		 if ((flag=system(comm)) != 0)  goto err_syst;
		 sprintf (comm, "alm2dl -Cl -lmax %d %s -o %s -nocom",
		      lmax_stat, tmpr_alm, tmpr_cl);
		 if ((flag=system(comm)) != 0)  goto err_syst;

		 /* read with spectrum */
		 read_cl (tmpr_cl, &est_arr, &l);

		 /* estimator with parity spectrum */
		 {  double sum_odd = 0.0, sum_even = 0.0;
		    for (l=2; l<=lmax_stat; l++) {
			 if (l%2)
			      sum_odd  += l*(l+1) * est_arr[l];
			 else
			      sum_even += l*(l+1) * est_arr[l];
		    }
		    estim = sum_odd/sum_even;
		 }
		 if (est_arr != (double *)0) {
		     free((void *)est_arr);
		     est_arr = (double *)0;
		 }
		 break;
	     case EST_PIX_MINIMAX:   /* (MAX-MIN)_north / (MAX-MIN)_south */
	     case EST_PIX_DISP:      /* Sigma_north / Sigma_south */
		 {
		    int ir, jr, nxr2=(nxr+1)/2;
		    double max_s = -1e30, min_s = +1e30,
			   max_n = -1e30, min_n = +1e30;
		    double aver_s=0.0, aver_n=0.0, sig_s=0.0, sig_n=0.0;
		    int n_n = 0;
		    double *r = (double *)0, *xq = (double *)0;
		    int nxq, npq, *nphq = (int *)0;

		    /* reading rotated map */
		    read_glesp_tmap(&r, &xq, &nxq, &npq, &nphq, tmpr_map);

		    /* calculation of statistics */
		    for (ir=0; ir<nxr2; ir++) {
			 int iir  = ir * npr;
			 int iir2 = (nxr - ir - 1) * npr;
			 for (jr=0; jr<nphq[ir]; jr++) {
			   if (estimator_type == EST_PIX_DISP) {
			     /* Sigma_north / Sigma_south */
			     aver_n += r[iir+jr];
			     aver_s += r[iir2+jr];
			     n_n++;
			   } else {
			     /* (MAX-MIN)_north / (MAX-MIN)_south */
			     if (max_n < r[iir+jr])  max_n = r[iir+jr];
			     if (min_n > r[iir+jr])  min_n = r[iir+jr];
			     if (max_s < r[iir2+jr]) max_s = r[iir2+jr];
			     if (min_s > r[iir2+jr]) min_s = r[iir2+jr];
			   }
			 }
		    }
		    if (estimator_type == EST_PIX_DISP) {
			 aver_n /= n_n;
			 aver_s /= n_n;
			 for (ir=0; ir<nxr2; ir++) {
			      int iir  = ir * npr;
			      int iir2 = (nxr - ir - 1) * npr;
			      for (jr=0; jr<nphq[ir]; jr++) {
				if (estimator_type == EST_PIX_DISP) {
				  /* Sigma_north / Sigma_south */
				  double ds= r[iir2+jr] - aver_s,
					 dn= r[iir +jr] - aver_n;
				  sig_n += dn*dn;
				  sig_s += ds*ds;
				}
			      }
			 }
			 estim = sig_n / sig_s;
		    } else
			 estim = (max_n - min_n) / (max_s - min_s);

		    if (r    != (double *)0)  free ((void *)r);
		    if (xq   != (double *)0)  free ((void *)xq);
		    if (nphq != (int *)0)     free ((void *)nphq);
		 }
		 break;
	    }

	    /* appropriate calculated value to a map */
	    m_est[ij] = estim;
	}
    }

    unlink(tmp_alm);
    unlink(tmpr_alm);
    unlink(tmprt_alm);
    unlink(tmpr_map);
    unlink(tmprt_map);
    unlink(tmpr_cl);

    if (est_arr != (double *)0) free((void *)est_arr);

    return;
}
