#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR            fprintf
#define ER            stderr
#define MAX(x,y)      ((x) > (y) ?  (x) : (y))
#define ABS(x)        ((x) < 0   ? -(x) : (x))

/* create GD-image and copy map to output file */
void t2rect(f, nx, nphi, x, nph, img, xs, ys, lon_0, mask)
double *f;
int nx, nphi;
double *x;
int    *nph;
double  **img;
int xs, ys;
double lon_0;   /* longitude to be added for rotation */
int **mask;
{
    register int i, j;
    unsigned int n = xs*ys;
    double theta, phi;
    double pi = M_PI, dt, dp;
    double x_c, y_c, x_r, y_r;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
nomem:
	PR(ER, "t2rec: no memory for temporal array...\n");
	exit(1);
    }
    if ( (*mask = (int *)calloc(n, sizeof **mask)) == (int *)0 )
	goto nomem;

    /* center of image */
    y_c = ys / 2;
    x_c = xs / 2;
    x_r = xs / 2;
    y_r = ys / 2;
    dt = pi   / ys;
    dp = 2*pi / xs;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs;
	theta = pi - i*dt;

	for(j=0; j<xs; j++) {
	    (*img)[ii+j] = 0.0;
	    (*mask)[ii+j] = 1;
	    phi = -lon_0 + (j-x_c) * dp;
	    if ( phi < 0.0 )    phi += 2. * pi;

	    if (ABS(theta) < 1.e5) {
		int n = ang2n(theta, phi, x, nx, nphi, nph);
//                if ( (*img)[ii+j] > 1.e-29)
		     (*img)[ii+j] = f[n];
	    }
	}
    }

    return;
}


void ht2rect(f, nx, nphi, x, offs, nph, img, xs, ys, lon_0, mask)
double *f;
int nx, nphi;
double *x, *offs;
int    *nph;
double  **img;
int xs, ys;
double lon_0;   /* longitude to be added for rotation */
int **mask;
{
    register int i, j;
    unsigned int n = xs*ys;
    double theta, phi;
    double pi = M_PI, dt, dp;
    double x_c, y_c, x_r, y_r;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
nomem:
	PR(ER, "t2mol: no memory for temporal array...\n");
	exit(1);
    }
    if ( (*mask = (int *)calloc(n, sizeof **mask)) == (int *)0 )
	goto nomem;

    /* center of image */
    y_c = ys / 2;
    x_c = xs / 2;
    x_r = xs / 2;
    y_r = ys / 2;
    dt = pi / ys;
    dp = 2*pi / xs;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs;
	theta = pi - i*dt;

	for(j=0; j<xs; j++) {
	    (*img)[ii+j] = 0.0;
	    (*mask)[ii+j] = 1;
	    phi = -lon_0 + (j-x_c) * dp;
	    if ( phi < 0.0 )    phi += 2. * pi;

	    /* keep only meaningful points */
	    if (ABS(theta) < 1.e5) {
		int n = hang2n(theta, phi, x, nx, nphi, nph, offs);
//                if ( (*img)[ii+j] > 1.e-29)
		     (*img)[ii+j] = f[n];
	    }
	}
    }

    return;
}


/* create GD-image, output overlaid map */
void overlay2rect(overlay_col, overlay_wid, lon_0, f,x,nx,np,nph, flag)
int overlay_col, overlay_wid;
double lon_0;
double *f, *x;
int nx, np, *nph;
int flag;           /* mask for v=1, or for v=0 */
{
    extern int get_nx_gd(), get_ny_gd();
    extern void color_gd(), move_gd();
    extern void marker_gd();
    register int i, j, k;
    struct _cr_  {double theta, phi; int v; int x, y;} *cr;
    int nx_gd = get_nx_gd(), ny_gd = get_ny_gd();
    int nn = nx_gd*ny_gd;
    int y_c, x_c, x_r, y_r;
    double theta=0., phi=0.;
    double pi = M_PI, dt, dp;
    int overlay_color = overlay_col;
    int overlay_width = overlay_wid;
    int i_x, j_phi;

    cr = (struct _cr_ *)calloc((unsigned)nn, sizeof *cr);
    if (cr == (struct _cr_ *)0) {
	PR(ER, "overlay2mol: no memory for temporal array...\n");
	exit(1);
    }

    y_c = ny_gd / 2;
    x_c = nx_gd / 2;
    x_r = nx_gd / 2;
    y_r = ny_gd / 2;
    dt = pi   / ny_gd;
    dp = 2*pi / nx_gd;

    /* put copy of base map to memory */
    for(k=i=0; i<ny_gd; i++) {
	theta = pi - i*dt;

	for(j=0; j<nx_gd; j++) {
	    phi = -lon_0 + (j-x_c) * dp;

	    /* phi in [0, 2pi] */
	    if ( phi < 0.0 )      phi += 2. * pi;
	    if ( phi > 2.*pi )    phi -= 2. * pi;
	    angle2pixel(x, nx, nph, &i_x, &j_phi, theta, phi);

	    cr[k].theta = theta;
	    cr[k].phi   = phi;
	    cr[k].y     = i;
	    cr[k].x     = j;

	    if (flag) {
		if (f[i_x * np + j_phi] > 0.5)  cr[k].v = 1;
	    } else {
		if (f[i_x * np + j_phi] < 0.5)  cr[k].v = 1;
	    }
	    if (cr[k].v) {
		/* put marker */
		int marker=0;
		color_gd(overlay_color);
		move_gd(cr[k].x, cr[k].y);
		marker_gd(overlay_width, marker); /* point */
	    }
	    k++;
	}
    }

    free((void *)cr);

    return;
}


#define ABS(x)  ((x) < 0 ? -(x) : (x))
/* create GD-image, output grid on map and copy map to output file */
/* despite too many loops, it works fast                           */
void grid2rect(grid_type, grid_col, lon_0, d1_step, d2_step)
int grid_type;  /* 1-galactic, 2-equatorial, 3-ecliptic */
int grid_col;
double lon_0;
double d1_step, d2_step;
{
    extern int get_nx_gd(), get_ny_gd();
    register int i, j, k, pa, pb;
    struct _cr_  {double theta, phi; double a, b; int x, y;} *cr;
    int nx_gd = get_nx_gd(), ny_gd = get_ny_gd();
    int nn = nx_gd*ny_gd, na=0, nb=0;
    int y_c, x_c, x_r, y_r, x_m, y_m;
    double theta=0., phi=0., a, b, a0=0., b0=0., da=15., db=15.;
    double pi = M_PI, dt, dp;
    double pmin= 1.e+30, pmax= -1.e+30;
    int grid_color = grid_col;
    int grid_width = 1;
    extern void pol2gal(double, double, double *, double *);
    extern void gal2equat(double, double, double*, double*);
    extern void equat2eclip(double, double, double*, double*);
    extern void color_gd(), move_gd();
    extern void marker_gd();

    cr = (struct _cr_ *)calloc((unsigned)nn, sizeof *cr);
    if (cr == (struct _cr_ *)0) {
	PR(ER, "grid2mol: no memory for temporal array...\n");
	exit(1);
    }

    y_c = ny_gd / 2;
    x_c = nx_gd / 2;
    x_r = nx_gd / 2;
    y_r = ny_gd / 2;
    dt = pi   / ny_gd;
    dp = 2*pi / nx_gd;

    /* put copy of base map to memory */
    for(k=i=0; i<ny_gd; i++) {
	theta = pi - i*dt;

	for(j=0; j<nx_gd; j++) {
	    phi = -lon_0 + (j-x_c) * dp;
	    cr[k].a = cr[k].b = 1e+30;

	    /* phi in [0, 2pi] */
	    if ( phi < 0.0 )    phi += 2. * pi;
	    if ( phi > 2.*pi )  phi -= 2. * pi;

	    /* coordinates of a grid point */
	    if (grid_type == 0 || grid_type == 1) {
		/* convert polar coordinates to galactic coordinates */
		pol2gal(theta,phi, &a, &b);
		/* radians -> degrees */
		cr[k].a = a*180/M_PI;
		cr[k].b = b*180/M_PI;
	    } else if (grid_type == 2) {
		pol2gal(theta,phi, &a, &b);
		cr[k].a = a*180/M_PI;
		cr[k].b = b*180/M_PI;
		gal2equat(cr[k].a, cr[k].b, &a, &b);
		cr[k].a = a; /* right ascension */
		cr[k].b = b; /* declination     */
	    } else if (grid_type == 3) {
		pol2gal(theta,phi, &a, &b);
		cr[k].a = a*180/M_PI;
		cr[k].b = b*180/M_PI;
		gal2equat(cr[k].a, cr[k].b, &a, &b);
		cr[k].a = a; /* right ascension, hrs */
		cr[k].b = b; /* declination, degs    */
		equat2eclip(cr[k].a, cr[k].b, &a, &b);
//PR(ER, "%d, %d,    %g,%g  %g,%g\n",i,j, cr[k].a,cr[k].b,  a,b);
		cr[k].a = a; /* ecliptic latitude, degs  */
		cr[k].b = b; /* ecliptic longitude, degs */
	    } else {
		pol2gal(theta,phi, &a, &b);
		cr[k].a = a*180/M_PI;
		cr[k].b = b*180/M_PI;
		grid_type = 0; /* then galactic coordinates */
	    }

	    if (phi < 1.e+3){
		if (pmin > phi) pmin = phi;
		if (pmax < phi) pmax = phi;
	    }
	    cr[k].theta = theta;
	    cr[k].phi   = phi;
	    cr[k].y     = i;
	    cr[k].x     = j;
	    k++;
	}
    }

    if (grid_type == 0 || grid_type == 1) {
	/* galactic coordinates */
	a0 = -90;        /* gal. latitude  */
	b0 = 0;          /* gal. longitude */
	da = d1_step<0.?15.:d1_step; /* step by spher.coor. a (lat.),degs */
	db = d2_step<0.?15.:d2_step; /* step by spher.coor. b (lon.),degs */
	na = 180. / da;  /* number of parallels */
	nb = 360. / db;  /* number of meridians */
    } else if (grid_type == 2) {
	a0 = 0;          /* right ascension  */
	b0 = -90;        /* declination      */
	da = d1_step<0.? 2.:d1_step;/* step by spher.coor. a (R.A.),hours */
	db = d2_step<0.?15.:d2_step;/* step by spher.coor. b (Dec.), degs */
	na = 24.  / da;  /* number of parallels */
	nb = 180. / db;  /* number of meridians */
    } else if (grid_type == 3) {
	/* ecliptic coordinates */
	a0 = -90;        /* eclip. latitude  */
	b0 = 0;          /* eclip. longitude */
	da = d1_step<0.?15.:d1_step; /* step by spher.coor. a (lat.),degs */
	db = d2_step<0.?15.:d2_step; /* step by spher.coor. b (lon.),degs */
	na = 180. / da;  /* number of parallels */
	nb = 360. / db;  /* number of meridians */
    }

    /* loop over grid a-coordinate */
    /*******************************/
    for(pa=0; pa<=na; pa++) {
	a = a0 + pa * da;
	x_m = y_m = -1;

	/* find closest pixel on image */
	for(i=0; i<ny_gd; i++) {
	    int ii = i * nx_gd;        /* beginning of the row */
	    int ii1 = (i+1) * nx_gd;
	    int ic = i * nx_gd + x_c;  /* middle of the row    */

	    if (i+1 >= ny_gd)   /* last line */
		   break;

	    /* we are still out of picture ellipse */
	    if (cr[ic].theta > 1.e5)  continue;

	    for(j=0; j<nx_gd; j++) {
		int jj  = ii+j;
		int jj1 = ii+j+1;
		int ij  = ii1+j;

		if (cr[jj].phi > 1.e5 || cr[jj1].phi > 1.e5)
		    continue;

		/* check conditions */
		if (
		    ((cr[jj].a <= a && a < cr[ij].a) && cr[ij].a<1.e5) ||
		     (cr[jj].a >= a && a > cr[ij].a) ||
		     (cr[jj].a >= a && a > cr[jj1].a) ||
		     (cr[jj].a <= a && a < cr[jj1].a) ||
		     ABS(cr[jj].a-a) < 1.e-6
		   ) {
		    /* then set point */
		    y_m = i;
		    if (ABS(cr[jj].a-a) < ABS(cr[jj1].a-a))
			x_m = j;
		    else
			x_m = j+1;
		    /* put marker */
		    color_gd(grid_color);
		    move_gd(x_m, y_m);
		    marker_gd(grid_width, grid_color); /* point */
		}
	    }
	}
    }

    /* loop over grid b-coordinate */
    /*******************************/
    for(pb=0; pb<=nb; pb++) {
	b = b0 + pb * db;
	x_m = y_m = -1;

	/* find closest pixel on image */
	for(i=0; i<ny_gd; i++) {
	    int ii = i * nx_gd;        /* beginning of the row */
	    int ii1 = (i+1) * nx_gd;
	    int ic = i * nx_gd + x_c;  /* middle of the row    */

	    if (i+1 >= ny_gd)   /* last line */
		   break;

	    /* we are still out of picture ellipse */
	    if (cr[ic].theta > 1.e5)  continue;

	    for(j=0; j<nx_gd; j++) {
		int jj  = ii+j;
		int jj1 = ii+j+1;
		int ij  = ii1+j;

		if (cr[jj].phi > 1.e5 || cr[jj1].phi > 1.e5)
		    continue;

		/* check conditions */
		if (
		    ((cr[jj].b <= b && b < cr[ij].b) && cr[ij].b<1.e5) ||
		     (cr[jj].b >= b && b > cr[ij].b) ||
		     (cr[jj].b >= b && b > cr[jj1].b) ||
		     (cr[jj].b <= b && b < cr[jj1].b) ||
		     ABS(cr[jj].b-b) < 1.e-6
		   ) {
		    /* then set point */
		    y_m = i;
		    if (ABS(cr[jj].b-b) < ABS(cr[jj1].b-b))
			x_m = j;
		    else
			x_m = j+1;
			/* put marker */
			color_gd(grid_color);
			move_gd(x_m, y_m);
			marker_gd(grid_width, grid_color); /* point */
		}
	    }
	}
    }
    free((void *)cr);

    return;
}


/* create GD-image of rect. area in Galactic coordinates */
/* and copy map to output file                           */
void sq2rect(f, nx, nphi, x, nph, img, xs, ys, lat1,lon1,lat2,lon2, mask)
double *f;
int nx, nphi;
double *x;
int    *nph;
double  **img;
int xs, ys;
double lat1,lon1,lat2,lon2;   /* cut rectagular, radians */
int **mask;
{
    register int i, j;
    unsigned int n = xs*ys;
    double theta, phi;
    double pi = M_PI, dt, dp;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
nomem:
	PR(ER, "t2rec: no memory for temporal array...\n");
	exit(1);
    }
    if ( (*mask = (int *)calloc(n, sizeof **mask)) == (int *)0 )
	goto nomem;

    /* steps by axes */
    dt = (lat2-lat1) / ys;
    dp = (lon2-lon1) / xs;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs;
	theta = pi/2 - (lat2 - i*dt);

	for(j=0; j<xs; j++) {
	    (*img)[ii+j] = 0.0;
	    (*mask)[ii+j] = 1;
	    phi = lon1 + j * dp;
	    if ( phi < 0.0 )  phi += 2. * pi;

	    if (ABS(theta) < 1.e5) {
		int n = ang2n(theta, phi, x, nx, nphi, nph);
//                if ( (*img)[ii+j] > 1.e-29)
		     (*img)[ii+j] = f[n];
	    }
	}
    }

    return;
}


/* create GD-image of rect. area in equatorial coordinates */
/* and copy map to output file                             */
void eq2rect(f, nx, nphi, x, nph, img, xs, ys, ra1,dec1,ra2,dec2, mask)
double *f;
int nx, nphi;
double *x;
int    *nph;
double  **img;
int xs, ys;
double ra1,dec1,ra2,dec2;   /* cut rectagular, radians */
int **mask;
{
    register int i, j;
    unsigned int n = xs*ys;
    double theta, phi;
    double pi = M_PI, d_ra, d_dec;
    double glat,glon;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
nomem:
      PR(ER, "eq2rect: no memory for temporal array of (%dx%d elements)...\n",
	xs, ys);
      exit(1);
    }
    if ( (*mask = (int *)calloc(n, sizeof **mask)) == (int *)0 )
	goto nomem;

    /* steps by axes */
    d_ra  = (ra2-ra1) / xs;
    d_dec = (dec2-dec1) / ys;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs;
	double ra, dec;

	dec = dec1 + i * d_dec;

	for(j=0; j<xs; j++) {
	    (*img)[ii+j] = 0.0;
	    (*mask)[ii+j] = 1;

	    ra = ra1 + j * d_ra;

	    /* equatoral to galactic coordinates*/
	    equat2gal(ra*12/M_PI, dec*180/M_PI, &glat, &glon);

	    /* galactic to polar  coordinates */
	    theta = pi/2 - glat*M_PI/180;
	    phi = glon*M_PI/180;

	    if ( phi < 0.0 )  phi += 2. * pi;

	    if (ABS(theta) < 1.e5) {
		int n = ang2n(theta, phi, x, nx, nphi, nph);
//                if ( (*img)[ii+j] > 1.e-29)
		     (*img)[ii+j] = f[n];
	    }
	}
    }

    return;
}

/* galactic cut for hpix-file */
void sq2hrect(f, nx, nphi, x, offs, nph, img, xs, ys,
				 lat1,lon1,lat2,lon2, mask)
double *f;
int nx, nphi;
double *x, *offs;
int    *nph;
double  **img;
int xs, ys;
double lat1,lon1,lat2,lon2;   /* coordinates of area */
int **mask;
{
    register int i, j;
    unsigned int n = xs*ys;
    double theta, phi;
    double pi = M_PI, dt, dp;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
nomem:
	PR(ER, "t2mol: no memory for temporal array...\n");
	exit(1);
    }
    if ( (*mask = (int *)calloc(n, sizeof **mask)) == (int *)0 )
	goto nomem;


    dt = (lat2-lat1) / ys;
    dp = (lon2-lon1) / xs;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs;
	theta = pi/2 - (lat2 - i*dt);

	for(j=0; j<xs; j++) {
	    (*img)[ii+j] = 0.0;
	    (*mask)[ii+j] = 1;

	    phi = lon1 + j * dp;
	    if ( phi < 0.0 )    phi += 2. * pi;

	    /* keep only meaningful points */
	    if (ABS(theta) < 1.e5) {
		int n = hang2n(theta, phi, x, nx, nphi, nph, offs);
//                if ( (*img)[ii+j] > 1.e-29)
		     (*img)[ii+j] = f[n];
	    }
	}
    }

    return;
}

/* equatorial cut for hpix-file */
void eq2hrect(f, nx, nphi, x, offs, nph, img, xs, ys,
				 ra1,dec1,ra2,dec2, mask)
double *f;
int nx, nphi;
double *x, *offs;
int    *nph;
double  **img;
int xs, ys;
double ra1,dec1,ra2,dec2;   /* coordinates of area */
int **mask;
{
    register int i, j;
    unsigned int n = xs*ys;
    double theta, phi;
    double pi = M_PI, d_ra, d_dec;
    double glat,glon;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
nomem:
	PR(ER, "t2mol: no memory for temporal array...\n");
	exit(1);
    }
    if ( (*mask = (int *)calloc(n, sizeof **mask)) == (int *)0 )
	goto nomem;

    /* steps by axes */
    d_ra  = (ra2-ra1) / xs;
    d_dec = (dec2-dec1) / ys;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs;
	double ra, dec;

	dec = dec1 + i * d_dec;

	for(j=0; j<xs; j++) {
	    (*img)[ii+j] = 0.0;
	    (*mask)[ii+j] = 1;

	    ra = ra1 + j * d_ra;

	    /* equatoral to galactic coordinates*/
	    equat2gal(ra*12/M_PI, dec*180/M_PI, &glat, &glon);

	    /* galactic to polar  coordinates */
	    theta = pi/2 - glat*M_PI/180;
	    phi = glon*M_PI/180;

	    if ( phi < 0.0 )    phi += 2. * pi;

	    /* keep only meaningful points */
	    if (ABS(theta) < 1.e5) {
		int n = hang2n(theta, phi, x, nx, nphi, nph, offs);
//                if ( (*img)[ii+j] > 1.e-29)
		     (*img)[ii+j] = f[n];
	    }
	}
    }

    return;
}


/* create GD-image, output overlaid map */
void overlay2sq(overlay_col, overlay_wid, lat1,lon1,lat2,lon2,
		 f,x,nx,np,nph, flag)
int overlay_col, overlay_wid;
double lat1,lon1,lat2,lon2;   /* coordinates of area */
double *f, *x;
int nx, np, *nph;
int flag;           /* mask for v=1, or for v=0 */
{
    extern int get_nx_gd(), get_ny_gd();
    extern void color_gd(), move_gd();
    extern void marker_gd();
    register int i, j, k;
    struct _cr_  {double theta, phi; int v; int x, y;} *cr;
    int nx_gd = get_nx_gd(), ny_gd = get_ny_gd();
    int nn = nx_gd*ny_gd;
    int y_c, x_c, x_r, y_r;
    double theta=0., phi=0.;
    double pi = M_PI, dt, dp;
    int overlay_color = overlay_col;
    int overlay_width = overlay_wid;
    int i_x, j_phi;

    cr = (struct _cr_ *)calloc((unsigned)nn, sizeof *cr);
    if (cr == (struct _cr_ *)0) {
	PR(ER, "overlay2mol: no memory for temporal array...\n");
	exit(1);
    }

    y_c = ny_gd / 2;
    x_c = nx_gd / 2;
    x_r = nx_gd / 2;
    y_r = ny_gd / 2;

    dt = (lat2-lat1) / ny_gd;
    dp = (lon2-lon1) / nx_gd;

    /* put copy of base map to memory */
    for(k=i=0; i<ny_gd; i++) {
	theta = pi/2 - (lat2 - i*dt);

	for(j=0; j<nx_gd; j++) {
	    phi = lon1 + j * dp;

	    /* phi in [0, 2pi] */
	    if ( phi < 0.0 )    phi += 2. * pi;
	    angle2pixel(x, nx, nph, &i_x, &j_phi, theta, phi);

	    cr[k].theta = theta;
	    cr[k].phi   = phi;
	    cr[k].y     = i;
	    cr[k].x     = j;

	    if (flag) {
		if (f[i_x * np + j_phi] > 0.5)  cr[k].v = 1;
	    } else {
		if (f[i_x * np + j_phi] < 0.5)  cr[k].v = 1;
	    }
	    if (cr[k].v) {
		/* put marker */
		int marker=0;
		color_gd(overlay_color);
		move_gd(cr[k].x, cr[k].y);
		marker_gd(overlay_width, marker); /* point */
	    }
	    k++;
	}
    }

    free((void *)cr);

    return;
}


/* create GD-image, output grid on map and copy map to output file */
/* despite too many loops, it works fast                           */
void grid2srect(grid_type, grid_col, lat1,lon1,lat2,lon2, d1_step, d2_step)
int grid_type;  /* 1-galactic, 2-equatorial, 3-ecliptic */
int grid_col;
double lat1,lon1,lat2,lon2;   /* coordinates of area */
double d1_step, d2_step;
{
    extern int get_nx_gd(), get_ny_gd();
    register int i, j, k, pa, pb;
    struct _cr_  {double theta, phi; double a, b; int x, y;} *cr;
    int nx_gd = get_nx_gd(), ny_gd = get_ny_gd();
    int nn = nx_gd*ny_gd, na=0, nb=0;
    int y_c, x_c, x_r, y_r, x_m, y_m;
    double theta=0., phi=0., a, b, a0=0., b0=0., da=15., db=15.;
    double pi = M_PI, dt, dp;
    double pmin= 1.e+30, pmax= -1.e+30;
    int grid_color = grid_col;
    int grid_width = 1;
    extern void pol2gal(double, double, double *, double *);
    extern void gal2equat(double, double, double*, double*);
    extern void equat2eclip(double, double, double*, double*);
    extern void color_gd(), move_gd();
    extern void marker_gd();

    cr = (struct _cr_ *)calloc((unsigned)nn, sizeof *cr);
    if (cr == (struct _cr_ *)0) {
	PR(ER, "grid2mol: no memory for temporal array...\n");
	exit(1);
    }

    y_c = ny_gd / 2;
    x_c = nx_gd / 2;
    x_r = nx_gd / 2;
    y_r = ny_gd / 2;

    dt = (lat2-lat1) / ny_gd;
    dp = (lon2-lon1) / nx_gd;

    /* put copy of base map to memory */
    for(k=i=0; i<ny_gd; i++) {
	theta = pi/2 - (lat2 - i*dt);

	for(j=0; j<nx_gd; j++) {
	    phi = lon1 + j * dp;
	    cr[k].a = cr[k].b = 1e+30;

	    /* phi in [0, 2pi] */
	    if ( phi < 0.0 )    phi += 2. * pi;

	    /* coordinates of a grid point */
	    if (grid_type == 0 || grid_type == 1) {
		/* convert polar coordinates to galactic coordinates */
		pol2gal(theta,phi, &a, &b);
		/* radians -> degrees */
		cr[k].a = a*180/M_PI;
		cr[k].b = b*180/M_PI;
	    } else if (grid_type == 2) {
		pol2gal(theta,phi, &a, &b);
		cr[k].a = a*180/M_PI;
		cr[k].b = b*180/M_PI;
		gal2equat(cr[k].a, cr[k].b, &a, &b);
		cr[k].a = a; /* right ascension */
		cr[k].b = b; /* declination     */
	    } else if (grid_type == 3) {
		pol2gal(theta,phi, &a, &b);
		cr[k].a = a*180/M_PI;
		cr[k].b = b*180/M_PI;
		gal2equat(cr[k].a, cr[k].b, &a, &b);
		cr[k].a = a; /* right ascension, hrs */
		cr[k].b = b; /* declination, degs    */
		equat2eclip(cr[k].a, cr[k].b, &a, &b);
//PR(ER, "%d, %d,    %g,%g  %g,%g\n",i,j, cr[k].a,cr[k].b,  a,b);
		cr[k].a = a; /* ecliptic latitude, degs  */
		cr[k].b = b; /* ecliptic longitude, degs */
	    } else {
		pol2gal(theta,phi, &a, &b);
		cr[k].a = a*180/M_PI;
		cr[k].b = b*180/M_PI;
		grid_type = 0; /* then galactic coordinates */
	    }

	    if (phi < 1.e+3){
		if (pmin > phi) pmin = phi;
		if (pmax < phi) pmax = phi;
	    }
	    cr[k].theta = theta;
	    cr[k].phi   = phi;
	    cr[k].y     = i;
	    cr[k].x     = j;
	    k++;
	}
    }

    if (grid_type == 0 || grid_type == 1) {
	/* galactic coordinates */
	a0 = -90;        /* gal. latitude  */
	b0 = 0;          /* gal. longitude */
	da = d1_step<0.?15.:d1_step; /* step by spher.coor. a (lat.),degs */
	db = d2_step<0.?15.:d2_step; /* step by spher.coor. b (lon.),degs */
	na = 180. / da;  /* number of parallels */
	nb = 360. / db;  /* number of meridians */
    } else if (grid_type == 2) {
	a0 = 0;          /* right ascension  */
	b0 = -90;        /* declination      */
	da = d1_step<0.? 2.:d1_step;/* step by spher.coor. a (R.A.),hours */
	db = d2_step<0.?15.:d2_step;/* step by spher.coor. b (Dec.), degs */
	na = 24.  / da;  /* number of parallels */
	nb = 180. / db;  /* number of meridians */
    } else if (grid_type == 3) {
	/* ecliptic coordinates */
	a0 = -90;        /* eclip. latitude  */
	b0 = 0;          /* eclip. longitude */
	da = d1_step<0.?15.:d1_step; /* step by spher.coor. a (lat.),degs */
	db = d2_step<0.?15.:d2_step; /* step by spher.coor. b (lon.),degs */
	na = 180. / da;  /* number of parallels */
	nb = 360. / db;  /* number of meridians */
    }

    /* loop over grid a-coordinate */
    /*******************************/
    for(pa=0; pa<=na; pa++) {
	a = a0 + pa * da;
	x_m = y_m = -1;

	/* find closest pixel on image */
	for(i=0; i<ny_gd; i++) {
	    int ii = i * nx_gd;        /* beginning of the row */
	    int ii1 = (i+1) * nx_gd;
	    int ic = i * nx_gd + x_c;  /* middle of the row    */

	    if (i+1 >= ny_gd)   /* last line */
		   break;

	    /* we are still out of picture ellipse */
	    if (cr[ic].theta > 1.e5)  continue;

	    for(j=0; j<nx_gd; j++) {
		int jj  = ii+j;
		int jj1 = ii+j+1;
		int ij  = ii1+j;

		if (cr[jj].phi > 1.e5 || cr[jj1].phi > 1.e5)
		    continue;

		/* check conditions */
		if (
		    ((cr[jj].a <= a && a < cr[ij].a) && cr[ij].a<1.e5) ||
		     (cr[jj].a >= a && a > cr[ij].a) ||
		     (cr[jj].a >= a && a > cr[jj1].a) ||
		     (cr[jj].a <= a && a < cr[jj1].a) ||
		     ABS(cr[jj].a-a) < 1.e-6
		   ) {
		    /* then set point */
		    y_m = i;
		    if (ABS(cr[jj].a-a) < ABS(cr[jj1].a-a))
			x_m = j;
		    else
			x_m = j+1;
		    /* put marker */
		    color_gd(grid_color);
		    move_gd(x_m, y_m);
		    marker_gd(grid_width, grid_color); /* point */
		}
	    }
	}
    }

    /* loop over grid b-coordinate */
    /*******************************/
    for(pb=0; pb<=nb; pb++) {
	b = b0 + pb * db;
	x_m = y_m = -1;

	/* find closest pixel on image */
	for(i=0; i<ny_gd; i++) {
	    int ii = i * nx_gd;        /* beginning of the row */
	    int ii1 = (i+1) * nx_gd;
	    int ic = i * nx_gd + x_c;  /* middle of the row    */

	    if (i+1 >= ny_gd)   /* last line */
		   break;

	    /* we are still out of picture ellipse */
	    if (cr[ic].theta > 1.e5)  continue;

	    for(j=0; j<nx_gd; j++) {
		int jj  = ii+j;
		int jj1 = ii+j+1;
		int ij  = ii1+j;

		if (cr[jj].phi > 1.e5 || cr[jj1].phi > 1.e5)
		    continue;

		/* check conditions */
		if (
		    ((cr[jj].b <= b && b < cr[ij].b) && cr[ij].b<1.e5) ||
		     (cr[jj].b >= b && b > cr[ij].b) ||
		     (cr[jj].b >= b && b > cr[jj1].b) ||
		     (cr[jj].b <= b && b < cr[jj1].b) ||
		     ABS(cr[jj].b-b) < 1.e-6
		   ) {
		    /* then set point */
		    y_m = i;
		    if (ABS(cr[jj].b-b) < ABS(cr[jj1].b-b))
			x_m = j;
		    else
			x_m = j+1;
			/* put marker */
			color_gd(grid_color);
			move_gd(x_m, y_m);
			marker_gd(grid_width, grid_color); /* point */
		}
	    }
	}
    }
    free((void *)cr);

    return;
}


