#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr

/*
 *   calculation of function I(l,m,l',m') = Sum w_p * F(l,m,p) * F(l',m',p)
 *   VO, 28/03/2002
 *      (questions to Oleg Verkhodanov (vo@sao.ru))
 *
 *  compilation:
 *      cc -o test_sum test_sum.c libglesp.a -lm
 *
 *  Lmax (LMX) should be set manually in code below, depends on free memory
 *
 *  flags of program:
 *  [-lmax lmax]    - set lmax (should less or equal LMX)
 *  [-l l_fix]      - fix L in one matrix
 *  [-m m_fix]      - fix M in one matrix
 *  [-lp l_fix]     - fix L in another matrix
 *
 *
 */

#define LMX     69                  /* L maximum  */
#define NMX     (2*(LMX+1))         /* Number of rings */
#define NMPHI   (2*NMX)             /* number of pixels on equator */

int l_inp = -1,                     /* input l */
    m_inp = -1,                     /* input , */
    lp_inp = -1,
    mp_inp = -1;
int hp_grid = 0;
int lmax=100, nx, nphi;
double th0=0, th1=M_PI;
int lx_add = -10;

extern void setopt(int, char **), usage(void);
extern void lgndr_m_(int *, int *, double *, double *, double *, double *);

char *myname="test_sum";

int main(argc, argv)
int argc;
char **argv;
{
    register int i, l, m, lp, mp;
/*
    double pf[LMX+1][LMX+1][LMX+1][LMX+1];
*/
    double pf[LMX+1][LMX+1][LMX+1];  /* array for integral         */
    double f_lm[NMX][LMX+1][LMX+1];  /* Legendre polynomials       */
    double x[NMX], w[NMX];           /* ring positions and weights */
    int nph[NMX];                    /* pixels per ring            */
    double cf[NMPHI];
    int nx0 = -1, nx1 = -1;;

    /* make value zero */
    for(m=0; m<=LMX; m++) {
	for(l=0; l<=LMX; l++) {
	    for(i=0; i<NMX; i++)  f_lm[i][m][l] = 0.0;
/*
 *           for(lp=0; lp<=LMX; lp++)
 */
		for(mp=0; mp<=LMX; mp++)
/*
 *                  pf[l][m][lp][mp] = 0.0;
 */
		    pf[l][m][mp] = 0.0;
	}
    }

    /* loop over external parameters */
    setopt(argc, argv);

    if (lmax > LMX)  lmax = LMX;
    if (nx   > NMX)  nx = NMX;
PR(ER,"l_fixed = %d, m_fixed = %d, lp_fixed = %d, mp_fixed = %d, lmax = %d\n",
l_inp, m_inp, lp_inp, mp_inp, lmax);
PR(ER, "th0=%g, th1=%g\n", th0, th1);

    if ((lp_inp < 0 || mp_inp < 0) && (l_inp  < 0 || m_inp  < 0) &&
	(m_inp  < 0 || mp_inp < 0) && (l_inp  < 0 || lp_inp < 0) ) {
	      PR(ER,"%s: set pair of input multipoles...\n", myname);
	      usage();
    }

    if (hp_grid) {
	/****************/
	/* HEALPix grid */
	/****************/
	double fc1, fc2;
	int j, nt, ns2, nside= -1, np = NMPHI;
	PR(ER, "Healpix grid...\n");

	if (nside < 1)  {
		int l_n;
		/* ln 2 = M_LN2 = 0.69314718055994530942 */
		i = (int) (log((double)(np))/M_LN2  + 0.0);
		for(l_n=1,j=0; j<i; j++)  l_n *= 2;
		nside = l_n / 4;
	}

nside_again:
	if (nside*4 > NMX) {
		nside /= 2;
		goto nside_again;
	}
PR(ER, "nside=%d\n", nside);
	nx = nt = 4*nside - 1;           /* X dimension=nx       */
	ns2 = nside*2;                   /* half of nt           */

	/* fill in 'theta_hp'-array */
	fc1 = 1./3./nside/nside;
	fc2 = 2./3./nside;
	for (i=0; i<ns2; i++) {
		int i1 = i+1;

		/* size by phi for North pole */
		nph[i] = 4 * (i1 < nside ? i1 : nside);
		/* value of cos(theta) for North pole */
		if ( i1 < nside )
		      x[i] = 1. - fc1 * i1 * i1;
		else
		      x[i] = (ns2 - i1) * fc2;

		j = nt-i1; /* symmentrical index for South */
		/* size by phi for South pole */
		nph[j] = nph[i];
		/* value of cos(theta) for South pole */
		if ( j > nt - nside - 1)
		      x[j] = -x[i];
		else
		      x[j] = -x[i];
	}
	for (i=0; i<nx; i++)  w[i] = 1.0;

    } else {
	/**************/
	/* GLESP grid */
	/**************/

	/* higher lmax & nx */
	nx = 2 * (lmax+1);
	nphi = 2 * nx;

	/* coordinates and weights */
	/* if lx_add = -10, then pixels have quasiequal area    */
	/* if lx_add =   0, then nuber pixels per ring is equal */
	coord(1, lx_add, -1., 1., 2*M_PI, x, w, nx, nph, nphi);

	if (th0 > -100.) {
	    nx0 = ind_x(nx, x, cos(M_PI/2+th0));
//            nx1 = ind_x(nx, x, cos(M_PI/2-th0));
//            if (nx0 > nx1) {i=nx0; nx0 = nx1; nx1=i;}
	}
	if (th1 > -100.) {
	    nx1 = ind_x(nx, x, cos(M_PI/2-th1));
	}
PR(ER, "nx0=%d, nx1=%d\n", nx0, nx1);

    }

    /*************************************/
    /* Legendre polynomials calculations */
    /*************************************/
    for(m=0; m<=lmax; m++) {
	    int mv = m;
	    double cmm;
	    start_clm_(&lmax, &mv, &cmm, cf);   /* start of calc */

	    for(i=0; i<nx; i++) {
		    double xc = x[i];
		    double *glm = &f_lm[i][m][0]; /*  reference to memory */
		    /* calculation of Legendre polynomials */
		    lgndr_m_(&lmax, &mv, &cmm, cf, &xc, glm);
	    }
    }

    if (nx0 < 0)  nx0=0;
    if (nx1 > nx) nx1=nx;

    /*******************************/
    /* basic function: integration */
    /*******************************/
    for(i=0; i<nx; i++)  {
	int np = (lx_add == 0 ? nphi :
		 get_nphi_by_x(x, nx, i)); /* number of pixels per ring */
	double v = w[i] * np;              /* weight */

	if (i < nx0)  continue;
	if (i > nx1)  break;
	for(l=2; l<=lmax; l++) {
	    for(m=0; m<=l; m++) {
//                for(lp=2; lp<=lmax; lp++) {
   lp = l;
		    for(mp=0; mp<=lp; mp++) {
//                       pf[l][m][lp][mp]+=v*f_lm[i][m][l]*f_lm[i][mp][lp];
			 /* integration, v is the weight */
			 pf[l][m][mp]   += v*f_lm[i][m][l]*f_lm[i][mp][lp];
		    }
//                }
	    }
	}
    }

    /*********************/
    /* output intergrals */
    /*********************/
    if (lp_inp >= 0 && mp_inp >= 0) {
	for(l=2; l<=lmax; l++) {
	    for(m=0; m<=lmax; m++) {
//              printf(" %5d  %5d  %14.7f\n",l,m,pf[l][m][lp_inp][mp_inp]);
		printf(" %5d  %5d  %14.7e\n",l,m,pf[l][m][mp_inp]);
	    }
	}
    } else if (l_inp >= 0 && m_inp >= 0) {
	for(lp=2; lp<=lmax; lp++) {
	    for(mp=0; mp<=lmax; mp++) {
//              printf(" %5d  %5d  %14.7f\n",lp,mp,pf[l_inp][m_inp][lp][mp]);
		printf(" %5d  %5d  %14.7e\n",lp,mp,pf[l_inp][m_inp][mp]);
	    }
	}
    } else if (m_inp >= 0 && mp_inp >= 0) {
	for(l=2; l<=lmax; l++) {
	    for(lp=2; lp<=lmax; lp++) {
//              printf(" %5d  %5d  %14.7f\n",l,lp,pf[l][m_inp][mp_inp]);
		printf(" %5d  %5d  %14.7e\n",l,lp,pf[l][m_inp][mp_inp]);
	    }
	}
    } else if (l_inp >= 0 && lp_inp >= 0) {
	for(m=0; m<=lmax; m++) {
	    for(mp=0; mp<=lmax; mp++) {
//              printf(" %5d  %5d  %14.7f\n",m,mp,pf[l_inp][m][lp_inp][mp]);
		printf(" %5d  %5d  %14.7e\n",m,mp,pf[l_inp][m][mp]);
	    }
	}
    }

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    char s1[30], s2[30];
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'l':
			if (strncmp(argv[i]+1, "lma", 3) == 0) {
			     lmax  = atoi(argv[++i]);
			} else if (strncmp(argv[i]+1, "lp", 2) == 0) {
			     lp_inp = atoi(argv[++i]);
			} else {
			     l_inp = atoi(argv[++i]);
			}
			break;
		case 'm':
			if (strncmp(argv[i]+1, "mp", 2) == 0) {
			     mp_inp = atoi(argv[++i]);
			} else {
			     m_inp = atoi(argv[++i]);
			}
			break;
		case 'h':
			if (strncmp(argv[i]+1, "hp", 2) == 0) {
				hp_grid = 1;  /* HEALPix */
			} else {
				usage();
			}
			break;
		case 'p':
			if (strncmp(argv[i]+1, "pin", 3) == 0) {
				/* phi increment number :       */
				/* addition to Nphi numbers for */
				/* 'coord'-function             */
				int v = atoi(argv[++i]);
				if (argv[i][0] == 'o')   v = -10;
				else if (v < -1)
				     lx_add = -1000 + v;
				else lx_add = v;
			}
			break;
		case 't':
			get_2_str(argv[++i], s1, s2);
			if(*s1 !='\0') th0=deg2rad(s1);
			if(*s2 !='\0') th1=deg2rad(s2);
		      /*
			th0=deg2rad(s1);
		       */
			break;
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		usage();
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s [-lmax lmax] [-l l_fix] [-m m_fix] [-lp l_fix] [-mp m_fix]\n", myname);
PR(ER, "[-t th0,th1] [-hp(Healpix)grid] [-pin 0]\n");
exit(2);
}
