#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"
#include "glespol.h"

#define ABS(x)  ((x) < 0 ? -(x) : (x))
#define PR      fprintf
#define ER      stderr

int sym_grid=1;
extern void setopt(int, char **), usage(void);
int lmax = -1;
int nx= -1, np= -1;
double eps = 1.e-16;
extern int _lmin_;
extern  double _enorm_;
extern int _lmin_gl;
extern  double _enorm_gl;
extern int max_double_(void);

char *myname="tst_plm";

int main(argc, argv)
int argc;
char **argv;
{
    register int i, j, l, m;
    double *x = (double *)0, *w = (double *)0, *wqu = (double *)0;
    int *nph = (int *)0;
    double *e, *c, *glm, *xlbd, *pnrm;
    int Kx, Lmn, Lmx1;
    int Ksym = 1, nx1;
    _lmin_gl = _lmin_ = max_double_();
    _enorm_gl =_enorm_ = pow(10., (double)_lmin_);   /* 10^Lmin */
PR(ER, "_lmin_ = %d\n", _lmin_);

    /* loop over external parameters */
    setopt(argc, argv);

    if (lmax < 0) lmax = 2000;
    nx = 2 * lmax + 1;
    np = nx * 2;
    Lmx1 = lmax + 1;
    Lmn = 2;
    nx1 = nx-1;
    Kx = (nx+1)/2;

PR(ER, "lmax=%d, nx=%d, np=%d\n", lmax, nx, np);
    if ((x = (double *)calloc((unsigned)nx, sizeof *x)) == (double *)0) {
nomem:
	PR(ER, "%s: no memory for arrays...\n", myname);
	exit(1);
    }
    if ((w = (double *)calloc((unsigned)nx, sizeof *w)) == (double *)0)
		goto nomem;
    if ((wqu = (double *)calloc((unsigned)nx, sizeof *wqu)) == (double *)0)
		goto nomem;
    if ((nph=(int *)calloc((unsigned)nx, sizeof *nph))  == (int *)0)
		goto nomem;
    l = (2*nx > np ? 2 * nx : np) * 2 + 4;
    if ((glm =(double *)calloc(l, sizeof *glm)) == (double *)0) goto nomem;
    if ((e =(double *)calloc(l, sizeof *e)) == (double *)0) goto nomem;
    if ((c =(double *)calloc(l, sizeof *c)) == (double *)0) goto nomem;
    if ((xlbd=(double*)calloc(2*nx+3,sizeof *xlbd))==(double*)0) goto nomem;
    if ((pnrm=(double*)calloc(nx+1, sizeof *pnrm))==(double*)0) goto nomem;

    coord_pol(1, -10, -1., 1., M_PI*2, x, w, wqu, nx, nph, np);

//    for(m=0; m<= lmax; m++) {   /* for all m */
     for(m=1000; m<= 2000; m++) {   /* for all m */
	int L0=m, l1, l2;
	double cmm;
	int ax;
	int mv=m;

	if(L0 < Lmn) L0 = Lmn;

	start_clm2_(&lmax,&mv,&cmm,c);
	{
	    start_plm_(&lmax,&mv,e,&ax);
	    if(m == 1)  ax = -ax;
	    for(i=0; i<nx; i++)   pnrm[i] = 0.0;
	}

//        for(i=0; i<Kx; i++) {       /*  for all x(ix)  */
	for(i=0; i<1; i++) {       /*  for all x(ix)  */
//            double xc = x[i];
	    double xc = 0.5;

//            if (*Lmx < lmin_ylm_hpix(m, xc))
//                continue;

	    if(Ksym > 0) {
		j = nx1-i;
		if(j <= i)  j = -100;
	    }
	    lgndr2_m_(&lmax,&mv,&cmm,c,&xc,glm);

	    l1 = lmax+1; l2 = -1;
	    for(l=m; l<= lmax; l++) {
	      if (ABS(glm[l])<eps) {
//                    if(l < l1)  l1 = l;
//                    if(l > l2)  l2 = l;
//               printf("%4d %12.5e %4d %4d %17.10e %17.10e\n",
//                    m, xc, l1, l2, glm[l1], glm[l2]);
		 PR(ER, "M=%4d L=%4d %17.10e\n",
		     m, l, glm[l]);
	      }
	    }
//            if (l2 > 0)
	}
    }

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'e':
			if (strncmp(argv[i]+1, "ep", 2) == 0) {
			      /* accuracy for Legendre plynomials */
			      eps = atof(argv[++i]);
			} else {
			      usage();
			}
			break;
		case 'h':
			usage();
			break;
		case 'l':
			lmax = atoi(argv[++i]);
			break;
		case 'n':
			switch(argv[i][2]) {
			case 'p':
				np = atoi(argv[++i]);
				break;
			case 'x':
				nx = atoi(argv[++i]);
				break;
			default:
				usage();
				break;
			}
			break;
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		usage();
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s \n",
myname);
exit(2);
}
