#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define PR  fprintf
#define ER  stderr

extern void setopt(int, char **), usage(void);
extern void lb2xyz(double, double, double *, double *, double *);
extern void xyz2lb(double, double, double, double *, double *);

/* vector statistics by Copi et al., 2006: astro-ph/0605135 */
/*  w, d, delta, r                                          */

struct _q_ {
    double x, y,z;
    double l, b;
};
char file[100];

char *myname;

int main(argc, argv)
int argc;
char **argv;
{
    struct _q_  r1, r2, s1, s2, w1, w2;

    double d_1, d_2, delta, r, mod_w1, mod_w2;
    FILE *fp;
    char str[81];

    setopt(argc, argv);

    /* read galactic coordinates of 1st & 2nd vector from file */
    if ((fp = fopen(file, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'\n", myname, file);
	exit(1);
    }
    str[0]='\0'; fgets(str, 80, fp); sscanf(str, "%lf %lf", &r1.l, &r1.b);
    str[0]='\0'; fgets(str, 80, fp); sscanf(str, "%lf %lf", &r2.l, &r2.b);
    str[0]='\0'; fgets(str, 80, fp); sscanf(str, "%lf %lf", &s1.l, &s1.b);
    str[0]='\0'; fgets(str, 80, fp); sscanf(str, "%lf %lf", &s2.l, &s2.b);
    fclose(fp);

    lb2xyz(r1.l, r1.b, &r1.x, &r1.y, &r1.z);
    lb2xyz(r2.l, r2.b, &r2.x, &r2.y, &r2.z);
    lb2xyz(s1.l, s1.b, &s1.x, &s1.y, &s1.z);
    lb2xyz(s2.l, s2.b, &s2.x, &s2.y, &s2.z);


//PR(ER, "r1: (l=%5.1f, b=%5.1f), (%6.4f, %6.4f, %6.4f)\n", r1.l, r1.b, r1.x, r1.y, r1.z);
//PR(ER, "r2: (l=%5.1f, b=%5.1f), (%6.4f, %6.4f, %6.4f)\n", r2.l, r2.b, r2.x, r2.y, r2.z);
//PR(ER, "s1: (l=%5.1f, b=%5.1f), (%6.4f, %6.4f, %6.4f)\n", s1.l, s1.b, s1.x, s1.y, s1.z);
//PR(ER, "s2: (l=%5.1f, b=%5.1f), (%6.4f, %6.4f, %6.4f)\n", s2.l, s2.b, s2.x, s2.y, s2.z);
    /* vector multiplication */
    w1.x = r1.y * r2.z - r1.z * r2.y;
    w1.y = r1.z * r2.x - r1.x * r2.z;
    w1.z = r1.x * r2.y - r1.y * r2.z;

    w2.x = s1.y * s2.z - s1.z * s2.y;
    w2.y = s1.z * s2.x - s1.x * s2.z;
    w2.z = s1.x * s2.y - s1.y * s2.z;

    xyz2lb(w1.x, w1.y, w1.z, &w1.l, &w1.b);
    xyz2lb(w2.x, w2.y, w2.z, &w2.l, &w2.b);

    /* scalar multiplication */
    d_1 = s1.x * r1.x + s1.y * r1.y + s1.z * r1.z;
    d_2 = s2.x * r2.x + s2.y * r2.y + s2.z * r2.z;

    mod_w1 = sqrt(w1.x * w1.x + w1.y * w1.y + w1.z * w1.z);
    mod_w2 = sqrt(w2.x * w2.x + w2.y * w2.y + w2.z * w2.z);

    delta = (w1.x * w2.x + w1.y * w2.y + w1.z * w2.z) / mod_w1 / mod_w2;
    r = mod_w2 / mod_w1;


    PR(stdout, "w1: (%6.4f,%6.4f,%6.4f)  w2: (%6.4f,%6.4f,%6.4f)\n",
		      w1.x, w1.y, w1.z, w2.x, w2.y, w2.z);
    PR(stdout, "w1: (l=%5.1f, b=%5.1f),m=%6.4f;  w2: (l=%5.1f, b=%5.1f),m=%6.4f\n",
		      w1.l, w1.b, mod_w1, w2.l, w2.b, mod_w2);
    PR(stdout, "d1 = %6.4f, d2 = %6.4f, delta=%6.4f, r=%6.4f\n",
		      d_1, d_2, delta, r);

    exit(0);
}



void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'h':
			usage();
			break;
		case 'f':
			strcpy(file, argv[++i]);
			break;
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		strcpy(file, argv[i]);
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s file\n",     myname);
exit(2);
}


void lb2xyz(l, b, x, y, z)
double l, b;
double *x, *y, *z;
{
    double phi, theta;
    double c = 180 / M_PI;

    phi = l / c;
    theta = (90. - b) / c;

    *x = sin(theta) * cos(phi);
    *y = sin(theta) * sin(phi);
    *z = cos(theta);

    return;
}

void xyz2lb(x, y, z, l, b)
double x, y, z;
double *l, *b;
{
    double r = sqrt(x*x + y*y + z*z);
    double cos_theta, phi=0.0, theta;
    double c = 180 / M_PI;

    /*****************************************/
    /* convert (x,y,z) to angle (theta, phi) */
    /*****************************************/
    cos_theta = z / r;
    theta = acos ( cos_theta );

    if (fabs(x) > 1.e-30)   phi = atan2(y, x);
    if (phi < 0.)  phi += 2*M_PI;

    *l = c * phi;
    *b = c * (M_PI/2 - theta);

    return;
}
