#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define PR  fprintf
#define ER  stderr

extern void setopt(int, char **), usage(void);
extern void lb2xyz(double, double, double *, double *, double *);
extern void xyz2lb(double, double, double, double *, double *);

/* vector statistics by Copi et al., 2006: astro-ph/0605135 */
/*  among multipoles: A1, A2, A3, S(A),                     */

struct _q_ {
    double x, y,z;
    double l, b;
};
char file[100];

char *myname;

int main(argc, argv)
int argc;
char **argv;
{
    struct _q_  w_212, w_312, w_323, w_331;
    double a1, a2, a3, s;

    FILE *fp;
    char str[81];

    setopt(argc, argv);

    /* read galactic coordinates of 1st & 2nd vector from file */
    if ((fp = fopen(file, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'\n", myname, file);
	exit(1);
    }
    str[0]='\0'; fgets(str,80,fp); sscanf(str, "%lf %lf", &w_212.l,&w_212.b);
    str[0]='\0'; fgets(str,80,fp); sscanf(str, "%lf %lf", &w_312.l,&w_312.b);
    str[0]='\0'; fgets(str,80,fp); sscanf(str, "%lf %lf", &w_323.l,&w_323.b);
    str[0]='\0'; fgets(str,80,fp); sscanf(str, "%lf %lf", &w_331.l,&w_331.b);
    fclose(fp);

    lb2xyz(w_212.l, w_212.b, &w_212.x, &w_212.y, &w_212.z);
    lb2xyz(w_312.l, w_312.b, &w_312.x, &w_312.y, &w_312.z);
    lb2xyz(w_323.l, w_323.b, &w_323.x, &w_323.y, &w_323.z);
    lb2xyz(w_331.l, w_331.b, &w_331.x, &w_331.y, &w_331.z);


PR(ER, "input data:\n");
PR(ER, "w_212: (l=%5.1f, b=%5.1f), (%6.4f, %6.4f, %6.4f)\n", w_212.l, w_212.b, w_212.x, w_212.y, w_212.z);
PR(ER, "w_312: (l=%5.1f, b=%5.1f), (%6.4f, %6.4f, %6.4f)\n", w_312.l, w_312.b, w_312.x, w_312.y, w_312.z);
PR(ER, "w_323: (l=%5.1f, b=%5.1f), (%6.4f, %6.4f, %6.4f)\n", w_323.l, w_323.b, w_323.x, w_323.y, w_323.z);
PR(ER, "w_331: (l=%5.1f, b=%5.1f), (%6.4f, %6.4f, %6.4f)\n", w_331.l, w_331.b, w_331.x, w_331.y, w_331.z);

    /* scalar multiplication */
    a1 = w_212.x * w_312.x + w_212.y * w_312.y + w_212.z * w_312.z;
    a2 = w_212.x * w_323.x + w_212.y * w_323.y + w_212.z * w_323.z;
    a3 = w_212.x * w_331.x + w_212.y * w_331.y + w_212.z * w_331.z;
    s = (fabs(a1)+fabs(a2)+fabs(a3)) / 3;

    PR(stdout, "a1=%6.4f, a2=%6.4f, a3=%6.4f, s=%6.4f\n",  a1,a2,a3, s);

    exit(0);
}



void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'h':
			usage();
			break;
		case 'f':
			strcpy(file, argv[++i]);
			break;
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		strcpy(file, argv[i]);
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s file\n",     myname);
PR(ER, "------------------------------\n");
PR(ER, "file format (input in degs):\n");
PR(ER, "w_212.l  w_212.b\n");
PR(ER, "w_312.l  w_312.b\n");
PR(ER, "w_323.l  w_323.b\n");
PR(ER, "w_331.l  w_331.b\n");
exit(2);
}


void lb2xyz(l, b, x, y, z)
double l, b;
double *x, *y, *z;
{
    double phi, theta;
    double c = 180 / M_PI;

    phi = l / c;
    theta = (90. - b) / c;

    *x = sin(theta) * cos(phi);
    *y = sin(theta) * sin(phi);
    *z = cos(theta);

    return;
}

void xyz2lb(x, y, z, l, b)
double x, y, z;
double *l, *b;
{
    double r = sqrt(x*x + y*y + z*z);
    double cos_theta, phi=0.0, theta;
    double c = 180 / M_PI;

    /*****************************************/
    /* convert (x,y,z) to angle (theta, phi) */
    /*****************************************/
    cos_theta = z / r;
    theta = acos ( cos_theta );

    if (fabs(x) > 1.e-30)   phi = atan2(y, x);
    if (phi < 0.)  phi += 2*M_PI;

    *l = c * phi;
    *b = c * (M_PI/2 - theta);

    return;
}
