/*
 *  functions to write a_{lm}-s
 *      by Oleg Verkhodanov (oleg@tac.dk), 2003-2004
 */
# include <stdio.h>
# include <math.h>
#ifndef __DARWIN__      /* by Tristan Beau <beau@in2p3.fr>, 17/06/2005 */
# include <values.h>
#endif

# include <ff.h>
# include "glesp.h"

extern char *myname;

extern void put_ff_vector(double *, int, FILE *);

void ff_write_alm(alm, nbuf, lmax, lmax_print, fp_out)
double *alm;
int lmax, lmax_print;
int nbuf;
FILE *fp_out;
{
	register int l, m;
	int nlm = (lmax_print+1)*(lmax_print+2) / 2;
	int tfields = 3, nsize = 0;
	char ttype1[30], tform1[30], tunit1[30],
	     ttype2[30], tform2[30], tunit2[30],
	     ttype3[30], tform3[30], tunit3[30];
	int bitp1, bitp2, bitp3;

	strcpy(tform1, "1J");
	strcpy(ttype1,  "INDEX");
	strcpy(tunit1,  "UNITLESS");

	strcpy(tform2, "1E");
	strcpy(ttype2,  "REAL");
	strcpy(tunit2,  "UNITLESS");
	strcpy(tform3, "1E");
	strcpy(ttype3,  "IMAGINE");
	strcpy(tunit3,  "UNITLESS");

	/* TFORM 1 */
	bitp1=LONGPIX;
	if (strcmp(tform1, "1J") == 0) {
		nsize += 4;
	} else if (strcmp(tform1, "1I") == 0) {
		nsize += 2;
		bitp1=SHORTPIX;
	}
	/* TFORM 2 */
	bitp2=FLOATPIX;
	if (strcmp(tform2, "1E") == 0) {
		nsize += 4;
	} else if (strcmp(tform2, "1D") == 0) {
		nsize += 8;
		bitp2=DOUBLEPIX;
	}
	/* TFORM 3 */
	bitp3=FLOATPIX;
	if (strcmp(tform3, "1E") == 0) {
		nsize += 4;
	} else if (strcmp(tform3, "1D") == 0) {
		nsize += 8;
		bitp3=DOUBLEPIX;
	}
	/************/
	/* set head */
	/************/
	Ckey_set("SIMPLE",    "T");
	Skey_set("BITPIX",    8);
	Skey_set("NAXIS",     0);
	Ckey_set("EXTEND",    "T");
	Ckey_set("COMMENTS",  "a_{lm} coefficients");

	Hupdate(fp_out);
	Hfree();

	Ckey_set("XTENSION",  "BINTABLE");
	Skey_set("BITPIX",  8);
	Skey_set("NAXIS",   2);
	Lkey_set("NAXIS1",  nsize);
	Lkey_set("NAXIS2",  nlm);
	Lkey_set("PCOUNT",  0);
	Lkey_set("GCOUNT",  1);

	Lkey_set("TFIELDS", tfields);
	Ckey_set("EXTNAME", "A_lm DATA");

	Ckey_set("TFORM1",  tform1); /* 1st field: index        */
	Ckey_set("TTYPE1",  ttype1);
	Ckey_set("TUNIT1",  tunit1);
	Ckey_set("TFORM2",  tform2); /* 2nd field: real a_lm    */
	Ckey_set("TTYPE2",  ttype2);
	Ckey_set("TUNIT2",  tunit2);
	Ckey_set("TFORM3",  tform3); /* 3d field:  imagine a_lm */
	Ckey_set("TTYPE3",  ttype3);
	Ckey_set("TUNIT3",  tunit3);

	Ckey_set("OBJECT",  "A_LM");
	Lkey_set("LMAX",    lmax_print);
	Ckey_set("ORIGIN",  "TAC");

	Hupdate(fp_out);
	Hfree();

	for (l=0; l<=lmax_print; l++) {
		register int ll=l*l+l+1,
			     bb = l*nbuf;
		for (m=0; m<=l; m++) {
			float v = ll+m;
			bitpix = bitp1;
			put_pix(fp_out, v);
			bitpix = bitp2;
			v = alm [bb+m];
			put_pix(fp_out, v);
			bitpix = bitp3;
			v = alm [bb+m+lmax+1];
			put_pix(fp_out, v);
		}
	}

	return;
}

void ff_write_cl(cl, lmax, fp_out)
double *cl;
int lmax;
FILE *fp_out;
{
	return;
}

void alm2ffile(alm, nalm, lmax, l_calc, m_calc, fp_out, p_flag)
double *alm;
int nalm, lmax;
int l_calc, m_calc;
FILE *fp_out;
int p_flag;     /* if p_flag =0, then both parts participate in output    */
		/* if p_flga =1, then real one, if =2, then imagine one   */
{
	register int l, m, n;
	int l_start = 2;
	double *v = (double *)0;
	double *d;
	int naxis2=0;

	n = 2*lmax + 1;

	if((v=(double *)calloc((unsigned)n, sizeof *v)) == (double *)0) {
		PR(ER, "alm_ks_test: no memory for %d element array.\n", n);
		exit(1);
	}
	d = v;

	/*************************/
	/* create HEAD of F-file */
	/*************************/
	bitpix = FLOATPIX;
	if (l_calc > 0 || m_calc >= 0) {
		/* one-dimensional file */
		Ckey_set("SIMPLE",    "T");
		Skey_set("BITPIX",    bitpix);
		Skey_set("NAXIS",     1);
		if (l_calc > 0) {
		     naxis1 = 2*l_calc+1;
		     /* pf=1, keep only real parts. If pf=2, only imagine */
		     if (p_flag)  naxis1 -= (p_flag==1 ? l_calc : l_calc+1);
		     crval1 = 0.0;
		     sprintf(ctype1, "Multipole, M (L=%d)", l_calc);
		} else {
		     naxis1 = lmax -(m_calc<l_start ?l_start:m_calc)+1;
		     /* pf=1, keep only real parts. If pf=2, only imagine */
		     if (m_calc  && !p_flag)  naxis1 *= 2;
		     crval1 = m_calc < l_start ? l_start : m_calc;
		     sprintf(ctype1, "Multipole, L (M=%d)", m_calc);
		}
		Lkey_set("NAXIS1",    naxis1);
		Skey_set("BSCALE",    1.0);
		Skey_set("BZERO",     0.0);
		Skey_set("BUNIT",     "UNITLESS");

		crpix1 = 1.0;
		if (p_flag)
		   cdelt1 = 0.5;  /* != 1, because Real and Imagine parts */
		else
		   cdelt1 = 1.0;
		Ckey_set("CTYPE1",  ctype1);
		Dkey_set("CRVAL1",  crval1);
		Dkey_set("CRPIX1",  crpix1);
		Dkey_set("CDELT1",  cdelt1);
		Ckey_set("OBJECT",  "A_LM");
	} else {
		/* two-dimensional file */
		double crval2, crpix2, cdelt2;
		char ctype2[30];

		Ckey_set("SIMPLE",    "T");
		Skey_set("BITPIX",    bitpix);
		Skey_set("NAXIS",     2);

		naxis1 = lmax-l_start+1;        /* l from 2 upto lmax */
		naxis2 = 2*lmax-1;              /* m from 0 upto lmax */
		if (p_flag)  naxis2 -= (p_flag==1 ? lmax : lmax+1);

		Lkey_set("NAXIS1",    naxis1);  /* l from 2 upto lmax */
		Lkey_set("NAXIS2",    naxis2);  /* m from 0 upto lmax */
		Skey_set("BSCALE",    1.0);
		Skey_set("BZERO",     0.0);
		Skey_set("BUNIT",     "UNITLESS");

		crval1 = 2.;
		crpix1 = 1.0;
		cdelt1 = 1.0;

		if (p_flag)
		   cdelt2 = 0.5;  /* != 1, because Real and Imagine parts */
		else
		   cdelt2 = 1.0;

		crval2 = 0.;
		crpix2 = 1.0;

		strcpy(ctype1, "Multipole, L");
		strcpy(ctype2, "Multipole, M");

		Ckey_set("CTYPE1",  ctype1);
		Dkey_set("CRVAL1",  crval1);
		Dkey_set("CRPIX1",  crpix1);
		Dkey_set("CDELT1",  cdelt1);

		Ckey_set("CTYPE2",  ctype2);
		Dkey_set("CRVAL2",  crval2);
		Dkey_set("CRPIX2",  crpix2);
		Dkey_set("CDELT2",  cdelt2);

		Ckey_set("OBJECT",  "A_LM");
	}

	/* output head of F-file */
	Hupdate(fp_out);

	/*************************/
	/* write binary data     */
	/*************************/
	if (l_calc > 0) {
		int ll = l_calc*nalm;
		n=0;
		if( !p_flag || p_flag==1) {n++; *d++ = alm[ll];}
		for (m=1; m<=lmax; m++) {
		    if( !p_flag || p_flag==1)
			{n++; *d++ = alm[ll+m];       }
		    if( !p_flag || p_flag==2)
			{n++; *d++ = alm[ll+m+lmax+1];}
		}
		/* n = 2*l_calc+1; */
		put_ff_vector(v, naxis1, fp_out);
	} else if (m_calc >= 0) {
		for(n=0,l= m_calc<l_start?l_start:m_calc ; l<=lmax; l++) {
			int ll = l*nalm;
			if( !p_flag || p_flag==1)
			    {n++; *d++ = alm[ll+m_calc];       }
			if(m_calc && (!p_flag || p_flag==2))
			    {n++; *d++ = alm[ll+m_calc+lmax+1];}
		}
		/* n = 2 * (lmax - (m_calc<2 ? 2 : m_calc) + 1); */
		put_ff_vector(v, naxis1, fp_out);
	} else {
		for (l=l_start; l<=lmax; l++) {
		     int ll = l*nalm;
		     d=v;
		     n=0;
		     if( !p_flag || p_flag==1) {*d++ = alm[ll]; n++;}
		     for (m=1; m<=l; m++) {
			if( !p_flag || p_flag==1)
			    {n++; *d++ = alm[ll+m];       }
			if( !p_flag || p_flag==2)
			    {n++; *d++ = alm[ll+m+lmax+1];}
		     }
		     put_ff_vector(v, naxis2, fp_out);
		}
	}

	/* free memory */
	if (v != (double *)0)  free((double *)v);

	return;
}

void put_ff_vector(v, n, fp_out)
double *v;
int n;
FILE *fp_out;
{
	register int i;
	for(i=0; i<n; i++)  put_pix(fp_out, (float)v[i]);
	return;
}
